/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global Proxy, Hammer, ArrayBuffer */

var Client = Client || {};
var InDe = InDe || {};

// command and properties translation maps
Client.transCmdMap = {ue: "updateElement", rc: "removeChild", ib: "insertBefore", ls: "loadStyle",
  cv: "closeView", ov: "openView"};
Client.transPropMap = {v: "value", s: "style", d: "visible", n: "name", cn: "className", c: "class",
  rn: "rownum", t: "innerText", h: "innerHTML"};
Client.transStyleMap = {b: "background", bc: "backgroundColor", c: "color", fs: "fontStyle"};


/**
 * Create the application
 */
Client.createApp = function ()
{
  this.eleMap = {};
  this.mainFrame = new this.MainFrame();
  this.id = Math.random().toString().substr(2, 6);
  //
  // If I don't have a proxy I need to create one
  // (N.B.: in offline mode a proxy has been already initialized by callee)
  if (!this.proxy) {
    // Get the SID/CID parameters from query string, session storage or cookie
    var sid = Client.Utils.getQueryVariable("sid");
    sid = sid || sessionStorage.getItem("sid");
    sid = sid || Client.Utils.getCookie("sid");
    //
    var cid = Client.Utils.getQueryVariable("cid");
    cid = cid || sessionStorage.getItem("cid");
    cid = cid || Client.Utils.getCookie("cid");
    //
    // Move from cookie store to session store
    sessionStorage.setItem("sid", sid);
    Client.Utils.deleteCookie("sid");
    sessionStorage.setItem("cid", cid);
    Client.Utils.deleteCookie("cid");
    //
    // acid/appid are specific to ideapp
    var acid = Client.Utils.getQueryVariable("acid");
    var appid = Client.Utils.getQueryVariable("appid");
    //
    // if there is a sid
    if (sid) {
      // Remember session ID
      this.mainFrame.sid = sid;
      //
      // Init the socket (i.e. proxy)
      this.initSocket(sid, cid, acid, appid);
    }
    else {
      // No SID. If I'm not inside the IDE (where the app is called "app")
      if (window.location.href.indexOf("/app/") === -1) {
        // I need to check if the user used the client/index.html url
        // If the URL ends with "client/index.html", remove the last part and enter using the main "entry-point"
        var url = window.location.href;
        if (url.substr(-18) === "/client/index.html")
          url = url.substring(0, url.length - 18);
        else if (url.substr(-23) === "/client/indexLocal.html")
          url = url.substring(0, url.length - 23);
        //
        // If the URL has been "fixed"
        if (url !== window.location.href)
          window.location = url;
      }
    }
  }
  //
  // Now, if I have a proxy, I can start the app
  if (this.proxy)
    this.mainFrame.start();
};


/**
 * Initialize socket
 * @param {string} sid - session id
 * @param {string} cid - client id
 * @param {string} acid - ide client id
 * @param {string} appid - ide app id
 */
Client.initSocket = function (sid, cid, acid, appid)
{
  var pthis = this;
  var appname = Client.resourceName || window.location.pathname.split("/")[1];
  //
  // Create the socket (if it's a reconnect, don't use socket connection pool)
  var reconnect = (this.socket !== undefined);
  this.socket = io(Client.resourceOrigin, (reconnect ? {forceNew: true} : undefined));
  this.socket.reconnected = reconnect;    // Remember that this socket has been reconnected
  //
  // Define a simpler interface for send
  this.socket.send = function (sender, events) {
    this.emit("appmsg", {appid: appid, sid: sid, events: events});
  };
  //
  // Listen for socket messages
  this.socket.on("redirect", function (url) {
    sessionStorage.clear();
    Client.Utils.deleteCookie("sid");
    Client.Utils.deleteCookie("cid");
    window.location = url;
  });
  //
  this.socket.on("appmsg", function (m) {
    pthis.mainFrame.processRequest(m.content);
  });
  //
  this.socket.on("disconnect", function () {
    // If I'm reconnecting, skip this event
    if (this.reconnecting)
      return;
    //
    // If the app is terminating, skip
    if (Client.terminating)
      return;
    //
    // If I'm inside IDE, don't reconnect -> terminate
    if (acid || appid)
      return pthis.mainFrame.terminate();
    //
    // I'm offline -> tell it to the user
    Client.realizeOfflineScreen();
    //
    // I'm inside MASTER -> try to reconnect
    delete pthis.proxy;       // Proxy is no longer valid
    //
    // First "detach" the old socket
    pthis.socket.reconnecting = true;
    pthis.socket.close();
    delete pthis.socket.reconnecting;
    //
    // Then try with a new one
    pthis.initSocket(sid, cid, acid, appid);
  });
  //
  // If this is a reconnection, listen for connection success
  if (reconnect) {
    this.socket.on("connect", function () {
      // Reconnection succesfull -> reload the page to resync with server
      window.location.reload();
    });
  }
  //
  // This is my proxy
  this.proxy = this.socket;
  //
  // Send a message with the sid/acid (IDE case), and sid/cid (master case)
  // N.B.: appname is used in the master case for redirecting the right app when the session
  // is no longer available in the server)
  this.socket.emit("asid", {sid: sid, acid: acid, cid: cid, appname: appname});
};


/**
 * Sets the proxy
 * @param {Object} proxy
 */
Client.setProxy = function (proxy)
{
  this.proxy = proxy;
};


/**
 * Complete loading requirement:
 * some external object require a "callback" function to tell they are ready to operate
 */
Client.callback = function ()
{
  this.mainFrame.reqComplete("callback");
};


/**
 * @class
 */
Client.MainFrame = function ()
{
  this.views = [];
  this.requireMap = {};
  this.cmdCache = [];
  this.device = new Client.Device();
  this.head = new Client.HeaderList();
  this.attachEvents();
  this.addedObjects = [];
  this.appStarted = false;
  this.cmdSeq = 0;
  this.windows = {};
  //
  this.lastHeadObject = document.head.lastChild;
  //
  // Change it to false to disable hammer support
  this.hammerEnabled = true;
  //
  // Change to false to push updates as soon as possible
  Client.useAnimationFrame = true;
  //
  // Try catch is needed to avoid security error when the app is launched on shell
  try {
    Client.isTestAuto = window.top && window.top.location.pathname.indexOf("client/testautoPreview.html") > -1;
  } catch (ex) {
  }
};


/**
 * Use Request Animation Frame to execute requests
 */
Client.fireProcessCache = function ()
{
  delete Client.rafId;
  if (Client.mainFrame) {
    //
    Client.mainFrame.cmdSeq++;
    //
    Client.mainFrame.processCache();
    //
    // Engage RAF again
    if (Client.mainFrame.cmdCache.length > 0)
      Client.rafId = requestAnimationFrame(Client.fireProcessCache);
  }
};


/**
 * Delete overlapping commands
 * @param {Object} req - Server request
 */
Client.MainFrame.prototype.purgeCmdCache = function (req)
{
  if (!this.cmdCache.length)
    return;
  //
  // Waiting for an opening view...
  if (this.paused)
    return;
  //
  // For each new command
  for (var i = 0; i < req.length; i++) {
    // cmd is the new command
    var cmd = req[i];
    //
    // If the structure of the view is changing, don't purge
    if (cmd.id === "insertBefore" || cmd.id === "removeChild")
      return;
    //
    // if we are changing an object property
    if (cmd.id === "updateElement") {
      // Let's see if we had to change the same property
      for (var j = 0; j < this.cmdCache.length; j++) {
        // old is a command already in the cache
        var old = this.cmdCache[j];
        //
        // If the structure of the view is changing, don't purge
        if (old.id === "insertBefore" || old.id === "removeChild")
          return;
        //
        // if we were changing the same object
        if (old.obj === cmd.obj && old.id === "updateElement" && !old.cnt.clid) {
          // let's see the new properties
          var k = Object.keys(cmd.cnt);
          // for each new property, delete it from the previous command
          for (var n = 0; n < k.length; n++)
            delete old.cnt[k[n]];
          // if the previous command is empty
          k = Object.keys(old.cnt);
          // delete the command from the cache
          if (!k.length) {
            this.cmdCache.splice(j, 1);
            j--;
          }
        }
      }
    }
  }
};


/**
 * Execute a server request
 * @param {Object} req - Server request
 */
Client.MainFrame.prototype.processRequest = function (req)
{
  if (!Client.isOffline())
    this.translate(req);
  //
  // If the app is terminating, skip any other command
  // as we don't want to change UI while the app is closing
  var terminating = false;
  if (req.length) {
    var l = req[req.length - 1];
    if (!l.obj && l.id === "terminate") {
      terminating = true;
      this.cmdCache.push(l);
    }
  }
  //
  if (!terminating) {
    this.purgeCmdCache(req);
    for (var i = 0; i < req.length; i++) {
      this.cmdCache.push(req[i]);
    }
  }
  //
  // Execute request
  if (Client.useAnimationFrame && !this.isEditing() && Client.mainFrame.appStarted && !document.hidden) {
    if (!Client.rafId) {
      Client.rafId = requestAnimationFrame(Client.fireProcessCache);
    }
  }
  else {
    this.processCache();
  }
};


/**
 * Execute cached commands
 */
Client.MainFrame.prototype.processCache = function ()
{
  // Not now
  if (this.paused)
    return;
  //
  this.sendChg = true;
  var x = new Date();
  var exit = true;
  var report = true;
  //
  while (this.cmdCache.length > 0) {
    var cmd = this.cmdCache[0];
    //
    if (Client.logcmd) {
      console.warn(this.cmdSeq, cmd.obj, cmd.id, cmd.cnt);
    }
    //
    // Extract the command from cache so if an error occurs, we don't get stuck on it
    this.cmdCache.splice(0, 1);
    //
    // Do not stop on error, but try to go on
    try {
      this.processCommand(cmd);
    }
    catch (ex) {
      // Report only the first error in this batch
      if (report) {
        report = false;
        // Tell the server that we had a problem for this change
        var s = ex.message || ex;
        if (typeof s === "object") {
          try {
            s = JSON.stringify(s);
          }
          catch (ex2) {
          }
        }
        s = "FE: " + s;
        //
        var cnt = {id: cmd.id, error: {message: s, stack: ex.stack}};
        var e = [{id: "onUpdateElementError", content: cnt}];
        Client.mainFrame.sendEvents(e);
      }
    }
    //
    if (this.paused) {
      // Insert cmd again because it requires a resource that has not loaded yet
      this.cmdCache.splice(0, 0, cmd);
      break;
    }
    //
    // If we are changing the dom, don't exit as we need to complete before
    // showing it to the user
    if (exit && (cmd.id === "insertBefore" || cmd.id === "removeChild"))
      exit = false;
    //
    // No more than 10 msec
    if (exit && Client.useAnimationFrame && !this.isEditing() && new Date() - x > 10) {
      break;
    }
  }
  //
  this.sendChg = false;
};


/**
 * Process a single command
 * @param {object} cmd - Command to process
 */
Client.MainFrame.prototype.processCommand = function (cmd)
{
  if (!cmd.obj) { // App command
    this[cmd.id](cmd.cnt);
  }
  else if (cmd.obj.substring(0, 7) === "device-") { // device command
    this.device.processRequest(cmd);
  }
  else if ((cmd.obj === "preview-exception" || cmd.obj === "preview-warning") && parent && parent.postMessage) {
    parent.postMessage({type: cmd.obj, txt: cmd.txt}, "*");
  }
  else {
    //
    // Object command
    var ele = Client.eleMap[cmd.obj];
    if (ele && ele[cmd.id]) {
      // Remote custom method or standard one
      if (cmd.cnt && cmd.cnt.params)
        ele[cmd.id].apply(ele, cmd.cnt.params);
      else if (cmd.content)
        ele[cmd.id](cmd.content);
      else
        ele[cmd.id](cmd.cnt);
    }
  }
};


/**
 * Decompress commands
 * @param {Array} cmds
 */
Client.MainFrame.prototype.translate = function (cmds)
{
  // Decompress each command ID in the list
  for (var i = 0; i < cmds.length; i++) {
    var c = cmds[i].id;
    var t = Client.transCmdMap[c];
    if (t)
      cmds[i].id = t;
    //
    // The updateElement command is special because we can compress
    // object properties and style properties
    if (t === "updateElement")
      this.translateProperties(cmds[i].cnt, Client.transPropMap);
  }
};


/**
 * Decompress object properties
 * @param {object} obj
 * @param {object} transMap - translation map
 */
Client.MainFrame.prototype.translateProperties = function (obj, transMap)
{
  var k = Object.keys(obj);
  for (var i = 0; i < k.length; i++) {
    var p = k[i];
    //
    // Let's see if the P property is compressed
    var t = transMap[p];
    if (t) {
      // Convert it
      obj[t] = obj[p];
      delete obj[p];
      //
      // If the style property is concerned, we need to decompress its content
      if (t === "style")
        this.translateProperties(obj[t], Client.transStyleMap);
    }
  }
};


/**
 * Execute activateView command
 * @param {Object} viewid - object containing the view id
 */
Client.MainFrame.prototype.activateView = function (viewid)
{
  var view = Client.eleMap[viewid.id];
  if (view)
    view.activate();
};


/**
 * Send an event collection to the server
 * @param {Array} events
 */
Client.MainFrame.prototype.sendEvents = function (events)
{
  // Old shells cannot route the first getCookie message right.
  // Getting things done.
  if (this.waitingForCookies && events && events.length) {
    if (events[0].source === "shell" && events[0].id === "getCookiesCB" && events[0].content) {
      this.startWithCookies(events[0].content.result);
      return;
    }
  }
  //
  if (Client.logcmd) {
    events.forEach(function (ev) {
      console.warn("<--", ev.id, ev.content);
    });
  }
  //
  // if there are events, send them
  if (Client.proxy)
    Client.proxy.send(this, events);
  else
    console.log("no proxy set");
};


/**
 * Execute openView command
 * @param {Object} view - view representation
 */
Client.MainFrame.prototype.openView = function (view)
{
  // let's see if there are some requirements for the view
  if (!this.loadRequirements(view.req))
    return;
  //
  var v = new Client.View(view);
  this.views.push(v);
  Client.eleMap[v.id] = v;
  //
  // send a message to the IDE if the view is open in the view editor
  if (this.isEditing())
    Client.eleMap.editm.editProxy.onOpenView();
  //
  // Resend app start if required
  this.sendAppStart();
};


/**
 * Execute closeView command
 * @param {Object} viewid - object containing the view id
 */
Client.MainFrame.prototype.closeView = function (viewid)
{
  var view = Client.eleMap[viewid.id];
  if (view)
    view.close();
};


/**
 * Execute alert command
 * @param {String} txt
 */
Client.MainFrame.prototype.alert = function (txt)
{
  alert(txt);
};


/**
 * Open a standard popup
 * @param {object} cnt
 */
Client.MainFrame.prototype.popup = function (cnt)
{
  // The standard implementation does nothing and returns undefined
  if (cnt.cbId) {
    var e = [{id: "popupBoxReturn", content: {res: undefined, cbId: cnt.cbId}}];
    this.sendEvents(e);
  }
};


/**
 * Execute alert command
 * @param {ojg} cnt
 */
Client.MainFrame.prototype.setCookie = function (cnt)
{
  var d = new Date();
  d.setTime(d.getTime() + (cnt.options.exdays * 24 * 60 * 60 * 1000));
  var expires = "; expires=" + d.toUTCString();
  //
  var path = "; path=" + (cnt.options.path || "/");
  var secure = (cnt.options.secure ? "; secure=true" : "");
  //
  // set the cookie
  if (Client.isOffline() || this.device.operatingSystem === "ios") {
    // send request to device
    // lscookies doesn't answer with a callback
    this.device.shell.postMessage({obj: "device-lscookies", id: "setCookie", cnt: {name: cnt.name, value: cnt.value, exdays: cnt.options.exdays}}, "*");
  }
  else
    document.cookie = cnt.name + "=" + cnt.value + expires + path + secure;
};


/**
 * Execute open command using window.open on browser, inappbrowser on device
 * @param {Object} cnt ( = {href, target, options, cbId} )
 */
Client.MainFrame.prototype.open = function (cnt)
{
  // Inplace app needs absolutization
  if (cnt.href)
    cnt.href = Client.Utils.abs(cnt.href);
  //
  var onmob = this.device.isMobile && this.device.shell && !(this.device.shell instanceof Client.ShellEmulator);
  //
  // Optimize target for different OS in some special cases
  if (!cnt.target) {
    var ext = ["tel:", "mailto:", "sms:", "geo:", "itms:", "itms-apps:", "market:", "http://maps.apple.com"];
    for (var i = 0; i < ext.length; i++) {
      if (cnt.href.startsWith(ext[i])) {
        cnt.target = onmob ? "_system" : "_self";
        //
        // Map is a special case on Android
        if (this.device.operatingSystem === "android" && cnt.href.indexOf("maps") > 0) {
          var qidx = cnt.href.indexOf("?q=");
          if (qidx > 0) {
            cnt.href = "geo:0,0" + cnt.href.substring(qidx);
          }
        }
        if ((this.device.operatingSystem === "ios" || !onmob) && cnt.href.startsWith("geo")) {
          var qidx = cnt.href.indexOf("?q=");
          if (qidx > 0) {
            cnt.href = "http://maps.apple.com/" + cnt.href.substring(qidx);
          }
        }
        break;
      }
    }
  }
  //
  var w;
  //
  if (onmob) {
    this.device.shell.postMessage({obj: "device-inappbrowser", id: "open", cnt: cnt}, "*");
    w = true;
  }
  else {
    if (cnt.options === "close") {
      var ris = this.windows[cnt.target];
      if (ris) {
        w = true;
        ris.close();
      }
    }
    else {
      var ris = window.open(cnt.href, cnt.target, cnt.options);
      if (ris) {
        w = true;
        if (cnt.target && !cnt.target.startsWith("_")) {
          this.windows[cnt.target] = ris;
          ris.onclose = function () {
            this.windows[cnt.target] = undefined;
          }.bind(this);
        }
      }
    }
  }
  if (cnt.cbId) {
    var e = [{id: "popupBoxReturn", content: {res: w, cbId: cnt.cbId}}];
    this.sendEvents(e);
  }
};


/**
 * Display a confirmation box
 * @param {String} opt - contains the text to display in the confirm box and the callback id
 */
Client.MainFrame.prototype.confirm = function (opt)
{
  var r = confirm(opt.txt);
  var e = [{id: "popupBoxReturn", content: {res: r, cbId: opt.cbId}}];
  this.sendEvents(e);
};


/**
 * Display a prompt box
 * @param {Object} opt - contains the text to display in the prompt box and the default input text
 */
Client.MainFrame.prototype.prompt = function (opt)
{
  var r = prompt(opt.txt, opt.def);
  var e = [{id: "popupBoxReturn", content: {res: r, cbId: opt.cbId}}];
  this.sendEvents(e);
};


/**
 * Handle session termination
 * @param {Object} options
 */
Client.MainFrame.prototype.terminate = function (options)
{
  Client.terminating = true;
  if (Proxy)
    Proxy.stopApplication();
  if (this.device.shell && !(this.device.shell instanceof Client.ShellEmulator))
    this.device.shell.postMessage("appStopped", "*");
  else {
    window.top.close();
    window.location = "http://www.instantdeveloper.com";
  }
};


/**
 * Remove any added objects
 */
Client.MainFrame.prototype.removeAddedObjects = function ()
{
  for (var i = 0; i < this.addedObjects.length; i++) {
    var obj = this.addedObjects[i];
    try {
      obj.parentNode.removeChild(obj);
    }
    catch (ex) {
      console.log(ex);
    }
  }
  this.addedObjects = [];
};


/**
 * Start the application
 */
Client.MainFrame.prototype.start = function ()
{
  // Skip start in case of testauto. In fact will be the testauto itself that will send onStart event to server
  if (Client.skipStart)
    return;
  //
  if (Client.isOffline() || this.device.operatingSystem === "ios") {
    // get local storage cookies
    this.waitingForCookies = true;
    this.device.shell.postMessage({obj: "device-lscookies", id: "getCookies", cnt: {mode: "startWithCookies"}}, "*");
  }
  else {
    // normal cookie
    var e = [{id: "onStart", content: this.device.getProp()}, {id: "defineStyle"}];
    this.sendEvents(e);
  }
};


/**
 * Start the application
 * @param {object} cookies
 */
Client.MainFrame.prototype.startWithCookies = function (cookies)
{
  // Skip start in case of testauto. In fact will be the testauto itself that will send onStart event to server
  if (Client.skipStart)
    return;
  //
  // local storage cookie
  this.waitingForCookies = false;
  var e = [{id: "onStart", content: this.device.getProp(), cookies: cookies}, {id: "defineStyle"}];
  this.sendEvents(e);
};


/**
 * Send the terminate signal to the server
 */
Client.MainFrame.prototype.onTerminate = function ()
{
  var ok = true;
  try {
    var ae = document.activeElement;
    var we = window.event;
    //
    // non http link won't cause session termination
    if (ae && ae.tagName === "A" && ae.href.indexOf("http") === -1 && we && new Date() - we.timeStamp < 100)
      ok = false;
  }
  catch (ex) {

  }
  //
  if (ok) {
    var e = [{id: "onTerminate"}];
    this.sendEvents(e);
  }
};


/**
 * Insert a new state in the browser history
 * @param {object} obj
 */
Client.MainFrame.prototype.historyPush = function (obj)
{
  if (obj.replace)
    window.history.replaceState(obj, obj.title, obj.url);
  else
    window.history.pushState(obj, obj.title, obj.url);
  if (obj.title) {
    var t = document.getElementsByTagName("title")[0];
    if (!t) {
      t = document.createElement("title");
      document.head.appendChild(t);
    }
    t.innerHTML = obj.title;
  }
};


/**
 * Send the popped obj to the server after a "back" action
 * @param {object} obj
 */
Client.MainFrame.prototype.historyPop = function (obj)
{
  var ev = [({id: "onHistoryPop", content: obj})];
  Client.mainFrame.sendEvents(ev);
};


/**
 * Load the CSS
 * @param {sty} sty - the css's to apply
 */
Client.MainFrame.prototype.loadStyle = function (sty)
{
  // let's see if there are some requirements to load
  if (!this.loadRequirements(sty.req))
    return;
  //
  // If there's some css rule to apply to the document, process it
  if (sty.css !== "") {
    //
    // Create a new stylesheet after the other ones
    var head = document.head || document.getElementsByTagName("head")[0];
    var style = document.createElement("style");
    style.type = "text/css";
    //
    // Styles in an inplace app need absolutization
    if (Client.resourceHome)
      sty.css = this.absStyle2(sty.css);
    //
    if (style.styleSheet)
      style.styleSheet.cssText = sty.css; // IE only
    else
      style.appendChild(document.createTextNode(sty.css));
    //
    // removing previous defined styles
    if (this.mystyle)
      head.removeChild(this.mystyle);
    //
    // and applying new ones
    head.appendChild(style);
    this.addedObjects.push(style);
    //
    // Styles in an inplace app need absolutization
    // UPDATE: IOS 10.3.1 breaks font-face substitution
    // if (Client.resourceHome)
    //   this.absStyle(style);
    //
    // Store the created style to remote it later
    this.mystyle = style;
  }
  //
  // the style has also some svg images to create
  if (sty.svg) {
    var body = document.body;
    //
    // For each svg, create an svg element
    for (var i = 0; i < sty.svg.length; i++) {
      //
      var x = document.createElement("x");
      x.innerHTML = sty.svg[i];
      //
      // To append it, first search for the SVG tag inside the X element
      for (var j = 0; j < x.childNodes.length; j++) {
        if (x.childNodes[j].tagName === "svg") {
          this.addedObjects.push(x.childNodes[j]);
          body.insertBefore(x.childNodes[j], body.firstChild);
          break;
        }
      }
    }
  }
  //
  this.theme = sty.theme || {};
};


/**
 * Absolutize an entire stylesheet
 * @param {style} style
 */
Client.MainFrame.prototype.absStyle = function (style)
{
  var r = style.sheet.rules;
  if (r) {
    for (var i = 0; i < r.length; i++) {
      var s = r[i].style;
      if (s) {
        for (var j = 0; j < s.length; j++) {
          var sp = s[j];
          if (Client.Utils.requireAbs(sp))
            s[sp] = Client.Utils.absStyle(s[sp]);
        }
      }
    }
  }
};


/**
 * Absolutize a css text
 * @param {string} csstext
 */
Client.MainFrame.prototype.absStyle2 = function (csstext)
{
  var idx = -1;
  var edx = 0;
  while (true) {
    idx = csstext.indexOf("url('", idx + 1);
    edx = csstext.indexOf("')", idx + 1);
    if (idx === -1 || edx === -1)
      break;
    var before = csstext.substring(idx, edx + 2);
    var after = Client.Utils.absStyle(before);
    if (before !== after) {
      csstext = csstext.substring(0, idx) + after + csstext.substring(edx + 2);
    }
  }
  return csstext;
};


/**
 * Load requirements for a view
 * @param {Array} reqList - requirement list
 */
Client.MainFrame.prototype.loadRequirements = function (reqList)
{
  var i, k, r, ok;
  //
  ok = true;
  k = Object.keys(reqList);
  for (i = 0; i < k.length; i++) {
    r = k[i];
    //
    // Already requested?
    if (this.requireMap[r] === undefined) {
      // No, put the request into the map to load it in a while
      this.paused = true;
      ok = false;
      this.requireMap[r] = reqList[r].type;
    }
  }
  //
  // Load missing requirements
  this.processRequirements();
  return ok;
};


/**
 * Go on with requirement loading
 */
Client.MainFrame.prototype.processRequirements = function ()
{
  var i, k, r, ok;
  //
  // Let's see if another requirement is already loading
  k = Object.keys(this.requireMap);
  for (i = 0; i < k.length; i++) {
    r = k[i];
    if (this.requireMap[r] === "loading")
      return false;
  }
  //
  // No requirement is loading, let's load the next one
  k = Object.keys(this.requireMap);
  for (i = 0; i < k.length; i++) {
    r = k[i];
    ok = false;
    switch (this.requireMap[r]) {
      case "jc":
        this.loadScript(r);
        ok = true;
        break;
      case "cs":
      case "fo":
        this.loadCss(r);
        ok = true;
        break;
      case "svg":
        this.loadSvg(r);
        ok = true;
        break;
    }
    if (ok) {
      this.requireMap[r] = "loading";
      break;
    }
  }
};


/**
 * Check if all requirements are loaded
 * @param {string} req url completed or "callback" if a callback has completed
 * @param {boolean} err - true if there was an error loading the resource
 */
Client.MainFrame.prototype.reqComplete = function (req, err)
{
  var r, k, i;
  //
  if (req === "callback") {
    // A callback has completed. Search for callback request and complete them
    k = Object.keys(this.requireMap);
    for (i = 0; i < k.length; i++) {
      r = k[i];
      //
      // A callback has beeing received... move to COMPLETE state all resource
      // that are in a CALLBACK state (i.e. waiting for a callback).
      // There could be a problem is the callback has beeing received BEFORE the onload/onerror event.
      // In this configuration the resource state is still LOADING and it would have been changed to CALLBACK
      // state in the if block below... Thus change to COMLPETE also resource that are still in LOADING state
      // but that were waiting for a callback.
      if ((this.requireMap[r] === "loading" || this.requireMap[r] === "callback") && r.indexOf("callback") > -1)
        this.requireMap[r] = "complete";
    }
  }
  else {
    // If there was an error, the resource is completed (the app can continue loading
    // even if the app will, probably, not work as espected)
    // If the resource was in LOADING state and was waiting for a callback, move it to CALLBACK state
    // When the callback will arrive it will be changed to COMPLETE state (see above if block)
    if (err)
      this.requireMap[req] = "complete";
    else if (this.requireMap[req] === "loading" && req.indexOf("callback") > -1)
      this.requireMap[req] = "callback";
    else
      this.requireMap[req] = "complete";
  }
  //
  // Check if other requirements are pending
  k = Object.keys(this.requireMap);
  for (i = 0; i < k.length; i++) {
    r = k[i];
    if (this.requireMap[r] !== "complete") {
      this.processRequirements();
      // let's wait for them
      return false;
    }
  }
  //
  // All requirements has been processed, we can go on
  this.paused = false;
  //
  // Restart commands
  if (Client.useAnimationFrame)
    Client.fireProcessCache();
  else
    this.processCache();
};


/**
 * Send an "app started" message to the shell
 */
Client.MainFrame.prototype.sendAppStart = function ()
{
  if (!this.appStarted) {
    setTimeout(function () {
      // If all commands were processed, start app
      if (!this.cmdCache.length) {
        this.appStarted = true;
        //
        if (this.device.shell)
          this.device.shell.postMessage({viewmsg: "appStarted", params: {sid: this.sid}}, "*");
      }
      else {
        // try again later
        this.processCache();
        this.sendAppStart();
      }
    }.bind(this), 100);
  }
};


/**
 * Load a new script into the document
 * @param {string} url
 */
Client.MainFrame.prototype.loadScript = function (url)
{
  var pthis = this;
  var script = document.createElement("script");
  script.type = "text/javascript";
  //
  if (script.readyState) {  //IE
    script.onreadystatechange = function () {
      if (script.readyState === "loaded" || script.readyState === "complete") {
        script.onreadystatechange = null;
        pthis.reqComplete(url);
      }
    };
  }
  else {  //Others
    script.onload = function () {
      pthis.reqComplete(url);
    };
    script.onerror = function () {
      pthis.reqComplete(url, true);
    };
  }
  //
  script.src = Client.Utils.abs(url);
  document.body.appendChild(script);
  this.addedObjects.push(script);
};


/**
 * Load a new stylesheet into the document
 * @param {string} url
 */
Client.MainFrame.prototype.loadCss = function (url)
{
  var pthis = this;
  var link = document.createElement("link");
  link.type = "text/css";
  link.rel = "stylesheet";
  //
  if (link.readyState) {  //IE
    link.onreadystatechange = function () {
      if (link.readyState === "loaded" || link.readyState === "complete") {
        link.onreadystatechange = null;
        pthis.reqComplete(url);
      }
    };
  }
  else {  //Others
    link.onload = function () {
      pthis.reqComplete(url);
    };
    link.onerror = function () {
      pthis.reqComplete(url, true);
    };
  }
  //
  link.href = Client.Utils.abs(url);
  //
  // Required CSS should be loaded BEFORE previous ones as they refers to
  // library elements that could be reconfigured by app CSS
  document.head.insertBefore(link, this.lastHeadObject.nextSibling);
  this.addedObjects.push(link);
};


/**
 * Load a new stylesheet into the document
 * @param {string} url
 */
Client.MainFrame.prototype.loadSvg = function (url)
{
  var pthis = this;
  var xhr = new XMLHttpRequest();
  xhr.open("GET", Client.Utils.abs(url));
  xhr.onload = function () {
    var body = document.body, x = document.createElement("x");
    x.innerHTML = xhr.responseText;
    for (var i = 0; i < x.childNodes.length; i++) {
      if (x.childNodes[i].tagName === "svg") {
        pthis.addedObjects.push(x.childNodes[i]);
        body.insertBefore(x.childNodes[i], body.firstChild);
        break;
      }
    }
    pthis.reqComplete(url);
  };
  xhr.onerror = function () {
    pthis.reqComplete(url, true);
  };
  xhr.send();
};


Client.MainFrame.prototype.attachEvents = function ()
{
  var pthis = this;
  var appui = document.getElementById("app-ui");
  //
  // set the onresize handler
  window.onresize = function (ev) {
    // update device props
    pthis.device.updateProp();
    // don't want to send resize too often (100 ms is enough)
    if (pthis.resizeTimeout)
      clearTimeout(pthis.resizeTimeout);
    pthis.resizeTimeout = setTimeout(function () {
      pthis.resizeTimeout = undefined;
      for (var i = 0; i < pthis.views.length; i++)
        pthis.views[i].onResize(ev);
      var e = [({id: "onResize", content: {width: appui.clientWidth, height: appui.clientHeight}})];
      Client.mainFrame.sendEvents(e);
      //
      // Scroll android input is needed
      if (pthis.device.operatingSystem === "android") {
        if (document.activeElement && (document.activeElement.tagName === "INPUT" ||
                document.activeElement.tagName === "TEXTAREA")) {
          document.activeElement.scrollIntoViewIfNeeded();
        }
      }
      //
    }, ev ? 100 : 0);
  };
  //
  // OnPopState
  window.addEventListener("popstate", function (ev) {
    if (!ev.state && pthis.device.isMobile && parent !== window)
      pthis.terminate();
    else
      pthis.historyPop(ev.state);
  }, true);
  //
  // Focus
  appui.addEventListener("focus", function (ev) {
    var el = Client.Utils.findElementFromDomObj(ev.target);
    if (el && el.view)
      el.view.setActiveElement(el);
  }, true);
  //
  // Key down
  appui.tabIndex = "0";
  appui.addEventListener("keydown", function (ev) {
    var key = Client.Utils.getKey(ev);
    //
    // Get the Client.Element from the html element
    var targetElem = Client.Utils.findElementFromDomObj(ev.target);
    //
    // If the target element doesn't want to handle the key down
    if (targetElem) {
      if (!targetElem.handleKeyDown || !targetElem.handleKeyDown(ev)) {
        // Get the view of the target element
        var currView = targetElem.view ? targetElem.view : targetElem;
        //
        // Find the element to activate
        var elemToAct = currView.findElementToActivate(key);
        //
        // If an element was found, activate it
        if (elemToAct) {
          ev.preventDefault();
          elemToAct.activate();
        }
      }
    }
  }, true);
  //
  // Register last scroll action
  appui.addEventListener("scroll", function (ev) {
    var now = new Date().getTime();
    pthis.lastScrollEvent = {srcElement: ev.srcElement, timeStamp: now};
    //
    // ...cancel touch start if near to a scroll event
    if (pthis.lastTouchEvent) {
      if (now - pthis.lastTouchEvent.timeStamp < 100 &&
              pthis.lastScrollEvent.srcElement.contains(pthis.lastTouchEvent.srcElement)) {
        pthis.preventClick();
      }
    }
  }, {passive: true, capture: true});
  //
  // Register a global touchstart event to...
  appui.addEventListener("touchstart", function (ev) {
    var now = new Date().getTime();
    //
    // Intercept start edge gesture
    var ex = undefined;
    if (pthis.edgeStart === undefined) {
      if (ev.touches.length === 1) {
        var x = ev.touches[0].screenX;
        if (x < 30)
          ex = 0;
        if (window.screen.width - x < 30)
          ex = window.screen.width;
      }
    }
    //
    pthis.lastTouchEvent = {srcElement: ev.srcElement, timeStamp: now, edge: ex};
    //
    // ...cancel touch start if near to a scroll event
    if (pthis.lastScrollEvent) {
      if (now - pthis.lastScrollEvent.timeStamp < 100 &&
              pthis.lastScrollEvent.srcElement.contains(ev.srcElement)) {
        pthis.preventClick();
      }
    }
    //
    // If the user touches outside an uncommitted input
    if (pthis.dirtyInput && ev.target.id !== pthis.dirtyInput) {
      var inp = Client.eleMap[pthis.dirtyInput];
      if (inp)
        inp.commit(ev);
    }
    //
  }, {passive: true, capture: true});
  //
  // Register a global touchmove event to intercept left/right edge gesture
  appui.addEventListener("touchmove", function (ev) {
    if (pthis.lastTouchEvent.edge !== undefined) {
      if (ev.changedTouches.length === 1) {
        var x = ev.changedTouches[0].screenX;
        var dx = Math.abs(x - pthis.lastTouchEvent.edge);
        if (dx > 60 && new Date() - pthis.lastTouchEvent.timeStamp < 200) {
          pthis.edgeGesture(pthis.lastTouchEvent.edge === 0 ? "left" : "right");
          pthis.lastTouchEvent.edge = undefined;
        }
      }
    }
  }, {passive: true, capture: true});
  //
  // Register a global mouseup event to...
  appui.addEventListener("mouseup", function (ev) {
    var now = new Date().getTime();
    //
    // ...cancel touch start if near to a scroll event
    if (pthis.lastScrollEvent) {
      if (now - pthis.lastScrollEvent.timeStamp < 200 &&
              pthis.lastScrollEvent.srcElement.contains(ev.srcElement)) {
        pthis.preventClick();
      }
    }
    //
  }, {passive: true, capture: true});
  //
  // If we are running inside a shell, attach an exit gesture. Not for root apps
  var onmob = this.device.isMobile && this.device.shell && !(this.device.shell instanceof Client.ShellEmulator);
  if (onmob) {
    var h = this.getHammerManager();
    if (h) {
      h.add(new Hammer.Swipe({event: 'exitswipe', pointers: 2, direction: Hammer.DIRECTION_DOWN, threshold: 10, velocity: 0.3}));
      h.on("exitswipe", function () {
        if (!this.device.rootApp)
          this.terminate();
      }.bind(this));
    }
  }
};


/**
 * Save the id of an uncommitted input
 * @param {String} inputId - the id of the input
 */
Client.MainFrame.prototype.setDirtyInput = function (inputId)
{
  // Dirty input is disabled because TH hanlde this edge case
  //this.dirtyInput = inputId;
};


/**
 * Get the id of the current uncommitted input
 * @returns {String}
 */
Client.MainFrame.prototype.getDirtyInput = function ()
{
  return this.dirtyInput;
};


/**
 * returns true if the application is in EditMode, false otherwise
 */
Client.MainFrame.prototype.isEditing = function ()
{
  return false;
};


/**
 * returns true if the application is in EditMode, false otherwise
 */
Client.MainFrame.prototype.isPlanning = function ()
{
  return false;
};


/**
 * Set client role and rtc properties
 * @param {Object} obj
 */
Client.MainFrame.prototype.setClientProperties = function (obj)
{
  // load requirements for webrtc connection
  var req = {"objects/base/webrtc.js": {name: "Webrtc", type: "jc"}, "objects/base/peer.js": {name: "Peer", type: "jc"}};
  if (!this.loadRequirements(req))
    return;
  else {
    if (Client.clientType)
      return;
  }
  //
  if (Client.clientType === undefined) {
    Client.clientType = obj.client;
    //
    // set client role for guest
    if (Client.clientType === "guest")
      Client.clientRole = obj.settings.clientRole;
  }
  //
  var sendAudio = true;
  var sendVideo = true;
  //
  // set guest's/owner's audio/video permission
  var av = (Client.clientType === "guest") ? obj.settings.guestStream : obj.settings.ownerStream;
  if (av.indexOf("v") === -1)
    sendVideo = false;
  if (av.indexOf("a") === -1)
    sendAudio = false;
  //
  if (sendVideo || sendAudio) {
    //
    // create webrtc object & popup
    var pthis = this;
    pthis.createWebRTC({sendVideo: sendVideo, sendAudio: sendAudio}, obj.settings.webRTCsettings);
    //
    // if session was already started, guest has to call owner to connect with this session
    if (obj.settings.ownerPeerId)
      pthis.callOwner(obj.settings.ownerPeerId);
  }
};


/**
 * Create WebRTC element for client
 * @param {Object} streamSettings
 * @param {Object} styleSettings
 */
Client.MainFrame.prototype.createWebRTC = function (streamSettings, styleSettings)
{
  // create and open webrtc popup
  this.rtc = new Client.WebRTC(streamSettings);
  this.rtc.prepareMultiClientPopup(this, styleSettings);
};


/**
 * Set owner's peerId received from server and start call
 * @param {String} id
 */
Client.MainFrame.prototype.callOwner = function (id)
{
  if (Client.clientType !== "owner") {
    //
    // update peerId property
    this.rtc.updateElement({peerId: id});
    //
    // connect to owner
    this.rtc.startCall();
  }
};


/**
 * Disconnect guest from current session
 */
Client.MainFrame.prototype.closeGuest = function ()
{
  this.rtc.endCall();
  this.rtc.removePopup();
  this.rtc = undefined;
  if (Client.clientType === "guest") {
    window.location = "http://www.instantdeveloper.com";
  }
};


/**
 * Close owner rtc popup & disconnect the last guest
 */
Client.MainFrame.prototype.closeOwnerRtc = function ()
{
  if (Client.clientType === "owner") {
    this.rtc.endCall();
    this.rtc.removePopup();
    this.rtc = undefined;
  }
};


Client.MainFrame.prototype.getHammerManager = function ()
{
  if (!this.hammer) {
    // Hammer 2.0.4 does not work on safari iOS 9.2
    var ok = true;
    //
    if (this.device.operatingSystem === "ios" && this.device.operatingSystemVersion &&
            this.device.operatingSystemVersion.trim() === "9.2") {
      ok = false;
      this.hammerEnabled = false;
    }
    //
    if (ok)
      this.hammer = new Hammer.Manager(document.getElementById("app-ui"));
  }
  //
  if (!this.hammer) {
    // create a fake hammer object
    this.hammer = {};
    this.hammer.on = function () {
    };
    this.hammer.off = function () {
    };
    this.hammer.add = function () {
    };
    this.hammer.get = function () {
    };
  }
  //
  return this.hammer;
};

/*
 * Checks if running in offline mode
 */
Client.isOffline = function ()
{
  return Proxy !== undefined;
};


/**
 * Set webRTC popup properties
 * @param {Object} properties
 */
Client.MainFrame.prototype.setWebRTCProperties = function (properties)
{
  if (properties && properties.parent) {
    if (this.rtcParent !== properties.parent) {
      var newParent = document.getElementById(properties.parent);
      if (newParent && this.rtc) {
        this.rtc.dialog.parentNode.removeChild(this.rtc.dialog);
        //
        newParent.appendChild(this.rtc.dialog);
        this.rtc.dialog.classList.remove("rtc-popup-default");
        this.rtc.dialog.classList.add("rtc-popup-custom");
        if (this.rtcParent)
          this.restoreCall();
        this.rtcParent = properties.parent;
      }
    }
  }
};


/**
 * Restore call when there is a change in rtc parent container or in device orientation
 */
Client.MainFrame.prototype.restoreCall = function ()
{
  if (this.rtc)
    this.rtc.restoreCall(Client.clientType);
};


/**
 * Let View know if there is a device somewhere, calling native events (gps..)
 * @param {object} obj
 */
Client.MainFrame.prototype.setPhDevice = function (obj)
{
  Client.phDeviceConnected = obj.device || "";
};


/**
 * Send a message to test auto preview
 * @param {Object} msg
 */
Client.MainFrame.prototype.sendMessageToTestAuto = function (msg)
{
  if (parent.postMessage)
    parent.postMessage(msg, "*");
};


/**
 * Send tag results to test auto preview
 * @param {Object} res
 */
Client.MainFrame.prototype.getTaggingDataResults = function (res)
{
  this.sendMessageToTestAuto({type: "testAutoMsg", content: {id: "taggingDataResults", taggingData: res}});
};


/**
 * Handle console.test command
 * @param {String} name
 * @param {Object} value
 */
Client.MainFrame.prototype.consoleTest = function (name, value) // jshint ignore:line
{

};


/**
 * Command to notify exception to testauto
 * @param {String} msg
 * @param {Object} stack
 */
Client.MainFrame.prototype.sendExceptionToTestAuto = function (msg, stack) // jshint ignore:line
{

};


/**
 * Save testauto tagging data
 * @param {String} data
 */
Client.MainFrame.prototype.setTaggingData = function (data) // jshint ignore:line
{
  var e = [{id: "saveTaggingData", content: data}];
  this.sendEvents(e);
};


/**
 * Close a popup
 */
Client.MainFrame.prototype.closePopup = function ()
{

};


/**
 * A swipe from edge was recognized
 * to override in UI framework
 * @param {string} side left or right
 */
Client.MainFrame.prototype.edgeGesture = function (side)
{
};


/**
 * Creates the offline screen for an app that goes offline
 */
Client.realizeOfflineScreen = function ()
{
  var appui = document.getElementById("app-ui");
  var offlineScreen = document.createElement("DIV");
  offlineScreen.className = "app-offline-screen";
  //
  var screenContent = document.createElement("DIV");
  screenContent.className = "app-screencover-content";
  offlineScreen.appendChild(screenContent);
  //
  var backImg = document.createElementNS("http://www.w3.org/2000/svg", "svg");
  backImg.setAttribute("class", "screencover-icon");
  var use = document.createElementNS("http://www.w3.org/2000/svg", "use");
  use.setAttributeNS("http://www.w3.org/1999/xlink", "xlink:href", "#icon-offlinestate");
  backImg.appendChild(use);
  screenContent.appendChild(backImg);
  //
  var offlineTitle = document.createElement("H1");
  offlineTitle.textContent = "You're offline";
  var offlineSubTitle = document.createElement("H2");
  offlineSubTitle.innerHTML = "Check your connection<br>to continue the session";
  screenContent.appendChild(offlineTitle);
  screenContent.appendChild(offlineSubTitle);
  appui.appendChild(offlineScreen);
  //
  window.setTimeout(function () {
    offlineScreen.classList.add("visible");
  }, 10);
};


/**
 * Get absolute IP address
 */
Client.MainFrame.prototype.getIP = function ()
{
  var cb = function (ip) {
    var e = [{id: "onIPAddress", content: ip}];
    this.sendEvents(e);
  }.bind(this);
  //
  var xhr = new XMLHttpRequest();
  xhr.open("GET", "https://api.ipify.org");
  //
  xhr.onload = function () {
    cb(xhr.responseText);
  };
  //
  xhr.onerror = function () {
    cb();
  };
  //
  xhr.send();
};


/**
 * Prevent Click
 * @param {int} ts
 */
Client.MainFrame.prototype.preventClick = function (ts)
{
  if (!ts)
    ts = 0;
  //
  this._preventClick = new Date().getTime() + ts;
  if (this._preventListeners) {
    for (var i in this._preventListeners) {
      this._preventListeners[i]();
    }
  }
};


/**
 * See if there are click prevented
 * @param {int} ts
 */
Client.MainFrame.prototype.isClickPrevented = function (ts)
{
  if (ts === undefined)
    ts = 250;
  return this._preventClick && new Date() - this._preventClick < ts;
};


/**
 * Register a click prevention listener
 * @param {string} id
 * @param {function} cb
 */
Client.MainFrame.prototype.registerPreventListener = function (id, cb)
{
  if (!this._preventListeners)
    this._preventListeners = {};
  this._preventListeners[id] = cb;
};


/**
 * unregister a click prevention listener
 * @param {string} id
 */
Client.MainFrame.prototype.unregisterPreventListener = function (id)
{
  if (!this._preventListeners)
    return;
  delete this._preventListeners[id];
};
