/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global Client, autosize */

/**
 * @class An input element
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the input
 */
Client.Input = function (element, parent, view)
{
  // The constructor will be called twice during object initialization
  // first time, without any parameter, then with them.
  if (element === undefined)
    return;
  if (view === undefined)
    return;
  //
  Client.Element.call(this, element, parent, view);
  //
  this.domObj = document.createElement(this.getDOMObjectType(element.type));
  //
  // Set readonly attribute to input element in case of readonly role
  if (Client.clientRole === "ro")
    this.domObj.setAttribute("readonly", "");
  //
  parent.appendChildObject(this, this.domObj);
  this.updateElement(element);
  this.attachEvents(element.events);
  //
  // No click on input if editing (prevents opening on input type file)
  if (Client.mainFrame.isEditing()) {
    this.domObj.addEventListener("click", function (ev) {
      ev.preventDefault();
      return false;
    });
  }
};


// Make Client.Input extend Client.Element
Client.Input.prototype = new Client.Element();


/**
 * Returns the type of domObj to create
 * @param {string} type
 */
Client.Input.prototype.getDOMObjectType = function (type)
{
  if (type === "static")
    return "p";
  //
  if (type === "textarea" || type === "select")
    return type;
  //
  return "input";
};


/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.Input.prototype.updateElement = function (el)
{
  this.purgeMyProp(el);
  //
  // Save properties to restore them in case of object recreation
  if (Client.mainFrame.isEditing()) {
    if (!this.el || el === this.el)
      this.el = {};
    for (var i in el)
      this.el[i] = el[i];
  }
  //
  // If there is a list of options, add them to the select
  // List example: [{"n":"Andrew","v":1},{"n":"Mary","v":2},{"n":"Luke","v":3}]
  if (el.list) {
    //
    // If el.list is a string make it a json object
    if (typeof el.list === "string")
      el.list = JSON.parse(el.list);
    //
    // Remove element children
    if (this.domObj.options)
      this.domObj.options.length = 0;
    //
    // Create an option for each item in the list
    for (var i = 0; i < el.list.length; i++) {
      var o = el.list[i];
      this.addOption(o.v, o.n);
    }
    //
    // restore value if present
    if (this.value)
      this.domObj.value = this.value;
    //
    // Base class must not use the list property
    delete el.list;
  }
  //
  if (el.data) {
    if (el.data.columns) {
      this.initInput(el);
      if (this.value)
        this.domObj.value = this.value;
    }
    else if (el.data.pos !== undefined) {
      if (el.data.data) // Update option
        this.updateOption(el.data);
      else  // Remove option
        this.removeOptionAtPos(el.data.pos);
    }
    else { // Add option
      this.addOption(el.data.data[0], el.data.data[1]);
      //
      // restore value if present
      if (this.value)
        this.domObj.value = this.value;
    }
    //
    // Base class must not use the data property
    delete el.data;
  }
  // If there is an activation key remember it
  if (el.cmdKey) {
    this.cmdKey = el.cmdKey.toUpperCase();
    //
    // Base class must not use the cmdKey property
    delete el.cmdKey;
  }
  //
  // Activate autosize if requird
  if (el.style && el.style.height && this.domObj.tagName === "TEXTAREA") {
    if (el.style.height === "auto") {
      setTimeout(function () {
        autosize(this.domObj);
      }.bind(this), 0);
      this.autosize = true;
    }
    else if (this.autosize) {
      autosize.destroy(this.domObj);
      this.autosize = false;
    }
  }
  //
  // Changing type? Maybe should recreate the domobj
  if (el.type && this.domObj.parentNode) {
    if (this.domObj.tagName.toLowerCase() !== this.getDOMObjectType(el.type)) {
      var newEl = document.createElement(this.getDOMObjectType(el.type));
      this.focusAttached = false;
      this.domObj.parentNode.insertBefore(newEl, this.domObj);
      this.domObj.parentNode.removeChild(this.domObj);
      var oldEl = this.domObj;
      this.domObj = newEl;
      if (this.el)
        this.updateElement(this.el);
    }
    //
    // Empty fake value for password fields
    if (this.domObj.tagName === "INPUT" && !this.focusAttached) {
      this.domObj.addEventListener("focus", function (ev) {
        if (this.domObj.type === "password") {
          if ("§".repeat(this.domObj.value.length) === this.domObj.value)
            this.domObj.value = "";
        }
      }.bind(this));
      this.focusAttached = true;
    }
  }
  //
  // converting value as innerText if type is static
  if (el.value !== undefined && this.domObj.tagName === "P") {
    el.innerText = el.value;
    delete el.value;
  }
  //
  if (el.value !== undefined) {
    // Storing value for list / option selection
    this.value = el.value;
    this.domObj.value = el.value;
    //
    // Update autosize textarea if required
    if (this.autosize)
      autosize.update(this.domObj);
    //
    delete el.value;
  }
  //
  // setting radio group
  if (el.radioGroup !== undefined) {
    this.domObj.setAttribute("name", el.radioGroup);
    delete el.radioGroup;
  }

  //
  Client.Element.prototype.updateElement.call(this, el);
};


/**
 * Tell the server the value is changed
 * @returns {undefined}
 */
Client.Input.prototype.commit = function (ev) {
  var e = this.changeValueEvent();
  //
  if (this.sendOnChange)
    e.push({obj: this.id, id: "onChange", content: this.saveEvent(ev)});
  //
  Client.mainFrame.sendEvents(e);
  //
  // I'm not the dirty input anymore
  Client.mainFrame.setDirtyInput(null);
};


/**
 * Attach events handler
 * @param {Array} events - array of the events to handle
 */
Client.Input.prototype.attachEvents = function (events)
{
  if (!events)
    return;
  //
  Client.Element.prototype.attachEvents.call(this, events);
  //
  var pthis = this;
  //
  // file uploader
  var idxup = events.indexOf("onUpload");
  if (idxup >= 0 && this.domObj.type === "file") {
    //
    // do not attach this event because the DOM object does not allow it
    events.splice(idxup, 1);
    //
    // attach change listener to see when new files are uploaded
    this.domObj.addEventListener("change", function (ev) {
      //
      // generate url form upload
      var qstring = Client.Utils.getRESTQueryString({msgType: "input-upload", objId: pthis.id});
      //
      // generate multipart form and send it
      var formData = new FormData();
      for (var i = 0; i < pthis.domObj.files.length; i++)
        formData.append("file", pthis.domObj.files[i]);
      //
      var req = new XMLHttpRequest();
      req.open("POST", qstring, true);
      req.send(formData);
    }, false);
  }
  //
  if (events.indexOf("onChange") >= 0) {
    this.domObj.onchange = function (ev) {
      pthis.checkError(true);
      var e = pthis.changeValueEvent();
      e.push({obj: pthis.id, id: "onChange", content: pthis.saveEvent(ev)});
      Client.mainFrame.sendEvents(e);
    };
  }
  //
  if (events.indexOf("onInput") >= 0) {
    this.domObj.oninput = function (ev) {
      pthis.checkError();
      var e = pthis.changeValueEvent();
      e.push({obj: pthis.id, id: "onInput", content: pthis.saveEvent(ev)});
      Client.mainFrame.sendEvents(e);
    };
  }
  //
  if (events.indexOf("onChange") < 0 && events.indexOf("onInput") < 0) {
    this.domObj.onchange = function () {
      pthis.checkError(true);
      var e = pthis.changeValueEvent();
      Client.mainFrame.sendEvents(e);
    };
  }
  //
  if (events.indexOf("onInput") < 0) {
    this.domObj.oninput = function () {
      pthis.checkError();
      if (Client.mainFrame.device.isMobile)
        Client.mainFrame.setDirtyInput(pthis.id);
    };
  }
};


/**
 * Prepare the event of change value
 * @returns {array}
 */
Client.Input.prototype.changeValueEvent = function () {
  var propName = "value";
  if (this.domObj.type === "checkbox")
    propName = "checked";
  var x = {obj: this.id, id: "chgProp", content: {name: propName, value: this.domObj[propName], clid: Client.id}};
  var y = {obj: this.id, id: "chgProp", content: {name: "errorText", value: this.domObj.validationMessage, clid: Client.id}};
  return [y, x];
};


/**
 * Initialize the input
 * @param {Object} el - element representation
 */
Client.Input.prototype.initInput = function (el)
{
  // First remove old options
  while (this.domObj.childNodes.length > 0)
    this.domObj.removeChild(this.domObj.childNodes[0]);
  //
  for (var i in el.data.data)
    this.addOption(el.data.data[i][0], el.data.data[i][1]);
};


/**
 * Remove an option from the select at the given position
 * @param {int} pos - the position of the option to remove
 */
Client.Input.prototype.removeOptionAtPos = function (pos)
{
  this.domObj.removeChild(this.domObj.childNodes[pos]);
};


/**
 * Add an option to the select
 * @param {String} value - the value of the option
 * @param {String} name - the text of the option
 */
Client.Input.prototype.addOption = function (value, name)
{
  var opt = document.createElement("option");
  opt.value = value ? value : "";
  opt.text = name;
  this.domObj.appendChild(opt);
};


/**
 * Update a select option
 * @param {Object} data - the data of the option to update. Example: {pos: 3, data:[3, "Andrea"]}
 */
Client.Input.prototype.updateOption = function (data)
{
  var opt = this.domObj.childNodes[data.pos];
  opt.value = data.data[0];
  opt.text = data.data[1];
};


/**
 * Set custom error message
 * @param {String} message - the error message
 * @param {bool} srv
 */
Client.Input.prototype.setError = function (message, srv)
{
  Client.Element.prototype.setError.call(this, message, srv);
  this.domObj.setCustomValidity(message);
};


/**
 * Remove the element and its children from the element map
 * @param {boolean} firstLevel - if true remove the dom of the element too
 * @param {boolean} triggerAnimation - if true and on firstLevel trigger the animation of 'removing'
 */
Client.Input.prototype.close = function (firstLevel, triggerAnimation)
{
  if (this.autosize)
    autosize.destroy(this.domObj);
  //
  if (this.id === Client.mainFrame.getDirtyInput())
    Client.mainFrame.setDirtyInput(null);
  //
  Client.Element.prototype.close.call(this, firstLevel, triggerAnimation);
};
