/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global Client */

/**
 * @class A bootstrap navigation element
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the button
 */
Client.BSNav = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  // Creating dropdown menu container
  this.domObj = document.createElement(element.type === "bc" ? "ol" : "ul");
  //
  this.updateElement(element);
  this.attachEvents(element.events);
  //
  parent.appendChildObject(this, this.domObj);
  //
  // Create children after attaching element as external components, rely on dom structure
  this.createChildren(element);
};

// Make BSNav extend Client.Element
Client.BSNav.prototype = new Client.Element();

/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.BSNav.prototype.updateElement = function (el)
{
  // Force update for the first time
  var update = this.domObj.parentNode === null;
  //
  Client.BSHelper.updateElement(el, this);
  //
  // specific class name for this layout
  if (el.className !== undefined) {
    this.className = el.className;
    update = true;
    delete el.className;
  }
  //
  if (el.type !== undefined) {
    this.type = el.type;
    update = true;
    delete el.type;
  }
  //
  if (el.stacked !== undefined) {
    this.stacked = el.stacked;
    update = true;
    delete el.stacked;
  }
  //
  if (el.justified !== undefined) {
    this.justified = el.justified;
    update = true;
    delete el.justified;
  }
  //
  if (update) {
    //
    var cs = "nav";
    if (this.type === "bc")
      cs = "breadcrumb";
    if (this.type === "pag")
      cs = "pagination";
    if (this.type === "pgr")
      cs = "pager";
    //
    if (this.type === "tab")
      cs += " nav-tabs";
    if (this.type === "nav")
      cs += " navbar-nav";
    if (this.type === "pill")
      cs += " nav-pills";
    if (this.stacked)
      cs += " nav-stacked";
    if (this.justified)
      cs += " nav-justified";
    //
    if (this.className)
      cs += " " + this.className;
    //
    this.domObj.className = cs;
  }
  //
  Client.Element.prototype.updateElement.call(this, el);
};



/**
 * @class A bootstrap navigation link
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the button
 */
Client.BSNavLink = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  var pthis = this;
  //
  // Search for buttons / dropdowns
  if (element.children) {
    for (var i = 0; i < element.children.length; i++) {
      var c = element.children[i].class || element.children[i].c;
      if (c === "BSDDMenu") {
        this.hasDropDown = true;
      }
    }
  }
  //
  // Creating dropdown menu container
  this.liObj = document.createElement("li");
  this.domObj = document.createElement("a");
  this.domObj.href = "javascript:void(0)";
  this.liObj.appendChild(this.domObj);
  //
  this.updateElement(element);
  this.attachEvents(element.events);
  //
  // Attach a click handler to close the menu if the object has also the click event
  //if (element.events.indexOf("onClick") > -1) {
  //  this.domObj.addEventListener("click", function () {
  //    $('#' + pthis.parent.id).dropdown('toggle');
  //  });
  //}
  //
  parent.appendChildObject(this, this.liObj);
  //
  // Create children after attaching element as external components, rely on dom structure
  this.createChildren(element);
};

// Make BSNavLink extend Client.Element
Client.BSNavLink.prototype = new Client.Element();

/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.BSNavLink.prototype.updateElement = function (el)
{
  // Force update for the first time
  var update = this.liObj.parentNode === null;
  //
  Client.BSHelper.updateElement(el, this);
  //
  // specific class name for this layout
  if (el.className !== undefined) {
    this.className = el.className;
    update = true;
    delete el.className;
  }
  if (el.active !== undefined) {
    this.active = el.active;
    //
    if (this.parent.type === "bc")
      this.changeTag(this.active ? "span" : "a");
    //
    update = true;
    delete el.active;
  }
  if (el.disabled !== undefined) {
    this.disabled = el.disabled;
    update = true;
    delete el.disabled;
  }
  //
  if (update) {
    //
    var cs = this.hasDropDown ? "dropdown" : "";
    if (this.disabled)
      cs += " disabled";
    if (this.active)
      cs += " active";
    if (this.className)
      cs += " " + this.className;
    //
    this.liObj.className = cs;
    //
    if (this.hasDropDown) {
      this.domObj.setAttribute("data-toggle", "dropdown");
      this.domObj.className = "dropdown-toggle";
      if (!this.caretObj) {
        this.caretObj = document.createElement("span");
        this.caretObj.className = "caret";
      }
    }
    //
    if (this.parent.type === "tab")
      this.domObj.setAttribute("data-toggle", "tab");
    if (this.parent.type === "pill")
      this.domObj.setAttribute("data-toggle", "pill");
  }
  //
  Client.Element.prototype.updateElement.call(this, el);
  //
  if (this.caretObj)
    this.domObj.appendChild(this.caretObj);
};

/**
 * Return the dom root object of this element
 * @returns {DomElement}
 */
Client.BSNavLink.prototype.getRootObject = function ()
{
  return this.liObj;
};


/**
 * Append a child DOM Object to this element DOM Object
 * Usually the child element is not contained in this element array
 * it will be pushed there just after this function call.
 * @param {Element} child child element that requested the insertion
 * @param {HTMLElement} domObj child DOM object to add
 */
Client.BSNavLink.prototype.appendChildObject = function (child, domObj)
{
  this.domObj.appendChild(domObj);
};


/**
 * @class A bootstrap navigation bar element
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the button
 */
Client.BSNavBar = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  // Creating dropdown menu container
  this.domObj = document.createElement("nav");
  this.divObj = document.createElement("div");
  this.divObj.className = "container-fluid";
  this.domObj.appendChild(this.divObj);
  //
  this.updateElement(element);
  this.attachEvents(element.events);
  //
  parent.appendChildObject(this, this.domObj);
  //
  // Create children after attaching element as external components, rely on dom structure
  this.createChildren(element);
};

// Make BSNavBar extend Client.Element
Client.BSNavBar.prototype = new Client.Element();

/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.BSNavBar.prototype.updateElement = function (el)
{
  // Force update for the first time
  var update = this.domObj.parentNode === null;
  //
  Client.BSHelper.updateElement(el, this);
  //
  // specific class name for this layout
  if (el.className !== undefined) {
    this.className = el.className;
    update = true;
    delete el.className;
  }
  //
  if (el.type !== undefined) {
    this.type = el.type;
    update = true;
    delete el.type;
  }
  //
  if (el.inverted !== undefined) {
    this.inverted = el.inverted;
    update = true;
    delete el.inverted;
  }
  //
  if (update) {
    //
    var cs = "navbar";
    //
    if (this.inverted)
      cs += " navbar-inverse";
    else
      cs += " navbar-default";
    //
    if (this.type === "ftop")
      cs += " navbar-fixed-top";
    if (this.type === "fbot")
      cs += " navbar-fixed-bottom";
    if (this.type === "stop")
      cs += " navbar-static-top";
    //
    if (this.className)
      cs += " " + this.className;
    //
    this.domObj.className = cs;
  }
  //
  Client.Element.prototype.updateElement.call(this, el);
};


/**
 * @class A bootstrap navigation bar section
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the button
 */
Client.BSNavSection = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  // Creating dropdown menu container
  this.domObj = document.createElement("div");
  //
  this.updateElement(element);
  this.attachEvents(element.events);
  //
  parent.appendChildObject(this, this.domObj);
  //
  // Create children after attaching element as external components, rely on dom structure
  this.createChildren(element);
};

// Make BSNavSection extend Client.Element
Client.BSNavSection.prototype = new Client.Element();

/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.BSNavSection.prototype.updateElement = function (el)
{
  // Force update for the first time
  var update = this.domObj.parentNode === null;
  //
  Client.BSHelper.updateElement(el, this);
  //
  // specific class name for this layout
  if (el.className !== undefined) {
    this.className = el.className;
    update = true;
    delete el.className;
  }
  //
  if (el.type !== undefined) {
    this.type = el.type;
    update = true;
    delete el.type;
  }
  //
  if (update) {
    //
    var cs = "";
    //
    if (this.type === "header")
      cs += " navbar-header";
    else
      cs += " collapse navbar-collapse";
    //
    if (this.className)
      cs += " " + this.className;
    //
    this.domObj.className = cs;
    //
    if (this.type === "header") {
      if (!this.btnObj) {
        this.btnObj = document.createElement("button");
        this.domObj.appendChild(this.btnObj);
        this.btnObj.type = "button";
        this.btnObj.className = "navbar-toggle collapsed";
        this.btnObj.setAttribute("data-toggle", "collapse");
        this.btnObj.innerHTML = "<span class='icon-bar'></span><span class='icon-bar'></span><span class='icon-bar'></span>";
      }
    }
    else if (!this.targetSet && !Client.mainFrame.isEditing()) {
      // Try to set header btn target
      this.targetSet = true;
      var idx = this.parent.ne();
      if (idx && idx > 0) {
        var h = this.parent.elements[idx - 1];
        if (h.btnObj)
          h.btnObj.setAttribute("data-target", "#" + this.id);
      }
    }
  }
  //
  Client.Element.prototype.updateElement.call(this, el);
};

/**
 * Append a child DOM Object to this element DOM Object
 * Usually the child element is not contained in this element array
 * it will be pushed there just after this function call.
 * @param {Element} child child element that requested the insertion
 * @param {HTMLElement} domObj child DOM object to add
 */
Client.BSNavSection.prototype.appendChildObject = function (child, domObj)
{
  this.domObj.appendChild(domObj);
  if (!domObj.classList.contains("navbar-right") && !domObj.classList.contains("navbar-left"))
    domObj.classList.add("navbar-left");
};


/**
 * Shows or hides section content
 * @param {bool} flag
 */
Client.BSNavSection.prototype.showContent = function (flag)
{
  $("#" + this.id).collapse(flag ? "show" : "hide");
};
