/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/**
 * @class A calendar Object
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the calendar
 */
Client.Calendar = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  this.domObj = document.createElement("div");
  //
  // Create the div that will contain the calendar
  var calendarDiv = document.createElement("div");
  //
  // Generate an id for the calendar formed by a letter and 5 numbers
  this.calendarId = "c" + Math.floor(Math.random() * (99999 - 10000) + 10000);
  calendarDiv.id = this.calendarId;
  this.domObj.appendChild(calendarDiv);
  //
  parent.appendChildObject(this, this.domObj);
  //
  this.eventsMap = {};
  this.optionsCache = {};
  this.updateElement(element);
  this.attachEvents(element.events);
  //
  // If in edit mode load sample data
  if (Client.mainFrame.isEditing()) {
    var el = this.loadSampleData();
    this.updateElement(el);
  }
};


// Make Client.Calendar extend Client.Element
Client.Calendar.prototype = new Client.Element();


/**
 * Define the calendar and event options
 */
Client.Calendar.default = {
  global: {
    // General Display
    header: {
      left: 'title',
      center: '',
      right: 'today prev,next'
    },
    buttonIcons: {// takes effect when "theme" is false
      prev: 'left-single-arrow',
      next: 'right-single-arrow',
      prevYear: 'left-double-arrow',
      nextYear: 'right-double-arrow'
    },
    theme: false,
    themeButtonIcons: {// takes effect when "theme" is true
      prev: 'circle-triangle-w',
      next: 'circle-triangle-e',
      prevYear: 'seek-prev',
      nextYear: 'seek-next'
    },
    firstDay: 0, // (Sunday) - it depends on the current "lang"
    isRTL: false,
    weekends: true,
    hiddenDays: [],
    fixedWeekCount: true,
    weekNumbers: false,
    weekNumberCalculation: "local", // ISO (for ISO8601 week numbers)
    businessHours: false, // boolean or an object, example: {start: "10:00", end: "18:00", dow: [1,2,3,4]}
    height: "auto", // "auto" or an integer
    contentHeight: "auto", // "auto" or an integer
    aspectRatio: 1.35, // float
    eventLimit: false, // boolean or integer
    eventLimitClick: "popover", // Other possible values: "week", "day", view name
    //
    // Views
    defaultView: "month", // Other possible values: basicWeek, basicDay, agendaWeek, agendaDay
    //
    // Current Date - for further informations visit http://fullcalendar.io/docs/utilities/Moment/
    defaultDate: "2014-05-01",
    //
    // Text/Time Customization
    lang: "en",
    timeFormat: "H:mm",
    //
    // Selection
    selectable: false,
    selectHelper: false, // only agenda views
    unselectAuto: true, // takes effect when selectable is set to true
    unselectCancel: "", // a way to specify elements that will ignore the unselectAuto option.
    selectOverlap: true,
    selectConstraint: "" // only applicable when the selectable option is activated - accept an event ID, "businessHours", object
  },
  event: {
    // Event Rendering
    eventColor: "#3a87ad", // background and border colors for all events on the calendar
    eventBackgroundColor: "#3a87ad", // background color for all events on the calendar
    eventBorderColor: "#3a87ad", // border color for all events on the calendar
    eventTextColor: "#FFFFFF", // text color for all events on the calendar
    nextDayThreshold: "09:00:00",
    //
    // Event Dragging & Resizing
    editable: false,
    eventOverlap: true
  },
  agenda: {
    allDaySlot: true,
    allDayText: "all-day", // The default value depends on the current "lang"
    axisFormat: "h(:mm)a", // The default value depends on the current "lang"
    slotDuration: "00:30:00",
    snapDuration: "00:30:00",
    scrollTime: "06:00:00",
    minTime: "00:00:00",
    maxTime: "24:00:00",
    slotEventOverlap: true
  }
};


/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.Calendar.prototype.updateElement = function (el)
{
  // Get options
  if (el.calendarOptions) {
    this.calendarOptions = JSON.parse(el.calendarOptions);
    this.changeCalendarOptions(el.calendarOptions);
    //
    delete el.calendarOptions;
  }
  if (el.eventsOptions) {
    this.eventsOptions = JSON.parse(el.eventsOptions);
    this.changeCalendarOptions(el.eventsOptions);
    //
    delete el.eventsOptions;
  }
  if (el.view) {
    this.view = el.view;
    this.changeView(el.view);
    delete el.view;
  }
  //
  if (el.currentDate) {
    if (this.currentDate && !this.currentDate.isSame(el.currentDate)) {
      this.currentDate = moment(el.currentDate);
      if (this.calWrapper) {
        this.calWrapper.fullCalendar("goToDate", this.currentDate);
      }
    }
  }
  //
  // Get fullcalendar data
  if (el.data) {
    // Load initial data
    if (el.data.columns)
      this.initCalendar(el);
    else if (el.data.pos !== undefined) {
      if (el.data.data) // Update data
        this.updateData(el.data);
      else  // Remove data
        this.removeData(el.data);
    }
    else  // Add data
      this.addData(el.data.data);

    //
    Client.Element.prototype.updateElement.call(this, el);
  }
  //
  if (el.style) {
    var style = el.style;
    if (typeof style === "string")
      style = JSON.parse(el.style);
    //
    if (style.height)
      this.changeDimension("height", style.height);
    if (style.width)
      this.changeDimension("width", style.width);
  }
};


/**
 * Change the width or the height of the calendar
 * @param {String} dim - width or height
 * @param {Int} size
 */
Client.Calendar.prototype.changeDimension = function (dim, size)
{
  if (!this.calWrapper) {
    this.optionsCache[dim] = size;
    return;
  }
  //
  var idx = size.indexOf("px");
  if (idx > -1) {
    this.calWrapper.fullCalendar("option", dim, parseInt(size.substring(0, idx)));
    delete this.percDim[dim];
  }
  else {
    idx = size.indexOf("%");
    if (idx > -1) {
      var s = window.getComputedStyle(this.domObj.parentNode, null);
      var parHeight = s.height;
      if (parHeight) {
        var parNumHeight = parHeight.match(/\d/g);
        parNumHeight = parseInt(parNumHeight.join(""));
        //
        var perc = parseInt(size.substring(0, idx));
        //
        this.calWrapper.fullCalendar("option", dim, parNumHeight * perc / 100);
        //
        if (!this.percDim)
          this.percDim = {};
        this.percDim[dim] = size;
      }
    }
  }
};


/**
 * Initialize the calendar
 * @param {Object} el - initial calendar data
 */
Client.Calendar.prototype.initCalendar = function (el)
{
  // Get the jQuery wrapper containing the calendar div
  this.calWrapper = $("#" + this.calendarId);
  //
  // Build the events array
  this.data = this.buildEventsArray(el);
  //
  // Create an object to remember which column represents the id, which one the title, ecc.
  var cols = el.data.columns;
  this.columns = {};
  this.columns["id"] = cols[0];
  this.columns["title"] = cols[1];
  this.columns["start"] = cols[2];
  if (cols.length > 2 && cols[3])
    this.columns["end"] = cols[3];
  //
  // Create empty calendar
  if (!this.options)
    this.options = {};
  //
  this.attachCalendarEvents(this.options);
  this.calWrapper.fullCalendar(this.options);
  //
  // Add events
  this.calWrapper.fullCalendar("addEventSource", this.data);
  //
  if (this.optionsCache) {
    if (this.optionsCache.options)
      this.changeCalendarOptions(this.optionsCache.options);
    if (this.optionsCache.height)
      this.changeDimension("height", this.optionsCache.height);
    if (this.optionsCache.width)
      this.changedimension("width", this.optionsCache.width);
    delete this.optionsCache;
  }
};


/**
 * Builds the array of the events
 * @param {Object} el - the events to be represented by the calendar
 * @returns {Object}
 */
Client.Calendar.prototype.buildEventsArray = function (el)
{
  var events = [];
  //
  for (var i = 0; i < el.data.data.length; i++) {
    // Build an event object for each set of data
    var evObj = this.buildEvent(el.data.data[i]);
    events.push(evObj);
  }
  //
  return {events: events};
};


/**
 * Builds an event
 * @param {Array} ev - event data
 * @returns {EventObject}
 */
Client.Calendar.prototype.buildEvent = function (ev) {
  var evObj = {};
  evObj["id"] = ev[0];
  evObj["title"] = ev[1];
  evObj["start"] = ev[2];
  if (ev.length > 2 && ev[3])
    evObj["end"] = ev[3];
  //
  return evObj;
};


/**
 * Update calendar event after a datamap update
 * @param {Object} data - data of the event to update
 */
Client.Calendar.prototype.updateData = function (data)
{
  // Get the event id
  var id = data.data[0];
  //
  // Search the event in the calendar
  var events = this.searchEvents({id: id});
  for (var i = 0; i < events.length; i++) {
    var currEv = events[i];
    //
    // If the client modifies an event the server will receive a message so it can modify the datamap.
    // When the server modifies the datamap the client will receive a message so it can modify its data.
    // To avoid loops I skip the event to update if it's the same of the event on the calendar.
    if (this.sameEvent(data, currEv))
      continue;
    //
    // Update the event properties
    currEv["id"] = id;
    currEv["title"] = data.data[1];
    currEv["start"] = moment(data.data[2]);
    if (data.data.length > 2)
      currEv["end"] = moment(data.data[3]);
    //
    this.calWrapper.fullCalendar("updateEvent", currEv);
  }
};


/**
 * Remove a calendar event after a datamap update
 * @param {Object} data - the event to remove
 */
Client.Calendar.prototype.removeData = function (data)
{
  // Find the event to remove
  var event = this.data.events[data.pos];
  if (event) {
    // Remove the event from the data array
    this.data.events.splice(data.pos, 1);
    //
    // Remove the event from the calendar
    this.calWrapper.fullCalendar("removeEvents", event.id);
  }
};


/**
 * Add an event to the calendar after a datamap update
 * @param {Object} data - data of the event to add
 */
Client.Calendar.prototype.addData = function (data)
{
  if (this.calWrapper && data.length > 1) {
    // Build an event object
    var newEvent = this.buildEvent(data);
    //
    // Add the event to the data array
    this.data.events.push(newEvent);
    //
    // Add the event to the calendar
    this.calWrapper.fullCalendar("renderEvent", newEvent, true);
  }
};


/**
 * Attach the events handlers to the calendar
 * @param {Object} opt - the object to attach the handlers to
 */
Client.Calendar.prototype.attachCalendarEvents = function (opt)
{
  var pthis = this;
  //
  opt.dayClick = function (date, jsEvent, view) {
    var e = [{obj: pthis.id, id: "onDayClick", content: {event: date.format()}}];
    Client.mainFrame.sendEvents(e);
  };
  //
  if (!Client.mainFrame.device.isMobile) {
    opt.eventClick = function (event, jsEvent, view) {
      var e = [{obj: pthis.id, id: "onEventClick", content: {event: event.id}}];
      Client.mainFrame.sendEvents(e);
    };
    //
    opt.eventDrop = function (event, delta, revertFunc) {
      pthis.updateEvent(event.id, event);
      var returnedEvent = {};
      returnedEvent["id"] = event.id;
      returnedEvent["title"] = event.id;
      returnedEvent["start"] = event.start;
      returnedEvent["end"] = event.end;
      //
      var e = [{obj: pthis.id, id: "onEventDrop", content: {event: returnedEvent}}];
      Client.mainFrame.sendEvents(e);
    };
    //
    opt.eventResize = function (event, delta, revertFunc) {
      pthis.updateEvent(event.id, event);
      var returnedEvent = {};
      returnedEvent["id"] = event.id;
      returnedEvent["end"] = event.end;
      //
      var e = [{obj: pthis.id, id: "onEventResize", content: {event: returnedEvent}}];
      Client.mainFrame.sendEvents(e);
    };
  }
  else {
    // This callback will be triggered when the dom objects representing the events are drawn on the calendar
    opt.eventRender = function (event, element, view) {
      pthis.eventsMap[event.id] = event;
      element[0].setAttribute("id", event.id);
    };
    //
    var hammer = Client.mainFrame.getHammerManager();
    if (!hammer.get("tap")) {
      var recognizer = new Hammer.Tap();
      hammer.add(recognizer);
    }
    //
    hammer.on("tap", function (ev) {
      var elem = Client.Utils.getParentWithClass(ev.srcEvent.target, "fc-event");
      if (elem) {
        var event = pthis.eventsMap[elem.id];
        if (event) {
          var e = [{obj: pthis.id, id: "onEventClick", content: {event: event.id}}];
          Client.mainFrame.sendEvents(e);
        }
      }
    });
  }
  //
  // This callback will get triggered when the user changes the view, or when any of the date navigation methods are called
  opt.viewRender = function (view, element) {
    if (pthis.calWrapper) {
      var currentDate = pthis.calWrapper.fullCalendar("getDate");
      if (!pthis.currentDate || !pthis.currentDate.isSame(currentDate)) {
        pthis.currentDate = currentDate;
        //
        // Inform the server
        var e = [{obj: pthis.id, id: "chgProp", content: {name: "currentDate", value: pthis.currentDate, clid: Client.id}}];
        Client.mainFrame.sendEvents(e);
      }
    }
  };
};


/**
 * Build sample data object
 * @returns {Object}
 */
Client.Calendar.prototype.loadSampleData = function ()
{
  var data = {};
  data["columns"] = ["id", "title", "start", "end"];
  var today = new Date();
  var todayEnd = today.setHours(today.getHours() + 1);
  var tomorrow = today.setDate(today.getDate() + 1);
  var tomorrowEnd = tomorrow + 60 * 60;
  data["data"] = [[0, "Meeting", today, todayEnd], [1, "Meeting", tomorrow, tomorrowEnd]];
  //
  return {data: data};
};


/**
 * Update calendar options
 * @param {Object} el - options to update
 */
Client.Calendar.prototype.changeCalendarOptions = function (el)
{
  // If there is a calendar, update its options
  if (this.calWrapper) {
    if (!this.options)
      this.options = {};
    //
    // Remember the old height
    var oldHeight = this.options.height;
    //
    var newOpt = el;
    if (typeof el === "string")
      newOpt = JSON.parse(el);
    var keys = Object.keys(newOpt);
    for (var i = 0; i < keys.length; i++)
      this.options[keys[i]] = newOpt[keys[i]];
    //
    // Update the height separately and only if it's different from the old one
    if (this.options.height && this.options.height !== oldHeight)
      this.changeDimension("height", this.options.height);
    //
    this.calWrapper.fullCalendar("destroy");
    this.calWrapper.fullCalendar(this.options);
    this.calWrapper.fullCalendar("addEventSource", this.data);
  }
  else {
    // If there isn't a calendar yet, cache the options
    var newOpt = el;
    if (typeof el === "string")
      newOpt = JSON.parse(el);
    var keys = Object.keys(newOpt);
    if (!this.optionsCache.options)
      this.optionsCache.options = {};
    for (var i = 0; i < keys.length; i++)
      this.optionsCache.options[keys[i]] = newOpt[keys[i]];
  }
};


/**
 Move the calendar one step back (either by a month, week, or day).
 If the calendar is in month view, will move the calendar back one month.
 If the calendar is in basicWeek or agendaWeek, will move the calendar back one week.
 If the calendar is in basicDay or agendaDay, will move the calendar back one day.
 */
Client.Calendar.prototype.prev = function ()
{
  if (this.calWrapper)
    this.calWrapper.fullCalendar("prev");
};


/**
 Move the calendar one step forward (either by a month, week, or day).
 If the calendar is in month view, will move the calendar forward one month.
 If the calendar is in basicWeek or agendaWeek, will move the calendar forward one week.
 If the calendar is in basicDay or agendaDay, will move the calendar forward one day.
 */
Client.Calendar.prototype.next = function ()
{
  if (this.calWrapper)
    this.calWrapper.fullCalendar("next");
};
/**
 Move the calendar back one year.
 */
Client.Calendar.prototype.prevYear = function ()
{
  if (this.calWrapper)
    this.calWrapper.fullCalendar("prevYear");
};
/**
 Move the calendar forward one year.
 */
Client.Calendar.prototype.nextYear = function ()
{
  if (this.calWrapper)
    this.calWrapper.fullCalendar("nextYear");
};
/**
 Move the calendar to the current date.
 */
Client.Calendar.prototype.today = function ()
{
  if (this.calWrapper)
    this.calWrapper.fullCalendar('today');
};


/**
 * Remove an event from the calendar
 * @param {String} eventId - id of the event to remove, all events with the same id will be removed.
 *  If it's omitted, all events are removed.
 */
Client.Calendar.prototype.removeEvents = function (eventId)
{
  if (!this.calWrapper)
    return;
  //
  this.calWrapper.fullCalendar("removeEvents", eventId);
  //
  var e = [{obj: this.id, id: "chgData", content: {name: "delete", id: eventId}}];
  if (!this.deleteEventTimeout) {
    var pthis = this;
    pthis.deleteEventTimeout = setTimeout(function () {
      pthis.deleteEventTimeout = undefined;
      Client.mainFrame.sendEvents(e);
    }, 0);
  }
};


/**
 * Adds events to the calendar
 * @param {Object} options - the event to add
 */
Client.Calendar.prototype.addEvent = function (options)
{
  // If there isn't a calendar or the title or the start time are missing I can't create the new event
  if (!this.calWrapper || !options.title || !options.start)
    return;
  //
  // Build an event object
  var newEvent = {};
  var keys = Object.keys(options);
  for (var i = 0; i < keys.length; i++) {
    newEvent[keys[i]] = options[keys[i]];
  }
  //
  // Add the event to the calendar
  this.calWrapper.fullCalendar("renderEvent", newEvent, true);
  //
  // Build the object used to inform the datamap
  var data = {};
  data[this.columns["title"]] = options.title;
  data[this.columns["start"]] = options.start;
  if (data[this.columns["end"]])
    data [this.columns["end"]] = options.end;
  //
  // Inform the server
  var e = [{obj: this.id, id: "chgData", content: {name: "insert", data: data}}];
  if (!this.addEventTimeout) {
    var pthis = this;
    pthis.addEventTimeout = setTimeout(function () {
      pthis.addEventTimeout = undefined;
      Client.mainFrame.sendEvents(e);
    }, 0);
  }
};


/**
 * Update an event
 * @param {String} id - the id of the event to update
 * @param {Object} options - properties to update
 */
Client.Calendar.prototype.updateEvent = function (id, event)
{
  if (!this.calWrapper)
    return;
  //
  this.calWrapper.fullCalendar("updateEvent", event);
  //
  // Build the object used to inform the datamap
  var data = {};
  data.id = event.id;
  data.title = event.title;
  data.start = event.start._i;
  if (event.end)
    data.end = event.end._i;
  //
  // Inform the server
  var e = [];
  e.push({obj: this.id, id: "chgData", content: {name: "update", data: data, id: id}});
  if (!this.updateTimeout) {
    var pthis = this;
    pthis.updateTimeout = setTimeout(function () {
      pthis.updateTimeout = undefined;
      Client.mainFrame.sendEvents(e);
    }, 0);
  }
};


/**
 * Retrieve events that FullCalendar has in memory.
 * @param {Object} filter - If filter is an ID, all events with the same ID will be returned.
 * filter may also be a filter function that accepts one Event Object argument and returns true if
 * it should be included in the result set.
 * @returns {Array}
 */
Client.Calendar.prototype.searchEvents = function (filter, cbId)
{
  var res = [];
  if (this.calWrapper) {
    if (!filter)
      res = this.calWrapper.fullCalendar("clientEvents");
    else
      res = this.calWrapper.fullCalendar("clientEvents", function (event) {
        if (filter.id && event.id !== filter.id)
          return false;
        //
        if (filter.title && event.title !== filter.title)
          return false;
        //
        if (filter.start && event.start._i !== filter.start)
          return false;
        //
        if (filter.end && event.end._i !== filter.end)
          return false;
        //
        return true;
      });
  }
  //
  var result = [];
  for (var i = 0; i < res.length; i++) {
    var ev = res[i];
    var event = {};
    event.id = ev.id;
    event.title = ev.title;
    event.start = ev.start._i;
    if (ev.end)
      event.end = ev.end._i;
    result.push(event);
  }
  //
  if (cbId) {
    var e = [{obj: this.id, id: "cb", content: {res: result, cbId: cbId}}];
    Client.mainFrame.sendEvents(e);
  }
  else
    return result;
};


/**
 * Switch to a different view
 * @param {String} view - a string representing a way of displaying days and events.
 */
Client.Calendar.prototype.changeView = function (view)
{
  if (!this.calWrapper || !view)
    return;
  //
  this.calWrapper.fullCalendar("changeView", view);
};


/**
 * Show the fullcalendar editor
 */
Client.Calendar.prototype.showEditor = function ()
{
  // Create or show the fullcalendar editor
  this.realizeFullCalendarEditor();
};


/**
 * Realize the fullcalendar editor popup window
 */
Client.Calendar.prototype.realizeFullCalendarEditor = function ()
{
  // Set the body owner with overflow hidden to prevent the scrollbar
  // due to devices bigger than the view editor window view
  var bodyOwner = document.body.parentNode;
  if (bodyOwner)
    bodyOwner.style.overflow = "hidden";
  //
  // Creation of overlay layer container
  this.overlay = document.createElement("div");
  this.overlay.id = "widget-editor-overlay";
  this.overlay.style.top = document.body.scrollTop + "px";
  this.overlay.style.left = document.body.scrollLeft + "px";
  var pthis = this;
  this.overlay.onclick = function (ev) {
    if (pthis.overlay && ev.target.id === pthis.overlay.id) {
      var cancButt = document.getElementById("cancel-button");
      if (cancButt)
        cancButt.onclick();
    }
  };
  //
  // Creation of fullcalendar editor container
  this.fullCalendarEditor = document.createElement("div");
  this.fullCalendarEditor.id = "widget-editor";
  window.onresize = function (ev) {
    if (!pthis.overlay) {
      ev.preventDefault();
      return false;
    }
    //
    var targCont = document.getElementById("widget-options-container");
    if (targCont)
      pthis.resizeContainer(targCont);
  };
  //
  // Creation of title container
  var titCont = document.createElement("div");
  titCont.id = "widget-editor-title";
  titCont.innerHTML = "FullCalendar Settings";
  this.fullCalendarEditor.appendChild(titCont);
  //
  // Creation of button that redirect to the chartjs documentation website
  var helpBut = document.createElement("div");
  helpBut.id = "widget-editor-help";
  helpBut.textContent = "i";
  helpBut.onmouseover = function (ev) {
    pthis.highlightOptions(ev, "rgb(0, 0, 0)");
    document.getElementById(ev.target.id).style.color = "rgb(255, 255, 255)";
  };
  helpBut.onmouseout = function (ev) {
    pthis.highlightOptions(ev, "whitesmoke");
    document.getElementById(ev.target.id).style.color = "rgb(0, 0, 0)";
  };
  helpBut.onclick = function () {
    window.open("http://fullcalendar.io/docs/", "fullcalendardoc");
  };
  titCont.appendChild(helpBut);
  //
  // Creation of buttons container
  var buttCont = document.createElement("div");
  buttCont.id = "widget-editor-buttons";
  //
  // Creation of the confirm button
  var confirmButDiv = document.createElement("div");
  confirmButDiv.id = "confirm-btn-div";
  //
  var confirmBut = document.createElement("button");
  confirmBut.id = "confirm-button";
  confirmBut.textContent = "Set";
  confirmBut.onmouseover = function (ev) {
    pthis.highlightOptions(ev, "#257831");
  };
  confirmBut.onmouseout = function (ev) {
    pthis.highlightOptions(ev, "#37B349");
  };
  confirmBut.onclick = function () {
    if (!pthis.calendarOptions)
      pthis.calendarOptions = {};
    if (!pthis.eventsOptions)
      pthis.eventsOptions = {};
    //
    // Copy the temporary changes into the real options object
    for (var i in pthis.tempCalendarOptions)
      pthis.calendarOptions[i] = pthis.tempCalendarOptions[i];
    //
    for (var i in pthis.tempEventsOptions)
      pthis.eventsOptions[i] = pthis.tempEventsOptions[i];
    //
    pthis.view = pthis.tempView;
    //
    // Send all the changes to the ide
    if (Client.mainFrame.isEditing()) {
      var e = Client.eleMap["editm"];
      if (e) {
        // No need to force changeChartOptions with new options: it will be updated automagically
        e.editProxy.appCmd([{id: pthis.id, c: "changeProp", n: "cst_calendarOptions", v: JSON.stringify(pthis.calendarOptions)}]);
        e.editProxy.appCmd([{id: pthis.id, c: "changeProp", n: "cst_eventsOptions", v: JSON.stringify(pthis.eventsOptions)}]);
        e.editProxy.appCmd([{id: pthis.id, c: "changeProp", n: "cst_view", v: pthis.view}]);
      }
      //
      // Close the fullcalendar editor window
      document.body.removeChild(pthis.overlay);
      pthis.overlay = null;
      pthis.fullCalendarEditor = null;
    }
    //
    pthis.tempCalendarOptions = {};
    pthis.tempEventsOptions = {};
    pthis.tempView = "";
  };
  confirmButDiv.appendChild(confirmBut);
  buttCont.appendChild(confirmButDiv);
  //
  // Creation of button that close the color picker without changes
  var cancelButDiv = document.createElement("div");
  cancelButDiv.id = "cancel-btn-div";
  //
  var cancelBut = document.createElement("button");
  cancelBut.id = "cancel-button";
  cancelBut.textContent = "Cancel";
  cancelBut.onmouseover = function (ev) {
    pthis.highlightOptions(ev, "#434D44");
  };
  cancelBut.onmouseout = function (ev) {
    pthis.highlightOptions(ev, "#9DA69F");
  };
  cancelBut.onclick = function () {
    // Undo changes
    pthis.tempCalendarOptions = {};
    pthis.tempEventsOptions = {};
    pthis.tempView = "";
    //
    // Close the fullcalendar editor window
    document.body.removeChild(pthis.overlay);
    //
    // Restore the original overflow property
    var bodyOwner = document.body.parentNode;
    if (bodyOwner)
      bodyOwner.style.overflow = "";
    //
    pthis.overlay = null;
    pthis.fullCalendarEditor = null;
  };
  cancelButDiv.appendChild(cancelBut);
  buttCont.appendChild(cancelButDiv);
  //
  // Creation of fixed properties container
  var fixedPropCont = document.createElement("div");
  fixedPropCont.id = "widget-editor-tabbar";
  //
  this.fullCalendarEditor.appendChild(fixedPropCont);
  //
  // Creation of the options container
  var optContainer = document.createElement("div");
  optContainer.id = "widget-options-container";
  //
  // Initialization of temporary changes objects
  this.tempCalendarOptions = (!this.calendarOptions) ? {} : this.calendarOptions;
  this.tempEventsOptions = (!this.eventsOptions) ? {} : this.eventsOptions;
  this.tempView = (!this.view) ? Client.Calendar.default.global.defaultView : this.view;
  //
  // Realize all tabs
  this.realizeTabBarLabel("fullcalendar-calendar-options", fixedPropCont, optContainer);
  this.realizeTabBarLabel("fullcalendar-event-options", fixedPropCont, optContainer);
  //
  // If the view is an "agenda" type, then I realize the "agenda options" label
  if (this.tempView === "agendaWeek" || this.tempView === "agendaDay")
    this.realizeTabBarLabel("fullcalendar-agenda-options", fixedPropCont, optContainer);
  //
  // Realize all the options available for the current calendar
  this.realizeOptions(Client.Calendar.default.global, optContainer, "fullcalendar-calendar-options");
  //
  // Append all container to create the fullcalendar editor well formatted
  this.fullCalendarEditor.appendChild(optContainer);
  this.fullCalendarEditor.appendChild(buttCont);
  this.overlay.appendChild(this.fullCalendarEditor);
  document.body.appendChild(this.overlay);
  //
  // Select fullcalendar global options label
  this.selectTab("fullcalendar-calendar-options");
  //
  // Readapt the container
  this.resizeContainer(optContainer);
};


/**
 * Resize an obj container
 * @param {HTMLElement} targetObj - dom object to resize
 */
Client.Calendar.prototype.resizeContainer = function (targetObj)
{
  // Resize editor options container
  var wEditor = document.getElementById("widget-editor");
  var dim = wEditor.offsetHeight - document.getElementById("widget-editor-title").offsetHeight - document.getElementById("widget-editor-buttons").offsetHeight - document.getElementById("widget-editor-tabbar").offsetHeight - 87;
  targetObj.style.height = dim + "px";
  //
  // Resize editor buttons
  var editorWidth = wEditor.offsetWidth;
  var cancBut = document.getElementById("cancel-button");
  var confBut = document.getElementById("confirm-button");
  if (editorWidth < 260) {
    cancBut.style.minWidth = "85px";
    confBut.style.minWidth = "85px";
  }
  else {
    cancBut.style.minWidth = "120px";
    confBut.style.minWidth = "120px";
  }
};


/**
 * Create the options for the current calendar
 * @param {Object} fullcalendarOptions - the object containing the options
 * @param {HTMLElement} targetObj - the dom object to append the options to
 * @param {MouseEvent} ev - the event occured when the user clicked on a tab of the calendar editor
 */
Client.Calendar.prototype.realizeOptions = function (fullcalendarOptions, targetObj, ev)
{
  // If I do not have any target obj in which append the options, I have to exit
  if (!fullcalendarOptions || !targetObj)
    return;
  //
  this.activatedObj = ev;
  if (typeof ev === "object")
    this.activatedObj = (ev.target.id) ? ev.target.id : ev.target.parentNode.id;
  //
  targetObj.innerHTML = "";
  //
  var currSect = "";
  var nextSect = "";
  var sections = {
    header: "General display",
    defaultView: "Views",
    defaultDate: "Current Date",
    lang: "Text/Time Customization",
    selectable: "Selection",
    eventColor: "Event Rendering",
    editable: "Event Dragging & Resizing",
    allDaySlot: "Agenda Options"
  };
  //
  for (var name in fullcalendarOptions) {
    var value = fullcalendarOptions[name];
    //
    // Set groups of properties
    for (var i in sections) {
      if (name.indexOf(i) !== -1) {
        nextSect = sections[i];
        break;
      }
    }
    //
    // If I have custom preferences they must be shown instead of the default values
    var opt = (this.activatedObj.indexOf("-calendar-") !== -1) ? this.tempCalendarOptions : this.tempEventsOptions;
    if (opt && opt[name] !== undefined && opt[name] !== fullcalendarOptions[name])
      value = opt[name];
    //
    // If I have a custom view, then I keep it. Otherwise I keep the default one
    if (this.activatedObj.indexOf("-calendar-") !== -1 && name === "defaultView")
      value = (!this.view) ? Client.Calendar.default.global.defaultView : ((this.view !== this.tempView) ? this.tempView : this.view);
    //
    // Find and set next section title
    if (currSect !== nextSect) {
      var gHeader = document.createElement("div");
      gHeader.className = "widget-section-groupHeader";
      if (currSect !== "")
        gHeader.classList.add("widget-next-section-groupHeader");
      //
      gHeader.innerHTML = nextSect[0].toUpperCase() + nextSect.substr(1);
      targetObj.appendChild(gHeader);
      //
      currSect = nextSect;
    }
    //
    // Create a container for the property label and the property value
    var pContainer = document.createElement("div");
    pContainer.id = "widget-property-container";
    //
    // Label to identify the property name
    var pHeader = document.createElement("label");
    pHeader.id = "widget-property-label";
    pHeader.innerHTML = name;
    pContainer.appendChild(pHeader);
    //
    var pValue;
    var pValueText;
    //
    // If it's a color property
    if (value !== null && (name.toLowerCase().indexOf("color") !== -1)) {
      // For color properties, creates span with a backgroundColor
      pValue = document.createElement("span");
      pValue.id = name + "-color";
      pValue.className = "widget-color-label";
      pValue.style.backgroundColor = (value === undefined) ? "" : value;
      //
      // The click on the color div, creates a color picker popup window
      pValue.onclick = function (ev) {
        var evTarget = ev;
        var targetColorBg = ev.target;
        var targetColorTxt = document.getElementById(ev.target.id.split("-")[0] + "-textcolor");
        var oldColorVal = targetColorTxt.value;
        //
        // Creation of upper overlay container,
        // need to manage the click out of popup container and do things i.e. close popup without changes
        var overlay = document.createElement("div");
        overlay.id = "widget-editor-colorpicker-overlay";
        var ppthis = pthis;
        overlay.onclick = function (ev) {
          if (ev.target.id === overlay.id) {
            var cancButt = document.getElementById("cancel-button-colorpicker");
            if (cancButt)
              cancButt.onclick();
          }
        };
        //
        // Creation of popup container in which append the color picker and its controls
        var popup = document.createElement("div");
        popup.id = "widget-editor-colorpicker-popup";
        //
        // Creation of buttons container
        var buttCont = document.createElement("div");
        buttCont.id = "widget-editor-colorpicker-buttons";
        //
        // Creation of button that set the color into the ide property
        var confirmBut = document.createElement("div");
        confirmBut.id = "confirm-button-colorpicker";
        confirmBut.textContent = "Set color";
        confirmBut.onmouseover = function (ev) {
          ppthis.highlightOptions(ev, "rgb(235, 248, 233)");
        };
        confirmBut.onmouseout = function (ev) {
          ppthis.highlightOptions(ev, "rgb(255, 255, 255)");
        };
        confirmBut.onclick = function (ev) {
          ppthis.saveTemporaryDataOption(evTarget);
          ppthis.overlay.removeChild(overlay);
        };
        buttCont.appendChild(confirmBut);
        //
        // Creation of button that close the color picker without changes
        var cancelBut = document.createElement("div");
        cancelBut.id = "cancel-button-colorpicker";
        cancelBut.textContent = "Cancel";
        cancelBut.onmouseover = function (ev) {
          ppthis.highlightOptions(ev, "rgb(254, 215, 222)");
        };
        cancelBut.onmouseout = function (ev) {
          ppthis.highlightOptions(ev, "rgb(255, 255, 255)");
        };
        cancelBut.onclick = function (ev) {
          targetColorBg.style.backgroundColor = oldColorVal;
          targetColorTxt.value = oldColorVal;
          //
          ppthis.overlay.removeChild(overlay);
        };
        buttCont.appendChild(cancelBut);
        //
        // Append the popup to the upper overlay
        overlay.appendChild(popup);
        //
        // Append the upper overlayto the lower overlay
        pthis.overlay.appendChild(overlay);
        //
        // Create and append to the fullcalendar editor the color picker
        var cp = new Client.ColorPicker({domObj: evTarget.target, id: targetColorBg.id, value: oldColorVal, type: "basic"}, {domObj: popup});
        cp.domObj.classList.add("widget-colorpicker");
        //
        // Append the button container to the popup container
        popup.appendChild(buttCont);
        //
        // Subscribe this color picker to the next changes (ev contains the Color obj)
        UIColorPicker.subscribe(cp.domObj.id, function (ev) {
          targetColorTxt.value = cp.getColor("rgba", ev);
          targetColorBg.style.backgroundColor = cp.getColor("rgba", ev);
        });
        //
        // Reposition the popup panel
        popup.style.left = ((document.body.offsetWidth - popup.offsetWidth) / 2) + "px";
        if (popup.offsetHeight + targetColorBg.offsetHeight + ev.clientY - ev.offsetY > document.body.scrollHeight)
          popup.style.top = (ev.clientY - ev.offsetY - popup.offsetHeight - 4) + "px";
        else
          popup.style.top = (ev.clientY - ev.offsetY + targetColorBg.offsetHeight + 4) + "px";
      };
      pContainer.appendChild(pValue);
      //
      pValueText = document.createElement("input");
      pValueText.id = name + "-textcolor";
      pValueText.className = "widget-property-textvalue";
      pValueText.value = (value === undefined) ? "" : value;
      pValueText.oninput = function (ev) {
        var c = document.getElementById(ev.target.id.split("-")[0] + "-color");
        if (c)
          c.style.backgroundColor = ev.target.value;
        //
        pthis.saveTemporaryDataOption(ev);
      };
      //
      pContainer.appendChild(pValueText);
    }
    else if (value !== null && typeof value === "object") {
      // Create the container of obj properties
      var pObjCont = document.createElement("div");
      pObjCont.id = name + "-container";
      pObjCont.style.width = "60%";
      //
      var k = Object.keys(value);
      var c = 0;
      for (var n in value) {
        // For each sub property belonging to the obj, create a div
        pValue = document.createElement("div");
        pValue.id = name + "-" + n;
        pValue.className = "fullcalendar-subproperty-name";
        pValue.innerHTML = (n === undefined) ? "" : n;
        //
        pObjCont.appendChild(pValue);
        //
        // Create the input with the sub property
        pValueText = document.createElement("input");
        pValueText.id = name + "-" + n + "-textvalueFromObj";
        pValueText.className = "fullcalendar-subproperty-textvalue";
        pValueText.type = "text";
        //
        if (c !== k.length - 1)
          pValueText.classList.add("fullcalendar-subproperty-notlast-textvalue");
        //
        pValueText.value = (value[n] === undefined) ? "" : value[n];
        pValueText.oninput = function (ev) {
          pthis.saveTemporaryDataOption(ev);
        };
        //
        pObjCont.appendChild(pValueText);
        c++;
      }
      pContainer.appendChild(pObjCont);
    }
    else {
      pValue = document.createElement("input");
      pValue.className = "widget-input-field";
      //
      if (value !== null && (value === true || value === false)) {
        // For boolean properties, creates a checkbox
        pValue.type = "checkbox";
        pValue.classList.add("widget-checkbox");
        if (value === true)
          pValue.setAttribute("checked", "checked");
      }
      else {
        // For text properties, creates an input text
        pValue.type = "text";
        pValue.classList.add("widget-text");
        pValue.style.textAlign = isNaN((value === undefined) ? "" : value) ? "left" : "right";
        pValue.value = (value === null ? "" : value);
      }
      pContainer.appendChild(pValue);
    }
    if (!pValue.id)
      pValue.id = name;
    //
    // Attach events to the value obj
    var pthis = this;
    if (pValue.type !== "text") {
      pValue.onchange = function (ev) {
        pthis.saveTemporaryDataOption(ev);
      };
    }
    else { // Attach onInput event to the input text
      pValue.oninput = function (ev) {
        pthis.saveTemporaryDataOption(ev);
        //
        // If the input field is the "defaultView" option,
        // then handle the visibility of Agenda label and its options based on its value
        if (ev.target.id === "defaultView")
          pthis.handleAgendaOptions();
      };
    }
    //
    // Append any property to the target obj container
    targetObj.appendChild(pContainer);
  }
};


/**
 * Handle the visibility of Agenda label and its options
 */
Client.Calendar.prototype.handleAgendaOptions = function ()
{
  var targetObj = document.getElementById("widget-editor-tabbar");
  var optionsObj = document.getElementById("widget-options-container");
  if ((this.tempView === "agendaDay" || this.tempView === "agendaWeek") && !document.getElementById("fullcalendar-agenda-options")) {
    if (targetObj && optionsObj)
      this.realizeTabBarLabel("fullcalendar-agenda-options", targetObj, optionsObj);
  }
  if (this.tempView !== "agendaDay" && this.tempView !== "agendaWeek" && document.getElementById("fullcalendar-agenda-options"))
    this.unrealizeAgendaTab(document.getElementById("fullcalendar-agenda-options"));
};


/**
 * Unrealize Agenda label from tabbar
 * @param {HTMLElement} contObj - the tab from which remove the label
 */
Client.Calendar.prototype.unrealizeAgendaTab = function (contObj)
{
  var agLabel = document.getElementById("fullcalendar-agenda-options");
  if (contObj && contObj.parentNode && agLabel)
    contObj.parentNode.removeChild(agLabel);
};


/**
 * Manage the temporary changes of fullcalendar properties
 * @param {Event} ev - the event occured when the user changed a property
 */
Client.Calendar.prototype.saveTemporaryDataOption = function (ev)
{
  var val;
  var nam;
  var subName;
  //
  // If an input was modified
  if (ev.type === "input") {
    // Get the name of the modified property
    nam = ev.target.id.split("-")[0];
    //
    // If the property is an object
    if (ev.target.id.indexOf("textvalueFromObj") !== -1) {
      // If it's a calendar property (and not an event property)
      if (this.activatedObj.indexOf("-calendar-") !== -1) {
        // Get the current property
        var obj = Client.Calendar.default.global[nam];
        for (var i in this.tempCalendarOptions) {
          if (obj[i])
            obj[i] = this.tempCalendarOptions[i];
        }
      }
      //
      // Get the name of the property inside the object property
      subName = ev.target.id.split("-")[1];
      //
      // Set the new value
      obj[subName] = ev.target.value;
      val = JSON.stringify(obj);
    }
    else
      val = ev.target.value;
  }
  else if (ev.type === "change" || ev.type === "click") {
    // Get the name of the modified property
    nam = ev.target.id.split("-")[0];
    //
    // If a checkbox was modified
    if (ev.target.type === "checkbox")
      val = ev.target.checked;
    else {
      var docCol = document.getElementById(ev.target.id.split("-")[0] + "-textcolor");
      //
      // If a color property was modified
      if (docCol)
        val = docCol.value;
      else
        val = ev.target.style.backgroundColor;
    }
  }
  //
  // Save changes into the temporary options object
  if (this.activatedObj.indexOf("-calendar-") !== -1) {
    if (nam !== "defaultView") {
      if (!this.tempCalendarOptions)
        this.tempCalendarOptions = {};
      if (typeof Client.Calendar.default.global[nam] === "object") {
        if (!this.tempCalendarOptions[nam])
          this.tempCalendarOptions[nam] = {};
        this.tempCalendarOptions[nam] = JSON.parse(val);
      }
      else
        this.tempCalendarOptions[nam] = val;
    }
    else
      this.tempView = val;
  }
  else {
    if (!this.tempEventsOptions)
      this.tempEventsOptions = {};
    this.tempEventsOptions[nam] = val;
  }
};


/**
 * Realize the label of the tabbar into the char editor
 * @param {String} sectionId - id of macro group of properties
 * @param {Object} targetObj - the dom object to append the options to
 * @param {Object} optionsObj - dom object used by "realizeOptions" for any tab section
 */
Client.Calendar.prototype.realizeTabBarLabel = function (sectionId, targetObj, optionsObj)
{
  // Create the tab containing the options
  var tabbed = document.createElement("div");
  tabbed.id = sectionId;
  tabbed.className = "fullcalendar-tab";
  //
  switch (sectionId) {
    case "fullcalendar-calendar-options":
      // Show the calendar options
      tabbed.textContent = "Calendar options";
      var pthis = this;
      tabbed.onclick = function (ev) {
        pthis.selectTab(ev.target.id);
        //
        // Realize all the options available for the current fullcalendar
        pthis.realizeOptions(Client.Calendar.default.global, optionsObj, ev);
      };
      //
      targetObj.appendChild(tabbed);
      break;
    case "fullcalendar-event-options":
      // Show the events options
      tabbed.textContent = "Event options";
      var pthis = this;
      tabbed.onclick = function (ev) {
        pthis.selectTab(ev.target.id);
        //
        // Realize all the options available for the current fullcalendar
        pthis.realizeOptions(Client.Calendar.default.event, optionsObj, ev);
      };
      //
      targetObj.appendChild(tabbed);
      break;
    case "fullcalendar-agenda-options":
      // Show the calendar options when the calendar is in agenda mode
      tabbed.textContent = "Agenda options";
      var pthis = this;
      tabbed.onclick = function (ev) {
        pthis.selectTab(ev.target.id);
        //
        // Realize all the options available for the current fullcalendar
        pthis.realizeOptions(Client.Calendar.default.agenda, optionsObj, ev);
      };
      //
      targetObj.appendChild(tabbed);
      break;
  }
};


/**
 * Select clicked tab
 * @param {String} tabId - the id of the selected tab
 */
Client.Calendar.prototype.selectTab = function (tabId)
{
  if (!tabId)
    return;
  //
  var tabbedFound;
  var tabBarDefaultLabels = ["fullcalendar-calendar-options", "fullcalendar-event-options", "fullcalendar-agenda-options"];
  for (var i = 0; i < tabBarDefaultLabels.length; i++) {
    var label = document.getElementById(tabBarDefaultLabels[i]);
    if (label) {
      label.style.backgroundColor = "whitesmoke";
      if (tabBarDefaultLabels[i] === tabId)
        tabbedFound = tabId;
    }
  }
  if (tabbedFound)
    document.getElementById(tabbedFound).style.backgroundColor = "#FFFFFF";
};


/**
 * Highlight or not a ev.target object
 * @param {obj} ev - mouse event
 * @param {string} color
 */
Client.Calendar.prototype.highlightOptions = function (ev, color)
{
  var tg = document.getElementById(ev.target.id);
  if (!tg)
    tg = document.getElementById(ev.target.parentNode.id);
  if (tg)
    tg.style.backgroundColor = color;
};


/**
 * Return true if two events have the same id, title, start date and end date
 * @param {String|Moment} event1
 * @param {String|Moment} event2
 * @returns {Boolean}
 */
Client.Calendar.prototype.sameEvent = function (event1, event2)
{
  if (event1.id !== event2.id)
    return false;
  //
  if (event1.title !== event2.id)
    return false;
  //
  var start1 = moment(event1.start);
  var start2 = moment(event2.start);
  if (!moment(start1).isSame(start2))
    return false;
  var end1 = moment(event1.end);
  var end2 = moment(event2.end);
  if (!moment(end1).isSame(end2))
    return false;
  //
  return true;
};


/**
 * Resize the element when the document view is resized
 * @param {Event} ev - the event occured when the document view was resized
 */
Client.Calendar.prototype.onResize = function (ev)
{
  if (this.percDim) {
    if (this.percDim.height)
      this.changeDimension("height", this.percDim.height);
    if (this.percDim.width)
      this.changeDimension("width", this.percDim.width);
  }
  //
  Client.Element.prototype.onResize.call(this, ev);
};
