/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/**
 * @class Color Picker
 * @param {Object} element
 * @param {View|Element} parent
 * @param {View|Element} view
 * */
Client.ColorPicker = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  this.domObj = document.createElement("div");
  if (!this.id)
    this.id = element.id;
  this.domObj.id = this.id;
  this.domObj.className = "ui-color-picker";
  this.domObj.setAttribute("data-topic", this.domObj.id);
  this.domObj.setAttribute("data-mode", "HSL");
  //
  parent.domObj.appendChild(this.domObj);
  //
  this.updateElement(element);
  this.attachEvents(element.events);
};


// Make Client.ColorPicker extend Client.Element
Client.ColorPicker.prototype = new Client.Element();


/**
 * Update an element
 * @param {Object} el - Element representation
 */
Client.ColorPicker.prototype.updateElement = function (el)
{
  if (el.type === undefined && this.picker === undefined)
    this.createPicker();
  if (el.type) {
    this.createPicker(el.type);
    delete el.type;
  }
  //
  if (el.format === undefined)
    el.format = "HSL";
  if (el.format) {
    var c = new UIColorPicker.Color();
    c.setFormat(el.format);
    delete el.format;
  }
  //
  // If a color value is given, decode and set the color to the picker
  if (el.value) {
    if (typeof el.value === "string") {
      if (el.value[0] === "#") {
        var v = el.value.substr(1);
        if (v.length === 6)
          this.picker.setColor(UIColorPicker.RGBColor(parseInt(v.substring(0, 2), 16), parseInt(v.substring(2, 4), 16), parseInt(v.substring(4, 6), 16)));
        else if (v.length === 3)
          this.picker.setColor(UIColorPicker.RGBColor(parseInt(v.substring(0, 1) + v.substring(0, 1), 16), parseInt(v.substring(1, 2) + v.substring(1, 2), 16), parseInt(v.substring(2, 3) + v.substring(2, 3), 16)));
      }
      else if (el.value.substr(0, 3) === "rgb") {
        var rgba = el.value.substring(el.value.indexOf("(") + 1, el.value.indexOf(")")).split(",");
        if (rgba.length > 3)
          this.picker.setColor(UIColorPicker.RGBAColor(parseInt(rgba[0]), parseInt(rgba[1]), parseInt(rgba[2]), parseFloat(rgba[3])));
        else
          this.picker.setColor(UIColorPicker.RGBColor(parseInt(rgba[0]), parseInt(rgba[1]), parseInt(rgba[2])));
      }
      else { // any valid string color
        var v = this.getHexFromStringColor(el.value).substr(1);
        if (v.length === 6)
          this.picker.setColor(UIColorPicker.RGBColor(parseInt(v.substring(0, 2), 16), parseInt(v.substring(2, 4), 16), parseInt(v.substring(4, 6), 16)));
      }
    }
    //
    delete el.value;
  }
  //
  Client.Element.prototype.updateElement.call(this, el);
};


/**
 * Attach events to the color picker obj
 * @param {Object} events
 */
Client.ColorPicker.prototype.attachEvents = function (events)
{
  if (!events)
    return;
  //
  Client.Element.prototype.attachEvents.call(this, events);
  //
  var pthis = this;
  UIColorPicker.subscribe(this.domObj.id, function (ev) {
    var e = [{obj: pthis.id, id: "chgProp", content: {name: "value", value: {hex: pthis.getColor("hex", ev), rgba: pthis.getColor("rgba", ev)}, clid: Client.id}}];
    if (events.indexOf("onChange") >= 0)
      e.push({obj: pthis.id, id: "onChange", content: pthis.saveEvent(ev)});
    Client.mainFrame.sendEvents(e);
  });
};


/**
 * Get the selected color
 * @param {String} format
 * @param {Object} color
 */
Client.ColorPicker.prototype.getColor = function (format, color)
{
  var f = (format ? format : "rgba");
  var c = (color ? color : new UIColorPicker.Color());
  if (!color)
    c.getColor(this.picker.id);
  //
  if (f === "rgba")
    return c.getRGBA();
  else if (f === "hex")
    return c.getHexa();
};


/**
 * Convert a color string into hex format color
 * @param {String} colorStr
 * @returns {Boolean|String}
 */
Client.ColorPicker.prototype.getHexFromStringColor = function (colorStr)
{
  var a = document.createElement('div');
  a.style.color = colorStr;
  var colors = window.getComputedStyle(document.body.appendChild(a)).color.match(/\d+/g).map(function (a) {
    return parseInt(a, 10);
  });
  document.body.removeChild(a);
  return (colors.length >= 3) ? '#' + (((1 << 24) + (colors[0] << 16) + (colors[1] << 8) + colors[2]).toString(16).substr(1)) : false;
};


/**
 * Create palette container
 */
Client.ColorPicker.prototype.createPalette = function ()
{
  var palette = document.createElement("div");
  palette.id = "palette";
  palette.className = "block";
  //
  var colPalette = document.createElement("div");
  colPalette.id = "color-palette";
  var colInfo = document.createElement("div");
  colInfo.id = "color-info";
  //
  palette.appendChild(colPalette);
  palette.appendChild(colInfo);
  //
  return palette;
};


/**
 * Create picker container
 * @param {String} type - can be empty
 */
Client.ColorPicker.prototype.createPicker = function (type)
{
  this.picker = new UIColorPicker.ColorPicker(this.domObj);
  this.picker.setColor(UIColorPicker.RGBColor(193, 66, 66));
  if (type === "full") {
    // Full feature picker
    this.domObj.appendChild(this.createPickerControls());
    this.domObj.appendChild(this.createPickerVoidSample());
    this.domObj.appendChild(this.createPickerSamples());
    this.domObj.appendChild(this.createPalette());
    //
    ColorPickerTool.init();
  }
};


/**
 * Create picker samples container
 */
Client.ColorPicker.prototype.createPickerSamples = function ()
{
  var pSamples = document.createElement("div");
  pSamples.id = "picker-samples";
  pSamples.setAttribute("sample-id", "master");
  //
  return pSamples;
};


/**
 * Create picker void sample container
 */
Client.ColorPicker.prototype.createPickerVoidSample = function ()
{
  var pVoidSample = document.createElement("div");
  pVoidSample.id = "void-sample";
  pVoidSample.className = "icon";
  //
  return pVoidSample;
};


/**
 * Create picker controls container
 */
Client.ColorPicker.prototype.createPickerControls = function ()
{
  var pControls = document.createElement("div");
  pControls.id = "controls";
  var pDelete = document.createElement("div");
  pDelete.id = "delete";
  var pTrash = document.createElement("div");
  pTrash.id = "trash-can";
  //
  pDelete.appendChild(pTrash);
  pControls.appendChild(pDelete);
  //
  return pControls;
};
