/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global Client, google */

/**
 * @class Google Maps
 * @param {Object} element
 * @param {View|Element} parent
 * @param {View} view
 * @extends Client.Element
 * */
Client.Gmap = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  this.domObj = document.createElement("div");
  this.domObj.id = btoa(element.id);
  parent.appendChildObject(this, this.domObj);
  //
  this.initMap(element);
  //
  // the following properties were already used by initMap
  delete element.options;
  delete element.zoom;
  delete element.mapTypeId;
  delete element.center;
  //
  this.updateElement(element);
  this.attachEvents(element.events);
  setTimeout(function () {
    this.resizeMap();
  }.bind(this), 200);
};


// Make Client.Gmap extend Client.Container
Client.Gmap.prototype = new Client.Element();


/**
 * Update an element
 * @param {Object} el - Element representation
 */
Client.Gmap.prototype.updateElement = function (el)
{
  this.purgeMyProp(el);
  var pthis = this;
  //
  if (this.map) {
    //
    // If a new map type is given, set the new map type
    try {
      if (el.type) {
        this.map.setMapTypeId(google.maps.MapTypeId[el.type.toUpperCase()]);
        this.typeChangeTime = new Date();
        delete el.type;
      }
      //
      // If a new zoom is given, set the new zoom to the map
      if (el.zoom) {
        this.map.setZoom(parseInt(el.zoom));
        this.zoomChangeTime = new Date();
        delete el.zoom;
      }
      //
      // If a new center is given, set the new center to the map
      if (el.center) {
        var nc = this.toGmapPosition(el.center);
        var ac = this.map.getCenter();
        if (!ac || !(nc instanceof google.maps.LatLng) ||
                ac.lat() !== nc.lat() || ac.lng() !== nc.lng()) {
          this.centerChangeTime = new Date();
          this.map.setCenter(nc);
        }
        delete el.center;
      }
      //
      if (el.visible)
        this.resizeMap();
      //
      if (el.address) {
        this.decodeAddress(el.address, function (results, status) {
          if (status === google.maps.GeocoderStatus.OK)
            pthis.map.setCenter(results[0].geometry.location);
        });
        delete el.address;
      }
      //
      if (el.options) {
        this.map.setOptions(el.options);
        this.zoomChangeTime = new Date();
        this.typeChangeTime = new Date();
        this.centerChangeTime = new Date();
      }
    }
    catch (ex) {
      this.notifyError("updateElement", ex);
    }
  }
  //
  Client.Container.prototype.updateElement.call(this, el);
};


/**
 * Convert the object into a gmap position
 * @param {object} center - Element representation
 */
Client.Gmap.prototype.toGmapPosition = function (center)
{
  if (!center)
    return new google.maps.LatLng(0, 0);
  //
  if (center instanceof google.maps.LatLng)
    return center;
  //
  var nc;
  if (typeof center === "string") {
    try {
      nc = JSON.parse(center);
    }
    catch (ex) {
      nc = center;
    }
  }
  else if (center.latLng)
    nc = new google.maps.LatLng(center.latLng.lat, center.latLng.lng);
  else if (center.lat)
    nc = new google.maps.LatLng(center.lat, center.lng);
  //
  return nc;
};


/**
 * Init the map with default options
 * @param {Object} el - Element representation
 */
Client.Gmap.prototype.initMap = function (el)
{
  var mapOptions = el.options ? JSON.parse(JSON.stringify(el.options)) : {};
  mapOptions.zoom = (el.zoom !== undefined) ? parseInt(el.zoom) : 8;
  mapOptions.mapTypeId = (el.type !== undefined) ? el.type : google.maps.MapTypeId.ROADMAP;
  mapOptions.center = this.toGmapPosition(el.center);
  if (!mapOptions.gestureHandling)
    mapOptions.gestureHandling = "greedy";
  //
  try {
    this.map = new google.maps.Map(document.getElementById(this.domObj.id), mapOptions);
  }
  catch (ex) {
    // Wait for attachevent
    setTimeout(function () {
      this.notifyError("initMap", ex);
    }.bind(this), 100);
  }
};


/**
 * Decode an address string into a location object
 * @param {Object} address
 */
Client.Gmap.prototype.decodeAddress = function (address, cb)
{
  try {
    if (this.geocoder === undefined)
      this.geocoder = new google.maps.Geocoder();
    //
    if (address && typeof address === "string")
      this.geocoder.geocode({"address": address}, cb);
    else
      cb();
  }
  catch (ex) {
    this.notifyError("decodeAddress", ex);
    if (cb)
      cb();
  }
};


/**
 * Set a marker
 * @param {String} id - marker unique identifier. If empty a random ID will be generated
 * @param {Object} pos - location of the marker
 * @param {String|Object} options
 */
Client.Gmap.prototype.addMarker = function (id, pos, options)
{
  try {
    var nc = this.toGmapPosition(pos);
    if (nc === undefined)
      return;
    //
    // Could be an address
    var pthis = this;
    if (typeof nc === "string") {
      this.decodeAddress(nc, function (results, status) {
        if (status === google.maps.GeocoderStatus.OK)
          pthis.addMarker(id, results[0].geometry.location, options);
      });
      return;
    }
    //
    if (id === "")
      id = "M" + Math.random();
    //
    if (this.markerList === undefined)
      this.markerList = [];
    //
    if (!options)
      options = {};
    if (typeof options === "string")
      options = {title: options};
    options.id = id;
    options.position = nc;
    options.map = this.map;
    if (options.animation === "drop")
      options.animation = google.maps.Animation.DROP;
    if (options.animation === "bounce")
      options.animation = google.maps.Animation.BOUNCE;
    //
    // Creating the infowindow
    var infowindow;
    if (options.infowindow) {
      infowindow = new google.maps.InfoWindow(options.infowindow);
      delete options.infowindow;
    }
    //
    // Create a new marker at the passed position
    var m = new google.maps.Marker(options);
    m.setMap(this.map);
    //
    if (infowindow) {
      m.infowindow = infowindow;
      m.addListener("click", function () {
        m.infowindow.open(this.map, m);
      });
    }
    //
    this.markerList.push(m);
    //
    this.attachMarkerEvents(m);
  }
  catch (ex) {
    this.notifyError("addMarker", ex);
  }
};


/**
 * Delete all marker
 */
Client.Gmap.prototype.deleteMarkers = function ()
{
  if (this.markerList === undefined)
    return;
  //
  // Cycle between any marker and unset them from the map
  for (var i = 0; i < this.markerList.length; i++)
    this.markerList[i].setMap();
  //
  // Remove all references of markers from the array
  this.markerList = [];
};


/**
 * Remove selected marker
 * @param {String} id - id of the marker to remove
 */
Client.Gmap.prototype.removeMarker = function (id)
{
  if (!this.markerList)
    return;
  for (var i = 0; i < this.markerList.length; i++) {
    if (this.markerList[i].id === id) {
      this.markerList[i].setMap();
      this.markerList.splice(i, 1);
      break;
    }
  }
};


/**
 * Change marker animation
 * @param {String} id - id of the marker to animate
 * @param {String} animation - animation for a marker
 * @param {String} timeout - reset animation after timeout is expired
 */
Client.Gmap.prototype.setMarkerAnimation = function (id, animation, timeout)
{
  if (animation === "drop")
    animation = google.maps.Animation.DROP;
  if (animation === "bounce")
    animation = google.maps.Animation.BOUNCE;
  //
  for (var i = 0; i < this.markerList.length; i++) {
    var m = this.markerList[i];
    //
    if (m.id === id) {
      m.setAnimation(animation);
      if (timeout) {
        // we can set a reset animation timeout within this loop only because we are about to exit
        setTimeout(function () {
          m.setAnimation();
        }, timeout); // jshint ignore:line
      }
      break;
    }
  }
};


/**
 * Change marker icon
 * @param {String} id - id of the marker to animate
 * @param {String} icon - icon of the marker
 */
Client.Gmap.prototype.setMarkerIcon = function (id, icon)
{
  for (var i = 0; i < this.markerList.length; i++) {
    if (this.markerList[i].id === id) {
      this.markerList[i].setIcon(icon);
      break;
    }
  }
};


/**
 * Attach custom events
 * @param {Object} events
 */
Client.Gmap.prototype.attachEvents = function (events)
{
  var pthis = this;
  if (this.map) {
    if (events.indexOf("onError") >= 0)
      this.sendError = true;
    //
    try {
      // ON CLICK
      if (events.indexOf("onClick") >= 0)
        google.maps.event.addListener(this.map, "click", function (ev) {
          var e = [{obj: pthis.id, id: "onClick", content: {lat: ev.latLng.lat(), lng: ev.latLng.lng()}}];
          Client.mainFrame.sendEvents(e);
        });
      //
      // ON CHANGE (center)
      google.maps.event.addListener(this.map, "center_changed", function (ev) {
        if (!pthis.centerChangeTime || new Date() - pthis.centerChangeTime > 100) {
          var v = {lat: pthis.map.getCenter().lat(), lng: pthis.map.getCenter().lng()};
          var e = [{obj: pthis.id, id: "chgProp", content: {name: "center", value: v, clid: Client.id}}];
          if (events.indexOf("onChange") >= 0)
            e.push({obj: pthis.id, id: "onChange", content: pthis.saveEvent(ev)});
          Client.mainFrame.sendEvents(e);
        }
      });
      //
      // ON CHANGE (zoom)
      google.maps.event.addListener(this.map, "zoom_changed", function (ev) {
        if (!pthis.zoomChangeTime || new Date() - pthis.zoomChangeTime > 100) {
          var e = [{obj: pthis.id, id: "chgProp", content: {name: "zoom", value: pthis.map.getZoom(), clid: Client.id}}];
          if (events.indexOf("onChange") >= 0)
            e.push({obj: pthis.id, id: "onChange", content: pthis.saveEvent(ev)});
          Client.mainFrame.sendEvents(e);
        }
      });
      //
      // ON CHANGE (map type)
      google.maps.event.addListener(this.map, "maptypeid_changed", function (ev) {
        if (!pthis.typeChangeTime || new Date() - pthis.typeChangeTime > 100) {
          var e = [{obj: pthis.id, id: "chgProp", content: {name: "type", value: pthis.map.getMapTypeId(), clid: Client.id}}];
          if (events.indexOf("onChange") >= 0)
            e.push({obj: pthis.id, id: "onChange", content: pthis.saveEvent(ev)});
          Client.mainFrame.sendEvents(e);
        }
      });
    }
    catch (ex) {
      this.notifyError("attachEvents", ex);
    }
  }
};


/**
 * Attach events to new markers
 * @param {Marker} marker
 */
Client.Gmap.prototype.attachMarkerEvents = function (marker)
{
  var pthis = this;
  google.maps.event.addListener(marker, "click", function () {
    var e = [{obj: pthis.id, id: "onClickMarker", content: marker.id}];
    Client.mainFrame.sendEvents(e);
  });
};


/**
 * Execute a nearby search on a location
 * @param {object} request - request to check
 * @param {string} cbId - callback id
 */
Client.Gmap.prototype.nearbySearch = function (request, cbId)
{
  try {
    var pthis = this;
    //
    if (request.location)
      request.location = this.toGmapPosition(request.location);
    //
    if (!this.PlacesService)
      this.PlacesService = new google.maps.places.PlacesService(this.map);
    //
    this.PlacesService.nearbySearch(request, function (results, status) {
      var e = [{obj: pthis.id, id: "cb", content: {res: results, cbId: cbId}}];
      Client.mainFrame.sendEvents(e);
    });
  }
  catch (ex) {
    this.notifyError("nearbySearch", ex);
  }
};


/**
 * Execute reverse geocode of the given location
 * @param {object} location
 * @param {string} cbId - callback id
 */
Client.Gmap.prototype.geocodeLocation = function (location, cbId)
{
  try {
    var pthis = this;
    //
    location = this.toGmapPosition(location);
    var prop = typeof location === "string" ? "address" : "location";
    var opt = {};
    opt[prop] = location;
    //
    if (!this.geocoder)
      this.geocoder = new google.maps.Geocoder();
    //
    this.geocoder.geocode(opt, function (results, status) {
      results = JSON.parse(JSON.stringify(results));    // "Adapt" result before sending it to server
      var e = [{obj: pthis.id, id: "cb", content: {res: results, cbId: cbId}}];
      Client.mainFrame.sendEvents(e);
    });
  }
  catch (ex) {
    this.notifyError("geocodeLocation", ex);
  }
};


/**
 * Add or set a polyline on the map
 * @param {object} polyLine
 */
Client.Gmap.prototype.polyLine = function (polyLine)
{
  try {
    if (!this.polyLines)
      this.polyLines = {};
    //
    var id = polyLine.id;
    delete polyLine.id;
    if (!id)
      return;
    //
    var setBounds = polyLine.setBounds;
    delete polyLine.setBounds;
    //
    var p = this.polyLines[id];
    if (!p) {
      if (polyLine.path && polyLine.path.length) {
        p = new google.maps.Polyline(polyLine);
        this.polyLines[id] = p;
        p.setMap(this.map);
      }
    }
    else {
      if (polyLine.path && polyLine.path.length)
        p.setOptions(polyLine);
      else
        p.setMap(null);
    }
    //
    if (setBounds && polyLine.path && polyLine.path.length) {
      var bounds = new google.maps.LatLngBounds();
      var min = 0;
      var max = polyLine.path.length - 1;
      if (setBounds.length === 2) {
        min = Math.max(setBounds[0], min);
        max = Math.min(setBounds[1], max);
      }
      //
      for (var i = min; i <= max; i++) {
        bounds.extend(this.toGmapPosition(polyLine.path[i]));
      }
      //
      this.map.fitBounds(bounds);
    }
  }
  catch (ex) {
    this.notifyError("polyLine", ex);
  }
};


/**
 * gets directions and show on map on request
 * @param {object} request - route to find
 * @param {string} cbId - callback id
 */
Client.Gmap.prototype.getDirections = function (request, cbId)
{
  try {
    var pthis = this;
    var show = request.show;
    delete request.show;
    //
    // If none is passed, hide the previous directions
    if (!request.origin && !show) {
      if (pthis.directionsDisplay)
        pthis.directionsDisplay.setMap(null);
      var e = [{obj: pthis.id, id: "cb", content: {res: {}, cbId: cbId}}];
      Client.mainFrame.sendEvents(e);
      return;
    }
    //
    // Converting coordinates
    if (request.origin)
      request.origin = this.toGmapPosition(request.origin);
    if (request.destination)
      request.destination = this.toGmapPosition(request.destination);
    if (request.waypoints && request.waypoints.length) {
      for (var i = 0; i < request.waypoints.length; i++) {
        if (request.waypoints[i].location)
          request.waypoints[i].location = this.toGmapPosition(request.waypoints[i].location);
      }
    }
    if (!request.travelMode)
      request.travelMode = google.maps.TravelMode.DRIVING;
    //
    if (!this.DirectionsService)
      this.DirectionsService = new google.maps.DirectionsService(this.map);
    //
    this.DirectionsService.route(request, function (result, status) {
      var e = [{obj: pthis.id, id: "cb", content: {res: result, cbId: cbId}}];
      Client.mainFrame.sendEvents(e);
      if (show) {
        if (!pthis.directionsDisplay)
          pthis.directionsDisplay = new google.maps.DirectionsRenderer();
        pthis.directionsDisplay.setMap(pthis.map);
        pthis.directionsDisplay.setDirections(result);
        if (typeof show === "object")
          pthis.directionsDisplay.setOptions(show);
      }
      else {
        if (pthis.directionsDisplay)
          pthis.directionsDisplay.setMap(null);
      }
    });
  }
  catch (ex) {
    this.notifyError("getDirections", ex);
  }
};


/**
 * Resize the element when the document view is resized
 * @param {Event} ev - the event occured when the document view was resized
 */
Client.Gmap.prototype.onResize = function (ev)
{
  this.resizeMap();
  Client.Element.prototype.onResize.call(this, ev);
};

Client.Gmap.prototype.resizeMap = function ()
{
  if (!this.map)
    return;
  //
  var pthis = this;
  if (!this.resizeTimerID) {
    this.resizeTimerID = setTimeout(function () {
      try {
        var x = pthis.map.getZoom();
        var c = pthis.map.getCenter();
        google.maps.event.trigger(pthis.map, 'resize');
        pthis.map.setZoom(x);
        pthis.map.setCenter(c);
      }
      catch (ex) {
        pthis.notifyError("resizeMap", ex);
      }
      //
      // enable resizing again only after a while to let map events fire.
      setTimeout(function () {
        pthis.resizeTimerID = undefined;
      }, 100);
    }, 0);
  }
};


/**
 * Execute a place autocomplete prediction
 * @param {object} request - request
 * @param {string} cbId - callback id
 */
Client.Gmap.prototype.getPlacePredictions = function (request, cbId)
{
  try {
    var pthis = this;
    //
    if (request.location)
      request.location = this.toGmapPosition(request.location);
    //
    if (!this.AutocompleteService)
      this.AutocompleteService = new google.maps.places.AutocompleteService();
    //
    this.AutocompleteService.getPlacePredictions(request, function (results, status) {
      var e = [{obj: pthis.id, id: "cb", content: {res: results, cbId: cbId}}];
      Client.mainFrame.sendEvents(e);
    });
  }
  catch (ex) {
    pthis.notifyError("getPlacePredictions", ex);
  }
};


/**
 * Execute a query autocomplete prediction
 * @param {object} request - request
 * @param {string} cbId - callback id
 */
Client.Gmap.prototype.getQueryPredictions = function (request, cbId)
{
  try {
    var pthis = this;
    //
    if (request.location)
      request.location = this.toGmapPosition(request.location);
    //
    if (!this.AutocompleteService)
      this.AutocompleteService = new google.maps.places.AutocompleteService();
    //
    this.AutocompleteService.getQueryPredictions(request, function (results, status) {
      var e = [{obj: pthis.id, id: "cb", content: {res: results, cbId: cbId}}];
      Client.mainFrame.sendEvents(e);
    });
  }
  catch (ex) {
    pthis.notifyError("getQueryPredictions", ex);
  }
};


/**
 * Tell to the children that the visibility has changed
 * @param {Boolean} visible
 */
Client.Gmap.prototype.visibilityChanged = function (visible)
{
  // Update content size after becoming visible
  if (visible)
    this.resizeMap();
  Client.Element.prototype.visibilityChanged.call(this, visible);
};


/**
 * Tell to the children that the visibility has changed
 * @param {string} location
 * @param {exception} exc
 */
Client.Gmap.prototype.notifyError = function (location, exc)
{
  if (!this.sendError)
    return;
  var e = [{obj: this.id, id: "onError", content: {location: location, name: exc.name, message: exc.message}}];
  Client.mainFrame.sendEvents(e);
};


/**
 * resize the map to show an array of points
 * @param {object} points - points array
 */
Client.Gmap.prototype.setBounds = function (points)
{
  var bounds = new google.maps.LatLngBounds();
  for (var i = 0; i < points.length; i++) {
    bounds.extend(this.toGmapPosition(points[i]));
  }
  //
  this.map.fitBounds(bounds);
};
