/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global Client */

/**
 * @class A index scrollbar widget
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the progress bar
 */
Client.IndexScrollbar = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  this.domObj = document.createElement("div");
  this.domObj.classList.add("index-scrollbar");
  //
  parent.appendChildObject(this, this.domObj);
  //
  this.updateElement(element);
  this.attachEvents(element.events);
  //
  var pthis = this;
  if (!Client.mainFrame.isEditing()) {
    this.domObj.addEventListener("mousemove", function (ev) {
      pthis.onMouseMove(ev);
    });
    this.domObj.addEventListener("touchmove", function (ev) {
      pthis.onMouseMove(ev);
    });
    this.domObj.addEventListener("mouseup", function (ev) {
      pthis.onClick(ev);
    });
    this.domObj.addEventListener("touchup", function (ev) {
      pthis.onClick(ev);
    });
  }
  //
  this.renderImmediate();
};

// Make Client.IndexScrollbar extend Client.Element
Client.IndexScrollbar.prototype = new Client.Element();


/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.IndexScrollbar.prototype.updateElement = function (el)
{
  this.purgeMyProp(el);
  //
  if (el.minValue !== undefined) {
    this.minValue = el.minValue;
    delete el.minValue;
    this.render();
  }
  //
  if (el.maxValue !== undefined) {
    this.maxValue = el.maxValue;
    delete el.maxValue;
    this.render();
  }
  //
  if (el.list !== undefined) {
    if (typeof el.list === "string" && false) {
      try {
        this.preList = JSON.parse(el.list);
      }
      catch (ex) {
      }
    }
    else
      this.preList = el.list;
    delete el.list;
    this.render();
  }
  //
  if (el.listStyle !== undefined) {
    this.listStyle = el.listStyle;
    delete el.listStyle;
    this.render();
  }
  //
  if (el.value !== undefined) {
    this.setValue(el.value);
    delete el.value;
  }
  //
  if (el.valueStyle !== undefined) {
    this.valueStyle = el.valueStyle;
    this.setValue(this.value);
    delete el.valueStyle;
  }
  //
  Client.Element.prototype.updateElement.call(this, el);
};


/**
 * Attach events handler
 * @param {Array} events - array of the events to handle
 */
Client.IndexScrollbar.prototype.attachEvents = function (events)
{
  if (!events)
    return;
  //
  var pos = events.indexOf("onChange");
  if (pos >= 0) {
    events.splice(pos, 1);
    this.sendChange = true;
  }
  //
  pos = events.indexOf("onInput");
  if (pos >= 0) {
    events.splice(pos, 1);
    this.sendInput = true;
  }
  //
  Client.Element.prototype.attachEvents.call(this, events);
};


/**
 * Prepare to render widget
 */
Client.IndexScrollbar.prototype.render = function ()
{
  var pthis = this;
  if (!this.renderID) {
    this.renderID = setTimeout(function () {
      pthis.renderID = undefined;
      pthis.renderImmediate();
    });
  }
};


/**
 * Render widget
 */
Client.IndexScrollbar.prototype.renderImmediate = function ()
{
  if (this.renderID) {
    this.renderID = undefined;
    clearTimeout(this.renderID);
  }
  //
  // Clear content
  while (this.domObj.firstChild)
    this.domObj.removeChild(this.domObj.firstChild);
  //
  this.prepareList();
  //
  if (this.list)
    this.renderList();
  //
  if (this.value)
    this.showValue(this.value);
};


/**
 * Prepare value list base on min and max values
 */
Client.IndexScrollbar.prototype.prepareList = function ()
{
  var i;
  this.list = [];
  //
  if (this.preList) {
    for (i = 0; i < this.preList.length; i++)
      this.list.push(this.preList[i]);
  }
  //
  if (this.minValue !== undefined && this.maxValue !== undefined) {
    if (isNaN(parseInt(this.minValue))) {
      for (i = this.minValue.charCodeAt(0); i <= this.maxValue.charCodeAt(0); i++)
        this.list.push(String.fromCharCode(i));
    }
    else {
      for (i = parseInt(this.minValue, 10); i <= parseInt(this.maxValue, 10); i++)
        this.list.push(i);
    }
  }
};


/**
 * Prepare a filtered value list base on size
 */
Client.IndexScrollbar.prototype.filterList = function ()
{
  // TODO
  return this.list;
};


/**
 * Create child objects
 */
Client.IndexScrollbar.prototype.renderList = function ()
{
  var list = this.filterList();
  //
  // Each element in the list is a span
  for (var i = 0; i < list.length; i++) {
    var c = document.createElement("span");
    c.innerText = list[i];
    //
    // We can set the span style using a class or a style string
    if (this.listStyle) {
      if (this.listStyle.indexOf(":") > -1)
        c.style.cssText = this.listStyle;
      else
        c.className = this.listStyle;
    }
    //
    this.domObj.appendChild(c);
  }
};


/**
 * sets a new value for this named range
 * @param {string|number} value
 */
Client.IndexScrollbar.prototype.setValue = function (value)
{
  // If I need to change, reset value
  if (this.value !== value) {
    this.showValue(this.value, true);
  }
  //
  this.value = value;
  this.showValue(this.value);
};


/**
 * shows a new value for this named range
 * @param {string|number} value
 * @param {bool} reset - reset the styles of a previous value
 */
Client.IndexScrollbar.prototype.showValue = function (value, reset)
{
  // Search corresponding element
  for (var i = 0; i < this.domObj.childNodes.length; i++) {
    var s = this.domObj.childNodes[i];
    if (s.innerText === value) {
      // Style or class?
      if (this.valueStyle && this.valueStyle.indexOf(":") > -1) {
        if (reset) {
          // to reset the value style, need to see if the liststyle
          // is a class or a style
          if (this.listStyle && this.listStyle.indexOf(":") > -1)
            s.style.cssText = this.listStyle;
          else
            s.style.cssText = "";
        }
        else {
          // Setting the value style
          s.style.cssText += this.valueStyle;
        }
      }
      else {
        // Setting or resetting class
        if (reset)
          s.classList.remove(this.valueStyle);
        else
          s.classList.add(this.valueStyle);
      }
    }
  }
};


/**
 * send a new value for this named range
 * @param {event} ev
 * @param {string|number} value
 * @param {bool} transient
 */
Client.IndexScrollbar.prototype.changeValue = function (ev, value, transient)
{
  // No real change, exit
  if (value === this.value)
    return;
  //
  if (this.sendInput || !transient) {
    var e = [{obj: this.id, id: "chgProp", content: {name: "value", value: value, clid: Client.id}}];
    if (this.sendChange)
      e.push({obj: this.id, id: "onChange", content: this.saveEvent(ev)});
    if (this.sendInput)
      e.push({obj: this.id, id: "onInput", content: this.saveEvent(ev)});
    Client.mainFrame.sendEvents(e);
    this.setValue(value);
  }
};


/**
 * mouse is moving, let's see if this event changes value
 * @param {event} event
 */
Client.IndexScrollbar.prototype.onMouseMove = function (event)
{
  // a button was pressed while moving mouse or a slide on the list is recognized
  if (event.which || event.touches) {
    //
    // the span touched is contined in the event only in mousemove events, otherwise we have to see
    // which span is under the finger
    var tg = event.target;
    if (event.touches) {
      var x = event.touches[0].pageX;
      var y = event.touches[0].pageY;
      //
      if (this.domObj.clientWidth > this.domObj.clientHeight)
        y = this.domObj.offsetTop + this.domObj.offsetHeight / 2;
      else
        x = this.domObj.offsetLeft + this.domObj.offsetWidth / 2;
      //
      tg = document.elementFromPoint(x, y);
    }
    //
    // Better to check if the found object is a child of this element
    var isChild = false;
    if (tg) {
      var obj = tg.parentNode;
      while (obj && !isChild) {
        if (obj === this.domObj)
          isChild = true;
        obj = obj.parentNode;
      }
    }
    //
    if (isChild) {
      var v = tg.innerText;
      this.changeValue(event, v, true);
    }
  }
};


/**
 * mouse is moving, let's see if this event changes value
 * @param {event} event
 */
Client.IndexScrollbar.prototype.onClick = function (event)
{
  // a button was pressed while moving mouse
  if (event.target !== this.domObj) {
    var v = event.target.innerText;
    this.changeValue(event, v);
  }
};
