/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/**
 * @class A popup container
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the dialog
 * @extends Client.Container
 */
Client.IonModal = function (element, parent, view)
{
  if (!element.options.modal)
    element.tag = "ion-popover-inner";
  Client.Container.call(this, element, parent, view);
  //
  if (element === undefined)
    return;
  //
  this.opt = element.options;
  var t = this.opt.modal ? "modal" : "popover";
  //
  // Create objects
  this.rootObj = document.createElement("ion-" + t);
  //
  var classname = t + "-cmp show-page";
  if (this.opt.ref && this.opt.ref.position === "top")
    classname += " popover-bottom";
  if (this.opt.extcls)
    classname += " " + this.opt.extcls;
  this.rootObj.className = classname;
  //
  this.rootObj.style.zIndex = "100";
  this.rootObj.id = this.id;
  //
  this.backgroundDiv = document.createElement("ion-backdrop");
  this.backgroundDiv.ontouchmove = function (ev) {
    ev.preventDefault();
  };
  this.rootObj.appendChild(this.backgroundDiv);
  //
  if (this.opt.modal) {
    //
    // Modal structure:
    // modal-cmp(root) / ion-backdrop(backgroundDiv) | modal-wrapper(domObj,extObj,wrapperObj)
    //
    this.domObj.className = "modal-wrapper";
    this.rootObj.appendChild(this.domObj);
    this.wrapperObj = this.domObj;
    this.extObj = this.wrapperObj;
    //
    // Delete ref property
    delete this.opt.ref;
  }
  else {
    //
    // Popover structure:
    // popover-cmp(root) / ion-backdrop(backgroundDiv) | popover-wrapper(extObj) / popover-content (wrapperObj) / popover-viewport (viewportObj) / ion-popover-inner (domObj)
    //
    this.extObj = document.createElement("div");
    this.extObj.className = "popover-wrapper";
    this.rootObj.appendChild(this.extObj);
    this.wrapperObj = document.createElement("div");
    this.wrapperObj.className = "popover-content";
    this.extObj.appendChild(this.wrapperObj);
    this.viewportObj = document.createElement("div");
    this.viewportObj.className = "popover-viewport";
    this.wrapperObj.appendChild(this.viewportObj);
    this.viewportObj.appendChild(this.domObj);
    //
    if (this.opt.width)
      this.wrapperObj.style.width = this.opt.width;
    if (this.opt.height)
      this.wrapperObj.style.height = this.opt.height;
    //
    // Adjust ref property
    this.opt.autoclose = true;
    if (this.opt.ref) {
      if (!this.opt.ref.id) {
        delete this.opt.ref;
      }
      else {
        if (Client.Ionic.platform === "md") {
          this.opt.ref.whisker = false;
          var refEl = Client.eleMap[this.opt.ref.id];
          if (refEl && refEl.domObj && this.opt.ref.offset === undefined)
            this.opt.ref.offset = -refEl.domObj.offsetHeight;
        }
        else {
          this.opt.ref.whisker = true;
          delete this.opt.ref.offset;
        }
      }
    }
  }
  //
  // Append the rootObj div to the app-ui
  var appui = document.getElementById("app-ui");
  appui.appendChild(this.rootObj);
  //
  // If the popup has to close when clicking outside it
  if (this.opt && this.opt.autoclose) {
    // Define the function to close the dialog
    this.autoclose = true;
    var pthis = this;
    this.autoCloseFunct = function (ev) {
      pthis.autoClose(ev, pthis);
    };
    //
    this.backgroundDiv.addEventListener("mousedown", this.autoCloseFunct);
    this.backgroundDiv.addEventListener("touchend", this.autoCloseFunct);
  }
  //
  if (this.opt.modal) {
    // modal enter by sliding content from bottom
    var duration = (Client.Ionic.platform === "md") ? 250 : 400;
    this.extObj.style.transition = "transform " + duration + "ms cubic-bezier(.36,.66,.04,1),opacity " + duration + "ms cubic-bezier(.36,.66,.04,1)";
    this.backgroundDiv.style.transition = "opacity " + duration + "ms cubic-bezier(.36,.66,.04,1)";
  }
  else {
    // popover: programming ios animations, as md animation require positioning
    this.extObj.style.transition = "opacity 100ms ease";
    this.backgroundDiv.style.transition = "opacity 100ms ease";
  }
};


// Make Client.IonModal extend Client.Dialog
Client.IonModal.prototype = new Client.Dialog();


/**
 * Close the popup
 * @param {Function} cb - callback
 * @param {String} elemId - callback param
 */
Client.IonModal.prototype.close = function (cb, elemId)
{
  this.closing = true;
  var duration = (Client.Ionic.platform === "md") ? 250 : 400;
  //
  Client.Container.prototype.close.call(this);
  //
  if (!this.opt.modal) {
    // popover exits with a fade out animation
    this.extObj.style.transitionDuration = "500ms";
    this.backgroundDiv.style.transitionDuration = "500ms";
    duration = 500;
  }
  //
  // Resetting properties to slide modal out
  this.extObj.style.transform = "";
  this.extObj.style.opacity = "";
  this.backgroundDiv.style.opacity = "0.01";
  setTimeout(function () {
    this.rootObj.remove();
    //
    // Delete any view element
    if (this.view) {
      this.view.dialog = null;
      this.view.close();
    }
    //
    // Callback is needed to allow other clients to close popup menu during telecollaboration or testauto cases.
    // See Client.IonHelper.createMenu for more details
    if (cb)
      cb(elemId);
  }.bind(this), duration);
};


/**
 * Position the popup on the screen
 */
Client.IonModal.prototype.positionElement = function ()
{
  Client.Dialog.prototype.positionElement.call(this);
  //
  // No absolute position of the content if we don't have a reference object
  if (!this.opt.ref)
    this.rootObj.classList.add("popover-no-ref");
  //
  // The positionElement could have changed the position of the popup set in the options.
  // If the popup is moved on top of the reference element, I need to add a class to draw
  // the whisker properly
  if (this.opt.ref && this.opt.ref.whiskerPosition === "bottom")
    this.rootObj.classList.add("popover-bottom");
  //
  if (this.opt.modal) {
    // Start sliding modal content up
    this.extObj.style.transform = "translate3d(0,0,0)";
    this.backgroundDiv.style.opacity = "0.4";
  }
  else {
    if (Client.Ionic.platform === "md") {
      //
      // md popover zoom in: select zoom direction
      var to1 = "center";
      var to2 = "center";
      if (this.opt.ref && this.opt.ref.id) {
        var refEl = Client.eleMap[this.opt.ref.id];
        if (refEl && refEl.domObj) {
          var r1 = refEl.domObj.getBoundingClientRect();
          var r2 = this.wrapperObj.getBoundingClientRect();
          var c1x = (r1.left + r1.right) / 2;
          var c2x = (r2.left + r2.right) / 2;
          var c1y = (r1.top + r1.bottom) / 2;
          var c2y = (r2.top + r2.bottom) / 2;
          if (Math.abs(c1x - c2x) > 10)
            to2 = (c1x < c2x) ? "left" : "right";
          if (c1y < c2y)
            to1 = "top";
          else {
            to2 = "bottom";
            //
            // I have to calculate again the top position of the element because the position
            // of the Client.Dialog is different when it's on top
            var x = r1.bottom - r2.height;
            if (this.opt.ref.offset !== undefined)
              x = r1.top - r2.height - this.opt.ref.offset;
            //
            this.wrapperObj.style.top = x + "px";
          }
        }
      }
      this.wrapperObj.style.transformOrigin = to1 + " " + to2;
      this.wrapperObj.style.transform = "scale(0.001)";
      //
      setTimeout(function () {
        this.wrapperObj.style.transition = "transform 300ms cubic-bezier(.36, .66, .04, 1)";
        this.wrapperObj.style.transform = "";
      }.bind(this), 50);
    }
    else {
      // ios popover fade in
      this.backgroundDiv.style.opacity = "0.1";
    }
  }
  //
  // Start fade in
  this.extObj.style.opacity = "1";
  if (this.viewportObj) {
    this.viewportObj.style.opacity = "1";
  }
};


/**
 * Create the wisker element for this dialog
 */
Client.IonModal.prototype.createWhisker = function (style, popupCoord, refAbsPos, extWhiskerSize, intWhiskerSize, whiskerDiff)
{
  var whisker = document.createElement("div");
  whisker.className = "popover-arrow";
  //
  // Position the whisker
  var borderW = this.getBorderWidth(style);
  //
  // Get the whisker position
  var wCoord = this.calculateWhiskerCoord(popupCoord.wSide, refAbsPos, extWhiskerSize, borderW, whiskerDiff);
  //
  // Set left and top position
  whisker.style.left = (wCoord.wExtLeft) + "px";
  whisker.style.top = (wCoord.wExtTop) + "px";
  //
  // Append whisker
  this.rootObj.appendChild(whisker);
};


/*
 * Calculate the whisker position
 * @param {String} side - side of the whisker relative to the popup
 * @param {Object} - position of the reference object
 * @param {Int} extWhiskerSize - whisker size
 * @param {Float} whiskerDiff - width difference between the internal and the external whisker
 * @returns {Object}
 */
Client.IonModal.prototype.calculateWhiskerCoord = function (side, refAbsPos, extWhiskerSize, borderWidth, whiskerDiff)
{
  var coord = Client.Dialog.prototype.calculateWhiskerCoord.call(this, side, refAbsPos, extWhiskerSize, borderWidth, whiskerDiff);
  //
  // The ionmodal whisker is different from the dialog whisker
  // and it needs a slightly different positioning
  if (side === "bottom")
    coord.wExtTop = coord.wExtTop - 2;
  if (side === "top" && this.opt.ionmenu)
    coord.wExtTop = coord.wExtTop + 5;
  return coord;
};


/**
 * Close the popup if mouse clicked outside it
 * @param {MouseEvent} ev - the event occured when the user clicked
 * @param {Dialog} pthis - the dialog that has to be closed
 */
Client.IonModal.prototype.autoClose = function (ev) {
  //
  // If the clicked element is not one of my children close the popup
  var clickedEl = ev.target;
  var parent = clickedEl.parentNode;
  while (parent) {
    if (parent === this.wrapperObj)
      return;
    parent = parent.parentNode;
  }
  //
  if (this.view && !this.closing) {
    //
    // Search for a back-button in the page
    var el = this.rootObj.getElementsByClassName("back-button");
    if (el && el.length) {
      // If the navbar has the back button, click on it.
      el[0].click();
    }
    else {
      // Try to go back. we cannot close the view directly as ionic use a page controller
      window.history.back();
    }
  }
};
