/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global Client,componentHandler */

/**
 * @class A Material Design Slider Component
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the button
 */
Client.MDSlider = function (element, parent, view)
{
  // To recognize the first updateElement call
  this.firstTime = true;
  element.type = "range";
  Client.Input.call(this, element, parent, view);
  componentHandler.upgradeElement(this.domObj);
};

Client.MDSlider.prototype = new Client.Input();

/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.MDSlider.prototype.updateElement = function (el)
{
  var update = Client.MDHelper.updateElement(el, this) || this.firstTime;
  //
  // To change the value of the slider after first call we need
  // to use the material object directly
  if (el.value !== undefined) {
    if (this.domObj.MaterialSlider) {
      this.domObj.MaterialSlider.change(el.value);
      delete el.value;
    }
  }
  //
  if (update) {
    var cs = "mdl-slider mdl-js-slider ";
    cs += Client.MDHelper.updateClass(this);
    this.domObj.className = cs;
    this.firstTime = false;
  }
  //
  Client.Input.prototype.updateElement.call(this, el);
};



/**
 * @class A container for a card
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the button
 */
Client.MDInput = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  // This object contains the entire input structure, that is fairly complex
  this.divObj = document.createElement("div");
  //
  if (element.type === "textarea")
    this.domObj = document.createElement("textarea");
  else
    this.domObj = document.createElement("input");
  //
  // This is the label object
  this.labelObj = document.createElement("label");
  this.labelObj.setAttribute("for", this.id);
  //
  // This is the error text object
  this.spanObj = document.createElement("span");
  //
  this.divObj.appendChild(this.domObj);
  this.divObj.appendChild(this.labelObj);
  this.divObj.appendChild(this.spanObj);
  //
  // Set readonly attribute to input element in case of readonly role
  if (Client.clientRole === "ro")
    this.domObj.setAttribute("readonly", "");
  //
  // The object that should be added to the DOM is the div, not the input
  parent.appendChildObject(this, this.divObj);
  //
  this.updateElement(element, true);
  this.attachEvents(element.events);
  //
  componentHandler.upgradeElement(this.divObj);
};

Client.MDInput.prototype = new Client.Input();

/**
 * Update element properties
 * @param {Object} el - properties to update
 * @param {Bool} firstTime - true if called during initialization
 */
Client.MDInput.prototype.updateElement = function (el, firstTime)
{
  var update = !el || Client.MDHelper.updateElement(el, this) || firstTime;
  if (!el)
    el = {};
  //
  this.purgeMyProp(el);
  //
  // Generate a floting label over the input
  if (el.floatingLabel !== undefined) {
    this.floatingLabel = el.floatingLabel;
    update = true;
    delete el.floatingLabel;
  }
  //
  if (el.expandIcon !== undefined) {
    // Generate an expandable input with a clickable icon
    if (el.expandIcon) {
      if (!this.expLabel) {
        this.holder = document.createElement("div");
        this.holder.className = "mdl-textfield__expandable-holder";
        this.expLabel = document.createElement("label");
        this.expLabel.className = "mdl-button mdl-js-button mdl-button--icon want-click";
        this.expLabel.setAttribute("for", this.id);
        this.expIcon = document.createElement("i");
        this.expIcon.className = "material-icons";
        //
        this.divObj.removeChild(this.domObj);
        this.divObj.removeChild(this.labelObj);
        this.divObj.removeChild(this.spanObj);
        this.holder.appendChild(this.domObj);
        this.holder.appendChild(this.labelObj);
        this.holder.appendChild(this.expLabel);
        this.expLabel.appendChild(this.expIcon);
        //
        this.divObj.appendChild(this.expLabel);
        this.divObj.appendChild(this.holder);
      }
      this.expandable = true;
      this.expIcon.innerText = el.expandIcon;
    }
    else {
      // Remove expandable icon
      if (this.expLabel) {
        this.divObj.removeChild(this.expLabel);
        this.divObj.removeChild(this.holder);
        this.divObj.appendChild(this.domObj);
        this.divObj.appendChild(this.labelObj);
        this.divObj.appendChild(this.spanObj);
        this.holder = undefined;
        this.expLabel = undefined;
        this.expIcon = undefined;
      }
      this.expandable = false;
    }
    update = true;
    delete el.expandIcon;
  }
  //
  // To change the value after initialization we need to act in the material design object
  if (el.value !== undefined) {
    if (this.divObj.MaterialTextfield)
      this.divObj.MaterialTextfield.change(el.value);
    //
    // it seems that an empty value is not well recognized so we need to force class update
    this.divObj.classList.toggle("is-dirty", el.value);
  }
  //
  if (el.disabled !== undefined) {
    this.divObj.classList.toggle("is-disabled", el.disabled);
  }
  // Set label text
  if (el.label !== undefined) {
    this.labelObj.innerText = el.label;
    delete el.label;
  }
  // Set error text
  if (el.errorText !== undefined)
    update = true;
  //
  if (update) {
    //
    // Setting container classes
    var cs = "mdl-textfield mdl-js-textfield ";
    //
    if (this.floatingLabel)
      cs += "mdl-textfield--floating-label ";
    //
    if (this.expandable)
      cs += "mdl-textfield--expandable ";
    //
    if (this.fullWidth)
      cs += "mdl-textfield--full-width ";
    //
    if (!firstTime)
      cs += "is_upgraded ";
    //
    var err = false;
    if (this.spanObj.textContent) {
      cs += "is-invalid ";
      err = true;
    }
    //
    if (this.domObj.value)
      cs += "is-dirty ";
    //
    cs += Client.MDHelper.updateClass(this);
    //
    this.divObj.className = cs;
    //
    // Setting input object classes
    cs = "mdl-textfield__input";
    if (err)
      cs += " element-invalid";
    //
    this.domObj.className = cs;
    //
    this.labelObj.className = "mdl-textfield__label";
    this.spanObj.className = "mdl-textfield__error";
  }
  //
  Client.Input.prototype.updateElement.call(this, el);
};

/**
 * Return the dom root object of this element
 * @returns {DomElement}
 */
Client.MDInput.prototype.getRootObject = function ()
{
  return this.divObj || this.domObj;
};

/**
 * Set custom error message
 * @param {String} message - the error message
 * @param {bool} srv
 */
Client.MDInput.prototype.setError = function (message, srv)
{
  Client.Element.prototype.setError.call(this, message, srv);
  if (this.spanObj.textContent !== message)
    this.spanObj.textContent = message;
  //
  var pthis = this;
  setTimeout(function () {
    if (message)
      pthis.divObj.classList.add("is-invalid");
    else
      pthis.divObj.classList.remove("is-invalid");
  }, 10);
};


/**
 * @class A container for a material design toggle (checkbox, switch...)
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the button
 */
Client.MDToggle = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  this.domObj = document.createElement("label");
  this.inputObj = document.createElement("input");
  this.spanObj = document.createElement("span");
  this.domObj.appendChild(this.inputObj);
  this.domObj.appendChild(this.spanObj);
  //
  this.updateElement(element, true);
  this.attachEvents(element.events);
  parent.appendChildObject(this, this.domObj);
  //
  componentHandler.upgradeElement(this.domObj);
  //
  // To ripple correctly we need to upgrade the ripple component too
  if (this.ripple)
    componentHandler.upgradeElements(this.domObj.children);
  //
  if (Client.mainFrame.isEditing()) {
    this.inputObj.style.visibility = "hidden";
    this.inputObj.style.position = "absolute";
  }
};

Client.MDToggle.prototype = new Client.Element();

/**
 * Update element properties
 * @param {Object} el - properties to update
 * @param {Bool} firstTime - true if called during initialization
 */
Client.MDToggle.prototype.updateElement = function (el, firstTime)
{
  var update = Client.MDHelper.updateElement(el, this) || firstTime;
  //
  // toggle type (checkbox, radio, icon-toggle, switch)
  if (el.type !== undefined) {
    this.type = el.type;
    update = true;
    delete el.type;
  }
  //
  if (el.label !== undefined) {
    this.spanObj.innerText = el.label;
    delete el.label;
  }
  //
  // group is used to group radio buttons
  if (el.group !== undefined) {
    this.inputObj.name = el.group;
    delete el.group;
  }
  //
  // To set disabled and checked properties we need to act on the material design object
  if (el.disabled !== undefined) {
    this.inputObj.disabled = el.disabled;
    var x = this.getComponent();
    if (x) {
      if (el.disabled)
        x.disable();
      else
        x.enable();
    }
    delete el.disabled;
  }
  if (el.checked !== undefined) {
    if (typeof el.checked === "string")
      el.checked = (el.checked === "true");
    this.inputObj.checked = el.checked;
    var x = this.getComponent();
    if (x) {
      if (el.checked) {
        if (x.check)
          x.check();
        if (x.on)
          x.on();
        if (this.type === "radio")
          x.onChange_();
      }
      else {
        if (x.uncheck)
          x.uncheck();
        if (x.off)
          x.off();
      }
    }
    delete el.checked;
  }
  //
  if (update) {
    var cs = "mdl-" + this.type + " mdl-js-" + this.type + " ";
    if (!firstTime)
      cs += "is_upgraded ";
    cs += Client.MDHelper.updateClass(this);
    //
    this.domObj.className = cs;
    //
    if (firstTime) {
      if (this.type === "radio") {
        this.inputObj.type = "radio";
        this.inputObj.className = "mdl-radio__button";
      }
      else {
        this.inputObj.type = "checkbox";
        this.inputObj.className = "mdl-" + this.type + "__input";
      }
      this.spanObj.className = "mdl-" + this.type + "__label";
      if (this.type === "icon-toggle")
        this.spanObj.className += " material-icons";
    }
  }
  //
  Client.Element.prototype.updateElement.call(this, el);
};

/**
 * Returns the underline material design object that is different among toggle types
 */
Client.MDToggle.prototype.getComponent = function ()
{
  if (this.domObj.MaterialCheckbox)
    return this.domObj.MaterialCheckbox;
  if (this.domObj.MaterialRadio)
    return this.domObj.MaterialRadio;
  if (this.domObj.MaterialIconToggle)
    return this.domObj.MaterialIconToggle;
  if (this.domObj.MaterialSwitch)
    return this.domObj.MaterialSwitch;
};

/**
 * Attach events handler
 * @param {Array} events - array of the events to handle
 */
Client.MDToggle.prototype.attachEvents = function (events)
{
  if (!events)
    return;
  //
  var idx = events.indexOf("onChange");
  if (idx >= 0) {
    this.sendChange = true;
    events.splice(idx, 1);
  }
  //
  var pthis = this;
  this.inputObj.onchange = function (ev) {
    var e = [{obj: pthis.id, id: "chgProp", content: {name: "checked", value: pthis.inputObj.checked, clid: Client.id}}];
    if (pthis.sendChange)
      e.push({obj: pthis.id, id: "onChange", content: pthis.saveEvent(ev)});
    Client.mainFrame.sendEvents(e);
  };
  //
  Client.Element.prototype.attachEvents.call(this, events);
};
