/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/**
 * @class A container that use iScroll to scroll the content
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the container
 * @extends Client.Element
 */
Client.ScrollContainer = function (element, parent, view)
{
  this.revealOffset = 20;
  this.refreshOffset = 60;
  //
  Client.Element.call(this, element, parent, view);
  //
  if (element === undefined)
    return;
  //
  this.domObj = document.createElement("div");
  this.scrollDiv = document.createElement("div");
  this.scrollDiv.style.minHeight = "101%";
  this.domObj.appendChild(this.scrollDiv);
  //
  this.updateElement(element);
  this.attachEvents(element.events);
  //
  parent.appendChildObject(this, this.domObj);
  //
  // Create children after attaching element as external components, rely on dom structure
  this.createChildren(element);
  //
  if (this.domObj) {
    var _this = this;
    window.setTimeout(function () {
      _this.myScroll = new IScroll(_this.domObj, _this.scrollOptions);
      //
      _this.myScroll.on('scroll', function (ret) {
        _this.onScrollHandler(ret);
      });
      _this.hideRevealTemplate();
      if (_this.revealDomObj)
        _this.revealDomObj.style.visibility = "";
    }, 0);
  }
};


// Make Client.ScrollContainer extend Client.Container
Client.ScrollContainer.prototype = new Client.Container();


/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.ScrollContainer.prototype.updateElement = function (el)
{
  // Layouting not supported
  if (el.layout)
    delete el.layout;
  //
  Client.Container.prototype.updateElement.call(this, el);
  //
  if (!this.scrollOptions)
    this.scrollOptions = {};
  if (el.scrollOptions) {
    if (typeof el.scrollOptions === "string") {
      try {
        this.scrollOptions = JSON.parse(el.scrollOptions);
      } catch (e) {
      }
    }
    else
      this.scrollOptions = el.scrollOptions;
    //
    if (this.scrollOptions.revealOffset)
      this.revealOffset = this.scrollOptions.revealOffset;
    if (this.scrollOptions.refreshOffset)
      this.refreshOffset = this.scrollOptions.refreshOffset;
  }
  //
  this.scrollOptions.probeType = 3;
  this.scrollOptions.mouseWheel = true;
  if (this.rowCount)
    this.scrollOptions.maxSpeed = 2;
  //
  // Handle design-time exposed properties
  if (el.showScrollbar)
    this.scrollOptions.scrollbars = el.showScrollbar;
  if (el.snap)
    this.scrollOptions.snap = el.snap;
  if (el.maxSpeed)
    this.scrollOptions.maxSpeed = el.maxSpeed;
  if (el.scrollDirection) {
    switch (el.scrollDirection) {
      case 1 :
        this.scrollOptions.scrollY = true;
        this.scrollOptions.scrollX = false;
        break;

      case 2 :
        this.scrollOptions.scrollY = false;
        this.scrollOptions.scrollX = true;
        break;

      case 3 :
        this.scrollOptions.scrollX = true;
        this.scrollOptions.scrollY = true;
        break;
    }
  }
};


/**
 * Attach events handler
 * @param {Array} events - array of the events to handle
 */
Client.ScrollContainer.prototype.attachEvents = function (events)
{
  if (!events)
    return;
  //
  Client.Element.prototype.attachEvents.call(this, events);
  //
  // When attachEvents is called the scroll doesn't exists, the onscroll event is attached at the end of construction so we
  // can use a boolean to tell the handler that must send the event to the server
  if (events.indexOf("onScroll") >= 0)
    this.handleScroll = true;
};


/**
 * Append a child DOM Object to this element DOM Object
 * Usually the child element is not contained in this element array
 * it will be pushed there just after this function call.
 * @param {Element} child - child element that requested the insertion
 * @param {HTMLElement} domObj - child DOM object to add
 */
Client.ScrollContainer.prototype.appendChildObject = function (child, domObj)
{
  if (!Client.mainFrame.isEditing() && domObj.classList.contains("pulltemplate")) {
    this.pullDomObj = domObj;
    this.domObj.appendChild(domObj);
    this.pullDomObj.style.display = "";
    this.pullDomObj.style.visibility = "";
  }
  else {
    if (domObj.classList.contains("revealtemplate")) {
      this.revealDomObj = domObj;
      this.revealDomObj.style.visibility = "hidden";
    }
    //
    // Show the pull header also in the viewview
    if (Client.mainFrame.isEditing() && domObj.classList.contains("pulltemplate")) {
      domObj.style.transform = "translate3d(0px, 0px, 0px)";
      domObj.style.position = "static";
    }
    this.scrollDiv.appendChild(domObj);
  }
};


/**
 * Returns the scrollTop of this object
 */
Client.ScrollContainer.prototype.getScrollTop = function ()
{
  if (!this.myScroll || this.myScroll.y === undefined)
    return 0;
  return Math.round(Math.abs(this.myScroll.y));
};


/**
 * Sets the scrollTop of this object
 * @param {int} top - scrollTop
 */
Client.ScrollContainer.prototype.setScrollTop = function (top)
{
  if (this.myScroll)
    this.myScroll.scrollTo(0, -top);
};


/**
 * Returns the scrollLeft of this object
 */
Client.ScrollContainer.prototype.getScrollLeft = function ()
{
  if (!this.myScroll || this.myScroll.x === undefined)
    return 0;
  return Math.round(Math.abs(this.myScroll.x));
};


/**
 * Sets the scrollLeft of this object
 * @param {int} left - scrollTop
 */
Client.ScrollContainer.prototype.setScrollLeft = function (left)
{
  if (this.myScroll)
    this.myScroll.scrollTo(-left, 0);
};


/**
 * Update this container as a list box
 */
Client.ScrollContainer.prototype.initListBox = function ()
{
  if (!this.elements)
    return;
  //
  if (!this.listInitialized) {
    if (this.domObj.style.position === "")
      this.domObj.style.position = "relative";
    if (this.domObj.style.overflow === "" && this.domObj.style.overflowY === "")
      this.domObj.style.overflowY = "hidden";
    //
    // Now I'm able to see the size of a single record
    this.tmpl = null;
    for (var i = 0; i < this.ne(); i++)
      if (this.elements[i] && this.elements[i].domObj && (!this.elements[i].domObj.className || this.elements[i].domObj.className.indexOf("pulltemplate") === -1 && this.elements[i].domObj.className.indexOf("revealtemplate") === -1)) {
        // The template is the element wich classname doesn't contains pulltemplate and revealtemplate
        this.tmpl = this.elements[i];
        break;
      }
    if (this.tmpl === null)
      this.tmpl = this.elements[0];
    var s = window.getComputedStyle(this.tmpl.domObj);
    this.recWidth = this.tmpl.domObj.offsetWidth + parseInt(s.marginLeft) + parseInt(s.marginRight);
    this.recHeight = this.tmpl.domObj.offsetHeight + parseInt(s.marginTop) + parseInt(s.marginBottom);
    this.cacheProp();
    //
    // Hide template object instead of making it invisible
    this.tmpl.domObj.style.display = "none";
    //
    // Attach onscroll to myscroller
    var pthis = this;
    if (this.myScroll)
      this.hideRevealTemplate();
    //
    // Attach key down
    this.domObj.tabIndex = "0";
    this.domObj.addEventListener("keydown", function (ev) {
      pthis.moveListBox(ev);
    }, true);
    //
    this.listInitialized = true;
  }
};


/**
 * Refreshes the scroller
 */
Client.ScrollContainer.prototype.refreshScroll = function ()
{
  if (this.myScroll && !this.refreshTimer) {
    var _this = this;
    this.refreshTimer = window.setTimeout(function () {
      _this.refreshTimer = null;
      _this.myScroll.refresh();
    }, 0);
  }
};


/**
 * Remove a child from the element
 * @param {Object} content - an object with the id of the element to remove
 */
Client.ScrollContainer.prototype.removeChild = function (content)
{
  Client.Element.prototype.removeChild.call(this, content);
  //
  // When adding or removing children we must refresh the scroller
  if (this.myScroll)
    this.refreshScroll();
};


/**
 * Insert a child element before another element
 * @param {Object} content - contains the element to insert and the id of the existing element. The new element is inserted before this element
 */
Client.ScrollContainer.prototype.insertBefore = function (content)
{
  Client.Container.prototype.insertBefore.call(this, content);
  //
  // When adding or removing children we must refresh the scroller
  if (this.myScroll)
    this.refreshScroll();
};


/**
 * Scrolls to an object
 * @param {string} elementId
 */
Client.ScrollContainer.prototype.scrollTo = function (elementId)
{
  var el = document.getElementById(elementId);
  if (el && this.myScroll)
    this.myScroll.scrollToElement(el);
};


/**
 * Goes to next page when snapped
 */
Client.ScrollContainer.prototype.nextPage = function ()
{
  if (this.myScroll)
    this.myScroll.nextPage();
};


/**
 * Goes to prev page when snapped
 */
Client.ScrollContainer.prototype.prevPage = function ()
{
  if (this.myScroll)
    this.myScroll.prevPage();
};


/**
 * Goes to a page
 * @param {int} x - page in the horizontal axis (0 if no such axis)
 * @param {int} y - page in the vertical axis (0 if no such axis)
 */
Client.ScrollContainer.prototype.goToPage = function (x, y)
{
  if (this.myScroll)
    this.myScroll.goToPage(x, y);
};


/**
 * Handler fo OnScroll on the ScrollContainer
 * @param {bool} returning - true if is a scroll of return to position
 */
Client.ScrollContainer.prototype.onScrollHandler = function (returning) {
  // OnScroll event handler function (notify to server)
  if (this.handleScroll && this.scrollSetTime === undefined || new Date() - this.scrollSetTime > 100) {
    var e = [];
    if (this.getScrollTop() !== this.lastScrollTop)
      e.push({obj: this.id, id: "chgProp", content: {name: "scrollTop", value: this.getScrollTop(), clid: Client.id}});
    if (this.getScrollLeft() !== this.lastScrollLeft)
      e.push({obj: this.id, id: "chgProp", content: {name: "scrollLeft", value: this.getScrollLeft(), clid: Client.id}});
    e.push({obj: this.id, id: "onScroll", content: {}});
    Client.mainFrame.sendEvents(e);
    this.lastScrollTop = this.getScrollTop();
    this.lastScrollLeft = this.getScrollLeft();
  }
  //
  // Handle Pull to Refresh
  if (this.pullDomObj) {
    // When scrolling over the refresh offset add the class so the user can change the
    // visualization and send the event
    if (this.myScroll.y > this.refreshOffset && this.myScroll.directionY === -1 && !this.pullDomObj.classList.contains("refreshing")) {
      this.pullDomObj.classList.add("refreshing");
      Client.mainFrame.sendEvents([{obj: this.id, id: "onPullRefresh"}]);
    }
    else if (this.myScroll.y > this.revealOffset && this.myScroll.directionY === -1 && !returning) {
      // Show the header when we have scrolled over the offset
      this.showPullTemplate();
    }
    else if (this.myScroll.y > 0 && this.myScroll.y < this.revealOffset && !this.pullDomObj.classList.contains("refreshing")) {
      // Hide the header but only if the user not surpassed the refreshOffset, in that case is the user that must hide the
      // template in its code when done refreshing
      this.hidePullTemplate();
    }
    else if ((this.myScroll.y < this.revealOffset && returning && this.hidePullTemplateOnReturning) || (!this.hasRealChildren() && this.hidePullTemplateOnReturning)) {
      // We want that the 'refreshing' status be hidden only when the scroller returns to the top (returning)
      // if the user code calls the 'hide' when the user has not left the scroll we don't hide the message but memorize that we have to do so
      // when returning
      this.pullDomObj.classList.remove("refreshing");
      this.hidePullTemplate();
    }
  }
  //
  // Handle useWindow scroll
  if (this.listInitialized)
    this.updateListBox();
  //
  // Set the global mainframe variable to separate scrolls from clicks
  var ev = {};
  ev.timeStamp = new Date();
  ev.srcElement = this.domObj;
  Client.mainFrame.lastScrollEvent = ev;
};


/**
 * Show the pull template
 */
Client.ScrollContainer.prototype.showPullTemplate = function ()
{
  if (this.pullDomObj)
    this.pullDomObj.classList.add("visible");
};


/**
 * hides the pull template
 */
Client.ScrollContainer.prototype.hidePullTemplate = function ()
{
  if (this.pullDomObj) {
    if (!this.pullDomObj.classList.contains("refreshing") || this.myScroll.y <= 0) {
      this.pullDomObj.classList.remove("visible");
      this.pullDomObj.classList.remove("refreshing");
      this.hidePullTemplateOnReturning = false;
    }
    else
      this.hidePullTemplateOnReturning = true;
  }
};


/**
 * Show the reveal template
 */
Client.ScrollContainer.prototype.showRevealTemplate = function ()
{
  if (this.revealDomObj && this.myScroll)
    this.myScroll.scrollTo(0, 0);
};


/**
 * hides the reveal template
 */
Client.ScrollContainer.prototype.hideRevealTemplate = function ()
{
  if (this.revealDomObj && this.myScroll) {
    var _this = this;
    window.setTimeout(function () {
      _this.myScroll.scrollTo(0, -_this.revealDomObj.offsetHeight);
    }, 0);
  }
};


/**
 * If there is a 'top element' in the list that is not the template returns its dimension
 * (for using with the reveal scrollcontainer + usewindow datamap)
 */
Client.ScrollContainer.prototype.getListTopOffset = function ()
{
  if (this.revealDomObj)
    return this.revealDomObj.offsetHeight;
  else
    return 0;
};


/**
 * If there is a 'top element' in the list that is not the template returns its dimension
 * (for using with the reveal scrollcontainer + usewindow datamap)
 */
Client.ScrollContainer.prototype.hasRealChildren = function ()
{
  for (var i = 0; i < this.ne(); i++) {
    var el = this.elements[i];
    if (el.visible && el.domObj && el !== this.tmpl && el.domObj !== this.pullDomObj && el.domObj !== this.revealDomObj)
      return true;
  }
  return false;
};