/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2015
 * All rights reserved
 */

/* global Client, SignaturePad */

/**
 * @class A signature pad element
 * @param {Object} element - the element description
 * @param {View|Element} parent - the parent element
 * @param {View} view - the view containing the input
 */
Client.SignaturePad = function (element, parent, view)
{
  Client.Element.call(this, element, parent, view);
  //
  this.domObj = document.createElement("canvas");
  this.options = {};
  this.drawPad();
  parent.appendChildObject(this, this.domObj);
  this.updateElement(element);
  this.attachEvents(element.events);
  this.resizePad();
};

// Make Client.SignaturePad extend Client.Element
Client.SignaturePad.prototype = new Client.Element();


/**
 * Update element properties
 * @param {Object} el - properties to update
 */
Client.SignaturePad.prototype.updateElement = function (el)
{
  this.purgeMyProp(el);
  //
  var needsUpdate = false;
  if (el.dotSize !== undefined) {
    this.options.dotSize = el.dotSize;
    needsUpdate = true;
    delete el.widdotSizeth;
  }
  //
  if (el.minWidth !== undefined) {
    this.options.minWidth = el.minWidth;
    needsUpdate = true;
    delete el.minWidth;
  }
  //
  if (el.maxWidth !== undefined) {
    this.options.maxWidth = el.maxWidth;
    needsUpdate = true;
    delete el.maxWidth;
  }
  //
  if (el.backgroundColor) {
    this.options.backgroundColor = el.backgroundColor;
    needsUpdate = true;
    delete el.backgroundColor;
  }
  //
  if (el.penColor) {
    this.options.penColor = el.penColor;
    needsUpdate = true;
    delete el.penColor;
  }
  //
  if (el.velocityFilterWeight !== undefined) {
    this.options.velocityFilterWeight = el.velocityFilterWeight;
    needsUpdate = true;
    delete el.velocityFilterWeight;
  }
  //
  var pthis = this;
  if (el.events && el.events.indexOf("onBegin") >= 0) {
    this.options.onBegin = function () {
      var e = [{obj: pthis.id, id: "onBegin"}];
      Client.mainFrame.sendEvents(e);
    };
    needsUpdate = true;
  }
  //
  if (el.events && el.events.indexOf("onEnd") >= 0) {
    this.options.onEnd = function () {
      var e = [{obj: pthis.id, id: "onEnd"}];
      Client.mainFrame.sendEvents(e);
    };
    needsUpdate = true;
  }
  //
  // If something has changed update the signature pad
  if (needsUpdate) {
    this.drawPad();
    needsUpdate = false;
  }
  //
  Client.Element.prototype.updateElement.call(this, el);
};


/*
 * Clears the signature canvas
 * @returns {undefined}
 */
Client.SignaturePad.prototype.clear = function ()
{
  this.pad.clear();
};


/*
 * Returns whether the canvas is empty or not
 * @returns {Bool}
 */
Client.SignaturePad.prototype.isEmpty = function (cbId)
{
  var isEmpty = this.pad.isEmpty();
  var e = [{obj: this.id, id: "cb", content: {res: isEmpty, cbId: cbId}}];
  Client.mainFrame.sendEvents(e);
};


/*
 * Makes a post request uploading the signature png file
 */
Client.SignaturePad.prototype.uploadSignature = function ()
{
  // Returns signature image as data URL
  var signatureDataUrl = this.pad.toDataURL();
  //
  // convert base64 string into bynary blob (in our case image/png)
  var signatureBlob = Client.Utils.base64FileDecode(signatureDataUrl);
  //
  // create querystring for post. convert base64 pic into binary blob
  var qstring = Client.Utils.getRESTQueryString({msgType: "signature-pad", objId: this.id});
  //
  // generate multipart form and send it
  var formData = new FormData();
  formData.append("file", signatureBlob, "signature.png");
  var req = new XMLHttpRequest();
  req.open("POST", qstring, true);
  req.send(formData);
};


/*
 * Draw the signature pad
 */
Client.SignaturePad.prototype.drawPad = function ()
{
  this.pad = new SignaturePad(this.domObj, this.options);
};


Client.SignaturePad.prototype.onUpload = function (picture)
{
};


/**
 * Resize the element when the document view is resized
 * @param {Event} ev - the event occured when the document view was resized
 */
Client.SignaturePad.prototype.onResize = function (ev)
{
  this.resizePad();
  //
  Client.Element.prototype.onResize.call(this, ev);
};


/**
 * Resize the signature pad
 */
Client.SignaturePad.prototype.resizePad = function ()
{
  var ratio = Math.max(window.devicePixelRatio || 1, 1);
  this.domObj.width = this.domObj.offsetWidth * ratio;
  this.domObj.height = this.domObj.offsetHeight * ratio;
  this.domObj.getContext("2d").scale(ratio, ratio);
  this.pad.clear(); // otherwise isEmpty() might return incorrect value
};
