/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module */


var App = App || {};


/**
 * @class ActiveDirectory
 * Represents an active directory object
 * @param {Object} app
 * @param {Object} config
 * */
App.ActiveDirectory = function (app, config)
{
  if (!module)
    throw new Error("Cannot create App.ActiveDirectory instance in offline mode");
  //
  if (module && !App.View)
    App = require("./requireAll");
  //
  this.config = config;
  //
  if (this.config.cc)
    this.remotead = new App.RemotePlugin(app, this.config.cc, this.config.APIKey);
  else
    this.ad = new require("activedirectory")(this.config);
};


/*
 * Execute a command
 * @param {type} cid
 * @param {type} args
 */
App.ActiveDirectory.prototype.exec = function (cid, args)
{
  var argsArray = [];
  for (var i = 0; i < args.length - 1; i++)
    argsArray.push(args[i]);
  //
  if (this.config.cc)
    this.remotead.callMethod("ActiveDirectory", cid, argsArray, args[args.length - 1]);
  else {
    var cb = args[args.length - 1];
    var newCb = function (err, res) {
      cb(res, err);
    };
    argsArray.push(newCb);
    //
    this.ad[cid].apply(this.ad, argsArray);
  }
};


/**
 * Execute authentication using username and password
 * @param {String} username
 * @param {String} password
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.authenticate = function (username, password, callback)
{
  this.exec("authenticate", arguments);
};


/**
 * Return true if username is member of groupName
 * @param {String} username
 * @param {String} groupName
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.isUserMemberOf = function (username, groupName, options, callback)
{
  // Make options to be the first parameter
  this.exec("isUserMemberOf", [options, username, groupName, callback]);
};


/**
 * Perform a generic search and return both groups and users that match the specified filter
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.find = function (options, callback)
{
  this.exec("find", [options || {}, callback]);
};


/**
 * Find a user by given username
 * @param {String} username
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.findUser = function (username, options, callback)
{
  // Make options to be the first parameter
  this.exec("findUser", [options, username, callback]);
};


/**
 * Find a group by given groupName
 * @param {String} groupName
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.findGroup = function (groupName, options, callback)
{
  // Make options to be the first parameter
  this.exec("findGroup", [options, groupName, callback]);
};


/**
 * Find users that match the specified filter
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.findUsers = function (options, callback)
{
  this.exec("findUsers", arguments);
};


/**
 * Find groups that match the specified filter
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.findGroups = function (options, callback)
{
  this.exec("findGroups", arguments);
};


/**
 * Return true if groupName exists
 * @param {String} groupName
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.groupExists = function (groupName, options, callback)
{
  // Make options to be the first parameter
  this.exec("groupExists", [options, groupName, callback]);
};


/**
 * Return true if username exists
 * @param {String} username
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.userExists = function (username, options, callback)
{
  // Make options to be the first parameter
  this.exec("userExists", [options, username, callback]);
};


/**
 * Get all of the groups that groupName is a member of
 * @param {String} groupName
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.getGroupMembershipForGroup = function (groupName, options, callback)
{
  // Make options to be the first parameter
  this.exec("getGroupMembershipForGroup", [options, groupName, callback]);
};


/**
 * Get all of the groups that username belongs to
 * @param {String} username
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.getGroupMembershipForUser = function (username, options, callback)
{
  // Make options to be the first parameter
  this.exec("getGroupMembershipForUser", [options, username, callback]);
};


/**
 * Get all of the users that belong to groupName
 * @param {String} groupName
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.getUsersForGroup = function (groupName, options, callback)
{
  // Make options to be the first parameter
  this.exec("getUsersForGroup", [options, groupName, callback]);
};


/**
 * Get the root DSE for the specified url
 * @param {String} url
 * @param {Object} attributes
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.getRootDSE = function (url, attributes, callback)
{
  this.exec("getRootDSE", arguments);
};


/**
 * Get items in the active directory recycle bin
 * @param {Object} options
 * @param {Function} callback
 */
App.ActiveDirectory.prototype.findDeletedObjects = function (options, callback)
{
  this.exec("findDeletedObjects", arguments);
};


//export module for node
if (module)
  module.exports = App.ActiveDirectory;
