/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module, process, Client */

var App = App || {};

/**
 * @class Definition of Analytics object
 * @param {Session} app that owns this analytics instance
 * */
App.Analytics = function (app)
{
  if (module && !App.View) {
    App = require("./requireAll");
  }
  //
  this.app = app; // App that owns this analytics instance
  this.enabled = false;
  this.serverUrl = ""; // url of server application to connect
  this.appName = "AnalyticsServer"; // name of app on server that receives/sends data
  this.apiKey; // authentication code
  this.sessionStats = {};
  this.sessionStats.appId = (App.info ? App.info.id : undefined);
  this.sessionStats.buildId = (App.info ? App.info.buildVersion : undefined);
  this.sessionStats.buildName = (App.info ? App.info.buildName : undefined);
  this.sessionStats.sessionId = this.app.id && this.app.id.length === 36 ? this.app.id : App.Utils.generateUID36();
  this.sessionStats.events = [];
  this.sessionStats.exceptions = [];
  this.saveTimer = 30000;
  //
  this.pagesIds = []; // List of shown pages
  this.lastShowIndex = -1;
  this.totIdleTime = 0;
  this.currentIdleTime = 0;
  this.dir;
  //
  this.isSaving = false;
};


App.Analytics.lists = {
  apps: "apps",
  browsers: "browsers",
  os: "os",
  countries: "countries",
  languages: "languages",
  customEvents: "customEvents"
};


App.Analytics.categories = {
  sessions: "sessions",
  users: "users",
  events: "events",
  pageviews: "pageviews",
  exceptions: "exceptions"
};


App.Analytics.sessions = {
  groupBy: {
    duration: "duration",
    weeklyUsage: "weeklyUsage"
  },
  statsType: {
    newUsers: "newUsers",
    returningUsers: "returningUsers"
  }
};


App.Analytics.users = {
  groupBy: {
    place: "place",
    os: "os",
    browser: "browser",
    lang: "lang",
    city: "city"
  },
  statsType: {
    sessions: "sessions",
    sessionsPerUsers: "sessionsPerUsers",
    sessionsDuration: "sessionsDuration",
    interactionsPerPage: "interactionsPerPage",
    interactionsPerMinute: "interactionsPerMinute",
    newUsers: "newUsers",
    returningUsers: "returningUsers",
    allUsers: "allUsers",
    pageviews: "pageviews",
    exceptions: "exceptions",
    customEvent: "customEvent"
  }
};


App.Analytics.events = {
  groupBy: {
    view: "view",
    properties: "properties"
  },
  statsType: {
    interactionsPerPage: "interactionsPerPage",
    interactionsPerSession: "interactionsPerSession",
    interactionsPerPageNewUsers: "interactionsPerPageNewUsers",
    interactionsPerPageReturningUsers: "interactionsPerPageReturningUsers",
    interactionsPerSessionNewUsers: "interactionsPerSessionNewUsers",
    interactionsPerSessionReturningUsers: "interactionsPerSessionReturningUsers"
  }
};


App.Analytics.pageviews = {
  groupBy: {
    view: "view",
    navigationFlow: "navigationFlow"
  },
  statsType: {
    duration: "duration",
    allUsers: "allUsers",
    newUsers: "newUsers",
    returningUsers: "returningUsers",
    interactions: "interactions"
  }
};


App.Analytics.exceptions = {
  groupBy: {
    type: "type"
  },
  statsType: {
    totalExceptions: "totalExceptions",
    sessions: "sessions",
    usersAffected: "usersAffected",
    newUsersAffected: "newUsersAffected",
    returningUsersAffected: "returningUsersAffected"
  }
};


App.Analytics.timeUnit = {
  hour: "hour",
  day: "day",
  week: "week",
  month: "month",
  dayHour: "dayHour",
  weekDay: "weekDay"
};


App.Analytics.cmds = {
  send: "analyticsSend",
  query: "analyticsQuery",
  setAlert: "analyticsSetAlert",
  getAlert: "analyticsGetAlert",
  delete: "analyticsDelete",
  config: "analyticsConfig",
  status: "analyticsStatus"
};


// Time limit to switch session (15 minutes)
App.Analytics.TIMELIMIT = 900000;

// File size limit (500 elements)
App.Analytics.SIZELIMIT = 500;


Object.defineProperty(App.Analytics.prototype, "enabled", {
  get: function () {
    return (this._enabled === true);
  },
  set: function (newValue) {
    if (newValue !== this._enabled) {
      this._enabled = newValue;
      if (this._enabled) {
        this.start();
        //
        setImmediate(function () {
          // Save user data only the first time I enable analytics for this session
          if (!this.sessionStats.userData && this._enabled)
            this.saveUserData();
          //
          // If there isn't a start session timestamp, get it
          if (!this.sessionStats.startTimeStamp)
            this.sessionStats.startTimeStamp = this.app.locale.moment().valueOf();
          //
          // Get session type
          var sessionType = "user";
          if (!this.app.runningSession) {
            if (this.app.sync && this.app.sync.isServerProxy())
              sessionType = "sync";
            else
              sessionType = "rest";
          }
          //
          this.sessionStats.sessionType = sessionType;
        }.bind(this));
        //
        // Start save interval
        if (!App.ide) {
          this.saveInterval = setInterval(function () {
            this.save();
          }.bind(this), this.saveTimer);
        }
      }
      else
        clearInterval(this.saveInterval);
    }
  }
});


/**
 * Save user data
 */
App.Analytics.prototype.saveUserData = function ()
{
  // Check if this is a new or a returning user
  var userid = this.app.cookies.analyticsUserId;
  var isnew = false;
  if (!userid) {
    isnew = true;
    //
    // If new user, generate UID and set as a session cookie
    userid = App.Utils.generateUID24();
    this.app.setCookie("analyticsUserId", userid, {exdays: 999});
  }
  //
  // Save user data
  this.sessionStats.userData = {};
  this.sessionStats.userData.userId = userid;
  this.sessionStats.userData.isNew = isnew;
  this.sessionStats.userData.language = this.app.device.language;
  this.sessionStats.userData.device = {platform: this.app.device.platform, version: this.app.device.version, model: this.app.device.model};
  this.sessionStats.userData.os = {name: this.app.device.operatingSystem, version: this.app.device.operatingSystemVersion};
  this.sessionStats.userData.browser = {name: this.app.device.browserName, version: this.app.device.browserVersion};
  this.sessionStats.userData.screenResolution = this.app.device.screenWidth + "x" + this.app.device.screenHeight;
  //
  // Get IP address and set geolocation
  this.getIP();
};


/**
 * Send session data to server
 */
App.Analytics.prototype.sendData = function ()
{
  if (App.ide || !this.enabled)
    return;
  //
  if (!this.dir)
    return;
  //
  this.dir.list(function (result, error) {
    if (error) {
      console.error("App.Analytics.sendData -", error.message);
      return;
    }
    //
    result = result || [];
    var toSend = [];
    var fileToSendNumber = result.length;
    for (var i = 0; i < result.length; i++) {
      var f = result[i];
      //
      // Check if file is ready to be sent
      if (f.path.indexOf("__@") !== -1) {
        f.readAll(function (f, result, error) {
          fileToSendNumber--;
          if (error) {
            console.error("App.Analytics.sendData -", error.message);
            return;
          }
          //
          try {
            var data = JSON.parse(result);
            toSend.push({data: data, file: f});
          }
          catch (ex) {
            f.remove(function (result, error) {
              if (error)
                console.error("App.Analytics.sendData - ", error);
            }.bind(this));
          }
          //
          if (fileToSendNumber <= 0)
            this.cascadeSend(toSend);
        }.bind(this, f));
      }
      else {
        fileToSendNumber--;
        if (fileToSendNumber <= 0)
          this.cascadeSend(toSend);
        //
        // The file could be ready to be sent. Try to detect if it's old
        f.dateTime(function (f, result, error) {
          if (error) {
            console.error("App.Analytics.sendData - ", error);
            return;
          }
          //
          // If the file it's old, rename it. It's closed
          if (new Date().getTime() - new Date(result).getTime() > App.Analytics.TIMELIMIT * 2) {
            var fileName = f.path.substring(f.path.lastIndexOf("/") + 1, f.path.length);
            var newName = fileName.replace(".json", "__@.json");
            f.rename(newName, function (result, error) {
              if (error) {
                console.error("App.Analytics.save - ", error.message);
                return;
              }
            }.bind(this));
          }
        }.bind(this, f));
      }
    }
  }.bind(this));
};


/**
 * Get IP address
 */
App.Analytics.prototype.getIP = function ()
{
  var e = [{id: "getIP"}];
  this.app.sendCommands(e);
};


/**
 * Save ip and set session country and city
 * @param {String} ip
 */
App.Analytics.prototype.setGeolocation = function (ip)
{
  // Save ip
  this.sessionStats.userData.ip = ip;
  //
  // Perform a GET request to know country and city from ip address
  var u = this.app.fs.url("https://www.freegeoip.net/json/" + ip);
  u.get(undefined, function (res, err) {
    var geo = {country: "Unknown", city: "Unknown"};
    if (err) {
      this.sessionStats.userData.geolocation = geo;
      return;
    }
    // Parse data
    if (res && res.status === 200) {
      var geo = JSON.parse(res.body);
      //
      // Set geolocation
      this.sessionStats.userData.geolocation = {country: geo.country_name || "Unknown", city: geo.city || "Unknown"};
    }
  }.bind(this));
};


/**
 * Save exception
 * @param {Object} ex
 */
App.Analytics.prototype.logException = function (ex)
{
  if (!this.enabled || App.ide)
    return;
  //
  var exception = {};
  exception.id = App.Utils.generateUID24();
  exception.timeStamp = this.app.locale.moment().valueOf();
  exception.msg = ex.msg;
  exception.uncought = ex.uncought || false;
  //
  // Parse stack trace
  var stack = ex.stack || "";
  var newStack = "";
  //
  // Split stack trace at end line
  var lines = stack.split("\n");
  for (var i = 0; i < lines.length; i++) {
    // Find .js file
    var jsIndex = lines[i].indexOf(".js");
    var startIndex = -1;
    var endIndex = null;
    //
    // Get file name start
    while (jsIndex >= 0) {
      if (lines[i].charAt(jsIndex) === "/" || lines[i].charAt(jsIndex) === "(") {
        startIndex = jsIndex + 1;
        break;
      }
      //
      jsIndex--;
    }
    //
    if (startIndex >= 0) {
      jsIndex = lines[i].indexOf(".js");
      //
      // If I found a start point, try to get end point
      while (jsIndex < lines[i].length) {
        if (lines[i].charAt(jsIndex) === ")") {
          endIndex = jsIndex;
          break;
        }
        //
        jsIndex++;
      }
      //
      // Get substring I need
      var newLine = lines[i].substring(startIndex, endIndex);
      //
      // Add new line to parsed stack trace
      if (newLine)
        newStack += newLine.trim() + "\n";
    }
  }
  //
  exception.stack = newStack || ex.stack;
  //
  this.sessionStats.exceptions.push(exception);
};


/**
 * Save given event and set it as custom if needed
 * @param {String/Object} ev
 * @param {Object} params
 */
App.Analytics.prototype.logEvent = function (ev, params)
{
  if (!this.enabled || App.ide)
    return;
  //
  // If no event to log, do nothing..
  if (!ev)
    return;
  //
  var i;
  var evObj = {};
  //
  // If the event is a string, it means is the name of a custom event
  if (typeof ev === "string") {
    evObj.custom = true;
    evObj.eventName = ev;
  }
  else if (typeof ev === "object") {
    evObj = ev;
    evObj.custom = false;
    //
    if (evObj.eventName === "analytics_show") {
      // If ev is a show event and parentViewId is unkonwn, push this id into pages array
      if (this.pagesIds.indexOf(evObj.parentViewId) === -1)
        this.pagesIds.push(evObj.parentViewId);
      //
      // Remember position of this show event
      this.lastShowIndex = this.sessionStats.events.length;
    }
    else {
      // Add current idle time to last show event idle time
      if (this.sessionStats.events[this.lastShowIndex]) {
        this.sessionStats.events[this.lastShowIndex].currentIdleTime = this.sessionStats.events[this.lastShowIndex].currentIdleTime || 0;
        this.sessionStats.events[this.lastShowIndex].currentIdleTime += this.currentIdleTime;
      }
    }
    //
    // Try to detect proper parent page for the event
    var pageId;
    for (i = 0; i < this.pagesIds.length; i++) {
      // If event parent page is a known page, this event belongs to that page
      if (this.pagesIds[i] === evObj.parentViewId) {
        pageId = evObj.parentViewId;
        break;
      }
      //
      // Otherwise make event belong to last opened page
      pageId = this.pagesIds[i];
    }
    evObj.parentViewId = pageId;
  }
  //
  evObj.id = App.Utils.generateUID24();
  evObj.timeStamp = this.app.locale.moment().valueOf();
  evObj.params = params || {};
  //
  evObj.currentIdleTime = this.currentIdleTime;
  evObj.totalIdleTime = this.totIdleTime;
  this.sessionStats.events.push(evObj);
  //
  this.currentIdleTime = 0;
};


/**
 * Extract data
 * @param {Object} queryObj
 *                    - category: stats category (App.Analytics.categories)
 *                    - groupBy: groupBy value (App.Analytics.[category].groupBy)
 *                    - statsType: stats type required for this grouping (App.Analytics.[category].statsType)
 *                    - params: optional parameters
 *                    - filter: optional query filter
 *                    - serverUrl: server to post to
 *                    - appName: name of server app to post to
 * @param {Function} callback
 */
App.Analytics.prototype.query = function (queryObj, callback)
{
  if (!queryObj || !queryObj.category) {
    if (callback)
      callback(undefined, "Cannot execute query: missing category property");
    return;
  }
  //
  if (!App.Analytics.lists[queryObj.category]) {
    // Check for errors in filter property
    if (!queryObj.filter) {
      if (callback)
        callback(undefined, "Cannot execute query: no filter set");
      return;
    }
    else {
      if (!queryObj.filter.apps) {
        if (callback)
          callback(undefined, "Cannot execute query: missing apps property in filter");
        return;
      }
      else if (!queryObj.filter.from) {
        if (callback)
          callback(undefined, "Cannot execute query: missing from property in filter");
        return;
      }
      else if (!queryObj.filter.to) {
        if (callback)
          callback(undefined, "Cannot execute query: missing to property in filter");
        return;
      }
      else if (!queryObj.filter.compareWith || ((!queryObj.filter.compareWith.from ||
              !queryObj.filter.compareWith.to) && !queryObj.filter.compareWith.build)) {
        if (callback)
          callback(undefined, "Cannot execute query: bad compareWith property in filter");
        return;
      }
    }
  }
  queryObj.filter = queryObj.filter || {};
  //
  // If user set a string as apps filter, make apps filter to be an array
  if (queryObj.filter && typeof queryObj.filter.apps === "string")
    queryObj.filter.apps = [queryObj.filter.apps];
  //
  // Make params to be an object in it is not
  if (!queryObj.params || typeof queryObj.params !== "object")
    queryObj.params = {};
  //
  // Get serverUrl and appName
  var serverUrl = queryObj.serverUrl;
  var appName = queryObj.appName;
  var apiKey = queryObj.apiKey;
  delete queryObj.serverUrl;
  delete queryObj.appName;
  delete queryObj.apiKey;
  //
  // Send post request
  this.postRequest({cmd: App.Analytics.cmds.query, data: queryObj, serverUrl: serverUrl, appName: appName, autk: apiKey}, function (res, err) {
    if (callback)
      callback(res, err);
  });
};


/**
 * Execute a post request
 * @param {Object} options - data: to send
 *                         - cmd: command to execute server side
 *                         - serverUrl: server to post to
 *                         - appName: name of the app to post to
 * @param {Function} callback
 */
App.Analytics.prototype.postRequest = function (options, callback)
{
  // Data to send
  var postData = JSON.stringify(options.data || {});
  var serverUrl = options.serverUrl || this.serverUrl;
  var appName = options.appName || this.appName;
  var apiKey = options.autk || this.apiKey;
  //
  // Get url params
  var urlParams = options.urlParams || {};
  var params = Object.keys(urlParams);
  var urlParamsString = "";
  for (var i = 0; i < params.length; i++)
    urlParamsString += "&" + params[i] + "=" + urlParams[params[i]];
  //
  var postOptions = {
    usePlugin: false,
    body: postData,
    headers: {
      "Content-Type": "text/plain"
    }
  };
  //
  var u = this.app.fs.url(serverUrl + "/" + appName + "?mode=rest&cmd=" + options.cmd + "&autk=" + apiKey + urlParamsString);
  u.post(postOptions, function (res) {
    if (res.error) {
      if (callback)
        callback(undefined, res.error.message || res.error);
      //
      return;
    }
    //
    if (res.body) {
      try {
        var body = JSON.parse(res.body);
        if (res.status === 200) {
          if (callback)
            callback(body);
        }
        else {
          if (callback)
            callback(undefined, body.message || body);
        }
      }
      catch (ex) {
        if (callback)
          callback(undefined, res.body);
      }
    }
  });
};


/**
 * Raised on app blur
 */
App.Analytics.prototype.onPause = function ()
{
  if (!this.enabled || App.ide)
    return;
  //
  this.onPauseTimeStamp = this.app.locale.moment().valueOf();
  //
  // Save session end time stamp (session may never be resumed...)
  this.sessionStats.endTimeStamp = this.onPauseTimeStamp;
  //
  // Clear save interval. It will be reset on app resume
  clearInterval(this.saveInterval);
  //
  // Save data on app pause and try to send them to server
  this.save({close: true, send: true});
};


/**
 * Raised on app focus
 */
App.Analytics.prototype.onResume = function ()
{
  if (!this.enabled || App.ide)
    return;
  //
  // Update app idle time
  var lastIdleTime = (this.app.locale.moment().valueOf() - this.onPauseTimeStamp) || 0;
  //
  // If idle time exceeds time limit, treat session as a new session
  if (lastIdleTime > App.Analytics.TIMELIMIT) {
    this.sessionStats.sessionId = App.Utils.generateUID36();
    this.totIdleTime = 0;
    this.currentIdleTime = 0;
    delete this.currentFileName;
  }
  else {
    this.totIdleTime += lastIdleTime;
    this.currentIdleTime += lastIdleTime;
  }
  //
  delete this.onPauseTimeStamp;
  //
  // Start save interval
  this.saveInterval = setInterval(function () {
    this.save();
  }.bind(this), this.saveTimer);
};


/**
 * Create a show event using given view as a new opened page
 * @param {Object} view
 */
App.Analytics.prototype.logPage = function (view)
{
  if (!this.enabled || App.ide || !view)
    return;
  //
  this.logEvent({eleId: view.pid, eleName: view.name || view.className, eleType: "page", eventName: "analytics_show", parentViewId: view.pid});
};


/**
 * Log an uncought exception
 * @param {Object} ex
 */
App.Analytics.logUncaughtException = function (ex)
{
  // If there is no App.sessionMap it means that this file has not executed the
  //    App = require("./requireAll");
  // or, in other words, no one created an instance of the Analytics object. This could happen
  // if an uncaught exception is cought inside app.js and the process has not created a session
  // (like during an app install)
  if (!App.sessionMap)
    return;
  //
  var sids = Object.keys(App.sessionMap);
  var ses;
  //
  // Default parent session for ex is the first session I find in the map
  var chosenSes = App.sessionMap[sids[0]];
  var lastExceptionTime = 0;
  for (var i = 0; i < sids.length; i++) {
    ses = App.sessionMap[sids[i]];
    //
    // Look for a session having analytics plugin enabled and in which there was the last logged exception
    if (ses.analytics && ses.analytics.enabled && ses.sessionStats && ses.sessionStats.exceptions &&
            ses.sessionStats.exceptions[ses.sessionStats.exceptions.length - 1].timeStamp > lastExceptionTime) {
      lastExceptionTime = ses.sessionStats.exceptions[ses.sessionStats.exceptions.length - 1].timeStamp;
      chosenSes = ses;
    }
  }
  //
  // Log exception to chosen session
  if (chosenSes)
    chosenSes.analytics.logException(ex);
};


/**
 * Stop analytics plugin and send last data
 */
App.Analytics.prototype.terminate = function ()
{
  if (App.ide || !this.enabled)
    return;
  //
  // Clear save interval
  clearInterval(this.saveInterval);
  //
  // Save data
  this.save({recursive: true});
  //
  // Disable analytics
  this.enabled = false;
};


/**
 * Send given alert to analytics server
 * @param {Object} options
 * @param {Function} callback
 */
App.Analytics.prototype.setAlert = function (options, callback)
{
  if (!options) {
    callback(undefined, "alert is undefined");
    return;
  }
  else if (!options.alertId) {
    callback(undefined, "Missing alertId property");
    return;
  }
  else if (!options.appId) {
    callback(undefined, "Missing appId property");
    return;
  }
  //
  // Get serverUrl and appName
  options = options || {};
  var serverUrl = options.serverUrl;
  var appName = options.appName;
  var apiKey = options.apiKey;
  delete options.serverUrl;
  delete options.appName;
  delete options.apiKey;
  //
  this.postRequest({cmd: App.Analytics.cmds.setAlert, data: options, serverUrl: serverUrl, appName: appName, autk: apiKey}, function (res, err) {
    if (callback)
      callback(res, err);
  }.bind(this));
};


/**
 * Ask to server the alerts that match options
 * @param {Object} options
 * @param {Function} callback
 */
App.Analytics.prototype.getAlerts = function (options, callback)
{
  options = options || {};
  //
  // Get serverUrl and appName
  var serverUrl = options.serverUrl;
  var appName = options.appName;
  var apiKey = options.apiKey;
  delete options.serverUrl;
  delete options.appName;
  delete options.apiKey;
  //
  this.postRequest({cmd: App.Analytics.cmds.getAlert, data: options, serverUrl: serverUrl, appName: appName, autk: apiKey}, function (res, err) {
    if (callback)
      callback(res, err);
  }.bind(this));
};


/**
 * Create directory where to save analytics files
 */
App.Analytics.prototype.createDir = function ()
{
  var dir = this.app.fs.directory("_logs");
  dir.exists(function (result, error) {
    if (error) {
      console.error("App.Analytics.createDir - ", error.message);
      return;
    }
    //
    if (!result) {
      dir.create(function (err) {
        if (err) {
          console.error("App.Analytics.createDir", err.message);
          return;
        }
        //
        this.dir = dir;
      }.bind(this));
    }
    else
      this.dir = dir;
  }.bind(this));
};


/**
 * Save analytics data
 * @param {Object} options :
 *                         - recursive: true if need to close file recursively
 *                         - close: true if need to force file closing
 *                         - send: true if need to send data after saving
 */
App.Analytics.prototype.save = function (options)
{
  if (App.ide || !this.enabled || this.isSaving)
    return;
  //
  if (!this.dir)
    return;
  //
  this.isSaving = true;
  //
  options = options || {};
  //
  var writeData = function (filePath, data, options) {
    options = options || {};
    var f = this.app.fs.file(filePath);
    //
    f.create(undefined, function (result, error) {
      if (error) {
        console.error("App.Analytics.save - ", error.message);
        this.isSaving = false;
        return;
      }
      //
      f.write(JSON.stringify(data), undefined, undefined, undefined, function (result, error) {
        if (error) {
          console.error("App.Analytics.save - ", error.message);
          this.isSaving = false;
          return;
        }
        // Mark the file as closed adding "__@" at the end of his name
        if (options.close) {
          var fileName = f.path.substring(f.path.lastIndexOf("/") + 1, f.path.length);
          var newName = fileName.replace(".json", "__@.json");
          f.rename(newName, function (result, error) {
            if (error) {
              console.error("App.Analytics.save - ", error.message);
              return;
            }
            //
            // Send data right now if needed
            if (options.send)
              this.sendData();
            //
            // Delete current file name. It cannot be used anymore because the file is full
            delete this.currentFileName;
            //
            if (options.recursive)
              this.save(options);
          }.bind(this));
        }
        //
        this.isSaving = false;
      }.bind(this));
    }.bind(this));
  }.bind(this);
  //
  // Set end session time stamp
  // If the app is paused, the session end time is the last pause time (excluding idle time)
  if (this.onPauseTimeStamp)
    this.sessionStats.endTimeStamp = this.onPauseTimeStamp - (this.totIdleTime || 0);
  else // Otherwise set current time as session end time (excluding idle time)
    this.sessionStats.endTimeStamp = this.app.locale.moment().valueOf() - (this.totIdleTime || 0);
  //
  this.currentFileName = this.currentFileName || App.Utils.generateUID36();
  //
  var filePath = this.dir.path + "/" + this.currentFileName + ".json";
  var f = this.app.fs.file(filePath);
  //
  f.readAll(function (result, error) {
    if (error && error.code !== "ENOENT" && error.message !== "NOT_FOUND_ERR") {
      console.error("App.Analytics.save - ", error.message);
      this.isSaving = false;
      return;
    }
    //
    // Get data
    try {
      var data = result || JSON.stringify({events: [], exceptions: []});
      data = JSON.parse(data);
      //
      // Get size of data saved and of data I'm going to save
      var dataSize = data.events.length + data.exceptions.length;
      var sessionStatsSize = this.sessionStats.events.length + this.sessionStats.exceptions.length;
      //
      var i;
      if (dataSize + sessionStatsSize > App.Analytics.SIZELIMIT) {
        var inserted = 0;
        //
        // Push on data all events that I can
        for (i = 0; i < this.sessionStats.events.length; i++) {
          dataSize = data.events.length + data.exceptions.length;
          if (dataSize < App.Analytics.SIZELIMIT) {
            data.events.push(this.sessionStats.events[i]);
            inserted++;
          }
          else
            break;
        }
        //
        // Delete from session the events just inserted
        this.sessionStats.events.splice(0, inserted);
        //
        inserted = 0;
        //
        // Push on data all exceptions that I can
        for (i = 0; i < this.sessionStats.exceptions.length; i++) {
          dataSize = data.events.length + data.exceptions.length;
          if (dataSize < App.Analytics.SIZELIMIT) {
            data.exceptions.push(this.sessionStats.exceptions[i]);
            inserted++;
          }
          else
            break;
        }
        //
        // Delete from session the exceptions just inserted
        this.sessionStats.exceptions.splice(0, inserted);
        //
        // Copy all property but events and exceptions from sessionStats to data
        var sessionStatsKeys = Object.keys(this.sessionStats);
        for (i = 0; i < sessionStatsKeys.length; i++) {
          if (["events", "exceptions"].indexOf(sessionStatsKeys[i]) !== -1)
            continue;
          //
          data[sessionStatsKeys[i]] = this.sessionStats[sessionStatsKeys[i]];
        }
        //
        // Write data and close the file. It's full
        options.close = true;
        writeData(filePath, data, options);
      }
      else {
        // No new data to save. Do nothing
        if (dataSize + sessionStatsSize === 0) {
          this.isSaving = false;
          return;
        }
        //
        // Concat old data with new ones
        for (i = 0; i < this.sessionStats.events.length; i++)
          data.events.push(this.sessionStats.events[i]);
        //
        for (i = 0; i < this.sessionStats.exceptions.length; i++)
          data.exceptions.push(this.sessionStats.exceptions[i]);
        //
        // Delete all events and exceptions
        this.sessionStats.events = [];
        this.sessionStats.exceptions = [];
        //
        // Copy all property but events and exceptions from sessionStats to data
        var sessionStatsKeys = Object.keys(this.sessionStats);
        for (i = 0; i < sessionStatsKeys.length; i++) {
          if (["events", "exceptions"].indexOf(sessionStatsKeys[i]) !== -1)
            continue;
          //
          data[sessionStatsKeys[i]] = this.sessionStats[sessionStatsKeys[i]];
        }
        //
        // Write data on file
        delete options.recursive;
        writeData(filePath, data, options);
      }
    }
    catch (ex) {
      f.remove(function (result, error) {
        if (error)
          console.error("App.Analytics.save - ", error);
        //
        this.isSaving = false;
      }.bind(this));
    }
  }.bind(this));
};


/**
 * Send analytics data to server
 */
App.Analytics.staticSend = function ()
{
  var keys = Object.keys(App.sessionMap);
  for (var i = 0; i < keys.length; i++) {
    var ses = App.sessionMap[keys[i]];
    if (ses.analytics) {
      ses.analytics.sendData();
      break;
    }
  }
};


/**
 * Start analytics plugin
 */
App.Analytics.prototype.start = function ()
{
  if (!App.ide && this.enabled) {
    // Wait for fs
    var fsInterval = setInterval(function () {
      if (this.app.fs) {
        clearInterval(fsInterval);
        //
        //Create directory if outside IDE
        this.createDir();
        //
        // If not already set, start send interval
        if (!App.sendAnalyticsInterval) {
          App.sendAnalyticsInterval = setInterval(function () {
            App.Analytics.staticSend();
          }, 17000);
        }
      }
    }.bind(this), 300);
  }
};


/**
 * Send session data to server
 * @param {Object} options
 * @param {Function} callback
 */
App.Analytics.prototype.send = function (options, callback)
{
  if (!this.enabled || App.ide) {
    if (callback) {
      if (!this.enabled)
        callback(undefined, "analytics plugin is not enabled");
      else if (App.ide)
        callback(undefined, "Cannot send data from IDE");
    }
    return;
  }
  //
  // Set end session time stamp
  // If the app is paused, the session end time is the last pause time (excluding idle time)
  if (this.onPauseTimeStamp)
    this.sessionStats.endTimeStamp = this.onPauseTimeStamp - (this.totIdleTime || 0);
  else // Otherwise set current time as session end time (excluding idle time)
    this.sessionStats.endTimeStamp = new Date(this.app.locale.moment()._d).getTime() - (this.totIdleTime || 0);
  //
  options = options || {};
  this.postRequest({cmd: App.Analytics.cmds.send, data: this.sessionStats, serverUrl: options.serverUrl, appName: options.appName, autk: options.apiKey}, function (res, err) {
    if (callback)
      callback(res, err);
  }.bind(this));
};


/**
 * Send a set of session data when previous response is received
 * @param {Array} dataArray
 */
App.Analytics.prototype.cascadeSend = function (dataArray)
{
  if (!dataArray || dataArray.length <= 0)
    return;
  //
  var item = dataArray.pop();
  var data = item.data || {};
  var f = item.file;
  this.postRequest({cmd: App.Analytics.cmds.send, data: data, autk: this.apiKey}, function (f, res, err) {
    // Recall function until the array is empty
    setTimeout(function () {
      this.cascadeSend(dataArray);
    }.bind(this), 0);
    //
    if (err)
      return;
    //
    // The data have been saved. I can delete the file
    if (res && f) {
      f.remove(function (result, error) {
        if (error)
          return;
      }.bind(this));
    }
  }.bind(this, f));
};


/**
 * Delete a set of data
 * @param {Object} options
 * @param {Function} callback
 */
App.Analytics.prototype.delete = function (options, callback)
{
  if (App.ide)
    return;
  //
  options = options || {};
  //
  // Get serverUrl and appName
  var serverUrl = options.serverUrl;
  var appName = options.appName;
  var apiKey = options.apiKey;
  delete options.serverUrl;
  delete options.appName;
  delete options.apiKey;
  //
  this.postRequest({cmd: App.Analytics.cmds.delete, data: options, serverUrl: serverUrl, appName: appName, autk: apiKey}, function (res, err) {
    if (callback)
      callback(res, err);
  }.bind(this));
};


/**
 * Set analytics server configuration
 * @param {Object} options
 * @param {Function} callback
 */
App.Analytics.prototype.config = function (options, callback)
{
  if (App.ide)
    return;
  //
  options = options || {};
  //
  // Get serverUrl, appName and urlParams
  var serverUrl = options.serverUrl;
  var appName = options.appName;
  var urlParams = options.urlParams;
  var apiKey = options.apiKey;
  delete options.serverUrl;
  delete options.appName;
  delete options.urlParams;
  delete options.apiKey;
  //
  this.postRequest({cmd: App.Analytics.cmds.config, data: options, serverUrl: serverUrl, appName: appName, urlParams: urlParams, autk: apiKey}, function (res, err) {
    if (callback)
      callback(res, err);
  }.bind(this));
};


/**
 * Get configuration status
 * @param {Object} options
 * @param {Function} callback
 */
App.Analytics.prototype.getStatus = function (options, callback)
{
  if (App.ide)
    return;
  //
  options = options || {};
  //
  // Get serverUrl, appName and urlParams
  var serverUrl = options.serverUrl;
  var appName = options.appName;
  var urlParams = options.urlParams;
  var apiKey = options.apiKey;
  delete options.serverUrl;
  delete options.appName;
  delete options.urlParams;
  delete options.apiKey;
  //
  this.postRequest({cmd: App.Analytics.cmds.status, data: options, serverUrl: serverUrl, appName: appName, urlParams: urlParams, autk: apiKey}, function (res, err) {
    if (callback)
      callback(res, err);
  }.bind(this));
};


//export module for node
if (module)
  module.exports = App.Analytics;
