/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, Uint8Array, Buffer, ArrayBuffer */


var App = App || {};


/**
 * @class Definition of crypt object
 * */
App.Crypt = function ()
{
  if (module && !App.View)
    App = require("./requireAll");
};


// Set crypto engine
App.Crypt.engine = module ? require("crypto") : window.crypto;

// Set default encrypting algorithm
App.Crypt.algorithm = {name: "AES-CBC", keyLength: 256};


/**
 * Generate encrypting key
 * @param {Session} app
 * @param {function} cb
 */
App.Crypt.generateKey = function (app, cb)
{
  // Set algorithm
  var algorithm = App.Crypt.algorithm;
  //
  if (module) {
    // Create random initialization vector
    var iv = App.Crypt.engine.randomBytes(16);
    //
    // Generate random password
    var password = App.Crypt.engine.randomBytes(algorithm.keyLength / 8);
    //
    // Return key object
    App.Crypt.exportKey(app, {iv: iv, password: password}, function (key, error) {
      if (cb)
        cb(key, error);
    });
  }
  else {
    var algo = {};
    //
    // Set algorithm
    algo.name = algorithm.name;
    algo.length = algorithm.keyLength;
    //
    // Create random initialization vector
    algo.iv = App.Crypt.engine.getRandomValues(new Uint8Array(16));
    //
    // Generate key
    App.Crypt.engine.subtle.generateKey(algo, true, ["encrypt", "decrypt"]).then(function (key) {
      App.Crypt.exportKey(app, {iv: algo.iv, password: key}, function (res, err) {
        if (cb)
          cb(res, err);
      });
    }).catch(function (error) {
      if (cb)
        cb(null, error);
    });
  }
};


/**
 * Encrypt data
 * @param {Session} app
 * @param {String} data
 * @param {String} key
 * @param {Object} options
 * @param {function} cb
 */
App.Crypt.encrypt = function (app, data, key, options, cb)
{
  if (module)
    App.Utils = require("./utils");
  //
  if (!key)
    return cb(null, new Error("key cannot be null"));
  if (!data)
    return cb(null, new Error("No data to encrypt"));
  //
  // Adapt key to make it an AES-CBC 256 bits key (fill missing bits or slice extra bits)
  if (key.length < 64)
    key = App.Utils.toBase64(key + new Array(48 - App.Utils.stringToBuffer(key).byteLength + 1).join("a"));
  else if (key.length > 64)
    key = App.Utils.toBase64(App.Utils.bufferToString(App.Utils.stringToBuffer(key).slice(0, 48)));
  //
  options = options || {};
  //
  // Set algorithm
  var algorithm = App.Crypt.algorithm;
  //
  // Create CryptoKey object using given key
  App.Crypt.importKey(app, {raw: key, algorithm: algorithm}, function (res, error) {
    if (error)
      return cb(null, error);
    //
    key = res;
    //
    if (module) {
      var algoTokens = algorithm.name.split("-");
      var algo = algoTokens[0] + "-" + algorithm.keyLength + "-" + algoTokens[1];
      //
      // Check if algorithm is supported by Node.crypto
      if (App.Crypt.engine.getCiphers().indexOf(algo.toLowerCase()) === -1)
        return cb(null, new Error("Algorithm " + algo + " is not supported"));
      //
      try {
        // Encrypt data
        var cipher = App.Crypt.engine.createCipheriv(algo, key.password, key.iv);
        var encrypted = cipher.update(new Buffer(data));
        encrypted = Buffer.concat([encrypted, cipher.final()]);
        //
        // Return encrypted data
        cb(App.Crypt.getOutput(encrypted, options.output, "encrypt"));
      } catch (ex) {
        cb(null, ex);
      }
    }
    else {
      var algo = {};
      algo.name = algorithm.name;
      algo.length = algorithm.keyLength;
      algo.iv = key.iv;
      //
      // Encrypt data
      var dataBuffer = typeof data === "string" ? App.Utils.stringToBuffer(data) : new Uint8Array(data).buffer;
      App.Crypt.engine.subtle.encrypt(algo, key.password, dataBuffer).then(function (encrypted) {
        // Return encrypted data
        if (cb)
          cb(App.Crypt.getOutput(encrypted, options.output, "encrypt"));
      }).catch(function (error) {
        if (cb)
          cb(null, error);
      });
    }
  });
};


/**
 * Decrypt data
 * @param {Session} app
 * @param {String} data
 * @param {String} key
 * @param {Object} options
 * @param {function} cb
 */
App.Crypt.decrypt = function (app, data, key, options, cb)
{
  if (module)
    App.Utils = require("./utils");
  //
  if (!key)
    return cb(null, new Error("key cannot be null"));
  if (!data)
    return cb(null, new Error("No data to decrypt"));
  //
  // Adapt key to make it an AES-CBC 256 bits key (fill missing bits or slice extra bits)
  if (key.length < 64)
    key = App.Utils.toBase64(key + new Array(48 - App.Utils.stringToBuffer(key).byteLength + 1).join("a"));
  else if (key.length > 64)
    key = App.Utils.toBase64(App.Utils.bufferToString(App.Utils.stringToBuffer(key).slice(0, 48)));
  //
  options = options || {};
  //
  // If data is a base64 string, convert it to buffer
  data = typeof data === "string" ? App.Utils.base64ToBuffer(data) : data;
  //
  // Set algorithm
  var algorithm = App.Crypt.algorithm;
  //
  // Create CryptoKey object using given key
  App.Crypt.importKey(app, {raw: key, algorithm: algorithm}, function (res, error) {
    if (error)
      return cb(null, error);
    //
    key = res;
    //
    if (module) {
      var algoTokens = algorithm.name.split("-");
      var algo = algoTokens[0] + "-" + algorithm.keyLength + "-" + algoTokens[1];
      //
      // Check if algorithm is supported by Node.crypto
      if (App.Crypt.engine.getCiphers().indexOf(algo.toLowerCase()) === -1)
        return cb(null, new Error("Algorithm " + algo + " is not supported"));
      //
      try {
        // Decrypt data
        var decipher = App.Crypt.engine.createDecipheriv(algo, new Buffer(key.password), key.iv);
        var decrypted = decipher.update(new Buffer(data));
        decrypted = Buffer.concat([decrypted, decipher.final()]);
        //
        // Return decrypted data
        cb(App.Crypt.getOutput(decrypted, options.output, "decrypt"));
      } catch (ex) {
        cb(null, ex);
      }
    }
    else {
      var algo = {};
      algo.name = algorithm.name;
      algo.length = algorithm.keyLength;
      algo.iv = key.iv;
      var dataBuffer = typeof data === "string" ? App.Utils.stringToBuffer(data) : new Uint8Array(data).buffer;
      App.Crypt.engine.subtle.decrypt(algo, key.password, dataBuffer).then(function (decrypted) {
        // Return decrypted data
        if (cb)
          cb(App.Crypt.getOutput(decrypted, options.output, "decrypt"));
      }).catch(function (error) {
        if (cb)
          cb(null, error);
      });
    }
  });
};


/**
 * Export a key
 * @param {Session} app
 * @param {Object} key to export
 * @param {function} cb
 */
App.Crypt.exportKey = function (app, key, cb)
{
  if (!key || typeof key !== "object")
    return cb(null, new Error("Invalid key"));
  //
  var iv, pwd, k;
  if (module) {
    if (cb) {
      iv = new Uint8Array(key.iv);
      pwd = new Uint8Array(key.password);
      k = new Uint8Array(iv.length + pwd.length);
      k.set(iv);
      k.set(pwd, iv.length);
      cb(Buffer.from(k).toString("base64"));
    }
  }
  else {
    App.Crypt.engine.subtle.exportKey("raw", key.password).then(function (result) {
      if (cb) {
        iv = new Uint8Array(key.iv);
        pwd = new Uint8Array(result);
        k = new Uint8Array(iv.length + pwd.length);
        k.set(iv);
        k.set(pwd, iv.length);
        cb(btoa(String.fromCharCode.apply(null, k)));
      }
    }).catch(function (error) {
      if (cb)
        cb(null, error);
    });
  }
};


/**
 * Import a key
 * @param {Session} app
 * @param {Object} key to import
 *                  - raw: raw key data
 *                  - algorithm: algorithm to uso for importo key
 * @param {function} cb
 */
App.Crypt.importKey = function (app, key, cb)
{
  if (!key)
    return cb(null, new Error("Invalid key"));
  //
  key.algorithm = key.algorithm || App.Crypt.algorithm;
  //
  var binaryKey, iv, password;
  if (module) {
    // Reconstruct iv and password from base64 key
    binaryKey = new Uint8Array(Buffer.from(key.raw, "base64"));
    iv = binaryKey.slice(0, 16);
    password = binaryKey.slice(16);
    if (cb)
      cb({iv: iv, password: password});
  }
  else {
    var algo = {};
    algo.name = key.algorithm.name;
    algo.length = key.algorithm.keyLength;
    //
    // Reconstruct iv and password from base64 key
    binaryKey = Uint8Array.from(atob(key.raw), function (c) {
      return c.charCodeAt(0);
    });
    iv = binaryKey.slice(0, 16);
    password = binaryKey.slice(16);
    algo.iv = iv;
    App.Crypt.engine.subtle.importKey("raw", password, algo, true, ["encrypt", "decrypt"]).then(function (result) {
      if (cb)
        cb({iv: iv, password: result});
    }).catch(function (error) {
      if (cb)
        cb(null, error);
    });
  }
};


/**
 * Generate private and public RSA keys
 * @param {Session} app
 * @param {function} cb
 */
App.Crypt.generateKeyPair = function (app, cb)
{
  if (module) {
    try {
      var nodeRSA = require("node-rsa");
      //
      // Generate key pair
      var key = new nodeRSA();
      key.generateKeyPair();
      //
      var keyPair = {};
      //
      // Export private key
      keyPair.privateKey = key.exportKey("pkcs8");
      //
      // Export public key
      keyPair.publicKey = key.exportKey("pkcs8-public");
      //
      // Return key pair
      if (cb)
        cb(keyPair);
    } catch (ex) {
      if (cb)
        cb(null, ex);
    }
  }
  else {
    var formatKey = function (keyData, keyType) {
      // Set start and end lines
      var startLine = keyType === "public" ? "-----BEGIN PUBLIC KEY-----\n" : "-----BEGIN PRIVATE KEY-----\n";
      var endLine = keyType === "public" ? "-----END PUBLIC KEY-----" : "-----END PRIVATE KEY-----";
      //
      var finalString = "";
      for (var i = 0; i < keyData.length; i++) {
        finalString += keyData.substring(0, 64) + "\n";
        keyData = keyData.substring(64);
      }
      finalString += keyData;
      //
      // Ensure last key line ends with a \n character
      if (finalString[finalString.length - 1] !== "\n")
        finalString += "\n";
      //
      return startLine + finalString + endLine;
    };
    //
    var exportKey = function (key, format, callback) {
      // Export key
      App.Crypt.engine.subtle.exportKey(format, key).then(function (keydata) {
        // Return key in base64 format
        callback(App.Utils.bufferToBase64(keydata));
      }).catch(function (err) {
        callback(null, err);
      });
    };
    //
    // Set options for key pair generation
    var opts = {};
    opts.name = "RSA-OAEP";
    opts.modulusLength = 2048;
    opts.publicExponent = new Uint8Array([0x01, 0x00, 0x01]);
    opts.hash = {name: "SHA-1"};
    //
    // Generate key pair
    App.Crypt.engine.subtle.generateKey(opts, true, ["encrypt", "decrypt"]).then(function (key) {
      // Export private key
      exportKey(key.privateKey, "pkcs8", function (privateKeyData, privateError) {
        if (privateError)
          return cb(null, privateError);
        //
        var keyPair = {privateKey: formatKey(privateKeyData, "private")};
        //
        // Export public key
        exportKey(key.publicKey, "spki", function (publicKeyData, publicError) {
          if (publicError)
            return cb(null, publicError);
          //
          keyPair.publicKey = formatKey(publicKeyData, "public");
          //
          // Return key pair
          cb(keyPair);
        });
      });
      //
    }).catch(function (err) {
      if (err)
        cb(null, err);
    });
  }
};


/**
 * Encrypt data using a public RSA key
 * @param {Session} app
 * @param {String} data
 * @param {String} publicKey
 * @param {Object} options
 * @param {function} cb
 */
App.Crypt.publicEncrypt = function (app, data, publicKey, options, cb)
{
  if (!publicKey)
    return cb(null, new Error("Invalid public key"));
  //
  options = options || {};
  //
  if (module) {
    try {
      // Set encrypt options
      var opts = {};
      opts.key = publicKey;
      opts.padding = App.Crypt.engine.constants.RSA_PKCS1_OAEP_PADDING;
      //
      // Encrypt data
      var encrypted = App.Crypt.engine.publicEncrypt(opts, new Buffer(data));
      //
      // Return encrypted data
      if (cb)
        cb(App.Crypt.getOutput(encrypted, options.output, "encrypt"));
    } catch (ex) {
      if (cb)
        cb(null, ex);
    }
  }
  else {
    // Set RSA-OAEP as asymmetric encryption algorithm
    var algo = {};
    algo.name = "RSA-OAEP";
    algo.hash = {name: "SHA-1"};
    //
    // Create CryptoKey object using given publicKey
    App.Crypt.engine.subtle.importKey("spki", App.Crypt.pemToBuffer(publicKey), algo, false, ["encrypt"]).then(function (key) {
      // Ensure data to encrypt is a buffer
      var dataBuffer = typeof data === "string" ? App.Utils.stringToBuffer(data) : new Uint8Array(data).buffer;
      //
      // Encrypt data
      App.Crypt.engine.subtle.encrypt(algo, key, dataBuffer).then(function (encrypted) {
        // Return encrypted data
        if (cb)
          cb(App.Crypt.getOutput(encrypted, options.output, "encrypt"));
      }).catch(function (error) {
        if (cb)
          cb(null, error);
      });
    }).catch(function (error) {
      if (cb)
        cb(null, error);
    });
  }
};


/**
 * Decrypt data using a private RSA key
 * @param {Session} app
 * @param {String} data
 * @param {String} privateKey
 * @param {Object} options
 * @param {function} cb
 */
App.Crypt.privateDecrypt = function (app, data, privateKey, options, cb)
{
  if (module)
    App.Utils = require("./utils");
  //
  if (!privateKey)
    return cb(null, new Error("Invalid private key"));
  //
  options = options || {};
  //
  if (module) {
    try {
      // Set decryption options
      var opts = {};
      opts.key = privateKey;
      opts.padding = App.Crypt.engine.constants.RSA_PKCS1_OAEP_PADDING;
      //
      // Ensure data to decrypt is a buffer
      data = typeof data === "string" ? App.Utils.base64ToBuffer(data) : data;
      //
      // Decrypt data
      var decrypted = App.Crypt.engine.privateDecrypt(opts, new Buffer(data));
      if (cb)
        // Return decrypted data
        cb(App.Crypt.getOutput(decrypted, options.output, "decrypt"));
    } catch (ex) {
      if (cb)
        cb(null, ex);
    }
  }
  else {
    // Set RSA-OAEP as asymmetric decryption algorithm
    var algo = {};
    algo.name = "RSA-OAEP";
    algo.hash = {name: "SHA-1"};
    //
    // Create CryptoKey object using given privateKey
    App.Crypt.engine.subtle.importKey("pkcs8", App.Crypt.pemToBuffer(privateKey), algo, false, ["decrypt"]).then(function (key) {
      // Ensure data to decrypt is a buffer
      var dataBuffer = typeof data === "string" ? App.Utils.base64ToBuffer(data) : new Uint8Array(data).buffer;
      //
      // Decrypt data
      App.Crypt.engine.subtle.decrypt(algo, key, dataBuffer).then(function (decrypted) {
        // Return decrypted data
        if (cb)
          cb(App.Crypt.getOutput(decrypted, options.output, "decrypt"));
      }).catch(function (error) {
        if (cb)
          cb(null, error);
      });
    }).catch(function (error) {
      if (cb)
        cb(null, error);
    });
  }
};


/**
 * Convert RSA key to buffer
 * @param {String} pem
 */
App.Crypt.pemToBuffer = function (pem) {
  if (module)
    App.Utils = require("./utils");
  //
  var lines = pem.split("\n");
  var encoded = "";
  for (var i = 0; i < lines.length; i++) {
    if (lines[i].trim().length > 0 &&
            lines[i].indexOf("-BEGIN RSA PRIVATE KEY-") < 0 &&
            lines[i].indexOf("-BEGIN RSA PUBLIC KEY-") < 0 &&
            lines[i].indexOf("-BEGIN PRIVATE KEY-") < 0 &&
            lines[i].indexOf("-BEGIN PUBLIC KEY-") < 0 &&
            lines[i].indexOf("-END PUBLIC KEY-") < 0 &&
            lines[i].indexOf("-END PRIVATE KEY-") < 0 &&
            lines[i].indexOf("-END RSA PRIVATE KEY-") < 0 &&
            lines[i].indexOf("-END RSA PUBLIC KEY-") < 0) {
      encoded += lines[i].trim();
    }
  }
  //
  return App.Utils.base64ToBuffer(encoded);
};


/**
 * Convert data to given format
 * @param {String/Buffer} data
 * @param {String} format
 * @param {String} operation - "encrypt" or "decrypt"
 */
App.Crypt.getOutput = function (data, format, operation) {
  if (module)
    App.Utils = require("./utils");
  //
  var output;
  switch (format) {
    case "buffer":
      output = new Uint8Array(data).buffer;
      break;
    case "base64":
      output = App.Utils.bufferToBase64(data);
      break;
    default:
      if (operation === "decrypt")
        output = App.Utils.bufferToString(data);
      else
        output = App.Utils.bufferToBase64(data);
      break;
  }
  //
  return output;
};



//export module for node
if (module)
  module.exports = App.Crypt;
