/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module, Error */

var App = App || {};

/**
 * @class A Database Object
 * @param {Session} app that owns this device instance;
 * if the object is created to extend a class the param req will be true and require
 * will be avoided. That's to prevent circular requires
 * */
App.Database = function (app)
{
  if (module && !App.View && app !== true)
    App = require("./requireAll");
  //
  this.schema = {};        // JSON schema representation
  this.dmCodesMap = {};
  this.app = app;
  this.typemap = App.Database.DataTypeMap; // Map of data types (override in extended class)
  this.keywords = [];                      // List of reserved keywords (override in extended class)
  this.defaultSchema = "";
  this.functions = App.Database.Functions; // Definition of common functions and operators
  // Any extended class can redefine every single function if it is not supported or has a different implementation
  this.quoteChar = "\""; // Character used to quote entity names
  this.indent = 2;       // Number of space for indentation
  //
  // TODO: togliere quando tutti hanno corelibraries aggionato (issue #2395)
  App.Database.DataTypes.JSON = App.Database.DataTypes.JSON || "j";
};

App.Database.DataTypes = {
  INT: "i",
  AUTOINCR: "ai",
  FLOAT: "f",
  DECIMAL: "d",
  VARCHAR: "vc",
  CHAR: "c",
  BOOL: "b",
  DATE: "da",
  TIME: "ti",
  DATETIME: "dt",
  CLOB: "cl",
  UUID: "id",
  BLOB: "bl",
  JSON: "j"
};


App.Database.DriverTypes = {
  WebSQL: "WebSQL",
  Native: "Native",
  None: "None"
};


// static initialization
App.Database.DataTypeMap = {
  i: ["integer"],
  f: ["float"],
  d: ["decimal"],
  vc: ["nvarchar"],
  c: ["nchar"],
  bl: ["blob"],
  cl: ["clob"],
  j: ["varchar"]
};

App.Database.DataTypeClasses = {
  NUMBER: "n",
  DATETIME: "dt",
  STRING: "s",
  BLOB: "bl",
  BOOL: "b",
  JSON: "j"
};


App.Database.FKActions = {
  RESTRICT: "r",
  CASCADE: "c",
  SETDEFAULT: "d",
  NOACTION: "n",
  SETNULL: "l"
};

App.Database.QueryParts = {
  SELECT: "s",
  FROM: "f",
  WHERE: "w",
  GROUPBY: "g",
  HAVING: "h",
  ORDERBY: "o",
  TOP: "t"
};

App.Database.QueryTypes = {
  SELECT: "s",
  INSERT: "i",
  UPDATE: "u",
  DELETE: "d"
};

/**
 * Definitions of functions and operators
 *   expr: implementation of the function/operator
 *   op: true if is an operator
 *   aggr: true if is an aggregated function
 * */
App.Database.Functions = {
  // Operators
  "^": {expr: "power(@1, @2)", op: true},
  "%": {expr: "", op: true},
  "*": {expr: "@1 * @2", op: true},
  "/": {expr: "@1 / @2", op: true},
  "+": {expr: "@1 + @2", op: true},
  "-": {expr: "@1 - @2", op: true},
  "<": {expr: "@1 < @2", op: true},
  ">": {expr: "@1 > @2", op: true},
  "=": {expr: "@1 = @2", op: true},
  ">=": {expr: "@1 >= @2", op: true},
  "<=": {expr: "@1 <= @2", op: true},
  "<>": {expr: "@1 <> @2", op: true},
  "and": {expr: "@1 and @2", op: true},
  "or": {expr: "@1 or @2", op: true},
  "like": {expr: "@1 like @2", op: true},
  "not like": {expr: "@1 not like @2", op: true},
  "&": {expr: "@1 & @2", op: true},
  "|": {expr: "@1 | @2", op: true},
  //"between": {expr: "@1 between @2 and @3", op: true},
  "not": {expr: "not @1", op: true},
  "is null": {expr: "@1 is null", op: true},
  "is not null": {expr: "@1 is not null", op: true},
  "in": {expr: "@1 in @2", op: true},
  "not in": {expr: "@1 not in @2", op: true},
  "exists": {expr: "exists @1", op: true},
  "all": {expr: "all @1", op: true},
  "any": {expr: "any @1", op: true},
  "||": {expr: "@1 || @2", op: true},
  //
  // Aggregate functions
  sum: {expr: "sum(@1)", aggr: true},
  average: {expr: "avg(@1)", aggr: true},
  count: {expr: "count(@1)", aggr: true},
  countDistinct: {expr: "count(distinct @1)", aggr: true},
  max: {expr: "max(@1)", aggr: true},
  min: {expr: "min(@1)", aggr: true},
  stdDev: {expr: "stddev(@1)", aggr: true},
  variance: {expr: "variance(@1)", aggr: true},
  //
  // Date functions
  now: {rt: App.Database.DataTypes.DATETIME},
  today: {rt: App.Database.DataTypes.DATE},
  time: {rt: App.Database.DataTypes.TIME},
  toDate: {rt: App.Database.DataTypes.DATE},
  toTime: {rt: App.Database.DataTypes.TIME},
  // year: {},
  // month: {},
  // day: {},
  // hour: {},
  // minute: {},
  // second: {},
  // weekDay: {},
  // yearDay: {},
  // week: {},
  getDate: {rt: App.Database.DataTypes.DATE},
  getTime: {rt: App.Database.DataTypes.TIME},
  //
  // Math functions
  abs: {expr: "ABS(@1)"},
  // int: {},
  // ceil: {},
  // floor: {},
  // round: {},
  // trunc: {},
  sign: {expr: "SIGN(@1)"},
  exp: {expr: "EXP(@1)"},
  // sign: {}
  // sqrt: {},
  // log: {},
  // log10: {},
  sin: {expr: "SIN(@1)"},
  cos: {expr: "COS(@1)"},
  tan: {expr: "TAN(@1)"},
  // cot: {},
  arcSin: {expr: "ASIN(@1)"},
  arcCos: {expr: "ACOS(@1)"},
  artTan: {expr: "ATAN(@1)"},
  //
  // String functions
  lower: {expr: "LOWER(@1)"},
  upper: {expr: "UPPER(@1)"},
  // left: {},
  // right: {},
  // substr: {},
  length: {expr: "LENGTH(@1)"},
  // spaces: {},
  trimLeft: {expr: "LTRIM(@1)"},
  trimRight: {expr: "RTRIM(@1)"},
  trim: {expr: "LTRIM(RTRIM(@1))"},
  replace: {expr: "REPLACE(@1, @2, @3)"},
  // fillLeft: {},
  // fillRight: {}
  //
  // Various function
  if : {expr: "CASE WHEN @1 THEN @2 ELSE @3 END"}
  // nullValue: {},
  // blobSize: {}
};


/*
 * Sets connection options for this database
 * @param {Object} opt - connect options
 */
App.Database.prototype.setOptions = function (opt)
{
  this.options = opt;
};


/*
 * Sets remote connector
 * @param {String} name
 * @param {String} key
 */
App.Database.prototype.setRemoteConnector = function (name, key)
{
  if (!module)
    return;
  //
  this.remoteConnector = name;
  this.APIKey = key;
};


/*
 * Returns the default instance of this database
 */
App.Database.prototype.getDefaultInstance = function ()
{
  return this.app.getDefaultInstance(this.schema.name);
};


/**
 * Handles the auto-close connection feature
 * (auto-closes a connection 10 seconds after the last query or transaction)
 * @param {Integer} delta - +/-1 for start/end query and transaction
 */
App.Database.prototype.handleAutoClose = function (delta)
{
  clearTimeout(this.closeTimeout);
  delete this.closeTimeout;
  //
  // Update counter of pending operations
  if (delta)
    this.pendingOperations = (this.pendingOperations || 0) + delta;
  //
  // No pending operations -> close connection
  if (!this.pendingOperations) {
    // With cloud connectors delay closure of connection
    if (this.remoteConnector)
      this.closeTimeout = setTimeout(App.Database.prototype.closeConnection.bind(this), 10 * 1000);
    else
      this.closeConnection();
  }
};


/**
 * Open the connection to the database
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.openConnection = function (callback)
{
  if (this.connection)
    return callback();
  //
  var logObj = this.log("Open connection" + (this.connectingCallbacks ? " (already connecting)" : ""));
  var done = function (connection, error) {
    if (!error)
      this.connection = connection;
    //
    this.log(error || "OK", logObj);
    //
    // Call all pending callbacks
    if (this.connectingCallbacks) {
      this.connectingCallbacks.forEach(setImmediate);
      delete this.connectingCallbacks;
    }
    //
    callback(null, error);
  }.bind(this);
  //
  // Remember this new callback... it will be called when the connection will be opened
  if (!this.connectingCallbacks)
    this.connectingCallbacks = [];
  else
    // If I'm already connecting (i.e. a previous call is still pending)
    return this.connectingCallbacks.push(function () {
      this.openConnection(callback);
    }.bind(this));
  //
  if (this.remoteConnector) {
    var conn = new App.RemoteDriver(this);
    conn.connect(function (result, error) {
      if (error)
        conn.disconnect();
      done(conn, error);
    });
  }
  else
    this._openConnection(done);
};


/**
 * Close the connection to the database
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.closeConnection = function (callback)
{
  var logObj = this.log("Release connection" + (this.connection ? "" : " (already released)"));
  var done = function (result, error) {
    delete this.connection;
    this.log(error || "OK", logObj);
    if (callback)
      callback(null, error);
  }.bind(this);
  //
  // No connection -> done
  if (!this.connection)
    return done();
  //
  // Release connection to pool
  if (this.remoteConnector)
    this.connection.disconnect(done);
  else
    this._closeConnection(done);
};


/**
 * Execute a list of commands
 * @param {Array} sqls - list of sql statement to execute
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.executeCommands = function (sqls, callback)
{
  var s = sqls.shift();
  this.bindParameters(s);
  this.executeSQL(s.sql, false, function (result, error) {
    if (error)
      callback(null, error);
    else if (sqls.length === 0)
      callback();
    else
      this.executeCommands(sqls, callback);
  }.bind(this));
};


/**
 * Update the database schema
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.updateSchema = function (callback)
{
  var done = function (result, error) {
    if (callback)
      callback(result, error);
  };
  //
  this.createDb(function (result, error) {
    if (error)
      return done(null, error);
    //
    // Read from Z_SCHEMA table the actual schema
    this.getActualSchema(function (as, error) {
      if (error)
        return done(null, error);
      //
      // Create a list of sql statement to update the schema
      var sqls = [];
      this.actualSchema = as;
      this.updateDatabase(sqls, this.schema, as);
      delete this.actualSchema;
      //
      // If not aligned ... updating schema
      if (sqls.length) {
        this.saveZSchema(sqls, this.schema);
        //
        // Open the transaction
        this.beginTransaction(function (result, error) {
          if (error)
            return done(null, error);
          //
          // Execute all commands
          this.executeCommands(sqls, function (result, error) {
            if (error) {
              // Rollback the transaction
              this.rollbackTransaction(function (res, err) {
                done(null, error || err);
              });
            }
            else // Commit the transaction
              this.commitTransaction(done);
          }.bind(this));
        }.bind(this));
      }
      else
        done();
    }.bind(this));
  }.bind(this));
};


/**
 * Create a new database
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.createDb = function (callback)
{
  // Not supported
  callback();
};


/**
 * Bind parameters of a SQL statement
 * @param {Object} stmt statement to bind
 */
App.Database.prototype.bindParameters = function (stmt)
{
  if (!stmt.params)
    return;
  //
  var newSql = stmt.sql;
  var inQuote = false;
  var parIndex = 0;
  for (var i = 0; i < newSql.length; i++) {
    if (newSql.charAt(i) === "'")
      inQuote = !inQuote;
    else if (newSql.charAt(i) === "?" && !inQuote) {
      var par = stmt.params[parIndex];
      if (typeof par === "string")
        par = "'" + par.replace(/'/g, "''") + "'";
      newSql = newSql.substr(0, i) + par.toString() + newSql.substr(i + 1);
      i += par.toString().length - 1;
      parIndex++;
    }
  }
  stmt.sql = newSql;
  delete stmt.params;
};


/**
 * Get the actual schema from the database
 * @param {Function} callback - function (schema, err)
 */
App.Database.prototype.getActualSchema = function (callback)
{
  this.executeSQL('select * from z_schema', true, function (result, error) {
    // If there is an error it means that the z_schema table does not exists
    // Return immediately to callee
    if (error)
      return callback({});
    //
    this.zSchemaExists = true;
    var s;
    try {
      s = JSON.parse(this.getRow(result, 0).schema);
    }
    catch (e) {
      return callback(null, e);
    }
    callback(s);
  }.bind(this));
};


/**
 * Create a list of sql statement to update the schema
 * @param {Array} sqls - list of sql statements
 * @param {Object} news - new schema
 * @param {Object} olds - old schema
 */
App.Database.prototype.updateDatabase = function (sqls, news, olds)
{
  // Drop unuseless objects
  if (olds.tables) {
    this.dropUnuselessFKs(sqls, news, olds);
    this.dropUnuselessIndexes(sqls, news, olds);
    this.dropUnuselessTables(sqls, news, olds);
  }
  //
  // Create or modify tables without FKs and indexes
  if (news.tables) {
    for (var i = 0; i < news.tables.length; i++) {
      var nt = news.tables[i];
      var ot = this.getObject(olds.tables, nt);
      if (ot)
        this.modifyTable(sqls, nt, ot);
      else {
        ot = this.createTable(sqls, nt);
        //
        if (!olds.tables)
          olds.tables = [];
        olds.tables.push(ot);
      }
    }
    //
    // Create or modify FKs and indexes
    for (var i = 0; i < news.tables.length; i++) {
      var nt = news.tables[i];
      var ot = this.getObject(olds.tables, nt);
      this.modifyTableIndexes(sqls, nt, ot);
      this.modifyTableFKs(sqls, nt, ot);
    }
  }
};


/**
 * Save the schema definition in z_schema table
 * @param {Array} sqls - list of sql statement where to add
 * @param {Object} news - new schema
 */
App.Database.prototype.saveZSchema = function (sqls, news)
{
  sqls.push({sql: "create table if not exists z_schema (schema text)"});
  sqls.push({sql: "delete from z_schema"});
  sqls.push({sql: "insert into z_schema values (?)", params: [JSON.stringify(news)]});
};


/**
 * Set a dummy value into not nullable field
 * @param {Array} sqls - list of sql statement where to add
 * @param {Object} t table to modify
 * @param {Object} f field to update
 */
App.Database.prototype.updateRequiredField = function (sqls, t, f)
{
  var defv = this.getFieldDefVal(f);
  if (!defv) {
    switch (App.Database.classDT(f.datatype)) {
      case App.Database.DataTypeClasses.NUMBER:
        defv = 0;
        break;
      case App.Database.DataTypeClasses.BOOL:
        defv = false;
        break;
      case App.Database.DataTypeClasses.DATETIME:
        defv = new Date(0);
        break;
      case App.Database.DataTypeClasses.STRING:
      case App.Database.DataTypeClasses.BLOB:
        defv = "";
        break;
    }
    defv = this.toSql(defv, f.datatype);
  }
  sqls.push({sql: "update " + this.quote(t.name) + " set " + this.quote(f.name)
            + " = " + defv + " where " + this.quote(f.name) + " is null"});
};


/**
 * Execute an SQL query
 * @param {String/Object} sql to execute
 * @param {Function} callback - function to be called at the end
 * @return {DataMap/Object}
 */
App.Database.prototype.query = function (sql, callback)
{
  if (typeof sql === "string")
    sql = {qry: sql};
  //
  try {
    var startTime = new Date();
    var dm = this.prepare(sql);
    dm.load(null, function (result, error) {
      if (result && result.elapsedTimes)
        result.elapsedTimes.tot = new Date() - startTime;
      callback(result, error);
    });
  }
  catch (ex) {
    callback(null, ex);
  }
};


/**
 * Prepare an SQL query
 * @param {String/Object} sql to execute
 * @return {DataMap/Object}
 */
App.Database.prototype.prepare = function (sql)
{
  if (typeof sql !== "string") {
    var dm = new App.DataMap();
    dm.dataStore = this;
    //
    // Pass "local" (i.e. in context) objects to datamap
    dm.app = this.app;
    dm.view = sql.view;
    dm.this = sql.this;
    //
    dm.addQuery(sql.qry);
    dm.addSchema(sql.sch);
    return dm;
  }
};


/**
 * Execute an SQL query with offset paging
 * @param {Object} params
 * @param {Function} callback - function to be called at the end
 * @return {DataMap/Object}
 */
App.Database.prototype.queryOffsetPaging = function (params, callback)
{
  callback(null, new Error("Pagination mode 'offset' not supported"));
};


/**
 * Execute an SQL statement
 * @param {String} sql to execute
 * @param {Function} callback - function to be called at the end
 * @return {Integer} number of rows affected
 */
App.Database.prototype.execute = function (sql, callback)
{
  // If the sql param is an object compose sql statement
  if (typeof sql !== "string") {
    var cliRefs = App.DataMap.getClientRefs(sql.qry);
    if (cliRefs.length) {
      // This query countains one or more client references...
      // Use a DataMap to "convert" them
      var dm = this.prepare(sql);
      dm.attachClientRefsObserver();
    }
    var parameters = [];
    sql = this.createSqlStatement(sql.qry, parameters);
  }
  //
  this.executeSQL(sql, false, parameters, callback);
};


/**
 * Execute an SQL query
 * @param {String} sql to execute
 * @param {Boolean} readOnly - true if is a SELECT
 * @param {Array} parameters
 * @param {Function} callback - function to be called at the end
 * @return {Object/Integer}
 */
App.Database.prototype.executeSQL = function (sql, readOnly, parameters, callback)
{
  var logObj = this.log("Execute SQL: " + sql);
  if (logObj)
    logObj.readOnly = readOnly;
  if (!callback && typeof parameters === "function") {
    callback = parameters;
    parameters = [];
  }
  else if (!parameters)
    parameters = [];
  //
  // Suspend auto closing
  this.handleAutoClose(1);
  //
  // Reset some info
  if (!readOnly) {
    delete this.lastRowsAffected;
    delete this.lastInsertId;
  }
  //
  // Add DTT item
  var dttitStart;
  if (this.app) {
    dttitStart = this.app.dttobj.addItem(this.pid, App.DTT.ItemTypeMap.QUERY, this.schema.name + " query");
    this.app.dttobj.addItemParam("DataModel", this.schema.name, dttitStart);
    this.app.dttobj.addItemParam("SQL", sql.replace(/\n/g, " "), dttitStart);
    this.app.dttobj.addItemParam("parameters", parameters, dttitStart);
  }
  //
  var startTime = new Date();
  var done = function (result, error) {
    // Program auto closing
    this.handleAutoClose(-1);
    //
    this.log(error || result, logObj);
    //
    if (error) {
      if (dttitStart) {
        this.app.dttobj.addException(dttitStart, error, dttitStart.parentItem);
        //
        // Send exception to testAuto
        if (this.app.testAuto) {
          var e = [{id: "sendExceptionToTestAuto", cnt: {msg: error}}];
          this.app.sendCommands(e);
        }
      }
      //
      if (callback)
        callback(null, error);
      return;
    }
    //
    var elapsedTime = new Date() - startTime;
    //
    // Dump some useful info
    var dttit;
    if (dttitStart)
      dttit = this.app.dttobj.addItem(dttitStart, App.DTT.ItemTypeMap.QUERYRES, undefined, undefined, dttitStart.parentItem);
    //
    // Query is terminated with success
    if (readOnly) {
      result.times = result.times || {qry: elapsedTime};
      //
      // Dump some useful info
      if (dttit) {
        this.app.dttobj.addItemParam("Total rows", this.getRowCount(result), dttit);
        //
        // Dump only some rows
        var rows = [];
        var nr = Math.min(App.DTT.QueryRows, this.getRowCount(result));
        for (var i = 0; i < nr; i++)
          rows.push(this.getRow(result, i));
        this.app.dttobj.addItemParam("Data", JSON.stringify(rows), dttit);
      }
    }
    else {
      this.readInfoFromResult(sql, result);
      result = this.lastRowsAffected;
      //
      // Dump some useful info
      if (dttit)
        this.app.dttobj.addItemParam("Records affected", this.lastRowsAffected, dttit);
    }
    //
    // Dump elapsed time
    if (dttit)
      this.app.dttobj.addItemParam("Elapsed time (ms)", elapsedTime, dttit);
    //
    if (callback)
      callback(result);
  }.bind(this);
  //
  // Open connection
  this.openConnection(function (result, error) {
    if (error)
      return done(null, error);
    //
    // Run the query
    if (this.remoteConnector)
      this.connection.execute(sql, parameters, done);
    else
      this._executeSQL(sql, readOnly, parameters, done);
  }.bind(this));
};


/**
 * Return column's definition from resultSet
 * @param {Object} result
 */
App.Database.prototype.getColumnsDefByResult = function (result)
{
  var columns = [];
  var undetectedColumns = 0;
  var colNames = null;
  var nrows = this.getRowCount(result);
  //
  // Detect dataType from data
  for (var i = 0; i < nrows; i++) {
    var row = this.getRow(result, i);
    //
    // If first row detect number and names of columns
    if (i === 0) {
      var colNames = Object.keys(row);
      undetectedColumns = colNames.length;
      for (var j = 0; j < colNames.length; j++)
        columns.push({name: colNames[j]});
    }
    //
    for (var colIdx = 0; colIdx < columns.length; colIdx++) {
      var c = columns[colIdx];
      if (c.datatype && c.datatype !== App.Database.DataTypes.INT)
        continue;
      //
      var value = row[c.name];
      switch (typeof (value)) {
        case "string":
          c.datatype = App.Database.DataTypes.VARCHAR;
          break;

        case "boolean":
          c.datatype = App.Database.DataTypes.BOOL;
          break;

        case "number":
          c.datatype = (value % 1 ? App.Database.DataTypes.FLOAT : App.Database.DataTypes.INT);
          break;
      }
      //
      if (!c.datatype || c.datatype === App.Database.DataTypes.INT)
        undetectedColumns--;
    }
    //
    // If I have detect datatype of all columns I finished
    if (!undetectedColumns)
      break;
  }
  return columns;
};


/*
 * Read some info from result
 * @param {String} sql
 * @param {Object} result
 */
App.Database.prototype.readInfoFromResult = function (sql, result)
{
  if (this.remoteConnector)
    this.connection.readInfoFromResult(result);
  else
    this._readInfoFromResult(sql, result);
};


/**
 * Get a corresponding object between two schema arrays
 * @param {Array} sc tables, fields,...
 * @param {Object} obj to find
 */
App.Database.prototype.getObject = function (sc, obj)
{
  if (!sc)
    return;
  //
  // Search for ID
  for (var i = 0; i < sc.length; i++) {
    if (sc[i].id === obj.id)
      return sc[i];
  }
  // Search for Name
  for (var i = 0; i < sc.length; i++) {
    if (sc[i].name === obj.name)
      return sc[i];
  }
};


/**
 * Drop unuseless foreign keys
 * @param {Array} sqls list of sql statements
 * @param {Object} news new schema
 * @param {Object} olds old schema
 */
App.Database.prototype.dropUnuselessFKs = function (sqls, news, olds)
{
  for (var i = 0; i < olds.tables.length; i++) {
    var ot = olds.tables[i];
    var nt = this.getObject(news.tables, ot);
    var needDrop = false;
    if (ot.fks) {
      for (var j = 0; j < ot.fks.length; j++) {
        var ofk = ot.fks[j];
        if (nt) {
          var nfk = this.getObject(nt.fks, ofk);
          if (!nfk)
            needDrop = !nfk;
        }
        else
          needDrop = true;
        //
        if (needDrop)
        {
          this.dropFK(sqls, ot, ofk);
          ot.fks.splice(j, 1);
          j--;
        }
      }
    }
  }
};


/**
 * Drop unuseless indexes
 * @param {Array} sqls list of sql statements
 * @param {Object} news new schema
 * @param {Object} olds old schema
 */
App.Database.prototype.dropUnuselessIndexes = function (sqls, news, olds)
{
  for (var i = 0; i < olds.tables.length; i++) {
    var ot = olds.tables[i];
    var nt = this.getObject(news.tables, ot);
    var needDrop = false;
    if (ot.indexes) {
      for (var j = 0; j < ot.indexes.length; j++) {
        var oidx = ot.indexes[j];
        if (nt) {
          var nidx = this.getObject(nt.indexes, oidx);
          if (!nidx)
            needDrop = !nidx;
        }
        else
          needDrop = true;
        //
        if (needDrop) {
          this.dropIndex(sqls, ot, oidx);
          ot.indexes.splice(j, 1);
          j--;
        }
      }
    }
  }
};


/**
 * Drop unuseless tables
 * @param {Array} sqls list of sql statements
 * @param {Object} news new schema
 * @param {Object} olds old schema
 */
App.Database.prototype.dropUnuselessTables = function (sqls, news, olds)
{
  for (var i = 0; i < olds.tables.length; i++) {
    var ot = olds.tables[i];
    var nt = this.getObject(news.tables, ot);
    if (!nt) {
      this.dropTable(sqls, ot);
      olds.tables.splice(i, 1);
      i--;
    }
  }
};


/**
 * Drop unuseless table fields
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} ot old version
 */
App.Database.prototype.dropUnuselessTableFields = function (sqls, t, ot)
{
  for (var i = 0; i < ot.fields.length; i++) {
    var of = ot.fields[i];
    var nf = this.getObject(t.fields, of);
    if (!nf && !this.dropField(sqls, ot, of))
      return false;
  }
  return true;
};


/**
 * Modify table fields
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} ot old version
 */
App.Database.prototype.modifyTableFields = function (sqls, t, ot)
{
  var newPK = "";
  for (var i = 0; i < t.fields.length; i++) {
    var nf = t.fields[i];
    var of = this.getObject(ot.fields, nf);
    if (of) {
      if (!this.modifyField(sqls, ot, nf, of))
        return false;
    }
    else if (!this.createField(sqls, ot, nf))
      return false;
    //
    if (nf.pk)
      newPK += (newPK ? "," : "") + nf.name;
  }
  //
  var oldPK = "";
  for (var i = 0; i < ot.fields.length; i++) {
    var of = ot.fields[i];
    if (of.pk)
      oldPK += (oldPK ? "," : "") + of.name;
  }
  //
  if (newPK !== oldPK)
    this.recreatePK(sqls, t, ot);
  //
  return true;
};


/**
 * Recreate the primary key of table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to recreate
 * @param {Object} ot old version
 */
App.Database.prototype.recreatePK = function (sqls, t, ot)
{
};


/**
 * Create a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to create
 */
App.Database.prototype.createTable = function (sqls, t)
{
};


/**
 * Copy data between two table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to copy data
 * @param {Object} ot table from copy data
 */
App.Database.prototype.copyDataTable = function (sqls, t, ot)
{
  var s = "insert into " + this.quote(t.name) + "(";
  var s1 = "";
  for (var i = 0; i < t.fields.length; i++) {
    var f = t.fields[i];
    s += (i === 0 ? "" : ",") + "\n  " + this.quote(f.name);
    s1 += (i === 0 ? "" : ",") + "\n  ";
    var of = this.getObject(ot.fields, f);
    if (of)
      s1 += this.quote(of.name);
    else {
      if (this.getFieldDefVal(f))
        s1 += this.getFieldDefVal(f);
      else if (f.notn) {
        switch (App.Database.classDT(f.datatype)) {
          case App.Database.DataTypeClasses.NUMBER:
          case App.Database.DataTypeClasses.BOOL:
            s1 += "0";
            break;
          case App.Database.DataTypeClasses.DATETIME:
            s1 += this.functions.now.expr;
            break;
          case App.Database.DataTypeClasses.STRING:
          case App.Database.DataTypeClasses.BLOB:
            s1 += "''";
            break;
        }
      }
      else
        s1 += "null";
    }
  }
  s += ")\nselect " + s1 + "\nfrom " + this.quote(ot.name);
  //
  sqls.push({sql: s});
};


/**
 * Modify a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} ot old version
 */
App.Database.prototype.modifyTable = function (sqls, t, ot)
{
  var ss = [];
  var needRecreate = !this.dropUnuselessTableFields(ss, t, ot);
  if (!needRecreate)
    needRecreate = !this.modifyTableFields(ss, t, ot);
  //
  if (needRecreate)
    this.recreateTable(sqls, t, ot);
  else {
    for (var i = 0; i < ss.length; i++)
      sqls.push(ss[i]);
    if (t.name.toLowerCase() !== ot.name.toLowerCase())
      this.renameTable(sqls, t, ot);
  }
  ot.name = t.name;
  ot.fields = t.fields;
};


/**
 * Recreate a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to recreate
 * @param {Object} ot old version
 */
App.Database.prototype.recreateTable = function (sqls, t, ot)
{
  var tt = {name: ot.name + "_old", fields: ot.fields};
  this.renameTable(sqls, tt, ot);
  //
  // Delete all table internal fks
  if (ot.fks) {
    for (var i = 0; i < ot.fks.length; i++) {
      this.dropFK(sqls, ot, ot.fks[i]);
      ot.fks.splice(i, 1);
      i--;
    }
  }
  //
  // Delete all table external fks
  this.dropExternalTableFKS(sqls, ot);
  //
  // Delete all table indexes
  if (ot.indexes) {
    for (var i = 0; i < ot.indexes.length; i++) {
      this.dropIndex(sqls, ot, ot.indexes[i]);
      ot.indexes.splice(i, 1);
      i--;
    }
  }
  //
  this.createTable(sqls, t);
  this.copyDataTable(sqls, t, tt);
  this.dropTable(sqls, tt);
};


/**
 * Delete all table external fks
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to delete external fks
 */
App.Database.prototype.dropExternalTableFKS = function (sqls, t)
{
  for (var i = 0; i < this.actualSchema.tables.length; i++) {
    var rt = this.actualSchema.tables[i];
    if (!rt.fks)
      continue;
    //
    for (var j = 0; j < rt.fks.length; j++) {
      var fk = rt.fks[j];
      if (fk.t === t.name) {
        this.dropFK(sqls, rt, rt.fks[j]);
        rt.fks.splice(j, 1);
        j--;
      }
    }
  }
};


/**
 * Drop a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to drop
 */
App.Database.prototype.dropTable = function (sqls, t)
{
  sqls.push({sql: "drop table " + this.quote(t.name)});
  return true;
};


/**
 * Rename a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to rename
 * @param {Object} ot old version
 */
App.Database.prototype.renameTable = function (sqls, t, ot)
{
  sqls.push({sql: "alter table " + this.quote(ot.name) + " rename to " + this.quote(t.name)});
  this.modifyTableFKs(sqls, t, ot);
  return true;
};


/**
 * Create a field
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} f field to create
 */
App.Database.prototype.createField = function (sqls, t, f)
{
  // If new field is not null create it as nullable ...
  var notn = false;
  if (f.notn && !this.getFieldDefVal(f)) {
    notn = true;
    f.notn = false;
  }
  //
  sqls.push({sql: "alter table " + this.quote(t.name) + " add " + this.getFieldDef(f)});
  //
  // ... finally update values and set it not null
  if (notn) {
    var of = JSON.parse(JSON.stringify(f));
    f.notn = true;
    this.modifyField(sqls, t, f, of);
  }
  //
  return true;
};


/**
 * Modify a field
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} f field to modify
 * @param {Object} of old version
 */
App.Database.prototype.modifyField = function (sqls, t, f, of)
{
};


/**
 * Drop a field
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of field
 * @param {Object} f field to drop
 */
App.Database.prototype.dropField = function (sqls, t, f)
{
  sqls.push({sql: "alter table " + this.quote(t.name) + " drop column " + this.quote(f.name)});
  return true;
};


/**
 * Create a foreign key
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of foreign key
 * @param {Object} fk foreign key to create
 */
App.Database.prototype.createFK = function (sqls, t, fk)
{
  var s = "alter table " + this.quote(t.name);
  s += " add constraint " + this.quote(this.getFKName(t, fk)) + " foreign key (";
  var fkn = "";
  for (var ref in fk.refs) {
    s += (!fkn ? "" : ",") + this.quote(this.getObject(t.fields, {id: ref}).name);
    fkn += (!fkn ? "" : ",") + this.quote(fk.refs[ref]);
  }
  s += ") references " + this.quote(fk.t) + " (" + fkn + ")";
  s += this.getFKActions(fk);
  //
  sqls.push({sql: s});
  return true;
};


/**
 * Modify a foreign key
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of foreign key
 * @param {Object} ot table old version
 * @param {Object} fk foreign key to modify
 * @param {Object} ofk old version
 */
App.Database.prototype.modifyFK = function (sqls, t, ot, fk, ofk)
{
};


/**
 * Calculate the name of a foreign key
 * @param {Object} t table owner of foreign key
 * @param {Object} fk
 */
App.Database.prototype.getFKName = function (t, fk)
{
  return t.name + "_" + fk.name;
};


/**
 * Drop an foreign key
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of foreign key
 * @param {Object} fk foreign key to drop
 */
App.Database.prototype.dropFK = function (sqls, t, fk)
{
  var s = "alter table " + this.quote(t.name) + " drop constraint " + this.quote(this.getFKName(t, fk));
  sqls.push({sql: s});
  return true;
};


/**
 * Create an index
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of index
 * @param {Object} idx index to create
 */
App.Database.prototype.createIndex = function (sqls, t, idx)
{
  // Getting create index stmt
  var s = "create ";
  if (idx.un)
    s += "unique ";
  s += "index " + this.quote(idx.name) + " on " + this.quote(t.name) + " (";
  for (var i = 0; i < idx.fields.length; i++)
    s += (i === 0 ? "" : ", ") + this.quote(this.getObject(t.fields, {id: idx.fields[i]}).name);
  s += ")";
  //
  sqls.push({sql: s});
  return true;
};


/**
 * Modify an index
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of index
 * @param {Object} idx index to modify
 * @param {Object} oidx old version
 */
App.Database.prototype.modifyIndex = function (sqls, t, idx, oidx)
{
};


/**
 * Drop an index
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of index
 * @param {Object} idx index to drop
 */
App.Database.prototype.dropIndex = function (sqls, t, idx)
{
  sqls.push({sql: "drop index " + this.quote(idx.name)});
  return true;
};


/**
 * Create or modify all index of a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table
 * @param {Object} ot old version
 */
App.Database.prototype.modifyTableIndexes = function (sqls, t, ot)
{
  if (!t.indexes)
    return true;
  //
  for (var i = 0; i < t.indexes.length; i++) {
    var ni = t.indexes[i];
    var oi = (ot ? this.getObject(ot.indexes, ni) : null);
    if (oi) {
      if (!this.modifyIndex(sqls, t, ni, oi))
        return false;
    }
    else if (!this.createIndex(sqls, t, ni))
      return false;
  }
  ot.indexes = t.indexes;
  return true;
};


/**
 * Create or modify all foreign key of a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table
 * @param {Object} ot old version
 */
App.Database.prototype.modifyTableFKs = function (sqls, t, ot)
{
  if (!t.fks)
    return true;
  //
  for (var i = 0; i < t.fks.length; i++) {
    var nf = t.fks[i];
    var of = (ot ? this.getObject(ot.fks, nf) : null);
    if (of) {
      if (!this.modifyFK(sqls, t, ot, nf, of))
        return false;
    }
    else if (!this.createFK(sqls, t, nf))
      return false;
  }
  ot.fks = t.fks;
  return true;
};


/**
 * Get the definition for the f field
 * @param {Object} f field to define
 */
App.Database.prototype.getFieldDef = function (f)
{
  var s = this.quote(f.name) + " " + this.getDataType(f);
  if (f.notn)
    s += " not null";
  if (this.getFieldDefVal(f))
    s += " default(" + this.getFieldDefVal(f) + ")";
  return s;
};


/**
 * Get the datatype for the f field
 * @param {Object} f
 */
App.Database.prototype.getDataType = function (f)
{
  var s = this.typemap[f.datatype][0];
  if (this.getFieldMaxLength(f)) {
    s += "(" + this.getFieldMaxLength(f);
    if (this.getFieldScale(f) && f.datatype === App.Database.DataTypes.DECIMAL)
      s += "," + this.getFieldScale(f);
    s += ")";
  }
  return s;
};


/**
 * Get the action for the fk foreign key
 * @param {Object} fk
 */
App.Database.prototype.getFKActions = function (fk)
{
  var s = "";
  for (var i = 0; i < 2; i++) {
    var r = (i === 0 ? fk.ur : fk.dr);
    s += (i === 0 ? " on update " : " on delete ");
    if (r === App.Database.FKActions.RESTRICT)
      s += "restrict";
    else if (r === App.Database.FKActions.CASCADE)
      s += "cascade";
    else if (r === App.Database.FKActions.SETDEFAULT)
      s += "set default";
    else if (r === App.Database.FKActions.NOACTION)
      s += "no action";
    else if (r === App.Database.FKActions.SETNULL)
      s += "set null";
  }
  return s;
};


/**
 * Get the class of a data type
 * @param {String} dataType
 */
App.Database.classDT = function (dataType)
{
  switch (dataType) {
    case App.Database.DataTypes.DECIMAL:
    case App.Database.DataTypes.FLOAT:
    case App.Database.DataTypes.INT:
    case App.Database.DataTypes.AUTOINCR:
      return App.Database.DataTypeClasses.NUMBER;

    case App.Database.DataTypes.DATE:
    case App.Database.DataTypes.TIME:
    case App.Database.DataTypes.DATETIME:
      return App.Database.DataTypeClasses.DATETIME;

    case App.Database.DataTypes.CHAR:
    case App.Database.DataTypes.VARCHAR:
    case App.Database.DataTypes.CLOB:
    case App.Database.DataTypes.UUID:
      return App.Database.DataTypeClasses.STRING;

    case App.Database.DataTypes.BLOB:
      return App.Database.DataTypeClasses.BLOB;

    case App.Database.DataTypes.BOOL:
      return App.Database.DataTypeClasses.BOOL;

    case App.Database.DataTypes.JSON:
      return App.Database.DataTypeClasses.JSON;
  }
};


/**
 * Quote a name
 * @param {String} s name to quote
 */
App.Database.prototype.quote = function (s)
{
  if (!s)
    s = "";
  //
  // Check if is a keyword
  var toQuote = false;
  if (this.keywords.indexOf(s.toUpperCase()) !== -1)
    toQuote = true;
  else if (this.caseSensitive && this.requireCaseSensitivity(s))
    toQuote = true;
  else {
    // Check if contains not alphanumeric character
    for (var i = 0; i < s.length; i++) {
      var code = s.charCodeAt(i);
      if (!(code > 47 && code < 58) && // numeric (0-9)
              !(code > 64 && code < 91) && // upper alpha (A-Z)
              !(code > 96 && code < 123) && // lower alpha (a-z)
              code !== 95) { // underscore
        toQuote = true;
        break;
      }
    }
  }
  //
  if (toQuote)
    return this.quoteChar + s + this.quoteChar;
  else
    return s;
};


/**
 * Quote a string
 * @param {String} s string to quote
 */
App.Database.quoteString = function (s)
{
  var ris = s;
  var i = 0;
  while (i < ris.length) {
    if (ris.charAt(i) === "'") {
      var tmp = ris.slice(0, i) + "'";
      tmp = tmp + ris.slice(i);
      ris = tmp;
      i++;
    }
    i++;
  }
  ris = "'" + ris + "'";
  return ris;
};


/**
 * Compose sql statement of a query
 * @param {Object} query to compose
 * @param {Array} parameters
 * @return {String} sql statement
 */
App.Database.prototype.createSqlStatement = function (query, parameters)
{
  this.increasePadding();
  this.queries = [query];
  this.parameters = parameters;
  //
  try {
    switch (query.type) {
      case App.Database.QueryTypes.SELECT:
        return this.createSelectStatement(query);

      case App.Database.QueryTypes.UPDATE:
        return this.createUpdateStatement(query);

      case App.Database.QueryTypes.INSERT:
        return this.createInsertStatement(query);

      case App.Database.QueryTypes.DELETE:
        return this.createDeleteStatement(query);

      default:
        return this.createCustomStatement(query);
    }
  }
  catch (e) {
    throw e;
  }
  finally {
    this.decreasePadding();
    delete this.queries;
    delete this.parameters;
  }
};


/**
 * Compose select statement of a query
 * @param {Object} query to compose
 */
App.Database.prototype.createSelectStatement = function (query)
{
  if (this.queries.indexOf(query) === -1)
    this.queries.push(query);
  //
  var sql = "";
  var useAggregate = (query.group || query.having);
  query.group = query.group || [];
  query.having = query.having || [];
  //
  // Select list
  if (query.select && query.select.length > 0) {
    sql += this.padding + "select";
    //
    // Distinct clause
    if (query.distinct)
      sql += " distinct";
    //
    // Top clause
    if (query.top)
      sql += " " + this.limitClause(query.top, true);
    //
    this.increasePadding();
    for (var i = 0; i < query.select.length; i++) {
      var tk = query.select[i];
      var sqltk = this.createSqlExpression(tk, App.Database.QueryParts.SELECT);
      //
      sql += (i === 0 || sql.substr(-1) === "," ? "" : ",") + "\n" + this.padding + sqltk;
      //
      // Don't aggregate commented selects
      if (sqltk.trim().substring(0, 2) === "--")
        continue;
      //
      // If not use aggregate functions add it to groupBy section (if there is not already)
      if (!this.useAggregateFuctions(tk)) {
        var found = false;
        var tksql = this.createSqlExpression(tk, App.Database.QueryParts.GROUPBY);
        var tkalias = tk.alias || (tk.from !== undefined ? tk.expr : "");
        for (var j = 0; j < query.group.length; j++) {
          var tkgsql = this.createSqlExpression(query.group[j], App.Database.QueryParts.GROUPBY);
          var index = Number(tkgsql);
          //
          // Find token by sql, alias or index
          if (tksql === tkgsql || tkalias === tkgsql || index === i + 1) {
            found = true;
            break;
          }
        }
        if (!found)
          query.group.push(tk);
      }
      else
        useAggregate = true;
    }
    this.decreasePadding();
  }
  //
  // From list
  if (query.from && query.from.length > 0) {
    sql += "\n" + this.padding + "from";
    this.increasePadding();
    for (var i = 0; i < query.from.length; i++) {
      var tk = query.from[i];
      sql += (i > 0 && !tk.join ? "," : "") + "\n" + this.padding + this.createSqlExpression(tk, App.Database.QueryParts.FROM);
    }
    this.decreasePadding();
  }
  //
  // Where clauses
  if (query.where && query.where.length > 0) {
    sql += "\n" + this.padding + "where";
    this.increasePadding();
    for (var i = 0; i < query.where.length; i++) {
      var tk = query.where[i];
      //
      // If use aggregate functions move it in having section (if there is not already)
      if (this.useAggregateFuctions(tk)) {
        var found = false;
        var tksql = this.createSqlExpression(tk, App.Database.QueryParts.HAVING);
        for (var j = 0; j < query.having.length; j++) {
          if (tksql === this.createSqlExpression(query.having[j], App.Database.QueryParts.HAVING)) {
            found = true;
            break;
          }
        }
        if (!found) {
          query.having.push(tk);
          useAggregate = true;
        }
      }
      else
        sql += "\n" + this.padding + (i === 0 ? "" : "and ") + this.createSqlExpression(query.where[i], App.Database.QueryParts.WHERE);
    }
    this.decreasePadding();
  }
  //
  if (!useAggregate) {
    delete query.group;
    delete query.having;
  }
  //
  // GroupBy clauses
  if (query.group && query.group.length > 0) {
    sql += "\n" + this.padding + "group by";
    this.increasePadding();
    for (var i = 0; i < query.group.length; i++)
      sql += (i === 0 ? "" : ",") + "\n" + this.padding + this.createSqlExpression(query.group[i], App.Database.QueryParts.GROUPBY);
    this.decreasePadding();
  }
  //
  // Having clauses
  if (query.having && query.having.length > 0) {
    sql += "\n" + this.padding + "having";
    this.increasePadding();
    for (var i = 0; i < query.having.length; i++)
      sql += "\n" + this.padding + (i === 0 ? "" : "and ") + this.createSqlExpression(query.having[i], App.Database.QueryParts.HAVING);
    this.decreasePadding();
  }
  //
  if (query.union)
    sql += "\n" + this.padding + "union" + (query.union.all ? " all" : "") + "\n" + this.padding + this.createSelectStatement(query.union);
  //
  // OrderBy clauses
  if (query.order && query.order.length > 0) {
    sql += "\n" + this.padding + "order by";
    this.increasePadding();
    for (var i = 0; i < query.order.length; i++)
      sql += (i === 0 ? "" : ",") + "\n" + this.padding + this.createSqlExpression(query.order[i], App.Database.QueryParts.ORDERBY);
    this.decreasePadding();
  }
  //
  // Limit clause
  if (query.top) {
    var t = this.limitClause(query.top, false);
    if (t)
      sql += "\n" + this.padding + t;
    //
    // Offset clause
    if (query.offset) {
      var o = this.offsetClause(query.offset, false);
      if (o)
        sql += " " + o;
    }
  }
  //
  this.queries.pop();
  return sql;
};


/**
 * Compose update statement of a query
 * @param {Object} query to compose
 */
App.Database.prototype.createUpdateStatement = function (query)
{
  var sql = "update " + this.quote(this.getObjDMCode(query.table)) + " set";
  if (query.select) {
    this.increasePadding();
    for (var i = 0; i < query.select.length; i++) {
      var s = query.select[i];
      sql += (i === 0 ? "" : ",") + "\n";
      if (s.field) {
        sql += this.padding + this.quote(this.getObjDMCode(query.table, s.field));
        sql += " = " + this.createSqlExpression(s);
      }
      else
        sql += this.padding + this.createSqlExpression(s);
    }
    this.decreasePadding();
  }
  sql += this.createSqlWhereClauses(query.where);
  return sql;
};


/**
 * Compose insert statement of a query
 * @param {Object} query to compose
 */
App.Database.prototype.createInsertStatement = function (query)
{
  var sql = "insert into " + this.quote(this.getObjDMCode(query.table));
  //
  this.increasePadding();
  if (query.select) {
    var fields = "";
    var values = "";
    for (var i = 0; i < query.select.length; i++) {
      var s = query.select[i];
      //
      // Compose list of fields to value
      if (s.field)
        fields += (i === 0 ? "" : ",\n") + this.padding
                + this.quote(this.getObjDMCode(query.table, s.field));
      //
      // Compose list of values
      if (s.expr || s.p || "locVar" in s || s.asPar)
        values += (i === 0 ? "" : ",\n") + this.padding + this.createSqlExpression(s);
    }
    //
    if (fields.length > 0)
      sql += " (\n" + fields + "\n)";
    //
    if (values.length > 0)
      sql += " values (\n" + values + "\n)";
  }
  this.decreasePadding();
  //
  if (query.subquery)
    sql += "\n" + this.createSelectStatement(query.subquery);
  //
  return sql;
};


/**
 * Compose delete statement of a query
 * @param {Object} query to compose
 */
App.Database.prototype.createDeleteStatement = function (query)
{
  var sql = "delete from ";
  sql += this.quote(this.getObjDMCode(query.table));
  //
  // Table alias
  if (query.alias)
    sql += " " + this.quote(query.alias);
  //
  sql += this.createSqlWhereClauses(query.where);
  return sql;
};


/**
 * Compose a custom statement of a query
 * @param {Object} query to compose
 */
App.Database.prototype.createCustomStatement = function (query)
{
  var sql = "";
  if (query.custom) {
    for (var i = 0; i < query.custom.length; i++)
      sql += this.createSqlExpression(query.custom[i]);
  }
  return sql;
};


/**
 * Compose sql limit clause
 * @param {Object} token top
 * @param {Boolean} afterSelect
 */
App.Database.prototype.limitClause = function (token, afterSelect)
{
  if (afterSelect)
    return "";
  else
    return "limit " + this.createSqlExpression(token);
};


/**
 * Compose sql offset clause
 * @param {Object} token offset
 */
App.Database.prototype.offsetClause = function (token)
{
  return "offset " + this.createSqlExpression(token);
};


/**
 * Compose sql expression
 * @param {String} token to compose
 * @param {String} part of expression (see App.Database.QueryParts)
 */
App.Database.prototype.createSqlExpression = function (token, part)
{
  if (part === App.Database.QueryParts.FROM)
    return this.createSqlJoinExpression(token);
  else if (token.select)
    return "(" + this.createSelectStatement(token).trim() + ")";
  //
  var sql = "";
  if (token.from !== undefined) { // Is a table field
    // Add alias of table
    sql += (token.from ? this.quote(token.from) + "." : "");
    sql += this.quote(this.getObjDMCode(this.getTableFromAlias(token.from), token.expr));
  }
  else if (token.asPar)
    sql += this.defineQueryParameter(token);
  else if (token.expr) {
    var funct = this.functions[token.expr];
    if (!funct || !funct.expr) { // Function not found -> watch info supplied by IDE
      if (token.funct) { // Is a generic function
        sql += token.expr;
        sql += "(";
        if (token.p) {
          for (var i = 0; i < token.p.length; i++)
            sql += (i === 0 ? "" : ", ") + this.createSqlExpression(token.p[i]);
        }
        sql += ")";
      }
      else if (token.op) { // Is an generic operator
        if (!token.p)
          sql += token.expr;
        else if (token.p.length === 1)
          sql += token.expr + " " + this.createSqlExpression(token.p[0]);
        else if (token.p.length === 2)
          sql += this.createSqlExpression(token.p[0]) + " " + token.expr + " " + this.createSqlExpression(token.p[1]);
      }
      else if (token.clientRef) // Is a client object reference
        sql += (token.value !== undefined ? this.toSql(token.value) : token.expr);
      else // Whatever
        sql += this.createSqlToken(token);
    }
    else { // Is a specific function
      if (token.expr === "*" && !token.p)
        sql += token.expr;
      else {
        funct = funct.expr;
        if (token.p) {
          for (var i = token.p.length - 1; i >= 0; i--)
            funct = funct.replace(new RegExp("@" + (i + 1), "g"), this.createSqlExpression(token.p[i]));
        }
        sql += funct;
      }
    }
  }
  else if ("locVar" in token)  // It's a local variable
    sql += this.toSql(token.locVar);
  else if (token.p) { // Is a simple list of tokens
    var posOp = -1;
    var pars = [];
    for (var i = 0; i < token.p.length; i++) {
      var tk = token.p[i];
      //
      // Find first operator
      if (posOp === -1 && tk.expr !== "between" && !tk.p) {
        var funct = this.functions[tk.expr];
        if (funct && funct.op) {
          posOp = i;
          continue;
        }
      }
      pars.push(tk);
    }
    //
    // If found an operator detect operands
    if (posOp >= 0) {
      var nt = {expr: token.p[posOp].expr, op: true, p: []};
      if (posOp === 0)
        nt.p.push(pars.length === 1 ? pars[0] : {p: pars});
      else {
        var opExpr = this.functions[nt.expr].expr;
        if (opExpr.indexOf(nt.expr) === 0)
          sql += this.createSqlExpression({p: pars.splice(0, posOp)});
        else
          nt.p.push(posOp === 1 ? pars.shift() : {p: pars.splice(0, posOp)});
        if (pars.length > 0)
          nt.p.push(pars.length === 1 ? pars.shift() : {p: pars});
      }
      sql += " " + this.createSqlExpression(nt);
    }
    else { // Concats list of tokens
      for (var i = 0; i < pars.length; i++) {
        if (i > 0) {
          var prev = pars[i - 1].expr;
          var curr = pars[i].expr;
          if (pars[i - 1] && pars[i - 1].op)
            sql += " ";
          else if (prev && prev[prev.length - 1] !== " " && ["(", ".", "["].indexOf(prev) === -1 &&
                  curr && curr[0] !== " " && ["(", ")", ".", "[", "]", ","].indexOf(curr.trim()) === -1)
            sql += " ";
        }
        sql += this.createSqlExpression(pars[i]);
      }
    }
  }
  else // In other all cases add null value
    sql += "null";
  //
  if (part === App.Database.QueryParts.SELECT) {
    // If is a field set alias to logical name
    if (!token.alias && token.from !== undefined) {
      var a = this.getObjDMCode(this.getTableFromAlias(token.from), token.expr);
      if (a && a !== token.expr)
        token.alias = token.expr;
    }
    if (token.alias) {
      // If the last token I've added was a comment, "inject" the alias before that comment
      var lasttk = (token.p && token.p[token.p.length - 1].expr ? token.p && token.p[token.p.length - 1].expr : "");
      if (lasttk && lasttk.trim().substring(0, 2) === "--")
        sql = sql.substring(0, sql.length - lasttk.length - 1) + " as " + this.quote(token.alias) + lasttk;
      else
        sql += " as " + this.quote(token.alias);
    }
  }
  //
  if (token.order)
    sql += " " + (token.order === "a" ? "asc" : "desc");
  //
  return sql;
};


/**
 * Compose sql for a simple token
 * @param {Object} token to compose
 */
App.Database.prototype.createSqlToken = function (token)
{
  return token.expr;
};


/**
 * Compose sql join expression
 * @param {Object} token to compose
 */
App.Database.prototype.createSqlJoinExpression = function (token)
{
  var sql = "";
  //
  // Join type
  if (token.join)
    sql += token.join + " ";
  //
  // Table name
  sql += this.quote(this.getObjDMCode(token.expr));
  //
  // Table alias
  if (token.alias)
    sql += " " + this.quote(token.alias);
  //
  // Join clause
  if (token.p)
    sql += " on " + this.createSqlExpression({p: token.p});
  //
  return sql;
};


/**
 * Compose sql where clauses
 * @param {Array} where clauses
 */
App.Database.prototype.createSqlWhereClauses = function (where)
{
  if (!where || where.length === 0)
    return "";
  //
  var sql = "\nwhere";
  this.increasePadding();
  for (var i = 0; i < where.length; i++)
    sql += "\n" + this.padding + (i === 0 ? "" : "and ") + this.createSqlExpression(where[i], App.Database.QueryParts.WHERE);
  this.decreasePadding();
  return sql;
};


/**
 * Verify if expression use aggregate functions
 * @param {String} token to test
 */
App.Database.prototype.useAggregateFuctions = function (token)
{
  if (token.from !== undefined)
    return false;
  //
  if (token.funct && token.aggr)
    return true;
  //
  if (token.p) {
    for (var i = 0; i < token.p.length; i++) {
      if (this.useAggregateFuctions(token.p[i]))
        return true;
    }
  }
  //
  var funct = this.functions[token.expr];
  if (!funct || !funct.expr)
    return false;
  //
  if (funct.aggr)
    return true;
  //
  return false;
};


/**
 * Return table with the given alias
 * @param {String} alias
 */
App.Database.prototype.getTableFromAlias = function (alias)
{
  for (var j = this.queries.length - 1; j >= 0; j--) {
    var qry = this.queries[j];
    switch (qry.type) {
      case App.Database.QueryTypes.SELECT:
        var from = qry.from;
        if (from) {
          for (var i = 0; i < from.length; i++)
            if (alias === "" || alias === (from[i].alias || from[i].expr))
              return from[i].expr;
        }
        break;

      case App.Database.QueryTypes.UPDATE:
      case App.Database.QueryTypes.INSERT:
      case App.Database.QueryTypes.DELETE:
        return qry.table;
    }
  }
};


/**
 * Return phisical name of an object (table or field)
 * @param {String} table
 * @param {String} field
 */
App.Database.prototype.getObjDMCode = function (table, field)
{
  var tmap = this.dmCodesMap[table];
  return (tmap && tmap[field || "_"]) || field || table;
};


/**
 * Returns the number of rows of a resultSet
 * @param {Object} rs
 */
App.Database.prototype.getRowCount = function (rs)
{
  return rs.rows.length;
};


/**
 * Returns the i-th row of a resultSet
 * @param {Object} rs - recordset
 * @param {int} pos - position
 * @param {DataSchema} schema
 */
App.Database.prototype.getRow = function (rs, pos, schema)
{
};


/**
 * Returns the default field maxlength
 * @param {Object} f
 */
App.Database.prototype.getFieldMaxLength = function (f)
{
  if (f.maxlen >= 0)
    return parseInt(f.maxlen);
  //
  switch (f.datatype) {
    case App.Database.DataTypes.INT:
    case App.Database.DataTypes.AUTOINCR:
      return 9;

    case App.Database.DataTypes.FLOAT:
      return 7;

    case App.Database.DataTypes.DECIMAL:
      return 10;

    case App.Database.DataTypes.UUID:
      return 36;

    default:
      return 0;
  }
};


/**
 * Returns the default field scale
 * @param {Object} f
 */
App.Database.prototype.getFieldScale = function (f)
{
  if (f.scale >= 0)
    return parseInt(f.scale);
  //
  if (f.datatype === App.Database.DataTypes.DECIMAL)
    return 2;
  else
    return 0;
};


/**
 * Returns the field default value
 * @param {Object} f
 */
App.Database.prototype.getFieldDefVal = function (f)
{
  if (App.Database.classDT(f.datatype) === App.Database.DataTypeClasses.DATETIME) {
    if (f.defval === "now")
      return this.functions.now.expr;
    else if (f.defval === "today")
      return this.functions.today.expr;
    else if (f.defval === "time")
      return this.functions.time.expr;
  }
  else if (App.Database.classDT(f.datatype) === App.Database.DataTypeClasses.STRING) {
    if (f.defval)
      return App.Database.quoteString(f.defval);
  }
  return f.defval;
};


/**
 * Prepare a constant string for SQL
 * @param {Object} v
 * @param {String} dt - dataType
 * @param {Integer} ml - maxLength
 * @param {Integer} sc - scale
 */
App.Database.prototype.toSql = function (v, dt, ml, sc)
{
  if (v instanceof Array && dt !== App.Database.DataTypes.JSON) {
    var s = "(";
    for (var i = 0; i < v.length; i++)
      s += (i > 0 ? "," : "") + this.toSql(v[i], dt);
    s += ")";
    return s;
  }
  //
  v = this.forSql(v, dt);
  //
  if (typeof v === "number" && dt === App.Database.DataTypes.DECIMAL)
    return v.toFixed(sc);
  //
  if (typeof v === "string")
    return App.Database.quoteString(v);
  //
  return v + "";
};


/**
 * Prepare a value for SQL
 * @param {Object} v
 * @param {String} dt - dataType
 */
App.Database.prototype.forSql = function (v, dt)
{
  if (v === null || v === undefined)
    return null;
  //
  if (v instanceof Date) {
    if (this.app && dt === App.Database.DataTypes.TIME)
      return this.app.locale.moment(v).format("HH:mm:ss.SSS");
    else if (this.app && dt === App.Database.DataTypes.DATE)
      return this.app.locale.moment(v).format("YYYY-MM-DD");
    else
      return v.toISOString();
  }
  //
  if (dt === App.Database.DataTypes.BLOB)
    return new Buffer(v);
  //
  if (typeof v === "object" || dt === App.Database.DataTypes.JSON) {
    try {
      return JSON.stringify(v);
    }
    catch (e) {
      return v.toString();
    }
  }
  //
  if (dt === App.Database.DataTypes.BOOL)
    return v ? 1 : 0;
  //
  if (typeof v !== "string" && App.Database.classDT(dt) === App.Database.DataTypeClasses.STRING)
    return v.toString ? v.toString() : v + "";
  //
  return v;
};


/**
 * Convert a native value into javascript value
 * @param {String} v
 * @param {String} dt
 */
App.Database.prototype.convertValue = function (v, dt)
{
  if (v === null || v === undefined)
    return null;
  //
  // Adapt the datatype to the expected one
  if (App.Database.classDT(dt) === App.Database.DataTypeClasses.DATETIME) {
    if (v instanceof Date) {
      if (this.app) {
        var date = App.moment(v);
        if (dt === App.Database.DataTypes.TIME)
          return this.app.locale.moment([1970, 0, 1, date.hour(), date.minute(), date.second(), date.millisecond()]).toDate();
        else if (dt === App.Database.DataTypes.DATE)
          return this.app.locale.moment([date.year(), date.month(), date.date()]).toDate();
      }
    }
    else {
      // Time in Postgres & SQLite has format HH:mm:ss.SSS
      if (dt === App.Database.DataTypes.TIME) {
        if (v.startsWith("1970-01-01T")) {
          if (this.app)
            v = "1970-01-01T" + this.app.locale.moment(v).utc().format("HH:mm:ss.SSS");
        }
        else
          v = "1970-01-01T" + v;
      }
      else if (dt === App.Database.DataTypes.DATE)
        v = v.substring(0, 10);
      //
      var date = null;
      if (this.app && dt !== App.Database.DataTypes.DATETIME)
        date = this.app.locale.moment(v).toDate();
      else
        date = new Date(v);
      return (isNaN(date.getTime()) ? null : date);
    }
  }
  //
  if (App.Database.classDT(dt) === App.Database.DataTypeClasses.NUMBER && typeof v !== "number") {
    if (dt === App.Database.DataTypes.INT)
      return parseInt(v);
    else
      return parseFloat(v);
  }
  //
  if (App.Database.classDT(dt) === App.Database.DataTypeClasses.BOOL && typeof v !== "boolean")
    return (v === "0" || v === "false" ? false : true);
  //
  if (App.Database.classDT(dt) === App.Database.DataTypeClasses.STRING && typeof v !== "string")
    return v + "";
  //
  return v;
};


/*
 * Set the connection String
 * @param {String} conString
 */
App.Database.prototype.setDbConString = function (conString, dbnamePrefix)
{
  this.dbnamePrefix = dbnamePrefix;
  this.setOptions(conString + "/" + this.getDBName());
};


/*
 * Initialize the connection String
 * @param {String} conString
 */
App.Database.prototype.initDbConString = function (conString, dbnamePrefix)
{
  var url = (App.remoteDBurls ? App.remoteDBurls[this.getDBName()] : undefined);
  if (url)
    return this.setOptions(url);
  //
  this.setDbConString(conString, dbnamePrefix);
};


/*
 * Return the phisical name of database
 * @return {String}
 */
App.Database.prototype.getDBName = function ()
{
  return (this.dbnamePrefix ? this.dbnamePrefix + "-" : "") + this.schema.name;
};


/*
 * Increase the padding
 */
App.Database.prototype.increasePadding = function ()
{
  if (this.padding === undefined)
    this.padding = "";
  else
    this.padding += new Array(this.indent + 1).join(" ");
};


/*
 * Decrease the padding
 */
App.Database.prototype.decreasePadding = function ()
{
  if (this.padding.length >= this.indent)
    this.padding = this.padding.substr(0, this.padding.length - this.indent);
  else
    delete this.padding;
};


/*
 * Begin a transaction
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.beginTransaction = function (callback)
{
  // If just in transaction exit adding warning
  var logObj = this.log("Begin transaction" + (!this.isInTransaction() ? "" : " (already in transaction)"));
  if (this.isInTransaction()) {
    if (this.app)
      this.app.dttobj.addWarning("", "begin called in transaction");
    return callback();
  }
  //
  // Suspend auto closing
  this.handleAutoClose(1);
  //
  var done = function (tr, error) {
    if (error)
      this.handleAutoClose(-1); // Program auto closing
    else
      this.transaction = tr || true;
    //
    this.log(error || "OK", logObj);
    //
    callback(null, error);
  }.bind(this);
  //
  this.openConnection(function (result, error) {
    if (error)
      return done(null, error);
    //
    if (this.remoteConnector)
      this.connection.beginTransaction(done);
    else
      this._beginTransaction(done);
  }.bind(this));
};


/*
 * Commit a transaction
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.commitTransaction = function (callback)
{
  var logObj = this.log("Commit transaction" + (this.isInTransaction() ? "" : " (out of transaction)"));
  var done = function (result, error) {
    delete this.transaction;
    //
    // Program auto closing
    this.handleAutoClose(-1);
    //
    this.log(error || "OK", logObj);
    //
    // Call function interested to end transaction (if there is)
    if (this.endTransactionCallback)
      this.endTransactionCallback(!error);
    //
    callback(null, error);
  }.bind(this);
  //
  // If out of transaction add warning
  if (!this.isInTransaction()) {
    if (this.app)
      this.app.dttobj.addWarning("", "commit called out transaction");
    return done();
  }
  //
  if (this.remoteConnector)
    this.connection.commitTransaction(done);
  else
    this._commitTransaction(done);
};


/*
 * Rollback a transaction
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.rollbackTransaction = function (callback)
{
  var logObj = this.log("Rollback transaction" + (this.isInTransaction() ? "" : " (out of transaction)"));
  var done = function (result, error) {
    delete this.transaction;
    //
    // Program auto closing
    this.handleAutoClose(-1);
    //
    this.log(error || "OK", logObj);
    //
    // Call function interested to end transaction (if there is)
    if (this.endTransactionCallback)
      this.endTransactionCallback(false);
    //
    callback(null, error);
  }.bind(this);
  //
  // If out of transaction add warning
  if (!this.isInTransaction()) {
    if (this.app)
      this.app.dttobj.addWarning("", "rollback called out transaction");
    return done();
  }
  //
  if (this.remoteConnector)
    this.connection.rollbackTransaction(done);
  else
    this._rollbackTransaction(done);
};


/**
 * Read list of objects
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.listSchema = function (options, callback)
{
  if (!options.schema)
    options.schema = this.defaultSchema;
  //
  var list = {};
  this.listTables(options, function (tables, error) {
    if (error)
      return callback(null, error);
    //
    if (tables.length > 0)
      list.tables = tables;
    //
    this.listViews(options, function (views, error) {
      if (error)
        return callback(null, error);
      //
      if (views.length > 0)
        list.views = views;
      //
      callback(list);
    });
  }.bind(this));
};


/**
 * Read list of tables
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.listTables = function (options, callback)
{
  var sql = "select table_name";
  sql += " from INFORMATION_SCHEMA.TABLES";
  sql += " where table_schema = '" + options.schema + "' and table_type = 'BASE TABLE'";
  sql += " order by table_name";
  this.executeSQL(sql, true, function (result, error) {
    if (error)
      return callback(null, error);
    //
    var tables = [];
    for (var i = 0; i < this.getRowCount(result); i++)
      tables.push(this.getRow(result, i).table_name);
    callback(tables);
  }.bind(this));
};


/**
 * Read list of views
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.listViews = function (options, callback)
{
  var sql = "select table_name";
  sql += " from INFORMATION_SCHEMA.TABLES";
  sql += " where table_schema = '" + options.schema + "' and table_type = 'VIEW'";
  sql += " order by table_name";
  this.executeSQL(sql, true, function (result, error) {
    if (error)
      return callback(null, error);
    //
    var views = [];
    for (var i = 0; i < this.getRowCount(result); i++)
      views.push(this.getRow(result, i).table_name);
    callback(views);
  }.bind(this));
};


/**
 * Import schema of database
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importSchema = function (options, callback)
{
  if (!options.schema)
    options.schema = this.defaultSchema;
  //
  // Import tables
  this.importTables(options, function (tables, error) {
    if (error)
      return callback(null, error);
    //
    var schema = {};
    if (tables.length > 0)
      schema.tables = tables;
    //
    // Import views
    this.importViews(options, function (views, error) {
      if (error)
        return callback(null, error);
      //
      if (views.length > 0)
        schema.views = views;
      //
      callback(schema);
    });
  }.bind(this));
};


/**
 * Import tables of database
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importTables = function (options, callback)
{
  if (!options.tables)
    return callback([]);
  //
  var tables = [];
  var next = function (result, error) {
    if (error)
      return callback(null, error);
    //
    if (result)
      tables.push(result);
    //
    var table = options.tables.pop();
    if (!table)
      return callback(tables);
    //
    this.importTable({table: table, schema: options.schema}, next);
  }.bind(this);
  next();
};


/**
 * Import views of database
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importViews = function (options, callback)
{
  if (!options.views)
    return callback([]);
  //
  var views = [];
  var next = function (result, error) {
    if (error)
      return callback(null, error);
    //
    if (result)
      views.push(result);
    //
    var view = options.views.pop();
    if (!view)
      return callback(views);
    //
    this.importView({view: view, schema: options.schema}, next);
  }.bind(this);
  next();
};


/**
 * Import a table
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importTable = function (options, callback)
{
  var table = {};
  table.name = options.table;
  //
  // Import fields
  this.importTableFields(options, function (fields, error) {
    if (error)
      return callback(null, error);
    //
    if (fields.length > 0)
      table.fields = fields;
    //
    // Import fks
    this.importTableFKs(options, function (fks, error) {
      if (error)
        return callback(null, error);
      //
      if (fks.length > 0)
        table.fks = fks;
      //
      // Import indexes
      this.importTableIndexes(options, function (indexes, error) {
        if (error)
          return callback(null, error);
        //
        if (indexes.length > 0)
          table.indexes = indexes;
        //
        callback(table);
      });
    }.bind(this));
  }.bind(this));
};


/**
 * Import a view
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importView = function (options, callback)
{
  var view = {};
  view.name = options.view;
  //
  // Import fields
  this.importViewFields(options, function (fields, error) {
    if (error)
      return callback(null, error);
    //
    if (fields.length > 0)
      view.fields = fields;
//
    callback(view);
  });
};


/**
 * Import fiels of a table
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importTableFields = function (options, callback)
{
  var fields = [];
  //
  var sql = "select column_name, table_schema, column_default, ";
  sql += " is_nullable, data_type, character_maximum_length, numeric_precision, numeric_scale, ";
  sql += " (select count(*)";
  sql += " from INFORMATION_SCHEMA.TABLE_CONSTRAINTS TC";
  sql += " join INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE CCU";
  sql += " on (C.column_name = CCU.column_name";
  sql += "  and TC.constraint_name = CCU.constraint_name";
  sql += "  and TC.table_name = CCU.table_name";
  sql += "  and TC.table_schema = CCU.table_schema)";
  sql += " where TC.constraint_type = 'PRIMARY KEY'";
  sql += "  and C.table_schema = TC.table_schema";
  sql += "  and C.table_name = TC.table_name) as pk";
  sql += " from INFORMATION_SCHEMA.COLUMNS C";
  sql += " where C.table_name = '" + options.table + "' and C.table_schema = '" + options.schema + "'";
  sql += " order by ordinal_position";
  this.executeSQL(sql, true, function (result, error) {
    if (error)
      return callback(null, error);
    //
    for (var i = 0; i < this.getRowCount(result); i++) {
      var row = this.getRow(result, i);
      var field = {};
      field.name = row.column_name;
      if (row.pk > 0)
        field.pk = true;
      if (row.is_nullable === "NO")
        field.notNull = true;
      field.dataType = this.convertDataType(row.data_type);
      if ((row.character_maximum_length || row.numeric_precision) > 0)
        field.maxLen = row.character_maximum_length || row.numeric_precision;
      if (field.dataType === App.Database.DataTypes.DECIMAL
              && (row.numeric_scale > 0 || row.numeric_scale === 0))
        field.scale = row.numeric_scale;
      if (row.column_default) {
        if (row.column_default.startsWith("'") && row.column_default.endsWith("'"))
          field.defaultValue = row.column_default.substring(1, row.column_default.length - 1).replace(/''/g, "");
        else
          field.defaultValue = row.column_default;
      }
      fields.push(field);
    }
    //
    callback(fields);
  }.bind(this));
};


/**
 * Import fiels of a view
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importViewFields = function (options, callback)
{
  options.table = options.view;
  delete options.view;
  App.Database.prototype.importTableFields.call(this, options, callback);
};


/**
 * Import fks of a table
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importTableFKs = function (options, callback)
{
  var fks = [];
  //
  // Import fks
  var sql = "select";
  sql += " rc.constraint_name,";
  sql += " rc.update_rule,";
  sql += " rc.delete_rule,";
  sql += " kcu1.column_name as fk_column_name,";
  sql += " kcu2.table_name as ref_table_name,";
  sql += " kcu2.column_name as ref_column_name ";
  sql += "from INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS as rc ";
  sql += " inner join INFORMATION_SCHEMA.KEY_COLUMN_USAGE as kcu1 ";
  sql += "  on (kcu1.constraint_schema = rc.constraint_schema ";
  sql += "   and kcu1.constraint_name = rc.constraint_name) ";
  sql += " inner join INFORMATION_SCHEMA.KEY_COLUMN_USAGE as kcu2 ";
  sql += "  on (kcu2.constraint_schema = rc.unique_constraint_schema ";
  sql += "   and kcu2.constraint_name = rc.unique_constraint_name ";
  sql += "   and kcu2.ordinal_position = kcu1.ordinal_position) ";
  sql += "where kcu1.table_name = '" + options.table + "' and kcu1.table_schema = '" + options.schema + "' ";
  sql += "order by rc.constraint_name, kcu1.ordinal_position";
  this.executeSQL(sql, true, function (result, error) {
    if (error)
      return callback(null, error);
    //
    var fk = {};
    for (var i = 0; i < this.getRowCount(result); i++) {
      var row = this.getRow(result, i);
      if (row.constraint_name !== fk.name) {
        fk = {};
        fk.name = row.constraint_name;
        fk.derived = row.ref_table_name;
        fk.refs = {};
        switch (row.update_rule) {
          case "NO ACTION":
            fk.updateRule = App.Database.FKActions.NOACTION;
            break;
          case "CASCADE":
            fk.updateRule = App.Database.FKActions.CASCADE;
            break;
          case "SET NULL":
            fk.updateRule = App.Database.FKActions.SETNULL;
            break;
          case "SET DEFAULT":
            fk.updateRule = App.Database.FKActions.SETDEFAULT;
            break;
        }
        switch (row.delete_rule) {
          case "NO ACTION":
            fk.deleteRule = App.Database.FKActions.NOACTION;
            break;
          case "CASCADE":
            fk.deleteRule = App.Database.FKActions.CASCADE;
            break;
          case "SET NULL":
            fk.deleteRule = App.Database.FKActions.SETNULL;
            break;
          case "SET DEFAULT":
            fk.deleteRule = App.Database.FKActions.SETDEFAULT;
            break;
        }
        fks.push(fk);
      }
      //
      fk.refs[row.fk_column_name] = row.ref_column_name;
    }
    //
    callback(fks);
  }.bind(this));
};


/**
 * Check if a name require case sensitivity
 * @param {String} name
 */
App.Database.prototype.requireCaseSensitivity = function (name)
{
};


/**
 * Import indexes of a table
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.importTableIndexes = function (options, callback)
{
  callback([]);
};


/*
 * Convert a native dataType
 * @param {String}
 */
App.Database.prototype.convertDataType = function (dt)
{
  for (var i in this.typemap) {
    if (this.typemap[i].indexOf(dt.toLowerCase()) !== -1)
      return i;
  }
};


/*
 * Indicates if there is an open transaction
 * @return {Boolean}
 */
App.Database.prototype.isInTransaction = function ()
{
  return (this.transaction ? true : false);
};


/*
 * Returns the inserted id by the last sql statement
 * @return {Integer}
 */
App.Database.prototype.getLastInsertId = function ()
{
  return this.lastInsertId;
};


/*
 * Returns the rows affected by the last sql statement
 * @return {Integer}
 */
App.Database.prototype.getLastRowsAffected = function ()
{
  return this.lastRowsAffected;
};


/*
 * Do nothing
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.ping = function (callback)
{
  if (!this.remoteConnector)
    return callback();
  //
  var rd = new App.RemoteDriver(this);
  rd.ping(callback);
};


/*
 * Define a query parameter
 * @param {Object} token
 */
App.Database.prototype.defineQueryParameter = function (token)
{
};


/*
 * Store a message for sync object
 * @param {String} handler - handler of message
 * @param {Object} message
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncStoreMessage = function (handler, message, callback)
{
  this.syncCheckTable(handler, function (result, error) {
    if (error)
      return callback(false);
    //
    this.syncInsertIntoTable(handler, message, function (result, error) {
      if (error)
        callback(false);
      else
        callback(true);
    });
  }.bind(this));
};


/*
 * Create z_sync table if not exists
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncCheckTable = function (handler, callback)
{
  if (!this.syncTableChecked)
    this.syncTableChecked = [];
  //
  // only if I have not already done
  if (this.syncTableChecked[handler])
    return callback(true);
  //
  if (this.syncTableChecked[handler] === 0) {
    // There is already a syncCheckTable underway
    // I have to wait for it to finish
    setTimeout(function () {
      this.syncCheckTable(handler, callback);
    }.bind(this), 100);
    return;
  }
  //
  this.syncTableChecked[handler] = 0;
  //
  this.syncExistsTable(handler, function (result) {
    if (result) {
      this.syncTableChecked[handler] = true;
      return callback(true);
    }
    //
    this.syncCreateTable(handler, function (result, error) {
      if (error) {
        this.syncTableChecked[handler] = false;
        return callback(false);
      }
      //
      this.syncInsertIntoTable(handler, {ts: null}, function (result, error) {
        if (error) {
          this.syncTableChecked[handler] = false;
          return callback(false);
        }
        //
        this.syncTableChecked[handler] = true;
        callback(true);
      }.bind(this));
    }.bind(this));
  }.bind(this));
};


/*
 * Verify if z_sync table exists
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncExistsTable = function (handler, callback)
{
  var sql = "select table_name";
  sql += " from INFORMATION_SCHEMA.TABLES";
  sql += " where table_schema = '" + this.defaultSchema + "'";
  sql += " and table_type = 'BASE TABLE'";
  sql += " and table_name = 'z_sync" + handler + "'";
  this.executeSQL(sql, true, function (result) {
    callback(result && this.getRowCount(result) === 1);
  }.bind(this));
};


/*
 * Returns the number of message stored
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncStoredMessageCount = function (handler, callback)
{
  this.executeSQL("select count(*) as " + this.quote("count") + " from z_sync" + handler + " where id > 1", true, function (result, error) {
    if (error)
      callback(null, error);
    else
      callback(this.getRow(result, 0).count);
  }.bind(this));
};


/*
 * Create z_sync table
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncCreateTable = function (handler, callback)
{
};


/*
 * Insert a message into z_sync table
 * @param {String} handler - handler of message
 * @param {Object} message to insert
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncInsertIntoTable = function (handler, message, callback)
{
  // Compose the query
  var q = {
    type: App.Database.QueryTypes.INSERT,
    table: "z_sync" + handler,
    select: [],
    counter: "id"
  };
  //
  var ts = message.ts;
  if (ts === undefined)
    ts = new Date();
  if (App.Database.sequences)
    q.select.push({field: "id", expr: ++App.Database.sequences[handler]});
  if (message.sender)
    q.select.push({field: "sender", expr: this.toSql(message.sender)});
  if (message.uid)
    q.select.push({field: "uid", expr: this.toSql(message.uid)});
  if (message.device)
    q.select.push({field: "device", expr: this.toSql(message.device)});
  q.select.push({field: "ts", expr: this.toSql(ts, App.Database.DataTypes.DATETIME)});
  q.select.push({field: "topics", expr: this.toSql(JSON.stringify(message.topics))});
  if (message.body)
    q.select.push({field: "body", expr: this.toSql(JSON.stringify(message.body))});
  if (message.ttl)
    q.select.push({field: "expire", expr: this.toSql(new Date(ts.getTime() + (message.ttl * 1000)), App.Database.DataTypes.DATETIME)});
  //
  // Execute the query
  this.execute({qry: q}, function (result, error) {
    if (error)
      return callback(null, error);
    //
    if (App.Database.sequences)
      message.id = App.Database.sequences[handler];
    else
      message.id = this.lastInsertId;
    callback();
  }.bind(this));
};


/*
 * Return z_sync table schema
 */
App.Database.prototype.syncTableSchema = function ()
{
  if (!this._syncTableSchema) {
    this._syncTableSchema = {
      p: [
        {name: "id", datatype: App.Database.DataTypes.AUTOINCR, pk: true},
        {name: "uid", datatype: App.Database.DataTypes.CHAR},
        {name: "device", datatype: App.Database.DataTypes.CHAR},
        {name: "sender", datatype: App.Database.DataTypes.VARCHAR},
        {name: "topics", datatype: App.Database.DataTypes.VARCHAR},
        {name: "body", datatype: App.Database.DataTypes.CLOB},
        {name: "ts", datatype: App.Database.DataTypes.DATETIME},
        {name: "expire", datatype: App.Database.DataTypes.DATETIME}
      ]
    };
  }
  return this._syncTableSchema;
};


/*
 * Load some messages from z_sync table
 * @param {String} handler - handler of message
 * @param {Object} filters
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncLoadMessages = function (handler, filters, callback)
{
  this.syncCheckTable(handler, function (result, error) {
    if (error)
      return callback(null, error);
    //
    // Compose the query
    var q = {};
    var sch = null;
    q.type = App.Database.QueryTypes.SELECT;
    q.from = [{expr: "z_sync" + handler}];
    //
    if (filters.maxId) {
      sch = {p: [{name: "id", datatype: App.Database.DataTypes.INT}]};
      q.select = [{expr: "max", p: [{expr: "id"}], alias: "id"}];
    }
    else {
      sch = this.syncTableSchema();
      q.select = [
        {expr: "id", from: ""},
        {expr: "uid", from: ""},
        {expr: "device", from: ""},
        {expr: "sender", from: ""},
        {expr: "topics", from: ""},
        {expr: "body", from: ""},
        {expr: "ts", from: ""},
        {expr: "expire", from: ""}
      ];
      q.where = [];
      q.order = [{expr: "id"}];
      //
      // Add requested filters
      if (filters.fromId !== undefined)
        q.where.push({expr: ">=", p: [{expr: "id"}, {expr: filters.fromId}]});
      //
      if (filters.toId !== undefined)
        q.where.push({expr: "<=", p: [{expr: "id"}, {expr: filters.toId}]});
      //
      if (filters.expired !== undefined)
        q.where.push({expr: (filters.expired ? "<=" : ">"), p: [{expr: "expire"}, {expr: this.toSql(new Date(), App.Database.DataTypes.DATETIME)}]});
      //
      if (filters.uid)
        q.where.push({expr: "=", p: [{expr: "uid"}, {expr: this.toSql(filters.uid)}]});
      //
      if (filters.max)
        q.top = {expr: filters.max};
    }
    //
    // Execute the query
    this.query({qry: q, sch: sch}, function (result, error) {
      if (error)
        return callback(null, error);
      //
      // Load messages from datamap
      var messages = [];
      for (var i = 0; i < result.rows.length; i++) {
        var row = result.rows[i];
        var msg = {};
        msg.id = row.id;
        if (filters.maxId !== true) {
          msg.uid = row.uid;
          msg.device = row.device;
          msg.sender = row.sender;
          msg.topics = (row.topics === null ? undefined : JSON.parse(row.topics));
          msg.body = JSON.parse(row.body);
          msg.ts = row.ts;
          if (row.expire)
            msg.ttl = (row.expire.getTime() - row.ts.getTime()) / 1000;
        }
        messages.push(msg);
      }
      //
      callback(messages);
    });
  }.bind(this));
};


/*
 * Delete some messages from z_sync table
 * @param {String} handler - handler of message
 * @param {Object} filters
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncDeleteMessages = function (handler, filters, callback)
{
  this.syncCheckTable(handler, function (result, error) {
    if (error)
      return callback(null, error);
    //
    // Compose the query
    var q = {
      type: App.Database.QueryTypes.DELETE,
      table: "z_sync" + handler,
      where: []
    };
    //
    // Add requested filters
    if (filters.expired)
      q.where.push({expr: "<=", p: [{expr: "expire"}, {expr: this.toSql(new Date(), App.Database.DataTypes.DATETIME)}]});
    //
    if (filters.fromId)
      q.where.push({expr: ">=", p: [{expr: "id"}, {expr: filters.fromId}]});
    else
      q.where.push({expr: ">", p: [{expr: "id"}, {expr: 1}]});
    //
    if (filters.toId)
      q.where.push({expr: "<=", p: [{expr: "id"}, {expr: filters.toId}]});
    //
    if (filters.uid)
      q.where.push({expr: "=", p: [{expr: "uid"}, {expr: this.toSql(filters.uid)}]});
    //
    // Execute the statement
    this.execute({qry: q}, callback);
  }.bind(this));
};


/*
 * Update a messages into z_sync table
 * @param {String} handler - handler of message
 * @param {Object} message
 * @param {Function} callback - function to be called at the end
 */
App.Database.prototype.syncUpdateMessage = function (handler, message, callback)
{
  this.syncCheckTable(handler, function (result, error) {
    if (error)
      return callback(null, error);
    //
    // Compose the query
    var q = {
      type: App.Database.QueryTypes.UPDATE,
      table: "z_sync" + handler,
      select: [],
      where: [{expr: "=", p: [{expr: "id"}, {expr: message.id}]}]
    };
    //
    if (message.sender)
      q.select.push({field: "sender", expr: this.toSql(message.sender)});
    if (message.uid)
      q.select.push({field: "uid", expr: this.toSql(message.uid)});
    if (message.device)
      q.select.push({field: "device", expr: this.toSql(message.device)});
    if (message.topics)
      q.select.push({field: "topics", expr: this.toSql(JSON.stringify(message.topics))});
    if (message.type)
      q.select.push({field: "type", expr: this.toSql(message.type)});
    if (message.body)
      q.select.push({field: "body", expr: this.toSql(JSON.stringify(message.body))});
    if (message.ts)
      q.select.push({field: "ts", expr: this.toSql(message.ts, App.Database.DataTypes.DATETIME)});
    if (message.expire)
      q.select.push({field: "expire", expr: this.toSql(message.expire, App.Database.DataTypes.DATETIME)});
    //
    // Execute the statement
    this.execute({qry: q}, callback);
  }.bind(this));
};


/*
 * Log a message to file
 * @param {String/ResultSet} obj
 * @param {Object} ref
 */
App.Database.prototype.log = function (obj, ref)
{
  // Within IDE or with log disabled do nothing
  if (!this.app || !App.Session.getClassByName(this.schema.type).logEnabled)
    return;
  //
  // Initilize logger
  if (!this.logger) {
    this.logger = {};
    this.logger.id = App.Utils.generateUID24();
    this.logger.msgCounter = 0;
  }
  //
  // Compose log row
  var data = {};
  data.ts = new Date();
  data.dbId = this.logger.id;
  data.msgId = (ref ? ref.msgId : this.logger.msgCounter++);
  data.pending = this.pendingOperations || 0;
  if (ref) {
    if (obj instanceof Error)
      data.ERROR = obj.message;
    else {
      data.result = "OK";
      if (ref.readOnly)
        data.rows = this.getRowCount(obj);
    }
  }
  else {
    data.command = obj;
    data.tr = this.isInTransaction();
  }
  if (ref)
    data.elapsed = data.ts - ref.ts;
  //
  var write = function (result, error) {
    if (error)
      return;
    //
    this.logger.file.write(JSON.stringify(data) + "\n", 0, 0, 0, function () {
    });
  }.bind(this);
  //
  if (!this.logger.file)
    this.openLogFile(write);
  else
    write();
  //
  return data;
};


/*
 * Open log file
 * @param {function} callback
 */
App.Database.prototype.openLogFile = function (callback)
{
  var f = this.app.fs.file(this.schema.name + ".log", App.FS.internalType.temp);
  f.append(function (result, error) {
    if (!error)
      this.logger.file = f;
    callback(result, error);
  }.bind(this));
};


/*
 * Close log file
 * @param {function} callback
 */
App.Database.prototype.closeLogFile = function (callback)
{
  if (!this.logger.file) {
    if (callback)
      callback();
    return;
  }
  //
  this.logger.file.close(function () {
    delete this.logger.file;
    if (callback)
      callback();
  }.bind(this));
};


//export module for node
if (module)
  module.exports = App.Database;
