/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module */

var App = App || {};

/**
 * @class A DataColumm Object
 * @param {Object} parent
 * @param {Object} defprop
 * */
App.DataColumn = function (parent, defprop)
{
  if (module && !App.View && parent !== true)
    App = require("./requireAll");
  //
  this.index = undefined;
  this.name = undefined;
  this.datatype = undefined;
  this.maxlength = undefined;
  this.scale = undefined;
  this.nullable = true;
  this.pk = false;
  this.default = undefined;
  this.unbound = true;
  //
  this.elements = [];   // RemElem to update
  this.parent = (parent !== true ? parent : null); // My owner schema
  this.filters = [];    // QBEFilters to add to the query
  this._searchOptions = {}; // Options of search criteria
  //
  // Load default properties
  if (defprop) {
    for (var x in defprop)
      this[x] = defprop[x];
  }
};

// Getter/setter of some properties
Object.defineProperty(App.DataColumn.prototype, "searchOptions", {
  get: function () {
    var opts = JSON.parse(JSON.stringify(this.parent.parent.searchOptions));
    for (var o in this._searchOptions)
      opts[o] = this._searchOptions[o];
    return opts;
  },
  set: function (newValue) {
    this._searchOptions = newValue;
  },
  enumerable: true,
  configurable: true
});


/**
 * Link an element to this column
 * @param {RemElem} el - element to link
 * @param {String} prop - prop of element to set/get
 */
App.DataColumn.prototype.addElement = function (el, prop)
{
  var oel = {element: el, prop: prop};
  //
  // If you did not specify the property to associate
  // choose the most appropriate based on the type of element
  if (!prop) {
    switch (el.class) {
      case "Input":
        switch (el.type) {
          case "checkbox":
            oel.prop = "checked";
            break;
          default:
            oel.prop = "value";
            break;
        }
        break;
      case "Meter":
      case "Autocomplete":
        oel.prop = "value";
        break;
      case "Image":
      case "Iframe":
        oel.prop = "src";
        break;
      default:
        oel.prop = "innerText";
    }
  }
  //
  // Let's see if elements is a template one
  if (this.parent.template && (el === this.parent.template || el.isChildOf(this.parent.template)))
    oel.template = true;
  //
  // Let's see if elements is a parent template one
  if (this.parent.parent.parentElement instanceof App.DataMap && this.parent.parent.parentElement.template && (el === this.parent.parent.parentElement.template || el.isChildOf(this.parent.parent.parentElement.template)))
    oel.templateParent = true;
  //
  if (this.parent.header && (el === this.parent.header || el.isChildOf(this.parent.header)))
    oel.groupHeader = true;
  //
  if (this.parent.footer && (el === this.parent.footer || el.isChildOf(this.parent.footer)))
    oel.groupFooter = true;
  //
  // Calc if property is a style one
  if (oel.prop && oel.prop.substring(0, 6) === "style.") {
    oel.style = true;
    oel.prop = prop.substring(6);
  }
  //
  this.elements.push(oel);
};


/**
 * Verify if expression use aggregate functions
 */
App.DataColumn.prototype.useAggregateFuctions = function ()
{
  if (this.getSelectToken())
    return this.parent.parent.dataStore.useAggregateFuctions(this.getSelectToken());
  else
    return false;
};


/**
 * Prepare a constant string for SQL
 * @param {AnyType} v
 * @param {Database} dm
 */
App.DataColumn.prototype.toSql = function (v, dm) {
  if (!dm && this.parent.parent)
    dm = this.parent.parent.dataStore;
  if (!dm)
    return v;
  //
  var f = {maxlen: this.maxlength, scale: this.scale, datatype: this.datatype};
  var ml = parseInt(dm.getFieldMaxLength(f));
  var sc = parseInt(dm.getFieldScale(f));
  return dm.toSql(v, this.datatype, ml, sc);
};


/**
 * Clear all filters of this column
 */
App.DataColumn.prototype.clearFilters = function ()
{
  this.filters.length = 0;
};


/**
 * Add a filter to this column
 * @param {String} f - qbe criteria
 * @param {Boolean} or - true if it must be considered in or with other criteria
 * @param {Boolean} qbe - true if it must be considered a qbe criteria
 */
App.DataColumn.prototype.addFilter = function (f, or, qbe)
{
  var ff = {};
  if (f instanceof Array)
    ff.v = f;
  else if (f instanceof App.moment)
    ff.v = f.format("YYYY-MM-DD");
  else
    ff.v = f + "";
  ff.or = (or === true);
  ff.qbe = (qbe !== false);
  this.filters.push(ff);
  return ff;
};


/**
 * Get all conditions for filters
 * @param {Boolean} or - true if want only the criteria in or
 */
App.DataColumn.prototype.getAllFilterConditions = function (or)
{
  var wcs = [];
  for (var i = 0; i < this.filters.length; i++) {
    var f = this.filters[i];
    if (f.or !== or)
      continue;
    //
    wcs.push(this.createCondition(f));
  }
  return wcs;
};


/**
 * Get char separator of QBE criteria
 */
App.DataColumn.prototype.getQBESep = function ()
{
  return ";";
};


/**
 * Create a where clause token that represents the filter criteria
 * @param {Object} filter - filter criteria
 * @returns {Object}
 */
App.DataColumn.prototype.createCondition = function (filter)
{
  if (!this.getSelectToken())
    return null;
  //
  // Clone the select token
  var s = JSON.parse(JSON.stringify(this.getSelectToken()));
  //
  // Remove the alias
  delete s.alias;
  //
  var useLike = true;
  if (this.searchOptions.likeType === App.DataMap.likeTypes.noLike || this.datatype === App.Database.DataTypes.UUID)
    useLike = false;
  //
  // Empty criteria -> null
  var text = filter.v;
  if (!text)
    text = "!";
  //
  // Parsing criteria
  // Tokenize getQBESep()
  var cplxCri = null;
  var cri = "";
  var p1 = -1;
  var p2 = -1;
  var init = true;
  while (p1 >= 0 || init) {
    init = false;
    //
    // Filters as array are be translated as in operator
    if (text instanceof Array) {
      cplxCri = {expr: "in", op: true, p: [s, {expr: this.toSql(text)}]};
      break;
    }
    //
    if (text.charAt(0) !== '=' && filter.qbe)
      p1 = text.indexOf(this.getQBESep(), p1 + 1);
    //
    if (p1 >= 0)
      cri = text.substr(p2 + 1, p1 - (p2 + 1));
    else
      cri = text.substr(p2 + 1);
    //
    // Now cri contains a single criteria
    // Parse the first character, if any
    var sqlcri = "";
    var par = new Array(2);
    par[0] = null;
    par[1] = null;
    if (cri.length > 0) {
      if (filter.qbe) {
        if (cri.charAt(0) === '>') {
          sqlcri = ">";
          par[0] = cri.substr(1);
          //
          // If is a DATE / DATETIME try to complete the data
          if (this.datatype === App.Database.DataTypes.DATETIME || this.datatype === App.Database.DataTypes.DATE) {
            var d = App.Utils.parsePartialDate(par[0], 1);
            if (d) {
              sqlcri = ">=";
              par[0] = d;
            }
          }
        }
        else if (cri.charAt(0) === '<') {
          sqlcri = "<";
          par[0] = cri.substr(1);
          //
          // If is a DATE / DATETIME try to complete the data
          if (this.datatype === App.Database.DataTypes.DATETIME || this.datatype === App.Database.DataTypes.DATE) {
            var d = App.Utils.parsePartialDate(par[0], 0);
            if (d) {
              sqlcri = "<";
              par[0] = d;
            }
          }
        }
        else if (cri.charAt(0) === '.')
          sqlcri = "is not null";
        else if (cri.charAt(0) === '!')
          sqlcri = "is null";
        else if (cri.charAt(0) === '#') {
          cri = cri.substr(1);
          if (useLike && App.Database.classDT(this.datatype) === App.Database.DataTypeClasses.STRING) {
            if (cri.indexOf("*") === -1 && cri.indexOf("%") === -1) {
              if (this.searchOptions.likeType === App.DataMap.likeTypes.startsWith)
                cri = cri + "*";
              else if (this.searchOptions.likeType === App.DataMap.likeTypes.contains)
                cri = "*" + cri + "*";
            }
            //
            // E' un like!
            cri = cri.replace(/\*/g, '%');
            sqlcri = "not like";
            par[0] = cri;
          }
          else {
            sqlcri = "<>";
            par[0] = cri;
            //
            // If is a DATE / DATETIME try to complete the data
            if (this.datatype === App.Database.DataTypes.DATETIME || this.datatype === App.Database.DataTypes.DATE) {
              var date0 = App.Utils.parsePartialDate(par[0], 0);
              var date1 = App.Utils.parsePartialDate(par[0], 1);
              if (date0) {
                sqlcri = {expr: "or", p: [{expr: "<", p: [s, {expr: this.toSql(date0)}]}, {expr: ">=", p: [s, {expr: this.toSql(date1)}]}]};
                par[0] = date0;
                par[1] = date1;
              }
            }
          }
        }
        else if (cri.charAt(0) === '=') {
          sqlcri = "=";
          par[0] = cri.substr(1);
        }
        else if (App.Database.classDT(this.datatype) === App.Database.DataTypeClasses.STRING && cri.charAt(cri.length - 1) === '&') {
          /*
           var soundex = this.parent.parent.dataStore.functions["soundex"];
           if (soundex && soundex.expr.length > 0)
           {
           sqlcri = "soundex0 = soundex(?)";
           par[0] = cri.substr(0, cri.length - 1);
           }
           else
           */
          cri = cri.substr(0, cri.length - 1) + "*";
        }
        else {
          // Check interval
          if (sqlcri.length === 0) {
            var p = cri.indexOf(":");
            if (p >= 0) {
              var handled = false;
              //
              // If the field is a DATE / TIME / DATETIME, better control
              if (App.Database.classDT(this.datatype) === App.Database.DataTypeClasses.DATETIME) {
                // Handle it head
                handled = true;
                //
                // Check if there are any partial dates
                var date0 = App.Utils.parsePartialDate(cri.substr(0, p), 0);
                var date1 = App.Utils.parsePartialDate(cri.substr(p + 1), 1);
                var partialDate = (date0 !== null || date1 !== null);
                //
                // If the criteria contains a date part , I search the ':' before the second date
                if (cri.indexOf('/') !== -1 || cri.indexOf('-') !== -1 || partialDate) {
                  // I start with the first ':' I have found and try the first '/' or '-' ...
                  // If I find the ':' as I try the bars, I remember the last found
                  var pp1;
                  for (pp1 = p + 1; pp1 < cri.length; pp1++) {
                    // If, while walking around, I find a ':' I memorize the last found
                    if (cri.charAt(pp1) === ':')
                      p = pp1;
                    else if (cri.charAt(pp1) === '/' || cri.charAt(pp1) === '-')
                      break;
                  }
                  //
                  // If you have not got to the end, p points to the last ':' before the second date
                  if (pp1 <= cri.length) {
                    var par0 = cri.substr(0, p);
                    var par1 = cri.substr(p + 1);
                    if (par0.length > 0 && par1.length > 0) {
                      if (date0 === null && date1 === null) {
                        sqlcri = {p: []};
                        sqlcri.p.push(s);
                        sqlcri.p.push({expr: "between"});
                        sqlcri.p.push({expr: this.toSql(par0)});
                        sqlcri.p.push({expr: "and"});
                        sqlcri.p.push({expr: this.toSql(par1)});
                      }
                      else {
                        sqlcri = {expr: ">=", p: [s]};
                        //
                        if (date0 === null)
                          par[0] = this.stringToValue(par0);
                        else
                          par[0] = date0;
                        sqlcri.p.push({expr: this.toSql(par[0])});
                        sqlcri = {expr: "and", p: [sqlcri]};
                        //
                        if (date1 === null) {
                          par[1] = this.stringToValue(par1);
                          sqlcri.p.push({expr: "<=", p: [s, {expr: this.toSql(par[1])}]});
                        }
                        else {
                          par[1] = date1;
                          sqlcri.p.push({expr: "<", p: [s, {expr: this.toSql(par[1])}]});
                        }
                      }
                    }
                    else if (par0.length > 0) {
                      sqlcri = ">=";
                      if (date0 === null)
                        par[0] = par0;
                      else
                        par[0] = date0;
                    }
                    else if (par1.length > 0) {
                      if (date1 === null) {
                        sqlcri = "<=";
                        par[0] = par1;
                      }
                      else {
                        sqlcri = "<";
                        par[0] = date1;
                      }
                    }
                  }
                }
                else { // The criteria does not contain dates ...
                  var dp = [];
                  //
                  // Find all ':'
                  for (var pp1 = 0; pp1 < cri.length; pp1++)
                    if (cri.charAt(pp1) === ':')
                      dp.push(pp1);
                  //
                  // If the number of ':' is odd and >2, take the ':' 'middle'
                  if (dp.length > 2 && (dp.length % 2) === 1) {
                    // I break the ':' in the middle
                    p = dp[(dp.length - 1) / 2];
                    //
                    var par0 = cri.substr(0, p);
                    var par1 = cri.substr(p + 1);
                    if (par0.length > 0 && par1.length > 0) {
                      sqlcri = {p: []};
                      sqlcri.p.push(s);
                      sqlcri.p.push({expr: "between"});
                      sqlcri.p.push({expr: this.toSql(par0)});
                      sqlcri.p.push({expr: "and"});
                      sqlcri.p.push({expr: this.toSql(par1)});
                    }
                    else if (par0.length > 0) {
                      sqlcri = ">=";
                      par[0] = par0;
                    }
                    else if (par1.length > 0) {
                      sqlcri = "<=";
                      par[0] = par1;
                    }
                  }
                }
              }
              //
              // If I have not done anything yet, I continue normally
              if (!handled) {
                // There is an interval!
                var par0 = cri.substr(0, p);
                var par1 = cri.substr(p + 1);
                if (par0.length > 0 && par1.length > 0) {
                  sqlcri = {p: []};
                  sqlcri.p.push(s);
                  sqlcri.p.push({expr: "between"});
                  sqlcri.p.push({expr: this.toSql(par0)});
                  sqlcri.p.push({expr: "and"});
                  sqlcri.p.push({expr: this.toSql(par1)});
                }
                else if (par0.length > 0) {
                  sqlcri = ">=";
                  par[0] = par0;
                }
                else if (par1.length > 0) {
                  sqlcri = "<=";
                  par[0] = par1;
                }
              }
            }
          }
        }
      }
      //
      // Equality
      if (sqlcri.length === 0) {
        // If is a DATE / DATETIME try to complete the data
        var date0 = App.Utils.parsePartialDate(cri, 0);
        var date1 = App.Utils.parsePartialDate(cri, 1);
        if ((this.datatype === App.Database.DataTypes.DATETIME || this.datatype === App.Database.DataTypes.DATE) && date0) {
          sqlcri = {expr: "and", p: [{expr: ">=", p: [s, {expr: this.toSql(date0)}]}, {expr: "<", p: [s, {expr: this.toSql(date1)}]}]};
          par[0] = date0;
          par[1] = date1;
        }
        else if (useLike && App.Database.classDT(this.datatype) === App.Database.DataTypeClasses.STRING) {
          if (cri.indexOf("*") === -1 && cri.indexOf("%") === -1) {
            if (this.searchOptions.likeType === App.DataMap.likeTypes.startsWith)
              cri = cri + "*";
            else if (this.searchOptions.likeType === App.DataMap.likeTypes.contains)
              cri = "*" + cri + "*";
          }
          //
          // It's a like!
          cri = cri.replace(/\*/g, '%');
          sqlcri = "like";
          par[0] = cri;
        }
        else {
          sqlcri = "=";
          par[0] = cri;
        }
      }
      //
      if (typeof sqlcri === "string") {
        if (sqlcri.length > 0) {
          // Adjust parameters
          var p = [s];
          for (var np = 0; np < par.length; np++) {
            if (par[np]) {
              if (typeof par[np] === "string")
                par[np] = this.stringToValue(par[np]);
              p.push({expr: this.toSql(par[np])});
            }
          }
          //
          sqlcri = {expr: sqlcri, p: p};
        }
        else
          sqlcri = null;
      }
      //
      if (sqlcri) {
        if (cplxCri)
          cplxCri = {expr: "or", p: [cplxCri, sqlcri]};
        else
          cplxCri = sqlcri;
      }
    }
    //
    p2 = p1;
  }
  return {p: [{expr: "("}, cplxCri, {expr: ")"}]};
};


/**
 * Convert a string to native value
 * @param {String} s
 * @param {Object} el element for wich convert the value
 * @param {App.Session} app
 * @returns {Any type}
 */
App.DataColumn.prototype.stringToValue = function (s, el, app)
{
  if (!app && this.parent.parent)
    app = this.parent.parent.app;
  //
  if (!app)
    return s;
  //
  var options = {};
  options.dstDT = this.datatype;
  if (el)
    options.format = App.DataColumn.getElementFormat(el.element);
  //
  return app.locale.convertValue(s, options);
};


/**
 * Convert a native value to string
 * @param {AnyType} v
 * @param {Object} el element for wich convert the value
 * @param {App.Session} app
 * @returns {String}
 */
App.DataColumn.prototype.valueToString = function (v, el, app)
{
  if (!app && this.parent.parent)
    app = this.parent.parent.app;
  //
  if (!app)
    return v;
  //
  var options = {};
  options.srcDT = this.datatype;
  options.dstDT = App.Database.DataTypes.VARCHAR;
  if (el)
    options.format = App.DataColumn.getElementFormat(el.element);
  //
  return app.locale.convertValue(v, options);
};


/**
 * Return format of an element
 * @param {RemElem} el
 * @returns {String}
 */
App.DataColumn.getElementFormat = function (el) {
  switch (el.type) {
    case "date":
      return "YYYY-MM-DD";
    case "datetime-local":
      return "YYYY-MM-DDTHH:mm";
    case "time":
      return "HH:mm";
  }
};


/**
 * Return a copy of this object
 * @returns {DataColumn}
 */
App.DataColumn.prototype.clone = function ()
{
  var c = new App.DataColumn(this.parent);
  var keys = Object.keys(this);
  for (var k in keys) {
    var v = keys[k];
    //
    // Don't copy some objects
    if (["filters", "selectToken", "_searchOptions"].indexOf(v) >= 0)
      continue;
    //
    // Copying children
    if (v === "elements")
      c.elements = this.elements.slice(0);
    else
      c[v] = this[v];
  }
  //
  return c;
};


/**
 * Return the elements linked to this column to update for a row
 * @param {DataRow} row
 * @returns {Array}
 */
App.DataColumn.prototype.getElementsToUpdateForRow = function (row)
{
  var elements = [];
  var ownerDataMap = this.parent.parent;
  //
  // Only list or also view?
  var ar = ownerDataMap.row === row;
  var par = (ownerDataMap.parentElement instanceof App.DataMap && ownerDataMap.parentElement.row === ownerDataMap.parent);
  //
  for (var j = 0; j < this.elements.length; j++) {
    var oel = this.elements[j];
    var ael = oel.element;
    var update = false;
    if (oel.template)
      update = true;
    else if (ar && !oel.groupHeader && !oel.groupFooter) {
      if (!(ownerDataMap.parentElement instanceof App.DataMap))
        update = true;
      else if (par)
        update = true;
      else
        update = oel.templateParent;
    }
    //
    if (update) {
      if (oel.template && row)
        ael = row.map[oel.element.id];
      else if (oel.templateParent && ownerDataMap.parent && ownerDataMap.parent.map)
        ael = ownerDataMap.parent.map[oel.element.id];
      //
      if (ael)
        elements.push({element: ael, prop: oel.prop, style: oel.style, origElem: oel.element});
    }
  }
  //
  return elements;
};


/**
 * Return the elements linked to this column to update for a row
 * @param {DataGroup} group
 * @param {String} type
 * @returns {Array}
 */
App.DataColumn.prototype.getElementsToUpdateForGroup = function (group, type)
{
  var elements = [];
  //
  for (var j = 0; j < this.elements.length; j++) {
    var oel = this.elements[j];
    if (oel.template)
      continue;
    //
    var ael = null;
    if (oel.groupHeader && !oel.gropFooter && type === "header")
      ael = group.map[oel.element.id];
    else if (oel.groupFooter && type === "footer")
      ael = group.map[oel.element.id];
    //
    if (ael)
      elements.push({element: ael, prop: oel.prop, style: oel.style, origElem: oel.element});
  }
  //
  return elements;
};


/**
 * Search my select token in the query
 */
App.DataColumn.prototype.getSelectToken = function ()
{
  if (this.selectToken === undefined) {
    this.selectToken = null;
    //
    // Search my select token of the query
    if (this.parent.parent && this.parent.parent.query instanceof Object) {
      var q = this.parent.parent.query;
      for (var i = 0; i < q.select.length; i++) {
        if ((q.select[i].alias || q.select[i].expr) === this.name) {
          this.selectToken = q.select[i];
          this.selectTokenIndex = i;
          break;
        }
      }
    }
  }
  //
  return this.selectToken;
};


/**
 * Check if column is a field of master table
 */
App.DataColumn.prototype.isMaster = function ()
{
  // Unbound -> no master
  if (this.unbound)
    return false;
  //
  // No from table -> no master
  var q = this.parent.parent.query;
  if (!q.from || q.from.length === 0)
    return false;
  //
  // Linked to unique from table -> master
  if (this.getSelectToken().from === "")
    return true;
  //
  // Master only if linked to first from table
  return this.getSelectToken().from === (q.from[0].alias || q.from[0].expr);
};


/**
 * Check if column si comparable
 */
App.DataColumn.prototype.isComparable = function ()
{
  switch (this.datatype) {
    case App.Database.DataTypes.BLOB:
    case App.Database.DataTypes.CLOB:
    case App.Database.DataTypes.JSON:
    case undefined:
      return false;

    default:
      return true;
  }
};


/**
 * Define the property for this column in the rowClass
 * @param {Function} rowClass
 */
App.DataColumn.prototype.defineClassProperty = function (rowClass)
{
  var dc = this;
  Object.defineProperty(rowClass.prototype, dc.name, {
    get: function () {
      return this.getValueContainer(dc)[dc.name];
    },
    set: function (newValue) {
      this.setValue(dc.name, newValue);
    },
    enumerable: true
  });
};


/**
 * Save to JSON
 */
App.DataColumn.prototype.toJSON = function ()
{
  var json = {};
  json.nm = this.name;
  json.dt = this.datatype;
  if (this.pk)
    json.pk = true;
  if (!this.unbound)
    json.ub = false;
  if (this.default)
    json.def = this.default;
  return json;
};


/**
 * Load from JSON
 * @param{Object} json
 */
App.DataColumn.prototype.loadFromJSON = function (json)
{
  this.name = json.nm;
  this.datatype = json.dt;
  if (json.pk)
    this.pk = true;
  if (json.ub === false)
    this.unbound = false;
  if (json.def)
    this.default = json.def;
};


//export module for node
if (module)
  module.exports = App.DataColumn;
