/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, Infinity */

var App = App || {};

/**
 * @class A DataRow Object
 * @param {DataMap/DataGroup} owner
 * */
App.DataGroup = function (owner)
{
  if (module && !App.View)
    App = require("./requireAll");
  //
  if (owner instanceof App.DataMap)
    this.datamap = owner;
  else if (owner instanceof App.DataGroup)
    this.datamap = owner.datamap;
  //
  this.owner = owner;
  this.map = {};
  this.groups = null;     // child groups
  //
  // Rows in this group
  this.rows = [];
  //
  // Header and footer element
  this.header = null;
  this.footer = null;
  //
  this.column = null;
  this.value = null;
};


/**
 * Cardinality of this group
 */
Object.defineProperty(App.DataGroup.prototype, "cardinality", {
  get: function () {
    return this.rows.length;
  }
});


/**
 * First row of this group
 */
Object.defineProperty(App.DataGroup.prototype, "firstRow", {
  get: function () {
    return this.rows[0];
  }
});


/**
 * Last row of this group
 */
Object.defineProperty(App.DataGroup.prototype, "lastRow", {
  get: function () {
    return this.rows.slice(-1)[0];
  }
});


/**
 * Index of this group
 */
Object.defineProperty(App.DataGroup.prototype, "index", {
  get: function () {
    if (this.oldIndex !== undefined)
      return this.oldIndex;
    else
      return Object.keys(this.owner.groups).indexOf(this.value + " ");
  }
});


/**
 * Sum of all values of a column
 * @param {String} column
 */
App.DataGroup.prototype.sum = function (column)
{
  var sum = 0;
  this.rows.forEach(function (row) {
    sum += row[column];
  });
  return sum;
};


/**
 * Min of all values of a column
 * @param {String} column
 */
App.DataGroup.prototype.min = function (column)
{
  var min = Infinity;
  this.rows.forEach(function (row) {
    min = Math.min(min, row[column]);
  });
  return min;
};


/**
 * Max of all values of a column
 * @param {String} column
 */
App.DataGroup.prototype.max = function (column)
{
  var max = -Infinity;
  this.rows.forEach(function (row) {
    max = Math.max(max, row[column]);
  });
  return max;
};


/**
 * Range of all values of a column
 * @param {String} column
 */
App.DataGroup.prototype.range = function (column)
{
  return this.max(column) - this.min(column);
};


/**
 * Median of all values of a column
 * @param {String} column
 */
App.DataGroup.prototype.median = function (column)
{
  var rows = this.rows;
  var mid = this.cardinality / 2;
  if (mid % 1)
    return rows[mid - 0.5][column];
  else
    return (rows[mid - 1][column] + rows[mid][column]) / 2;
};


/**
 * Average of all values of a column
 * @param {String} column
 */
App.DataGroup.prototype.average = function (column)
{
  return this.sum(column) / this.cardinality;
};


/**
 * Variance of all values of a column
 * @param {String} column
 */
App.DataGroup.prototype.variance = function (column)
{
  var variance = 0;
  var average = this.average(column);
  this.rows.forEach(function (row) {
    variance += Math.pow(row[column] - average, 2);
  });
  return variance;
};


/**
 * Standar deviation of all values of a column
 * @param {String} column
 */
App.DataGroup.prototype.standardDeviation = function (column)
{
  return Math.sqrt(this.variance(column));
};


/**
 * Remove a group
 * @param {Boolean} reusing
 */
App.DataGroup.prototype.remove = function (reusing)
{
  this.removeElement("header");
  this.removeElement("footer");
  //
  if (this.groups) {
    Object.keys(this.groups).forEach(function (k) {
      this.groups[k].remove(reusing);
    }.bind(this));
  }
  //
  if (!reusing)
    delete this.owner.groups[this.value + " "];
};


/**
 * Close the group
 */
App.DataGroup.prototype.close = function ()
{
  this.closeElement("header");
  this.closeElement("footer");
  //
  if (this.groups) {
    Object.keys(this.groups).forEach(function (k) {
      this.groups[k].close();
    }.bind(this));
  }
};


/**
 * Remove a row from group
 * @param {DataRow} row
 * @param {Boolean} reusing
 */
App.DataGroup.prototype.removeRow = function (row, reusing)
{
  // Close group if needed otherwise update header and footer
  var wasFirstRow = (row === this.firstRow);
  this.rows.splice(this.rows.indexOf(row), 1);
  if (this.cardinality === 0 || (wasFirstRow && !this.firstRow.hasElement()))
    this.remove(reusing);
  else if (!reusing) {
    this.updateElement("header");
    this.updateElement("footer");
  }
  //
  // Dispatch to my owner group
  if (this.owner instanceof App.DataGroup)
    this.owner.removeRow(row, reusing);
};


/**
 * Hide a row from group
 * @param {DataRow} row
 */
App.DataGroup.prototype.hideRow = function (row)
{
  if (!this.rows.some(function (r) {
    return r.hasElement();
  })) {
    this.removeElement("header");
    this.removeElement("footer");
  }
  //
  // Dispatch to my owner group
  if (this.owner instanceof App.DataGroup)
    this.owner.hideRow(row);
};


/**
 * Remove header/footer element
 * @param {String} key - header/footer
 */
App.DataGroup.prototype.removeElement = function (key)
{
  if (this[key]) {
    this[key].parent.removeChild(this[key]);
    delete this[key];
  }
};


/**
 * Close header/footer element
 * @param {String} key - header/footer
 */
App.DataGroup.prototype.closeElement = function (key)
{
  if (this[key])
    this[key].close();
};


/**
 * Update header/footer elemnt
 * @param {String} key - header/footer
 */
App.DataGroup.prototype.updateElement = function (key)
{
  var schema = this.datamap.schema;
  if (!schema[key])
    return;
  //
  // Create new header/footer template clone if missing
  if (!this[key]) {
    this[key] = schema[key].clone(this.map);
    //
    var ownerEl = null;
    var sibID = schema[key + "SiblingID"];
    if (key === "header") {
      if (this.owner === this.datamap || (schema.template && !schema.template.isChildOf(schema.header))) {
        ownerEl = schema.listBox || schema[key].parent;
        if (this.firstRow.hasElement())
          sibID = this.firstRow.element.id;
        else if (this.level > 0 && this.owner.footer)
          sibID = this.owner.footer.id;
      }
      else {
        if (schema.template)
          ownerEl = this.owner.map[schema.template.parent.id];
        else
          ownerEl = this.owner.header;
        if (schema.footer && schema.footer.isChildOf(schema.header))
          sibID = (schema.footer.id in this.owner.map ? this.owner.map[schema.footer.id].id : null);
      }
    }
    else {
      if (!schema.header || !schema.footer.isChildOf(schema.header)) {
        ownerEl = schema.listBox || schema[key].parent;
        var lastEl = null;
        if (!this.groups || !schema.footer)
          lastEl = this.lastRow.element;
        else if (this.groups)
          lastEl = this.lastRow.group.getGroup(this.level + 1).footer;
        if (lastEl) {
          var parElements = lastEl.parent.elements;
          var idx = parElements.indexOf(lastEl);
          if (idx < parElements.length - 1)
            sibID = parElements[idx + 1].id;
        }
      }
      else
        ownerEl = this.map[schema.footer.parent.id] || this.header;
    }
    //
    // Append element
    if (sibID)
      ownerEl.insertBefore(this[key], sibID);
    else
      ownerEl.appendChild(this[key]);
  }
  //
  // Set group to element
  if (this[key].group !== this)
    this[key].setDataGroup(this);
  //
  // Send values to elements
  var columns = this.datamap.schema.columns;
  for (var c in columns) {
    var dc = columns[c];
    var elements = dc.getElementsToUpdateForGroup(this, key);
    for (var j = 0; j < elements.length; j++) {
      var oel = elements[j];
      var ael = oel.element;
      //
      // Convert date to string
      var v = this.firstRow[c];
      if (v === undefined)
        v = null;
      else if (v instanceof Date)
        v = dc.valueToString(v, oel);
      //
      if (oel.style)
        ael.style[oel.prop] = v;
      else
        ael[oel.prop] = v;
    }
  }
  //
  // Notify on[Header/Footer]Composition event
  try {
    this.datamap["on" + key[0].toUpperCase() + key.substring(1) + "Composition"](this, this[key]);
  }
  catch (e) {
  }
};


/**
 * Get a group of a level
 * @param {Integer} level
 */
App.DataGroup.prototype.getGroup = function (level)
{
  if (this.level === level)
    return this;
  else if (this.owner instanceof App.DataGroup)
    return this.owner.getGroup(level);
};


/**
 * Compare order with another group
 * @param {DataGroup} group
 */
App.DataGroup.prototype.compare = function (group)
{
  if (this.owner !== this.datamap) {
    if (this.owner.index < group.owner.index)
      return -1;
    else if (this.owner.index > group.owner.index)
      return 1;
  }
  //
  if (this.index < group.index)
    return -1;
  else if (this.index > group.index)
    return 1;
  else
    return 0;
};


//export module for node
if (module)
  module.exports = App.DataGroup;
