/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module */

var App = App || {};


/**
 * @class
 * @param {View} view
 * @param {Object} defprop
 * @param {Object} parent
 * */
App.DataMap = function (view, defprop, parent)
{
  if (module && !App.View && view !== true)
    App = require("./requireAll");
  //
  if (view && view !== true) {
    this.view = view;
    this.parentElement = parent; // DataMap, View or RemElem
    this.app = view.app;
  }
  //
  // Objects for window's mode
  this.useWindow = false;
  this._pageSize = 0;
  this._pageLimit = 0;
  this.uiCache = [];      // Cache of UI elements
  this.uiRecycleBin = []; // Not used UI elements
  //
  // Options of search
  this._searchOptions = {
    likeType: App.DataMap.likeTypes.startsWith
  };
  //
  this.elements = [];       // Child elements by design-time (not only RemElem)
  this.outputElements = []; // Elements attached at run-time
  //
  this.rows = [];   // DataRows as well as loaded by the query
  this.uiRows = []; // Visible dataRows (sorted, non hidden and not deleted)
  this.schema = null;
  this._position = null; // Position of current row (don't set directly: use the property)
  this.oldRow = null;    // Last current row
  //
  this.query = null;    // Query data load
  this.clientRefs = null; // References in the query to interface objects
  this.needReload = false; // True if dataMap is changed when listBox was invisible
  this.autoLoad = false; // True if dataMap has to load automatically the data
  this.recursive = false;
  this.parent = null;    // Parent row
  this.maxRows = 0;
  this.loaded = false;
  this.rowsToUpdate = [];
//
  this.dataPagingMode = App.DataMap.dataPagingModes.none;
  this.dataPageSize = 100;
  this.dataPagingPreCount = false;
  this.loadingPages = [];
  //
  // Load default properties
  if (defprop) {
    for (var x in defprop) {
      if (x === "class")
        continue;
      if (x === "schema")
        this.schema = new App.DataSchema(this, defprop);
      else
        this[x] = defprop[x];
    }
    //
    // Save id from IDE if set
    if (this.id)
      this.pid = this.id;
    //
    if (this.app) {
      // Put the element in the map
      this.id = this.app.generateCounter();
      this.app.eleMap[this.id] = this;
    }
  }
};

Object.defineProperties(App.DataMap.prototype, {
  // Number of visible rows
  count: {
    get: function () {
      return this.uiRows.length;
    }
  },
  // Number of rows
  length: {
    get: function () {
      return this.rows.length;
    }
  },
  // Get/set position of current row
  position: {
    get: function () {
      return this._position;
    },
    set: function (newValue) {
      this.setPosition(newValue);
    }
  },
  // Get/set position of useWindow property
  useWindow: {
    get: function () {
      return this._useWindow && this.template;
    },
    set: function (newValue) {
      this._useWindow = newValue;
    }
  },
  // Get/set list template, stored into schema
  template: {
    get: function () {
      return this.schema.template;
    },
    set: function (newValue) {
      this.schema.setTemplate(newValue);
    }
  },
  // Get/set group header template, stored into schema
  headerTemplate: {
    get: function () {
      return this.schema.header;
    },
    set: function (newValue) {
      this.schema.setHeaderTemplate(newValue);
    }
  },
  // Get/set group footer template, stored into schema
  footerTemplate: {
    get: function () {
      return this.schema.footer;
    },
    set: function (newValue) {
      this.schema.setFooterTemplate(newValue);
    }
  },
  // Reference to current row
  row: {
    get: function () {
      return (this.uiRows[this.position] || null);
    }
  },
  // Get/set search options
  searchOptions: {
    get: function () {
      return this._searchOptions;
    },
    set: function (newValue) {
      this._searchOptions = newValue;
    }
  },
  // Get/set of modified status
  modified: {
    get: function () {
      return this.isModified();
    },
    //
    set: function (newValue) {
      if (!newValue)
        this.undoChanges();
    }
  },
  // Get/set of page size
  pageSize: {
    get: function () {
      return this._pageSize;
    },
    //
    set: function (newValue) {
      this._pageSize = (newValue <= 0 ? 0 : newValue);
      if (this.currentPageSize < this._pageSize)
        this.currentPageSize = this._pageSize;
    }
  },
  // Get/set of page limit
  currentPageSize: {
    get: function () {
      return this._pageLimit;
    },
    //
    set: function (newValue) {
      var old = this._pageLimit;
      this._pageLimit = (newValue <= 0 ? 0 : newValue);
      //
      // Create elements for the new rows if not done yet
      var m = Math.min(this._pageLimit, this.uiRows.length);
      var pos = old;
      for (var i = old; i < m; i++) {
        this.getRow(i, function () {
          var row = this.uiRows[pos++];
          if (row)
            row.updateElement();
        }.bind(this));
      }
      //
      // Reset more rows indicator
      if (this.schema && this.schema.listBox) {
        var mr = this.schema.listBox.moreRows;
        this.schema.listBox.moreRows = 0;
        if (this._pageLimit < this.uiRows.length || (this.dataPagingMode && !this.dataPagingPreCount))
          this.schema.listBox.moreRows = ((mr || 0) % 100000) + 1;
      }
    }
  },
  // Get/set of upsert
  upsert: {
    get: function () {
      return (this._upsert === true);
    },
    //
    set: function (newValue) {
      if (newValue) {
        this._upsert = true;
        //
        // Set all rows as inserted
        for (var i = 0; i < this.rows.length; i++)
          this.rows[i].inserted = true;
      }
    }
  },
  // Get/set query
  query: {
    get: function () {
      return this._query;
    },
    //
    set: function (newValue) {
      this.addQuery(newValue);
    }
  },
  // DataStore instance
  dataStore: {
    get: function () {
      // If datamap has own dataStore return it
      if (this._dataStore)
        return this._dataStore;
      //
      // Return default instance of dataModel
      if (this.getDataModel())
        return this.app.getDefaultInstance(this.getDataModel());
    },
    //
    set: function (newValue) {
      if (this.goodDataStore(newValue))
        this._dataStore = newValue;
    }
  }
});


App.DataMap.likeTypes = {
  noLike: 0,
  equals: 1,
  startsWith: 2,
  contains: 3
};


App.DataMap.dataPagingModes = {
  none: 0,
  offset: 1,
  keyset: 2,
  cursor: 3
};


App.DataMap.errorSources = {
  load: "load",
  save: "save",
  validation: "validation"
};


/**
 * Return the name of datamodel
 * */
App.DataMap.prototype.getDataModel = function ()
{
  if (this.query)
    return this.query.dm;
};


/**
 * Check if dataStore is good for me
 * @param {App.Database} ds
 * */
App.DataMap.prototype.goodDataStore = function (ds)
{
  var dm = this.getDataModel();
  return !ds || !dm || (ds.schema && ds.schema.name === dm);
};


/**
 * Tells something about this object to the DTT module
 * @param {Integer} childLevel - level of child to log
 */
App.DataMap.prototype.getDebugInfo = function (childLevel)
{
  var info = {};
  info._class = "DataMap";
  if (this.name)
    info._name = this.name;
  info.length = this.length;
  info.count = this.count;
  info.loaded = this.loaded;
  info.position = this.position;
  info.orderBy = this.orderBy;
  info.groupBy = this.groupBy;
  if (childLevel > 0)
    info.rows = App.DTT.convertValue(this.rows, undefined, undefined, childLevel - 1);
  if (childLevel > 0)
    info.uiRows = App.DTT.convertValue(this.uiRows, undefined, undefined, childLevel - 1);
  return info;
};


/**
 * Change a property by the IDE
 * @param {Object} pv property to change
 */
App.DataMap.prototype.chgProp = function (pv)
{
  var pname = pv.name;
  if (pname.substring(0, 4) === "cst_")
    pname = pv.name.substr(4);
  this[pname] = pv.value;
};


/**
 * Closes the datamap
 */
App.DataMap.prototype.close = function ()
{
  // Send close event to all my rows
  for (var i in this.rows)
    this.rows[i].close();
  //
  // Send close event to all my sub elements
  if (this.elements) {
    for (var i = 0; i < this.elements.length; i++)
      this.elements[i].close();
  }
  //
  // Send close event to all my groups
  if (this.groups) {
    Object.keys(this.groups).forEach(function (k) {
      this.groups[k].close();
    }.bind(this));
  }
  //
  // Detach client references observer
  this.detachClientRefsObserver();
  //
  // Remove me from eleMap
  delete this.app.eleMap[this.id];
  //
  // Delete any observer on me
  if (App.deleteObservers)
    App.deleteObservers(this);
};


/**
 * Changes the total rows in the Map
 * @param {int} c
 */
App.DataMap.prototype.setRowCount = function (c)
{
};


/**
 * Add a row to the datamap
 * @param {Object/DataRow} row to add
 * @param {Object} options
 */
App.DataMap.prototype.add = function (row, options)
{
  // If row is a simple object create a DataRow with it
  if (!(row instanceof this.getRowClass()))
    row = this.createRow(row);
  //
  // Now row is mine
  row.datamap = this;
  //
  // Init default value of unbound columns
  row.initDefaults();
  //
  // Add the row
  if (options && options.index >= 0)
    this.rows[options.index] = row;
  else
    this.rows.push(row);
  //
  // Update UI elements attached to this row
  if (row.isVisible())
    this.updateElementsVisibleRow(row, options);
  //
  return row;
};


/**
 * Remove a row from the datamap
 * @param {DataRow} row to delete
 */
App.DataMap.prototype.remove = function (row)
{
  // If not specified delete current row
  if (!row)
    row = this.row;
  //
  // Remove the row
  this.rows.splice(row.index, 1);
  //
  // Update UI elements attached to this row
  this.updateElementsInvisibleRow(row);
};


/**
 * Update elements of a visible row
 * @param {DataRow} row
 * @param {Object} options
 */
App.DataMap.prototype.updateElementsVisibleRow = function (row, options)
{
  // Add the row to that visible
  if (options && (options.index >= 0 || options.position >= 0)) {
    var pos = options.position || options.index || 0;
    if (pos < this.uiRows.length) {
      this.uiRows.splice(pos, 0, row);
      if (!this.useWindow)
        row.spliced = true;
    }
    else
      this.uiRows[pos] = row;
  }
  else
    this.uiRows.push(row);
  //
  if (this.useWindow) {
    // If I'm loading data don't update rowCount for performance reasons
    // I update it after the loading
    if (!this.loadingData)
      this.schema.listBox.rowCount = this.uiRows.length;
  }
  else {
    // Calculate row groups
    this.calcRowGroups(row);
    //
    // Reuse element if need
    var inPagination = row.inPagination();
    var e = inPagination ? this.uiRecycleBin.shift() : null;
    if (e)
      row.attachElement(e);
    else
      row.updateElement();
    //
    // When a row inside the pagination becomes visible
    // remove the element of the last row inside the pagination
    if (!this.loadingData && this.pageSize && inPagination) {
      var lr = this.uiRows[this.currentPageSize];
      if (lr) {
        if (!lr.hasElement()) {
          if (lr.ueTimer)
            this.clearUETimeout(lr);
        }
        else {
          lr.removeElement();
          //
          // Inform group that row disappear
          if (lr.group)
            lr.group.hideRow(lr);
        }
      }
    }
  }
  //
  if (!this.loadingData)
    this.addElementsData(row);
};


/**
 * Update elements of an invisible row
 * @param {DataRow} row
 */
App.DataMap.prototype.updateElementsInvisibleRow = function (row)
{
  var pos = row.position;
  //
  // If the rows is not inside the visible rows array, do nothing
  if (pos === -1)
    return;
  //
  // Reset updateElement
  if (row.ueTimer)
    this.clearUETimeout(row);
  //
  if (!this.loadingData)
    this.removeElementsData(row);
  //
  // Remove the row from that visible
  this.uiRows.splice(pos, 1);
  //
  // Inform group that row disappear
  if (row.group && row.hasElement())
    row.group.removeRow(row);
  //
  // Check current position
  if (!this.loadingData)
    this.setPosition();
  //
  if (this.useWindow) {
    // If I'm loading data don't update rowCount for performance reasons
    // I update it after the loading
    if (!this.loadingData)
      this.schema.listBox.rowCount = this.uiRows.length;
  }
  else {
    row.removeElement();
    //
    if (!this.loadingData && this.pageSize && pos < this.currentPageSize) {
      this.updatePageLimit();
      //
      // With paging by hiding a line you may need to create the element of row
      // that now falls into the visible part of the pagination
      var lr = this.uiRows[this.currentPageSize - 1];
      if (lr && !lr.hasElement())
        lr.updateElement();
    }
  }
  //
  this.fireOnDataChange();
};


/**
 * Fire onDataChange event
 */
App.DataMap.prototype.fireOnDataChange = function ()
{
  if (this.dataChangeTimer || this.onDataChange === App.DataMap.prototype.onDataChange)
    return;
  //
  // I plan notification of onChange event (only if implemented)
  this.dataChangeTimer = setImmediate(function () {
    delete this.dataChangeTimer;
    try {
      this.onDataChange();
    }
    catch (e) {
    }
  }.bind(this));
};


/**
 * Event fired when anything is changed
 */
App.DataMap.prototype.onDataChange = function ()
{
};


/**
 * Add a new row to the datamap
 * @param {Object} row to insert
 * @param {Object} options
 */
App.DataMap.prototype.insert = function (row, options)
{
  // If row is a simple object create a DataRow with it
  if (!(row instanceof this.getRowClass())) {
    row = this.createRow(row);
    row.initDefaults(true);
  }
  //
  // Add new row
  row = this.add(row, options);
  //
  // Set inserted status
  row.inserted = true;
  //
  return row;
};


/**
 * Set deleted a row of the datamap
 * @param {DataRow} row
 */
App.DataMap.prototype.delete = function (row)
{
  // If not specified set deleted the current row
  if (!row)
    row = this.row;
  //
  // Set deleted status
  row.deleted = true;
};


/**
 * Changes current position of the Map
 * @param {int} np - new position
 */
App.DataMap.prototype.setPosition = function (np)
{
  // Checking input value
  if (np === undefined)
    np = this._position;
  else if (np !== null) {
    np = parseInt(np);
    if (isNaN(np))
      np = null;
  }
  if (np !== null) {
    if (this.uiRows.length > 0) {
      if (np < 0)
        np = 0;
      if (np > this.uiRows.length - 1)
        np = this.uiRows.length - 1;
    }
    else
      np = null;
  }
  //
  // If the position has not changed I still have to check if it physically changed the actual row
  if (this._position === np) {
    this.checkChangeRow();
    return;
  }
  //
  // Firing event onPositionChange
  if (this.onPositionChange(this._position, np) === false)
    return;
  //
  // Changing position
  this._position = np;
  //
  // Updating user interface
  if (this.template) {
    this.schema.listBox.position = np;
    //
    // Update currentPageSize if needed
    if (!this.useWindow && this.pageSize > 0 && np >= this.currentPageSize)
      this.currentPageSize = (Math.floor(np / this.pageSize) + 1) * this.pageSize;
    //
    if (this.oldRow)
      this.oldRow.updateElement();
  }
  //
  if (this.row) {
    // Update elements of current row and of her childrens
    this.row.updateElement();
    for (var i in this.row.subDataMaps) {
      var dm = this.row.subDataMaps[i];
      for (var j in dm.uiRows)
        dm.uiRows[j].updateElement();
    }
  }
  else {
    // Reset all elements because the row is empty
    for (var c in this.schema.columns) {
      var elements = this.schema.columns[c].getElementsToUpdateForRow(null);
      for (var i in elements) {
        var oel = elements[i];
        oel.element[oel.prop] = null;
      }
    }
  }
  //
  // Check if it physically changed the actual row
  this.checkChangeRow();
};


/**
 * Check if actual row is physical changed
 */
App.DataMap.prototype.checkChangeRow = function ()
{
  if (this.oldRow === this.row)
    return;
  //
  // Notify to observers that is changed the actual row
  if (this.observed)
    Object.getNotifier(this).notify({type: "update", name: "row", oldValue: this.oldRow});
  //
  // Store acual row for next check
  this.oldRow = this.row;
};


/**
 * Notified during row update
 * @param {DataRow} row
 * @param {RemElem} template
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.onRowComposition = function (row, template, callback)
{
  callback();
};


/**
 * Notified during group header update
 * @param {DataGroup} group
 * @param {RemElem} template
 */
App.DataMap.prototype.onHeaderComposition = function (group, template)
{
};


/**
 * Notified during group footer update
 * @param {DataGroup} group
 * @param {RemElem} template
 */
App.DataMap.prototype.onFooterComposition = function (group, template)
{
};


/**
 * Notified before position change
 * @param {int} actualPosition
 * @param {int} newPosition
 * @return {bool} return false to cancel position change
 */
App.DataMap.prototype.onPositionChange = function (actualPosition, newPosition)
{
};


/**
 * Notified before load data
 * @param {Object} query
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.beforeLoad = function (query, callback)
{
  if (callback)
    callback();
};


/**
 * Notified after load data
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.afterLoad = function (callback)
{
  if (callback)
    callback();
};


/**
 * Notified during listbox scroll
 * @param {int} start
 * @param {int} end
 */
App.DataMap.prototype.onMissingRecord = function (start, end)
{
  // Copy parameter for a bug of V8 Engine
  // It's result undefined into callback of async functions
  var newStart = start;
  var newEnd = end;
  this.lastMissingRecordRequest = start + "|" + end;
  //
  // Estending window if another client has different size
  if (newEnd - newStart > this.maxWindowSize || this.maxWindowSize === undefined)
    this.maxWindowSize = newEnd - newStart;
  else {
    var delta = this.maxWindowSize - (newEnd - newStart);
    if (delta % 2 === 1)
      delta++;
    newStart -= delta / 2;
    newEnd += delta / 2;
    if (newStart < 0)
      newStart = 0;
    if (newEnd > this.uiRows.length)
      newEnd = this.uiRows.length;
  }
  //
  // Pagesize was specified, extend window
  if (newEnd - newStart < this.pageSize) {
    var ofs = this.pageSize - (newEnd - newStart);
    newEnd += Math.ceil(ofs / 2);
    newStart -= Math.floor(ofs / 2);
    if (newStart < 0) {
      newEnd -= newStart;
      newStart = 0;
    }
    if (newEnd > this.uiRows.length)
      newEnd = this.uiRows.length;
  }
  //
  var afterLoadPages = function () {
    // Skip the update if UI came another request in the meantime
    if (this.lastMissingRecordRequest !== start + "|" + end)
      return;
    delete this.lastMissingRecordRequest;
    //
    var newCache = [];
    //
    // Let's create element for the new cache, reusing the old ones
    for (var i = newStart; i < newEnd; i++) {
      if (this.uiStart !== undefined && i >= this.uiStart && i < this.uiEnd) {
        newCache.push(this.uiCache[i - this.uiStart]);
        this.uiCache[i - this.uiStart] = null;
      }
      else
        newCache.push(null);
    }
    //
    // Let's recycle not used elements
    for (var i = 0; i < this.uiCache.length; i++) {
      if (this.uiCache[i]) {
        this.uiCache[i].element.row.detachElement();
        this.uiRecycleBin.push(this.uiCache[i]);
      }
    }
    //
    // Registering new cache
    this.uiCache = newCache;
    this.uiStart = newStart;
    this.uiEnd = newEnd;
    //
    // recycling objects or creating new ones
    for (var i = 0; i < newCache.length; i++) {
      if (!newCache[i]) {
        if (this.uiRecycleBin.length > 0)
          newCache[i] = this.uiRecycleBin.pop();
        else {
          var map = {};
          var element = this.schema.template.clone(map);
          newCache[i] = {element: element, map: map, append: true};
        }
      }
      //
      // Attaching objects to row and updating them
      this.uiRows[newStart + i].attachElement(newCache[i], newStart + i);
    }
  }.bind(this);
  //
  // Load missing pages; without autoload skip request with end = 0
  if (this.dataPagingMode && newEnd > 0 && !this.autoLoad) {
    var firstPage = Math.floor(newStart / this.dataPageSize);
    var lastPage = Math.floor(newEnd / this.dataPageSize);
    this.loadPages(firstPage, lastPage, function (result, error) {
      if (!error)
        afterLoadPages();
    });
  }
  else
    afterLoadPages();
};


/**
 * Append a row element
 * @param {DataRow} row
 */
App.DataMap.prototype.appendRowElement = function (row)
{
  // Find siblingID into parent row map
  var sibID = this.schema.templateSiblingID;
  if (this.schema.templateSiblingID && this.parent && this.parent.map && this.parent.map[this.schema.templateSiblingID])
    sibID = this.parent.map[this.schema.templateSiblingID].id;
  //
  var ownerEl = this.schema.listBox;
  if (row.group) {
    // If data are grouped and group are nested append to parent's header
    if (row.group.header && this.schema.template.isChildOf(this.schema.header)) {
      ownerEl = row.group.map[this.schema.template.parent.id];
      sibID = (sibID in row.group.map ? row.group.map[sibID].id : null);
    }
    else {
      var lastEl = null;
      if (row === row.group.firstRow) {
        if (row.group.header)
          lastEl = row.group.header;
        else if (row.group.footer)
          sibID = row.group.footer.id;
        else {
          var prevRow = this.uiRows[row.position - 1];
          if (prevRow) {
            for (var l = row.group.level; l >= 0; l--) {
              var pg = prevRow.group.getGroup(l - 1);
              var g = row.group.getGroup(l - 1);
              if (pg === g) {
                lastEl = prevRow.group.getGroup(l).footer;
                break;
              }
            }
          }
        }
      }
      else {
        if (row.group.footer)
          sibID = row.group.footer.id;
        else {
          var prevRow = this.uiRows[row.position - 1];
          if (prevRow)
            lastEl = prevRow.element;
        }
      }
      if (lastEl) {
        var parElements = lastEl.parent.elements;
        var idx = parElements.indexOf(lastEl);
        if (idx < parElements.length - 1)
          sibID = parElements[idx + 1].id;
      }
    }
  }
  //
  // The row was not inserted at the bottom;
  // I have to find the first element of the first row of which position myself
  if (row.spliced) {
    delete row.spliced;
    var rows = (row.group ? row.group.rows : this.uiRows);
    for (var i = rows.indexOf(row) + 1; i < rows.length; i++) {
      var r = rows[i];
      if (r && r.hasElement()) {
        sibID = r.element.id;
        break;
      }
    }
  }
  //
  if (sibID)
    ownerEl.insertBefore(row.element, sibID);
  else
    ownerEl.appendChild(row.element);
};


/**
 * Notified when user press a key
 * @param {String} cmd - the command to execute
 * @param {int} columns - visible columns
 * @param {int} rows - visible rows
 * @param {int} pos - position to go
 */
App.DataMap.prototype.onNavigate = function (cmd, columns, rows, pos)
{
  switch (cmd) {
    case "prev":
      this.position--;
      break;
    case "next":
      this.position++;
      break;
    case "pageup":
      this.position -= columns * rows;
      break;
    case "pagedown":
      this.position += columns * rows;
      break;
    case "first":
      this.position = 0;
      break;
    case "last":
      this.position = this.uiRows.length - 1;
      break;
    case "up":
      this.position = this.position - columns;
      break;
    case "down":
      this.position = this.position + columns;
      break;
    case "goToPos":
      this.position = pos;
      break;
  }
};


/**
 * Notified after value change
 * @param {DataRow} row
 * @param {String} columnName
 * @param {Object} oldValue
 * @param {Object} newValue
 */
App.DataMap.prototype.onValueChange = function (row, columnName, oldValue, newValue)
{
};


/**
 * Notified before save data of datamap
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.beforeSave = function (callback)
{
  callback();
};


/**
 * Notified after save data of datamap
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.afterSave = function (callback)
{
  callback();
};


/**
 * Notified before save an inserted row
 * @param {DataRow} row
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.beforeInsert = function (row, callback)
{
  callback();
};


/**
 * Notified after save an inserted a row
 * @param {DataRow} row
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.afterInsert = function (row, callback)
{
  callback();
};


/**
 * Notified before save an updated row
 * @param {DataRow} row
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.beforeUpdate = function (row, callback)
{
  callback();
};


/**
 * Notified after save an updated a row
 * @param {DataRow} row
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.afterUpdate = function (row, callback)
{
  callback();
};


/**
 * Notified before save a deleted row
 * @param {DataRow} row
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.beforeDelete = function (row, callback)
{
  callback();
};


/**
 * Notified after save a deleted a row
 * @param {DataRow} row
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.afterDelete = function (row, callback)
{
  callback();
};


/**
 * Notified when occurs an error
 * @param {String} errorMessage
 * @param {Object} info
 */
App.DataMap.prototype.onError = function (errorMessage, info)
{
};


/**
 * Add a query to this datamap
 * @param {Object} query
 */
App.DataMap.prototype.addQuery = function (query)
{
  this._query = query;
  if (query && this.dataStore) {
    this.needReload = this.autoLoad;
    //
    // Load data only for dataMap with autoLoad
    // DataMap of higher level do it when are cloned
    if (this.autoLoad && !(this.templateDataMap && this.parentElement instanceof App.DataMap)) {
      setImmediate(function () {
        this.refreshData();
      }.bind(this));
    }
  }
};


/**
 * Add a schema to this datamap
 * @param {Object} schema
 */
App.DataMap.prototype.addSchema = function (schema)
{
  if (schema instanceof App.DataSchema) {
    this.schema = schema;
    this.schema.parent = this;
  }
  else
    this.schema = new App.DataSchema(this, schema);
  //
  // Adjust sort criterias by column index
  this.adjustSortByIndex(this._orderBy);
  this.adjustSortByIndex(this._groupBy);
};


/**
 * Load data of the query from db
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.load = function (options, callback)
{
  // TODO: remove when all projects have libraries updated (with DataMap.load.options parameter)
  if (typeof options === "function") {
    callback = options;
    options = undefined;
  }
  //
  options = Object.assign({}, options);
  //
  // Set supplied dataStore to use for load
  var oldDataStore = this._dataStore;
  if (options.dataStore) {
    this.dataStore = options.dataStore;
    delete options.dataStore;
  }
  //
  var done = function (result, error) {
    // Restore old dataStore
    this.dataStore = oldDataStore;
    //
    // Notify onError event
    if (error)
      this.onError(error, {source: App.DataMap.errorSources.load});
    //
    if (callback)
      callback(result, error);
  }.bind(this);
  //
  // If is already loaded do nothing
  if (this.loaded && !options.reload && !options.page)
    return done();
  //
  if (options.maxRows >= 0)
    this.maxRows = options.maxRows;
  if (options.orderBy)
    this.orderBy = options.orderBy;
  if (options.groupBy)
    this.groupBy = options.groupBy;
  //
  this.attachClientRefsObserver();
  //
  this.needReload = false;
  delete this.elapsedTimes;
  //
  // Clone the query
  var q = (this.query ? JSON.parse(JSON.stringify(this.query)) : null);
  //
  // Notify beforeLoad event
  this.fireBeforeLoad(q, options, function (skip, error) {
    if (error)
      return done(null, error);
    //
    if (skip) {
      // Notify afterLoad event
      options.skip = true;
      return this.fireAfterLoad(options, function (result, error) {
        done(this, error);
      }.bind(this));
    }
    //
    if (options.page)
      this.loadingPages.push(options.page);
    //
    options.clientSort = (this.sortBy.length > 0);
    if (q instanceof Object) {
      // Apply QBE filters
      this.applyFilters(q);
      //
      // Add sort clauses
      if (options.clientSort) {
        var u = q;
        while (u && u.union)
          u = u.union;
        //
        // If sorting involves at least one unbound column
        // I keep query's orderBy and sort in memory
        var querySort = this.getSortConditions();
        if (querySort.length > 0) {
          options.clientSort = false;
          u.order = querySort;
        }
      }
      //
      // Set top clause
      if (this.maxRows && !q.top)
        q.top = {expr: this.maxRows};
    }
    //
    // Execute the query
    this.execQuery(q, options, function (result, error) {
      this.loadingData = !options.page;
      //
      // Restart simulated observers (if any)
      if (App.markObservers)
        App.markObservers();
      //
      // Get Row Count
      var nr = 0;
      if (!error) {
        nr = this.getResultCount(result);
        if (this.maxRows && nr > this.maxRows)
          nr = this.maxRows;
      }
      //
      if (!options.page) {
        // Detach elements to save them
        this.reuseElements(nr);
        this.clear();
      }
      //
      var afterLoadCallback = function (result, error) {
        // Send data to elements, but not when I searching
        // because I do it to end of search
        if (!this.searching && !options.page)
          this.initElement();
        //
        if (options.page)
          this.loadingPages.splice(this.loadingPages.indexOf(options.page), 1);
        else
          this.loadingData = false;
        //
        // Without data notify onDataChange event
        if (this.count === 0)
          this.fireOnDataChange();
        //
        // If someone is waiting for a callback, notify him
        done(this, error);
      }.bind(this);
      //
      if (error) {
        // Reset position
        if (!options.page)
          this.position = null;
        //
        afterLoadCallback(null, error);
      }
      else {
        // Check if schema is present for query by string
        this.checkSchema(result);
        //
        // Insert rows
        for (var i = 0; i < nr; i++) {
          var opts = {};
          if (options.page)
            opts.index = (options.page * this.dataPageSize) + i;
          this.add(this.getResultRow(result, i), opts);
        }
        //
        // Set rowCount calculed for dataPagingPreCount
        if (this._calcRowCount) {
          this.rows.length = this._calcRowCount;
          this.uiRows.length = this._calcRowCount;
          delete this._calcRowCount;
        }
        //
        // Update listBox rowCount
        if (this.useWindow) {
          this.schema.listBox.rowCount = this.uiRows.length;
          Object.getNotifier(this.schema.listBox).notify({type: "update", name: "rowCount", oldValue: this.schema.listBox.rowCount});
        }
        else {
          this.removeUnusedElements();
          this.updatePageLimit();
        }
        //
        // Move to first row
        this.position = this.position || 0;
        //
        this.elapsedTimes = result.times;
        this.loaded = true;
        //
        // Notify afterLoad event
        this.fireAfterLoad(options, afterLoadCallback);
      }
    }.bind(this));
  }.bind(this));
};


/**
 * Detach elements of rows for reuse them
 * @param {Integer} nr - number of element to reuse
 */
App.DataMap.prototype.reuseElements = function (nr)
{
  if (this.useWindow)
    return;
  //
  if (this.groupBy && nr) {
    // Keep actual index of groups
    if (this.groups) {
      Object.keys(this.groups).forEach(function (k, index) {
        this.groups[k].oldIndex = index;
      }.bind(this));
    }
    //
    delete this.groups;
  }
  else if (this.groups)
    this.removeGroups();
  //
  if (this.pageSize)
    this.currentPageSize = this.pageSize;
  //
  for (var i = 0; i < this.uiRows.length; i++) {
    if (this.uiRows[i].hasElement())
      this.uiRecycleBin.push(this.uiRows[i].detachElement(true));
  }
};


/*
 * Remove detached elements unused
 */
App.DataMap.prototype.removeUnusedElements = function ()
{
  if (this.useWindow)
    return;
  //
  while (this.uiRecycleBin.length > 0) {
    var e = this.uiRecycleBin.shift();
    if (e.group)
      e.group.removeRow(e.row, true);
    e.element.parent.removeChild(e.element);
  }
};


/**
 * Remove all groups
 */
App.DataMap.prototype.removeGroups = function ()
{
  if (!this.groups)
    return;
  //
  Object.keys(this.groups).forEach(function (k) {
    var g = this.groups[k];
    g.remove();
    g.rows.forEach(function (r) {
      r.removeElement();
    });
  }.bind(this));
  //
  delete this.groups;
};


/**
 * Fire beforeLoad event
 * @param {Object} q
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.fireBeforeLoad = function (q, options, callback)
{
  try {
    this.beforeLoad(q, callback);
  }
  catch (e) {
    callback(null, e);
  }
};


/**
 * Fire afterLoad event
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.fireAfterLoad = function (options, callback)
{
  // Sort if required
  if (options.clientSort)
    this.sort();
  //
  // Notify afterLoad event
  try {
    this.afterLoad(callback);
  }
  catch (e) {
    callback(null, e);
  }
};


/**
 * Returns the number of rows of a resultSet
 * @param {Object} rs
 */
App.DataMap.prototype.getResultCount = function (rs)
{
  return this.dataStore.getRowCount(rs);
};


/**
 * Returns the i-th row of a resultSet
 * @param {Object} rs - recordset
 * @param {int} pos - position
 */
App.DataMap.prototype.getResultRow = function (rs, pos)
{
  return this.dataStore.getRow(rs, pos, this.schema);
};


/**
 * Reload datamap from datamodel
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.DataMap.prototype.reload = function (options, callback)
{
  // TODO: remove when all projects have libraries updated (with DataMap.reload.options parameter)
  if (typeof options === "function") {
    callback = options;
    options = undefined;
  }
  //
  options = Object.assign({}, options);
  options.reload = true;
  this.load(options, callback);
};


/**
 * Remove all data from datamap
 */
App.DataMap.prototype.clear = function ()
{
  // Remove all rows
  this.rows.forEach(function (row) {
    if (!row)
      return;
    //
    if (!this.useWindow)
      row.removeElement();
    row.close();
  }.bind(this));
  //
  this.rows.length = 0;
  this.uiRows.length = 0;
  this.loaded = false;
  //
  if (!this.loadingData) {
    this.updatePageLimit();
    this.setPosition();
    this.initElement();
    this.removeGroups();
  }
  else if (this.useWindow)
    this.schema.listBox.rowCount = 0;
};


/**
 * Save data of datamap
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * @return {Boolean} true if successful
 */
App.DataMap.prototype.save = function (options, callback)
{
  // TODO: remove when all projects have libraries updated (with DataMap.load.options parameter)
  if (typeof options === "function") {
    callback = options;
    options = undefined;
  }
  //
  options = Object.assign({}, options);
  //
  // Set supplied dataStore to use for save
  var oldDataStore = this._dataStore;
  if (options.dataStore) {
    this.dataStore = options.dataStore;
    delete options.dataStore;
  }
  //
  var done = function (result, error) {
    // Restore old dataStore
    this.dataStore = oldDataStore;
    //
    // Notify onError event
    if (error)
      this.onError(error, {source: App.DataMap.errorSources.save, row: r});
    //
    if (callback)
      callback(result);
  }.bind(this);
  //
  if (!(this.query instanceof Object) || !this.query.from || this.query.from.length === 0)
    return done(null, new Error("Datamap not updatable"));
  //
  // Find first row modified
  var r;
  var qt;
  var beforeEvent;
  for (var i = 0; i < this.rows.length; i++) {
    r = this.rows[i];
    if (r.inserted && !r.deleted) {
      qt = App.Database.QueryTypes.INSERT;
      beforeEvent = this.beforeInsert;
      break;
    }
    //
    if (!r.inserted && r.updated && !r.deleted) {
      qt = App.Database.QueryTypes.UPDATE;
      beforeEvent = this.beforeUpdate;
      break;
    }
    //
    if (!r.inserted && r.deleted) {
      qt = App.Database.QueryTypes.DELETE;
      beforeEvent = this.beforeDelete;
      break;
    }
  }
  //
  if (!qt) {
    // If there is not other changes terminate
    if (options.again) {
      // Notify afterSave event
      this.afterSave(function (result, error) {
        done(true, error);
      });
    }
    else
      done(true);
    //
    return;
  }
  //
  var skip = false;
  var afterSaveCallback = function (rowsAffected, error) {
    // If no error but no row affected throw error
    if (!error && rowsAffected === 0)
      error = new Error("No record affected");
    //
    if (error)
      return done(null, error);
    //
    var afterEventCallback = function (result, error) {
      if (error)
        return done(null, error);
      //
      // Proceed with next row changed
      this.save({again: true}, done);
    }.bind(this);
    //
    // Notify after event
    if (qt === App.Database.QueryTypes.INSERT) {
      if (!skip) {
        // Set autoincrement column
        var cc = this.schema.getAutoincrementColumn();
        if (this.dataStore.lastInsertId > 0 && cc)
          r[cc.name] = this.dataStore.lastInsertId;
      }
      //
      r.updated = false;
      r.inserted = false;
      this.afterInsert(r, afterEventCallback);
    }
    else if (qt === App.Database.QueryTypes.UPDATE) {
      r.updated = false;
      this.afterUpdate(r, afterEventCallback);
    }
    else if (qt === App.Database.QueryTypes.DELETE) {
      this.remove(r);
      this.afterDelete(r, afterEventCallback);
    }
  }.bind(this);
  //
  var beforeSaveCallback = function (result, error) {
    if (error)
      return done(null, error);
    //
    // Notify before event
    beforeEvent.call(this, r, function (result, error) {
      if (error)
        return done(null, error);
      //
      skip = result;
      if (skip)
        return afterSaveCallback(); // Call only after save operations
      //
      // Create the query object
      var q = null;
      if (qt === App.Database.QueryTypes.INSERT)
        q = r.getInsertQuery();
      else if (qt === App.Database.QueryTypes.UPDATE)
        q = r.getUpdateQuery();
      else if (qt === App.Database.QueryTypes.DELETE)
        q = r.getDeleteQuery();
      //
      // If no query need continue
      if (!q)
        return afterSaveCallback();
      //
      // Compose sql statement of the query
      var parameters = [];
      var sql = this.dataStore.createSqlStatement(q, parameters);
      //
      // Execute the query
      this.dataStore.executeSQL(sql, false, parameters, afterSaveCallback);
    }.bind(this));
  }.bind(this);
  //
  // If is the first change notify beforeSave event
  if (!options.again)
    this.beforeSave(beforeSaveCallback);
  else
    beforeSaveCallback();
};


/**
 * Add a filter to a column
 * @param {String} columnName
 * @param {String} f filter to add
 * @param {Boolean} or
 * @param {Boolean} qbe
 */
App.DataMap.prototype.addFilter = function (columnName, f, or, qbe)
{
  // If column not exists skip filter
  var dc = this.schema.columns[columnName];
  if (dc)
    dc.addFilter(f, or, qbe);
};


/**
 * Clear all filters
 * @param {String} columnName
 */
App.DataMap.prototype.clearFilters = function (columnName)
{
  if (columnName) {
    // Clear filters of column
    var dc = this.schema.columns[columnName];
    if (dc)
      dc.clearFilters();
  }
  else {
    // Clear filters of all columns
    for (var c in this.schema.columns)
      this.schema.columns[c].clearFilters();
  }
};


/**
 * Apply filters to the query
 * @param {Object} query
 */
App.DataMap.prototype.applyFilters = function (query)
{
  var wcH = null;
  var wcW = null;
  for (var c in this.schema.columns) {
    var col = this.schema.columns[c];
    //
    // Create all conditions of filters in AND
    var wcs = col.getAllFilterConditions(false);
    var aggr = col.useAggregateFuctions();
    for (var i = 0; i < wcs.length; i++) {
      if (aggr) {
        if (!query.having)
          query.having = [];
        query.having.push(wcs[i]);
      }
      else {
        if (!query.where)
          query.where = [];
        query.where.push(wcs[i]);
      }
    }
    //
    // Comopose the condition of filters in OR
    wcs = col.getAllFilterConditions(true);
    for (var i = 0; i < wcs.length; i++) {
      if (aggr) {
        if (!wcH)
          wcH = wcs[i];
        else
          wcH = {expr: "or", p: [wcH, wcs[i]]};
      }
      else {
        if (!wcW)
          wcW = wcs[i];
        else
          wcW = {expr: "or", p: [wcW, wcs[i]]};
      }
    }
  }
  //
  if (wcH) {
    if (!query.having)
      query.having = [];
    query.having.push({p: [{expr: "("}, wcH, {expr: ")"}]});
  }
  //
  if (wcW) {
    if (!query.where)
      query.where = [];
    query.where.push({p: [{expr: "("}, wcW, {expr: ")"}]});
  }
};


/**
 * Clear all error of row and columns
 */
App.DataMap.prototype.clearErrors = function ()
{
  for (var i = 0; i < this.rows.length; i++)
    this.rows[i].clearErrors();
};


/**
 * Set row o column error
 * @param {String} text
 * @param {String} columnName
 * @param {DataRow} row
 */
App.DataMap.prototype.setError = function (text, columnName, row)
{
  if (!row)
    row = this.row;
  row.setError(text, columnName);
};


/**
 * Check if there are errors in rows
 */
App.DataMap.prototype.hasErrors = function ()
{
  for (var i = 0; i < this.rows.length; i++) {
    if (this.rows[i].hasErrors())
      return true;
  }
  return false;
};


/**
 * Link an element to the datamap
 * @param {RemElem} el - the element to link
 * @param {Object} prop - the element properties
 * @param {Array} dcarray - the array of the column to select
 */
App.DataMap.prototype.attachElement = function (el, prop, dcarray)
{
  // If are not specified the columns add all columns
  if (!dcarray)
    dcarray = Object.keys(this.schema.columns);
  //
  el.onNavigate = function (ev) {
    this.onNavigate(ev.cmd, ev.columns, ev.rows, ev.pos);
  }.bind(this);
  //
  el.onFilter = function () {
    if (this.dataMap)
      this.dataMap.search(this.filter === " " ? "" : this.filter);
  };
  //
  this.outputElements.push({element: el, prop: prop, dcarray: dcarray});
  el.dataMap = this;
  //
  // If the view is already open, load data
  if (this.view.isOpen)
    this.loadElementData(el, prop, dcarray);
};


/**
 * Load initial data of an element
 * @param {RemElem} el - the element to add data to
 * @param {Object} prop - property of the element that will contain data
 * @param {Array} dcarray - columns to add
 */
App.DataMap.prototype.loadElementData = function (el, prop, dcarray)
{
  // If view is opening try later
  if (!this.view.isOpen) {
    setTimeout(function () {
      if (this.app.eleMap[this.view.id])
        this.loadElementData(el, prop, dcarray);
    }.bind(this), 10);
    return;
  }
  //
  // Select the columns to add
  var cols = this.selectColumns(dcarray);
  //
  // Prepare column schema
  var sc = {};
  for (var j = 0; j < cols.length; j++)
    sc[cols[j]] = this.schema.columns[cols[j]].toJSON();
  //
  var noLabel = dcarray.indexOf("") > -1 ? true : false;
  //
  // Retrieve data to load
  var r = [];
  for (var i = 0; i < this.uiRows.length; i++) {
    var a = [];
    if (noLabel)
      a.push("");
    for (j = 0; j < cols.length; j++)
      a.push(this.uiRows[i].getColumn(cols[j]));
    r.push(a);
  }
  //
  if (noLabel)
    cols.unshift("");
  //
  // Send command
  var d = {columns: cols, data: r, events: ["onNavigate"], schema: sc};
  var f = {};
  f[prop] = d;
  var r = {obj: el.id, id: "updateElement", cnt: f};
  this.view.app.sendCommands([r], true);
};


/**
 * Look for the output element and load its initial data
 * @param {String} id - the element to load data in
 */
App.DataMap.prototype.initElement = function (id)
{
  for (var i = 0; i < this.outputElements.length; i++) {
    var el = this.outputElements[i].element;
    if (this.templateDataMap && el === this.templateDataMap.parentElement)
      el = this.parentElement;
    if (el.id === id || id === undefined) {
      this.loadElementData(el, this.outputElements[i].prop, this.outputElements[i].dcarray);
      break;
    }
  }
};


/**
 * Remove data from elements
 * @param {DataRow} row - row to delete
 */
App.DataMap.prototype.removeElementsData = function (row)
{
  for (var i = 0; i < this.outputElements.length; i++) {
    var f = {};
    f[this.outputElements[i].prop] = {pos: row.position};
    var el = this.outputElements[i].element;
    if (this.templateDataMap && el === this.templateDataMap.parentElement)
      el = this.parentElement;
    var r = {obj: el.id, id: "updateElement", cnt: f};
    this.view.app.sendCommands([r], true);
  }
};


/**
 * Add data to elements
 * @param {DataRow} row - the row to add
 */
App.DataMap.prototype.addElementsData = function (row)
{
  for (var i = 0; i < this.outputElements.length; i++) {
    // Retrieve data from columns the element needs
    var cols = this.selectColumns(this.outputElements[i].dcarray);
    var a = [];
    for (var j = 0; j < cols.length; j++)
      a.push(row.getColumn(cols[j]));
    //
    // Send command
    var f = {};
    f[this.outputElements[i].prop] = {data: a};
    var el = this.outputElements[i].element;
    if (this.templateDataMap && el === this.templateDataMap.parentElement)
      el = this.parentElement;
    var r = {obj: el.id, id: "updateElement", cnt: f};
    this.view.app.sendCommands([r], true);
  }
};


/**
 * Update elements data
 * @param {DataRow} row
 */
App.DataMap.prototype.updateElementsData = function (row)
{
  if (this.loadingData)
    return;
  //
  for (var i = 0; i < this.outputElements.length; i++) {
    // Retrieve data from columns the element needs
    var cols = this.selectColumns(this.outputElements[i].dcarray);
    var a = [];
    for (var j = 0; j < cols.length; j++)
      a.push(row.getColumn(cols[j]));
    //
    // Send command
    var f = {};
    f[this.outputElements[i].prop] = {pos: row.position, data: a};
    var el = this.outputElements[i].element;
    if (this.templateDataMap && el === this.templateDataMap.parentElement)
      el = this.parentElement;
    var r = {obj: el.id, id: "updateElement", cnt: f};
    this.view.app.sendCommands([r], true);
  }
};


/**
 * Execute an operation on the datamap
 * @param {Object} op - the operation to execute and its parameters.
 */
App.DataMap.prototype.chgData = function (op) {
  switch (op.name) {
    case "delete":
      if (op.id) {
        var rows = this.find({id: op.id});
        if (rows.length > 0)
          this.remove(rows[0]);
      }
      break;
    case "insert":
      if (op.data)
        this.insert(op.data);
      break;
    case "update":
      if (op.id !== undefined && op.data) {
        var rows = this.find({id: op.id});
        if (rows.length > 0) {
          var row = rows[0];
          var keys = Object.keys(op.data);
          for (var i = 0; i < keys.length; i++)
            row[keys[i]] = op.data[keys[i]];
        }
      }
      break;
  }
};


/**
 * Find a row matching the filter criteria
 * @param {Object/String} filter
 * @param {Object/String} update
 * @returns {Array}
 */
App.DataMap.prototype.find = function (filter, update)
{
  // No rows -> no result
  if (this.rows.length === 0)
    return [];
  //
  var filters = {};
  if (filter && typeof filter === "object")
    filters = filter;
  else if (typeof filter === "string")
    filters = this.parseFindExpression(filter, false);
  //
  var setters = {};
  if (update && typeof update === "object")
    setters = update;
  else if (typeof update === "string")
    setters = this.parseFindExpression(update, true);
  //
  // Extract some special attributes
  var op = filters.op;
  delete filters.op;
  var cmp = filters.cmp;
  delete filters.cmp;
  //
  var keys = Object.keys(filters);
  //
  // Find rows that matchs with filters
  var rows = [];
  var loopRows = (filters.visible ? this.uiRows : this.rows);
  for (var i = 0; i < loopRows.length; i++) {
    var row = loopRows[i];
    //
    // No filters -> all rows
    var ok = (keys.length === 0);
    for (var k = 0; k < keys.length; k++) {
      var col = keys[k];
      var rv = row[col];
      var fv = filters[col];
      var match = false;
      switch (col) {
        case "inserted":
        case "updated":
        case "deleted":
        case "hidden":
          match = (rv === fv);
          break;

        case "visible":
          match = (filters.visible || !row.isVisible());
          break;

        case "error":
          match = (row.hasErrors() === fv);
          break;

        default:
          var dc = this.schema.columns[col];
          switch (App.Database.classDT(dc.datatype)) {
            case App.Database.DataTypeClasses.DATETIME:
              match = ((!rv && !fv) || (rv && fv && rv.getTime() === fv.getTime()));
              break;

            case undefined:
            case App.Database.DataTypeClasses.STRING:
              if (fv && rv) {
                if (cmp === "like") {
                  if (fv.startsWith("%") && fv.endsWith("%")) {
                    match = (rv.toLowerCase().indexOf(fv.substr(1, fv.length - 2).toLowerCase()) >= 0);
                    break;
                  }
                  else if (fv.startsWith("%")) {
                    match = rv.toLowerCase().endsWith(fv.substr(1, fv.length - 1).toLowerCase());
                    break;
                  }
                  else if (fv.endsWith("%")) {
                    match = rv.toLowerCase().startsWith(fv.substr(0, fv.length - 1).toLowerCase());
                    break;
                  }
                }
                else if (cmp === "regexp") {
                  match = new RegExp(fv, "i").test(rv);
                  break;
                }
              }
              match = (rv === fv);
              break;

            default:
              match = (rv === fv);
              break;
          }
      }
      //
      if (op === "or")
        ok = ok || match;
      else {
        ok = match;
        if (!ok)
          break;
      }
    }
    //
    if (ok)
      rows.push(row);
    //
    // Set boolean value of setter on each row
    // or (for not boolean value) only on rows that matchs
    var keysSetter = Object.keys(setters);
    for (var k = 0; k < keysSetter.length; k++) {
      var col = keysSetter[k];
      var sv = setters[col];
      //
      // Boolean attribute without value are setted true for all rows that matches, false for others
      if (typeof sv !== "boolean" && this.isBooleanRowCol(col))
        sv = ok;
      else if (!ok)
        continue;
      //
      // Visible attribute is treated as !hidden
      if (col === "visible") {
        col = "hidden";
        sv = !sv;
      }
      //
      row[col] = sv;
    }
  }
  //
  return rows;
};


/**
 * Return true if attribute of row is of boolean type
 * @param {String} col
 * @returns {Array}
 */
App.DataMap.prototype.isBooleanRowCol = function (col)
{
  switch (col) {
    case "inserted":
    case "updated":
    case "deleted":
    case "hidden":
    case "visible":
    case "error":
      return true;

    default:
    {
      var dc = this.schema.columns[col];
      return (dc && dc.datatype === App.Database.DataTypes.BOOL);
    }
  }
};


/**
 * Parse expression for find function
 * @param {String} expr
 * @param {Boolean} update
 * @returns {Array}
 */
App.DataMap.prototype.parseFindExpression = function (expr, update)
{
  if (!expr)
    return {};
  //
  // Parse string to object filters
  var obj = {};
  //
  // Separators handled
  var seps = [",", " or ", " and "];
  //
  // Split by each separator
  while (seps.length > 0) {
    var sep = seps.shift();
    var cond = expr.split(sep);
    //
    // If separator not found try with next separator
    if (cond.length === 1 && seps.length > 0)
      continue;
    //
    // Set if separator is or
    if (sep === " or ")
      obj.op = "or";
    //
    // Split name and value
    for (var i = 0; i < cond.length; i++) {
      // Comparators handled
      var cmps = [" like ", " regexp ", "="];
      var cmp = "";
      while (cmps.length > 0) {
        cmp = cmps.shift();
        var parts = cond[i].split(cmp);
        //
        // If comparator not found try with next comparator
        if (parts.length > 1) {
          if (!obj.cmp)
            obj.cmp = cmp.trim();
          break;
        }
      }
      //
      var col = parts[0].trim();
      var value = (parts.length > 1 ? parts.slice(1).join(cmp).trim() : "");
      var quoters = ["\"", "'"];
      //
      // If value starts with a quote char
      if (quoters.indexOf(value[0]) >= 0) {
        // If also ends with same char remove quote chars
        if (value[value.length - 1] === value[0])
          value = value.substring(1, value.length - 1);
        else if (i + 1 < cond.length) {
          // Separator is after quote char
          // Concatenate with next condition and try again
          cond[i] += sep + cond.splice(i + 1, 1)[0];
          i--;
          continue;
        }
      }
      //
      // Convert value
      if (this.isBooleanRowCol(col)) {
        if (value === "false")
          value = false;
        else if (value === "true" || (!value && !update))
          value = true;
        else
          value = null;
      }
      else {
        var dc = this.schema.columns[col];
        if (dc)
          value = dc.stringToValue(value);
      }
      obj[col] = value;
    }
    break;
  }
  return obj;
};


/**
 * Select columns
 * @param {Array} dcarray
 * @returns {Array}
 */
App.DataMap.prototype.selectColumns = function (dcarray)
{
  var cols = [];
  var schemaCols = this.schema.getColumns();
  for (var i = 0; i < dcarray.length; i++) {
    var dc = dcarray[i];
    if (schemaCols[dc] !== undefined)
      cols.push(dc);
  }
  //
  return cols;
};


/**
 * Find all token that refer to client objects
 * @param {Object} obj query or token
 */
App.DataMap.getClientRefs = function (obj)
{
  var clientRefs = [];
  if (obj.expr && obj.clientRef)
    clientRefs.push(obj);
  else {
    // Search recursively client references in various containers
    var tokens = [].concat(obj.select).concat(obj.from).concat(obj.where).concat(obj.group).concat(obj.having).concat(obj.order).concat(obj.p).concat(obj.top).concat(obj.offset);
    for (var i in tokens) {
      if (tokens[i])
        clientRefs = clientRefs.concat(App.DataMap.getClientRefs(tokens[i]));
    }
    //
    if (obj.union)
      clientRefs = clientRefs.concat(App.DataMap.getClientRefs(obj.union));
  }
  return clientRefs;
};


/**
 * Attach observer to client ref objects
 */
App.DataMap.prototype.attachClientRefsObserver = function ()
{
  // No structured query no client refs
  if (!(this.query instanceof Object))
    return;
  //
  if (!this.clientRefs)
    this.clientRefs = {};
  //
  var tokens = App.DataMap.getClientRefs(this.query);
  for (var i in tokens) {
    var tk = tokens[i];
    var ref = this.clientRefs[tk.expr];
    if (!ref) {
      ref = {tokens: []};
      this.clientRefs[tk.expr] = ref;
      //
      // Retreave object referenced
      if (tk.obj) {
        // Object just given by (see addDOFilter)
        ref.obj = tk.obj;
        delete tk.obj;
        ref.prop = tk.prop;
        delete tk.prop;
      }
      else {
        var exprTokens = tk.expr.split(".");
        for (var j = 0; (j < exprTokens.length); j++) {
          var etk = exprTokens[j];
          //
          // Split token as row["column"]
          var leftBracket = etk.indexOf("[");
          if (leftBracket > 0 && etk.endsWith("]")) {
            exprTokens.splice(j + 1, 0, etk.substring(leftBracket + 2, etk.length - 2));
            etk = etk.substring(0, leftBracket);
          }
          //
          // Resolve first token
          if (j === 0) {
            switch (etk) {
              case "view":
                ref.obj = this.view;
                break;
              case "app":
                ref.obj = this.app;
                break;
              case "App":
                ref.obj = App;
                break;
              case "this":
                ref.obj = this.this || this;
                break;
            }
          }
          else if (ref.obj instanceof App.DataMap && etk === "row") {
            // Reference to current row of a DataMap can be handle only if "row" is penultimate token
            if (j !== exprTokens.length - 2)
              delete ref.obj;
            break;
          }
          else if (j < exprTokens.length - 1)
            ref.obj = ref.obj[etk];
          //
          // Observe can be defined only if ref.obj is an istance of object
          if (!ref.obj) {
            delete ref.obj;
            break;
          }
        }
        //
        // Property to read is the last token
        if (ref.obj)
          ref.prop = exprTokens.pop();
      }
      //
      // Define observer callback (only for datamap that are elements)
      if (ref.obj && this.parentElement) {
        ref.observer = function (changes) {
          for (var i in changes) {
            var chg = changes[i];
            for (var j in this.clientRefs) {
              var ref = this.clientRefs[j];
              if (chg.object === ref.obj) {
                var value = undefined;
                // Is a reference to actual row
                if (ref.obj instanceof App.DataMap && (chg.name === ("row." + ref.prop) || chg.name === "row"))
                  value = this.getClientRefValue(ref);
                else if (chg.name === ref.prop)
                  value = this.getClientRefValue(ref);
                //
                // Update value of each token
                if (value !== undefined) {
                  for (var k in ref.tokens) {
                    if (value !== ref.tokens[k].value) {
                      this.needReload = true;
                      ref.tokens[k].value = value;
                    }
                  }
                }
              }
            }
          }
          //
          // Some reference has changed, reloading data
          this.refreshData();
        }.bind(this);
        //
        // Attach the observer to client object referenced
        Object.observe(ref.obj, ref.observer);
        if (ref.obj instanceof App.DataMap || ref.obj instanceof App.DataRow)
          ref.obj.observed = true;
      }
    }
    //
    // Add token to list (if I have not yet done)
    if (ref.tokens.indexOf(tk) === -1)
      ref.tokens.push(tk);
    //
    // Update token value with actual value
    tk.value = this.getClientRefValue(ref);
  }
};


/**
 * Detach observer to client ref objects
 */
App.DataMap.prototype.detachClientRefsObserver = function ()
{
  if (!this.clientRefs)
    return;
  //
  var keys = Object.keys(this.clientRefs);
  for (var i = 0; i < keys.length; i++) {
    var ref = this.clientRefs[keys[i]];
    if (ref.obj)
      Object.unobserve(ref.obj, ref.observer);
  }
};


/**
 * Refresh client references of query
 */
App.DataMap.prototype.refreshQuery = function ()
{
  this.detachClientRefsObserver();
  delete this.clientRefs;
};


/**
 * Returns the value of client object reference
 * @param {Object} ref
 */
App.DataMap.prototype.getClientRefValue = function (ref)
{
  if (!ref.obj)
    return null;
  //
  var value = null;
  if (ref.obj instanceof App.DataMap) {
    // If reference is to parent read the value from her
    if (this.parent && this.parent.parent === ref.obj)
      value = this.parent[ref.prop];
    else if (ref.obj.row) // else from actual row
      value = ref.obj.row[ref.prop];
  }
  else
    value = ref.obj[ref.prop];
  //
  return (value === undefined ? null : value);
};


/**
 * Return a copy of this object
 * @param {Object} map
 */
App.DataMap.prototype.clone = function (map)
{
  var dm = new App.DataMap(this.view);
  if (map)
    map[this.id] = dm;
  var keys = Object.keys(this);
  for (var k in keys) {
    var v = keys[k];
    //
    // Don't copy some objects
    if (["rows", "uiRows", "uiCache", "uiRecycleBin", "clientRefs", "needReload", "loadingPages", "rowsToUpdate"].indexOf(v) >= 0) {
      continue;
    }
    //
    if (v === "elements") {
      for (var i = 0; i < this.elements.length; i++) {
        var ne = this.elements[i].clone(map);
        dm.elements.push(ne);
        //
        // Relink original parent
        if (ne instanceof App.DataMap)
          ne.parentElement = dm;
        else
          ne.parent = dm;
      }
    }
    else if (v === "id") {
      // The id property is special and needs to be generated here
      dm[v] = this.app.generateCounter();
      this.app.eleMap[dm.id] = dm;
    }
    else if (v === "schema") {
      dm.schema = this.schema.clone();
      dm.schema.parent = dm;
    }
    else if (["_searchOptions", "query"].indexOf(v) >= 0)
      dm[v] = JSON.parse(JSON.stringify(this[v]));
    else if (v === "outputElements")
      dm.outputElements = this.outputElements.slice(0);
    else
      dm[v] = this[v];
  }
  //
  dm.templateDataMap = this;
  dm.needReload = this.autoLoad;
  //
  setImmediate(function () {
    if (dm.parentElement instanceof App.RemElem)
      dm.parentElement.dataMap = dm;
    //
    if (dm.autoLoad)
      dm.refreshData();
  });
  //
  return dm;
};


/**
 * Clone my sub DataMap (if I have not already done)
 * @param {DataRow} row
 */
App.DataMap.prototype.cloneChildsForRow = function (row)
{
  if (!row.subDataMaps) {
    row.subDataMaps = [];
    //
    // Add to row map also original elements
    var tdm = row.datamap.templateDataMap;
    if (tdm && tdm.template) {
      for (var i in row.map) {
        var elems = [];
        tdm.template.getElementsByPid(row.map[i].pid, elems);
        if (elems.length > 0)
          row.map[elems[0].id] = row.map[i];
      }
    }
    //
    // Clonus my child elements; if I are recursive clonus also myself
    var subDataMaps = (this.recursive ? this.elements.concat([this.templateDataMap]) : this.elements);
    for (var i in subDataMaps) {
      // Clone the sub dataMap
      var dm = subDataMaps[i].clone();
      //
      // Add it to array of subDataMaps of the row
      row.subDataMaps.push(dm);
      row[dm.name] = dm;
      //
      // If I'm cloning myself
      if (subDataMaps[i] === this.templateDataMap) {
        // His parent is me
        dm.parentElement = this;
        //
        if (this.templateDataMap.template) {
          // Clone my listBox and append it to the RemElem of the row
          var lb = this.templateDataMap.schema.listBox.clone(row.map);
          row.element.appendChild(lb);
          //
          // Assign cloned listBox
          dm.schema.setListBox(lb);
          //
          // Recreate also the direct pointer
          row.element[dm.schema.listBox.name] = lb;
        }
      }
      else {
        // Assign cloned listBox
        if (dm.template)
          dm.schema.setListBox(row.map[dm.schema.listBox.id]);
      }
      //
      // Link the dataMap to parent row
      dm.parent = row;
    }
  }
};


/**
 * Set the parent of the datamap
 * @param {View} view
 * @param {RemElem} parent
 */
App.DataMap.prototype.setParent = function (view, parent)
{
  this.view = view;
  this.parentElement = parent;
};


/**
 * Refresh data of the query from db
 */
App.DataMap.prototype.refreshData = function ()
{
  // Load data only if I have not a listBox or if my listBox is visible
  if (this.needReload && (!this.template || this.schema.listBox.visible))
    this.reload();
};


/**
 * Event fired when change visibility of an attached element
 * @param {RemElem} el element that has changed visibility
 */
App.DataMap.prototype.onChangeElementVisibility = function (el)
{
  // If the element is my listBox reload the data
  if (el === this.schema.listBox)
    this.refreshData();
};


/**
 * Fire onSearch event
 * @param {String} text
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.fireOnSearch = function (text, options, callback)
{
  this.onSearch(text, options, callback);
};


/**
 * Event fired when start an LKE search
 * @param {String} text to search
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.onSearch = function (text, options, callback)
{
  if (callback)
    callback();
};


/**
 * Do an LKE search on the datamap
 * @param {String} text to search
 */
App.DataMap.prototype.search = function (text)
{
  // No structured query no search
  if (!(this.query instanceof Object))
    return;
  //
  if (typeof text === "string") {
    // Notify onSearch event
    var options = {skip: false};
    this.fireOnSearch(text, options, function (result, error) {
      if (options.skip || error)
        return;
      //
      // Remome all the where clauses that refers to parent element
      this.removeParentFilters();
      //
      // Start the search
      this.tryLKESearch(text, 0);
    }.bind(this));
  }
  else {
    // Readd all the where clauses that refers to parent element previously removed from the query
    this.addParentFilters();
    //
    // Remove all qbe filters added previously for lke filter
    this.removeLKEFilters();
    //
    // Reload decode data
    this.reload();
  }
};


/**
 * Remome temporaly all the where clauses that refers to parent element
 */
App.DataMap.prototype.removeParentFilters = function ()
{
  if (!this.query.where)
    return;
  //
  for (var i = 0; i < this.query.where.length; i++) {
    var clientRefs = App.DataMap.getClientRefs(this.query.where[i]);
    for (var j = 0; j < clientRefs.length; j++) {
      if (this.clientRefs[clientRefs[j].expr].obj === this.parentElement) {
        if (!this.parentRefs)
          this.parentRefs = [];
        this.parentRefs.push(this.query.where.splice(i--, 1)[0]);
        break;
      }
    }
  }
  //
  if (this.query.where.length === 0)
    delete this.query.where;
};


/**
 * Re-add all the where clauses that refers to parent element previously removed from the query
 */
App.DataMap.prototype.addParentFilters = function ()
{
  if (!this.parentRefs)
    return;
  //
  if (!this.query.where)
    this.query.where = [];
  for (var i = 0; i < this.parentRefs.length; i++)
    this.query.where.push(this.parentRefs[i]);
  //
  delete this.parentRefs;
};


/**
 * Remove all qbe filters added previously for lke filter
 */
App.DataMap.prototype.removeLKEFilters = function ()
{
  if (!this.lkeFilters)
    return;
  //
  for (var i = 0; i < this.lkeFilters.length; i++) {
    for (var j in this.schema.columns) {
      var dc = this.schema.columns[j];
      var idx = dc.filters.indexOf(this.lkeFilters[i]);
      if (idx >= 0) {
        dc.filters.splice(idx, 1);
        break;
      }
    }
  }
  //
  delete this.lkeFilters;
};


/**
 * Do an attempt of LKE research
 * @param {String} text
 * @param {Integer} ntry
 */
App.DataMap.prototype.tryLKESearch = function (text, ntry)
{
  // Remove all qbe filters added previously
  this.removeLKEFilters();
  //
  // If filter is empty find all data
  if (text) {
    this.lkeFilters = [];
    var decodeColumnIndex;
    var cols = this.schema.getColumns(true);
    for (var i = 0; i < cols.length; i++) {
      var dc = cols[i];
      //
      // I do research only on character columns (no UUID)
      if (App.Database.classDT(dc.datatype) !== App.Database.DataTypeClasses.STRING
              || dc.datatype === App.Database.DataTypes.UUID)
        continue;
      //
      // I don't research on unbound columns
      if (dc.unbound)
        continue;
      //
      // Search a good decode column
      if (decodeColumnIndex === undefined && (cols.length === 0 || i > 0))
        decodeColumnIndex = i;
      //
      // In the first 3 attempts I do research only on the decode column
      if (ntry < 3 && dc.index !== decodeColumnIndex)
        continue;
      //
      // In the attempts forward to the 3rd I do research only on the not decode column
      if (ntry >= 3 && dc.index === decodeColumnIndex)
        continue;
      //
      var criteria = "";
      //
      // If text contains an * skip first two attempt
      if (text.indexOf("*") !== -1) {
        criteria = text;
        if (ntry < 2)
          ntry = 2;
      }
      else {
        switch (ntry) {
          case 0:
            criteria = "=" + text;
            break;
          case 1:
            criteria = text + "*";
            break;
          case 2:
          case 3:
            criteria = "*" + text + "*";
            break;
        }
      }
      //
      // Add the qbe criteria
      if (criteria)
        this.lkeFilters.push(dc.addFilter(criteria, true));
    }
    //
    // If you have not added any criteria I stop
    if (this.lkeFilters.length === 0) {
      this.initElement();
      return;
    }
  }
  //
  // Find data
  this.searching = true;
  this.reload(null, function (result, error) {
    this.searching = false;
    //
    // If the research did not get results I continue with the next attempt
    if (!error && this.rows.length === 0)
      this.tryLKESearch(text, ++ntry);
    else
      this.initElement();
  }.bind(this));
};


/**
 * Return the class for row
 * @return {Function}
 */
App.DataMap.prototype.getRowClass = function ()
{
  if (!this.rowClass)
    this.rowClass = this.createRowClass();
  //
  return this.rowClass;
};


/**
 * Define the class for rows of this datamap
 * @return {Function}
 */
App.DataMap.prototype.createRowClass = function ()
{
  // Define the class for rows of this datamap
  var rc = function (parent, data) {
    App.DataRow.call(this, parent, data);
  };
  //
  // Make rowClass extend App.DataRow
  rc.prototype = new App.DataRow(true);
  //
  // Define column properties
  for (var c in this.schema.columns)
    this.schema.columns[c].defineClassProperty(rc);
  //
  return rc;
};


/**
 * Create a row for this datamap
 * @param {Object} data - value of row
 * @return {DataRow}
 */
App.DataMap.prototype.createRow = function (data)
{
  // If data is a DataRow return it as is
  if (data instanceof this.getRowClass())
    return data;
  //
  // Otherwise create a DataRow with it
  return new (this.getRowClass())(this, data);
};


/**
 * Return true if there is at least one row is modified
 * @return {Boolean}
 */
App.DataMap.prototype.isModified = function ()
{
  for (var i = 0; i < this.rows.length; i++) {
    if (this.rows[i].isModified())
      return true;
  }
  return false;
};


/**
 * Reset each row to original status
 */
App.DataMap.prototype.undoChanges = function ()
{
  var rows = this.rows.slice();
  for (var i = 0; i < rows.length; i++)
    rows[i].restoreOriginal();
};


/**
 * Save to JSON
 */
App.DataMap.prototype.toJSON = function ()
{
  var json = {};
  //
  // Save the query
  json.query = JSON.parse(JSON.stringify(this.query));
  //
  // Save partial flag
  if (typeof this.partial === "boolean")
    json.partial = this.partial;
  //
  // Save the schema
  json.md = App.DataSchema.prototype.toJSON.call(this.schema);
  //
  // Save the rows
  if (this.rows.length > 0) {
    json.rows = [];
    for (var i = 0; i < this.rows.length; i++)
      json.rows.push(App.DataRow.prototype.toJSON.call(this.rows[i]));
  }
  //
  return json;
};


/**
 * Load from JSON
 * @param {Object} json
 */
App.DataMap.prototype.loadFromJSON = function (json)
{
  // Load the query
  this.addQuery(json.query);
  //
  // Load partial flag
  if (typeof json.partial === "boolean")
    this.partial = json.partial;
  //
  // Load the schema
  this.schema = new App.DataSchema(this);
  this.schema.loadFromJSON(json.md);
  //
  // Load the rows
  if (json.rows) {
    for (var i = 0; i < json.rows.length; i++) {
      var row = this.createRow();
      row.loadFromJSON(json.rows[i]);
      this.add(row);
    }
  }
  this.loaded = true;
};


/**
 * Give client the next page of a paged DataMap
 */
App.DataMap.prototype.nextPage = function ()
{
  var page = (this.dataPagingMode ? Math.floor(this.currentPageSize / this.dataPageSize) : 0);
  this.loadPage(page, function (result, error) {
    if (error)
      return;
    //
    var e = {limit: this.currentPageSize, increment: this.pageSize};
    this.onNextPage(e);
    this.currentPageSize += e.increment;
    //
    // Load next page of data
    if (this.dataPagingMode)
      this.loadPage(page + 1, function () {
      });
  }.bind(this));
};


/**
 * Handle next page event
 * @param {Object} event
 */
App.DataMap.prototype.onNextPage = function (event)
{
};


/**
 * When the uiRows.length falls under page boundaries, lower the currentPageSize
 */
App.DataMap.prototype.updatePageLimit = function ()
{
  if (this.pageSize) {
    var p1 = Math.floor(this.uiRows.length / this.pageSize);
    var p2 = Math.floor(this.currentPageSize / this.pageSize);
    if (p1 && p1 < p2)
      this.currentPageSize = p1 * this.pageSize;
    //
    // Reset more rows indicator if not necessary
    if (this.currentPageSize >= this.uiRows.length && this.schema && this.schema.listBox)
      this.schema.listBox.moreRows = 0;
  }
};


/**
 * Returns the orderBy contitions
 */
App.DataMap.prototype.getSortConditions = function ()
{
  var orderBy = [];
  var sortBy = this.sortBy;
  for (var i = 0; i < sortBy.length; i++) {
    var col = Object.keys(sortBy[i])[0];
    var dc = this.schema.columns[col];
    //
    // If it is not a known column don't sort
    if (!dc)
      return [];
    //
    // If column is not from query don't sort
    var tk = dc.getSelectToken();
    if (!tk)
      return [];
    //
    tk = JSON.parse(JSON.stringify(tk));
    delete tk.alias;
    if (sortBy[i][col] === "desc")
      tk.order = "d";
    orderBy.push(tk);
  }
  return orderBy;
};


/**
 * Define sort criteria and sort if data are already been loaded
 */
Object.defineProperty(App.DataMap.prototype, "orderBy", {
  get: function () {
    return (this._orderBy ? this._orderBy.orig : null);
  },
  set: function (newOrder) {
    this._orderBy = this.parseOrderCriteria(newOrder);
    this._sortBy = (this._groupBy || []).concat(this._orderBy || []);
    this.sort();
  }
});


/**
 * Define group criteria and sort if data are already been loaded
 */
Object.defineProperty(App.DataMap.prototype, "groupBy", {
  get: function () {
    return (this._groupBy ? this._groupBy.orig : null);
  },
  set: function (newOrder) {
    this._groupBy = this.parseOrderCriteria(newOrder);
    this._sortBy = (this._groupBy || []).concat(this._orderBy || []);
    this.sort();
  }
});


/**
 * Return all sort criteria
 */
Object.defineProperty(App.DataMap.prototype, "sortBy", {
  get: function () {
    return this._sortBy || [];
  }
});


/**
 * Parse order criteria
 * @param {String/Object} criteria
 */
App.DataMap.prototype.parseOrderCriteria = function (criteria)
{
  var orderBy = [];
  orderBy.orig = criteria;
  //
  if (!criteria || criteria.length === 0)
    return;
  //
  // format1: “columnName1 asc, columnName2 desc”
  if (typeof criteria === "string")
    criteria = criteria.split(",");
  //
  // Transform in format2
  if (criteria instanceof Array) {
    for (var i = 0; i < criteria.length; i++) {
      var cr = null;
      //
      // format2: [{“columnName1”:”asc/desc”}, { … }]
      if (typeof criteria[i] === "object")
        cr = criteria[i];
      else if (typeof criteria[i] === "string") {
        // format3: [“columnName1 asc”, “columnName2 desc”]
        var parts = criteria[i].trim().split(" ");
        cr = {};
        cr[parts[0]] = parts[1];
      }
      //
      // Asc/Desc lowerCase
      var col = Object.keys(cr)[0];
      cr[col] = (cr[col] ? cr[col].toLowerCase() : "asc");
      //
      orderBy.push(cr);
    }
  }
  this.adjustSortByIndex(orderBy);
  return orderBy;
};


/**
 * Adjust sort criterias by column index
 * @param {Array} sort
 */
App.DataMap.prototype.adjustSortByIndex = function (sort) {
  if (!sort || !this.schema)
    return;
  //
  var cols = Object.keys(this.schema.columns);
  sort.forEach(function (cr) {
    var col = Object.keys(cr)[0];
    if (cols.indexOf(col) >= 0)
      return;
    //
    // If col is a column index find column name
    var index = Number(col);
    if (!isNaN(index) && index >= 1 && index <= cols.length) {
      cr[cols[index - 1]] = cr[col];
      delete cr[col];
    }
  });
};


/**
 * Add a filter to query
 * @param {String} sql
 * @param {Integer} queryIndex
 */
App.DataMap.prototype.filterBy = function (sql, queryIndex)
{
  if (!this.query)
    return;
  //
  // If queryIndex is specified append to required query in union
  var q = this.query;
  if (queryIndex) {
    for (var i = 0; i <= queryIndex && q.union; i++)
      q = q.union;
    if (!q)
      return;
  }
  //
  if (!q.where)
    q.where = [];
  q.where.push({expr: sql});
};


/**
 * Sort UI rows
 */
App.DataMap.prototype.sort = function ()
{
  // If there are no data to be sorted
  if (this.uiRows.length === 0 || this.sortBy.length === 0)
    return;
  //
  // Detach all elements from rows
  App.DataMap.prototype.reuseElements.call(this, this.uiRows.length);
  //
  // Sort visible rows
  this.uiRows.sort(function (row1, row2) {
    return this.compareRows(row1, row2);
  }.bind(this));
  //
  // Reattach all elements to rows
  if (!this.useWindow) {
    this.uiRows.forEach(function (row) {
      this.calcRowGroups(row);
      var e = row.inPagination() ? this.uiRecycleBin.shift() : null;
      if (e)
        row.attachElement(e);
      else
        row.updateElement(true);
    }.bind(this));
    //
    App.DataMap.prototype.removeUnusedElements.call(this);
  }
  else
    Object.getNotifier(this.schema.listBox).notify({type: "update", name: "rowCount", oldValue: this.schema.listBox.rowCount});
};


/**
 * Compare two rows
 * @param {DataRow} row1
 * @param {DataRow} row2
 * */
App.DataMap.prototype.compareRows = function (row1, row2)
{
  if (!row1 && !row2)
    return 0;
  if (!row1)
    return 1;
  if (!row2)
    return -1;
  //
  var res = 0;
  var sortBy = this.sortBy;
  for (var i = 0; i < sortBy.length; i++) {
    var cr = sortBy[i];
    var col = Object.keys(cr)[0];
    //
    var v1 = row1[col];
    var v2 = row2[col];
    if (v1 < v2)
      res = -1;
    else if (v1 > v2)
      res = 1;
    if (cr[col] === "desc")
      res = -res;
    if (res !== 0)
      break;
  }
  return res;
};


/**
 * Find position for a row beetwen that visible
 * @param {DataRow} row
 * */
App.DataMap.prototype.findRowPosition = function (row)
{
  // If datamap is ordered ...
  if (this.sortBy.length > 0) {
    // find position by comparing rows (with binary search)
    var start = 0;
    var end = this.uiRows.length - 1;
    var pos = 0;
    while (start <= end)
    {
      pos = Math.round((start + end) / 2);
      var comp = this.compareRows(row, this.uiRows[pos]);
      if (comp < 0)
        end = pos - 1;
      else if (comp > 0)
        start = pos++ + 1;
      else
        break;
    }
    return pos;
  }
  else {
    // find position searching first before/after visible row
    var len = this.rows.length;
    var pos = row.index;
    for (var i = 1; ; i++) {
      if (pos - i >= 0) {
        var prevRow = this.rows[pos - i];
        if (prevRow.isVisible())
          return prevRow.position + 1;
      }
      else
        return 0;
      //
      if (pos + i < len) {
        var nextRow = this.rows[pos + i];
        if (nextRow.isVisible())
          return nextRow.position;
      }
      else
        break;
    }
  }
  //
  return this.uiRows.length;
};


/**
 * Create schema by resultSet if needed
 * @param {Object} result
 */
App.DataMap.prototype.checkSchema = function (result)
{
  if (Object.keys(this.schema.columns).length > 0)
    return;
  //
  this.addSchema({p: this.dataStore.getColumnsDefByResult(result)});
};


/**
 * Execute load query (paged if needed)
 * @param {Object/String} qry
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.execQuery = function (qry, options, callback)
{
  // Compose sql statement of the query
  var parameters = [];
  qry.schema = this.schema;
  var sql = (typeof qry === "string" ? qry : this.dataStore.createSqlStatement(qry, parameters));
  var loadingPage = options.page || 0;
  //
  var afterQuery = function (result, error) {
    if (error)
      callback(null, error);
    else if (!loadingPage && this.dataPagingMode && this.dataPagingPreCount) {
      this.calcRowCount(qry, function (rc, error) {
        this._calcRowCount = rc;
        callback(result, error);
      }.bind(this));
    }
    else
      callback(result);
  }.bind(this);
  //
  switch (this.dataPagingMode) {
    case App.DataMap.dataPagingModes.none:
      this.dataStore.executeSQL(sql, true, parameters, afterQuery);
      break;

    case App.DataMap.dataPagingModes.offset:
    {
      var params = {};
      params.limit = this.dataPageSize;
      params.offset = loadingPage * this.dataPageSize;
      params.sql = sql;
      params.parameters = parameters;
      this.dataStore.queryOffsetPaging(params, afterQuery);
      break;
    }

    case App.DataMap.dataPagingModes.keyset:
    {
      if (typeof qry === "string")
        return callback(null, new Error("Paging mode 'keyset' not supported with query by string"));
      //
      if (!qry.order || qry.order.length === 0)
        return callback(null, new Error("Paging mode 'keyset' not supported with query without order by"));
      //
      var q = JSON.parse(JSON.stringify(qry));
      if (this.rows.length > 0) {
        var keyCols = [];
        var keyVals = [];
        var lastRow = this.rows[(loadingPage * this.dataPageSize) - 1];
        if (lastRow) {
          for (var i = 0; i < q.order.length; i++) {
            var tkO = q.order[i];
            //
            // Order clause as column index
            if (tkO.expr && /^\d+$/.test(tkO.expr))
              keyCols.push(q.select[parseInt(tkO.expr) - 1]);
            else {
              for (var j = 0; j < q.select.length; j++) {
                var tkS = JSON.parse(JSON.stringify(q.select[j]));
                delete tkS.alias;
                if (JSON.stringify(tkS) === JSON.stringify(tkO)) {
                  keyCols.push(tkS);
                  keyVals.push(lastRow[tkS.alias || tkS.expr]);
                  break;
                }
              }
            }
            //
            if (!keyCols[i])
              return callback(null, new Error("Paging mode 'keyset' not supported without all order fields also in select list"));
          }
          //
          var numkeys = keyCols.length;
          var wcpart = [];
          var attpart = 0;
          for (var i = 0; i < keyCols.length; i++) {
            var col = this.schema.columns[keyCols[i].alias || keyCols[i].expr];
            //
            // I compose the where conditions for partialization
            for (var j = 0; j < numkeys - attpart; j++) {
              var wc = {expr: (j < numkeys - attpart - 1 ? "=" : ">"), p: []};
              wc.p.push(keyCols[col.index]);
              wc.p.push({expr: col.toSql(keyVals[col.index])});
              if (!wcpart[j])
                wcpart[j] = wc;
              else
                wcpart[j] = {p: [{expr: "("}, wcpart[j], {expr: "and"}, wc, {expr: ")"}]};
            }
            attpart++;
          }
          //
          if (wcpart.length > 0) {
            var wc = null;
            for (var i = 0; i < wcpart.length; i++) {
              if (i === 0)
                wc = wcpart[i];
              else {
                wc = {p: [{expr: "("}, wc, {expr: "or"}, wcpart[i], {expr: ")"}]};
              }
            }
            q.where = q.where || [];
            q.where.push(wc);
          }
        }
      }
      //
      // Set maxRows limit
      q.top = {expr: this.dataPageSize};
      //
      parameters = [];
      this.dataStore.executeSQL(this.dataStore.createSqlStatement(q, parameters), true, parameters, afterQuery);
      break;
    }
  }
};


/**
 * Detect if data of a page are loaded
 * @param {Integer} page
 */
App.DataMap.prototype.isPageLoaded = function (page)
{
  return (this.rows[page * this.dataPageSize] ? true : false);
};


/**
 * Load a page of data
 * @param {Integer} page to load
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.loadPage = function (page, callback)
{
  // If page is already loaded do nothing
  if (this.isPageLoaded(page))
    return callback(0);
  //
  // If datamap has not paging load throw error
  if (!this.dataPagingMode)
    return callback(null, new Error("Attempt to load a page without dataPagingMode"));
  //
  // If there is already underway loading a page postponement
  if (this.loadingPages.indexOf(page) >= 0)
    return setTimeout(function () {
      this.loadPage(page, callback);
    }.bind(this), 100);
  //
  // KeySet paging mode cannot load random page
  // If previous page is not already loaded load it
  if (this.dataPagingMode === App.DataMap.dataPagingModes.keyset && page > 0 && !this.isPageLoaded(page - 1)) {
    this.loadPage(page - 1, function (result, error) {
      if (error)
        callback(null, error);
      else if (result < this.dataPageSize)
        callback(0);
      else
        this.loadPage(page, callback);
    }.bind(this));
    return;
  }
  //
  // Load data of required page
  this.load({page: page}, function (result, error) {
    if (error)
      return callback(null, error);
    //
    // Detect how many rows are been inserted
    var i = 0;
    for (; i < this.dataPageSize; i++)
      if (!this.rows[(page * this.dataPageSize) + i])
        break;
    //
    callback(i);
  }.bind(this));
};


/**
 * Load a set of pages of data
 * @param {Integer} first page to load
 * @param {Integer} last page to load
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.loadPages = function (first, last, callback)
{
  var next = function (result, error) {
    if (error || first > last)
      callback(null, error);
    else
      this.loadPage(first++, next);
  }.bind(this);
  next();
};


/**
 * Load next page of data
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.loadNextPage = function (callback)
{
  if (!this.dataPagingMode)
    return callback(null, new Error("Attempt to load a page without dataPagingMode"));
  //
  var rowCount = this.rows.length;
  //
  // With precount the array of rows is filled with undefined
  // Detect position of first undefined
  if (this.dataPagingPreCount) {
    for (var i = 0; i < this.rows.length; i++)
      if (!this.rows[i]) {
        rowCount = i;
        break;
      }
  }
  //
  // If page is partially loaded I finished
  if (rowCount % this.dataPageSize > 0)
    return callback(0);
  //
  // Load page required
  this.loadPage(rowCount / this.dataPageSize, callback);
};


/**
 * Execute a query for detect rowCount
 * @param {Object/String} qry
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.calcRowCount = function (qry, callback)
{
  var parameters = [];
  var sql = "select count(*) as count from (";
  sql += (typeof qry === "string" ? qry : this.dataStore.createSqlStatement(qry, parameters));
  sql += ") a";
  //
  this.dataStore.executeSQL(sql, true, parameters, function (result, error) {
    if (error)
      callback(null, error);
    else
      callback(this.dataStore.getRow(result, 0).count);
  }.bind(this));
};


/**
 * Return a row loading page if needed
 * @param {Integer} index
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.getRow = function (index, callback)
{
  this.loadPage(Math.floor(index / this.dataPageSize), function (result, error) {
    callback(this.rows[index], error);
  }.bind(this));
};


/**
 * Load a collection of document with my data
 * @param {Function} cls
 * @param {Function} callback - function to be called at the end
 */
App.DataMap.prototype.toCollection = function (cls, callback)
{
  var coll = new App.Collection(this.app);
  coll.rowTemplate = new cls(this.app);
  //
  // Add a warning if missing any columns
  if (this.app) {
    var dmCols = Object.keys(this.schema.columns);
    var clCols = Object.keys(coll.rowTemplate.schema.columns);
    for (var c = 0; c < clCols.length; c++)
      if (dmCols.indexOf(clCols[c]) === -1 && !coll.rowTemplate.schema.columns[clCols[c]].unbound) {
        this.app.dttobj.addWarning("", "The datamap to load omits column " + clCols[c] + ". This property will not be valued");
        break;
      }
  }
  //
  coll.loadFromDataMap(this, function (result, error) {
    if (callback)
      callback(coll, error);
  });
  //
  return coll;
};


/*
 * Informs all rows that a document is changed
 * @param {Document} doc changed
 */
App.DataMap.prototype.handleDocUpdate = function (doc)
{
  for (var i = 0; i < this.rows.length; i++)
    this.rows[i].handleDocUpdate(doc);
};


/*
 * Update elements of each row
 */
App.DataMap.prototype.updateElements = function ()
{
  this.uiRows.forEach(function (r) {
    if (r)
      r.updateElement();
  }.bind(this));
};


/*
 * Update header of groups
 * @param {DataRow} row
 */
App.DataMap.prototype.updateHeaderElements = function (row)
{
  // Only if data are grouped
  if (!this.groupBy)
    return;
  //
  // Only if datamap has header template
  if (!this.schema.header)
    return;
  //
  // Create groups and headers
  for (var level = 0; level < this._groupBy.length; level++) {
    var group = row.group.getGroup(level);
    //
    if (this.schema.header) {
      // Update header (only one time when creating group)
      if (row === group.firstRow || !group.justCreated)
        group.updateElement("header");
      //
      // Reset flag justCreated
      if (row === group.lastRow && !this.schema.footer)
        delete group.justCreated;
    }
  }
};


/*
 * Update footer of groups
 * @param {DataRow} row
 */
App.DataMap.prototype.updateFooterElements = function (row)
{
  // Only if data are grouped
  if (!this.groupBy)
    return;
  //
  // Only if datamap has header or footer template
  if (!this.schema.footer)
    return;
  //
  var group = row.group;
  while (group !== this) {
    // Update footer (only one time when creating group)
    if (row === group.lastRow || !group.justCreated) {
      group.updateElement("footer");
      delete group.justCreated;
    }
    //
    // Next owner group
    group = group.owner;
  }
};


/*
 * Calculate groups for a row
 * @param {DataRow} row
 */
App.DataMap.prototype.calcRowGroups = function (row)
{
  if (!this.groupBy) {
    delete row.group;
    return;
  }
  //
  // Create groups
  var owner = this;
  for (var level = 0; level < this._groupBy.length; level++) {
    var lastLevel = (level === this._groupBy.length - 1);
    var col = Object.keys(this._groupBy[level])[0];
    var v = row[col];
    //
    if (!owner.groups)
      owner.groups = {};
    //
    var group = owner.groups[v + " "];
    if (!group) {
      // Create new group
      group = new App.DataGroup(owner);
      owner.groups[v + " "] = group;
      //
      group.level = level;
      group.column = col;
      group.value = v;
      group.justCreated = true;
    }
    //
    // If row was not inserted at the bottom find its position into group
    if (row.spliced) {
      var nextRow = this.uiRows[row.position + 1];
      if (nextRow && nextRow.group.getGroup(level) === group)
        group.rows.splice(group.rows.indexOf(nextRow), 0, row);
      else
        group.rows.push(row);
    }
    else
      group.rows.push(row);
    //
    // Set group of row
    if (lastLevel)
      row.group = group;
    //
    // Reuse old group elements
    if (row.inPagination()) {
      while (this.uiRecycleBin.length > 0) {
        var e = this.uiRecycleBin[0];
        if (!e)
          break;
        //
        var compare = 1;
        var oldGroup = e.group ? e.group.getGroup(level) : null;
        if (oldGroup)
          compare = oldGroup.compare(group);
        //
        // Old group of this element is at next position ... add element before
        if (compare > 0) {
          this.uiRecycleBin.splice(0, 0, null);
          break;
        }
        //
        // Old group of this element is at correct position
        if (compare === 0) {
          group.map = oldGroup.map;
          //
          // Reuse header element
          if (oldGroup.header) {
            group.header = oldGroup.header;
            delete oldGroup.header;
          }
          //
          // Reuse footer element
          if (oldGroup.footer) {
            group.footer = oldGroup.footer;
            delete oldGroup.footer;
          }
          //
          // At last level remove sub groups
          // At not last level remove old row elements
          if ((oldGroup.groups && lastLevel) || (!oldGroup.groups && !lastLevel)) {
            e.group.removeRow(e.row, true);
            var el = this.uiRecycleBin.shift().element;
            el.parent.removeChild(el);
            this.uiRecycleBin.splice(0, 0, null);
          }
          //
          break;
        }
        //
        // Old group of this element is at previous position ... destroy it
        if (compare < 0) {
          e.group.removeRow(e.row, true);
          var el = this.uiRecycleBin.shift().element;
          el.parent.removeChild(el);
        }
      }
    }
    //
    // Next child group
    owner = group;
  }
};


/*
 * Set timer for row's updateElement
 * @param {DataRow} row
 */
App.DataMap.prototype.setUETimeout = function (row)
{
  if (this.rowsToUpdate.length === 0)
    setTimeout(App.DataMap.prototype.updateElementRows.bind(this), 0);
  this.rowsToUpdate.push(row);
};


/*
 * Call updateElementTimer for some rows (almost for 5 ms)
 */
App.DataMap.prototype.updateElementRows = function ()
{
  var start = new Date();
  var n = this.rowsToUpdate.length;
  try {
    while (new Date() - start < 5 && n-- > 0)
      this.rowsToUpdate.shift().updateElementTimer();
  }
  catch (e) {
    throw e;
  }
  finally {
    if (n > 0)
      setTimeout(App.DataMap.prototype.updateElementRows.bind(this), 0);
  }
};


/*
 * Reset updateElementTimer for a row
 */
App.DataMap.prototype.clearUETimeout = function (row)
{
  this.rowsToUpdate.splice(this.rowsToUpdate.indexOf(row), 1);
  delete row.ueTimer;
};


//export module for node
if (module)
  module.exports = App.DataMap;
