/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, Client */

var App = App || {};

/**
 * @class A DataRow Object
 * @param {DataMap} parent - datamap owner
 * @param {Object} data
 * */
App.DataRow = function (parent, data)
{
  if (module && !App.View && parent !== true)
    App = require("./requireAll");
  //
  this.intProps = {};
  this.uiProps = {};
  //
  if (parent !== true) {
    if (parent instanceof App.DataMap) {
      this.uiProps.datamap = parent;
      this.intProps.app = parent.app;
    }
    else if (parent instanceof App.Session)
      this.intProps.app = parent;
  }
  //
  // Error handling
  this.intProps.rowError = "";      // Error text of the entire row
  this.intProps.columnsErrors = {}; // Error text of each column
  //
  // The object properties are the actual row values
  this.values = {};
  //
  if (data) {
    var okeys = Object.keys(data instanceof App.DataRow ? this.getSchema().columns : data);
    for (var i = 0; i < okeys.length; i++) {
      var p = okeys[i];
      this.assignValue(p, data[p]);
    }
  }
};

// Getter/Setter of some properties
Object.defineProperties(App.DataRow.prototype, {
  // Application
  app: {
    get: function () {
      return this.intProps.app;
    },
    set: function (newValue) {
      this.intProps.app = newValue;
    }
  },
  // Index of this row between visible rows
  position: {
    get: function () {
      if (this.datamap)
        return this.datamap.uiRows.indexOf(this);
      else
        return -1;
    }
  },
  // Index of this row between all rows
  index: {
    get: function () {
      if (this.datamap)
        return this.datamap.rows.indexOf(this);
      else
        return -1;
    }
  },
  // List template cloned for this row
  element: {
    get: function () {
      return this.uiProps.element;
    },
    set: function (newValue) {
      this.uiProps.element = newValue;
    }
  },
  // Map of cloned element from original list template
  map: {
    get: function () {
      return this.uiProps.map;
    },
    set: function (newValue) {
      this.uiProps.map = newValue;
    }
  },
  // Datamap owner
  datamap: {
    get: function () {
      return this.uiProps.datamap;
    },
    set: function (newValue) {
      this.uiProps.datamap = newValue;
    }
  },
  // Error text of the entire row
  rowError: {
    get: function () {
      return this.intProps.rowError;
    },
    set: function (newValue) {
      this.intProps.rowError = newValue;
    }
  },
  // Error text of each column
  columnsErrors: {
    get: function () {
      return this.intProps.columnsErrors;
    },
    set: function (newValue) {
      this.intProps.columnsErrors = newValue;
    }
  },
  // Original values as after load
  originalValues: {
    get: function () {
      return this.intProps.originalValues;
    },
    set: function (newValue) {
      this.intProps.originalValues = newValue;
    }
  },
  // Actual values
  values: {
    get: function () {
      return this.intProps.values;
    },
    set: function (newValue) {
      this.intProps.values = newValue;
    }
  },
  // Inserted status
  inserted: {
    get: function () {
      return (this.intProps.inserted === true);
    },
    set: function (newValue) {
      if (this.inserted !== newValue) {
        this.intProps.inserted = newValue;
        //
        // Update user interface
        if (!this.inserted)
          this.updateElement();
      }
    }
  },
  // Deleted status
  deleted: {
    get: function () {
      return (this.intProps.deleted === true);
    },
    set: function (newValue) {
      if (this.deleted !== newValue) {
        var wasVisible = this.isVisible();
        this.intProps.deleted = newValue;
        //
        // Update user interface
        if (wasVisible !== this.isVisible()) {
          if (newValue)
            this.datamap.updateElementsInvisibleRow(this);
          else {
            var options = {};
            options.position = this.datamap.findRowPosition(this);
            this.datamap.updateElementsVisibleRow(this, options);
          }
        }
      }
    }
  },
  // Updated status
  updated: {
    get: function () {
      return (this.intProps.originalValues !== undefined);
    },
    set: function (newValue) {
      if (this.updated !== newValue) {
        if (newValue)
          this.intProps.originalValues = this.intProps.originalValues || {};
        else
          this.intProps.originalValues = undefined;
        //
        // Update user interface
        if (!this.updated)
          this.updateElement();
      }
    }
  },
  // Hidden status
  hidden: {
    get: function () {
      return (this.uiProps.hidden === true);
    },
    set: function (newValue) {
      if (this.hidden !== newValue) {
        var wasVisible = this.isVisible();
        this.uiProps.hidden = newValue;
        //
        // Update user interface
        if (wasVisible !== this.isVisible()) {
          if (newValue)
            this.datamap.updateElementsInvisibleRow(this);
          else {
            var options = {};
            options.position = this.datamap.findRowPosition(this);
            this.datamap.updateElementsVisibleRow(this, options);
          }
        }
      }
    }
  },
  // Children datamap cloned for this row
  subDataMaps: {
    get: function () {
      return this.uiProps.subDataMaps;
    },
    set: function (newValue) {
      this.uiProps.subDataMaps = newValue;
    }
  }
});


/**
 * Setter of each column
 * @param {String} columnName
 * @param {AnyType} newValue
 * @param {Boolean} convert value to column datatype
 */
App.DataRow.prototype.setValue = function (columnName, newValue, convert) {
  // If new value is equal do nothing
  var oldValue = this[columnName];
  if (App.Locale.compareValues(newValue, oldValue))
    return;
  //
  var oldOrgValue = this.getOriginalValue(columnName);
  //
  // Set new value and original value
  this.assignValue(columnName, newValue, convert);
  this.setOriginalValue(columnName, oldOrgValue);
  //
  // Notify change value to datamap observer
  this.notifyChangeToParentObserver("update", columnName, oldValue);
  //
  // Update UI
  this.updateElement();
  this.datamap.updateElementsData(this);
};


/**
 * Assign a value to a column
 * @param {String} columnName
 * @param {AnyType} newValue
 * @param {Boolean} convert value to column datatype
 * */
App.DataRow.prototype.assignValue = function (columnName, newValue, convert)
{
  if (convert === undefined)
    convert = true;
  //
  var dc = this.getSchema().columns[columnName];
  if (dc) {
    if (convert && this.app)
      this.getValueContainer(dc)[columnName] = this.app.locale.convertValue(newValue, {dstDT: dc.datatype});
    else
      this.getValueContainer(dc)[columnName] = newValue;
  }
  else
    this[columnName] = newValue;
};


/**
 * Return the container of a value for a column
 * @param {DataColumn} dc
 */
App.DataRow.prototype.getValueContainer = function (dc)
{
  return this.values;
};


/**
 * Notify change value to datamap observer
 * @param{String} type
 * @param{String} name
 * @param{AnyType} oldValue
 */
App.DataRow.prototype.notifyChangeToParentObserver = function (type, name, oldValue)
{
  if (this.datamap && this.datamap.observed && this.datamap.row === this)
    Object.getNotifier(this.datamap).notify({type: type, name: "row." + name, oldValue: oldValue});
  if (this.observed)
    Object.getNotifier(this).notify({type: type, name: name, oldValue: oldValue});
};


/**
 * Set original value of a column
 * @param{String} columnName
 * @param{String} oldValue
 */
App.DataRow.prototype.setOriginalValue = function (columnName, oldValue)
{
  // If it is the first change to the row or column
  if (!App.Locale.compareValues(oldValue, this[columnName])) {
    // Change updated status and set original value
    this.updated = true;
    this.originalValues[columnName] = oldValue;
  }
  else if (this.updated) {
    // All values are like the original ones then reset the entire row in the original state
    delete this.originalValues[columnName];
    if (Object.keys(this.originalValues).length === 0)
      this.updated = false;
  }
};


/**
 * Get original value of a column
 * @param{String} columnName
 */
App.DataRow.prototype.getOriginalValue = function (columnName)
{
  if (this.originalValues && (columnName in this.originalValues))
    return this.originalValues[columnName];
  else
    return this[columnName];
};


/**
 * Close the datarow
 */
App.DataRow.prototype.close = function ()
{
  for (var i in this.subDataMaps)
    this.subDataMaps[i].close();
  this.subDataMaps = undefined;
  //
  // Delete any observer on me
  if (App.deleteObservers)
    App.deleteObservers(this);
};


/**
 * Create UI element for this row
 */
App.DataRow.prototype.createElement = function ()
{
  this.map = {};
  if (this.datamap.template) {
    this.element = this.datamap.template.clone(this.map);
    this.element.setDataRow(this);
  }
  //
  // Clone all childs of dataMap for this row
  this.datamap.cloneChildsForRow(this);
};


/**
 * Update UI
 * @param {boolean} resetTimer
 */
App.DataRow.prototype.updateElement = function (resetTimer)
{
  // For query by code no update
  if (!this.datamap.parentElement)
    return;
  //
  this.datamap.fireOnDataChange();
  //
  // For datamap without template update only current row
  if (!this.datamap.template && !this.datamap.groupBy && this.datamap.row !== this)
    return;
  //
  // If is already planned for this row do nothing
  if (this.ueTimer) {
    if (resetTimer)
      this.datamap.clearUETimeout(this);
    else
      return;
  }
  //
  if (!this.hasElement() && !this.inPagination() && this.isVisible()) {
    this.datamap.schema.listBox.moreRows = ((this.datamap.schema.listBox.moreRows || 0) % 100000) + 1;
    return; // DataMap paging is active and this element is beyond the page limit of the datamap
  }
  //
  this.ueTimer = true;
  this.datamap.setUETimeout(this);
};


/**
 * Update UI with timer
 */
App.DataRow.prototype.updateElementTimer = function ()
{
  delete this.ueTimer;
  //
  if (!this.hasElement() && this.datamap.useWindow)
    return; // the DM will handle better than me
  //
  // If the row is invisible does not need updated
  if (this.position === -1)
    return;
  //
  // Update header of groups
  this.datamap.updateHeaderElements(this);
  //
  // No window, any datarow should have its element
  var append = false;
  if (!this.hasElement()) {
    append = true;
    this.createElement();
  }
  //
  // Set error to row element
  if (this.hasElement() && (this.element.errorText || this.rowError))
    this.element.errorText = this.rowError || "";
  //
  // Send values to elements
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    var elements = dc.getElementsToUpdateForRow(this);
    for (var j = 0; j < elements.length; j++) {
      var oel = elements[j];
      var ael = oel.element;
      if (ael === oel.origElem)
        ael.setDataRow(this);
      //
      // Convert date to string
      var v = this[dc.name];
      if (v === undefined)
        v = null;
      else if ((v && typeof v === "object") || dc.datatype === App.Database.DataTypes.JSON)
        v = dc.valueToString(v, oel);
      //
      if (oel.style)
        ael.style[oel.prop] = v;
      else
        ael[oel.prop] = v;
      //
      // Set error on the column
      if (ael.errorText || this.columnsErrors[c])
        ael.errorText = this.columnsErrors[c] || "";
    }
  }
  //
  // Notify onRowComposition event
  // If the event is made asynchronous operations the element could change before the end of the event
  // To prevent the event code does not give any error in element references
  // when it is detached attach temporarily a fake element.
  // When last event finish destroy fake element
  this.uiProps.rowComp = (this.uiProps.rowComp || 0) + 1;
  this.datamap.onRowComposition(this, this.element, function (result, error) {
    if (--this.uiProps.rowComp === 0)
      delete this.uiProps.rowComp;
    //
    // Rethrow the error
    if (error)
      throw error;
  }.bind(this));
  //
  // Append my element to listBox
  if (append && this.hasElement())
    this.datamap.appendRowElement(this);
  //
  // Update footer of groups
  this.datamap.updateFooterElements(this);
};


/**
 * Delete UI element for this row
 */
App.DataRow.prototype.removeElement = function ()
{
  if (this.hasElement()) {
    this.element.parent.removeChild(this.element, true);
    this.element = null;
  }
};


/**
 * One of my element has changed
 * @param {RemElem} el - element changed
 * @param {String} n - property name
 * @param {String} v - property value
 */
App.DataRow.prototype.chgProp = function (el, n, v)
{
  // Original element
  var oids = [el.id];
  var dc = null;
  var oel = null;
  //
  // Search for original element in template
  for (var id in this.map) {
    if (this.map[id] === el)
      oids.push(id);
  }
  //
  // Search for datacolumn
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var d = columns[c];
    for (var i = 0; i < d.elements.length; i++) {
      var e = d.elements[i];
      if (oids.indexOf(e.element.id) >= 0 && e.prop === n) {
        dc = d;
        oel = e;
        break;
      }
    }
    if (dc)
      break;
  }
  //
  // Setting values
  if (dc) {
    var ov = this[dc.name];
    //
    // Convert from string to real value
    v = dc.stringToValue(v === "" ? null : v, oel);
    //
    // Reset errors or set to client error
    this.setError(el.errorText || "", dc.name);
    //
    // Set new value
    this.setValue(dc.name, v, false);
    //
    // Notify onValueChange event
    this.datamap.onValueChange(this, dc.name, ov, v);
    //
    return dc;
  }
};


/**
 * Attach a cached element to this row
 * @param {Object} cache cached element
 * @param {Int} rownum
 */
App.DataRow.prototype.attachElement = function (cache, rownum)
{
  this.element = cache.element;
  this.map = cache.map;
  this.element.setDataRow(this);
  //
  // Clone all childs of dataMap for this row
  this.datamap.cloneChildsForRow(this);
  //
  // Update elements of this row and all childs
  this.updateElement();
  for (var i in this.subDataMaps) {
    var dm = this.subDataMaps[i];
    for (var j in dm.uiRows)
      dm.uiRows[j].updateElement();
  }
  //
  if (cache.append) {
    this.datamap.appendRowElement(this);
    cache.append = false;
  }
  //
  this.element.visible = true;
  //
  if (rownum !== this.element.rownum) {
    this.element.rownum = rownum;
    // The list position of this datarow has changed.
    // If the app is offline, we push this modification
    // to the client side so it moves as soon as possible.
    // Remember: standard command notification implies at least
    // a setTimeout (updateElement) and a setImmediate (sendCommands)
    if (App.isOffline()) {
      var ele = Client.eleMap[this.element.id];
      if (ele)
        ele.setRownum(rownum);
      else
        this.app.sendCommands([{obj: this.element.id, id: "setRownum", cnt: rownum}]);
    }
  }
};


/**
 * Detach the cached element to this row
 * @param {bool} keepVisible
 */
App.DataRow.prototype.detachElement = function (keepVisible)
{
  var ris = {element: this.element, map: this.map};
  if (this.group) {
    ris.row = this;
    ris.group = this.group;
  }
  this.element.setDataRow(null);
  if (!keepVisible)
    this.element.visible = false;
  //
  if (this.uiProps.rowComp)
    this.createFakeElement();
  else {
    this.element = null;
    this.map = null;
  }
  //
  // I have to remove the elements of subdatamaps
  // why are recreated and old remain visible
  for (var i in this.subDataMaps) {
    var dm = this.subDataMaps[i];
    dm.uiRows.forEach(function (row) {
      if (row.hasElement()) {
        var r = row.detachElement();
        r.element.parent.removeChild(r.element);
      }
    });
  }
  //
  return ris;
};


/**
 * Check if element is a fake
 */
App.DataRow.prototype.isFakeElement = function ()
{
  return (this.element && this.element.id === "fake");
};


/**
 * Check if row has an element (not a fake)
 */
App.DataRow.prototype.hasElement = function ()
{
  return (this.element && !this.isFakeElement());
};


/**
 * Attach a fake element
 */
App.DataRow.prototype.createFakeElement = function ()
{
  // Create a fake clone of actual element
  this.map = {};
  this.element = this.element.clone(this.map);
  this.element.id = "fake";
  //
  // Destroy immediately the fake
  this.element.close();
  //
  // Relink template of subdatamaps
  this.subDataMaps.forEach(function (dm) {
    if (dm.template)
      dm.template = this.map[dm.template.id];
  }.bind(this));
};


/**
 * Clear all error of row and columns
 */
App.DataRow.prototype.clearErrors = function ()
{
  this.setError("");
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    this.setError("", dc.name);
  }
};


/**
 * Set row or column error
 * @param {String} text - the text of the error
 * @param {String} columnName - the name of the column in error
 */
App.DataRow.prototype.setError = function (text, columnName)
{
  // Only if is changed
  if (this.getError(columnName) === text)
    return;
  //
  // If you did not specify the column set the error on the row
  if (!columnName)
    this.rowError = text;
  else {
    if (text)
      this.columnsErrors[columnName] = text;
    else
      delete this.columnsErrors[columnName];
  }
  this.updateElement();
  //
  // Notify onError to datamap
  if (text) {
    var info = {};
    info.source = App.DataMap.errorSources.validation;
    info.row = this;
    if (columnName)
      info.column = columnName;
    this.datamap.onError(text, info);
  }
};


/**
 * Check if there are errors in row and/or columns
 */
App.DataRow.prototype.hasErrors = function ()
{
  if (this.rowError)
    return true;
  else
    return Object.keys(this.columnsErrors).length > 0;
};


/**
 * Return error of row or of a column
 * @param{String} columnName
 */
App.DataRow.prototype.getError = function (columnName)
{
  return (columnName ? this.columnsErrors[columnName] : this.rowError) || "";
};


/**
 * Return all error of row and of columns
 */
App.DataRow.prototype.getErrors = function ()
{
  var errors = [];
  if (this.rowError)
    errors.push(this.rowError);
  for (var c in this.columnsErrors)
    errors.push(this.columnsErrors[c]);
  return errors;
};


/**
 * Return query for update data of this row
 */
App.DataRow.prototype.getUpdateQuery = function ()
{
  var q = {
    type: App.Database.QueryTypes.UPDATE,
    table: this.datamap.query.from[0].expr,
    select: [],
    where: []
  };
  //
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    //
    // Skip not master column
    if (!dc.isMaster())
      continue;
    //
    var newValue = this[dc.name];
    if (dc.name in this.originalValues) {
      // BLOB as parameter
      if (dc.datatype === App.Database.DataTypes.BLOB)
        q.select.push({field: dc.name, value: newValue, asPar: true, dt: dc.datatype});
      else
        q.select.push({field: dc.name, expr: dc.toSql(newValue)});
    }
    //
    if (dc.pk || (this.datamap.optimisticLocking && dc.isComparable())) {
      var orgValue = this.getOriginalValue(dc.name);
      if (orgValue === null || orgValue === undefined)
        q.where.push({expr: "is null", p: [{expr: dc.name, from: ""}]});
      else
        q.where.push({expr: "=", p: [{expr: dc.name, from: ""}, {expr: dc.toSql(orgValue)}]});
    }
  }
  //
  if (q.select.length > 0)
    return q;
  else
    return null;
};


/**
 * Return query for delete this row
 */
App.DataRow.prototype.getDeleteQuery = function ()
{
  var q = {
    type: App.Database.QueryTypes.DELETE,
    table: this.datamap.query.from[0].expr,
    where: []
  };
  //
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    //
    // Skip not master column
    if (!dc.isMaster())
      continue;
    //
    if (dc.pk || (this.datamap.optimisticLocking && dc.isComparable())) {
      var orgValue = this.getOriginalValue(dc.name);
      if (orgValue === null || orgValue === undefined)
        q.where.push({expr: "is null", p: [{expr: dc.name, from: ""}]});
      else
        q.where.push({expr: "=", p: [{expr: dc.name, from: ""},
            {expr: dc.toSql(orgValue)}]});
    }
  }
  return q;
};


/**
 * Return query for insert this row
 */
App.DataRow.prototype.getInsertQuery = function ()
{
  var q = {
    type: App.Database.QueryTypes.INSERT,
    table: this.datamap.query.from[0].expr,
    select: []
  };
  //
  if (this.datamap.upsert)
    q.upsert = true;
  //
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    //
    // Skip not master column
    if (!dc.isMaster())
      continue;
    //
    // Set counter column (for Postgres)
    if (dc.datatype === App.Database.DataTypes.AUTOINCR)
      q.counter = dc.name;
    //
    // Skip omissis values
    var value = this[dc.name];
    if (value === undefined)
      continue;
    //
    // BLOB as parameter
    if (dc.datatype === App.Database.DataTypes.BLOB)
      q.select.push({field: dc.name, value: value, asPar: true, dt: dc.datatype});
    else
      q.select.push({field: dc.name, expr: dc.toSql(value)});
  }
  //
  if (q.select.length > 0)
    return q;
  else
    return null;
};


/**
 * Return a column value
 * @param{String} columnName
 */
App.DataRow.prototype.getColumn = function (columnName)
{
  return this[columnName];
};


/**
 * Return true if the column is unbound or unknown
 * @param{String} columnName
 */
App.DataRow.prototype.isColumnUnbound = function (columnName)
{
  var dc = this.getSchema().columns[columnName];
  return !dc || dc.unbound;
};


/**
 * Return the schema of this row
 */
App.DataRow.prototype.getSchema = function ()
{
  return this.datamap.schema;
};


/**
 * Init default values
 * @param {bool} all - all columns or unbound only
 */
App.DataRow.prototype.initDefaults = function (all)
{
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    if ((all || dc.unbound) && this[c] === undefined) {
      if (dc.default !== undefined)
        this[c] = dc.default;
      else if (dc.pk && dc.datatype === App.Database.DataTypes.UUID) {
        if (App.Utils["generateUID" + dc.maxlength])
          this[c] = App.Utils["generateUID" + dc.maxlength]();
      }
    }
  }
};


/**
 * Reset this row to original state
 * */
App.DataRow.prototype.restoreOriginal = function ()
{
  if (this.inserted)
    this.datamap.remove(this);
  else {
    if (this.deleted)
      this.deleted = false;
    //
    if (this.updated) {
      var columns = this.getSchema().columns;
      for (var c in columns)
        this.setValue(c, this.getOriginalValue(c));
      //
      this.updated = false;
      this.clearErrors();
    }
  }
};


/**
 * Check if row is modified
 * */
App.DataRow.prototype.isModified = function ()
{
  if (this.inserted && this.deleted)
    return false;
  //
  return (this.updated || this.inserted || this.deleted);
};


/**
 * Save to JSON
 */
App.DataRow.prototype.toJSON = function ()
{
  var json = [];
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    var v = this[dc.name];
    if ([App.Database.DataTypeClasses.DATETIME, App.Database.DataTypeClasses.BLOB].indexOf(App.Database.classDT(dc.datatype)) >= 0)
      json[dc.index] = dc.valueToString(v);
    else
      json[dc.index] = v;
  }
  return json;
};


/**
 * Load from JSON
 * @param{Object} json
 */
App.DataRow.prototype.loadFromJSON = function (json)
{
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    this.assignValue(dc.name, json[dc.index]);
  }
};


/**
 * Tells something about this object to the DTT module
 */
App.DataRow.prototype.getDebugInfo = function ()
{
  var info = {};
  info._class = "DataRow";
  info.position = this.position;
  info.inserted = this.inserted;
  info.deleted = this.deleted;
  info.updated = this.updated;
  info.hidden = this.hidden;
  var columns = this.getSchema().columns;
  for (var c in columns) {
    var dc = columns[c];
    info[dc.name] = App.DTT.convertValue(this[dc.name], true);
  }
  return info;
};


/**
 * Rows is visible?
 * */
App.DataRow.prototype.isVisible = function ()
{
  return (!this.hidden && !this.deleted);
};


/*
 * Informs all sub-datamaps that a document is changed
 * @param {Document} doc changed
 */
App.DataRow.prototype.handleDocUpdate = function (doc)
{
  for (var i in this.subDataMaps)
    this.subDataMaps[i].handleDocUpdate(doc);
};


/*
 * Returns true if row is before current pageLimit
 */
App.DataRow.prototype.inPagination = function ()
{
  if (!this.datamap.pageSize)
    return true;
  //
  var pos = this.position;
  return (pos >= 0 && pos < this.datamap.currentPageSize);
};


//export module for node
if (module)
  module.exports = App.DataRow;
