/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

var App = App || {};


/**
 * @class A DataSchema Object
 * @param {Object} dm - datamap owner
 * @param {Object} defprop - definition of the schema
 * */
App.DataSchema = function (dm, defprop)
{
  if (module && !App.View && dm !== true)
    App = require("./requireAll");
  //
  this.template = null;
  this.parent = (dm !== true ? dm : null);
  this.columns = {};
  this.listBox = null;
  //
  // Load default properties
  if (defprop) {
    for (var x in defprop) {
      if (x === "p") {
        for (var i = 0; i < defprop.p.length; i++)
          this.addColumn(defprop.p[i]);
      }
      else
        this[x] = defprop[x];
    }
  }
};


/**
 * Add a column to schema
 * @param {Object} c
 */
App.DataSchema.prototype.addColumn = function (c)
{
  var nc = new App.DataColumn(this, c);
  nc.index = Object.keys(this.columns).length;
  this.columns[c.name] = nc;
  return nc;
};


/**
 * Changes group header template
 * @param {RemElem} nt
 */
App.DataSchema.prototype.setHeaderTemplate = function (nt)
{
  this.header = nt;
  //
  // Remove group header template without destroying it
  if (this.header) {
    // Before removing the template we must remember its position, so we can do the correct insert
    var idx = this.header.parent.elements.indexOf(this.header);
    if (idx >= 0 && this.header.parent.elements.length > idx + 1)
      this.headerSiblingID = this.header.parent.elements[idx + 1].id;
    this.header.parent.removeChild(this.header);
  }
};


/**
 * Changes group footer template
 * @param {RemElem} nt
 */
App.DataSchema.prototype.setFooterTemplate = function (nt)
{
  this.footer = nt;
  //
  // Remove group footer template without destroying it
  if (this.footer) {
    // Before removing the template we must remember its position, so we can do the correct insert
    var idx = this.footer.parent.elements.indexOf(this.footer);
    if (idx >= 0 && this.footer.parent.elements.length > idx + 1)
      this.footerSiblingID = this.footer.parent.elements[idx + 1].id;
    this.footer.parent.removeChild(this.footer);
  }
};


/**
 * Changes template, reconnect elements by name
 * @param {RemElem} nt
 */
App.DataSchema.prototype.setTemplate = function (nt)
{
  this.template = nt;
  if (nt)
    this.setListBox((this.header && nt.isChildOf(this.header) ? this.header : nt).parent);
};


/**
 * Changes template, reconnect elements by name
 * @param {RemElem} lb
 */
App.DataSchema.prototype.setListBox = function (lb)
{
  this.listBox = lb;
  //
  // Set dataMap as owner of this listTemplate
  this.listBox.dataMap = this.parent;
  //
  // If DM uses record window, tell client it
  if (this.parent.useWindow) {
    lb.rowCount = this.parent.uiRows.length;
    lb.onMissingRecord = function (ev) {
      this.parent.onMissingRecord(ev.start, ev.end);
    }.bind(this);
    lb.onNavigate = function (ev) {
      this.parent.onNavigate(ev.cmd, ev.columns, ev.rows, ev.pos);
    }.bind(this);
  }
  else {
    // Before removing the template we must remember its position, so we can do the correct insert
    var idx = this.template.parent.elements.indexOf(this.template);
    if (idx >= 0 && this.template.parent.elements.length > (idx + 1))
      this.templateSiblingID = this.template.parent.elements[idx + 1].id;
    //
    // Remove template without destroying it
    this.template.parent.removeChild(this.template);
    //
    // Register next page event handler to get paging requests
    lb.onNextPage = function (ev) {
      this.parent.nextPage(ev);
    }.bind(this);
  }
};


/**
 * Get columns
 * @param {Boolean} asArray
 * @returns {Object/Array}
 */
App.DataSchema.prototype.getColumns = function (asArray)
{
  if (asArray) {
    var cols = [];
    for (var c in this.columns) {
      var col = this.columns[c];
      cols[col.index] = col;
    }
    return cols;
  }
  else
    return this.columns;
};


/**
 * Return column of type autoincrement
 * @returns {DataColumn}
 */
App.DataSchema.prototype.getAutoincrementColumn = function ()
{
  for (var c in this.columns) {
    if (this.columns[c].datatype === App.Database.DataTypes.AUTOINCR)
      return this.columns[c];
  }
  return null;
};


/**
 * Return a copy of this object
 * @returns {DataSchema}
 */
App.DataSchema.prototype.clone = function ()
{
  var sc = new App.DataSchema(this.parent);
  var keys = Object.keys(this);
  for (var k in keys) {
    var v = keys[k];
    if (v === "columns") {
      for (var c in this.columns) {
        var nc = this.columns[c].clone();
        sc.columns[c] = nc;
        nc.parent = sc;
      }
    }
    else
      sc[v] = this[v];
  }
  //
  return sc;
};


/**
 * Save to JSON
 */
App.DataSchema.prototype.toJSON = function ()
{
  var json = {cols: []};
  for (var c in this.columns) {
    var col = this.columns[c];
    json.cols[col.index] = col.toJSON();
  }
  return json;
};


/**
 * Load from JSON
 * @param{Object} json
 */
App.DataSchema.prototype.loadFromJSON = function (json)
{
  this.columns = {};
  for (var i = 0; i < json.cols.length; i++) {
    var col = new App.DataColumn(this);
    col.index = i;
    col.loadFromJSON(json.cols[i]);
    this.columns[col.name] = col;
  }
};


//export module for node
if (module)
  module.exports = App.DataSchema;
