/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, process */

var App = App || {};

/**
 * @class Definition of device
 * @param {Session} app that owns this device instance
 * */
App.Device = function (app)
{
  if (module && !App.View)
    App = require("./requireAll");
  this.app = app;       // App that owns this device instance
  this.viewport = null; // Object inited in App.Session.prototype.initSession
  //
  this.id = "device-ui"; // used for calling setProp, onChange
  this.app.eleMap[this.id] = this;
  //
  this.states = {};
  //
  // plug-in initialization
  this.barcodeScanner = new App.Device.BarcodeScanner(this);
  this.camera = new App.Device.Camera(this);
  this.vibration = new App.Device.Vibration(this);
  this.accelerometer = new App.Device.Accelerometer(this);
  this.compass = new App.Device.Compass(this);
  this.geolocation = new App.Device.Geolocation(this);
  this.notification = new App.Device.Notification(this);
  this.statusBar = new App.Device.StatusBar(this);
  this.contacts = new App.Device.Contacts(this);
  this.nfc = new App.Device.Nfc(this);
  this.lscookies = new App.Device.LSCookies(this);
  this.appManager = new App.Device.AppManager(this);
  this.keyboard = new App.Device.Keyboard(this);
  //
  this.viewport = {devicetype: "none"}; //default
  this._isMobile = undefined;
  this._isInsideShell = undefined;
  //
  // TODO: remove after upgrading shell in app stores
  this.lscookiesNotYetAsked = true;
};


//
// class properties
Object.defineProperties(App.Device.prototype, {
  isMobile: {
    get: function () {
      return this._isMobile;
    },
    set: function (obj) {
      return undefined;
    }
  },
  isInShell: {
    get: function () {
      return this._isInsideShell;
    },
    set: function (obj) {
      return undefined;
    }
  }
});


/*
 * Complete the plugin initalization
 */
App.Device.prototype.initPlugins = function ()
{
  // Notification.push is a server function, non a remote one
  if (this.notification && this.notification.push) {
    this.notification.push = App.Device.Notification.prototype.push;
    this.notification.alertCB = App.Device.Notification.prototype.alertCB;
  }
  if (this.camera) {
    this.camera.onPreUpload = App.Device.Camera.prototype.onPreUpload;
  }
};


/*
 * Returns true ore false depending whether we are running inside the shell
 */
App.Device.prototype.isInsideShell = function () {
  // double not casts value to strict true/false
  return !!this.isInShell;
};


/**
 * Set the device properties
 * @param {Object} p - device properties
 */
App.Device.prototype.setProp = function (p)
{
  // TODO: remove after releases the new shell to app stores
  if (this.lscookiesNotYetAsked && this.lscookies && p &&
          (App.isOffline() || p.operatingSystem === "ios")) {
    this.lscookies.getCookies();
    this.lscookiesNotYetAsked = false;
  }
  //
  var obj = this;
  if (p.plugin) {
    obj = this.app.eleMap["device-" + p.plugin];
    delete p.plugin;
  }
  //
  // Init locale object
  if (p.timeZone) {
    this.app.locale.timeZone = p.timeZone;
    this.app.locale.setLanguage(p.language);
    delete p.timeZone;
  }
  //
  // Init language using the first two characters of the language code, ie. "en" for english
  if (p.language && !this.app.langCode) {
    this.app.langCode = p.language.substring(0, 2);
    //
    if (!App.i18m[this.app.langCode] && App.i18m.defaultLangCode)
      this.app.langCode = App.i18m.defaultLangCode;
  }
  //
  if (obj) {
    var keys = Object.keys(p);
    for (var x in keys)
      obj[keys[x]] = p[keys[x]];
  }
  //
  // Create App.fs of ShellDriver type just I know to be inside shell
  if (!module && this.isInsideShell() && !(App.fs instanceof App.ShellDriver))
    App.fs = App.FS.createFS({app: this.app});
  //
  // If I'm inside a launcher, append launcher details into App.info structure
  if (!module && this.isInsideShell() && App.info) {
    if (p.launcherID)
      App.info.launcherID = p.launcherID;
    if (p.launcherName)
      App.info.launcherName = p.launcherName;
  }
  //
  // Update theme platform
  if (this.isMobile && this.app && this.app.theme && this.app.theme.ionPlatform)
    this.app.theme.ionPlatform = this.operatingSystem === "ios" ? "ios" : "md";
};


/**
 * Lock device orientation
 * @param {string} orientation
 **/
App.Device.prototype.lockOrientation = function (orientation)
{
  this.app.sendCommands([{obj: "device-device",
      id: "lockOrientation",
      cnt: orientation
    }]);
};


/*
 * Unlock device orientation
 */
App.Device.prototype.unlockOrientation = function ()
{
  this.app.sendCommands([{obj: "device-device",
      id: "unlockOrientation"
    }]);
};


/**
 * Prevent or allow standb
 * @param {boolean} status
 **/
App.Device.prototype.keepAwake = function (status)
{
  this.app.sendCommands([{obj: "device-device",
      id: "keepAwake",
      cnt: status
    }]);
};


/**
 * Send the viewport params
 * Called from App.Session.prototype.defineStyle
 * @param {string} origin
 * @param {boolean} physical
 */
App.Device.prototype.sendViewport = function (origin, physical)
{
  // Viewport send from not an origin and without physical,
  // its a new client for our session. If we have a phClient connected we set physical to true,
  // so the client preview adapts to the device
  if (!origin && !physical && App.phDeviceConnected)
    physical = true;
  //
  this.app.sendCommands([{obj: "device-ui",
      id: "setViewport",
      cnt: {params: [this.viewport.devicetype,
          this.viewport.orientation,
          origin || "server",
          physical || false
        ]}
    }]);
};


/**
 * Set the viewport params
 * @param {String} devicetype - none, desktop, tablet, smartphone
 * @param {String} orientation - horizontal, vertical
 * @param {String} origin - View.id from the device where the change is happening (or server if the server is initializing device viewports)
 * @param {boolean} physical
 */
App.Device.prototype.setViewport = function (devicetype, orientation, origin, physical)
{
  if (!App.phDeviceConnected                          // no physical device connected
          || App.phDeviceConnected !== "" && physical // physical device is asking for
          || origin === "server") {                   // server is asking for
    //
    if (devicetype)
      this.viewport.devicetype = devicetype;
    if (orientation)
      this.viewport.orientation = orientation;
    this.sendViewport(origin, physical);
  }
};


/*
 * Event handler fired when a App.Device property is changed
 * @param {Object} properties
 * @returns {undefined}
 */
App.Device.prototype.onChange = function (properties) // jshint ignore:line
{
};


/*
 * Go through all the states and reapply them
 * @returns {undefined}
 */
App.Device.prototype.readStates = function ()
{
  var cmd;
  for (var i in this.states) {
    cmd = this.states[i];
    if (cmd && cmd !== "") {
      this.app.eleMap[cmd.obj][cmd.id].apply(this.app.eleMap[cmd.obj], cmd.args);
    }
  }
};


/*
 * When doing multi/telecollaboration with a physical device, we only start the app inside the device first.
 * Then we open it on browsers, when the device is already connected (and hence master)
 * @returns {undefined}
 */
App.Device.prototype.startGuests = function ()
{
  // for telecollaboration. it only works online
  if (process)
    process.send({type: App.msgTypeMap.deviceMsg, cnt: {type: "startGuests", uuid: this.uuid}});
};


/**
 * Generate a new callback id
 * @param {Function} cb
 */
App.Device.prototype.gencb = function (cb)
{
  if (!cb)
    return null;
  //
  var id = App.Utils.generateUID24();
  if (!this.cbmap)
    this.cbmap = {};
  this.cbmap[id] = cb;
  return id;
};


/**
 * Called when a shell function returns
 * @param {Object} response
 */
App.Device.prototype.cb = function (response)
{
  var cb = this.cbmap[response.cbId];
  if (cb) {
    cb(response.result, response.error);
    delete this.cbmap[response.cbId];
  }
};


//export module for node
if (module)
  module.exports = App.Device;


/*
 * Verifies if the device is a mobile one (smartphone/tablet)
 * @returns {bool}
 */
App.Device.prototype.isMobileDevice = function (cb)
{
  if (!this.deviceType)
    throw "device not ready";

  if (cb)
    cb((this.deviceType in {"smartphone": 1, "tablet": 1}));
};


/**
 * Barcode scanner initialization
 * @param {Object} device
 */
App.Device.BarcodeScanner = function (device)
{
  this.device = device;
  this.id = "device-barcodeScanner";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.BarcodeScanner.prototype.scan = function (cb)
{
  // Scan already initiated?
  if (this.cb)
    throw "error: already scanning";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "scan"}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.BarcodeScanner.prototype.scanCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/**
 * Vibration initialization
 * @param {Object} device
 */
App.Device.Vibration = function (device)
{
  this.device = device;
  this.id = "device-vibration";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @param {Number} time
 * @returns {undefined}
 */
App.Device.Vibration.prototype.vibrate = function (time)
{
  // there's no callback
  this.device.app.sendCommands([{obj: this.id, id: "vibrate", cnt: time}]);
};


/**
 * Network initialization
 * @param {Object} device
 */
App.Device.Network = function (device)
{
  this.device = device;
  this.id = "device-network";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @returns {undefined}
 */
App.Device.Network.prototype.getNetworkStatus = function ()
{
  // there's no callback
  this.device.app.sendCommands([{obj: this.id, id: "getNetworkStatus"}]);
};


/**
 * Accelerometer initialization
 * @param {Object} device
 */
App.Device.Accelerometer = function (device)
{
  this.device = device;
  this.id = "device-accelerometer";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.Accelerometer.prototype.getCurrentAcceleration = function (cb)
{
  // Measurement already initiated?
  if (this.cb)
    throw "error: already measuring acceleration";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "getCurrentAcceleration"}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Accelerometer.prototype.getCurrentAccelerationCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Adds a "watch", that fires onAcceleration at given intervals
 * @param {type} frequency
 * @returns {undefined}
 */
App.Device.Accelerometer.prototype.watchAcceleration = function (frequency)
{
  var cmd = {obj: this.id, id: "watchAcceleration", cnt: {frequency: frequency}};
  var watch = {obj: cmd.obj, id: cmd.id, args: [frequency]};
  this.device.states.watchAcceleration = watch;
  //This sets an event listener
  this.device.app.sendCommands([cmd]);
};


/*
 * Event fired with the accelerometer data
 * @param {type} acceleration
 * @returns {undefined}
 */
App.Device.Accelerometer.prototype.onAcceleration = function (acceleration) // jshint ignore:line
{
};


/*
 * Removes the listening watch
 * @returns {undefined}
 */
App.Device.Accelerometer.prototype.clearWatch = function ()
{
  delete this.device.states.watchAcceleration;
  // removes the event listener
  this.device.app.sendCommands([{obj: this.id, id: "clearWatch"}]);
};


/**
 * Compass initialization
 * @param {Object} device
 */
App.Device.Compass = function (device)
{
  this.device = device;
  this.id = "device-compass";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.Compass.prototype.getCurrentHeading = function (cb)
{
  // Measurement already initiated?
  if (this.cb)
    throw "error: already measuring heading";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "getCurrentHeading"}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Compass.prototype.getCurrentHeadingCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Adds a "watch", that fires onHeading at given intervals
 * @param {type} frequency
 * @returns {undefined}
 */
App.Device.Compass.prototype.watchHeading = function (frequency)
{
  var cmd = {obj: this.id, id: "watchHeading", cnt: {frequency: frequency}};
  var watch = {obj: cmd.obj, id: cmd.id, args: [frequency]};
  this.device.states.watchHeading = watch;
  //This sets an event listener
  this.device.app.sendCommands([cmd]);
};


/*
 * Event fired with magnetic heading position
 * @param {type} heading
 * @returns {undefined}
 */
App.Device.Compass.prototype.onHeading = function (heading) // jshint ignore:line
{
};


/*
 * Removes the listening watch
 * @returns {undefined}
 */
App.Device.Compass.prototype.clearWatch = function ()
{
  delete this.device.states.watchHeading;
  // removes the event listener
  this.device.app.sendCommands([{obj: this.id, id: "clearWatch"}]);
};


/**
 * Geolocation initialization
 * @param {Object} device
 */
App.Device.Geolocation = function (device)
{
  this.device = device;
  this.id = "device-geolocation";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.Geolocation.prototype.getCurrentPosition = function (options, cb)
{
  // Measurement already initiated?
  if (this.cb)
    throw "error: already measuring location";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "getCurrentPosition", cnt: options}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Geolocation.prototype.getCurrentPositionCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Adds a "watch", that fires onPosition
 * @param {type} params
 * @returns {undefined}
 */
App.Device.Geolocation.prototype.watchPosition = function (params)
{
  var cmd = {obj: this.id, id: "watchPosition", cnt: params};
  var watch = {obj: cmd.obj, id: cmd.id, args: [params]};
  this.device.states.watchPosition = watch;
  //This sets an event listener
  this.device.app.sendCommands([cmd]);
};


/*
 * Event fired when new position data is available
 * @param {type} coords
 * @returns {undefined}
 */
App.Device.Geolocation.prototype.onPosition = function (coords) // jshint ignore:line
{
};


/*
 * Removes the listening watch
 * @returns {undefined}
 */
App.Device.Geolocation.prototype.clearWatch = function ()
{
//  this.watch.enabled = false;
  delete this.device.states.watchPosition;
  // removes the event listener
  this.device.app.sendCommands([{obj: this.id, id: "clearWatch"}]);
};


/**
 * Notification initialization
 * @param {Object} device
 */
App.Device.Notification = function (device)
{
  this.device = device;
  this.id = "device-notification";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.Notification.prototype.alert = function (message, title, buttonName, cb)
{
  if (this.cb)
    throw "error: already notifying";
  //
  this.cb = cb;
  //
  // native alert, confirm, prompt accept only strings
  if (typeof message !== "undefined" && typeof message !== "string") {
    try {
      message = JSON.stringify(message);
    }
    catch (ex) {
      message = "object";
    }
  }
  //
  if (typeof title !== "undefined" && typeof title !== "string") {
    try {
      title = JSON.stringify(title);
    }
    catch (ex) {
      title = "object";
    }
  }
  //
  if (typeof buttonName !== "undefined" && typeof buttonName !== "string") {
    try {
      buttonName = JSON.stringify(buttonName);
    }
    catch (ex) {
      buttonName = "object";
    }
  }
  //
  this.device.app.sendCommands([{obj: this.id, id: "alert", cnt: {message: message, title: title, buttonName: buttonName}}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Notification.prototype.alertCB = function (res)
{
  // since this.cb is shared between alert, confirm and prompt
  // if we have one in callback to the other, we move it and delete it before calling it
  if (this.cb) {
    var cb = this.cb;
    delete this.cb;
    cb(res);
  }
};


/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.Notification.prototype.confirm = function (message, title, buttonLabels, cb)
{
  if (this.cb)
    throw "error: already notifying";
  //
  this.cb = cb;
  //
  // native alert, confirm, prompt accept only strings
  if (typeof message !== "undefined" && typeof message !== "string") {
    try {
      message = JSON.stringify(message);
    }
    catch (ex) {
      message = "object";
    }
  }
  //
  if (typeof title !== "undefined" && typeof title !== "string") {
    try {
      title = JSON.stringify(title);
    }
    catch (ex) {
      title = "object";
    }
  }
  //
  if (typeof buttonLabels !== "undefined") {
    for (var i = 0; i < buttonLabels.length; i++) {
      if (typeof buttonLabels[i] !== "string") {
        try {
          buttonLabels[i] = JSON.stringify(buttonLabels[i]);
        }
        catch (ex) {
          buttonLabels[i] = "object";
        }
      }
    }
  }
  //
  this.device.app.sendCommands([{obj: this.id, id: "confirm", cnt: {message: message, title: title, buttonLabels: buttonLabels}}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Notification.prototype.confirmCB = function (res)
{
  // since this.cb is shared between alert, confirm and prompt
  // if we have one in callback to the other, we move it and delete it before calling it
  if (this.cb) {
    var cb = this.cb;
    delete this.cb;
    cb(res);
  }
  //
  this.device.app.sendCommands([{id: "closePopup"}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Notification.prototype.alertCB = function (res)
{
  // since this.cb is shared between alert, confirm and prompt
  // if we have one in callback to the other, we move it and delete it before calling it
  if (this.cb) {
    var cb = this.cb;
    delete this.cb;
    cb(res);
  }
  //
  this.device.app.sendCommands([{id: "closePopup"}]);
};


/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.Notification.prototype.prompt = function (message, title, buttonLabels, defaultText, cb)
{
  if (this.cb)
    throw "error: already notifying";
  //
  this.cb = cb;
  //
  // native alert, confirm, prompt accept only strings
  if (typeof message !== "undefined" && typeof message !== "string") {
    try {
      message = JSON.stringify(message);
    }
    catch (ex) {
      message = "object";
    }
  }
  //
  if (typeof title !== "undefined" && typeof title !== "string") {
    try {
      title = JSON.stringify(title);
    }
    catch (ex) {
      title = "object";
    }
  }
  //
  if (typeof buttonLabels !== "undefined") {
    for (var i = 0; i < buttonLabels.length; i++) {
      if (typeof buttonLabels[i] !== "string") {
        try {
          buttonLabels[i] = JSON.stringify(buttonLabels[i]);
        }
        catch (ex) {
          buttonLabels[i] = "object";
        }
      }
    }
  }
  //
  if (typeof defaultText !== "undefined" && typeof defaultText !== "string") {
    try {
      defaultText = JSON.stringify(defaultText);
    }
    catch (ex) {
      defaultText = "object";
    }
  }
  //
  this.device.app.sendCommands([{obj: this.id, id: "prompt", cnt: {message: message, title: title, buttonLabels: buttonLabels, defaultText: defaultText}}]); // jshint ignore:line
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Notification.prototype.promptCB = function (res)
{
  // since this.cb is shared between alert, confirm and prompt
  // if we have one in callback to the other, we move it and delete it before calling it
  if (this.cb) {
    var cb = this.cb;
    delete this.cb;
    cb(res);
  }
  //
  this.device.app.sendCommands([{id: "closePopup"}]);
};


/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @param {Number} number
 * @returns {undefined}
 */
App.Device.Notification.prototype.beep = function (number)
{
  // There is no callback
  this.device.app.sendCommands([{obj: this.id, id: "beep", cnt: number}]);
};

/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @param {Object} notification
 * @returns {undefined}
 */
App.Device.Notification.prototype.schedule = function (notification)
{
  // There is no callback
  this.device.app.sendCommands([{obj: this.id, id: "schedule", cnt: notification}]);
};

/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @param {Object} notification
 * @returns {undefined}
 */
App.Device.Notification.prototype.update = function (notification)
{
  // There is no callback
  this.device.app.sendCommands([{obj: this.id, id: "update", cnt: notification}]);
};

/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @param {Object} notification
 * @returns {undefined}
 */
App.Device.Notification.prototype.clear = function (notification)
{
  // There is no callback
  this.device.app.sendCommands([{obj: this.id, id: "clear", cnt: notification}]);
};

/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @returns {undefined}
 */
App.Device.Notification.prototype.clearAll = function ()
{
  // There is no callback
  this.device.app.sendCommands([{obj: this.id, id: "clearAll"}]);
};

/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @param {Object} notification
 * @returns {undefined}
 */
App.Device.Notification.prototype.cancel = function (notification)
{
  // There is no callback
  this.device.app.sendCommands([{obj: this.id, id: "cancel", cnt: notification}]);
};

/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @returns {undefined}
 */
App.Device.Notification.prototype.cancelAll = function ()
{
  // There is no callback
  this.device.app.sendCommands([{obj: this.id, id: "cancelAll"}]);
};


/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.Notification.prototype.hasPermission = function (cb)
{
  if (this.cb)
    throw "error: already quering permission";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "hasPermission"}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Notification.prototype.hasPermissionCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};

/*
 * Plugin call, sent using postMessage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.Notification.prototype.register = function (senderID, cb)
{
  if (this.cb)
    throw "error: already registering";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "register", cnt: {senderID: senderID}}]); // jshint ignore:line
};


/*
 * Push notifications
 * @param {object} message
 * @returns {object|array} token
 */
App.Device.Notification.prototype.push = function (msg, token, cb)
{
  var tokenList = (token instanceof Array) ? token : [token];
  var gcmtk = [];
  var apntk = [];
  //
  for (var i = 0; i < tokenList.length; i++) {
    if (tokenList[i].substr(0, 4) === "gcm-")
      gcmtk.push(tokenList[i].substr(4));
    else if (tokenList[i].substr(0, 4) === "apn-")
      apntk.push(tokenList[i].substr(4));
  }
  //
  // send to gcm if required
  var errorSent = false;
  if (gcmtk.length) {
    var gcm = require('node-gcm');
    //
    var message = new gcm.Message();
    var msgToSend = {};
    for (var v in msg) {
      var vv = v;
      if (v === "body" || v === "alert")
        vv = "message";
      msgToSend[vv] = msg[v];
    }
    message.addData(msgToSend);
    //
    var gkey = this.gcmKey || "AIzaSyB4NXshVT91fmbS3TjefdOvkV21uxTM170";
    var sender = new gcm.Sender(gkey);
    sender.send(message, gcmtk, function (error, result) {
      if (error)
        cb(error);
      else
        cb(result);
    });
  }
  //
  // send to apn if required
  if (apntk.length) {
    var apn = require('apn');
    var certp = this.certPath || "./cert/apn-prod.pem";
    var keyp = this.keyPath || "./cert/apn-key.pem";
    var prod = (this.production === undefined) ? true : this.production;
    var service = new apn.connection({cert: certp, key: keyp, production: prod});
    service.on("completed", function () {
      cb("completed");
    });
    service.on("error", function (error) {
      errorSent = true;
      cb(error.message);
    });
    service.on("transmissionError", function (error) {
      setImmediate(function () {
        if (!errorSent)
          cb(error);
      });
    });
    //
    var note = new apn.Notification();
    //
    for (var v in msg) {
      var vv = v;
      if (v === "body" || v === "message")
        vv = "alert";
      if (v === "payload")
        note[vv] = {payload: msg[v]};
      else
        note[vv] = msg[v];
    }
    //
    service.pushNotification(note, apntk);
  }
  //
  if (!gcmtk.length && !apntk.length) {
    if (cb)
      cb(0);
  }
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Notification.prototype.registerCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Event fired when a notification has been clicked
 * @param {notification}
 * @returns {undefined}
 */
App.Device.Notification.prototype.onClick = function (notification) // jshint ignore:line
{
};

/*
 * Event fired when a notification has been triggered
 * @param {notification}
 * @returns {undefined}
 */
App.Device.Notification.prototype.onTrigger = function (notification) // jshint ignore:line
{
};


/**
 * Statusbar initialization
 * @param {Object} device
 */
App.Device.StatusBar = function (device)
{
  this.device = device;
  this.id = "device-statusBar";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @returns {undefined}
 */
App.Device.StatusBar.prototype.hide = function ()
{
  // no callback
  this.device.app.sendCommands([{obj: this.id, id: "hide"}]);
};


/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @returns {undefined}
 */
App.Device.StatusBar.prototype.show = function ()
{
  // no callback
  this.device.app.sendCommands([{obj: this.id, id: "show"}]);
};


/*
 * Plugin call, sent using postMessage
 * It has no callback
 * @returns {undefined}
 */
App.Device.StatusBar.prototype.setBackgroundColor = function (color)
{
  // could this become a property?
  this.device.app.sendCommands([{obj: this.id, id: "setBackgroundColor", cnt: color}]);
};


/**
 * Contacts initialization
 * @param {Object} device
 */
App.Device.Contacts = function (device)
{
  this.device = device;
  this.id = "device-contacts";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.pick = function (cb)
{
  if (this.cb)
    throw "error: already picking contacts";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "pick", cnt: {}}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.pickCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Plugin call
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.find = function (filter, fieldsToReturn, fieldsToMatch, cb)
{
  if (this.cb)
    throw "error: already searching contacts";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "find", cnt: {filter: filter, fields: fieldsToMatch, desired: fieldsToReturn}}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.findCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Plugin call
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.read = function (contactId, cb)
{
  if (this.cb)
    throw "error: already reading a contact";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "read", cnt: contactId + ""}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.readCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Plugin call
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.create = function (info, cb)
{
  if (this.cb)
    throw "error: already creating a contact";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "create", cnt: info}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.createCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Plugin call
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.delete = function (contactId, cb)
{
  if (this.cb)
    throw "error: already deleting a contact";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "delete", cnt: contactId + ""}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.deleteCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Plugin call
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.update = function (contactId, info, cb)
{
  //
  if (this.cb)
    throw "error: already updating a contact";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "update", cnt: {id: contactId + "", data: info}}]);
};


/*
 * Plugin callback
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Contacts.prototype.updateCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/**
 * Camera initialization
 * @param {Object} device
 */
App.Device.Camera = function (device)
{
  this.device = device;
  this.id = "device-camera";
  device.app.eleMap[this.id] = this;
};


/*
 * Plugin call
 * @param {type} res
 * @returns {undefined}
 */
App.Device.Camera.prototype.getPicture = function (options)
{
  options.sid = this.device.app.id;
  this.device.app.sendCommands([{obj: this.id, id: "getPicture", cnt: options}]);
};


/*
 * Event fired with info about picture taken
 * @param {Object} picture = {file: fileURI, id: picid}
 * @returns {undefined}
 */
App.Device.Camera.prototype.onUpload = function (picture) // jshint ignore:line
{
};


/*
 * Private event fired directly from the shell, which creates an fs object and
 * passes it to onUpload
 * @param {Object} picture = {file: {path:"uploaded/path", type: undefined}, id: picid}
 * @returns {undefined}
 */
App.Device.Camera.prototype.onPreUpload = function (res) // jshint ignore:line
{
  if (res.err)
    return;
  //
  var f = this.device.app.fs.file(res.file[0].path, res.file[0].type, res.file[0].publicUrl);
  this.onUpload(f);
};


/**
 * Nfc initialization
 * @param {Object} device
 */
App.Device.Nfc = function (device)
{
  this.device = device;
  this.id = "device-nfc";
  device.app.eleMap[this.id] = this;
};


/*
 * Add a listener to a nfc event
 * @param {type} event type
 * @param {options} event options
 * @returns {undefined}
 */
App.Device.Nfc.prototype.listen = function (type, options)
{
  var cmd = {obj: this.id, id: "listen", cnt: {type: type, options: options}};
  this.device.app.sendCommands([cmd]);
};


/*
 * Event fired when a tag has been discovered
 * @param {nfcEvent}
 * @returns {undefined}
 */
App.Device.Nfc.prototype.onTag = function (nfcEvent) // jshint ignore:line
{
};


/*
 * Remove a listener to a nfc event
 * @param {type} event type
 * @returns {undefined}
 */
App.Device.Nfc.prototype.unListen = function (type)
{
  this.device.app.sendCommands([{obj: this.id, id: "unListen", cnt: {type: type}}]);
};


/**
 * Cookies using localstorage
 * @param {Object} device
 */
App.Device.LSCookies = function (device)
{
  this.device = device;
  this.id = "device-lscookies";
  device.app.eleMap[this.id] = this;
};


/*
 * Set a cookie in localstorage. A prefix is forced in order to distinguish
 * from other localstorage values
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.LSCookies.prototype.setCookie = function (name, value, exdays)
{
  //
  this.device.app.sendCommands([{obj: this.id, id: "setCookie", cnt: {name: name, value: value, exdays: exdays}}]);
};


/*
 * Read all the lscookies saved in localstorage
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.LSCookies.prototype.getCookies = function (cb)
{
  if (this.cb)
    throw "error: you are already reading the cookies";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "getCookies"}]);
};


/*
 * Load the cookies read from localstorage in the cookiemap
 * @param {type} cb
 * @returns {undefined}
 */
App.Device.LSCookies.prototype.getCookiesCB = function (res)
{
  res = res.result;
  //
  // copy in new cookies
  for (var k in res) {
    this.device.app.cookies[k] = res[k];
  }
  // return all cookies
  if (this.cb)
    this.cb(this.device.app.cookies);
  //
  delete this.cb;
};


/*
 * Let the application install and update sub-apps
 * Each sub-app is an Inde app packaged as zip file
 */
App.Device.AppManager = function (device)
{
  this.device = device;
  this.id = "device-appmanager";
  device.app.eleMap[this.id] = this;
};


/*
 * List the apps installed. IndeShell is not included,
 * rootapp (the first app installed, when building) is included (and updatable)
 */
App.Device.AppManager.prototype.listApps = function (cb)
{
  if (this.cb)
    throw "error: you are already listing the installed apps";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "listApps", cnt: {}}]);
};


/*
 * Callback for listApps, returns an list of the apps and info about them
 * @param {object} res
 *                 res.err It's the string error, if anythihng goes wrong.
 *                 res.result It's the resuslting list of apps.
 */
App.Device.AppManager.prototype.listAppsCB = function (res)
{
  //
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Command for launching a subapp.
 * @param {string} name. Select the app by name.
 * @param {bool} visible. If visible = false, the app is launched but not shown.
 * @param {bool} fullscreen. If fullscreen = true (and visible = true),
 *                           the app is shown and all other are automatically hidden.
 */
App.Device.AppManager.prototype.launchApp = function (name, visible, fullscreen, cb)
{
  if (this.cb)
    throw "error: you can only launch one app at a time";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "launchApp", cnt: {name: name, visible: visible, fullscreen: fullscreen}}]);
};


/*
 * Callback for launchApp.
 * @param {object} res
 *                 res.err It's the string error, if anythihng goes wrong.
 *                 res.result It's the resuslting value, if any.
 */
App.Device.AppManager.prototype.launchAppCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Command for stopping a launched app.
 * @param {string} name. It's the name of the app to stop.
 * @param {string} appToShow. It's the optional name of the app to show in place of the one stopped,
 * if it was being displayed.
 */
App.Device.AppManager.prototype.stopApp = function (name, appToShow, cb)
{
  if (this.cb)
    throw "error: you can only stop one app at a time";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "stopApp", cnt: {name: name, appToShow: appToShow}}]);
};


/*
 * Callback for stopApp.
 * @param {object} res
 *                 res.err It's the string error, if anythihng goes wrong.
 *                 res.result It's the resuslting value, if any.
 */
App.Device.AppManager.prototype.stopAppCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Command for showing a launched (but hidden) app.
 * @param {string} name. It's the name of the app to show.
 * @param {bool} fullscreen. If fullscreen = true the app is shown
 *                           and all other are automatically hidden.
 */
App.Device.AppManager.prototype.showApp = function (name, fullscreen, cb)
{
  if (this.cb)
    throw "error: you can only show one app at a time";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "showApp", cnt: {name: name, fullscreen: fullscreen}}]);
};


/*
 * Callback for showApp.
 * @param {object} res
 *                 res.err It's the string error, if anythihng goes wrong.
 *                 res.result It's the resuslting value, if any.
 */
App.Device.AppManager.prototype.showAppCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Command for hiding a launched app.
 * @param {string} name. It's the name of the app to hide.
 */
App.Device.AppManager.prototype.hideApp = function (name, cb)
{
  if (this.cb)
    throw "error: you can only hide one app at a time";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "hideApp", cnt: {name: name}}]);
};


/*
 * Callback for hideApp.
 * @param {object} res
 *                 res.err It's the string error, if anythihng goes wrong.
 *                 res.result It's the resuslting value, if any.
 */
App.Device.AppManager.prototype.hideAppCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Command for installing a sub-app.
 * @param {string} url. It's the url of the manifest.txt file for the app to install.
 * @param {string} version. It's the version of the app to install. It must be present in the manifest.
 *                      It accepts the format "1.0" (no minors as it parses the version as float)
 *                      It accepts version = "latest"
 * @param {string} name. It's the name to be give to the app when installed. If the name of an existing
 *                       app is given, setup will fail.
 */
App.Device.AppManager.prototype.installApp = function (url, version, name, cb)
{
  if (this.cb)
    throw "error: you can only install one app at a time";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "installApp", cnt: {url: url, version: version, name: name}}]);
};


/*
 * Callback for installApp.
 * @param {object} res
 *                 res.err It's the string error, if anythihng goes wrong.
 *                 res.result It's the resuslting value, if any.
 */
App.Device.AppManager.prototype.installAppCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Command for uninstalling a sub-app. The "rootapp" that was installed when first building the
 * package, can't be removed.
 * @param {string} name. It's the name of the sub-app to remove.
 */
App.Device.AppManager.prototype.uninstallApp = function (name, cb)
{
  if (this.cb)
    throw "error: you can only uninstall one app at a time";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "uninstallApp", cnt: {name: name}}]);
};


/*
 * Callback for uninstallApp.
 * @param {object} res
 *                 res.err It's the string error, if anythihng goes wrong.
 *                 res.result It's the resuslting value, if any.
 */
App.Device.AppManager.prototype.uninstallAppCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/*
 * Command for updating an app, included rootapp. It installs version = "latest".
 * @param {string} url. It's the url of the manifest.txt file for the app to install.
 * @param {string} name. It's the name to be give to the app when installed.
 */
App.Device.AppManager.prototype.updateApp = function (url, name, cb)
{
  if (this.cb)
    throw "error: you can only install one app at a time";
  //
  this.cb = cb;
  //
  this.device.app.sendCommands([{obj: this.id, id: "updateApp", cnt: {url: url, name: name}}]);
};


/*
 * Callback for uninstallApp.
 * @param {object} res
 *                 res.err It's the string error, if anythihng goes wrong.
 *                 res.result It's the resuslting value, if any.
 *
 */
App.Device.AppManager.prototype.updateAppCB = function (res)
{
  if (this.cb) {
    this.cb(res);
  }
  //
  delete this.cb;
};


/**
 * Show and hide the device's keyboard
 * @param {Object} device
 */
App.Device.Keyboard = function (device)
{
  this.device = device;
  this.id = "device-keyboard";
  device.app.eleMap[this.id] = this;
};


/*
 * Show the keyboard
 * @returns {undefined}
 */
App.Device.Keyboard.prototype.show = function ()
{
  //
  this.device.app.sendCommands([{obj: this.id, id: "show", cnt: {}}]);
};


/*
 * Hide the keyboard
 * @returns {undefined}
 */
App.Device.Keyboard.prototype.hide = function ()
{
  //
  this.device.app.sendCommands([{obj: this.id, id: "hide", cnt: {}}]);
};



/*
 * Enable or disable view shrinking when the keyboard is shown
 * @param {type} req - pluginmanager.js request obj
 * @returns {undefined}
 */
App.Device.Keyboard.prototype.shrinkView = function (shrink)
{
  this.device.app.sendCommands([{obj: this.id, id: "shrinkView", cnt: {shrink: shrink}}]);
};


/*
 * Show or hide view accessory bar
 * @param {type} req - pluginmanager.js request obj
 * @returns {undefined}
 */
App.Device.Keyboard.prototype.hideViewAccessoryBar = function (hide)
{
  this.device.app.sendCommands([{obj: this.id, id: "hideViewAccessoryBar", cnt: {hide: hide}}]);
};


/*
 * Enable or disable scrolling when the keyboard is shown
 * @param {type} req - pluginmanager.js request obj
 * @returns {undefined}
 */
App.Device.Keyboard.prototype.disableScrollingInShrinkView = function (disable)
{
  this.device.app.sendCommands([{obj: this.id, id: "disableScrollingInShrinkView", cnt: {disable: disable}}]);
};
