/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module */

var App = App || {};

if (module)
  App.DataMap = require("../datamap");


/**
 * @class
 * @param {View} view
 * @param {Object} defprop
 * @param {Object} parent
 * */
App.Collection = function (view, defprop, parent)
{
  if (module && !App.View)
    App = require("../requireAll");
  //
  var app = null;
  if (view instanceof App.Session) {
    app = view;
    view = undefined;
  }
  else if (view instanceof App.View)
    app = view.app;
  //
  App.DataMap.call(this, view, defprop, parent);
  //
  this.app = app;
  this.rowClass = App.Document;
};

// Make App.Collection extend App.DataMap
App.Collection.prototype = new App.DataMap(true);

Object.defineProperties(App.Collection.prototype, {
  // Document template
  rowTemplate: {
    get: function () {
      return this._docTemplate;
    },
    set: function (newValue) {
      this.setDocTemplate(newValue);
    }
  },
  collection: {
    get: function () {
      return (this.innerCollection || this);
    },
    set: function (newValue) {
      this.setCollection(newValue);
    }
  },
  document: {
    get: function () {
      return this.row;
    },
    set: function (newValue) {
      if (newValue === this.document)
        return;
      //
      // Empty the collection
      if (this.innerCollection)
        this.setCollection(new App.Collection());
      else
        this.clear();
      //
      // Add document, if not null
      if (newValue) {
        this.add(newValue, {takeOwnership: false});
        this.position = 0;
      }
    }
  },
  // DataStore instance
  dataStore: {
    get: function () {
      // If collection has a template with own dataStore return it
      if (this.rowTemplate && this.rowTemplate.intProps.dataStore)
        return this.rowTemplate.intProps.dataStore;
      //
      // If collection has a parent with own dataStore return it
      if (this.parent && this.parent.dataStore && this.goodDataStore(this.parent.dataStore))
        return this.parent.dataStore;
      //
      // Return default instance of dataModel
      if (this.getDataModel())
        return this.app.getDefaultInstance(this.getDataModel());
    },
    //
    set: function (newValue) {
      if (this.rowTemplate && this.goodDataStore(newValue))
        this.rowTemplate.dataStore = newValue;
    }
  }
});


/**
 * Return an array of my documents
 */
App.Collection.prototype.getLoopCollection = function ()
{
  if (this.documentProvider) {
    // Better to give only existing documents
    var nc = [];
    for (var i = 0; i < this.rows.length; i++) {
      if (this.rows[i])
        nc.push(this.rows[i]);
    }
    return nc;
  }
  else
    return this.rows.slice();
};


/**
 * Set the document template
 * @param {App.Document} doc
 */
App.Collection.prototype.setDocTemplate = function (doc)
{
  if (this._docTemplate === doc)
    return;
  //
  if (doc.app && !this.app)
    this.app = doc.app;
  //
  this._docTemplate = doc;
  this.freezeEvents = true;
  this.rowClass = doc.class;
  this.addQuery(this.getLoadQuery());
  this.addSchema(doc.schema.clone());
};


/**
 * Remove all deleted documents
 */
App.Collection.prototype.purgeDeleted = function ()
{
  for (var i = 0; i < this.rows.length; i++) {
    var d = this.rows[i];
    if (d.deleted) {
      this.remove(d);
      i--;
    }
  }
};


/**
 * Set loaded this collection and children
 * @param {Integer} childLevel
 * @param {Boolean} loaded
 * */
App.Collection.prototype.setLoaded = function (childLevel, loaded)
{
  // First the collection
  this.loaded = loaded;
  //
  // Now the children
  this.rows.forEach(function (d) {
    if (d)
      d.setLoaded(childLevel, loaded);
  });
};


/**
 * Set this object in a no-modified state
 * @param {Integer} childLevel
 * */
App.Collection.prototype.setOriginal = function (childLevel)
{
  for (var i = 0; i < this.rows.length; i++) {
    var d = this.rows[i];
    if (d)
      d.setOriginal(childLevel);
  }
};


/**
 * Extend the schema with more columns
 * @param {Array} columns
 */
App.Collection.prototype.extendSchema = function (columns)
{
  for (var i = 0; i < columns.length; i++) {
    var nc = this.schema.addColumn(columns[i]);
    nc.extended = true;
  }
  //
  this.rowClass = this.createRowClass();
  this.collection = new App.Collection(this.app);
};


/**
 * Load collection from datamodel
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Collection.prototype.load = function (options, callback)
{
  // TODO: remove when all projects have libraries updated (with DataMap.load.options parameter)
  if (typeof options === "function") {
    callback = options;
    options = undefined;
  }
  //
  // Initialize default options
  options = Object.assign({
    loadType: App.Document.loadTypes.loadCollection,
    remote: this.rowTemplate.remote
  }, options);
  //
  // Set supplied dataStore to use for load
  var oldDataStore = (this.rowTemplate && this.rowTemplate.intProps.dataStore);
  if (options.dataStore) {
    this.dataStore = options.dataStore;
    delete options.dataStore;
  }
  //
  var done = function (result, error) {
    // Restore old dataStore
    this.dataStore = oldDataStore;
    if (callback)
      callback(result, error);
  }.bind(this);
  //
  // If collection has a parent document ask it to load me
  if (this.parent && options.loadType !== App.Document.loadTypes.loadChildCollection) {
    var wasLoaded = this.loaded;
    this.parent.loadChildCollection(this, options, function (result, error) {
      if (!error && this.loaded && !wasLoaded && !this.parent.freezeEvents) {
        this.parent.afterLoad(true, function (result, error) {
          this.parent.markAsChanged();
          done(null, error);
        }.bind(this));
      }
      else
        done(null, error);
    }.bind(this));
    return;
  }
  //
  // Add a warning if query missing any columns
  if (this.app) {
    for (var c in this.schema.columns) {
      var dc = this.schema.columns[c];
      if (dc.dmCode && !dc.getSelectToken()) {
        if (options.loadType === App.Document.loadTypes.load)
          this.app.dttobj.addWarning("", "The load query failed to load the " + c + " column for the " + this.schema.className + " document. This property will not be set.");
        else
          this.app.dttobj.addWarning("", "The load query failed to load the " + c + " column for the " + this.schema.className + " document collection. This property will not be set.");
      }
    }
  }
  //
  App.DataMap.prototype.load.call(this, options, done);
};


/**
 * Fire afterLoad event
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Collection.prototype.fireAfterLoad = function (options, callback)
{
  // If there are no rows or skipped, continue
  if (options.skip || this.rows.length === 0)
    return App.DataMap.prototype.fireAfterLoad.call(this, options, callback);
  //
  var firstDoc = (options.page || 0) * this.dataPageSize;
  var docLen = this.rows.length;
  if (this.dataPagingMode)
    docLen = Math.min(this.rows.length - firstDoc, this.dataPageSize);
  var lastDoc = firstDoc + docLen - 1;
  //
  var firstError;
  var done = function (result, error) {
    firstError = firstError || error;
    if (--docLen > 0)
      return;
    //
    if (firstError)
      return callback(null, firstError);
    //
    App.DataMap.prototype.fireAfterLoad.call(this, options, callback);
  }.bind(this);
  //
  for (var i = firstDoc; i <= lastDoc; i++) {
    var doc = this.rows[i];
    doc.setOriginal(0);
    doc.stopEvents();
    doc.loaded = true;
    doc.loadChildren(options.childLevel - 1, function (result, error) {
      if (error)
        return done(null, error);
      //
      if (options.loadType === App.Document.loadTypes.load)
        return done();
      //
      doc.afterLoad(false, function (result, error) {
        doc.markAsChanged();
        doc.startEvents();
        done(null, error);
      });
    });
  }
};


/**
 * Detach elements of rows for reuse them
 * @param {Integer} nr - number of element to reuse
 */
App.Collection.prototype.reuseElements = function (nr)
{
  this.updateUI(function (dm) {
    App.DataMap.prototype.reuseElements.call(dm, nr);
  });
};


/*
 * Remove detached elements unused
 */
App.Collection.prototype.removeUnusedElements = function ()
{
  this.updateUI(function (dm) {
    App.DataMap.prototype.removeUnusedElements.call(dm);
  });
};


/**
 * Fire beforeLoad event
 * @param {Object} q
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Collection.prototype.fireBeforeLoad = function (q, options, callback)
{
  App.DataMap.prototype.fireBeforeLoad.call(this, q, options, function (result, error) {
    if (error)
      return callback(null, error);
    //
    if (!this.rowTemplate)
      return callback();
    //
    // Set filters into document's properties
    if (options.loadType !== App.Document.loadTypes.load) {
      for (var c in this.schema.columns) {
        var dc = this.schema.columns[c];
        var filter = dc.filters[0];
        if (filter && !filter.or)
          this.rowTemplate.setProp(dc.name, filter.v, !options.useQBE);
        else
          this.rowTemplate.setProp(dc.name);
      }
    }
    //
    // Notify beforeLoad event to template document
    this.rowTemplate.beforeLoad(this, options, callback);
  }.bind(this));
};


/**
 * Save data of collection
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * @return {Boolean} true if successful
 */
App.Collection.prototype.save = function (options, callback)
{
  // TODO: remove when all projects have libraries updated (with DataMap.load.options parameter)
  if (typeof options === "function") {
    callback = options;
    options = undefined;
  }
  //
  // Initialize default options
  options = Object.assign({
    childLevel: 9999,
    revalidate: true
  }, options);
  //
  // Set supplied dataStore to use for save
  var oldDataStore = (this.rowTemplate && this.rowTemplate.intProps.dataStore);
  if (options.dataStore) {
    this.dataStore = options.dataStore;
    delete options.dataStore;
  }
  //
  var done = function (result, error) {
    // Restore old dataStore
    this.dataStore = oldDataStore;
    //
    // Notify onError event
    var ok = !(error || result === false);
    if (!ok)
      this.onError(error, {source: App.DataMap.errorSources.save, row: this.rows[docIdx]});
    //
    if (callback)
      callback(ok);
  }.bind(this);
  //
  var atLeastOne = false;
  var docIdx = -1;
  var saveNextDoc = function (result, error) {
    if (error || result === false)
      return done(false);
    //
    // Remove deleted document
    if (docIdx >= 0 && this.rows[docIdx].deleted)
      this.remove(this.rows[docIdx--]);
    //
    if (++docIdx < this.rows.length) {
      var d = this.rows[docIdx];
      if (d.isModified()) {
        // If is the first change notify beforeSave event
        if (!atLeastOne) {
          atLeastOne = true;
          this.beforeSave(function (result, error) {
            if (error)
              return done(null, error);
            //
            d.save(options, saveNextDoc);
          });
        }
        else
          d.save(options, saveNextDoc);
      }
      else
        saveNextDoc();
    }
    else {
      if (atLeastOne) {
        // Notify afterSave event
        this.afterSave(function (result, error) {
          done(true, error);
        });
      }
      else
        done(true);
    }
  }.bind(this);
  saveNextDoc();
};


/**
 * Add a filter to a column
 * @param {String} columnName
 * @param {RemElem} obj
 * @param {String} prop
 */
App.Collection.prototype.addDOFilter = function (columnName, obj, prop)
{
  // If column not exists skip filter
  var dc = this.schema.columns[columnName];
  if (!dc)
    return;
  //
  var st = dc.getSelectToken();
  if (!st)
    return;
  //
  // Clone the select token
  st = JSON.parse(JSON.stringify(st));
  var token = {};
  if (typeof obj === "object" && prop) {
    token.expr = App.Utils.generateUID24();
    token.obj = obj;
    token.prop = prop;
    token.clientRef = true;
  }
  else
    token.expr = dc.toSql(obj);
  //
  if (!this.DOFilters)
    this.DOFilters = [];
  var filter = {};
  filter.col = columnName;
  filter.obj = obj;
  filter.prop = prop;
  this.DOFilters.push(filter);
  //
  filter.wc = {expr: "=", op: true, p: [st, token]};
  if (!this.query.where)
    this.query.where = [];
  this.query.where.push(filter.wc);
};


/**
 * Find the index of a document
 * @param {Document} doc to find
 */
App.Collection.prototype.findDoc = function (doc)
{
  var idx = this.rows.indexOf(doc);
  if (idx === -1 && doc.document && this.innerCollection)
    idx = this.innerCollection.findDoc(doc.document);
  return idx;
};


/**
 * Set the collection to show
 * @param {Collection} collection to linked
 */
App.Collection.prototype.setCollection = function (collection)
{
  // Skip if same collection
  if (this.innerCollection === collection || collection === this)
    return;
  //
  this.loadingData = true;
  //
  // If I am already linked to a collection, unlink them
  if (this.innerCollection) {
    this.innerCollection.unlinkDataMap(this);
    this.innerCollection = null;
  }
  //
  var oldLength = this.length;
  if (!collection)
    this.clear(); // Empty all data
  else {
    // Set rowTemplate if not defined
    if (!collection.rowTemplate)
      collection.rowTemplate = this.rowTemplate.clone();
    //
    // Detach elements to save them
    this.reuseElements(collection.uiRows.length);
    //
    // Empty all data
    this.clear();
    //
    // Inherit sort criterias if I don't have any
    if (JSON.stringify(this.sortBy) !== JSON.stringify(collection.sortBy)) {
      if (collection.sortBy.length > 0 && this.sortBy.length === 0) {
        this.orderBy = collection.orderBy;
        this.groupBy = collection.groupBy;
      }
    }
    //
    // Add linked rows to me
    for (var i = 0; i < collection.rows.length; i++)
      this.add(collection.rows[i], {takeOwnership: false});
    //
    this.removeUnusedElements();
    //
    this.sort();
    //
    // Link new collection
    collection.linkDataMap(this);
    this.innerCollection = collection;
  }
  //
  if (this.useWindow) {
    this.schema.listBox.rowCount = this.uiRows.length;
    Object.getNotifier(this.schema.listBox).notify({type: "update", name: 'rowCount', oldValue: this.schema.listBox.rowCount});
  }
  //
  this.initElement();
  //
  this.loadingData = false;
  //
  // Notify onDataChange only if I have not yet rows
  if (oldLength === 0 && this.length === 0)
    this.fireOnDataChange();
};


/**
 * Create a new row linked to a document
 * @param {Document} doc to link
 */
App.Collection.prototype.createLinkedRow = function (doc) {
  var r = (this.rowClass === App.Document ? new (doc.class)(this) : this.createRow());
  r.intProps = doc.intProps;
  r.hidden = doc.hidden;
  r.document = doc;
  r.datamap = this;
  return r;
};


/**
 * Add a link to a datamap
 * @param {Collection} datamap to link
 */
App.Collection.prototype.linkDataMap = function (datamap)
{
  if (!this.linkedDataMaps)
    this.linkedDataMaps = {};
  this.linkedDataMaps[datamap.id] = datamap;
};


/**
 * Remove a link to a datamap
 * @param {Collection} datamap to unlink
 */
App.Collection.prototype.unlinkDataMap = function (datamap)
{
  delete this.linkedDataMaps[datamap.id];
};


/**
 * Closes the collection
 */
App.Collection.prototype.close = function ()
{
  // Unlink innerCollection
  if (this.innerCollection)
    this.innerCollection.unlinkDataMap(this);
  //
  App.DataMap.prototype.close.call(this);
};


/**
 * Add a row to the collection
 * @param {Object/DataRow} row to add
 * @param {Object} options
 */
App.Collection.prototype.add = function (row, options)
{
  options = Object.assign({callerDM: this}, options);
  //
  // Call on innerCollection
  if (this.innerCollection) {
    if ("position" in options) {
      options.callerPos = options.position;
      delete options.position;
    }
    return this.innerCollection.add(row, options);
  }
  //
  if (options.takeOwnership === undefined)
    options.takeOwnership = (!row.collection || (this.parent && row.collection.parent) ? true : false);
  //
  var lrow = row;
  if (!options.takeOwnership) {
    lrow = this.createLinkedRow(row);
    row.refDocuments = row.refDocuments || [];
    row.refDocuments.push(lrow);
  }
  else if (row.collection) {
    // Really I have to stay here, I remove it and take off the takeOwnership
    if (row.collection === this) {
      options.takeOwnership = false;
      var oldFreeze = false;
      if (row.parent) {
        // Freeze events of his parent to avoid to notifies it onChildRemoved event
        oldFreeze = row.parent.freezeEvents;
        row.parent.freezeEvents = true;
      }
      //
      // Remove the document that will be added back just below
      row.collection.remove(row);
      //
      // Restoring the freezeEvents of parent
      if (row.parent)
        row.parent.freezeEvents = oldFreeze;
    }
    else
      row.collection.remove(row);
  }
  //
  // Control of the lock must have ...
  // but not in the early stages of loading, there have to be able to form the doc in memory!
  if (options.takeOwnership && this.parent && !this.parent.freezeEvents) {
    /*
     var s = null;
     if (!this.parent.getLock(9999, "", s))
     throw new Error("ERR " + MDOInit.DOERR_NOLOCK + " [ " + Parent() + "] " + s);
     */
  }
  //
  // If the collection still does not know what type of document contains
  // set it to the type of the first added document
  if (!this.rowTemplate && row instanceof App.Document)
    this.rowTemplate = new (row.class)(this.app);
  //
  row = App.DataMap.prototype.add.call(this, lrow, options);
  for (var i in this.linkedDataMaps) {
    var ldm = this.linkedDataMaps[i];
    //
    if ("callerPos" in options) {
      // If collection have different orderBy find a good new position
      // otherwise add linked row to same position
      if (JSON.stringify(ldm.sortBy) === JSON.stringify(options.callerDM.sortBy))
        options.position = options.callerPos;
      else if (ldm.sortBy.length > 0 && JSON.stringify(ldm.sortBy) !== JSON.stringify(options.callerDM.sortBy))
        options.position = ldm.findRowPosition(row);
      else
        options.position = row.position;
    }
    //
    App.DataMap.prototype.add.call(ldm, ldm.createLinkedRow(row), options);
  }
  //
  if (options.takeOwnership) {
    row.collection = this;
    if (this.parent) {
      row.parent = this.parent;
      //
      if (!this.parent.freezeEvents)
        this.parent.onChildAdded(this, row);
    }
    //
    // If I was included in a collection of support (no document) and I move in this collection
    // the change is not undoable (can not go back because I do not know
    // if the collection where I come from still exists)
    if (!row.oldParent)
      row.oldCollection = null;
    //
    // Mark the document as changed
    row.markAsChanged();
  }
  //
  return row;
};


/**
 * Remove a document from the collection
 * @param {DataRow} row to delete
 */
App.Collection.prototype.remove = function (row)
{
  // If not specified delete current row
  if (!row)
    row = this.row;
  //
  // Call on innerCollection
  if (this.innerCollection)
    return this.innerCollection.remove(row.document);
  //
  var p = row.index;
  this.updateUI(function (dm) {
    App.DataMap.prototype.remove.call(dm, dm.rows[p]);
  });
  //
  // If the document has a parent, and it does not have freezed events
  // notified the onChildRemoved evento to parent
  if (row.parent && !row.parent.freezeEvents) {
    row.parent.onChildRemoved(row.collection, row);
    //
    // The parent lost his child and then changed
    row.parent.markAsChanged();
  }
};


/**
 * Returns the number of rows of a resultSet
 * @param {Object} rs
 */
App.Collection.prototype.getResultCount = function (rs)
{
  if (rs instanceof App.Collection)
    return rs.length;
  else
    return App.DataMap.prototype.getResultCount.call(this, rs);
};


/**
 * Returns the i-th row of a resultSet
 * @param {Object} rs - recordset
 * @param {int} pos - position
 */
App.Collection.prototype.getResultRow = function (rs, pos)
{
  if (rs instanceof App.Collection) {
    // Reset owner collection so when the document will be added
    // to the final collection will have the ownership
    var r = rs.rows[pos];
    r.collection = null;
    return r;
  }
  else
    return App.DataMap.prototype.getResultRow.call(this, rs, pos);
};


/**
 * Remove all data from collection
 */
App.Collection.prototype.clear = function ()
{
  // Call on innerCollection
  if (this.innerCollection)
    return this.innerCollection.clear();
  //
  this.updateUI(function (dm) {
    App.DataMap.prototype.clear.call(dm);
  });
};


/**
 * Return a copy of this object
 * @param {Object} map
 */
App.Collection.prototype.clone = function (map)
{
  var dm = new App.Collection(this.view);
  if (map)
    map[this.id] = dm;
  var keys = Object.keys(this);
  for (var k in keys) {
    var v = keys[k];
    //
    // Don't copy some objects
    if (["rows", "uiRows", "uiCache", "uiRecycleBin", "clientRefs", "needReload", "loadingPages",
      "schema", "query", "rowsToUpdate"].indexOf(v) >= 0) {
      continue;
    }
    //
    if (v === "elements") {
      for (var i = 0; i < this.elements.length; i++) {
        var ne = this.elements[i].clone(map);
        dm.elements.push(ne);
        //
        // Relink original parent
        if (ne instanceof App.DataMap)
          ne.parentElement = dm;
        else
          ne.parent = dm;
      }
    }
    else if (v === "id") {
      // The id property is special and needs to be generated here
      dm[v] = this.app.generateCounter();
      this.app.eleMap[dm.id] = dm;
    }
    else if (v === "_searchOptions")
      dm._searchOptions = JSON.parse(JSON.stringify(this._searchOptions));
    else if (v === "outputElements")
      dm.outputElements = this.outputElements.slice(0);
    else if (v === "DOFilters") {
      for (var i = 0; i < this.DOFilters.length; i++) {
        var filter = this.DOFilters[i];
        var obj = map[filter.obj.id] || filter.obj;
        dm.addDOFilter(filter.col, obj, filter.prop);
      }
    }
    else if (v === "_docTemplate") {
      dm.rowTemplate = this.rowTemplate.clone();
      dm.schema = this.schema.clone();
      dm.schema.parent = dm;
    }
    else
      dm[v] = this[v];
  }
  //
  dm.templateDataMap = this;
  dm.needReload = this.autoLoad;
  //
  setImmediate(function () {
    if (dm.parentElement instanceof App.RemElem)
      dm.parentElement.dataMap = dm;
    //
    if (dm.autoLoad)
      dm.refreshData();
  });
  //
  return dm;
};


/**
 * Fire onSearch event
 * @param {String} text
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Collection.prototype.fireOnSearch = function (text, options, callback)
{
  // Call base method
  App.DataMap.prototype.fireOnSearch.call(this, text, options, function (result, error) {
    // No rowTemplate or event skipped -> do nothing
    if (!this.rowTemplate || options.skip)
      return callback();
    //
    // Ask to docTemplate if want do a custom search
    this.rowTemplate.search(text, function (datamap, error) {
      if (error)
        return callback(null, error);
      //
      // If docTemplate returns a datamap load data from it
      if (datamap) {
        this.clear();
        options.skip = true;
        return this.loadFromDataMap(datamap, callback);
      }
      //
      callback();
    }.bind(this));
  }.bind(this));
};


/**
 * Define the class for rows of this collection
 * @return {Function}
 */
App.Collection.prototype.createRowClass = function ()
{
  var baseClass = this.rowTemplate.class;
  var rc = function (parent, data) {
    baseClass.call(this, parent, data);
  };
  //
  // Make rowClass extend baseClass
  rc.prototype = new baseClass(true);
  //
  // Define extended column properties
  for (var c in this.schema.columns) {
    var dc = this.schema.columns[c];
    if (dc.extended)
      dc.defineClassProperty(rc);
  }
  //
  return rc;
};


/**
 * Return true if there is at least one document is modified
 * @return {Boolean}
 */
App.Collection.prototype.isModified = function ()
{
  // Call on innerCollection
  if (this.innerCollection)
    return this.innerCollection.isModified();
  else
    return App.DataMap.prototype.isModified.call(this);
};


/**
 * Reset each document to original status
 */
App.Collection.prototype.undoChanges = function ()
{
  // Call on innerCollection
  if (this.innerCollection) {
    this.innerCollection.undoChanges();
    return;
  }
  //
  App.DataMap.prototype.undoChanges.call(this);
};


/**
 * Validate the collection
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Collection.prototype.validate = function (options, callback)
{
  // Initialize default options
  options = Object.assign({
    reason: App.Document.validationReasons.save,
    childLevel: 9999
  }, options);
  //
  var c = this.getLoopCollection();
  var docCount = c.length;
  if (docCount === 0)
    return callback(true);
  //
  var firstError;
  var ok = true;
  var done = function (result, error) {
    firstError = firstError || error;
    ok = ok && result;
    if (--docCount === 0)
      callback(ok, firstError);
  };
  //
  // Validate documents in parallel
  c.forEach(function (doc) {
    doc.validate(options, done);
  });
};


/**
 * Tells something about this object to the DTT module
 * @param {Integer} childLevel - level of child to log
 */
App.Collection.prototype.getDebugInfo = function (childLevel)
{
  var info = App.DataMap.prototype.getDebugInfo.call(this, childLevel);
  info._class = "Collection";
  return info;
};


/**
 * Create the load query for this collection
 * */
App.Collection.prototype.getLoadQuery = function ()
{
  if (!this.rowTemplate)
    return;
  //
  if (this.parent) {
    var mdc = this.parent.schema.getCollDefinition(this.index);
    //
    // If collection has a load query by design time returns that
    if (!mdc.query) {
      // Returns load query of docTemplate with filter for parent
      var q = JSON.parse(JSON.stringify(this.rowTemplate.getLoadQuery()));
      if (mdc.pks) {
        q.where = q.where || [];
        for (var i = 0; i < mdc.pks.length; i++) {
          var w = {p: []};
          w.p.push({expr: mdc.fks[i], from: (q.from.length > 1 ? q.from[0].alias : "")});
          w.p.push({expr: "="});
          w.p.push({expr: "this.parent." + mdc.pks[i], clientRef: true});
          q.where.push(w);
        }
      }
      //
      // Cache the query
      mdc.query = q;
    }
    return JSON.parse(JSON.stringify(mdc.query));
  }
  else // Returns load query of docTemplate
    return this.rowTemplate.getLoadQuery();
};


/**
 * Save to JSON
 * @param {Object} options
 */
App.Collection.prototype.intoJSON = function (options)
{
  // Initialize default options
  options = Object.assign({asString: true}, options);
  //
  var asString = options.asString;
  var asObject = options.asObject;
  options.asString = false;
  options.asObject = false;
  //
  // Add the rows
  var rows = [];
  for (var i = 0; i < this.rows.length; i++)
    rows.push(this.rows[i].intoJSON(options));
  //
  var json;
  if (asObject) {
    json.rows = rows;
    App.Document.JSONReplacer(this, json, options);
  }
  else
    json = rows;
  //
  return (asString ? JSON.stringify(json) : json);
};


/**
 * Load from JSON
 * @param {Object} json
 * @param {Object} options
 */
App.Collection.prototype.loadFromJSON = function (json, options)
{
  if (typeof json === "string")
    json = JSON.parse(json);
  //
  var rows;
  if (json instanceof Array)
    rows = json;
  else {
    App.Document.JSONReplacer(this, json, options);
    rows = json.rows;
  }
  //
  // Load the rows
  if (rows) {
    var cls = (this.rowTemplate ? this.rowTemplate.class : App.Document);
    for (var i = 0; i < rows.length; i++)
      this.add(cls.fromJSON(this.app, rows[i], options));
  }
  //
  this.loaded = true;
};


/**
 * Create a collection from JSON
 * @param {Session} app
 * @param {Object/String} json
 * @param {Object} options
 */
App.Collection.fromJSON = function (app, json, options)
{
  options = Object.assign({}, options);
  //
  var coll = new App.Collection(app);
  if (options.contentType && options.contentType.prototype instanceof App.Document) {
    coll.rowTemplate = new (options.contentType)(app);
    delete options.contentType;
  }
  //
  coll.loadFromJSON(json, options);
  return coll;
};


/**
 * Load from DataMap
 * @param {DataMap} datamap
 * @param {Function} callback - function to be called at the end
 */
App.Collection.prototype.loadFromDataMap = function (datamap, callback)
{
  var firstError;
  var len = datamap.rows.length;
  var done = function (result, error) {
    firstError = firstError || error;
    if (--len > 0)
      return;
    //
    this.loaded = true;
    callback(null, firstError);
  }.bind(this);
  //
  // If no rows done
  if (len === 0)
    return done();
  //
  // Create a document for each row
  for (var i = 0; i < datamap.rows.length; i++) {
    var doc = this.add(datamap.rows[i]);
    doc.setOriginal(0);
    doc.stopEvents();
    doc.loaded = true;
    doc.afterLoad(false, function (result, error) {
      doc.markAsChanged();
      doc.startEvents();
      done(null, error);
    });
  }
};


/**
 * Compare two documents
 * @param {Document} doc1
 * @param {Document} doc2
 * */
App.Collection.prototype.compareRows = function (doc1, doc2)
{
  if (doc1.onCompare)
    return doc1.onCompare(doc2);
  else
    return App.DataMap.prototype.compareRows.call(this, doc1, doc2);
};


/**
 * Returns all document in hierarchy
 * @param {Integer} childLevel
 */
App.Collection.prototype.getAllDocuments = function (childLevel)
{
  var docs = [];
  //
  // Add all my documents
  for (var i = 0; i < this.rows.length; i++)
    Array.prototype.push.apply(docs, this.rows[i].getAllDocuments(childLevel));
  //
  return docs;
};


/**
 * Sort UI rows
 */
App.Collection.prototype.sort = function ()
{
  this.updateUI(function (dm) {
    // Sort linked datamap only if have same orderBy
    if (JSON.stringify(dm.sortBy) === JSON.stringify(this.sortBy))
      App.DataMap.prototype.sort.call(dm);
  }.bind(this));
};


/**
 * Call a function for all related datamap
 * @param {Function} funct
 */
App.Collection.prototype.updateUI = function (funct)
{
  // Call function for me
  funct(this);
  //
  // Call function for all linked datamap
  for (var i in this.linkedDataMaps)
    this.linkedDataMaps[i].updateUI(funct);
};


/**
 * Execute load query (paged if needed)
 * @param {Object/String} qry
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Collection.prototype.execQuery = function (qry, options, callback)
{
  // If docTemplate is not remote exec query as usually
  if (!options.remote)
    return App.DataMap.prototype.execQuery.call(this, qry, options, callback);
  //
  // DocTemplate is remote -> do remote load
  if (this.rowTemplate.webApiEndpoint)
    this.rowTemplate.webApiEndpoint.handleLoad(this, options, callback);
  else
    this.app.sync.DO.remoteLoad(this, options, callback);
};


//export module for node
if (module)
  module.exports = App.Collection;
