/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, Error, ArrayBuffer */

var App = App || {};

if (module)
  App.DataRow = require("../datarow");


/**
 * @class A Document Object
 * @param {DataMap/Session} parent - datamap owner or application session
 * @param {Object} data
 * */
App.Document = function (parent, data)
{
  if (module && !App.View && parent !== true) {
    App = require("../requireAll");
    t = App.Utils.t;
  }
  //
  var app = null;
  if (parent instanceof App.Session)
    app = parent;
  else if (parent instanceof App.Collection)
    app = parent.app;
  //
  App.DataRow.call(this, app, data);
  //
  // Notify onInit event
  this.onInit();
};

// Make App.Document extend App.DataRow
App.Document.prototype = new App.DataRow(true);


App.Document.ServiceTypes = {
  synchronization: "synchronization"
};


App.Document.validationReasons = {
  change: "change",
  save: "save",
  sync: "sync"
};


App.Document.savePhases = {
  beforeSave: "beforeSave",
  inserting: "inserting",
  updating: "updating",
  deleting: "deleting",
  afterSave: "afterSave"
};


/**
 * Types of loading
 * */
App.Document.loadTypes = {
  load: 0,
  loadChildCollection: 1,
  loadCollection: 2
};


// Getter/Setter of some properties
Object.defineProperties(App.Document.prototype, {
  // My owner collection
  collection: {
    get: function () {
      if (this.document !== this)
        return this.document.datamap;
      else
        return this.datamap;
    },
    set: function (newValue) {
      if (this.document !== this)
        this.document.datamap = newValue;
      else
        this.datamap = newValue;
    }
  },
  // My parent document
  parent: {
    get: function () {
      return this.intProps.parent;
    },
    set: function (newValue) {
      this.intProps.parent = newValue;
    }
  },
  // Loaded status
  loaded: {
    get: function () {
      return (this.intProps.loaded === true);
    },
    set: function (newValue) {
      this.intProps.loaded = newValue;
    }
  },
  // changed status
  changed: {
    get: function () {
      return this.intProps.changed;
    },
    set: function (newValue) {
      this.intProps.changed = newValue;
    }
  },
  // Linked document
  document: {
    get: function () {
      return (this.uiProps.document || this);
    },
    set: function (newValue) {
      this.uiProps.document = newValue;
    }
  },
  // My old parent document
  oldParent: {
    get: function () {
      return this.intProps.oldParent;
    },
    set: function (newValue) {
      this.intProps.oldParent = newValue;
    }
  },
  // My old parent collection
  oldCollection: {
    get: function () {
      return this.intProps.oldCollection;
    },
    set: function (newValue) {
      this.intProps.oldCollection = newValue;
    }
  },
  // If changes has be to synchronized
  syncChanges: {
    get: function () {
      if (this.intProps.syncChanges === false)
        return false;
      else if (this.parent)
        return this.parent.syncChanges;
      else
        return true;
    },
    set: function (newValue) {
      this.intProps.syncChanges = newValue;
    }
  },
  // If load and save operations are remote
  remote: {
    get: function () {
      return (this.intProps.remote || this.class.remote || App.Document.remote);
    },
    set: function (newValue) {
      this.intProps.remote = newValue;
    }
  },
  // DataStore instance
  dataStore: {
    get: function () {
      // If document has own dataStore return it
      if (this.intProps.dataStore)
        return this.intProps.dataStore;
      //
      // If document is in a collection with own dataStore return it
      if (this.collection && this.collection.dataStore)
        return this.collection.dataStore;
      //
      // Return default instance of dataModel
      var dm = this.getDataModel();
      if (this.app && dm)
        return this.app.getDefaultInstance(dm);
    },
    //
    set: function (newValue) {
      if (this.goodDataStore(newValue))
        this.intProps.dataStore = newValue;
    }
  },
  // WebAPI endpoint
  webApiEndpoint: {
    get: function () {
      return (this.intProps.webApiEndpoint || this.class.webApiEndpoint);
    },
    set: function (newValue) {
      this.intProps.webApiEndpoint = newValue;
    }
  }
});


/**
 * Return the name of datamodel
 * */
App.Document.prototype.getDataModel = function ()
{
  var md = this.schema;
  while (md) {
    if (md.datamodel)
      return md.datamodel;
    md = md.baseMD;
  }
};


/**
 * Check if dataStore is good for me
 * @param {App.Database} ds
 * */
App.Document.prototype.goodDataStore = function (ds)
{
  var dm = this.getDataModel();
  return !ds || !dm || (ds.schema && ds.schema.name === dm);
};


/**
 * Set loaded this document and children
 * @param {Integer} childLevel
 * @param {Boolean} loaded
 * */
App.Document.prototype.setLoaded = function (childLevel, loaded)
{
  // First the document
  this.loaded = loaded;
  //
  // Now the children
  if (childLevel > 0) {
    var md = this.schema;
    for (var i = 0; i < md.collections.length; i++)
      this[md.collections[i].name].setLoaded(childLevel - 1, loaded);
  }
};


/**
 * Set this object in a no-modified state
 * @param {Integer} childLevel
 * */
App.Document.prototype.setOriginal = function (childLevel)
{
  if (childLevel === undefined)
    childLevel = 9999;
  //
  this.oldParent = this.parent;
  this.oldCollection = this.collection;
  this.updated = false;
  this.inserted = false;
  if (childLevel > 0) {
    var md = this.schema;
    // Remove finally all objects deleted
    for (var i = 0; i < md.collections.length; i++) {
      var coll = this[md.collections[i].name];
      coll.purgeDeleted();
      coll.setOriginal(childLevel - 1);
    }
  }
};


/**
 * Reset this object to original state
 * @param {Integer} childLevel
 * */
App.Document.prototype.restoreOriginal = function (childLevel)
{
  if (childLevel === undefined)
    childLevel = 9999;
  //
  this.stopEvents();
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (this.getOriginalValue(mdp.name) !== this[mdp.name])
      this.setProp(mdp.name, this.getOriginalValue(mdp.name));
  }
  //
  // Restoring the parent
  if (this.oldCollection && this.oldCollection !== this.collection) {
    try {
      this.oldCollection.add(this);
    }
    catch (e) {
    }
  }
  //
  this.updated = false;
  if (this.inserted)
    this.deleted = true;
  else
    this.deleted = false;
  //
  if (childLevel > 0) {
    for (var i = 0; i < md.collections.length; i++) {
      // I have to copy the list because it is possible to restore the original
      // move objects and I would not have a comodification!
      var mdc = md.collections[i];
      var cl = this[mdc.name].getLoopCollection();
      for (var it = 0; it < cl.length; it++)
        cl[it].restoreOriginal(childLevel - 1);
      this[mdc.name].purgeDeleted();
    }
  }
  this.clearErrors();
  this.startEvents();
};


/**
 * Clear this document
 * */
App.Document.prototype.clear = function ()
{
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++)
    this.setProp(md.properties[i].name, null);
  //
  for (var i = 0; i < md.collections.length; i++)
    this[md.collections[i].name].clear();
  this.setOriginal();
  //
  // The document will reset completely
  this.loaded = false;
  this.inserted = false;
  this.deleted = false;
  this.updated = false;
  //
  this.clearErrors();
  this.onClear();
};


/**
 * Check if document is modified
 * @param {Integer} childLevel
 * */
App.Document.prototype.isModified = function (childLevel)
{
  if (childLevel === undefined)
    childLevel = 9999;
  //
  if (App.DataRow.prototype.isModified.call(this))
    return true;
  //
  if (childLevel > 0) {
    var md = this.schema;
    for (var i = 0; i < md.collections.length; i++) {
      var c = this[md.collections[i].name].getLoopCollection();
      for (var it = 0; it < c.length; it++) {
        if (c[it].isModified(childLevel - 1))
          return true;
      }
    }
  }
  return false;
};


/**
 * Validate the document
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.validate = function (options, callback)
{
  // Initialize default options
  options = Object.assign({
    reason: App.Document.validationReasons.save,
    childLevel: 9999
  }, options);
  //
  var oldRowError;
  var oldColumnsErrors;
  var done = function (result, error)
  {
    if (options.reason !== App.Document.validationReasons.change) {
      // Reset validating status
      delete this.intProps.validating;
      //
      // Reset the errors as they were before the validation started
      // and set them up again to trigger updatesElement if needed
      var newRowError = this.rowError;
      this.rowError = oldRowError;
      this.setError(newRowError || "");
      //
      var newColumnsErrors = this.columnsErrors;
      this.columnsErrors = oldColumnsErrors;
      for (var i = 0; i < this.schema.properties.length; i++) {
        var mdp = this.schema.properties[i];
        this.setError(newColumnsErrors[mdp.name] || "", mdp.name);
      }
    }
    //
    if (callback)
      callback(result, error);
  }.bind(this);
  //
  // Reconnect the hierarchy of documents
  this.reconnectFK(options.childLevel);
  //
  // Update process document if modified
  this.processChange(function (result, error) {
    if (error)
      return done(null, error);
    //
    // Clear all errors (expect for reason change)
    if (options.reason !== App.Document.validationReasons.change) {
      // Set validating status
      this.intProps.validating = true;
      //
      // Copy current errors and clear them
      oldColumnsErrors = Object.assign({}, this.columnsErrors);
      oldRowError = this.rowError;
      this.clearErrors();
    }
    //
    var opts = Object.assign({cancel: false, skip: false}, options);
    this.onValidate(opts, function (result, error) {
      if (error)
        return done(null, error);
      //
      if (opts.cancel) {
        if (this.app)
          this.app.dttobj.addWarning("", "Validation failed by custom event:\n" + this.getDocErrors(0));
        return done(false);
      }
      //
      if (!opts.skip) {
        this.internalValidate(opts);
        if (opts.cancel) {
          if (this.app)
            this.app.dttobj.addWarning("", "Validation failed for internal controls:\n" + this.getDocErrors(0));
          return done(false);
        }
        //
        // Check internal error document
        if (options.reason === App.Document.validationReasons.change ? this.getError(options.property) : this.hasErrors()) {
          if (this.app)
            this.app.dttobj.addWarning("", "Validation failed for errors:\n" + this.getDocErrors(0));
          return done(false);
        }
      }
      //
      var collCount = this.schema.collections.length;
      if (options.childLevel > 0 && options.reason !== App.Document.validationReasons.change && collCount > 0) {
        // Validate collections in parallel
        opts.childLevel--;
        var firstError;
        var ok = true;
        this.schema.collections.forEach(function (mdc) {
          this[mdc.name].validate(opts, function (result, error) {
            firstError = firstError || error;
            ok = ok && result;
            if (--collCount === 0)
              done(ok, firstError);
          });
        }.bind(this));
      }
      else
        done(true);
    }.bind(this));
  }.bind(this));
};


/**
 * Check internal constraint
 * @param {Object} options
 * */
App.Document.prototype.internalValidate = function (options)
{
  // Don't control the properties if it is deleted ...
  if (this.deleted)
    return;
  //
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    //
    // If validation is caused by change of a property validate only that
    if (options.reason === App.Document.validationReasons.change && options.property !== mdp.name)
      continue;
    //
    // If the property has already been hooked an error, then don't override it !
    var doe = this.getError(mdp.name);
    if (doe) {
      options.cancel = true;
      continue;
    }
    //
    var v = this[mdp.name];
    if (v === null || v === undefined) {
      if (!mdp.nullable && mdp.datatype !== App.Database.DataTypes.AUTOINCR) {
        var isError = true;
        //
        // Before giving an error I have to find out if maybe one of my father's documents
        // is connected to this field by another field counter which then will give me the value
        // after the father was saved.
        var dp = this.parent;
        var cp = this.collection;
        var nameToFind = mdp.dmCode;
        //
        while (dp && cp && isError && cp.parent === dp && cp.index >= 0) {
          var sdp = dp.schema;
          var mdc = sdp.getCollDefinition(cp.index);
          //
          // Cycle properties FK/PK of the collection to see if I'm there
          var yet = false;
          for (var j = 0; j < mdc.pks.length; j++) {
            var nmfk = mdc.fks[j];
            var nmpk = mdc.pks[j];
            if (nmfk === nameToFind) {
              yet = true;
              //
              // I am, let's see how it is made the property on the parent document
              // If the parent property that corresponds to me is counter, is ok
              var mpp = sdp.getPropDefinition(nmpk);
              if (mpp.datatype === App.Database.DataTypes.AUTOINCR)
                isError = false;
              //
              // I get on a level
              cp = dp.collection;
              dp = dp.parent;
              nameToFind = nmpk;
              //
              break;
            }
          }
          //
          // The variable is not in the FK?
          if (!yet)
            break;
        }
        //
        if (isError) {
          options.cancel = true;
          this.setError(mdp.name + " cannot be null", mdp.name);
        }
      }
    }
    else {
      /*
       switch (Glb.ClassDT(mdp.DataType)) {
       case Glb.DTC_CHAR:
       if (v.stringValue().length > mdp.MaxLength) {
       options.cancel = true;
       //
       e = "[|1] è troppo lungo; sono consentiti al massimo |2 caratteri";
       var z = v.stringValue();
       if (z.length > 40)
       z = z.substr(0, 37) + "...";
       var RTCObj = MDOInit.GetRTC(this);
       if (RTCObj != null)
       e = RTCObj.GetConst("0A52C2C4-2943-43C0-8254-7706B8E7E090", e);
       e = IDL.FormatMessage(new IDVariant(e), new IDVariant(z), new IDVariant(mdp.MaxLength)).stringValue();
       //
       if (MDOInit.GetDTT(this) != null)
       MDOInit.GetDTT(this).AddMsg(DTTEngine.DTTMSG_WARN, "", 206, "DO.InternalValidate", this.GetName(1) + ": " + mdp.UIName + " is too long");
       this.SetPropError(i, e);
       }
       else if (mdp.DataType == Glb.DT_CHAR_FIXED && v.stringValue().length < mdp.MaxLength)
       this.SetPropDirect(i, IDL.FillLR(v, new IDVariant(mdp.MaxLength), new IDVariant(" "), false));
       break;

       case Glb.DTC_DATE:
       if (!IDL.IsDate(v)) {
       options.cancel = true;
       //
       e = "[|1] non è una data";
       var z = v.stringValue();
       if (z.length > 40)
       z = z.substr(0, 37) + "...";
       var RTCObj = MDOInit.GetRTC(this);
       if (RTCObj != null)
       e = RTCObj.GetConst("D93EF94D-866A-4CC5-AE35-020FDD80C7D5", e);
       e = IDL.FormatMessage(new IDVariant(e), new IDVariant(z)).stringValue();
       //
       if (MDOInit.GetDTT(this) != null)
       MDOInit.GetDTT(this).AddMsg(DTTEngine.DTTMSG_WARN, "", 206, "DO.InternalValidate", this.GetName(1) + ": " + mdp.UIName + " is not a date value");
       this.SetPropError(i, e);
       }
       break;

       case Glb.DTC_NUMBER:
       case Glb.DTC_BOOL:
       if (!IDL.IsNumber(v)) {
       options.cancel = true;
       //
       e = "[|1] is not a number value";
       var z = v.stringValue();
       if (z.length > 40)
       z = z.substr(0, 37) + "...";
       var RTCObj = MDOInit.GetRTC(this);
       if (RTCObj != null)
       e = RTCObj.GetConst("D435495C-D17A-46F0-A73E-FD231C22A9B4", e);
       e = IDL.FormatMessage(new IDVariant(e), new IDVariant(z)).stringValue();
       //
       if (MDOInit.GetDTT(this) != null)
       MDOInit.GetDTT(this).AddMsg(DTTEngine.DTTMSG_WARN, "", 206, "DO.InternalValidate", this.GetName(1) + ": " + mdp.UIName + " is not a number value");
       this.SetPropError(i, e);
       }
       break;
       }
       */
    }
  }
};


/**
 * Fills the collection of child documents to be saved
 * @param {Array} collRef
 * */
App.Document.prototype.loadChildsToSave = function (collRef)
{
  var collSave = [];
  collRef = collRef || [];
  var onlyMemory = this.inserted && this.deleted;
  //
  // For each collection
  var md = this.schema;
  for (var i = 0; i < md.collections.length; i++) {
    var idcoll = this[md.collections[i].name];
    var coll = idcoll.getLoopCollection();
    for (var li = 0; li < coll.length; li++) {
      var d = coll[li];
      var isRef = d.collection !== idcoll;
      //
      if (isRef)
        collRef.push(d);
      else {
        // If the parent is only in memory, then all her children entered
        // must be only in memory, so set their deleted flag
        if (onlyMemory && d.inserted)
          d.deleted = true;
        collSave.push(d);
      }
    }
  }
  return collSave;
};


/**
 * Saving an entire document
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.save = function (options, callback)
{
  // Initialize default options
  options = Object.assign({
    childLevel: 9999,
    revalidate: true,
    remote: this.remote
  }, options);
  //
  // Set supplied dataStore to use for save
  var oldDataStore = this.intProps.dataStore;
  if (options.dataStore) {
    this.dataStore = options.dataStore;
    delete options.dataStore;
  }
  //
  var done = function (result, error) {
    // Restore old dataStore
    this.dataStore = oldDataStore;
    //
    // If error occurs set it to document
    if (error)
      this.setError(error);
    //
    if (callback)
      callback(result && !error);
  }.bind(this);
  //
  var afterValidate = function (result, error) {
    if (error || result === false)
      return done(false, error);
    //
    // If operation is remote -> do remote savage
    if (options.remote) {
      if (this.webApiEndpoint)
        return this.webApiEndpoint.handleSave(this, options, done);
      else
        return this.app.sync.DO.remoteSave(this, options, done);
    }
    //
    var transOpened = false;
    var closeTransaction = function (result, error) {
      var ok = !(error || result === false);
      if (ok)
        this.afterSaveReset(options.childLevel);
      //
      // Close the transaction
      if (transOpened)
        this.dataStore[(ok ? "commit" : "rollback") + "Transaction"](function (result, err) {
          done(ok, error || err);
        });
      else
        done(ok, error);
    }.bind(this);
    //
    var phases = [
      App.Document.savePhases.beforeSave,
      App.Document.savePhases.inserting,
      App.Document.savePhases.updating,
      App.Document.savePhases.deleting,
      App.Document.savePhases.afterSave
    ];
    var nextPhase = function (result, error) {
      if (error || result === false || phases.length === 0)
        return closeTransaction(result, error);
      //
      this.doSavePhase(phases.shift(), options.childLevel, nextPhase);
    }.bind(this);
    //
    // Open the transaction if required
    if (this.dataStore && !this.dataStore.isInTransaction()) {
      transOpened = true;
      this.dataStore.beginTransaction(function (result, error) {
        if (error)
          return done(null, error);
        //
        nextPhase();
      });
    }
    else
      nextPhase();
  }.bind(this);
  //
  // Revalidate the document if required
  if (options.revalidate) {
    var opts = {};
    opts.reason = App.Document.validationReasons.save;
    opts.childLevel = options.childLevel;
    this.validate(opts, afterValidate);
  }
  else
    afterValidate();
};


/**
 * Posting after save message
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.afterSavePhase = function (callback)
{
  // Before the afterSave, sending the doc at the service of sync
  this.app.sync.DO.addVariation(this, function (added, error) {
    if (error)
      return callback(null, error);
    //
    var options = {};
    options.cancel = false;
    options.phase = App.Document.savePhases.afterSave;
    this.onSave(options, function (result, error) {
      if (error)
        return callback(null, error);
      //
      // NotifyDocUpdate if required and only if I'm handling a variation or if I'm adding a variation server side
      if (this.app.sync.DO.notifyDocUpdates && added && (this.isSynchronizing() || !this.app.sync.isClientProxy()))
        this.notifyDocUpdate({allSessions: true});
      //
      if (options.cancel)
        return callback(false);
      //
      callback(true);
    }.bind(this));
  }.bind(this));
};


/**
 * Resetting state after save
 * @param {Integer} childLevel
 * */
App.Document.prototype.afterSaveReset = function (childLevel)
{
  // If class has cache remove cached copy of this document
  if (!this.inserted && this.isModified(0))
    this.class.clearCache(this.app, this.getDNA());
  //
  var md = this.schema;
  //
  // Posting to children
  if (childLevel > 0) {
    for (var i = 0; i < md.collections.length; i++) {
      var dc = this[md.collections[i].name];
      var c = dc.getLoopCollection();
      for (var it = 0; it < c.length; it++) {
        var d = c[it];
        // Only the real children and not those added for reference
        if (d.collection === dc)
          d.afterSaveReset(childLevel - 1);
      }
    }
  }
  //
  // Setting original status / purging deleted
  this.clearErrors();
  if (this.inserted) {
    this.loaded = true;
    for (var i = 0; i < md.collections.length; i++)
      this[md.collections[i].name].loaded = true;
  }
  //
  if (childLevel > 0) {
    for (var i = 0; i < md.collections.length; i++)
      this[md.collections[i].name].purgeDeleted();
  }
  //
  this.setOriginal(0);
};


/**
 * Execute a saving phase for the document and his childs
 * @param {String} phase
 * @param {Integer} childLevel
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.doSavePhase = function (phase, childLevel, callback) {
  // Execute phase for childs in parallel
  var forChilds = function (result, error) {
    if (error || result === false)
      return callback(result, error);
    //
    if (childLevel > 0) {
      var childs = this.loadChildsToSave();
      var childsCount = childs.length;
      if (childsCount === 0)
        return steps.shift()(true);
      //
      var firstError;
      var ok = true;
      var done = function (result, error) {
        firstError = firstError || error;
        ok = ok && result;
        if (--childsCount === 0)
          steps.shift()(ok, firstError);
      };
      //
      // Phase for child executed in parallel
      childs.forEach(function (child) {
        App.Document.prototype.doSavePhase.call(child, phase, childLevel - 1, done);
      });
    }
    else
      steps.shift()(true);
  }.bind(this);
  //
  // Execute phase for the document
  var forMe = function (result, error) {
    if (error || result === false)
      return callback(result, error);
    //
    if (phase === App.Document.savePhases.afterSave)
      App.Document.prototype.afterSavePhase.call(this, steps.shift());
    else
      App.Document.prototype.savePhase.call(this, phase, steps.shift());
  }.bind(this);
  //
  var steps;
  if (phase === App.Document.savePhases.deleting)
    steps = [forChilds, forMe, callback];
  else
    steps = [forMe, forChilds, callback];
  //
  steps.shift()();
};


/**
 * Execute a saving phase only for the document
 * @param {String} phase
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.savePhase = function (phase, callback)
{
  var options = {};
  options.skip = false;
  options.cancel = false;
  options.phase = phase;
  //
  // Fire onSave event
  this.onSave(options, function (result, error) {
    if (error) {
      var dttit = this.app.dttobj.addItem("", App.DTT.ItemTypeMap.EXCEPTION, this.getDNA() + ": exception in onSave event: " + error);
      this.app.dttobj.addItemParam("Error code", error.code, dttit);
      this.app.dttobj.addItemParam("Error message", error.message, dttit);
      //
      this.setError(error);
      return callback(false);
    }
    //
    if (options.cancel)
      return callback(false);
    //
    // Calculation list DOMD extended to be treated ...
    var docMD = [];
    var md = this.schema;
    while (md) {
      if (this.deleted)
        docMD.push(md);
      else
        docMD.unshift(md);
      md = md.baseMD;
    }
    //
    var i = -1;
    var nextMD = function (result, error) {
      if (error)
        return callback(false);
      //
      if (++i < docMD.length) {
        md = docMD[i];
        if ((this.updated || this.inserted || this.deleted) && !(this.deleted && this.inserted)) {
          if (!options.skip && (
                  (phase === App.Document.savePhases.inserting && this.inserted && !this.deleted)
                  || (phase === App.Document.savePhases.updating && !this.deleted && !this.inserted && this.loaded)
                  || (phase === App.Document.savePhases.deleting && this.deleted && !this.inserted && this.loaded))) {
            var q = null;
            if (phase === App.Document.savePhases.deleting)
              q = this.getDeleteQuery(md);
            else if (phase === App.Document.savePhases.inserting)
              q = this.getInsertQuery(md);
            else if (phase === App.Document.savePhases.updating)
              q = this.getUpdateQuery(md);
            //
            // I would not want it to be a fake update
            if (!this.dataStore) {
              /*
               if (this.GetDTT(this) != null)
               this.GetDTT(this).AddMsg(DTTEngine.DTTMSG_INFO, "", 210, "DO.SaveToDB", this.GetName(1) + " (" + MD.ObjTag + "): not saved because the class is not linked to a DM table");
               */
            }
            else if (q) {
              // Compose sql statement of the query
              var parameters = [];
              var sql = this.dataStore.createSqlStatement(q, parameters);
              this.dataStore.executeSQL(sql, false, parameters, function (result, error) {
                if (error) {
                  this.setError(error);
                  return callback(false);
                }
                //
                if (this.inserted && q.counter) {
                  for (var j = 0; j < md.properties.length; j++) {
                    var mdp = md.properties[j];
                    if (mdp.pk && mdp.datatype === App.Database.DataTypes.AUTOINCR) {
                      this[mdp.name] = this.dataStore.getLastInsertId();
                      this.reconnectFK(0);
                      break;
                    }
                  }
                }
                nextMD();
              }.bind(this));
              return;
            }
          }
        }
        nextMD();
      }
      else
        callback(true);
    }.bind(this);
    //
    nextMD();
  }.bind(this));
};


/**
 * Set fields FK - PK from fathers to sons
 * @param {Integer} childLevel
 * */
App.Document.prototype.reconnectFK = function (childLevel)
{
  var md = this.schema;
  for (var i = 0; i < md.collections.length; i++) {
    var mdc = md.collections[i];
    var c = this[mdc.name];
    var coll = c.getLoopCollection();
    for (var it = 0; it < coll.length; it++) {
      var d = coll[it];
      if (mdc.pks) {
        for (var j = 0; j < mdc.fks.length; j++)
          d[mdc.fks[j]] = this[mdc.pks[j]];
      }
      //
      if (childLevel > 1)
        d.reconnectFK(childLevel - 1);
    }
  }
};


/**
 * Return a good name for this document
 * @param{String} type
 * @param {Function} callback - function to be called at the end
 */
App.Document.prototype.getName = function (type, callback)
{
  var name = "";
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (mdp.datatype === App.Database.DataTypes.VARCHAR) {
      name = this[mdp.name];
      break;
    }
  }
  //
  var options = {name: name, type: type};
  this.onGetName(options, function (result, error) {
    callback(options.name, error);
  });
};


/**
 * Declare this object as deleted
 */
Object.defineProperty(App.Document.prototype, "deleted", {
  get: function () {
    return (this.intProps.deleted === true);
  },
  set: function (b) {
    // Set only to real document
    if (this.document !== this) {
      this.document.deleted = b;
      return;
    }
    //
    if (this.deleted !== b) {
      // Check Locking
      if (!this.freezeEvents && !this.getLock())
        return;
      //
      var wasVisible = this.isVisible();
      this.intProps.deleted = b;
      //
      // Update user interface
      if (wasVisible !== this.isVisible()) {
        this.updateUI(function (doc) {
          if (doc.datamap) {
            if (b)
              doc.datamap.updateElementsInvisibleRow(doc);
            else {
              var options = {};
              options.position = doc.datamap.findRowPosition(doc);
              doc.datamap.updateElementsVisibleRow(doc, options);
            }
          }
        });
      }
      //
      if (!this.freezeEvents)
        this.onDeleting();
      //
      this.markAsChanged();
    }
  }
});


/**
 * Returns true if this document or one of its parent are deleted
 */
App.Document.prototype.isDeleted = function ()
{
  if (this.deleted)
    return true;
  else if (this.parent)
    return this.parent.isDeleted();
  return false;
};


/**
 * Declare this object as inserted
 */
Object.defineProperty(App.Document.prototype, "inserted", {
  get: function () {
    return (this.intProps.inserted === true);
  },
  set: function (b) {
    // Set only to real document
    if (this.document !== this) {
      this.document.inserted = b;
      return;
    }
    //
    if (this.inserted !== b) {
      if (b)
        this.initDefaults(true);
      //
      this.intProps.inserted = b;
      //
      if (!this.freezeEvents)
        this.onInserting();
      //
      this.markAsChanged();
      //
      // Update user interface
      if (!this.inserted)
        this.updateUI(function (doc) {
          doc.updateElement();
        });
    }
  }
});


/**
 * Declare this object as updated
 */
Object.defineProperty(App.Document.prototype, "updated", {
  get: function () {
    return (this.intProps.originalValues !== undefined);
  },
  set: function (b) {
    // Set only to real document
    if (this.document !== this) {
      this.document.updated = b;
      return;
    }
    //
    if (this.updated !== b) {
      if (b)
        this.intProps.originalValues = this.intProps.originalValues || {};
      else
        this.intProps.originalValues = undefined;
      //
      if (!this.freezeEvents)
        this.onUpdating();
      //
      this.markAsChanged();
      //
      // Update user interface
      if (!this.updated)
        this.updateUI(function (doc) {
          doc.updateElement();
        });
    }
  }
});


/**
 * Declare this object as hidden
 */
Object.defineProperty(App.Document.prototype, "hidden", {
  get: function () {
    if (this.uiProps.hidden)
      return true;
    else if (this.document !== this)
      return this.document.hidden;
    else
      return false;
  },
  set: function (b) {
    if (this.uiProps.hidden !== b) {
      var wasVisible = [];
      this.document.updateUI(function (doc) {
        wasVisible.push(doc.isVisible());
      });
      this.uiProps.hidden = b;
      //
      this.document.updateUI(function (doc) {
        // Update user interface
        if (doc.datamap && wasVisible.shift() !== doc.isVisible()) {
          if (b)
            doc.datamap.updateElementsInvisibleRow(doc);
          else {
            var options = {};
            options.position = doc.datamap.findRowPosition(doc);
            doc.datamap.updateElementsVisibleRow(doc, options);
          }
        }
      });
    }
  }
});


/**
 * Notification events will be fired?
 * */
Object.defineProperty(App.Document.prototype, "freezeEvents", {
  get: function () {
    return this.intProps.freezeEvents || (this.intProps.stopEvents > 0);
  },
  set: function (b) {
    this.intProps.freezeEvents = b;
  }
});


/**
 * Stopping any freezable events
 * */
App.Document.prototype.stopEvents = function ()
{
  this.intProps.stopEvents = (this.intProps.stopEvents || 0) + 1;
};


/**
 * Restarting any freezable events
 * */
App.Document.prototype.startEvents = function ()
{
  this.intProps.stopEvents--;
};


/**
 * Event fired before loading data of this document
 * @param {Collection} collection
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.beforeLoad = function (collection, options, callback)
{
  callback();
};


/**
 * Event fired after loading the document
 * @param {Boolean} alreadyLoaded
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.afterLoad = function (alreadyLoaded, callback)
{
  if (callback)
    callback();
};


/**
 * Event fired when document is validating
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.onValidate = function (options, callback)
{
  callback();
};


/**
 * Event fired before document saving
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.onSave = function (options, callback)
{
  callback();
};


/**
 * Event fired when needs a good name for the document
 * @param {Function} callback - function to be called at the end
 * @param {Object} options
 * */
App.Document.prototype.onGetName = function (options, callback)
{
  callback();
};


/**
 * Event fired when document is inited
 * */
App.Document.prototype.onInit = function ()
{
};


/**
 * Event fired when document is cleared
 * */
App.Document.prototype.onClear = function ()
{
};


/**
 * Event fired when document need resync
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.onResyncClient = function (options, callback)
{
  if (callback)
    callback();
};


/**
 * Event (freezable) fired when a property is changing
 * @param {String} propName
 * @param {Object} params
 * */
App.Document.prototype.onPropertyChanged = function (propName, params)
{
};


/**
 * Event (freezable) fired when change inserted status
 * */
App.Document.prototype.onInserting = function ()
{
};


/**
 * Event (freezable) fired when change deleted status
 * */
App.Document.prototype.onDeleting = function ()
{
};


/**
 * Event (freezable) fired when change updated status
 * */
App.Document.prototype.onUpdating = function ()
{
};


/**
 * Event (freezable) I took a child
 * @param {Collection} coll
 * @param {Document} child
 * */
App.Document.prototype.onChildAdded = function (coll, child)
{
};


/**
 * Event (freezable) I lose a child
 * @param {Collection} coll
 * @param {Document} child
 * */
App.Document.prototype.onChildRemoved = function (coll, child)
{
};


/**
 * Event fired when a document variation is synchronized
 * @param {Object} params
 * @param {Function} callback - function to be called at the end
 */
App.Document.prototype.onGetTopics = function (params, callback)
{
  if (callback)
    callback();
};


/**
 * Search a resultset that match with text
 * @param {String} text
 * @param {Function} callback - function to be called at the end
 */
App.Document.prototype.search = function (text, callback)
{
  if (this.remote)
    this.rfc("search", [text], null, callback);
  else {
    var options = {};
    this.onSearch(text, options, function (result, error) {
      callback(options.datamap, error);
    });
  }
};


/**
 * Event fired when start an LKE search
 * @param {String} text to search
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Document.prototype.onSearch = function (text, options, callback)
{
  if (callback)
    callback();
};


/**
 * Gets a lock on the document
 * @param {Integer} parentLevel
 * @param {String} userInfo
 * @param {String} lockInfo
 * @returns {Boolean}
 * */
App.Document.prototype.getLock = function (parentLevel, userInfo, lockInfo)
{
  if (parentLevel === undefined)
    parentLevel = 9999;
  if (userInfo === undefined)
    userInfo = "";
  if (lockInfo === undefined)
    lockInfo = null;
  //
  try {
    // If the document is inserted there is no need to take the lock
    if (this.inserted)
      return true;
    //
    // If you already have a lock I finished
    if (this.locked)
      return true;
    //
    // Let's see if it supports locking
    if (false/* (MD.Services & MDOInit.SERV_LOCKING) != 0*/) {
      // OK, use the doc helper
      if (this.docHelper) {
        /*
         var ris = new IDVariant(IDVariant.TRUE);
         this.docHelper.getLock(this, new IDVariant(UserInfo), LockInfo, ris);
         if (ris.isTrue())
         this.locked = true;
         return ris.isTrue();
         */
      }
      else
        return false; // I can not say that the lock was taken !
    }
    else {
      // Let's see if I can ask the parent
      if (parentLevel > 0 && this.parent)
        return this.parent.getLock(parentLevel - 1, userInfo, lockInfo);
      else
        return true; // Reassure the caller ...
    }
  }
  catch (e) {
    return false;
  }
};


/**
 * Release the lock on the document
 * @param {Integer} parentLevel
 * @returns {Boolean}
 * */
App.Document.prototype.releaseLock = function (parentLevel)
{
  if (parentLevel === undefined)
    parentLevel = 9999;
  //
  try {
    var md = this.schema;
    //
    // Let's see if it supports locking
    if (false /*(md.Services & MDOInit.SERV_LOCKING) != 0*/) {
      // If it is not locked, I go out
      if (!this.locked)
        return false; // I can not say that the lock has been released!
      //
      // OK, use the doc helper
      if (this.docHelper) {
        this.locked = false;
        this.docHelper.releaseLock(this);
        return true;
      }
      else
        return false; // I can not say that the lock has been released!
    }
    else {
      // Let's see if I can ask the parent
      if (parentLevel > 0 && this.parent)
        return this.parent.releaseLock(parentLevel - 1);
      else
        return true; // Reassure the caller ...
    }
  }
  catch (e) {
    return false;
  }
};


/**
 * Locked status
 * */
Object.defineProperty(App.Document.prototype, "locked", {
  get: function () {
    return this.intProps.locked;
  },
  set: function (nv) {
    this.intProps.locked = nv;
  }
});


/**
 * Return identification of a document
 **/
App.Document.prototype.getDNA = function ()
{
  var md = this.schema;
  var dna = [];
  //
  // Add the className
  var cn = md.className;
  cn = cn.replace(/\./, "$");
  //
  // Preappend package for imported classes
  if (md.package)
    cn = md.package + "." + cn;
  dna.push(cn);
  //
  // Add values of primary keys
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (mdp.pk)
      dna.push(this[mdp.name]);
  }
  //
  dna = JSON.stringify(dna);
  return dna.substring(1, dna.length - 1);
};


/**
 * Return the document loaded by the DNA
 * @param {Session} app
 * @param {String} dna
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 **/
App.Document.loadByDNA = function (app, dna, options, callback)
{
  // Initialize default options
  options = Object.assign({skipLoad: false}, options);
  //
  var doc = null;
  try {
    var pks = JSON.parse("[" + dna.replace(/\\/g, "\\\\") + "]");
    var className = pks.shift();
    if (className.indexOf("$") > 0) {
      // Remove package
      if (className.indexOf(".") > 0)
        className = className.substring(className.lastIndexOf(".") + 1);
      className = className.replace(/\$/, ".");
    }
    //
    // Find the class by name
    var cls = App.Session.getClassByName(className);
    if (!cls)
      return callback(null, new Error("class " + className + " not found"));
    //
    // If it is a document, not good!
    if (!(cls.prototype instanceof App.Document))
      return callback(null, new Error("classCastException (className " + className + " is not an document)"));
    //
    // Create an instance of the class
    doc = new cls(app);
  }
  catch (e) {
    return callback(null, e);
  }
  //
  // Load primary key values
  var canBeLoaded = true;
  var md = doc.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (mdp.pk) {
      if (pks.length > 0)
        doc[mdp.name] = pks.shift();
      else
        canBeLoaded = false;
    }
  }
  //
  // Load document and childs if needed
  if (canBeLoaded && !options.skipLoad)
    doc.load(options, function (result, error) {
      callback(doc);
    });
  else
    callback(doc);
};


/**
 * Load a document given keys
 * @param {Session} app
 * @param {Object} key
 * @param {Object} options
 * @param {Function} callback
 * */
App.Document.loadByKey = function (app, key, options, callback)
{
  // If key is an Array or a simple value create template object
  var template = null;
  if (key instanceof Array || typeof key !== "object") {
    var keys = (key instanceof Array ? key.slice() : [key]);
    var md = this.schema;
    template = {};
    for (var i = 0; i < md.properties.length; i++) {
      var mdp = md.properties[i];
      if (mdp.pk)
        template[mdp.name] = keys.shift();
    }
  }
  else
    template = key;
  //
  // Load document from template
  var d = new this(app, template);
  d.load(options, function (result, error) {
    if (error)
      app.dttobj.addWarning("", error.message || error.toString());
    callback(d.loaded ? d : null);
  });
};


/**
 * Return a related document
 * @param {Function/Array} classList - list of class to load
 * @param {Object} options - options of load
 * @param {Function} callback - function to be called at the end
 **/
App.Document.prototype.getRelated = function (classList, options, callback)
{
  // Initialize default options
  options = Object.assign({
    cache: true,
    dataStore: this.dataStore
  }, options);
  //
  try {
    var cls = (classList instanceof Array ? classList[0] : classList);
    var md = this.schema;
    //
    // Find first related document definition for this classname
    var rd = null;
    for (var i = 0; i < md.relatedDocs.length; i++) {
      if (md.relatedDocs[i].class === cls.schema.className) {
        rd = md.relatedDocs[i];
        break;
      }
    }
    //
    if (!rd)
      return callback(null);
    //
    // Create an instance of the class
    var doc = new cls(this.app);
    //
    // Set primary key values
    for (var i = 0; i < rd.fks.length; i++)
      doc[rd.pks[i]] = this[rd.fks[i]];
    //
    var afterLoadCallback = function (result, error) {
      if (error || !doc.loaded)
        return callback(null);
      //
      // If there are more class to load ...
      if (classList instanceof Array && classList.length > 1)
        doc.getRelated(classList.slice(1), options, callback);
      else
        callback(doc);
    };
    //
    // Load the document
    doc.load(options, afterLoadCallback);
  }
  catch (e) {
    return callback(null);
  }
};


/**
 * Event fired when document is changed
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.onChange = function (callback)
{
  callback();
};


/**
 * Record that the document has changed
 * */
App.Document.prototype.markAsChanged = function ()
{
  this.updateUI(function (doc) {
    if (doc)
      doc.updateElement();
  });
  if (this.notifyOnChange() && !this.changed)
    this.changed = setImmediate(App.Document.prototype.processChange.bind(this));
  //
  if (this.parent)
    this.parent.markAsChanged();
};


/**
 * Notified that the document has changed
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.processChange = function (callback)
{
  if (!this.changed) {
    if (callback)
      callback();
    return;
  }
  //
  var afterOnChange = function (result, error) {
    clearImmediate(this.changed);
    delete this.changed;
    //
    if (callback)
      callback(result, error);
  }.bind(this);
  //
  try {
    this.onChange(afterOnChange);
  }
  catch (e) {
    afterOnChange(null, e);
  }
};


/**
 * Check if onChange event is implemented
 * */
App.Document.prototype.notifyOnChange = function ()
{
  return (this.class.prototype.onChange !== App.Document.prototype.onChange);
};


/**
 * Generic setter of each property
 * @param {String} propName
 * @param {AnyType} newValue
 * @param {Bool} flTransient
 * */
App.Document.prototype.changePropDirect = function (propName, newValue, flTransient)
{
  // Set only to real document
  if (this.document !== this)
    return this.document.changePropDirect(propName, newValue, flTransient);
  //
  // If new value is equal do nothing
  var oldValue = this[propName];
  if (App.Locale.compareValues(newValue, oldValue))
    return;
  //
  var oldOrgValue = this.getOriginalValue(propName);
  this.setProp(propName, newValue);
  this.setOriginalValue(propName, oldOrgValue);
  //
  if (!flTransient && !this.changingProperty(propName)) {
    this.setProp(propName, oldValue);
    this.setOriginalValue(propName, oldOrgValue);
  }
  else {
    // Update user interface
    this.updateUI(function (doc) {
      // Notify change value to datamap observer
      doc.notifyChangeToParentObserver("update", propName, oldValue);
      //
      doc.updateElement();
      if (doc.datamap)
        doc.datamap.updateElementsData(doc);
    });
  }
};


/**
 * Internal function called when a property is changing
 * @param {String} propName
 * */
App.Document.prototype.changingProperty = function (propName)
{
  var cancel = false;
  //
  if (!this.freezeEvents) {
    this.onPropertyChanged(propName, cancel);
    if (!cancel && !this.getLock())
      cancel = true;
  }
  //
  if (!cancel) {
    this.updated = true;
    //
    /*
     if (this.collection)
     this.collection.changeDoc(this, propName);
     */
    return true;
  }
  //
  return false;
};


/**
 * Get a property by name
 * @param {String} prop
 * */
App.Document.prototype.getProp = function (prop)
{
  return this.values[prop];
};


/**
 * Get a collection by name
 * @param {String} coll
 * */
App.Document.prototype.getColl = function (coll)
{
  var c = this.intProps["coll_" + coll];
  if (!c) {
    c = new App.Collection();
    this.setColl(coll, c);
  }
  return c;
};


/**
 * Set a property directly
 * @param {String} prop
 * @param {AnyType} newValue
 * @param {Boolean} convert value to column datatype
 * */
App.Document.prototype.setProp = function (prop, newValue, convert)
{
  this.assignValue(prop, newValue, convert);
};


/**
 * Return the container of a value for a column
 * @param {DataColumn} dc
 */
App.Document.prototype.getValueContainer = function (dc)
{
  if (dc.extended) {
    if (!this.uiProps.values)
      this.uiProps.values = {};
    return this.uiProps.values;
  }
  //
  return App.DataRow.prototype.getValueContainer.call(this, dc);
};


/**
 * Set a collection directly
 * @param {String/Integer} coll
 * @param {Collection} newValue
 * */
App.Document.prototype.setColl = function (coll, newValue)
{
  var mdc = this.schema.getCollDefinition(coll);
  if (newValue) {
    newValue.parent = this;
    //
    // Set docTemplate of the collection
    newValue.index = mdc.index;
    newValue.orderBy = mdc.orderBy;
    if (!newValue.rowTemplate && mdc.contentType) {
      var cls = App.Session.getClassByName(mdc.contentType);
      if (cls)
        newValue.rowTemplate = new cls(this.app);
    }
  }
  //
  this.intProps["coll_" + mdc.name] = newValue;
};


/**
 * Load a collection by template
 * @param {Session} app
 * @param {Object} template
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.loadCollection = function (app, template, options, callback)
{
  options = Object.assign({}, options);
  var coll = new App.Collection(app);
  var doc = new this(app);
  coll.searchOptions.likeType = (options.useQBE ? App.DataMap.likeTypes.startsWith : App.DataMap.likeTypes.noLike);
  doc.loadDocProp(template, !options.useQBE);
  doc.loadCollectionByExample(coll, options, function (result, error) {
    callback(coll, error);
  });
};


/**
 * Load a collection by example
 * @param {String/Integer} coll - collection to populate
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.loadCollectionByExample = function (coll, options, callback)
{
  // Initialize default options
  options = Object.assign({useQBE: false}, options);
  //
  // Set docTemplate of the collection
  coll.rowTemplate = this;
  //
  // Set a collection filter for each property valued
  var md = this.schema;
  coll.clearFilters();
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    var v = this[mdp.name];
    if (v === undefined || v === null)
      continue;
    //
    coll.addFilter(mdp.name, v, false, options.useQBE);
  }
  //
  // Ask to collection to load data
  coll.load(options, function (result, error) {
    callback(coll.rows.length, error);
  });
};


/**
 * Load document children
 * @param {Integer} childLevel - level of childs to load
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.loadChildren = function (childLevel, callback)
{
  // Do not load document inserted (are not in the dm!)
  if (this.inserted)
    return callback(false);
  //
  if (childLevel >= 0) {
    // If coll has no rows, return
    var md = this.schema;
    if (md.collections.length === 0)
      return callback(false);
    //
    var firstError;
    var res = false;
    var collLen = md.collections.length;
    var done = function (result, error) {
      firstError = firstError || error;
      if (--collLen === 0)
        callback(res, firstError);
    };
    //
    // Load childrens
    for (var i = 0; i < md.collections.length; i++) {
      var coll = this[md.collections[i].name];
      if (!coll.loaded)
        res = true;
      this.loadChildCollection(coll, {childLevel: childLevel}, done);
    }
  }
  else
    callback(false);
};


/**
 * Load a collection from datamodel
 * @param {Collection} coll - collection to load
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.loadChildCollection = function (coll, options, callback)
{
  options = Object.assign({}, options);
  //
  // Do not load documents inserted (are not in the dm!)
  if (coll.parent && coll.parent.inserted)
    return callback();
  //
  // Load my data only if not already loaded or if reload required
  if (!coll.loaded || options.reload) {
    this.stopEvents();
    //
    // Ask to collection to load data
    coll.searchOptions.likeType = App.DataMap.likeTypes.noLike;
    options.loadType = App.Document.loadTypes.loadChildCollection;
    coll.load(options, function (result, error) {
      this.startEvents();
      callback(null, error);
    }.bind(this));
  }
  else if (options.childLevel > 0) {
    // If coll has no rows, return
    if (coll.rows.length === 0)
      return callback();
    //
    var firstError;
    var docLen = coll.rows.length;
    var done = function (result, error) {
      firstError = firstError || error;
      if (--docLen === 0)
        callback(null, firstError);
    }.bind(this);
    //
    // Load childrens
    for (var i = 0; i < coll.rows.length; i++)
      coll.rows[i].load({childLevel: options.childLevel - 1}, done);
  }
  else
    callback();
};


/**
 * Reload the document from datamodel
 * @param {Object} options - options of load
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.reload = function (options, callback)
{
  options = Object.assign({}, options);
  options.reload = true;
  this.load(options, callback);
};


/**
 * Load the document from datamodel
 * @param {Object} options - options of load
 * @param {Function} callback - function to be called at the end
 * */
App.Document.prototype.load = function (options, callback)
{
  // Initialize default options
  options = Object.assign({
    childLevel: 0,
    useQBE: false
  }, options);
  //
  // Set supplied dataStore to use for load
  var oldDataStore = this.intProps.dataStore;
  if (options.dataStore) {
    this.dataStore = options.dataStore;
    delete options.dataStore;
  }
  //
  var id = this.getDNA().split(",").slice(1).join(",");
  var done = function (result, error) {
    // Restore old dataStore
    this.dataStore = oldDataStore;
    //
    // Call pending callbacks
    if (options.cache) {
      if (this.class.loadFromCacheCallbacks[id]) {
        this.class.loadFromCacheCallbacks[id].forEach(setImmediate);
        delete this.class.loadFromCacheCallbacks[id];
      }
    }
    //
    if (callback)
      callback(result, error);
  }.bind(this);
  //
  // Do not load document inserted (are not in the dm!)
  if (this.inserted)
    return done(false);
  //
  // If required load from cache search that
  if (options.cache) {
    if (this.class.docCache && id in this.class.docCache)
      this.copyFrom(this.class.docCache[id], 0);
    else {
      if (!this.class.loadFromCacheCallbacks)
        this.class.loadFromCacheCallbacks = {};
      var cbs = this.class.loadFromCacheCallbacks[id];
      //
      // Handle concurrent calls for load of same doc from cache
      if (!cbs)
        cbs = this.class.loadFromCacheCallbacks[id] = [];
      else
        return cbs.push(function () {
          this.load(options, callback);
        }.bind(this));
    }
  }
  //
  if (this.loaded && !options.reload) {
    this.loadChildren(options.childLevel - 1, function (ndl, error) {
      if (error)
        return done(null, error);
      //
      if (ndl) {
        // If I had loaded at least one collection ...
        this.afterLoad(true, function (result, error) {
          this.markAsChanged();
          done(true, error);
        }.bind(this));
      }
      else
        done(false);
    }.bind(this));
    return;
  }
  //
  var coll = new App.Collection();
  //
  // Set docTemplate of the collection
  coll.rowTemplate = this;
  //
  var md = this.schema;
  //
  // Set a collection filter for each property pk valued
  coll.clearFilters();
  var pkdef = true;
  var npk = 0;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (mdp.pk) {
      npk++;
      var v = this[mdp.name];
      if (v === undefined || v === null) {
        pkdef = false;
        break;
      }
      else
        coll.addFilter(mdp.name, v, false, options.useQBE);
    }
  }
  //
  if (npk === 0)
    pkdef = false;
  //
  // If not all pk properties are valued set collection filter for all other properties
  if (!pkdef) {
    coll.clearFilters();
    for (var i = 0; i < md.properties.length; i++) {
      var mdp = md.properties[i];
      var v = this[mdp.name];
      if (v === undefined || v === null)
        continue;
      //
      pkdef = true;
      coll.addFilter(mdp.name, v, false, options.useQBE);
    }
  }
  //
  if (!pkdef)
    return done(null, new Error("No property valued"));
  //
  // Ask to collection to load data
  coll.searchOptions.likeType = App.DataMap.likeTypes.noLike;
  options.loadType = App.Document.loadTypes.load;
  coll.load(options, function (result, error) {
    if (error)
      return done(null, error);
    //
    if (coll.rows.length === 0)
      return done(false);
    //
    if (coll.rows.length > 1)
      return done(null, new Error("The document was not loaded because it has been found more than one result"));
    //
    this.stopEvents();
    //
    // Copy values
    var srcDoc = coll.rows[0];
    this.loadDocProp(srcDoc);
    //
    // Copy collections
    if (options.childLevel > 0) {
      for (var i = 0; i < md.collections.length; i++) {
        var mdc = md.collections[i];
        var dstColl = this[mdc.name];
        var srcColl = srcDoc[mdc.name].getLoopCollection();
        //
        // Detach elements to save them
        dstColl.reuseElements(srcDoc[mdc.name].uiRows.length);
        dstColl.clear();
        for (var j = 0; j < srcColl.length; j++)
          dstColl.add(srcColl[j]);
        dstColl.removeUnusedElements();
        dstColl.loaded = true;
        //
        // Update listBox rowCount
        dstColl.updateUI(function (dm) {
          if (dm.useWindow) {
            dm.schema.listBox.rowCount = dm.uiRows.length;
            Object.getNotifier(dm.schema.listBox).notify({type: "update", name: 'rowCount', oldValue: dm.schema.listBox.rowCount});
          }
        });
      }
    }
    //
    this.setOriginal(options.childLevel);
    this.loaded = true;
    //
    this.afterLoad(false, function (result, error) {
      this.markAsChanged();
      this.startEvents();
      //
      // Handle cache
      if (options.cache) {
        // Init cache if not exists
        if (!this.class.docCache)
          this.class.docCache = {};
        //
        // Add a clone to cache
        if (!this.class.docCache[id])
          this.class.docCache[id] = this.clone(0);
      } // Update cache
      else if (this.class.docCache && this.class.docCache[id])
        this.class.docCache[id] = this.clone(0);
      //
      done(true, error);
    }.bind(this));
  }.bind(this));
};


/**
 * Create the query for this document
 * @param {Object} md - metaschema for witch build the query
 * @param {Boolean} aliasTable - true if almost one parent derived from a table
 * @returns{Object}
 * */
App.Document.prototype.getLoadQuery = function (md, aliasTable)
{
  if (!md)
    md = this.schema;
  //
  // If document has a custom load query use that
  if (md.query)
    return md.query;
  //
  var q = null;
  if (md.baseMD) {
    if (!aliasTable && md.dmCode)
      aliasTable = true;
    q = this.getLoadQuery(md.baseMD, aliasTable);
  }
  //
  if (!md.dmCode)
    return q;
  //
  if (!q) {
    if (!md.datamodel)
      return null;
    //
    q = {};
    q.dm = md.datamodel;
    q.type = App.Database.QueryTypes.SELECT;
    q.select = [];
    q.from = [];
  }
  //
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (mdp.dmCode && (!md.baseMD || i >= md.baseMD.properties.length)) {
      var tk = {expr: mdp.dmCode};
      tk.from = (aliasTable ? md.dmCode : "");
      if (tk.expr !== mdp.name)
        tk.alias = mdp.name;
      q.select.push(tk);
    }
  }
  //
  var fromTk = {expr: md.dmCode};
  if (q.from.length > 0) {
    fromTk.join = "inner join";
    fromTk.p = [];
    var keys = Object.keys(md.fieldMap);
    for (var i = 0; i < keys.length; i++) {
      if (i > 0)
        fromTk.p.push({expr: "and"});
      fromTk.p.push({expr: keys[i], from: md.baseMD.dmCode});
      fromTk.p.push({expr: "="});
      fromTk.p.push({expr: md.fieldMap[keys[i]], from: md.dmCode});
    }
  }
  q.from.push(fromTk);
  //
  return q;
};


/**
 * Return the schema of this row
 */
App.Document.prototype.getSchema = function ()
{
  if (this.datamap && this.datamap.schema)
    return this.datamap.schema;
  else
    return this.schema;
};


/**
 * Return the class of this document
 * @returns{Function}
 * */
Object.defineProperty(App.Document.prototype, "class", {
  get: function () {
    return App.Document;
  }
});


/**
 * Return the className of this document
 * */
App.Document.getClassName = function ()
{
  return "Document";
};


/**
 * Return the meta schema of the document
 * */
Object.defineProperty(App.Document.prototype, "schema", {
  get: function () {
    return this.class.schema;
  }
});


/**
 * Load document property from given data object
 * @param {Object} data - object with data to load
 * @param {Boolean} convert value to column datatype
 * */
App.Document.prototype.loadDocProp = function (data, convert)
{
  if (!data)
    return;
  //
  // The object properties are the actual row values
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (mdp.name in data)
      this.setProp(mdp.name, data[mdp.name], convert);
  }
};


/**
 * Return true if the column is unbound
 * @param {String} columnName
 */
App.Document.prototype.isColumnUnbound = function (columnName)
{
  var mdp = this.schema.getPropDefinition(columnName);
  return !(mdp && mdp.dmCode);
};


/**
 * Close the document
 */
App.Document.prototype.close = function ()
{
  App.DataRow.prototype.close.call(this);
  //
  // If I'm a reference document remove reference
  if (this.collection !== this.datamap && this.document.refDocuments) {
    var i = this.document.refDocuments.indexOf(this);
    if (i >= 0)
      this.document.refDocuments.splice(i, 1);
  }
};


/**
 * Update UI
 * @param {boolean} resetTimer
 */
App.Document.prototype.updateElement = function (resetTimer)
{
  if (this.datamap)
    App.DataRow.prototype.updateElement.call(this, resetTimer);
};


/**
 * One of my element has changed
 * @param {RemElem} el - element changed
 * @param {String} n - property name
 * @param {String} v - property value
 */
App.Document.prototype.chgProp = function (el, n, v)
{
  var column = App.DataRow.prototype.chgProp.call(this, el, n, v);
  if (column) {
    var opts = {};
    opts.reason = App.Document.validationReasons.change;
    opts.property = column.name;
    opts.childLevel = 0;
    this.validate(opts);
  }
};


/**
 * Return query for update data of this document
 * @param {DocumentSchema} md
 */
App.Document.prototype.getUpdateQuery = function (md)
{
  // If the document is not linked to a table nothing query
  if (!md.dmCode)
    return null;
  //
  var q = {
    type: App.Database.QueryTypes.UPDATE,
    table: md.dmCode,
    select: [],
    where: []
  };
  //
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    //
    // Skip unbound column
    if (!mdp.dmCode || !mdp.datatype)
      continue;
    //
    var code = mdp.dmCode;
    if (mdp.pk && md.baseMD && mdp.dmCode in md.fieldMap)
      code = md.fieldMap[mdp.dmCode];
    //
    // Add to the list of fields to be updated only if it is part of this md and if value is changed
    if (!md.baseMD || i >= md.baseMD.properties.length) {
      if (mdp.name in this.originalValues) {
        // BLOB as parameter
        if (mdp.datatype === App.Database.DataTypes.BLOB)
          q.select.push({field: code, value: this[mdp.name], asPar: true, dt: mdp.datatype});
        else
          q.select.push({field: code, expr: mdp.toSql(this[mdp.name], this.dataStore)});
      }
    }
    //
    if (mdp.pk || (md.optimisticLocking && mdp.isComparable())) {
      var orgValue = this.getOriginalValue(mdp.name);
      if (orgValue === null || orgValue === undefined)
        q.where.push({expr: "is null", p: [{expr: code, from: ""}]});
      else
        q.where.push({expr: "=", p: [{expr: code, from: ""}, {expr: mdp.toSql(orgValue, this.dataStore)}]});
    }
  }
  //
  if (q.select.length > 0)
    return q;
  else
    return null;
};


/**
 * Return query for delete this row
 * @param {DocumentSchema} md
 */
App.Document.prototype.getDeleteQuery = function (md)
{
  // If the document is not linked to a table nothing query
  if (!md.dmCode)
    return null;
  //
  var q = {
    type: App.Database.QueryTypes.DELETE,
    table: md.dmCode,
    where: []
  };
  //
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    //
    // Skip unbound column
    if (!mdp.dmCode || !mdp.datatype)
      continue;
    //
    var code = mdp.dmCode;
    if (mdp.pk && md.baseMD && mdp.dmCode in md.fieldMap)
      code = md.fieldMap[mdp.dmCode];
    //
    if (mdp.pk || (md.optimisticLocking && mdp.isComparable())) {
      var orgValue = this.getOriginalValue(mdp.name);
      if (orgValue === null || orgValue === undefined)
        q.where.push({expr: "is null", p: [{expr: code, from: ""}]});
      else
        q.where.push({expr: "=", p: [{expr: code, from: ""}, {expr: mdp.toSql(orgValue, this.dataStore)}]});
    }
  }
  //
  return q;
};


/**
 * Return query for insert this row
 * @param {DocumentSchema} md
 */
App.Document.prototype.getInsertQuery = function (md)
{
  // If the document is not linked to a table nothing query
  if (!md.dmCode)
    return null;
  //
  var q = {
    type: App.Database.QueryTypes.INSERT,
    table: md.dmCode,
    select: []
  };
  //
  // Mark query to use insert or replace for loaded document or for datamap that require it
  if ((this.datamap && this.datamap.upsert) || this.loaded)
    q.upsert = true;
  //
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    var value = this[mdp.name];
    //
    // Set counter column (for Postgres)
    if ((!md.baseMD || i >= md.baseMD.properties.length) && mdp.datatype === App.Database.DataTypes.AUTOINCR)
      q.counter = mdp.dmCode;
    //
    // Skip unbound column and omissis values
    if (!mdp.dmCode || value === undefined)
      continue;
    //
    var code = mdp.dmCode;
    if (mdp.pk && md.baseMD && mdp.dmCode in md.fieldMap)
      code = md.fieldMap[mdp.dmCode];
    //
    // Add to the list of fields to be updated only if it is part of this md
    if (!md.baseMD || i >= md.baseMD.properties.length) {
      // BLOB as parameter
      if (mdp.datatype === App.Database.DataTypes.BLOB)
        q.select.push({field: code, value: value, asPar: true, dt: mdp.datatype});
      else
        q.select.push({field: code, expr: mdp.toSql(value, this.dataStore)});
    }
    //
    // Add to the list of fields to be updated also pk fields
    if (md.baseMD && i < md.baseMD.properties.length && mdp.pk)
      q.select.push({field: code, expr: mdp.toSql(value, this.dataStore)});
  }
  //
  if (q.select.length > 0)
    return q;
  else
    return null;
};


/**
 * Set document or property error
 * @param {String} text - the text of the error
 * @param {String} propName - the name of the property to set in error
 */
App.Document.prototype.setError = function (text, propName)
{
  // Keep only message of error
  if (text instanceof Error)
    text = text.message;
  //
  // Only if is changed
  if (this.getError(propName) === text)
    return;
  //
  // Set only to real document
  if (this.document !== this) {
    this.document.setError(text, propName);
    return;
  }
  //
  // If you did not specify the column set the error on the row
  if (!propName)
    this.rowError = text;
  else {
    if (text)
      this.columnsErrors[propName] = text;
    else
      delete this.columnsErrors[propName];
  }
  //
  // During validation don't update the UI
  if (this.intProps.validating)
    return;
  //
  this.updateUI(function (doc) {
    doc.updateElement();
    //
    // Notify onError to datamap
    if (doc.datamap && text) {
      var info = {};
      info.source = App.DataMap.errorSources.validation;
      info.row = doc;
      if (propName)
        info.column = propName;
      doc.datamap.onError(text, info);
    }
  });
};


/**
 * Call a function for all datamap in which is shown
 * @param {Function} funct
 */
App.Document.prototype.updateUI = function (funct)
{
  // Call function for my owner collection
  funct(this);
  //
  // Call function for all datamap in which is shown
  if (this.datamap && this.datamap.linkedDataMaps) {
    var p = this.index;
    if (p >= 0) {
      for (var i in this.datamap.linkedDataMaps)
        funct(this.datamap.linkedDataMaps[i].rows[p]);
    }
  }
  //
  // Call function for all document that refer to me
  if (this.refDocuments) {
    for (var i in this.refDocuments)
      this.refDocuments[i].updateUI(funct);
  }
};


/**
 * Returns true if document was created for handle a variation
 */
App.Document.prototype.isSynchronizing = function ()
{
  return (this.syncVariation ? true : false);
};


/**
 * Find a property by name
 * @param {String} name
 * @param {Object} options
 */
App.Document.prototype.findPropByName = function (name, options)
{
  // Initialize default options
  options = Object.assign({
    byName: true,
    byTag: true
  }, options);
  //
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (options.byName && name === mdp.name)
      return mdp.name;
    if (options.byTag && name === mdp.tag)
      return mdp.name;
    if (options.byDMCode && name === mdp.dmCode)
      return mdp.name;
  }
  return "";
};


/**
 * Duplicate a document
 * @param {Integer} childLevel - level of children to duplicate
 * @param {Boolean} childDoc - indicate if is not the root document
 */
App.Document.prototype.duplicate = function (childLevel, childDoc)
{
  if (childLevel === undefined)
    childLevel = 9999;
  //
  // If is the main call create the object by copying
  var dcopy = (childDoc ? this : this.clone(childLevel));
  //
  dcopy.inserted = true;
  //
  // Generate a new uuid
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (mdp.pk && mdp.datatype === App.Database.DataTypes.UUID) {
      if (App.Utils["generateUID" + mdp.maxlength])
        dcopy[mdp.name] = App.Utils["generateUID" + mdp.maxlength]();
      dcopy.reconnectFK(0);
    }
  }
  //
  dcopy.onDuplicate(childDoc ? null : this);
  //
  if (childLevel > 0) {
    for (var i = 0; i < md.collections.length; i++) {
      var coll = dcopy[md.collections[i].name].getLoopCollection();
      for (var it = 0; it < coll.length; it++)
        coll[it].duplicate(childLevel - 1, true);
    }
  }
  //
  return dcopy;
};


/**
 * Event fired when document is duplicating
 * @param {Document} srcDoc - original document
 * */
App.Document.prototype.onDuplicate = function (srcDoc)
{
};


/**
 * Clone a document
 * @param {Integer} childLevel - level of children to clone
 */
App.Document.prototype.clone = function (childLevel)
{
  var d = new (this.class)(this.app);
  d.copyFrom(this, childLevel);
  return d;
};


/**
 * Copy from another document
 * @param {Document} doc - document to copy
 * @param {Integer} childLevel - level of children to clone
 */
App.Document.prototype.copyFrom = function (doc, childLevel)
{
  // First copy attributes and errors
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    var v = doc[mdp.name];
    //
    // Date, Array and Object must be treated differently
    if (v && typeof v === "object") {
      if (v instanceof Date)
        v = new Date(v);
      else if (mdp.datatype === App.Database.DataTypes.JSON)
        v = JSON.parse(JSON.stringify(v));
      else if (v instanceof Array || v instanceof ArrayBuffer)
        v = v.slice();
    }
    //
    this.changePropDirect(mdp.name, v, true);
    this.setError(doc.getError(mdp.name), mdp.name);
  }
  this.setError(doc.getError());
  //
  // Now the states
  this.updated = doc.updated;
  this.inserted = doc.inserted;
  this.deleted = doc.deleted;
  this.hidden = doc.hidden;
  this.loaded = doc.loaded;
  //
  // Now the children
  if (childLevel > 0) {
    for (var i = 0; i < md.collections.length; i++) {
      var mdc = md.collections[i];
      var cs = doc[mdc.name];
      var cd = this[mdc.name];
      for (var it = 0; it < cs.rows.length; it++) {
        var ds = cs.rows[it];
        var dna = ds.getDNA();
        var found = false;
        for (var j = 0; j < cd.rows.length; j++) {
          var dd = cd.rows[j];
          if (dd.getDNA() === dna) {
            dd.copyFrom(ds, childLevel - 1);
            cd.add(dd);
            found = true;
            break;
          }
        }
        if (!found)
          cd.add(ds.clone(childLevel - 1));
      }
      //
      // Remove documents those not found in source collection:
      // Are all documents at top of collection because those found are be readded at bottom
      while (cd.length > cs.length)
        cd.remove(cd.rows[0]);
      //
      cd.loaded = cs.loaded;
    }
  }
};


/**
 * Save to JSON
 * @param {Object} options
 */
App.Document.prototype.intoJSON = function (options)
{
  // Initialize default options
  options = Object.assign({
    skipStatus: true,
    skipType: "base",
    skipNulls: true,
    skipOriginals: true,
    skipEmptyColls: true,
    childLevel: 9999,
    skipErrors: true,
    asString: true
  }, options);
  //
  var json = {};
  //
  // Add statuses
  if (!options.skipStatus) {
    json["@s.inserted"] = this.inserted;
    json["@s.updated"] = this.updated;
    json["@s.deleted"] = this.deleted;
    json["@s.loaded"] = this.loaded;
    json["@s.hidden"] = this.hidden;
  }
  //
  // Add type name
  var md = this.schema;
  var skipType = false;
  if (options.skipType === "always")
    skipType = true;
  if (options.skipType === "base") {
    if (this.collection)
      skipType = (this.collection.schema.className === md.className);
    else
      skipType = true;
  }
  if (!skipType)
    json["@s.type"] = md.className;
  //
  // Add doc error
  if (!options.skipErrors && this.getError())
    json["@s.error"] = this.getError();
  //
  var asString = options.asString;
  options.asString = false;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    var v = this[mdp.name];
    //
    // Some tranformations
    if (v === undefined)
      v = null;
    if (v) {
      if ([App.Database.DataTypeClasses.DATETIME, App.Database.DataTypeClasses.BLOB].indexOf(App.Database.classDT(mdp.datatype)) >= 0)
        v = mdp.valueToString(v, null, this.app);
      else if (v instanceof App.Document)
        v = v.intoJSON(options);
    }
    //
    // Add value
    if (v !== null || !options.skipNulls)
      json[mdp.name] = v;
    //
    // Add error
    if (!options.skipErrors && this.getError(mdp.name))
      json["@e." + mdp.name] = this.getError(mdp.name);
    //
    if (options.skipOriginals || !this.updated || !(mdp.name in this.originalValues))
      continue;
    //
    // Add original value
    var o = this.originalValues[mdp.name];
    //
    // Some tranformations
    if (o === undefined)
      o = null;
    else if (o instanceof Date)
      o = mdp.valueToString(o, null, this.app);
    //
    json["@o." + mdp.name] = o;
  }
  //
  // Add collections
  if (options.childLevel > 0) {
    options.childLevel--;
    for (var i = 0; i < md.collections.length; i++) {
      var mdc = md.collections[i];
      var coll = this[mdc.name];
      if (options.skipEmptyColls && coll.rows.length === 0)
        continue;
      //
      json[mdc.name] = coll.intoJSON(options);
    }
  }
  //
  // If replacer function has been supplied call it
  App.Document.JSONReplacer(this, json, options);
  //
  return (asString ? JSON.stringify(json) : json);
};


/**
 * Load from JSON
 * @param {Object/String} json
 * @param {Object} options
 */
App.Document.prototype.loadFromJSON = function (json, options)
{
  // Initialize default options
  options = Object.assign({childLevel: 9999}, options);
  //
  if (typeof json === "string")
    json = JSON.parse(json);
  //
  if (!options.child) {
    this.stopEvents();
    this.clear();
  }
  //
  // If replacer function has been supplied call it
  App.Document.JSONReplacer(this, json, options);
  //
  var md = this.schema;
  Object.keys(json).forEach(function (k) {
    var v = json[k];
    //
    // Statuses and doc error
    if (k.startsWith("@s.")) {
      var st = k.substring(3);
      switch (st) {
        case "inserted":
        case "updated":
        case "deleted":
        case "loaded":
        case "hidden":
          if (!options.skipStatus)
            this[st] = v;
          return;

        case "error":
          if (!options.skipErrors)
            this.setError(v);
          return;
      }
    }
    //
    if (k === "@s.type")
      return;
    //
    // Properties (actual and original values, errors)
    var orgValue = k.startsWith("@o.");
    if (orgValue && options.skipOriginals)
      return;
    //
    var errValue = k.startsWith("@e.");
    if (errValue && options.skipErrors)
      return;
    //
    var propName = k;
    if (orgValue || errValue)
      propName = k.substring(3);
    //
    for (var i = 0; i < md.properties.length; i++) {
      var mdp = md.properties[i];
      if (mdp.name === propName) {
        // Some transformations
        if (v && !errValue) {
          if (mdp.objectType)
            v = App.Session.getClassByName(mdp.objectType).fromJSON(this.app, v, options);
          //
          if ([App.Database.DataTypeClasses.DATETIME, App.Database.DataTypeClasses.BLOB].indexOf(App.Database.classDT(mdp.datatype)) >= 0)
            v = mdp.stringToValue(v, null, this.app);
        }
        //
        if (orgValue)
          this.setOriginalValue(propName, v);
        else if (errValue)
          this.setError(v, propName);
        else
          this.setProp(propName, v);
        //
        return;
      }
    }
    //
    // Collections
    for (var i = 0; i < md.collections.length; i++) {
      var mdc = md.collections[i];
      if (mdc.name === k) {
        if (options.childLevel > 0) {
          var opts = Object.assign({}, options);
          opts.childLevel--;
          opts.child = true;
          this[mdc.name].loadFromJSON(v, opts);
        }
        return;
      }
    }
    //
    // Attribute is unknown; set it as unbound property
    this[k] = v;
  }.bind(this));
  //
  if (!options.child)
    this.startEvents();
};


/**
 * Create a document from JSON
 * @param {Session} app
 * @param {Object/String} json
 * @param {Object} options
 */
App.Document.fromJSON = function (app, json, options)
{
  // If method is called by App.Document find type into json
  var cls;
  if (this === App.Document) {
    if (typeof json === "string")
      json = JSON.parse(json);
    //
    var typeName = json["@s.type"];
    if (!typeName)
      throw new Error("Class not defined");
    //
    cls = App.Session.getClassByName(typeName);
    if (!cls)
      throw new Error("Class " + typeName + " not found");
  }
  else
    cls = this;
  //
  var d = new cls(app);
  d.loadFromJSON(json, options);
  return d;
};


/**
 * Tells something about this object to the DTT module
 * @param {Integer} childLevel - level of child to log
 */
App.Document.prototype.getDebugInfo = function (childLevel)
{
  var info = App.DataRow.prototype.getDebugInfo.call(this);
  info._class = this.schema.className;
  info.loaded = this.loaded;
  if (childLevel > 0) {
    var md = this.schema;
    for (var i = 0; i < md.collections.length; i++) {
      var mdc = md.collections[i];
      info[mdc.name] = this[mdc.name].getDebugInfo(childLevel - 1);
    }
  }
  return info;
};


/**
 * Returns all document in hierarchy
 * @param {Integer} childLevel
 */
App.Document.prototype.getAllDocuments = function (childLevel)
{
  childLevel = childLevel || 9999;
  //
  // Add myself
  var docs = [this];
  //
  // Add all document properties
  var md = this.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (mdp.objectType && this[mdp.name])
      docs.push(this[mdp.name]);
  }
  //
  // Add all documents of collections
  if (childLevel > 0) {
    for (var i = 0; i < md.collections.length; i++)
      Array.prototype.push.apply(docs, this[md.collections[i].name].getAllDocuments(childLevel - 1));
  }
  //
  return docs;
};


/**
 * Returns a string with all errors
 * @param {Integer} childLevel
 */
App.Document.prototype.getDocErrors = function (childLevel)
{
  if (childLevel === undefined)
    childLevel = 9999;
  //
  var md = this.schema;
  var s = "*** " + this.getDNA() + "\n";
  //
  // Add properties errors
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    var e = this.getError(mdp.name);
    if (e)
      s += e + "\n";
  }
  //
  // Add document error
  if (this.getError())
    s += this.getError() + "\n";
  //
  // Now the children
  if (childLevel > 0) {
    for (var i = 0; i < md.collections.length; i++) {
      var coll = this[md.collections[i].name];
      for (var ii = 0; ii < coll.length; ii++)
        s += coll[ii].getDocErrors(childLevel - 1);
    }
  }
  //
  return s;
};


/**
 * Informs all that document is changed
 * @param {Object} options
 */
App.Document.prototype.notifyDocUpdate = function (options)
{
  this.app.sync.DO.notifyDocUpdate(this, options);
};


/*
 * Informs all collection that a document is changed
 * @param {Document} doc changed
 */
App.Document.prototype.handleDocUpdate = function (doc)
{
  // Notify onDocUpdate event
  var options = {skip: false};
  this.onDocUpdate(doc, options);
  if (options.skip)
    return;
  //
  // If doc it's me ...
  var md = this.schema;
  if (!doc.inserted && doc.getDNA() === this.getDNA()) {
    // Copy status and properties into document
    if (doc.deleted) {
      if (this.collection)
        this.collection.remove(this);
    }
    else if (!doc.inserted) {
      for (var i = 0; i < md.properties.length; i++) {
        var mdp = md.properties[i];
        // Copy only changed properties and set their as original
        if (doc.updated && mdp.name in doc.originalValues) {
          this.changePropDirect(mdp.name, doc[mdp.name], true);
          this.setOriginalValue(mdp.name, doc[mdp.name]);
        }
      }
    }
  }
  else {
    // Dispatch message to base class
    App.DataRow.prototype.handleDocUpdate.call(this, doc);
    //
    // Dispatch message to all collections
    // Don't check contentType because doc could be a grandson
    for (var i = 0; i < md.collections.length; i++)
      this[md.collections[i].name].handleDocUpdate(doc);
  }
};


/*
 * Event fired when a document is changed
 * @param {Document} doc changed
 * @param {Object} options
 */
App.Document.prototype.onDocUpdate = function (doc, options)
{
};


/*
 * Call a method remotely
 * @param {String/Function} method to call
 * @param {Array} args
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Document.prototype.rfc = function (method, args, options, callback) {
  if (this.webApiEndpoint)
    this.webApiEndpoint.handleCall(this, method, args, options, callback);
  else
    this.app.sync.DO.remoteCall(this, method, args, options, callback);
};
App.Document.rfc = function (app, method, args, callback) {
  if (this.webApiEndpoint)
    this.webApiEndpoint.handleCall(this, method, args, null, callback);
  else
    app.sync.DO.remoteCall(this, method, args, null, callback);
};


/**
 * Returns true if document is remoted
 */
App.Document.prototype.isRemoted = function ()
{
  return (this.app && this.app.sync.isServerProxy());
};


/**
 * Clear cache for a document
 * @param {Session} app
 * @param {String} dna
 * */
App.Document.clearCache = function (app, dna)
{
  // Extract pks from dna
  var pks;
  if (dna)
    pks = dna.split(",").slice(1).join(",");
  //
  var classFound = function (cls) {
    if (cls.docCache) {
      if (pks)
        delete cls.docCache[pks];
      else
        delete cls.docCache;
    }
  };
  //
  // Search class
  if (this !== App.Document)
    classFound(this);
  else
    App.Document.loadByDNA(app, dna, {skipLoad: true}, function (doc) {
      if (doc)
        classFound(doc.class);
    });
};


/**
 * Call replacer function for parse e stringigy JSON
 * @param {Collection/Document} obj
 * @param {Object} json
 * @param {Object} options
 * */
App.Document.JSONReplacer = function (obj, json, options)
{
  if (!options.replacer || typeof options.replacer !== "function")
    return;
  //
  // If replacer function has been supplied call it
  Object.keys(json).forEach(function (k) {
    var v = json[k];
    var attr = {name: k, value: v};
    options.replacer(obj, attr, options);
    //
    // If name has been changed remove it
    if (attr.name !== k)
      delete json[k];
    //
    // Replacer function can change that
    if (attr.name && (attr.name !== k || attr.value !== v))
      json[attr.name] = attr.value;
  });
  //
  // Call replacer any times for add other attributes
  var other = 0;
  while (true) {
    var attr = {name: "", other: other++};
    options.replacer(obj, attr, options);
    if (!attr.name)
      break;
    //
    json[attr.name] = attr.value;
  }
};


//export module for node
if (module)
  module.exports = App.Document;
