/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module, InDe, process */

var App = App || {};

/**
 * @class DTT
 * @param {App.Session} app - app that owns this DTT instance
 */
App.DTT = function (app)
{
  if (module && !App.View)
    App = require("./requireAll");
  this.app = app;                 // Owner of the DTT instance
  //
  this.enabled = false;           // True if enabled
  this.fullTrace = false;         // True if I want to trace all data
  this.reqList = [];              // List of requests
  this.currRequest = null;        // Current request
};

App.DTT.ItemTypeMap = {
  INFO: "i",
  ENTERMETH: "e",
  EXITMETH: "q",
  QUERY: "r",
  QUERYRES: "rs",
  LOG: "l",
  EXCEPTION: "x",
  WARNING: "w"
};

App.DTT.CloseMargin = 500;      // Number of ms to wait when closing a request
App.DTT.ReuseMargin = 5000;     // Number of ms from the start of the previous request that allows merging
App.DTT.WaitAsyncMargin = 5000; // Number of ms to wait before sending an incomplete request that contains async items
App.DTT.QueryRows = 10;         // Number of rows that have to be dumped for a given query
App.DTT.MaxItems = 50000;       // Maximum nuber of items per request
App.DTT.DebugCycles = 10;       // Nuber of loops for which items are collected
App.DTT.AsyncLoopMargin = 1000; // Number of ms to wait before creating a new request after a BIG request


Object.defineProperty(App.DTT.prototype, "enabled", {
  get: function () {
    return this._enabled;
  },
  set: function (newValue) {
    if (newValue === this._enabled)
      return;
    this._enabled = newValue;
    if (this._enabled)
      this.redirectConsole();
  }
});


/**
 * Redirect console object
 */
App.DTT.prototype.redirectConsole = function ()
{
  console.log = function () {
    App.DTT.prototype.log.apply(this, arguments);
  }.bind(this);
};


/**
 * Redirect console object
 */
App.DTT.prototype.log = function ()
{
  this.addItem(null, App.DTT.ItemTypeMap.LOG, "Log", Array.prototype.slice.call(arguments));
};


/**
 * Starts a new DTT request
 * @param {Object} data - optional data to log
 */
App.DTT.prototype.beginRequest = function (data)
{
  // If not enabled, do nothing
  if (!this.enabled)
    return;
  //
  // If this new request is "close" enough to the previous request, reuse it (see DTT::endRequest)
  if (this.currRequest && this.currRequest.delayCloseTimer) {
    clearTimeout(this.currRequest.delayCloseTimer);
    delete this.currRequest.delayCloseTimer;
    //
    // If there is a current request, append to this only within 5 seconds from the start
    // (otherwise a timer less than App.DTT.CloseMargin will keep this request alive forever!)
    var dt = (new Date().getTime() - this.currRequest.start.getTime());
    if (dt < App.DTT.ReuseMargin)
      return;
  }
  //
  // If there was an open request, close it
  if (this.currRequest)
    this.currRequest.close();
  //
  // Create a new request
  this.currRequest = new App.DTTRequest(this);
  //
  // Set full trace property
  this.currRequest.fullTrace = this.fullTrace;
  //
  // Dump input data
  if (data) {
    for (var i = 0; i < data.length; i++) {
      var d = data[i];
      //
      // Try to see if it points to a model object (see App::processRequest)
      var objid;
      if (!d.obj)
        objid = this.appid;
      else {
        var ele = this.app.eleMap[d.obj];
        if (ele)
          objid = ele.pid;
      }
      this.addItem(objid, App.DTT.ItemTypeMap.INFO, d.id, JSON.stringify(d.content));
    }
  }
};


/**
 * Ends a new DTT request
 * @param{DTTRequest} request
 */
App.DTT.prototype.endRequest = function (request)
{
  // If not enabled, do nothing
  if (!this.enabled)
    return;
  //
  // If request is undefined set to currRequest
  if (!request)
    request = this.currRequest;
  //
  // Close the current request if present
  if (!request)
    return;
  //
  delete this.lastParentItem;
  //
  // Don't close it immediately... Wait for something more to arrive (for frequent events)
  if (request.delayCloseTimer)
    clearTimeout(request.delayCloseTimer);
  //
  request.delayCloseTimer = setTimeout(function () {
    // Do something only if there is a pending request... it could be that it has been already closed
    request.close();
    if (this.currRequest === request)
      delete this.currRequest;
  }.bind(this), App.DTT.CloseMargin);
};


/**
 * Sets the name of the current request
 * @param {String} name - name of the request
 */
App.DTT.prototype.setRequestName = function (name)
{
  // If I've a request give it a name
  if (this.currRequest && !this.currRequest.name)
    this.currRequest.name = name;
};


/**
 * Add an item to the current request
 * @param {String} objid - identifier of the object that created the message (any design time object)
 * @param {String} itemtype - item type
 * @param {String} objname - name of the object that created the message
 * @param {String} msg - message
 * @param {App.DTTItem} parentItem - enter method item
 * @returns {App.DTTItem} returns the new item
 */
App.DTT.prototype.addItem = function (objid, itemtype, objname, msg, parentItem)
{
  // If not enabled, do nothing
  if (!this.enabled)
    return;
  //
  // If full trace is not enabled, ignore blacklist-type item
  // Save objid of no-type items in order to attach it to next LOG/EXCEPTION item.
  // This info will be used by debugger to highlight line of code associated with LOG/EXCEPTION item
  if (!this.fullTrace) {
    var blackList = [App.DTT.ItemTypeMap.ENTERMETH, App.DTT.ItemTypeMap.EXITMETH];
    if (blackList.indexOf(itemtype) !== -1 || !itemtype) {
      if (!itemtype)
        this.prevItemObjId = objid;
      //
      return;
    }
  }
  //
  // For enter method item store last parentItem because is not given as parameter
  if (parentItem && !itemtype)
    this.lastParentItem = parentItem;
  else {
    if (itemtype === App.DTT.ItemTypeMap.ENTERMETH)
      parentItem = this.lastParentItem;
    delete this.lastParentItem;
  }
  //
  // If there is not a current request, create new one
  if (!parentItem && !this.currRequest) {
    // Before creating a new request I need to check if I can do that.
    // If the previous request was a "big" one it could be that the app code contains a loop
    // If the loop is syncrhonous there are not problem: the MaxItems check below will protect the system.
    // But if the loop code contains asynchronous code the DTT will have time to send the closed big request
    // and when a new item arrives a new request is created and filled as well...
    // So if the app code contains an asynchronous loop the system sends to the IDE a lot of big requests!
    // Solution: don't open a new request if the previous one was a full request... allow to create a
    // new request only after 1000 ms
    var prevreq = this.reqList[this.reqList.length - 1];
    if (prevreq && prevreq.items.length >= App.DTT.MaxItems) {
      console.log("[DTT.addItem] Previous request was BIG");
      // Previous request was full... check elapsed time
      var lastBigReqIt = prevreq.items[prevreq.items.length - 1];
      var et = new Date().getTime() - lastBigReqIt.getDate().getTime();
      if (et <= App.DTT.AsyncLoopMargin) {
        console.log("[DTT.addItem] Delta too short: " + et + " ms");
        return;
      }
    }
    //
    this.beginRequest();
    this.endRequest();
  }
  //
  // Get the correct request
  var req = (parentItem ? parentItem.parent : this.currRequest);
  if (objid instanceof App.DTTItem) {
    req = objid.parent;
    objid = objid.getID();
  }
  //
  // If there is a block due to a block cycle, don't add items
  if (req.loopFilter)
    return;
  //
  // If there are too many items in this request
  if (req.items.length >= App.DTT.MaxItems) {
    // Tell it to the programmer (only once)
    var lastit = req.items[req.items.length - 1];
    if ([App.DTT.ItemTypeMap.EXCEPTION, App.DTT.ItemTypeMap.WARNING].indexOf(lastit.type) === -1)
      this.addException("", "Too many items... Debug stopped", parentItem);
    //
    // Don't add new items
    return;
  }
  //
  var newit = new App.DTTItem(req, objid, itemtype, objname, msg, parentItem);
  //
  // If item is a log or an exception and there is a previous objid (i.e. full trace is not enabled)
  // attach it to newborn item
  if ([App.DTT.ItemTypeMap.LOG, App.DTT.ItemTypeMap.EXCEPTION, App.DTT.ItemTypeMap.WARNING].indexOf(itemtype) !== -1 && this.prevItemObjId) {
    newit.prevItemObjId = this.prevItemObjId;
    delete this.prevItemObjId;
  }
  //
  // Save timeStamp
  newit.timeStamp = new Date().getTime();
  //
  req.items.push(newit);
  //
  // If request is already sent send new items
  if (req.itemsSent)
    this.endRequest(req);
  //
  return newit;
};


/**
 * Add a parameter to the last item of the current request
 * @param {String} objname - name of the parameter
 * @param {String} objvalue - value of the parameter
 * @param {App.DTTItem} item - if given add the parameter to this item
 */
App.DTT.prototype.addItemParam = function (objname, objvalue, item)
{
  // If not enabled, do nothing
  if (!this.enabled)
    return;
  //
  // If there is a block due to a block cycle, don't add item params
  if (!item || item.parent.loopFilter)
    return;
  //
  // Don't add any param if full trace is disabled
  if (!this.fullTrace) {
    var blackList = [App.DTT.ItemTypeMap.ENTERMETH, App.DTT.ItemTypeMap.EXITMETH, App.DTT.ItemTypeMap.INFO];
    if (blackList.indexOf(item.type) !== -1 || !item.type)
      return;
  }
  //
  // Logged value
  var par = {};
  par._n = objname;
  //
  var i;
  if (objvalue instanceof Array) {
    par._l = objvalue.length;
    for (i = 0; i < Math.min(10, objvalue.length); i++)
      par["i" + i] = App.DTT.convertValue(objvalue[i]);
  }
  else if (objvalue && typeof objvalue === "object" && !(objvalue instanceof Date)) {
    if (objvalue.getDebugInfo)
      objvalue = objvalue.getDebugInfo();
    var k = Object.getOwnPropertyNames(objvalue);
    for (i = 0; i < k.length; i++)
      par[k[i]] = App.DTT.convertValue(objvalue[k[i]]);
  }
  else {
    var keep = [App.DTT.ItemTypeMap.QUERY,
      App.DTT.ItemTypeMap.QUERYRES,
      App.DTT.ItemTypeMap.EXCEPTION,
      App.DTT.ItemTypeMap.WARNING].indexOf(item.type) >= 0;
    par._v = App.DTT.convertValue(objvalue, keep);
  }
  //
  // For item of statement add param only if is really changed
  if (!item.type && item.parentItem) {
    var json = JSON.stringify(par);
    if (item.parentItem.lastParValues && item.parentItem.lastParValues.hasOwnProperty(objname)
            && item.parentItem.lastParValues[objname] === json)
      return;
    else {
      item.parentItem.lastParValues = item.parentItem.lastParValues || {};
      item.parentItem.lastParValues[objname] = json;
    }
  }
  //
  item.params = item.params || [];
  item.params.push(par);
};


/**
 * Store a value in a dtt parameter
 * @param {Object} v - value to convert
 * @param {Bool} keep - keep entire value or trim it
 * @param {Array} objlist - list of obj for circular references
 * @param {Integer} childLevel - level of child to log
 */
App.DTT.convertValue = function (v, keep, objlist, childLevel)
{
  if (childLevel === undefined)
    childLevel = 0;
  //
  if (v === undefined)
    return "undefined";
  //
  if (v === null)
    return null;
  //
  if (v instanceof Array) {
    if (childLevel < 0)
      return "Array(" + v.length + ")";
    var a = [];
    for (var i = 0; i < v.length; i++)
      a.push(App.DTT.convertValue(v[i], keep, objlist, childLevel - 1));
    return a;
  }
  //
  if (v instanceof Date)
    return new Date(v);
  //
  if (v instanceof App.moment)
    return v.toDate();
  //
  if (v instanceof ArrayBuffer)
    return "ArrayBuffer(" + v.byteLength + ")";
  //
  if (typeof v === "object") {
    // If object has getDebugInfo method call it
    if (v.getDebugInfo)
      return v.getDebugInfo(childLevel);
    //
    if (!objlist || childLevel < 0)
      return "{...}";
    if (objlist.indexOf(v) >= 0)
      return "[Circular reference]";
    else {
      var o = {};
      var parentList = objlist.concat([v]);
      var keys = Object.getOwnPropertyNames(v);
      for (var i = 0; i < keys.length; i++)
        o[keys[i]] = App.DTT.convertValue(v[keys[i]], keep, parentList);
      return o;
    }
  }
  //
  if (typeof v === "string") {
    if (!keep && v.length > 255)
      return v.substr(0, 255) + "...";
  }
  //
  if (typeof v === "function")
    return "function";
  //
  return v;
};


/**
 * Add an exception to the current request
 * @param {String} id - identifier of the object that created the exception (any design time object)
 * @param {Exception} ex - exception thrown
 * @param {DTTItem} parentItem - parent DTT Item
 */
App.DTT.prototype.addException = function (id, ex, parentItem)
{
  // If not enabled, do nothing
  if (!this.enabled)
    return;
  //
  // Dump the exception only once per request
  if (ex.handled)
    return;
  ex.handled = true;
  //
  var newit = this.addItem(id, App.DTT.ItemTypeMap.EXCEPTION, "Exception", undefined, parentItem);
  //
  // Add exception attributes
  var text = (typeof ex === "string" ? ex : ex.message || ex.toString());
  this.addItemParam("message", text, newit);
  //
  if (ex.stack)
    this.addItemParam("stack", ex.stack, newit);
  this.app.notifyClientError(text);
};


/**
 * Add a warning to the current request
 * @param {String} id - identifier of the object that created the warning (any design time object)
 * @param {String} message
 * @param {DTTItem} parentItem - parent DTT Item
 */
App.DTT.prototype.addWarning = function (id, message, parentItem)
{
  // If not enabled, do nothing
  if (!this.enabled)
    return;
  //
  var newit = this.addItem(id, App.DTT.ItemTypeMap.WARNING, "Warning", undefined, parentItem);
  //
  // Add exception attributes
  this.addItemParam("message", message, newit);
  this.app.notifyClientError(message, true);
};


/**
 * Checks if the debug must be switched off (debug of loop cycles)
 * @param {DTTItem} item - item of the block (loop) formula
 * @returns {Boolean} returns false if the loop must be broken
 */
App.DTT.prototype.checkLoop = function (item)
{
  if (!item)
    return true;
  //
  // Remember that for the given block ID a loop has been executed
  var req = item.parent;
  req.loops = req.loops || [];
  req.loopFilter = req.loopFilter || 0;
  //
  // Search the block
  var blk = null;
  for (var i = 0; i < req.loops.length && !blk; i++)
    if (req.loops[i].item.objid === item.objid && req.loops[i].item.parentItem === item.parentItem)
      blk = req.loops[i];
  if (!blk) {
    blk = {item: item, cnt: 1};
    req.loops.push(blk);
  }
  else
    blk.cnt++;
  //
  // If there was no active filtering and I've executed too many loops for this block,
  // stop the debug until we'll exit from this block
  if (!blk.filter && blk.cnt > App.DTT.DebugCycles) {
    blk.filter = true; // This is the block that is causing a filtering
    req.loopFilter++; // Stop the debug from now on
    item.filter = true;
  }
  //
  // Don't stop the loop
  return true;
};


/**
 * Checks if the debug must be re-enabled (debug of loop cycles)
 * @param {String} id - identifier of the block (loop) formula
 * @param {DTTItem} parentItem - parent DTT Item
 */
App.DTT.prototype.endLoop = function (id, parentItem)
{
  if (!parentItem)
    return;
  //
  var req = parentItem.parent;
  if (!req.loops || req.loops.length === 0)
    return;
  //
  var lastLoop = req.loops[req.loops.length - 1];
  if (lastLoop.item.objid !== id || lastLoop.item.parentItem !== parentItem)
    return;
  //
  var blk = req.loops.pop();
  //
  // If I'm exiting from the block that was causing the filter, disable it
  if (blk && blk.filter)
    req.loopFilter--;
};






/**
 * @class DTTRequest
 * @param {App.DTT} parent - parent (owner) of the DTTRequest
 */
App.DTTRequest = function (parent)
{
  this.parent = parent;
  this.id = App.Utils.generateUID24();
  this.start = new Date();    // Remember when I was born
  this.items = [];            // Request items
  //
  var prevreq = this.parent.reqList[this.parent.reqList.length - 1];
  this.delta = 0;
  if (prevreq) {
    var lastit = prevreq.items[prevreq.items.length - 1];
    this.delta = this.start.getTime() - prevreq.start.getTime() + lastit.dt;
  }
};


/**
 * Return the request identifier
 */
App.DTTRequest.prototype.getID = function ()
{
  return "req-" + this.id;
};


/**
 * Return the request icon
 */
App.DTTRequest.prototype.getIcon = function ()
{
  // If the request contains errors, show that it contains errors
  // If the request contains methods, show that it contains methods
  var isException = false;
  var isWarning = false;
  var isMethod = false;
  var arr = this.getVisibleChildren();
  for (var i = 0; i < arr.length; i++) {
    if (!isException && arr[i].isType(App.DTT.ItemTypeMap.EXCEPTION))
      isException = arr[i].isType(App.DTT.ItemTypeMap.EXCEPTION);
    if (!isWarning && arr[i].isType(App.DTT.ItemTypeMap.WARNING))
      isWarning = arr[i].isType(App.DTT.ItemTypeMap.WARNING);
    if (!isMethod && arr[i].isType(App.DTT.ItemTypeMap.ENTERMETH))
      isMethod = true;
  }
  //
  // Show the default icon for this request
  if (isException)
    return {icon: "icon-stop", color: "red"};
  if (isWarning)
    return {icon: "icon-stop", color: "yellow"};
  if (isMethod)
    return InDe.AMethod.DefaultIcon;
  else
    return InDe.AConst.DefaultIcon;
};


/**
 * Return an object that can be used to show request details
 * (like total time, delta with previous request, ...)
 */
App.DTTRequest.prototype.getInfo = function ()
{
  var r = {_n: this.name};
  //
  // Add the total request time (using the last item in the request)
  var tt = this.getTotalTime();
  if (tt)
    r[InDe.rh.t("dttview_reqinfo2")] = tt + " ms";
  //
  return r;
};




/**
 * Return the total time occurred for the request
 */
App.DTTRequest.prototype.getTotalTime = function ()
{
  var lastDT = 0;
  for (var i = 0; i < this.items.length; i++)
    lastDT = Math.max(lastDT, this.items[i].dt || 0);
  return lastDT;
};


/**
 * Return the list of visible children
 */
App.DTTRequest.prototype.getVisibleChildren = function ()
{
  var methLevel = 0;
  var res = [];
  for (var i = 0; i < this.items.length; i++) {
    var it = this.items[i];
    //
    // Add only the method at the right level
    // A request can "start" with a quit if an inputBox/messageConfirm with yield was used (thus the "methLevel > 0" for EXIT)
    if (it.type === App.DTT.ItemTypeMap.ENTERMETH)
      methLevel++;
    if (it.type === App.DTT.ItemTypeMap.EXITMETH && methLevel > 0)
      methLevel--;
    //
    // Add only the right items at the right level
    if ((methLevel === 0 && [App.DTT.ItemTypeMap.INFO, App.DTT.ItemTypeMap.QUERY, App.DTT.ItemTypeMap.EXCEPTION, App.DTT.ItemTypeMap.WARNING].indexOf(it.type) !== -1) ||
            (methLevel === 1 && it.type === App.DTT.ItemTypeMap.ENTERMETH))
      res.push(it);
  }
  return res;
};


/**
 * Return the list of visible children
 */
App.DTTRequest.prototype.getVisibleChildren2 = function ()
{
  var methLevel = 0;
  var res = [];
  for (var i = 0; i < this.items.length; i++) {
    var it = this.items[i];
    //
    // Add only the method at the right level
    // A request can "start" with a quit if an inputBox/messageConfirm with yield was used (thus the "methLevel > 0" for EXIT)
    if (it.type === App.DTT.ItemTypeMap.ENTERMETH)
      methLevel++;
    if (it.type === App.DTT.ItemTypeMap.EXITMETH && methLevel > 0)
      methLevel--;
    //
    // Add only the right items at the right level
    if (methLevel === 1 && it.type === App.DTT.ItemTypeMap.ENTERMETH)
      res.push(it);
  }
  return res;
};


/**
 * Reconnect item hierarchy
 * @param {Array} newItems - items to append
 */
App.DTTRequest.prototype.reconnectItems = function (newItems)
{
  // Append new items if needed
  var start = 0;
  if (newItems) {
    start = this.items.length;
    this.items = this.items.concat(newItems);
    this.itemsSent = this.items.length;
  }
  //
  var lastItem = null;
  for (var i = start; i < this.items.length; i++) {
    var item = this.items[i];
    item.parent = this;
    //
    // If item has a parentItem search it
    if (item.parentItem) {
      for (var j = i - 1; j >= 0; j--) {
        if (this.items[j].getID() === item.parentItem) {
          lastItem = this.items[j];
          item.parentItem = lastItem;
          break;
        }
      }
      if (typeof item.parentItem === "string") {
        InDe.theApp.doc.log("ERROR", "Bad DTT item", "DTTRequest.reconnectItems", {item: item, now: new Date()});
        return;
      }
    }
    //
    if (lastItem) {
      item.parentItem = lastItem;
      lastItem.items = lastItem.items || [];
      lastItem.items.push(item);
    }
    if (item.type === App.DTT.ItemTypeMap.ENTERMETH)
      lastItem = item;
    else if (item.type === App.DTT.ItemTypeMap.EXITMETH) {
      // If parentItem is not my enter method request is malformed
      if (!item.parentItem || item.parentItem.objid !== item.objid) {
        InDe.theApp.doc.log("ERROR", "Bad DTT request; exit method item without enter item", "DTTRequest.reconnectItems", {item: item, now: new Date()});
        return;
      }
      //
      lastItem = item.parentItem.parentItem;
    }
  }
};


/**
 * Saves/loads the request
 */
App.DTTRequest.prototype.toJSON = function ()
{
  var keys = Object.keys(this);
  var o = {};
  for (var i = 0; i < keys.length; i++) {
    var k = keys[i];
    // Don't serialize PARENT property... I don't need it
    if (k === "parent")
      continue;
    else if (k === "items")
      o[k] = this.items.slice(this.itemsSent);
    else
      o[k] = this[k];
  }
  return o;
};
/**
 * Saves/loads the request
 */
App.DTTRequest.prototype.save = function ()
{
  return JSON.stringify(this);
};
App.DTTRequest.load = function (s)
{
  var req = JSON.parse(s);
  //
  /* jshint ignore:start */
  // The JSON serialization looses all prototypes... restore them
  req.__proto__ = App.DTTRequest.prototype;
  //
  // Parse start date
  req.start = new Date(req.start);
  //
  for (var i = 0; i < req.items.length; i++) {
    var rit = req.items[i];
    rit.__proto__ = App.DTTItem.prototype;
    rit.parent = req;
  }
  /* jshint ignore:end */
  //
  return req;
};


/**
 * Return True if the request is complete
 * (i.e. the number of ENTER and EXIT methods matches)
 */
App.DTTRequest.prototype.isComplete = function ()
{
  var numenter = 0;
  var numexit = 0;
  for (var i = 0; i < this.items.length; i++) {
    var it = this.items[i];
    if (it.type === App.DTT.ItemTypeMap.ENTERMETH)
      numenter++;
    else if (it.type === App.DTT.ItemTypeMap.EXITMETH)
      numexit++;
  }
  //
  return (numexit >= numenter);
};


/**
 * Close the request
 */
App.DTTRequest.prototype.close = function ()
{
  // Keep only requests that are not EMPTY
  if (this.items && this.items.length > 0) {
    if (!this.itemsSent) {
      this.parent.reqList.push(this);
      //
      // If the request has no name, give it an automatic-name
      if (!this.name) {
        var reqName = "Request " + this.parent.reqList.length;
        //
        // Check if I could assign a better name
        for (var i = 0; i < this.items.length; i++) {
          if (this.items[i].name) {
            reqName = this.items[i].name;
            break;
          }
        }
        this.name = reqName;
      }
    }
    //
    // Remove timer
    delete this.delayCloseTimer;
    //
    // Save timeStamp
    this.timeStamp = new Date().getTime();
    //
    // Send this request to the IDE
    this.sendToIDE();
  }
};


/**
 * Send the request to the IDE
 */
App.DTTRequest.prototype.sendToIDE = function ()
{
  // Send this request to InDe
  var data = {appid: this.parent.appid, req: this.save()};
  if (module)
    process.send({type: "dtt", content: data});
  else if (this.parent.app.proxy) {
    // Send this message to my IDE and to the other clients (if any)
    // There is not a direct connection with the IDE instance that have generated me
    // Search a DTT view attached to my app and notify the message
    var app = this.parent.app.proxy;
    if (app.sendDTT) {
      // offline Proxy
      data.offline = true;
      app.sendDTT(data);
    }
    else {
      // Local testing
      var doc = app.getDocument();
      for (var i = 0; i < doc.listeners.length; i++) {
        if (doc.listeners[i].type === "dtt" && doc.listeners[i].obj.app === app) {
          doc.listeners[i].cb(data);
          break;
        }
      }
      //
      this.parent.app.proxy.getDocument().sendMessage({type: "dtt", cnt: data});
    }
  }
  //
  this.itemsSent = this.items.length;
};


/**
 * A method is been excluded then purge it
 * @param {InDe.AMethod} met excluded
 */
App.DTTRequest.prototype.excludeMethod = function (met)
{
  for (var i = 0; i < this.items.length; i++) {
    var item = this.items[i];
    //
    // Remove items of methods excluded
    if ((item.type === App.DTT.ItemTypeMap.ENTERMETH && item.objid === met.id)
            || (item.type !== App.DTT.ItemTypeMap.ENTERMETH && item.parentItem && item.parentItem.objid === met.id))
      this.items.splice(i--, 1);
    else {
      delete item.items;
      delete item.parentItem;
    }
  }
  //
  this.reconnectItems();
};







/**
 * @class DTTItem
 * @param {App.DTTRequest} parent - parent (owner) of the DTTItem
 * @param {String} objid - identifier of the object that created the message (any design time object)
 * @param {String} type - item type
 * @param {String} objname - name of the object that created the message
 * @param {String} msg - message
 * @param {DTTItem} parentItem - parent DTT Item
 */
App.DTTItem = function (parent, objid, type, objname, msg, parentItem)
{
  this.parent = parent;
  var t = (new Date()).getTime() - parent.start.getTime();
  if (t) {
    this.dt = t;
    //
    var previt = this.parent.items[this.parent.items.length - 1];
    if (previt)
      this.delta = this.dt - previt.dt;
  }
  this.objid = objid;
  this.type = type;
  if (objname)
    this.name = (this.type === App.DTT.ItemTypeMap.INFO ? "Info: " : "") + objname;
  //
  if (this.type === App.DTT.ItemTypeMap.LOG)
    this.msg = App.DTT.convertValue(msg, true, [], 9999);
  else
    this.msg = msg;
  //
  if (parentItem)
    this.parentItem = parentItem;
};



/**
 * Save the DTTItem in JSON
 */
App.DTTItem.prototype.toJSON = function ()
{
  var keys = Object.keys(this);
  var o = {};
  for (var i = 0; i < keys.length; i++) {
    var k = keys[i];
    // Don't serialize PARENT property... I don't need it
    if (k === "parent" || k === "lastParValues")
      continue;
    else if (k === "parentItem")
      o[k] = this.parentItem.getID();
    else
      o[k] = this[k];
  }
  return o;
};

/**
 * Return the item identifier
 */
App.DTTItem.prototype.getID = function ()
{
  return "ite-" + this.parent.id + "-" + (this.parent.items.indexOf(this) + 1);
};


/**
 * Return true if the item or one of its children are of the given type
 * @param {String} type
 */
App.DTTItem.prototype.isType = function (type)
{
  // Ask my children first
  var arr = this.getVisibleChildren();
  for (var i = 0; i < arr.length; i++)
    if (arr[i].isType(type))
      return true;
  //
  return (this.type === type);
};


/**
 * Return the list of visible children
 */
App.DTTItem.prototype.getVisibleChildren = function ()
{
  var arr = [];
  var idx = this.parent.items.indexOf(this);
  if (idx === -1 || this.type !== App.DTT.ItemTypeMap.ENTERMETH)
    return arr;
  //
  // Search other inner EnterMethod items (only at first level)
  var methLevel = 0;
  for (var i = idx + 1; i < this.parent.items.length; i++) {
    var it = this.parent.items[i];
    //
    // If it's the end of the method I'm pointing to -> I've done
    if (it.objid === this.objid && it.type === App.DTT.ItemTypeMap.EXITMETH)
      break;
    //
    // Add only the method at the right level
    if (it.type === App.DTT.ItemTypeMap.ENTERMETH)
      methLevel++;
    if (it.type === App.DTT.ItemTypeMap.EXITMETH)
      methLevel--;
    //
    // Add only the right items at the right level
    if ((methLevel === 0 && [App.DTT.ItemTypeMap.INFO, App.DTT.ItemTypeMap.QUERY, App.DTT.ItemTypeMap.EXCEPTION, App.DTT.ItemTypeMap.WARNING].indexOf(it.type) !== -1) ||
            (methLevel === 1 && it.type === App.DTT.ItemTypeMap.ENTERMETH))
      arr.push(it);
  }
  //
  return arr;
};


/**
 * Return the list of visible children
 */
App.DTTItem.prototype.getVisibleChildren2 = function ()
{
  var arr = [];
  if (this.items) {
    for (var i = 0; i < this.items.length; i++) {
      var it = this.items[i];
      if (it.type === App.DTT.ItemTypeMap.ENTERMETH)
        arr.push(it);
    }
  }
  //
  return arr;
};


/**
 * Return the list of invisible children
 */
App.DTTItem.prototype.getInvisibleChildren = function ()
{
  var arr = [];
  if (this.items) {
    for (var i = 0; i < this.items.length; i++) {
      var it = this.items[i];
      if (it.type !== App.DTT.ItemTypeMap.ENTERMETH)
        arr.push(it);
    }
  }
  //
  return arr;
};


/**
 * Check if is child of another item
 * @param {DTTItem} item - parent DTT Item
 */
App.DTTItem.prototype.isChildOf = function (item)
{
  if (!item.items)
    return false;
  //
  for (var i = 0; i < item.items.length; i++) {
    if (item.items[i] === this)
      return true;
    else if (this.isChildOf(item.items[i]))
      return true;
  }
  return false;
};


/**
 * Return the item Date
 */
App.DTTItem.prototype.getDate = function ()
{
  return new Date(this.parent.start.getTime() + (this.dt || 0));
};


/**
 * Return an object that can be used to show item details
 * (like total time, delta with previous request, ...)
 */
App.DTTItem.prototype.getInfo = function ()
{
  var r = {};
  //
  if ([App.DTT.ItemTypeMap.QUERY, App.DTT.ItemTypeMap.EXCEPTION, App.DTT.ItemTypeMap.WARNING].indexOf(this.type) >= 0) {
    r._i = this.getID();
    r._n = this.name;
    r._t = this.type;
    r._exp = this.type === App.DTT.ItemTypeMap.EXCEPTION;
    var params = this.params;
    //
    // For query append also params of result
    if (this.type === App.DTT.ItemTypeMap.QUERY) {
      for (var i = this.parent.items.indexOf(this) + 1; i < this.parent.items.length; i++) {
        var res = this.parent.items[i];
        if (res.objid !== this.getID())
          continue;
        params = params.concat(res.params);
        r._r = res.type;
      }
    }
    //
    for (var i = 0; i < params.length; i++) {
      var p = params[i];
      r[p._n] = p._v;
    }
    return r;
  }
  //
  if (this.type === App.DTT.ItemTypeMap.LOG) {
    r._i = this.getID();
    r._n = this.name;
    r._t = this.type;
    r._exp = false;
    r.msg = this.msg;
    return r;
  }
  //
  // Show parameter of previous item because changes to variables must to be shown after
  if (!this.type && this.parentItem) {
    for (var i = this.parentItem.items.indexOf(this) - 1; i >= 0; i--) {
      var it = this.parentItem.items[i];
      if (!it.type)
        return it.params || r;
    }
    return r;
  }
  //
  // If the item has parameters, return them
  if (this.params)
    return this.params;
  //
  // Return other info
  // If the item has a message, dump it
  if (this.msg) {
    r._n = this.name;
    r._t = this.type;
    r._exp = true;
    var m = this.msg;
    //
    // If it's a simple string, dump it as "TEXT:VALUE" object
    // And if the string starts with { try to parse it as a JSON
    if (typeof m === "string") {
      // Store the value in a dummy property
      m = {text: m};
      //
      // If it starts with a {, try to parse as a JSON value
      if (m.text[0] === "{") {
        try {
          m = JSON.parse(m.text);
        }
        catch (ex) {
        }
      }
    }
    //
    for (var prop in m)
      if (m.hasOwnProperty(prop))
        r[prop] = m[prop];
  }
  //
  return r;
};


// Export module for node
if (module)
  module.exports = {DTT: App.DTT, DTTRequest: App.DTTRequest, DTTItem: App.DTTItem};
