/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */


/* global module, Client */

var App = App || {};


/**
 * @class File
 * Represents a file object
 * @param {App.FS} fs
 * @param {string} type
 * @param {string} path
 * @param {string} publicUrl
 */
App.File = function (fs, type, path, publicUrl)
{
  // Import modules
  if (module && !App.View)
    App = require("../requireAll");
  //
  // String containing the type of the file
  this.type = type || App.FS.internalType.permanent;
  //
  // String containing the relative path of the file
  if (path)
    this.path = path;
  //
  this.fs = fs;
  //
  // Public url pointing to the file
  if (publicUrl)
    this._publicUrl = publicUrl;
  else
    this.fs.getFilePublicUrl(this, function (url) {
      this._publicUrl = url;
    }.bind(this));
  //
  // String containing the encoding of the file, null by default: decided by the user at the time
  // of the creation of the physical file
  this.encoding = null;
};


Object.defineProperties(App.File.prototype, {
  publicUrl: {
    get: function () {
      return this._publicUrl;
    }
  },
  path: {
    get: function () {
      return this._path;
    },
    set: function (newValue) {
      this._path = App.FS.normalizePath(newValue);
    }
  }
});


/**
 * Tells something about this object to the DTT module
 */
App.File.prototype.getDebugInfo = function ()
{
  var info = {};
  info._class = "File";
  info.path = this.path;
  info.type = this.type;
  info.publicUrl = this.publicUrl;
  return info;
};


/**
 * Returns the full internal path.
 * Only for internal, nodedriver use
 * @param {function} cb
 */
App.File.prototype._getFullPath = function (cb)
{
  if (this.fs instanceof App.NodeDriver) {
    this.fs.getFileFullPath(this, function (path) {
      if (cb)
        cb(path);
    });
  }
};


/**
 * Creates the file physically (Opens the file and overwrites it)
 * @param {string} encoding
 * @param {function} cb
 */
App.File.prototype.create = function (encoding, cb)
{
  // String containing the encoding of the file, null only for binary files
  this.encoding = encoding;
  this.fs.createFile(this, function (err) {
    cb(null, err);
  });
};


/**
 * Opens the file for reading
 * @param {function} cb
 */
App.File.prototype.open = function (cb)
{
  this.fs.openFile(this, function (err) {
    cb(null, err);
  });
};


/**
 * Opens the file to append data
 * @param {function} cb
 */
App.File.prototype.append = function (cb)
{
  this.fs.openFileForAppend(this, function (err) {
    cb(null, err);
  });
};


/**
 * Closes a file
 * @param {function} cb
 */
App.File.prototype.close = function (cb)
{
  this.fs.close(this, function (err) {
    cb(null, err);
  });
};


/**
 * Checks  the existence of the file
 * @param {function} cb
 */
App.File.prototype.exists = function (cb)
{
  this.fs.fileExists(this, cb);
};


/**
 * Reads a text file line by line through a loop
 * @param {function} cb
 * @param {function} cbEnd
 */
App.File.prototype.readLine = function (cb, cbEnd)
{
  this.fs.readLine(this, cb, function (err) {
    cbEnd(null, err);
  });
};


/**
 * Stops the reading line by line
 */
App.File.prototype.break = function ()
{
  this.fs.break(this);
};


/**
 * Reads a block of data, return a node buffer/array buffer
 * @param {int} length
 * @param {int} offset
 * @param {function} cb
 */
App.File.prototype.read = function (length, offset, cb)
{
  // The length must be greater than 0, if she's "null" is read up to the end of file
  if (length && length < 1)
    return cb(null, new Error("Length must be greater than 0"));
  //
  this.fs.read(this, length, offset, cb);
};


/**
 * Read the whole file as text
 * @param {function} cb
 */
App.File.prototype.readAll = function (cb)
{
  // If I have to download a resource (see App.loadResourceFile)
  if (!module && this.type === App.FS.internalType.resource && this.path && this.path[0] === "/" && this.path.indexOf("/resources/") !== -1) {
    // Download it from the server
    var path = (Client.resourceOrigin || window.location.origin) + this.path;
    this.fs.httpRequest(this.fs.url(path), "GET", {}, function (response) {
      cb(response.body, response.error);
    });
    return;
  }
  //
  this.fs.readAll(this, cb);
};


/**
 * Read the whole file as text, returns the file content
 */
App.File.prototype.readSync = function ()
{
  if (!this.fs.readSync)
    throw new Error("readSync not supported by this driver");
  //
  return this.fs.readSync(this);
};


/**
 * Writes the data or the string given
 * @param {string/buffer} data
 * @param {int} offset
 * @param {int} size
 * @param {int} position
 * @param {function} cb
 */
App.File.prototype.write = function (data, offset, size, position, cb)
{
  // offset: offset relative to the buffer
  // size: number of bytes to write
  // position: position(bytes) of the file from which you start writing
  this.fs.write(this, data, offset, size, position, function (err) {
    cb(null, err);
  });
};


/**
 * Copy the file
 * @param {string} newPath
 * @param {function} cb
 */
App.File.prototype.copy = function (newPath, cb)
{
  var newFile = this.fs.file(newPath, this.type);
  this.fs.copyFile(this, newFile, function (err) {
    cb(newFile, err);
  });
};


/**
 * Rename a file
 * @param {string} newName
 * @param {function} cb
 */
App.File.prototype.rename = function (newName, cb)
{
  this.fs.renameObject(this, newName, function (err) {
    if (err)
      return cb(null, err);
    //
    // I change the path only if the file has been renamed correctly
    // Takes only the name of the file from the new path
    this.path = this.path.substring(0, this.path.lastIndexOf("/") + 1) + newName;
    //
    // Update files' public url
    this.fs.getFilePublicUrl(this, function (url) {
      this._publicUrl = url;
      cb();
    }.bind(this));
  }.bind(this));
};


/**
 * Return the file size (in bytes)
 * @param {function} cb
 */
App.File.prototype.length = function (cb)
{
  this.fs.fileLength(this, cb);
};


/**
 * Return the extension of file (without the dot)
 * @param {function} cb
 */
App.File.prototype.extension = function (cb)
{
  // Get only the file name
  var fileName = this.path.replace(/^.*(\\|\/|\:)/, "");
  //
  /* jshint ignore:start */
  if (fileName.length)
    // Returns only the extension (null for a binary file)
    cb(fileName.substr((~-fileName.lastIndexOf(".") >>> 0) + 2));
  else
    cb("");
  /* jshint ignore:end */
};


/**
 * Return the last modified file date
 * @param {function} cb
 */
App.File.prototype.dateTime = function (cb)
{
  this.fs.fileDateTime(this, cb);
};


/**
 * Deletes the file
 * @param {function} cb
 */
App.File.prototype.remove = function (cb)
{
  this.fs.deleteFile(this, function (err) {
    cb(null, err);
  });
};


/**
 * Zip the file
 * @param {function} cb
 */
App.File.prototype.zip = function (cb)
{
  var zipFile = this.fs.file(this.path + ".zip", this.type);
  this.fs.zipFile(this, zipFile, function (err) {
    cb(zipFile, err);
  });
};


/**
 * Unzip the archive
 * @param {string} path
 * @param {function} cb
 */
App.File.prototype.unzip = function (path, cb)
{
  var unzipDir = this.fs.directory(path, this.type);
  this.fs.unzip(this, unzipDir, function (err) {
    cb(unzipDir, err);
  });
};


/**
 * Put file on remote/local file system
 * @param {App.File} dstFile - remote or local destination file
 * @param {function} cb
 */
App.File.prototype.put = function (dstFile, cb)
{
  this.fs.put(this, dstFile, function (err) {
    cb(null, err);
  });
};


/**
 * Encrypt file
 * @param {String} key
 * @param {String} output
 * @param {function} cb
 */
App.File.prototype.encrypt = function (key, output, cb)
{
  if (!key)
    return cb(new Error("Missing key"));
  //
  // Create encrypted file path
  var pathTokens = this.path.split("/");
  var fileName = App.Utils.generateUID36();
  var originalFileName = pathTokens[pathTokens.length - 1];
  pathTokens.splice(pathTokens.length - 1, 1, fileName);
  var encryptedFile = this.fs.file(pathTokens.join("/"));
  //
  var done = function (error) {
    if (error) {
      encryptedFile.remove(function () {
        cb(null, error);
      });
    }
    else
      cb();
  };
  //
  var readChunk = function (length, offset, callback) {
    // Read chunk
    this.read(length, offset, function (chunk, err) {
      if (err)
        return callback(err);
      //
      // Append current chunk to file content
      var chunkArray = new Uint8Array(chunk);
      var newRes = new Uint8Array(fileContent.length + chunkArray.length);
      newRes.set(fileContent);
      newRes.set(chunkArray, fileContent.length);
      fileContent = newRes;
      //
      // Check if I need to read the next chunk
      if (!chunk || chunk.byteLength !== length)
        return callback();
      else // Read next chunk
        readChunk(length, offset + length, callback);
    });
  }.bind(this);
  //
  var writeChunk = function (data, offset, length, callback) {
    encryptedFile.write(data, offset, length, undefined, function (writeRes, writeErr) {
      if (writeErr)
        return callback(writeErr);
      //
      // Check if I need to write the next chunk
      if (offset + length >= data.length || data.byteLength)
        return callback();
      else // Write next chunk
        writeChunk(data, offset + length, length, callback);
    });
  }.bind(this);
  //
  var fileContent = new Uint8Array(0);
  //
  // Open file
  this.open(function (res, err) {
    if (err)
      return done(err);
    //
    // Start reading chunk by chunk (256 KB is size of a chunk)
    readChunk(1024 * 256, 0, function (error) {
      if (error)
        return done(error);
      //
      // File read. Close it
      this.close(function (closeRes, closeError) {
        if (closeError)
          return done(closeError);
        //
        // Start encryption
        App.Crypt.encrypt(undefined, fileContent.buffer, key, {output: output}, function (encrypted, error) {
          if (error)
            return done(error);
          //
          encryptedFile.create(undefined, function (createError) {
            if (createError)
              return done(createError);
            //
            writeChunk(encrypted, 0, 1024 * 256, function (writeError) {
              if (writeError)
                return done(writeError);
              //
              encryptedFile.close(function (closeRes, closeError) {
                if (closeError)
                  return done(closeError);
                //
                encryptedFile.rename(originalFileName, function (renameRes, renameError) {
                  done(renameError);
                }.bind(this));
              }.bind(this));
            }.bind(this));
          }.bind(this));
        }.bind(this));
      }.bind(this));
    }.bind(this));
  }.bind(this));
};


/**
 * Decrypt file
 * @param {String} key
 * @param {String} output
 * @param {function} cb
 */
App.File.prototype.decrypt = function (key, output, cb)
{
  if (!key)
    return cb(new Error("Missing key"));
  //
  // Create encrypted file path
  var pathTokens = this.path.split("/");
  var fileName = App.Utils.generateUID36();
  var originalFileName = pathTokens[pathTokens.length - 1];
  pathTokens.splice(pathTokens.length - 1, 1, fileName);
  var decryptedFile = this.fs.file(pathTokens.join("/"));
  //
  var done = function (error) {
    if (error) {
      decryptedFile.remove(function () {
        cb(null, error);
      });
    }
    else
      cb();
  };
  //
  var readChunk = function (length, offset, callback) {
    // Read chunk
    this.read(length, offset, function (chunk, err) {
      if (err)
        return callback(err);
      //
      // Append current chunk to file content
      var chunkArray = new Uint8Array(chunk);
      var newRes = new Uint8Array(fileContent.length + chunkArray.length);
      newRes.set(fileContent);
      newRes.set(chunkArray, fileContent.length);
      fileContent = newRes;
      //
      // Check if I need to read the next chunk
      if (!chunk || chunk.byteLength !== length)
        return callback();
      else // Read next chunk
        readChunk(length, offset + length, callback);
    });
  }.bind(this);
  //
  var writeChunk = function (data, offset, length, callback) {
    decryptedFile.write(data, offset, length, undefined, function (writeRes, writeErr) {
      if (writeErr)
        return callback(writeErr);
      //
      // Check if I need to write the next chunk
      if (offset + length >= data.length || data.byteLength)
        return callback();
      else // Write next chunk
        writeChunk(data, offset + length, length, callback);
    });
  };
  //
  var fileContent = new Uint8Array(0);
  //
  // Open file
  this.open(function (res, err) {
    if (err)
      return done(err);
    //
    // Start reading chunk by chunk (256 KB is size of a chunk)
    readChunk(1024 * 256, 0, function (error) {
      if (error)
        return done(error);
      //
      // File read. Close it
      this.close(function (closeRes, closeError) {
        if (closeError)
          return done(closeError);
        //
        // Start decryption
        var decr = output === "buffer" ? fileContent.buffer : App.Utils.bufferToString(fileContent.buffer);
        App.Crypt.decrypt(undefined, decr, key, {output: output}, function (decrypted, error) {
          if (error)
            return done(error);
          //
          decryptedFile.create(undefined, function (createError) {
            if (createError)
              return done(createError);
            //
            writeChunk(decrypted, 0, 1024 * 256, function (writeError) {
              if (writeError)
                return done(writeError);
              //
              decryptedFile.close(function (closeRes, closeError) {
                if (closeError)
                  return done(closeError);
                //
                decryptedFile.rename(originalFileName, function (renameRes, renameError) {
                  done(renameError);
                });
              });
            }.bind(this));
          }.bind(this));
        }.bind(this));
      }.bind(this));
    }.bind(this));
  }.bind(this));
};


//  export module for node
if (module)
  module.exports = App.File;
