/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */


/* global ArrayBuffer, module, async, TextDecoder */

var App = App || {};


/**
 * @class LocalDriver
 * Represents a local driver object,that will handle files and folders
 * @param {Object} options
 */
App.LocalDriver = function (options)
{
  // Local file systems (temporary e persistent)
  this.fs = {};
};

// Make App.LocalDriver extend App.FS
App.LocalDriver.prototype = new App.FS(true);


/*
 * Given a relative path, return the public url to the file
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.getFilePublicUrl = function (file, cb)
{
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(null, err);
    //
    this.checkEntry(file, function (err) {
      if (err)
        return cb(null, err);
      //
      cb(file.entry.toURL());
    });
  }.bind(this));
};


/**
 * Check and create fs object
 * @param {String} type
 * @param {function} cb
 */
App.LocalDriver.prototype.initLocalFs = function (type, cb)
{
  // If the object already exists fs will not be created
  if (this.fs[type])
    return cb();
  //
  var typeWindow = window.PERSISTENT;
  var typeNavigator = navigator.webkitPersistentStorage;
  //
  if (type === App.FS.internalType.temp) {
    typeWindow = window.TEMPORARY;
    typeNavigator = navigator.webkitTemporaryStorage;
  }
  //
  typeNavigator.requestQuota(50 * 1024 * 1024, function (grantedBytes) {
    window.webkitRequestFileSystem(typeWindow, grantedBytes, function (fs) {
      // Save the html5 fs object
      this.fs[type] = fs;
      cb();
    }.bind(this), cb);
  }.bind(this), cb);
};


/**
 * Check and create entry
 * @param {File/Directory} obj
 * @param {function} cb
 */
App.LocalDriver.prototype.checkEntry = function (obj, cb)
{
  if (obj.entry)
    return cb();
  //
  var funcName = (obj instanceof App.File ? "getFile" : "getDirectory");
  this.fs[obj.type].root[funcName](obj.path, {create: false}, function (entry) {
    // Save entry object
    obj.entry = entry;
    cb();
  }, cb);
};


/**
 * Creates the file physically (Opens the file and overwrites it)
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.createFile = function (file, cb)
{
  var create = function () {
    // Sets the attribute "create" to true to create the file
    this.fs[file.type].root.getFile(file.path, {create: true, exclusive: false}, function (entry) {
      // Create the object writer
      entry.createWriter(function (fileWriter) {
        file.writer = fileWriter;
        //
        // entry object
        file.entry = entry;
        cb();
      }, cb); // Writer creation error
    }, cb); // getFile() error
  }.bind(this);
  //
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(err);
    //
    this.fs[file.type].root.getFile(file.path, {create: false}, function (entry) {
      entry.remove(function () {
        // If a file with the same name already exists, delete it and create
        delete file.entry;
        create();
      }, cb);// Deleting error
    }, create); // If a file with the same name not exists, create it
  }.bind(this));
};


/**
 * Opens the file to append data
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.openFileForAppend = function (file, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(file, function (err) {
      if (err)
        return cb(err);
      //
      // Create the object writer
      file.entry.createWriter(function (fileWriter) {
        file.writer = fileWriter;
        //
        // I bet the writer at the end of the file (to allow append mode)
        file.writer.seek(file.writer.length);
        cb();
      }, cb); // Writer creation error
    });
  }.bind(this));
};


/**
 * Opens the file for reading
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.openFile = function (file, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(file, function (err) {
      if (err)
        return cb(err);
      //
      // Gets object file from a entry and save it in a relative property
      file.entry.file(function (f) {
        file.fileObject = f;
        //
        // Create the object reader
        file.reader = new FileReader(f);
        //
        // Property that contains the current position of reader
        file.readerPosition = 0;
        cb();
      });
    });
  }.bind(this));
};


/**
 * Closes the file
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.close = function (file, cb)
{
  delete file.reader;
  delete file.lineReader;
  delete file.writer;
  cb();
};


/**
 * Checks existence of a file
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.fileExists = function (file, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(false);
    //
    // Try to get the entry to see if there is
    this.fs[file.type].root.getFile(file.path, {create: false}, function () {
      cb(true);
    }, function () {
      cb(false);
    });
  }.bind(this));
};


/**
 * Reads a text file line by line through a loop that reads each line of the file
 * @param {File} file
 * @param {function} cb
 * @param {function} cbEnd
 */
App.LocalDriver.prototype.readLine = function (file, cb, cbEnd)
{
  // Check that the file is opened (reading)
  if (!file.fileObject)
    return cbEnd(new Error("File not opened"));
  //
  // Creates a new "LineReader" object
  file.lineReader = new LineReader();
  //
  // Listen to line events (called for each line of the file)
  file.lineReader.on("line", function (line, next) {
    // "line" contains the text line
    cb(line);
    //
    // Calls the reading of the next line
    next();
  });
  //
  // Listen to error event
  file.lineReader.on("error", cbEnd);
  //
  // Listen to end event
  file.lineReader.on("end", cbEnd);
  //
  // Starts the reading
  file.lineReader.read(file.fileObject, file.encoding || "utf-8");
};


/**
 * Stops reading line by line
 * @param {File} file
 */
App.LocalDriver.prototype.break = function (file)
{
  if (file.lineReader)
    file.lineReader.abort();
};


/**
 * Reads a block of data, return a array buffer
 * @param {File} file
 * @param {int} length
 * @param {int} offset
 * @param {function} cb
 */
App.LocalDriver.prototype.read = function (file, length, offset, cb)
{
  // Check that the file is opened (reading)
  if (!file.reader || !file.fileObject)
    return cb(null, new Error("file not open"));
  //
  // Setting the correct limit: if length or offset are null take respectively the current position of the
  // reader and the position of the last byte of the file
  var start = file.readerPosition;
  if (typeof offset === "number")
    start = offset;
  //
  var end = file.fileObject.size;
  if (length)
    end = start + length;
  //
  // Get the part of file
  var blob = file.fileObject.slice(start, end);
  //
  // If the size of the blob is less than 1, the reader position is beyond the end of file (eof)
  if (blob.size < 1)
    return cb(null, new Error("File EOF"));
  //
  // Listen to read end events
  file.reader.onloadend = function () {
    // Updates the current position of the reader
    file.readerPosition = end;
    //
    // evt.target.result contains the string
    cb(file.reader.result);
  };
  //
  // Listen to error events
  file.reader.onerror = function (err) {
    cb(null, err);
  };
  //
  // Read cunck as array buffer
  file.reader.readAsArrayBuffer(blob);
};


/**
 * Read the whole file as text
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.readAll = function (file, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(null, err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(file, function (err) {
      if (err)
        return cb(null, err);
      //
      file.entry.file(function (f) {
        var reader = new FileReader(f);
        reader.onloadend = function () {
          cb(reader.result, reader.error);
        };
        reader.readAsText(f, file.encoding);
      });
    });
  }.bind(this));
};


/**
 * Writes the data or the string given
 * @param {File} file
 * @param {string/buffer} data
 * @param {int} offset
 * @param {int} size
 * @param {int} position
 * @param {function} cb
 */
App.LocalDriver.prototype.write = function (file, data, offset, size, position, cb)
{
  // Check that the file is opened (writing)
  if (!file.writer)
    return cb(new Error("File not open for write"));
  //
  // Updates the current position of the writer
  if (position >= 0 && position !== null)
    file.writer.seek(position);
  //
  var blob;
  //
  // If "data" is a string
  if (typeof data === "string") {
    // If not specified the default encoding is utf-8
    file.encoding = file.encoding || "utf-8";
    blob = new Blob([data], {type: "text/plain;charset=" + file.encoding});
  }
  else {
    // If "data" is a set of bytes
    // calculates the limits of writing
    if (!size)
      size = data.byteLength;
    //
    if (!offset || offset < 0)
      offset = 0;
    //
    blob = new Blob([data]);
    //
    // Get the part of "data"
    blob.slice(offset, offset + size);
  }
  //
  // Listen to error events
  file.writer.onerror = cb;
  //
  // Listen to write end events
  file.writer.onwriteend = function (progressEvent) {
    cb();
  };
  //
  //Write
  file.writer.write(blob);
};


/**
 * Copy the file
 * @param {File} file
 * @param {File} newFile
 * @param {function} cb
 */
App.LocalDriver.prototype.copyFile = function (file, newFile, cb)
{
  // Checks and creates  fs object (if it doesn't exist)
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(err);
    //
    // Get the parent directory
    var pathParts = newFile.path.split("/");
    var dirPath = pathParts.slice(0, pathParts.length - 1).join("/");
    if (dirPath[0] === "/")
      dirPath = dirPath.substr(1);
    //
    var newName = pathParts[pathParts.length - 1];
    //
    // Gets new dir entry
    this.fs[file.type].root.getDirectory(dirPath, {create: false}, function (destinationDirEntry) {
      // Checks and sets  the entry property (if it doesn't exist)
      this.checkEntry(file, function (err) {
        if (err)
          return cb(err);
        //
        // Copy
        file.entry.copyTo(destinationDirEntry, newName, function (newEntry) {
          // Create the new object file by setting the property correctly
          newFile.entry = newEntry;
          cb();
        }, cb); // Raise error
      });
    }.bind(this), cb); // Getting new dir entry error
  }.bind(this));
};


/**
 * Rename a file or directory
 * @param {File/Directory} obj
 * @param {string} newName
 * @param {function} cb
 */
App.LocalDriver.prototype.renameObject = function (obj, newName, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(obj.type, function (err) {
    if (err)
      return cb(err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(obj, function (err) {
      if (err)
        return cb(err);
      //
      // Rename
      obj.entry.getParent(function (parentDirEntry) {
        obj.entry.moveTo(parentDirEntry, newName, function (newEntry) {
          // Update the entry
          obj.entry = newEntry;
          cb();
        }, cb); // moveTo error
      }, cb); // getParent dir error
    });
  }.bind(this));
};


/**
 * Return the file size (in bytes)
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.fileLength = function (file, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(null, err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(file, function (err) {
      if (err)
        return cb(null, err);
      //
      // Get size
      file.entry.file(function (f) {
        cb(f.size);
      });
    });
  }.bind(this));
};


/**
 * Return the last modified file date
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.fileDateTime = function (file, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(null, err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(file, function (err) {
      if (err)
        return cb(null, err);
      //
      // Get dateTime (contained in a metadata entry object)
      file.entry.getMetadata(function (md) {
        cb(md.modificationTime);
      });
    });
  }.bind(this));
};


/**
 * Deletes a file
 * @param {File} file
 * @param {function} cb
 */
App.LocalDriver.prototype.deleteFile = function (file, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(file.type, function (err) {
    if (err)
      return cb(err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(file, function (err) {
      if (err)
        return cb(err);
      //
      // Remove file
      file.entry.remove(function () {
        delete file.entry;
        cb();
      }, cb);
    });
  }.bind(this));
};


/**
 * Zip a file
 * @param {File} file
 * @param {File} zipFile
 * @param {function} cb
 */
App.LocalDriver.prototype.zipFile = function (file, zipFile, cb)
{
  // Save the reader object
  var reader = file.reader;
  this.openFile(file, function (err) {
    if (err)
      return cb(err);
    //
    // Reads the file to zip
    this.read(file, null, 0, function (data, err) {
      if (err)
        return cb(err);
      //
      // Restore the reader to original status
      file.reader = reader;
      //
      // new JSZip object
      var zip = new JSZip();
      //
      // Add the file to archive
      zip.file(file.path, data);
      //
      // Generates compress file
      var blob = zip.generate({type: "blob"});
      //
      this.fs[file.type].root.getFile(zipFile.path, {create: true, exclusive: false}, function (entry) {
        // Create the object writer
        entry.createWriter(function (fileWriter) {
          // Listen to error events
          fileWriter.onerror = cb;
          //
          // Listen to write end events
          fileWriter.onwriteend = function (progressEvent) {
            cb();
          };
          //
          // Write the zip file
          fileWriter.write(blob);
        }, cb); // Writer creation error
      }, cb); // Creating file error
    }.bind(this));
  }.bind(this));
};


/**
 * Unzip an archiver
 * @param {File} file
 * @param {Directory} directory
 * @param {function} cb
 */
App.LocalDriver.prototype.unzip = function (file, directory, cb)
{
  // Save the reader object
  var reader = file.reader;
  //
  // Open the zipped file
  this.openFile(file, function (err) {
    if (err)
      return cb(err);
    //
    // Reads the file to zip
    this.read(file, null, 0, function (data, err) {
      if (err)
        return cb(err);
      //
      // Restore the reader
      file.reader = reader;
      //
      // Create the JSZip object
      var zip = new JSZip(data);
      //
      // Creates the root directory
      this.mkDir(directory, function (err) {
        if (err)
          return cb(err);
        //
        var files = [];
        for (var k in zip.files)
          files.push(zip.files[k]);
        //
        // Async each series:
        // applies the function iterator to each item in files array
        // and call the callback for when it has finished
        async.eachSeries(files, function (entry, cbDone) {
          // If the entry is a directory
          if (entry.dir)
            return this.mkDir(this.directory(directory.path + "/" + entry.name, file.type), cbDone);
          //
          // If the entry is a file create the file
          this.fs[directory.type].root.getFile(directory.path + "/" + entry.name, {create: true}, function (entry) {
            entry.createWriter(function (fileWriter) {
              var writer = fileWriter;
              writer.seek(0);
              //
              // Listen to error events
              writer.onerror = cbDone;
              //
              // Listen to write end events
              writer.onwriteend = function (progressEvent) {
                cbDone();
              };
              //
              // Write
              writer.write(new Blob([entry.asArrayBuffer()]));
            }, cbDone); // Writer creation error
          }, cbDone); // Creating file error
        }.bind(this), cb); // eachSeries error
      }.bind(this));
    }.bind(this));
  }.bind(this));
};


/**
 * Create the directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.LocalDriver.prototype.mkDir = function (directory, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(directory.type, function (err) {
    if (err)
      return cb(err);
    //
    // Check if the directory is the root
    if (directory.path === "/")
      return cb();
    //
    this.mkDirRecursive(this.fs[directory.type].root, directory.path.split("/"), function (entry, err) {
      directory.entry = entry;
      cb(err);
    });
  }.bind(this));
};


/**
 * Creates recursive folder
 * @param {Object} entry
 * @param {string} folders
 * @param {function} cb
 */
App.LocalDriver.prototype.mkDirRecursive = function (entry, folders, cb)
{
  // Throw out './' or '/' if present on the beginning of our path.
  var folder = folders.shift();
  if (folder === "." || folder === "")
    folder = folders.shift();
  //
  entry.getDirectory(folder, {create: true}, function (dirEntry) {
    // Recursively add the new subfolder (if we still another to create).
    if (folders.length)
      this.mkDirRecursive(dirEntry, folders, cb);
    else
      cb(dirEntry);
  }.bind(this), function (err) {
    // Creation error
    cb(null, err);
  });
};


/**
 * checks the existence of the directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.LocalDriver.prototype.dirExists = function (directory, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(directory.type, function (err) {
    if (err)
      return cb(false);
    //
    // Try to get the dirEntry to see if there is
    this.fs[directory.type].root.getDirectory(directory.path, {create: false}, function () {
      cb(true);
    }, function () {
      cb(false);
    });
  }.bind(this));
};


/**
 * Copies the entire directory
 * @param {Directory} srcDir
 * @param {Directory} dstDir
 * @param {function} cb
 */
App.LocalDriver.prototype.copyDir = function (srcDir, dstDir, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(srcDir.type, function (err) {
    if (err)
      return cb(err);
    //
    // Get the parent directory
    var pathParts = dstDir.path.split("/");
    var dirPath = pathParts.slice(0, pathParts.length - 1).join("/");
    if (dirPath[0] === "/")
      dirPath = dirPath.substr(1);
    //
    var newName = pathParts[pathParts.length - 1];
    //
    this.fs[srcDir.type].root.getDirectory(dirPath, {create: true}, function (destinationDirEntry) {
      // Checks and sets the entry property (if it doesn't exist)
      this.checkEntry(srcDir, function (err) {
        if (err)
          return cb(err);
        //
        // Copy
        srcDir.entry.copyTo(destinationDirEntry, newName, function (newEntry) {
          dstDir.entry = newEntry;
          cb();
        }, cb); // copyTo error
      });
    }.bind(this), cb); // getDirectory error
  }.bind(this));
};


/**
 * Read recursively the content of directory
 * @param {Directory} directory
 * @param {Integer} depth
 * @param {function} cb
 */
App.LocalDriver.prototype.readDirectory = function (directory, depth, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(directory.type, function (err) {
    if (err)
      return cb(null, err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(directory, function (err) {
      if (err)
        return cb(null, err);
      //
      // Recursive core
      var readDirRecursive = function (entries, content, cbRec) {
        // No directory unexplored: end of recursion
        if (!entries.length)
          return cbRec(content);
        //
        if (entries[0].depth > depth) {
          // Remove the folder because is deeper than depth
          entries.shift();
          //
          // recall the function
          return readDirRecursive(entries, content, cbRec);
        }
        //
        // Reads the current directory
        this.fs[directory.type].root.getDirectory(entries[0].entry.fullPath, {create: false}, function (entry) {
          // Create dirReader object
          var dirReader = entry.createReader();
          dirReader.readEntries(function (results) {
            //
            for (var i = 0; i < results.length; i++) {
              // Add the element to the content array
              content = content.concat(results[i]);
              //
              // If the element is a directory, I add it to the array of folders to be scanned
              if (results[i].isDirectory)
                entries.push({depth: entries[0].depth + 1, entry: results[i]});
            }
            //
            // Remove the folder just examined from the array
            entries.shift();
            //
            // Recall the function
            readDirRecursive(entries, content, cbRec);
          }, function (err) {
            //readEntries error
            cbRec(null, err);
          });
        });
      }.bind(this);
      //
      // Set default depth to 0
      depth = depth || 0;
      //
      // Array of files/directory objects
      var content = [];
      //
      // Array of directories yet to be examined
      var dir = [];
      dir.push({depth: 0, entry: directory.entry});
      //
      readDirRecursive(dir, content, function (entries, err) {
        if (err)
          return cb(null, err);
        //
        // No file/directory inside
        if (entries.length < 1)
          return cb();
        //
        // Order the entries by name
        var content = [];
        for (var i = 0; i < entries.length; i++) {
          // Create file/directory object by setting the property correctly
          var object;
          if (entries[i].isFile)
            object = this.file(entries[i].fullPath.substr(1), directory.type);
          else
            object = this.directory(entries[i].fullPath.substr(1), directory.type);
          content.push(object);
        }
        cb(content);
      }.bind(this));
    }.bind(this));
  }.bind(this));
};


/**
 * Zip directory
 * @param {Directory} directory
 * @param {File} zipFile
 * @param {function} cb
 */
App.LocalDriver.prototype.zipDirectory = function (directory, zipFile, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(directory.type, function (err) {
    if (err)
      return cb(err);
    //
    // Reads the files and the sub directories
    this.readDirectory(directory, function (files, err) {
      if (err)
        return cb(err);
      //
      // New JSZip object
      var zip = new JSZip();
      //
      // async each series:
      // applies the function iterator to each item in files array
      // and call the callback for when it has finished
      async.eachSeries(files, function (entry, cbDone) {
        // If the entry is a file
        if (entry instanceof App.File) {
          // Save the status of the readear of file
          var oldReader = entry.reader;
          //
          entry.open(function (err) {
            if (err)
              return cbDone(err);
            //
            // Reads the file to zip
            entry.read(null, 0, function (data, err) {
              if (err)
                return cbDone(err);
              //
              // Restore the reader to original status
              entry.reader = oldReader;
              //
              zip.file(entry.path, data);
              cbDone();
            });
          });
        }
        else {
          // If the entry is a directory add it to zip archive
          zip.folder(entry.path);
          cbDone();
        }
      }, function (err) {
        if (err)
          return cb(err);
        //
        // Generate compress file
        var blob = zip.generate({type: "blob"});
        //
        var zipPath = zipFile.path;
        this.fs[directory.type].root.getFile(zipPath, {create: true, exclusive: false}, function (entry) {
          // Create the object writer
          entry.createWriter(function (fileWriter) {
            // Listen to error events
            fileWriter.onerror = cb;
            //
            // Listen to write end events
            fileWriter.onwriteend = function (progressEvent) {
              cb();
            };
            //
            // Write the zip file
            fileWriter.write(blob);
          }, cb); // Writer creation error
        }, cb); // Creating file error
      }.bind(this));
    }.bind(this));
  }.bind(this));
};


/**
 * Removes the entire directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.LocalDriver.prototype.removeDirRecursive = function (directory, cb)
{
  // Checks and creates fs object (if it doesn't exist)
  this.initLocalFs(directory.type, function (err) {
    if (err)
      return cb(err);
    //
    // Checks and sets the entry property (if it doesn't exist)
    this.checkEntry(directory, function (err) {
      // If file not exists it's ok, do nothing
      if (err)
        return cb();
      //
      // Remove
      directory.entry.removeRecursively(function () {
        delete directory.entry;
        cb();
      }, cb); // Remove error
    });
  }.bind(this));
};


/**
 * Makes an HTTP GET request to a web server
 * @param {object} url
 * @param {string} method
 * @param {object} options
 * @param {function} cb
 */
App.LocalDriver.prototype.httpRequest = function (url, method, options, cb)
{
  options = Object.assign({}, options);
  //
  // Extract and encode params from  options
  // * @param {object} obj
  var parseParams = function (obj) {
    var str = [];
    for (var p in obj) {
      if (obj.hasOwnProperty(p))
        str.push(encodeURIComponent(p) + "=" + encodeURIComponent(obj[p]));
    }
    return str.join("&");
  };
  //
  var uri = url.url;
  //
  // Multipart request
  var multiPart = false;
  //
  // Download
  var download = false;
  //
  // Upload
  var upload = false;
  //
  // Creates request object
  var req = new XMLHttpRequest();
  //
  // Checks case
  switch (method) {
    case "POST":
      multiPart = true;
      break;

    case "DOWNLOAD":
      download = true;
      if (options.method === "POST")
        method = "POST";
      else
        method = "GET";
      break;

    case "UPLOAD":
      multiPart = true;
      upload = true;
      method = "POST";
      break;
  }
  //
  // Check user options
  if (!options.headers)
    options.headers = {};
  //
  // Get eventually values for the autentication
  if (options.authentication)
    options.headers.Authorization = "Basic " + btoa(options.authentication.username + ":" + options.authentication.password);
  //
  // Set timeout request
  if (options.timeOut)
    req.timeout = options.timeOut;
  //
  // If not specified, the post request type is multipart
  if (options.headers && options.headers["Content-Type"])
    multiPart = false;
  //
  // Create the complete query string
  var params = parseParams(options.params);
  //
  // Add it to the url (for GET method)
  if (method !== "POST" && params)
    uri += (uri.indexOf("?") > -1 ? "&" : "?") + params;
  //
  // If the request is a multipart, add onprogress event to track the upload
  if (upload) {
    req.upload.onprogress = function (evt) {
      if (evt.lengthComputable) {
        if (url.onUploadProgress(evt.loaded, evt.total) === false) {
          // Abort request
          req.abort();
          cb({error: new Error("upload stopped")});
        }
      }
    };
  }
  //
  // Initialize request
  req.open(method, uri, true);
  //
  // To handle better response, want receive binary data
  req.responseType = (download ? "arraybuffer" : options.responseType || "text");
  //
  var data = null;
  //
  // Custom body case
  if (options.body) {
    if (typeof options.bodyType === "string")
      options.headers["Content-Type"] = options.bodyType;
    else if (options.headers && !options.headers["Content-Type"])
      options.headers["Content-Type"] = "application/octet-stream";
    //
    // Types allowed for the custom body are: string and ArrayBuffer
    if (options.body instanceof ArrayBuffer || typeof options.body === "string")
      data = options.body;
    else
      return cb({error: new Error("Custom body must be string or ArrayBuffer")});
  }
  // Multipart request
  else if (multiPart) {
    // Create formdata object
    data = new FormData();
    //
    // Add params to formData object
    var sepParams = params.split("&");
    for (var i = 0; i < sepParams.length; i++) {
      var pair = sepParams[i].split('=');
      if (pair[0] && pair[1])
        data.append(decodeURIComponent(pair[0]), decodeURIComponent(pair[1]));
    }
    //
    // Upload
    if (upload) {
      // Get file object
      options._file.open(function (err) {
        if (err)
          return cb({error: err});
        //
        // Add the file
        var value = null;
        if (options._fileContentType)
          value = new Blob([options._file.fileObject], {type: options._fileContentType});
        else
          value = options._file.fileObject;
        data.append(options._nameField, value, options._fileName);
        //
        // Send request
        req.send(data);
      });
    }
  }
  //
  // Add custom headers
  if (options.headers) {
    for (var index in options.headers)
      req.setRequestHeader(index, options.headers[index]);
  }
  //
  if (!upload)
    req.send(data);
  //
  // Listen to timeout event
  req.ontimeout = function () {
    cb({error: new Error("timeout request error")});
  };
  //
  // Listen to progress events
  req.onprogress = function (evt) {
    if (!url.onDownloadProgress(evt.loaded, evt.total)) {
      // Stop response writing
      req.abort();
      cb({error: new Error("download stopped")});
    }
  };
  //
  // Listen to request error events
  req.onerror = function () {
    cb({error: new Error("unknown error")});
  };
  //
  // Listen to response load event
  req.onload = function () {
    // default encoding
    var encoding;
    var body = req.response;
    //
    if (body instanceof ArrayBuffer) {
      // get encoding
      var responseContentType = req.getResponseHeader("Content-Type");
      if (responseContentType && responseContentType.length >= 3) {
        if (responseContentType.indexOf("text") >= 0 || responseContentType.substring(responseContentType.length - 3,
                responseContentType.length) === "xml" || responseContentType.indexOf("application/json") >= 0) {
          //
          // Default encoding
          encoding = "utf-8";
          //
          // Get encoding from "text/... charset=..." header
          if (responseContentType.indexOf("charset") >= 0)
            encoding = responseContentType.substr(responseContentType.indexOf("charset") + 8);
        }
      }
      //
      // Encode the response
      if (encoding && typeof TextDecoder !== "undefined") {
        var dataView = new DataView(req.response);
        try {
          body = new TextDecoder(encoding).decode(dataView);
        }
        catch (err) {
          // If there is an error, the default decoding uses utf-8
          body = new TextDecoder("utf-8").decode(dataView);
        }
      }
    }
    //
    // Create the response object
    var response = {
      status: req.status,
      headers: {},
      body: body
    };
    //
    // Headers as map
    if (req.getAllResponseHeaders()) {
      req.getAllResponseHeaders().trim().split("\r\n").forEach(function (v) {
        var ss = v.split(":");
        response.headers[ss[0].trim()] = ss[1].trim();
      });
    }
    //
    // Download case
    if (download) {
      // Create the file
      options._file.create(null, function (file, err) {
        if (err)
          return cb({error: err});
        //
        // Write the file
        options._file.write(body, null, null, 0, function (file, err) {
          if (err)
            cb({error: err});
          else {
            this.getFilePublicUrl(options._file, function (ris) {
              options._file._publicUrl = ris;
              cb(response);
            });
          }
        }.bind(this));
      }.bind(this));
    }
    else
      cb(response);
  }.bind(this);
};


/**
 * Put a local file on the other side file system (local/remote)
 * @param {File} localFile
 * @param {File} dstFile
 * @param {function} cb
 */
App.LocalDriver.prototype.put = function (localFile, dstFile, cb)
{
  cb(new Error("Method not supported"));
};


//
/* LineReader
 * https://github.com/mgmeyers/LineReader
 *
 * Copyright 2014 Matthew Meyers <hello@matthewmeye.rs>
 * Released under the MIT license:
 * http://www.opensource.org/licenses/mit-license.php
 *
 * Reads a file one line at a time.
 *
 * @param {Object} options -- options for the line reader
 *
 * Available options:
 *   chunkSize {Integer} -- how much of the file to read at a time
 */


var LineReader = function (options) {
  /**
   * If 'this' isn't an instance of 'LineReader', then the user forgot to use
   * the 'new' keyword, let's do it for them, otherwise, 'this' will reference
   * the 'window' object
   */
  if (!(this instanceof LineReader)) {
    return new LineReader(options);
  }
  /**
   * We'll use '_internals' to store data we don't want public facing
   *
   * We'll also need a reference to 'this', as it will be overridden in the
   * 'onload' and 'onerror' events
   */
  var internals = this._internals = {};
  var self = this;
  /**
   * Let's create a 'FileReader' instance, we'll only need one per 'LineReader'
   * instance
   */
  internals.reader = new FileReader();
  /**
   * If 'chunkSize' has been set by the user, use that value, otherwise,
   * default to 1024
   */
  internals.chunkSize = (options && options.chunkSize)
          ? options.chunkSize
          : 1024;
  /**
   * Let's create an object to house user defined event callbacks
   */
  internals.events = {};
  /**
   * 'canRead' will be set to false if the LineReader#abort method is fired
   */
  internals.canRead = true;
  /**
   * FileReader#onload' event. This gets called when any read operations have
   * completed
   */
  internals.reader.onload = function () {
    /**
     * Store the processed text by tagging it on to any existing processed text
     * 'this' refers to our 'FileReader' instance
     */
    internals.chunk += this.result;
    /**
     * If the processed text contains a newline character
     */
    if (/\r|\n/.test(internals.chunk)) {
      /**
       * Split the text into an array of lines
       */
      internals.lines = internals.chunk.match(/^.*((\r\n|\n|\r)|$)/gm);
      /**
       * If there is still more data to read, save the last line, as it may be
       * incomplete
       */
      if (self._hasMoreData()) {
        internals.chunk = internals.lines.pop();
      }
      /**
       * Start stepping through each line
       */
      self._step();
      /**
       * If the text did not contain a newline character
       */
    }
    else {
      /**
       * Start another round of the read process if there is still data to read
       */
      if (self._hasMoreData()) {
        return self.read();
      }
      /**
       * If there is no data left to read, but there is still data stored in
       * 'chunk', emit it as a line
       */
      if (internals.chunk.length) {
        return self._emit('line', [
          internals.chunk,
          self._emit.bind(self, 'end')
        ]);
      }
      /**
       * if there is no data stored in 'chunk', emit the end event
       */
      self._emit('end');
    }
  };
  /**
   * 'FileReader#onerror' event. This gets called any time there is an error
   * reading a file
   */
  internals.reader.onerror = function () {
    /**
     * Emit the error event, passing along the error object to the callback
     * 'this' refers to our 'FileReader' instance
     */
    self._emit('error', [this.error]);
  };
};
/**
 * LineReader#on
 *
 * Binds events
 *
 * @param {String} eventName -- the name of the event to bind to
 * @param {Function} cb -- the function to execute when the event is triggered
 */
LineReader.prototype.on = function (eventName, cb) {
  this._internals.events[ eventName ] = cb;
};
/**
 * LineReader#read
 *
 * Starts the read process
 *
 * @param {File} file -- The file reference to process
 * @param {String} encoding -- The encoding of the file
 */
LineReader.prototype.read = function (file, encoding) {
  var internals = this._internals;
  /**
   * If 'file' is defined, then we want to get some information about it and
   * reset 'readPos', 'chunk', and 'lines'
   */
  if (typeof file !== 'undefined') {
    internals.file = file;
    internals.fileLength = file.size;
    internals.readPos = 0;
    internals.chunk = '';
    internals.lines = [];
  }
  /**
   * Extract a section of the file for reading starting at 'readPos' and
   * ending at 'readPos + chunkSize'
   */
  var blob = internals.file.slice(internals.readPos, internals.readPos + internals.chunkSize);
  /**
   * Update our current read position
   */
  internals.readPos += internals.chunkSize;
  /**
   * Read the blob as text
   */
  internals.reader.readAsText(blob, encoding);
};
/**
 * LineReader#abort
 *
 * Stops the read process
 */
LineReader.prototype.abort = function () {
  this._internals.canRead = false;
};
/**
 * LineReader#_step
 *
 * Internal: gets the next line and emits it as a `line` event
 */
LineReader.prototype._step = function () {
  var internals = this._internals;
  /**
   * If there are no lines left to emit and there is still data left to read,
   * start the read process again, otherwise, emit the 'end' event
   */
  if (internals.lines.length === 0) {
    if (this._hasMoreData()) {
      return this.read();
    }
    return this._emit('end');
  }


  /**
   * If the reading process hasn't been aborted, emit the first element of the
   * line array, and pass in '_step' for the user to call when they are ready
   * for the next line. We have to bind '_step' to 'this', otherwise it will be
   * in the wrong scope when the use calls it
   */
  if (internals.canRead) {
    this._emit('line', [
      internals.lines.shift(),
      this._step.bind(this)
    ]);
  }
  else {
    /**
     * If we can't read, emit the 'end' event
     */
    this._emit('end');
  }
};
/**
 * LineReader#_hasMoreData
 *
 * Internal: determines if there is still more data to read.
 */
LineReader.prototype._hasMoreData = function () {
  var internals = this._internals;
  return internals.readPos <= internals.fileLength;
};
/**
 * LineReader#_emit
 *
 * Internal: handles event emissions
 *
 * @param  {String} event -- The event name to emit
 * @param  {Array} args -- An array of arguments to send to the event callback
 */
LineReader.prototype._emit = function (event, args) {
  var boundEvents = this._internals.events;
  /**
   * if the user has bound the requested event
   */
  if (typeof boundEvents[event] === 'function') {
    /**
     * Use apply to ensure correct scope, and pass in the 'args' array to
     * be used as arguments for the callback
     */
    boundEvents[event].apply(self, args);
  }
};
