/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */


/* global ArrayBuffer, module */

var App = App || {};

if (module)
  App.FS = require("./fs");

/**
 * @class NodeDriver
 * Represents a node driver object,that will handle files and folders
 * @param {Object} options
 */
App.NodeDriver = function (options)
{
  // Import modules
  if (module && !App.View) {
    App = require("../requireAll");
    App.nodeFs = require("fs");
    App.pathNode = require("path");
    App.http = require("http");
    App.https = require("https");
    App.querystring = require("querystring");
    App.async = require("async"); //license and detail: https://github.com/caolan/async
    App.lazy = require("lazy"); //License and detail: https://github.com/pkrumins/node-lazy
    App.fsExtra = require("fs.extra"); //license and detail: https://www.npmjs.com/package/fs.extra
    App.archiver = require("archiver"); //license and detail: https://github.com/ctalkington/node-archiver
    App.yauzl = require("yauzl"); //license and detail: https://github.com/thejoshwolfe/yauzl
  }
  //
  // String containing the root path of the App
  this.appRoot = options.rootPath.replace(/\\/g, "\/");
  //
  // String containing the base url of the server
  this.publicUrl = options.publicUrl;
};

// Make App.NodeDriver extend App.FS
App.NodeDriver.prototype = new App.FS(true);

App.NodeDriver.typeMap = {
  private: "files/private",
  blob: "files/blobs",
  temp: "files/temp",
  permanent: "files",
  resource: "resources"
};


/*
 * Given a relative path, return the internal (server) path to the file
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.getFileFullPath = function (file, cb)
{
  cb([this.appRoot, App.NodeDriver.typeMap[file.type], file.path].join("/"));
};


/*
 * Given a relative path, return the public url to the file
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.getFilePublicUrl = function (file, cb)
{
  cb([this.publicUrl, App.NodeDriver.typeMap[file.type], file.path].join("/"));
};


/**
 * Checks the validity of path
 * @param {boolean} write
 * @param {File/Directory} obj
 * @param {string} forceType
 */
App.NodeDriver.prototype.checkPath = function (write, obj, forceType)
{
  // If the type is not defined then you work with a normal file, you need to check that it does
  // not touch the system folders
  if (!obj.type) {
    // Default directory that should not be changed or deleted
    if (["private", "blobs", "temp"].indexOf(obj.path.split("/")[0]) >= 0)
      return;
  }
  //
  // Parent folder of file/directory: determined by the type of file
  var parentFolder = App.NodeDriver.typeMap[forceType || obj.type];
  if ((forceType || obj.type) === App.FS.internalType.resource) {
    // Because the resource files can only be read
    if (write)
      return;
    //
    // $resource contains a bit of path, should be removed in this case
    if (obj.path.indexOf("resources/") > -1)
      obj.path = obj.path.split("/").pop();
  }
  //
  // Absolute path
  var realPath = [this.appRoot, parentFolder, obj.path].join("/");
  //
  // Remove final slash
  if (realPath.endsWith("/"))
    realPath = realPath.slice(0, -1);
  //
  var sepRealPath = realPath.split("/");
  //
  // validPath fills with paths accepted
  var validPaths = [];
  validPaths.push(([this.appRoot, parentFolder].join("/")).split("/"));
  //
  if (!write)
    validPaths.push(([this.appRoot, "resources"].join("/")).split("/"));
  //
  var validPathsNumber = validPaths.length;
  for (var i = 0; i < validPaths.length; i++) {
    for (var j = 0; j < validPaths[i].length; j++) {
      if (validPaths[i][j] !== sepRealPath[j]) {
        validPathsNumber--;
        break;
      }
    }
  }
  //
  // If the path is invalid returns null otherwise realpath
  if (validPathsNumber > 0)
    return realPath;
};


/**
 * Creates the file physically (Opens the file and overwrites it)
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.createFile = function (file, cb)
{
  // Check the validity of the path (writing)
  var path = this.checkPath(true, file);
  if (!path)
    return cb(new Error("Invalid path"));
  //
  // Object of type write stream used to write
  file.wstream = App.nodeFs.createWriteStream(path, {encoding: file.encoding});
  //
  // Listen to open event
  file.wstream.on("open", function () {
    file.wstream.removeAllListeners("error");
    cb();
  });
  //
  // Listen to next error event
  file.wstream.once("error", function (error) {
    delete file.wstream;
    cb(error);
  });
};


/**
 * Opens the file for reading
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.openFile = function (file, cb)
{
  // Check the validity of the path (reading)
  var fullPath = this.checkPath(false, file);
  if (!fullPath)
    return cb(new Error("Invalid path"));
  //
  // Checking the existence of the file
  this.fileExists(file, function (exists) {
    if (!exists)
      return cb(new Error("File does not exist"));
    //
    // Object of type read stream used to read
    file.rstream = App.nodeFs.createReadStream(fullPath, {encoding: file.encoding});
    //
    // Listen to open event
    file.rstream.on("open", function () {
      file.rstream.removeAllListeners("error");
      cb();
    });
    //
    // Listen to next error event
    file.rstream.once("error", function (error) {
      delete file.rstream;
      cb(error);
    });
  });
};


/**
 * Opens the file to append data
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.openFileForAppend = function (file, cb)
{
  // Check the validity of the path (wiriting)
  var path = this.checkPath(true, file);
  if (!path)
    return cb(new Error("Invalid path"));
  //
  // Object of type write stream used to write (setting the append flag)
  file.wstream = App.nodeFs.createWriteStream(path, {flags: "a", encoding: file.encoding});
  //
  // Listen to open event
  file.wstream.on("open", function () {
    file.wstream.removeAllListeners("error");
    cb();
  });
  //
  // Listen to next error event
  file.wstream.once('error', function () {
    delete file.wstream;
    cb(new Error("Open file for append error"));
  });
};


/**
 * Closes the file
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.close = function (file, cb)
{
  // Delete variables stream
  delete file.rstream;
  delete file.rspos;
  //
  if (file.wstream) {
    file.wstream.on("finish", function () {
      delete file.wstream;
      cb();
    });
    //
    // Listen to next error event
    file.wstream.once('error', function () {
      delete file.wstream;
      cb(new Error("Close file error"));
    });
    //
    file.wstream.end();
  }
  else
    cb();
};


/**
 * Checks existence of a file
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.fileExists = function (file, cb)
{
  // Check the validity of the path (reading)
  var path = this.checkPath(false, file);
  if (!path)
    return cb(false);
  //
  // Check stats of file
  App.nodeFs.stat(path, function (err, stats) {
    cb(err || !stats.isFile() ? false : true);
  });
};


/**
 * Reads a text file line by line through a loop that reads each line of the file
 * @param {File} file
 * @param {function} cb
 * @param {function} cbEnd
 */
App.NodeDriver.prototype.readLine = function (file, cb, cbEnd)
{
  // Check that the file is opened (reading)
  if (!file.rstream)
    return cbEnd(new Error("File not opened"));
  //
  file.lineStreamPause = false;
  var cbCalled = false;
  //
  // Lazy (external module) object used to read line by line
  var lazy = new App.lazy(file.rstream);
  //
  // "line" contains the text line
  lazy.lines.forEach(function (line) {
    // If was called the "break" function
    if (file.lineStreamPause) {
      if (!cbCalled) {
        cbCalled = true;
        cbEnd();
      }
      return;
    }
    //
    // To avoid printing zero when I find a blank line
    if (line.length === 1 && line[0] === 48)
      cb("");
    else
      cb(line.toString(file.encoding));
  });
  //
  // Necessary to issue the event "end" only after reading the last line
  file.rstream.on("end", function () {
    if (!cbCalled)
      cbEnd();
  });
  //
  // listen to error event
  file.rstream.on("error", function (err) {
    cbEnd(err);
  });
};


/**
 * stops reading line by line
 * @param {File} file
 */
App.NodeDriver.prototype.break = function (file)
{
  // Sets the variable flag to stop the reading line by line
  file.lineStreamPause = true;
};



/**
 * Reads a block of data, return an array buffer
 * @param {File} file
 * @param {int} length
 * @param {int} offset
 * @param {function} cb
 */
App.NodeDriver.prototype.read = function (file, length, offset, cb)
{
  // Check the validity of the path (reading)
  var path = this.checkPath(false, file);
  if (!path)
    return cb(null, new Error("Invalid path"));
  //
  // Check that the file is opened (reading)
  if (!file.rstream)
    return cb(null, new Error("File not opened"));
  //
  // Setting the correct limit: if length or offset are null take respectively the current position of the
  // reader and the position of the last byte of the file
  var opts = {};
  if (typeof offset === "number")
    opts.start = offset;
  else if (file.rspos)
    opts.start = file.rspos;
  //
  if (length)
    opts.end = opts.start + length - 1;
  opts.encoding = null;
  //
  // I recreate the reader with the correct limits
  file.rstream = App.nodeFs.createReadStream(path, opts);
  //
  // I create the buffer where the bytes read will be placed on
  var buf = new Buffer(length || 1024);
  var bytesRead = 0;
  //
  // Listen to next error event
  file.rstream.once("error", function (error) {
    cb(null, error);
  });
  //
  // Listen to data events
  file.rstream.on("data", function (chunk) {
    // Check if buffer is large enough
    if (buf.length < bytesRead + chunk.length) {
      // Enlarge buffer
      var bufTmp = new Buffer(Math.max(buf.length, chunk.length) * 2);
      buf.copy(bufTmp);
      buf = bufTmp;
    }
    //
    // Write the bytes in the buffer
    chunk.copy(buf, bytesRead);
    bytesRead += chunk.length;
  });
  //
  // Listen to end event
  file.rstream.once("end", function () {
    file.rspos = (opts.start || 0) + bytesRead;
    //
    // Converts the Buffer to ArrayBuffer
    cb(bytesRead ? (new Uint8Array(buf.slice(0, bytesRead))).buffer : null);
  });
};


/**
 * Read the whole file as text
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.readAll = function (file, cb)
{
  // Check the validity of the path (reading)
  var path = this.checkPath(false, file);
  if (!path)
    return cb(null, new Error("Invalid path"));
  //
  // Read content
  App.nodeFs.readFile(path, {encoding: file.encoding || "utf-8"}, function (err, content) {
    cb(content, err);
  });
};


/**
 * Read the whole file as text
 * @param {File} file
 */
App.NodeDriver.prototype.readSync = function (file)
{
  // Check the validity of the path (reading)
  var path = this.checkPath(false, file);
  if (!path)
    throw new Error("Invalid path");
  //
  return App.nodeFs.readFileSync(path, {encoding: file.encoding});
};


/**
 * Writes the data or the string given
 * @param {File} file
 * @param {string/buffer} data
 * @param {int} offset
 * @param {int} size
 * @param {int} position
 * @param {function} cb
 */
App.NodeDriver.prototype.write = function (file, data, offset, size, position, cb)
{
  // Check that the file is opened (writing)
  if (!file.wstream)
    return cb(new Error("File not open for write"));
  //
  if (!data)
    return cb(new Error("No data to write"));
  //
  if (!(data instanceof ArrayBuffer) && typeof data !== "string")
    data = JSON.stringify(data);
  //
  // Checks if param data is an array buffer or a string
  if (typeof data === "string") {
    // If not specified the default encoding is utf-8
    file.encoding = file.encoding || "utf-8";
    //
    file.wstream.write(data, file.encoding, function (err) {
      file.wstream.removeAllListeners("error");
      cb(err);
    });
  }
  else {
    // Setting the correct limit: if offset or size are null take respectively the current
    // position of the writer and the position of the last byte of the file
    if (position !== null && position >= 0)
      file.wstream.pos = position;
    //
    if (!size)
      size = data.length;
    //
    if (!offset || offset < 0)
      offset = 0;
    //
    var buffer = new Buffer(new Uint8Array(data));
    buffer.slice(offset, offset + size);
    //
    // Write buffer
    file.wstream.write(buffer, function (error) {
      file.wstream.removeAllListeners("error");
      cb(error);
    });
    //
    // Listen to next error event; if not listened don't is called write callback
    file.wstream.once("error", function (error) {
    });
  }
};


/**
 * Copy the file
 * @param {File} file
 * @param {File} newFile
 * @param {function} cb
 */
App.NodeDriver.prototype.copyFile = function (file, newFile, cb)
{
  // Check that the relative paths are valid and I get absolute paths
  var oldFullPath = this.checkPath(false, file);
  if (!oldFullPath)
    return cb(new Error("Invalid old path"));
  //
  var newFullPath = this.checkPath(true, newFile);
  if (!newFullPath)
    return cb(new Error("Invalid new path"));
  //
  // Check the file existence
  this.fileExists(file, function (exists) {
    if (!exists)
      return cb(new Error("Error: file to copy doesn't exist"));
    //
    var cbCalled = false;
    var done = function (err) {
      if (cbCalled)
        return;
      //
      cbCalled = true;
      cb(err);
    };
    //
    // Create reader and writer streams
    var rs = App.nodeFs.createReadStream(oldFullPath, {encoding: null});
    var ws = App.nodeFs.createWriteStream(newFullPath, {encoding: null});
    //
    // Listen to error and close events
    rs.on("error", done);
    ws.on("error", done);
    ws.on("close", done);
    //
    // Copy file trough pipe the streams
    rs.pipe(ws);
  });
};


/**
 * Rename a file or directory
 * @param {File/Directory} obj
 * @param {string} newName
 * @param {function} cb
 */
App.NodeDriver.prototype.renameObject = function (obj, newName, cb)
{
  // Check that the relative paths are valid and I get absolute paths
  var path = this.checkPath(false, obj);
  if (!path)
    return cb(null, new Error("Invalid old path"));
  //
  var newPath = path.substring(0, path.lastIndexOf("/") + 1) + newName;
  //
  // Rename
  App.nodeFs.rename(path, newPath, cb);
};


/**
 * Return the file size (in bytes)
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.fileLength = function (file, cb)
{
  // Check the validity of the path (reading)
  var path = this.checkPath(false, file);
  if (!path)
    return cb(null, new Error("Invalid path"));
  //
  // Get the size
  App.nodeFs.stat(path, function (err, stats) {
    if (err)
      cb(null, err);
    else
      cb(stats.size);
  });
};


/**
 * Return the last modified file date
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.fileDateTime = function (file, cb)
{
  // Check the validity of the path (reading)
  var path = this.checkPath(false, file);
  if (!path)
    return cb(null, new Error("Invalid path"));
  //
  // Get the date
  App.nodeFs.stat(path, function (err, stats) {
    if (err)
      cb(null, err);
    else
      cb(stats.mtime);
  });
};


/**
 * Deletes a file
 * @param {File} file
 * @param {function} cb
 */
App.NodeDriver.prototype.deleteFile = function (file, cb)
{
  // Check the validity of the path (writing)
  var path = this.checkPath(true, file);
  if (!path)
    return cb(new Error("Invalid path"));
  //
  // Remove
  App.nodeFs.unlink(path, cb);
};


/**
 * Zip a file
 * @param {File} file
 * @param {File} zipFile
 * @param {function} cb
 */
App.NodeDriver.prototype.zipFile = function (file, zipFile, cb)
{
  var path = this.checkPath(false, file);
  var zipPath = this.checkPath(true, zipFile);
  //
  if (!path || !zipPath)
    return cb(new Error("Invalid paths"));
  //
  var cbCalled = false;
  var done = function (err) {
    if (cbCalled)
      return;
    //
    cbCalled = true;
    cb(err);
  };
  //
  // Create the write stream
  var output = App.nodeFs.createWriteStream(zipPath);
  //
  // Create the archive object
  var archive = App.archiver('zip');
  //
  // Listen to next error event
  output.once("error", function (err) {
    deleteVoidZip(err);
  });
  //
  // Listen to close finalization archive
  output.on('close', function () {
    done();
  });
  //
  // Listen to error event
  archive.on("error", function (err) {
    deleteVoidZip(err);
  });
  //
  output.on("open", function () {
    // Push data into the archive
    archive.pipe(output);
    //
    var input = App.nodeFs.createReadStream(path);
    //
    // Listen to error event
    input.on("error", function (err) {
      deleteVoidZip(err);
    });
    //
    // Listen to error event
    input.on("open", function () {
      // Get the file name
      var sepPath = file.path.split("/");
      archive.append(input, {name: sepPath[sepPath.length - 1]}).finalize();
    });
  });
  //
  // Function that deletes the new zip file (if there is an error)
  var deleteVoidZip = function (err) {
    archive.finalize();
    App.nodeFs.unlink(zipPath, function (err1) {
      done(err || err1);
    });
  };
};


/**
 * Unzip the archiver
 * @param {File} file
 * @param {Directory} directory
 * @param {function} cb
 */
App.NodeDriver.prototype.unzip = function (file, directory, cb)
{
  // Check the validity of the path (writing)
  var zipPath = this.checkPath(false, file);
  if (!zipPath)
    return cb(new Error("Invalid path"));
  //
  var dirPath = this.checkPath(true, directory);
  if (!dirPath)
    return cb(new Error("Invalid newPath"));
  //
  // Create the parent directory
  App.fsExtra.mkdirs(dirPath, function (err) {
    if (err)
      return cb(err);
    //
    // Opens the archive for decompression using the library yauzl
    App.yauzl.open(zipPath, {lazyEntries: true}, function (err, zipfile) {
      if (err)
        return cb(err);
      //
      var ok = false;
      zipfile.readEntry();
      //
      // For each file/directory
      zipfile.on("entry", function (entry) {
        // Directory file names end with '/'
        if (entry.fileName.endsWith("/")) {
          // Create the directory
          App.fsExtra.mkdirs(dirPath + "/" + entry.fileName, function (err) {
            if (err)
              return cb(err);
            //
            zipfile.readEntry();
          });
        }
        else {
          zipfile.openReadStream(entry, function (err, readStream) {
            if (err)
              return cb(err);
            //
            // Ensure parent directory exists
            App.fsExtra.mkdirs(App.pathNode.dirname(entry.fileName), function (err) {
              if (err)
                return cb(err);
              //
              // Extract the files
              var output = App.nodeFs.createWriteStream(dirPath + "/" + entry.fileName);
              //
              // Listen to open output event
              output.on("open", function () {
                readStream.pipe(output);
              });
              //
              // Listen to next error output event
              output.once("error", function (err) {
                return cb(err);
              });
              //
              // Listen to close output event
              output.on("close", function (err) {
                zipfile.readEntry();
              });
            });
          });
        }
      });
      //
      // When the parsing is in error
      zipfile.on("error", function (error) {
        cb(error);
      });
      //
      // When the parsing is to end
      zipfile.on("end", function (error) {
        ok = true;
      });
      //
      // When the parsing is over
      zipfile.on("close", function () {
        // Create the parent directory object
        if (ok)
          cb();
      });
    });
  });
};


/**
 * Create the directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.NodeDriver.prototype.mkDir = function (directory, cb)
{
  // Check the validity of the path (writing)
  var path = this.checkPath(true, directory);
  if (!path)
    return cb(null, new Error("Invalid path"));
  //
  // Use fs extra to create even the parent hierarchy
  App.fsExtra.mkdirs(path, cb);
};


/**
 * Checks the existence of the directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.NodeDriver.prototype.dirExists = function (directory, cb)
{
  // Check the validity of the path (reading)
  var path = this.checkPath(false, directory);
  if (!path)
    return cb(false);
  //
  App.nodeFs.exists(path, cb);
};


/**
 * Copies the entire directory
 * @param {Directory} srcDir
 * @param {Directory} dstDir
 * @param {function} cb
 */
App.NodeDriver.prototype.copyDir = function (srcDir, dstDir, cb)
{
  // Check that the relative paths are valid and I get absolute paths
  var srcPath = this.checkPath(false, srcDir);
  var dstPath = this.checkPath(true, dstDir);
  //
  if (!srcPath || !dstPath)
    return cb("Invalid paths");
  //
  // Check if source directory exists
  App.nodeFs.exists(srcPath, function (exists) {
    if (!exists)
      return cb(new Error("Directory " + srcDir.path + " doesn't exist"));
    //
    // Use fs extra to copy the entire directory
    App.fsExtra.copyRecursive(srcPath, dstPath, function (err) {
      if (err)
        return cb(err);
      //
      cb();
    });
  });
};


/**
 * Reads recursively the content of directory
 * @param {Directory} directory
 * @param {Integer} depth
 * @param {function} cb
 */
App.NodeDriver.prototype.readDirectory = function (directory, depth, cb)
{
  // Check the validity of the path (reading)
  var path = this.checkPath(false, directory);
  if (!path)
    return cb(null, new Error("Invalid path"));
  //
  // Set default depth to 0
  depth = depth || 0;
  //
  // Array of files/directory objects
  var content = [];
  //
  // Array of directories yet to be examined
  var dir = [];
  dir.push({depth: 0, path: path});
  //
  // Recursive core
  var readDirRecursive = function (entries, content, cb) {
    // No directory unexplored: end of recursion
    if (!entries.length)
      return cb(content);
    //
    if (entries[0].depth > depth) {
      // Remove the folder because is deeper than depth
      entries.shift();
      //
      // recall the function
      return readDirRecursive(entries, content, cb);
    }
    //
    // Reads the current directory
    App.nodeFs.readdir(entries[0].path, function (err, files) {
      if (err)
        return cb(null, err);
      //
      for (var i = 0; i < files.length; i++) {
        // Add the element to the content array
        content = content.concat(entries[0].path + "/" + files[i]);
        var stats = App.nodeFs.statSync(entries[0].path + "/" + files[i]);
        //
        // if the element is a directory,i add it to the array of folders to be scanned
        if (stats.isDirectory())
          entries.push({depth: entries[0].depth + 1, path: entries[0].path + "/" + files[i]});
      }
      //
      // remove the folder just examined from the array
      entries.shift();
      //
      // recall the function
      readDirRecursive(entries, content, cb);
    });
  };
  //
  readDirRecursive(dir, content, function (files, err) {
    if (err)
      return cb(null, err);
    //
    // Empty directory
    if (files.length < 1)
      return cb();
    //
    files.sort();
    var content = [];
    //
    var index = path.split("/").length - (directory.path === "" ? 0 : directory.path.split("/").length);
    //
    // async for each
    App.async.concat(files, function (files, cb) {
      App.nodeFs.stat(files, function (err, stats) {
        if (err)
          return cb(null, err);
        //
        // adds to array an object (file or folder)
        var relativePath = files.split("/").slice(index);
        relativePath = relativePath.join("/");
        //
        var object;
        if (stats.isFile())
          object = this.file(relativePath, directory.type);
        else
          object = this.directory(relativePath, directory.type);
        //
        content.push(object);
        cb(null, object);
      }.bind(this));
    }.bind(this), function (err, files) {
      cb(files, err);
    });
  }.bind(this));
};


/**
 * Zip directory
 * @param {Directory} directory
 * @param {File} zipFile
 * @param {function} cb
 */
App.NodeDriver.prototype.zipDirectory = function (directory, zipFile, cb)
{
  var path = this.checkPath(false, directory);
  var zipPath = this.checkPath(true, zipFile);
  if (!path || !zipPath)
    return cb(new Error("Invalid paths"));
  //
  var cbCalled = false;
  var done = function (err) {
    if (cbCalled)
      return;
    //
    cbCalled = true;
    cb(err);
  };
  //
  // Check for existence of directory
  App.nodeFs.exists(path, function (exists) {
    if (!exists)
      return cb(new Error("Directory doesn't exist"));
    //
    // Create the write stream
    var output = App.nodeFs.createWriteStream(zipPath);
    //
    // Create the archive object
    var archive = App.archiver("zip");
    //
    // Listen to next error event
    output.once("error", function (err) {
      deleteVoidZip(err);
    });
    //
    // Listen to close finalization archive
    output.on("close", function () {
      done();
    });
    //
    // Listen to error event
    archive.on("error", function (err) {
      deleteVoidZip(err);
    });
    //
    output.on("open", function () {
      // Push data into the archive
      archive.pipe(output);
      //
      // Add to archive the folder to compress
      archive.glob(path + "/**/*");
      archive.finalize();
      //
    });
  });
  //
  // Function that deletes the new zip file (if there is an error)
  var deleteVoidZip = function (err) {
    App.nodeFs.unlink(zipPath, function (err1) {
      done(err || err1);
    });
  };
};


/**
 * Removes the entire directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.NodeDriver.prototype.removeDirRecursive = function (directory, cb)
{
  // Check the validity of the path (writing)
  var path = this.checkPath(true, directory);
  if (!path)
    return cb(new Error("Invalid path"));
  //
  // Use fs extra to remove the directory and its content
  App.fsExtra.rmrf(path, cb);
};


/**
 * Makes an HTTP GET request to a web server
 * @param {object} url
 * @param {string} method
 * @param {object} options
 * @param {function} cb
 */
App.NodeDriver.prototype.httpRequest = function (url, method, options, cb)
{
  options = Object.assign({}, options);
  //
  var uri = url.url;
  //
  // Multipart request
  var multiPart = false;
  //
  // Download
  var download = false;
  //
  // Upload
  var upload = false;
  //
  // Node module variable
  var request = require("request");
  //
  // Create internal request options object
  var opts = {};
  opts.headers = {};
  //
  switch (method) {
    case "POST":
      multiPart = true;
      break;

    case "DOWNLOAD":
      download = true;
      if (options.method === "POST")
        method = "POST";
      else
        method = "GET";
      break;

    case "UPLOAD":
      multiPart = true;
      upload = true;
      method = "POST";
      break;
  }
  //
  // Add an Accept-Encoding header to request compressed content encodings from the server, default true
  opts.gzip = options.hasOwnProperty("gzip") ? options.gzip : true;
  //
  if (method === "POST") {
    // If not specified, the post request type is multipart
    if (options.headers && options.headers["Content-Type"])
      multiPart = false;
  }
  //
  // Get custom headers
  if (options.headers)
    opts.headers = options.headers;
  //
  // Get eventually values for the autentication
  if (options.authentication)
    opts.auth = options.authentication;
  //
  // Set timeout
  if (options.timeOut)
    opts.timeout = options.timeOut;
  //
  opts.method = method;
  opts.uri = uri;
  //
  // Custom body case
  if (options.body) {
    if (typeof options.bodyType === "string")
      opts.headers["Content-Type"] = options.bodyType;
    else if (options.headers && !options.headers["Content-Type"])
      opts.headers["Content-Type"] = "application/octet-stream";
    //
    // Check type of custom body
    if (options.body instanceof ArrayBuffer || typeof options.body === "string")
      opts.body = options.body;
    else
      return cb({error: new Error("Custom body must be string or ArrayBuffer")});
  }
  else if (multiPart) { // multipart
    // Create multipart request for upload
    opts.formData = options.params || {};
    //
    // Delete custom Content-Type
    delete opts.headers["Content-Type"];
    //
    // File section (only fot the upload)
    if (upload) {
      // Check if the path of file is valid
      var path = this.checkPath(false, options._file);
      if (!path)
        return cb({error: new Error("Invalid path")});
      //
      // Add the files
      opts.formData[options._nameField] = {
        value: App.nodeFs.createReadStream(path),
        options: {
          filename: options._fileName,
          contentType: options._fileContentType
        }
      };
    }
  }
  else if (opts.headers["Content-Type"] === "application/x-www-form-urlencoded")
    opts.form = options.params;
  else {
    // GET request
    //
    // Concatenate options params and url params
    var posQuery = uri.indexOf("?");
    if (posQuery > 0) {
      var qs = App.querystring.parse(encodeURI(uri.substr(posQuery + 1)));
      if (options.params) {
        for (var propertyName in options.params)
          qs[propertyName] = options.params[propertyName];
      }
      opts.qs = qs;
    }
    else
      opts.qs = options.params;
    opts.useQuerystring = true;
  }
  //
  // For download check file path
  var downloadError = null;
  if (download) {
    // Check path
    var downloadPath = this.checkPath(true, options._file);
    if (!downloadPath)
      return cb({error: new Error("invalid download path")});
    //
    // Create stream
    var writeStream = App.nodeFs.createWriteStream(downloadPath, {encoding: null});
    writeStream.once("error", function (error) {
      downloadError = error;
    });
  }
  //
  // Make request
  var res = {};
  var req = request(opts, function (error, response, body) {
    // Stop the progress events when the response is complete
    clearInterval(uploadprogressTimer);
    //
    if (error)
      return cb({error: error});
    //
    // Remove bom utf-8
    if (typeof body === "string" && body.charCodeAt(0) === 65279)
      body = body.substring(1);
    //
    res.body = body;
    cb(res);
  });
  //
  // Listen to response event
  req.on("response", function (response) {
    // Listen to next abort event
    req.once("abort", function () {
      cb({error: downloadError || new Error("request aborted")});
    });
    //
    res.status = response.statusCode;
    //
    // Download
    if (download && response.statusCode === 200) {
      if (downloadError)
        req.abort();
      else
        req.pipe(writeStream);
    }
    //
    // Amount of byte downloaded
    var byteDownloaded = 0;
    //
    // Listen to data response event
    response.on('data', function (data) {
      res.headers = response.headers;
      var totalBytes = response.headers["Content-Length"];
      //
      byteDownloaded += data.length;
      if (url.onDownloadProgress(byteDownloaded, totalBytes) === false) {
        // Stop response writing
        req.abort();
      }
    });
  });
  //
  // Last value of byte sent
  var lastByteSent = 0;
  //
  // Upload progress handler
  var uploadprogressTimer = setInterval(function () {
    if (req.req) {
      var byteTotal = req.req._headers["Content-Length"];
      if (upload)
        byteTotal = options._fileSize;
      var byteSent = Math.min(req.req.connection._bytesDispatched, byteTotal);
    }
    else
      clearInterval(uploadprogressTimer);
    //
    // If the information is not avaible, stop the event
    if (byteTotal === undefined)
      return clearInterval(uploadprogressTimer);
    //
    // If the values have not changed, they are not notified
    if (lastByteSent === byteSent)
      return;
    //
    // Check if the upload was interrupted
    var abort = (url.onUploadProgress(byteSent, byteTotal) === false);
    lastByteSent = byteSent;
    //
    // In these cases, stops the event
    if (req._ended || abort)
      clearInterval(uploadprogressTimer);
    //
    if (abort)
      req.abort();
  }, 250);
};


/**
 * Put a local file on the other side file system (local/remote)
 * @param {File} localFile
 * @param {File} dstFile
 * @param {function} cb
 */
App.NodeDriver.prototype.put = function (localFile, dstFile, cb)
{
  if (!dstFile)
    return cb(new Error("Destination file cannot be null"));
  //
  var done = function (error) {
    localFile.close(function (result, remoteError) {
      dstFile.close(function (result, dstError) {
        cb(error || remoteError || dstError);
      });
    });
  };
  //
  var readChunk = function (length, offset, create, callback) {
    // Read chunk
    localFile.read(length, offset, function (res, err) {
      if (err)
        return callback(err);
      //
      var result = {};
      result.chunk = res;
      result.create = create;
      //
      // Write current chunk
      writeChunk(result, function (writeError) {
        if (writeError)
          return callback(writeError);
        //
        // Check if I need to read the next chunk
        if (!res || res.byteLength !== length)
          return callback();
        else // Read next chunk
          readChunk(length, offset + length, false, callback);
      });
    });
  };
  //
  var writeChunk = function (content, callback) {
    // Check if dstFile needs to be created
    if (content.create) {
      dstFile.create(undefined, function (createRes, createErr) {
        if (createErr)
          return callback(createErr);
        //
        // Write chunk into dstFile
        dstFile.write(content.chunk, undefined, undefined, undefined, function (writeRes, writeErr) {
          callback(writeErr);
        });
      });
    }
    else {
      // Write chunk into dstFile
      dstFile.write(content.chunk, undefined, undefined, undefined, function (writeRes, writeErr) {
        callback(writeErr);
      });
    }
  };
  //
  // Open remote file
  localFile.open(function (res, err) {
    if (err)
      return done(err);
    //
    // Start reading and writing chunk by chunk (256 KB is size of a chunk)
    readChunk(1024 * 256, 0, true, done);
  }.bind(this));
};


// export module for node
if (module)
  module.exports = App.NodeDriver;
