/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, process, ArrayBuffer, Buffer */

var App = App || {};

if (module)
  App.FS = require("./fs");

/**
 * @class RemoteFS
 * Represents a remote file system object interface
 * @param {Object} app
 * @param {String} root
 * @param {String} APIKey
 *
 */
App.RemoteFS = function (app, root, APIKey)
{
  if (!module)
    throw new Error("Cannot create App.RemoteFS instance in offline mode");
  //
  // Import modules
  if (module && !App.View)
    App = require("../requireAll");
  //
  this.id = "remotefs-" + App.Utils.generateUID24();
  //
  this.app = app;
  this.root = root;
  this.APIKey = APIKey;
  App.RemoteFS.objectsMap[this.id] = this;
};

// Make App.RemoteFS extend App.FS
App.RemoteFS.prototype = new App.FS(true);

App.RemoteFS.callbacks = {};
App.RemoteFS.objectsMap = {};

/*
 * Send message to remote file system
 * @param {String} cmd
 * @param {Object} args
 */
App.RemoteFS.prototype.sendMessage = function (cmd, args)
{
  // Compose the message
  var msg = {};
  var m = {};
  m.type = "ccm";
  m.cnt = {};
  m.cnt.conn = this.root.split("://")[0];
  m.cnt.type = "remoteCmd";
  m.cnt.data = msg;
  m.cnt.key = this.APIKey;
  msg.cmd = cmd;
  msg.cid = this.id;
  msg.fs = this.root.split("://")[1];
  //
  var argsArray = [];
  for (var i = 0; i < args.length - 1; i++) {
    var arg = args[i];
    if (arg instanceof App.File)
      arg = {_t: "file", path: arg.path, id: arg.id};
    else if (arg instanceof App.Directory)
      arg = {_t: "directory", path: arg.path};
    else if (arg instanceof ArrayBuffer)
      arg = {_t: "buffer", data: App.Utils.bufferToBase64(arg)};
    argsArray.push(arg);
  }
  //
  // Last argument is always the callback
  // Store the callback and send only a callbackID
  var cbid = App.Utils.generateUID24();
  msg.cbid = cbid;
  App.RemoteFS.callbacks[cbid] = {cb: args[args.length - 1], ccName: m.cnt.conn};
  //
  // Dispatch to Node.Worker (master) or to InDe.AApp (IDE)
  msg.args = argsArray;
  //
  process.send(m);
};


/*
 * Given a relative path, return the public url to the file.
 * It has file:// protocol, hence inside indeshell it won't work.
 * @param {File} path
 * @param {function} cb
 */
App.RemoteFS.prototype.getFilePublicUrl = function (file, cb)
{
  cb();
};


/**
 * Creates the file physically (Opens the file and overwrites it)
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.createFile = function (file, cb)
{
  file.id = App.Utils.generateUID24();
  this.sendMessage("createFile", arguments);
};


/**
 * Opens the file to append data
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.openFileForAppend = function (file, cb)
{
  file.id = App.Utils.generateUID24();
  this.sendMessage("openFileForAppend", arguments);
};


/**
 * Opens the file for reading
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.openFile = function (file, cb)
{
  file.id = App.Utils.generateUID24();
  this.sendMessage("openFile", arguments);
};


/**
 * Closes the file
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.close = function (file, cb)
{
  this.sendMessage("close", arguments);
};


/**
 * Checks existence of a file
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.fileExists = function (file, cb)
{
  this.sendMessage("fileExists", arguments);
};


/**
 * Reads a text file line by line through a loop that reads each line of the file
 * @param {File} file
 * @param {function} cb
 * @param {function} cbEnd
 */
App.RemoteFS.prototype.readLine = function (file, cb, cbEnd)
{
  file.lineStreamPause = false;
  var currentLine = new Buffer([]);
  //
  var done = function (error) {
    file.close(function (result, closeError) {
      cbEnd(error || closeError);
    });
  };
  //
  var getLines = function (chunk, callback) {
    var lines = [];
    var lastNewLine = -1;
    //
    // Create a buffer from chunk
    var buf = new Buffer(new Uint8Array(chunk));
    for (var i = 0; i < buf.length; i++) {
      // Look for lines inside buffer
      if (buf[i] === "\n".charCodeAt(0)) {
        var line;
        //
        // If there is a not empty pending buffer concatenate it with the buffer just created before reset it
        if (currentLine.length > 0) {
          line = Buffer.concat([currentLine, buf.slice(lastNewLine + 1, i)]);
          currentLine = new Buffer([]);
        }
        else // Otherwise extract line from buffer
          line = buf.slice(lastNewLine + 1, i);
        //
        // Push string representation of the line
        lines.push(line.toString(file.encoding || "utf-8"));
        //
        // Set new line index
        lastNewLine = i;
      }
    }
    //
    // If there is a pending line, store it inside currentLine buffer
    if (lastNewLine < buf.length - 1)
      currentLine = Buffer.concat([currentLine, buf.slice(lastNewLine + 1, buf.length)]);
    //
    callback(lines);
  };
  //
  var readChunk = function (length, offset, callback) {
    file.read(length, offset, function (res, err) {
      if (err)
        return callback(err);
      //
      // No result
      if (!res) {
        // Check if there is a pending line before exit
        if (currentLine.length > 0)
          cb(currentLine.toString(file.encoding || "utf-8"));
        //
        return callback();
      }
      //
      // Get lines from result
      getLines(res, function (lines) {
        // Return each line
        for (var i = 0; i < lines.length; i++) {
          // If break method was called, exit
          if (file.lineStreamPause)
            return callback();
          //
          cb(lines[i]);
        }
        //
        // Check if I need to read the next chunk
        if (res.byteLength !== length) {
          // Check if there is a pending line before exit
          if (currentLine.length > 0)
            cb(currentLine.toString(file.encoding || "utf-8"));
          //
          return callback();
        }
        else // Read next chunk
          readChunk(length, offset + length, callback);
      });
    });
  };
  //
  // Start reading chunk by chunk (256 KB is size of a chunk)
  readChunk(1024 * 256, 0, done);
};


/**
 * Stops reading line by line
 * @param {File} file
 */
App.RemoteFS.prototype.break = function (file)
{
  // Sets the variable flag to stop the reading line by line
  file.lineStreamPause = true;
};


/**
 * Read the whole file as text
 * @param {File} file
 */
App.RemoteFS.prototype.readSync = function (file)
{
  throw new Error("Method not supported in remote file system");
};


/**
 * Reads a block of data, return an array buffer
 * @param {File} file
 * @param {int} length
 * @param {int} offset
 * @param {function} cb
 */
App.RemoteFS.prototype.read = function (file, length, offset, cb)
{
  this.sendMessage("read", arguments);
};


/**
 * Read the whole file as text
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.readAll = function (file, cb)
{
  this.sendMessage("readAll", arguments);
};


/**
 * Writes the data or the  string given
 * @param {File} file
 * @param {string/buffer} data
 * @param {int} offset
 * @param {int} size
 * @param {int} position
 * @param {function} cb
 */
App.RemoteFS.prototype.write = function (file, data, offset, size, position, cb)
{
  this.sendMessage("write", arguments);
};


/**
 * Copy the file
 * @param {File} file
 * @param {File} newFile
 * @param {function} cb
 */
App.RemoteFS.prototype.copyFile = function (file, newFile, cb)
{
  this.sendMessage("copyFile", arguments);
};


/**
 * Rename a file or directory
 * @param {File/Directory} obj
 * @param {string} newName
 * @param {function} cb
 */
App.RemoteFS.prototype.renameObject = function (obj, newName, cb)
{
  this.sendMessage("renameObject", arguments);
};


/**
 * Return the file size (in bytes)
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.fileLength = function (file, cb)
{
  this.sendMessage("fileLength", arguments);
};


/**
 * Return the last modified file date
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.fileDateTime = function (file, cb)
{
  var newCb = function (result, error) {
    if (result)
      cb(new Date(result));
    else
      cb(null, error);
  };
  this.sendMessage("fileDateTime", [file, newCb]);
};


/**
 * Deletes a file
 * @param {File} file
 * @param {function} cb
 */
App.RemoteFS.prototype.deleteFile = function (file, cb)
{
  this.sendMessage("deleteFile", arguments);
};


/**
 * Zip a file
 * @param {File} file
 * @param {File} zipFile
 * @param {function} cb
 */
App.RemoteFS.prototype.zipFile = function (file, zipFile, cb)
{
  this.sendMessage("zipFile", arguments);
};


/**
 * Unzip an archiver
 * @param {File} file
 * @param {Directory} directory
 * @param {function} cb
 */
App.RemoteFS.prototype.unzip = function (file, directory, cb)
{
  this.sendMessage("unzip", arguments);
};


/**
 * Create the directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.RemoteFS.prototype.mkDir = function (directory, cb)
{
  this.sendMessage("mkDir", arguments);
};


/**
 * Checks the existence of the directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.RemoteFS.prototype.dirExists = function (directory, cb)
{
  this.sendMessage("dirExists", arguments);
};


/**
 * Copies the entire directory
 * @param {Directory} srcDir
 * @param {Directory} dstDir
 * @param {function} cb
 */
App.RemoteFS.prototype.copyDir = function (srcDir, dstDir, cb)
{
  this.sendMessage("copyDir", arguments);
};


/**
 * Read recursively the content of directory
 * @param {Directory} directory
 * @param {Integer} depth
 * @param {function} cb
 */
App.RemoteFS.prototype.readDirectory = function (directory, depth, cb)
{
  var newCB = function (res, err) {
    if (err || !res)
      return cb(null, err);
    //
    // res.path is an array of {path: relativepth, type: "file"|"directory"}
    var entries = [];
    for (var i = 0; i < res.length; i++)
      entries.push(this[res[i].type](res[i].path));
    cb(entries);
  }.bind(this);
  this.sendMessage("readDirectory", [directory, depth, newCB]);
};


/**
 * Zip directory
 * @param {Directory} directory
 * @param {File} zipFile
 * @param {function} cb
 */
App.RemoteFS.prototype.zipDirectory = function (directory, zipFile, cb)
{
  this.sendMessage("zipDirectory", arguments);
};


/**
 * Removes the entire directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.RemoteFS.prototype.removeDirRecursive = function (directory, cb)
{
  this.sendMessage("removeDirRecursive", arguments);
};


/**
 * Put a remote file on the other side file system (local/remote)
 * @param {File} remoteFile
 * @param {File} dstFile
 * @param {function} cb
 */
App.RemoteFS.prototype.put = function (remoteFile, dstFile, cb)
{
  if (!dstFile)
    return cb(new Error("Destination file cannot be null"));
  //
  var done = function (error) {
    remoteFile.close(function (result, remoteError) {
      dstFile.close(function (result, dstError) {
        cb(error || remoteError || dstError);
      });
    });
  };
  //
  var readChunk = function (length, offset, create, callback) {
    // Read chunk
    remoteFile.read(length, offset, function (res, err) {
      if (err)
        return callback(err);
      //
      var result = {};
      result.chunk = res;
      result.create = create;
      //
      // Write current chunk
      writeChunk(result, function (writeError) {
        if (writeError)
          return callback(writeError);
        //
        // Check if I need to read the next chunk
        if (!res || res.byteLength !== length)
          return callback();
        else // Read next chunk
          readChunk(length, offset + length, false, callback);
      });
    });
  };
  //
  var writeChunk = function (content, callback) {
    // Check if dstFile needs to be created
    if (content.create) {
      dstFile.create(undefined, function (createRes, createErr) {
        if (createErr)
          return callback(createErr);
        //
        // Write chunk into dstFile
        dstFile.write(content.chunk, undefined, undefined, undefined, function (writeRes, writeErr) {
          callback(writeErr);
        });
      });
    }
    else {
      // Write chunk into dstFile
      dstFile.write(content.chunk, undefined, undefined, undefined, function (writeRes, writeErr) {
        callback(writeErr);
      });
    }
  };
  //
  // Open remote file
  remoteFile.open(function (res, err) {
    if (err)
      return done(err);
    //
    // Start reading and writing chunk by chunk (256 KB is size of a chunk)
    readChunk(1024 * 256, 0, true, done);
  }.bind(this));
};


/**
 * Makes an HTTP GET request to a web server
 * @param {object} url
 * @param {string} method
 * @param {object} options
 * @param {function} cb
 */
App.RemoteFS.prototype.httpRequest = function (url, method, options, cb)
{
  throw new Error("Method not supported in remote file system");
};


/**
 * Handle response of a remote command
 * @param {Object} msg
 */
App.RemoteFS.handleResponse = function (msg)
{
  var cb = App.RemoteFS.callbacks[msg.cbid];
  if (cb) {
    delete App.RemoteFS.callbacks[msg.cbid];
    var args = [];
    if (msg.data.result !== undefined) {
      if (msg.data.result && typeof msg.data.result === "object" && msg.data.result._t === "buffer" && msg.data.result.data)
        args.push(App.Utils.base64ToBuffer(msg.data.result.data));
      else
        args.push(msg.data.result);
    }
    if (msg.data.error) {
      args.push(new Error(msg.data.error));
      //
      // Since error could be both in first or second place and callbacks always have two parameters,
      // if I have no results I add the error twice to fill both the parameters.
      // This happens when error occurs in user.js when there are no cloud connectors connected
      if (args.length === 1)
        args.push(new Error(msg.data.error));
    }
    //
    cb.cb.apply(this, args);
  }
};


/**
 * Handle disconnect of a remote connector
 * @param {Object} msg
 */
App.RemoteFS.handleDisconnect = function (msg)
{
  // Invoke all callback awaiting with error
  var cbids = Object.keys(App.RemoteFS.callbacks);
  for (var i = 0; i < cbids.length; i++) {
    var cbid = cbids[i];
    var cb = App.RemoteFS.callbacks[cbid];
    if (cb.ccName === msg.data.name) {
      delete App.RemoteFS.callbacks[cbid];
      setImmediate(function () {
        cb.cb(null, new Error("Remote connector disconnected"));
      });
    }
  }
};


// export module for node
if (module)
  module.exports = App.RemoteFS;