/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */


/* global module */

var App = App || {};

//
// import modules
if (module && !App.View) {
  App = require("../requireAll");
  App.nodeFs = require("fs");
  App.crypto = require("crypto");
}


/**
 * @class SecretKeeper
 * Encrypts using AES256 and RSA
 * Used to communicate with the buildserver
 */
App.SecretKeeper = {};


/*
 * Encrypt "encoding" plaintext into hex ciphertext using sring key
 * key: aes256 key. It must be a binary encoded string or a buffer object
 * plaintext: string
 * encoding: encoding for plaintext
 */
App.SecretKeeper.encryptAES = function (app, key, plaintext, encoding, cb) {
  if (!encoding)
    encoding = "utf8";
  //
  var cipher = App.crypto.createCipher("aes256", key);
  cb(Buffer.concat([cipher.update(plaintext, encoding), cipher.final()]).toString("hex"));
};


/*
 * Decrypt hex ciphertext into "encoding" plaintext using string key
 * key: aes256 key. It must be a binary encoded string or a buffer object
 * ciphertext: hex encoded string
 * encoding: encoding for resulting plaintext
 */
App.SecretKeeper.decryptAES = function (app, key, ciphertext, encoding, cb) {
  if (!encoding)
    encoding = "utf8";
  //
  var decipher = App.crypto.createDecipher("aes256", key);
  cb(Buffer.concat([decipher.update(ciphertext, "hex"), decipher.final()]).toString(encoding));
};


/*
 * keyPem: path to the secret rsa key generated by `openssl genrsa -out key.pem 2048`
 * secret: object
 * secret.symkey: aes256 key encoded as base64 using rsa
 * secret.ciphertext: aes256 encoded hex string ciphertext
 * secret.encoding: string encoding for ciphertext
 */
App.SecretKeeper.decryptSecret = function (app, keyPem, secret, cb) {
  // PLEASE NOTE: we use a CRYPTOGTAPHICALLY SECURE PRNG for generating our AES256 password
  // Hence we can simply crypto.createCipher.
  // Were we to prefere a password based system, node documentation strongly suggests to use
  // createCipheriv, with output from pbkdf2.
  // In such case a good article to read is http://www.boronine.com/2012/08/30/Strong-Password-Hashing-with-Node-Standard-Library/
  var symkey = App.crypto.privateDecrypt({key: App.nodeFs.readFileSync(keyPem)}, new Buffer(secret.symkey, "base64"));
  this.decryptAES(app, symkey.toString("binary"), secret.ciphertext, secret.encoding, function (plaintext) {
    cb(plaintext);
  });
};


/*
 * pubkey: path to the public rsa key exported from the pem by `openssl rsa -in key.pem -pubout -out my-server.pub`
 * symkey: aes256 symmetric key, encoded as utf8
 * plaintext: string to hide
 * encoding: encoding for plaintext
 */
App.SecretKeeper.encryptSecret = function (app, pubkey, plaintext, encoding, cb) {
  //
  // generate a secure AES key
  var rand = App.crypto.randomBytes(256); // read the security note in this.encryptAES
  var symkey = rand.toString("binary"); // the symkey must be a binary encoded string or a buffer object
  //
  var secret = {};
  secret.symkey = App.crypto.publicEncrypt(App.nodeFs.readFileSync(pubkey), rand); // rsa encoded aes256 key
  secret.symkey = secret.symkey.toString("base64"); // crypto.publicEncrypt returns a buffer
  secret.encoding = encoding;
  this.encryptAES(app, symkey, plaintext, secret.encoding, function (ciphertext) {
    secret.ciphertext = ciphertext;
    cb(secret);
  });
};


//  export module for node
if (module)
  module.exports = App.SecretKeeper;
