/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */


var App = App || {};


/**
 * @class ShellDriver
 * Represents a shell driver object,that will handle files and folders
 * @param {Object} options
 */
App.ShellDriver = function (options)
{
  this.id = "device-shelldriverhandler";
  this.nid = App.Utils.generateUID24();
  //
  this.cbId = 0; // id of added callback
  this.callbacks = {}; // list of callbacks
  this.app = options.app;
  this.app.eleMap[this.id + this.nid] = this;
  this.name = "shelldriver";
};

// Make App.ShellDriver extend App.FS
App.ShellDriver.prototype = new App.FS(true);


/*
 * Send a command to shell
 * @param {type} cid
 * @param {type} args
 * @param {type} cb
 */
App.ShellDriver.prototype.exec = function (cid, args)
{
  var argsArray = [];
  for (var i = 0; i < args.length - 1; i++) {
    var arg = args[i];
    if (arg instanceof App.File)
      arg = {_t: "File", path: arg.path, type: arg.type, id: arg.id};
    else if (arg instanceof App.Directory)
      arg = {_t: "Directory", path: arg.path, type: arg.type};
    argsArray.push(arg);
  }
  //
  // Last argument is always the callback
  // Store the callback and send only a callbackID
  var cbId = ++this.cbId;
  argsArray.push(cbId);
  this.callbacks[cbId] = args[args.length - 1];
  //
  // Send command
  parent.postMessage({obj: this.id, id: "exec", cnt: {nid: this.nid, cid: cid, args: argsArray}}, "*");
};


/*
 * Generic callback for shelldriver. Works like the callbacks in App.Device
 * @param {object} res = {cbID, arg, err}
 */
App.ShellDriver.prototype.CB = function (result)
{
  var res = result.result;
  var cb = this.callbacks[res.cbId];
  delete this.callbacks[res.cbId];
  //
  // Recreate error
  if (res.args[1])
    res.args[1] = new Error(res.args[1]);
  //
  cb.apply(cb, res.args);
};


/*
 * Given a relative path, return the public url to the file.
 * It has file:// protocol, hence inside indeshell it won't work.
 * @param {File} path
 * @param {function} cb
 */
App.ShellDriver.prototype.getFilePublicUrl = function (file, cb)
{
  this.exec("getFilePublicUrl", arguments);
};


/**
 * Creates the file physically (Opens the file and overwrites it)
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.createFile = function (file, cb)
{
  file.id = App.Utils.generateUID24();
  this.exec("createFile", arguments);
};


/**
 * Opens the file to append data
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.openFileForAppend = function (file, cb)
{
  file.id = App.Utils.generateUID24();
  this.exec("openFileForAppend", arguments);
};


/**
 * Opens the file for reading
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.openFile = function (file, cb)
{
  file.id = App.Utils.generateUID24();
  this.exec("openFile", arguments);
};


/**
 * Closes the file
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.close = function (file, cb)
{
  this.exec("close", arguments);
};


/**
 * Checks existence of a file
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.fileExists = function (file, cb)
{
  this.exec("fileExists", arguments);
};


/**
 * Reads a text file line by line through a loop that reads each line of the file
 * @param {File} file
 * @param {function} cb
 * @param {function} cbEnd
 */
App.ShellDriver.prototype.readLine = function (file, cb, cbEnd)
{
  cbEnd(new Error("Method not supported inside shell"));
};


/**
 * Stops reading line by line
 * @param {File} file
 */
App.LocalDriver.prototype.break = function (file)
{
  throw new Error("Method not supported inside shell");
};


/**
 * Reads a block of data, return a array buffer
 * @param {File} file
 * @param {int} length
 * @param {int} offset
 * @param {function} cb
 */
App.ShellDriver.prototype.read = function (file, length, offset, cb)
{
  this.exec("read", arguments);
};


/**
 * Read the whole file as text
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.readAll = function (file, cb)
{
  this.exec("readAll", arguments);
};


/**
 * Writes the data or the  string given
 * @param {File} file
 * @param {string/buffer} data
 * @param {int} offset
 * @param {int} size
 * @param {int} position
 * @param {function} cb
 */
App.ShellDriver.prototype.write = function (file, data, offset, size, position, cb)
{
  this.exec("write", arguments);
};


/**
 * Copy the file
 * @param {File} file
 * @param {File} newFile
 * @param {function} cb
 */
App.ShellDriver.prototype.copyFile = function (file, newFile, cb)
{
  this.exec("copyFile", arguments);
};


/**
 * Rename a file or directory
 * @param {File/Directory} obj
 * @param {string} newName
 * @param {function} cb
 */
App.ShellDriver.prototype.renameObject = function (obj, newName, cb)
{
  this.exec("renameObject", arguments);
};


/**
 * Return the file size (in bytes)
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.fileLength = function (file, cb)
{
  this.exec("fileLength", arguments);
};


/**
 * Return the last modified file date
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.fileDateTime = function (file, cb)
{
  this.exec("fileDateTime", arguments);
};


/**
 * Deletes a file
 * @param {File} file
 * @param {function} cb
 */
App.ShellDriver.prototype.deleteFile = function (file, cb)
{
  this.exec("deleteFile", arguments);
};


/**
 * Zip a file
 * @param {File} file
 * @param {File} zipFile
 * @param {function} cb
 */
App.ShellDriver.prototype.zipFile = function (file, zipFile, cb)
{
  cb(new Error("Method not supported inside shell"));
};


/**
 * Unzip an archiver
 * @param {File} file
 * @param {Directory} directory
 * @param {function} cb
 */
App.ShellDriver.prototype.unzip = function (file, directory, cb)
{
  this.exec("unzip", arguments);
};


/**
 * Create the directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.ShellDriver.prototype.mkDir = function (directory, cb)
{
  this.exec("mkDir", arguments);
};


/**
 * Checks the existence of the directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.ShellDriver.prototype.dirExists = function (directory, cb)
{
  this.exec("dirExists", arguments);
};


/**
 * Copies the entire directory
 * @param {Directory} srcDir
 * @param {Directory} dstDir
 * @param {function} cb
 */
App.ShellDriver.prototype.copyDir = function (srcDir, dstDir, cb)
{
  this.exec("copyDir", arguments);
};


/**
 * Read recursively the content of directory
 * @param {Directory} directory
 * @param {Integer} depth
 * @param {function} cb
 */
App.ShellDriver.prototype.readDirectory = function (directory, depth, cb)
{
  var newCB = function (res, err) {
    if (err || !res)
      return cb(null, err);
    //
    // res.path is an array of {path: relativepth, type: "file"|"directory"}
    var entries = [];
    for (var i = 0; i < res.length; i++)
      entries.push(this[res[i].type](res[i].path, directory.type));
    cb(entries);
  }.bind(this);
  this.exec("readDirectory", [directory, depth, newCB]);
};


/**
 * Zip directory
 * @param {Directory} directory
 * @param {File} zipFile
 * @param {function} cb
 */
App.ShellDriver.prototype.zipDirectory = function (directory, zipFile, cb)
{
  cb(new Error("Method not supported inside shell"));
};


/**
 * Removes the entire directory
 * @param {Directory} directory
 * @param {function} cb
 */
App.ShellDriver.prototype.removeDirRecursive = function (directory, cb)
{
  this.exec("removeDirRecursive", arguments);
};


/**
 * Makes an HTTP GET request to a web server
 * @param {object} url
 * @param {string} method
 * @param {object} options
 * @param {function} cb
 */
App.ShellDriver.prototype.httpRequest = function (url, method, options, cb)
{
  var argsArray = Array.prototype.slice.call(arguments);
  argsArray[0] = {url: url.url};
  if (options && options._file) {
    var file = options._file;
    options._file = {path: file.path, type: file.type};
  }
  this.exec("httpRequest", argsArray);
};


/**
 * Put a local file on the other side file system (local/remote)
 * @param {File} localFile
 * @param {File} dstFile
 * @param {function} cb
 */
App.ShellDriver.prototype.put = function (localFile, dstFile, cb)
{
  cb(new Error("Method not supported inside shell"));
};