/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */


/* global module */


var App = App || {};


/**
 * @class GAC
 * Represents an GAC Object
 * @param {Object} app
 * @param {Object} options
 */
App.GAC = function (app, options)
{
  options = options || {};
  //
  // import modules
  if (module && !App.View) {
    App = require("./requireAll");
  }
  // sets module variable
  this.google = require("googleapis");
  this.googleAuth = require("google-auth-library");
  //
  this._redirectURL = options.redirectUrl;
  //
  // if wasn't passed calculate the redirect url
  if (!this._redirectURL) {
    if (App.parentType === "ide")
      this._redirectURL = App.serverUrl + "/" + app.id + "/" + encodeURIComponent(App.id.replace(/\//g, "-")) + "/run?mode=rest&cmd=gauth";
    else
      this._redirectURL = App.serverUrl + "/" + App.name + "?mode=rest&cmd=gauth";
  }
  //
  // sets client id, the client secret and the Token
  this._clientId = options.clientId;
  this._clientSecret = options.clientSecret;
  this._token = options.token;
};


//
// class properties
Object.defineProperties(App.GAC.prototype, {
  clientId: {
    get: function () {
      return this._clientId;
    },
    set: function (newValue) {
      this.oauth2Client = null;
      this._clientId = newValue;
    }
  },
  clientSecret: {
    get: function () {
      return this._clientSecret;
    },
    set: function (newValue) {
      this.oauth2Client = null;
      this._clientSecret = newValue;
    }
  },
  token: {
    get: function () {
      return this._token;
    },
    set: function (newValue) {
      this.oauth2Client = null;
      this._token = newValue;
    }
  },
  redirectURL: {
    get: function () {
      return this._redirectURL;
    },
    set: function (newValue) {
      this.oauth2Client = null;
      this._redirectURL = newValue;
    }
  }
});


/*
 * Generates the url for the first authentication prompt
 * @param {Object} options
 *  -scopes {String}: scopes of the script, optional
 *  -redirectURL {String}: google OAuth redirect URL
 *  -access_type {String}: specific execution option of google oauth 2  authentication api
 *  -state {String}: optional parameter that will be added to the redirection url
 */
App.GAC.prototype.generateAuthURL = function (options) {
  // protects from null options
  options = options || {};
  //
  // sets the scopes for the script
  var scopes = ["https://www.googleapis.com/auth/documents", "https://www.googleapis.com/auth/drive", "https://www.googleapis.com/auth/script.external_request"];
  var auth = new this.googleAuth();
  //
  if (!this.oauth2Client) {
    // if i don't have all the properties to request url i will return an error
    if (!this._clientId || !this._clientSecret)
      throw new Error("Missing  parameters for generating the url");
    //
    this.oauth2Client = new auth.OAuth2(this._clientId, this._clientSecret, options.redirectURL || this._redirectURL);
  }
  try {
    // generates and return the auth url
    return this.oauth2Client.generateAuthUrl({access_type: options.access_type || "offline", scope: options.scopes || scopes, state: options.state, approval_prompt: "force"});
  }
  catch (ex) {
    throw new Error(ex);
  }
};


/*
 * Generates the token for authentication
 * @param {String} code
 */
App.GAC.prototype.generateToken = function (code, cb) {
  if (!this.oauth2Client) {
    // if i don't have all the properties to request token i will return an error
    if (!this._clientId || !this._clientSecret)
      return cb(null, "Missing  parameters for generating the url");
    //
    var auth = new this.googleAuth();
    this.oauth2Client = new auth.OAuth2(this._clientId, this._clientSecret, this._redirectURL);
  }
  //
  // gets the token via the code obtained from the first authentication
  this.oauth2Client.getToken(code, function (err, token) {
    if (err)
      return cb(false, "Error while trying to retrieve access token: " + err);
    //
    // set credentials
    this._token = token;
    this.oauth2Client.credentials = token;
    return cb(token);
  }.bind(this));
};


/**
 * execute the script
 * @param {Object} options
 *  -load {String}: name of the template document
 *  -save {String}: path,with name, where save the generated document
 *  -set {Object}: object containing the replace values
 *  -devMode {Boolean}: specific execution option of google api script
 *  -export: {App.File} if is specific, this file will be write with the content of new document
 *  -format: {App.String} specifies the document format when writing to files, by default the script creates a PDF
 * @param {function} cb
 */
App.GAC.prototype.exec = function (options, cb)
{
  // protects from null options
  options = options || {};
  //
  // if i don't have all the properties to create the new dociment i will return an error
  if (!options.load || !options.save)
    return cb(null, "Missing parameters for the creation of document");
  if (!this.oauth2Client) {
    // if i don't have all the properties to request permission i will return an error
    if (!this._token || !this._clientId || !this._clientSecret)
      return cb(null, "Missing parameters for authentication");
    //
    var auth = new this.googleAuth();
    this.oauth2Client = new auth.OAuth2(this._clientId, this._clientSecret, this._redirectUrl);
    this.oauth2Client.credentials = this._token;
  }
  var script = this.google.script("v1");
  //
  // make the API request
  script.scripts.run({
    auth: this.oauth2Client,
    resource: {
      "devMode": options.devMode ? true : false,
      function: "exec",
      "parameters": [
        {templateName: options.load, filePath: options.save, set: options.set}
      ]
    },
    scriptId: options.scriptId
  }, function (err, resp) {
    if (err) {
      // the API encountered a problem before the script started executing.
      return cb(null, "The API returned an error: " + err);
    }
    if (resp.error) {
      // the API executed, but the script returned an error.
      //
      // extract the first (and only) set of error details. The values of this
      // object are the script's 'errorMessage' and 'errorType', and an array
      // of stack trace elements.
      var error = resp.error.details && resp.error.details[0];
      //
      if (!error)
        return cb(null, "Error while script execution");
      var stringError = "Script error message: " + error.errorMessage;
      //
      if (error.scriptStackTraceElements) {
        stringError += ("Script error stacktrace: ");
        //
        // There may not be a stacktrace if the script didn't start executing.
        for (var i = 0; i < error.scriptStackTraceElements.length; i++) {
          var trace = error.scriptStackTraceElements[i];
          stringError += trace.function + ":" + trace.lineNumber + " ";
        }
      }
      return cb(null, stringError);
    }
    else {
      var newDocID = resp.response && resp.response.result && resp.response.result.id;
      if (!newDocID)
        return cb(null, "Error while script execution");
      //
      // if the developer has passed a file, i write this with the content of generated google document
      if (options.export) {
        var drive = this.google.drive("v3");
        //
        // get real path of file and check for his validity
        var filePath = options.export.fs.checkPath(true, options.export);
        //
        if (filePath) {
          var stream = App.nodeFs.createWriteStream(filePath);
          //
          // call the API for the export
          drive.files.export({
            auth: this.oauth2Client,
            fileId: newDocID,
            mimeType: options.format || "application/pdf"
          }).on('end', function () {
            return cb(newDocID);
          }).on('error', function (err) {
            return cb(null, "Error during download: " + err);
          }).pipe(stream);
        }
        else
          return cb(null, "Invalid path");
      }
      else
        return cb(newDocID);
    }
  }.bind(this));
};


/*
 * Download document from drive
 * @param {App.File} file
 * @param {String}  docID
 * @param {String}  format
 * @param {Function}  cb
 */
App.GAC.prototype.downloadDOC = function (file, docID, format, cb) {
  var drive = this.google.drive("v3");
  //
  if (!this.oauth2Client) {
    // if i don't have all the properties to request permission i will return an error
    if (!this._token || !this._clientId || !this._clientSecret)
      return cb(null, "Missing parameters for authentication");
    //
    var auth = new this.googleAuth();
    this.oauth2Client = new auth.OAuth2(this._clientId, this._clientSecret, this._redirectUrl);
    this.oauth2Client.credentials = this._token;
  }
  //
  // get real path of file and check for his validity
  var filePath = file.fs.checkPath(true, file);
  //
  if (filePath) {
    var stream = App.nodeFs.createWriteStream(filePath);
    //
    // call the API for the export
    drive.files.export({
      auth: this.oauth2Client,
      fileId: docID,
      mimeType: format || "application/pdf"
    }).on('end', function () {
      return cb();
    }).on('error', function (err) {
      return cb(null, "Error during download: " + err);
    }).pipe(stream);
  }
  else
    return cb(null, "Invalid path");
};


//  export module for node
if (module)
  module.exports = App.GAC;