/*
 * Instant Developer Cloud
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */


var App = App || {};


/**
 * @class GM
 * Represents a image object that
 * @param {Object} app
 * @param {Object} source
 */
App.GM = function (app, source)
{
  // Import modules
  if (module && !App.View)
    App = require("./requireAll");
  //
  //license and detail: https://github.com/aheckmann/gm
  App.gm = require('gm');
  //
  // the source is a file
  if (source instanceof App.File) {
    // save the file in a property for further validation of the file path
    this._file = source;
    //
    // calculate the real path
    this.path = this._file.fs.checkPath(true, this._file);
    //
    // creation of "gm" object
    this.gm = App.gm(this.path);
  }
  else if (source instanceof String || typeof (source) === 'string') { // if the source is a string, i create the instance directly from she
    // save the file in a property for further validation of the file path
    this._file = app.fs.file(source);
    //
    // calculate the real path
    this.path = this._file.fs.checkPath(true, this._file);
    //
    // creation of "gm" object
    this.gm = App.gm(this.path);
  }
  else {
    this._file = app.fs.file();
    //
    // creation of "gm" object for a new image
    this.gm = App.gm(source.width, source.height, source.color);
  }
};


/* ASYNC methods*/

/**
 * Write image to file. If receive a path, change the path of the instance
 * @param {string} path
 * @param {function} cb
 */
App.GM.prototype.save = function (path, cb)
{
  this._file.path = path || this._file.path;
  //
  // calculates the real path
  this.path = this._file.fs.checkPath(true, this._file);
  //
  // if the path is invalid, returns an error
  if (!this.path)
    return cb(null, new Error("Invalid path"));
  //
  this.gm.write(this.path, function (err) {
    cb(null, err);
  });
};


/**
 * Write image to buffer
 * @param {string} format
 * @param {function} cb
 */
App.GM.prototype.getBuffer = function (format, cb)
{
  // create the buffer from the image
  this.gm.toBuffer(format, function (err, buffer) {
    cb(buffer, err);
  });
};



/**
 * @param {App.File} file1
 * @param {App.File} file2
 * @param {Object} options
 * @param {function} cb
 */
App.GM.prototype.compare = function (file1, file2, options, cb)
{
  // protection  against null options
  options = options || {};
  //
  // checks and calculates the real path
  var realPath1 = file1.fs.checkPath(false, file1);
  var realPath2 = file1.fs.checkPath(false, file2);
  var realPath3 = options.file ? options.file.fs.checkPath(true, options.file) : true;
  //
  // sets the correct file path for the oprions
  if (options.file)
    options.file = realPath3;
  //
  // if all paths are valid proceed
  if (realPath1 && realPath2 && realPath3) {
    App.gm.compare(realPath1, realPath2, options, function (err, isEqual, equality, raw) {
      cb({err: err, isEqual: isEqual, equality: equality, raw: raw}, err);
    });
  }
  else
    return cb(null, new Error("Invalid path"));
};


/**
 * @param {function} cb
 */
App.GM.prototype.size = function (cb)
{
  this.gm.size(function (err, value) {
    cb(value, err);
  });
};


/**
 * @param {function} cb
 */
App.GM.prototype.format = function (cb)
{
  this.gm.format(function (err, value) {
    cb(value, err);
  });
};


/**
 * @param {function} cb
 */
App.GM.prototype.depth = function (cb)
{
  this.gm.depth(function (err, value) {
    cb(value, err);
  });
};


/**
 * @param {function} cb
 */
App.GM.prototype.color = function (cb)
{
  this.gm.color(function (err, value) {
    cb(value, err);
  });
};


/**
 * @param {function} cb
 */
App.GM.prototype.res = function (cb)
{
  this.gm.res(function (err, value) {
    cb(value, err);
  });
};


/**
 * @param {function} cb
 */
App.GM.prototype.filesize = function (cb)
{
  this.gm.filesize(function (err, value) {
    cb(value, err);
  });
};


/**
 * @param {function} cb
 */
App.GM.prototype.identify = function (cb)
{
  this.gm.identify(function (err, value) {
    cb(value, err);
  });
};


/**
 * @param {function} cb
 */
App.GM.prototype.orientation = function (cb)
{
  this.gm.orientation(function (err, value) {
    cb(value, err);
  });
};


/* SYNC methods*/


/**
 * @param {string} text
 */
App.GM.prototype.comment = function (text)
{
  this.gm.comment(text);
};


/**
 * @param {string} name
 */
App.GM.prototype.label = function (name)
{
  this.gm.label(name);
};


App.GM.prototype.adjoin = function ()
{
  this.gm.adjoin();
};


/**
 * @param {object} matrix
 */
App.GM.prototype.affine = function (matrix)
{
  this.gm.affine(matrix);
};


/**
 * @param {App.File} fileImg
 * @param {string} ltr
 */
App.GM.prototype.append = function (fileImg, ltr)
{
  // checks and calculates the real path
  var imgPath = fileImg.fs.checkPath(false, fileImg);
  //
  // if path is invalid, returns an exception
  if (!imgPath)
    throw new Error("Invalid path");
  else
    this.gm.append(imgPath, ltr);
};


/**
 * @param {string} string
 */
App.GM.prototype.authenticate = function (string)
{
  this.gm.authenticate(string);
};


App.GM.prototype.average = function ()
{
  this.gm.average();
};


App.GM.prototype.backdrop = function ()
{
  this.gm.backdrop();
};


/**
 * @param {int} red
 * @param {int} green
 * @param {int} blue
 * @param {int} opacity
 */
App.GM.prototype.blackThreshold = function (red, green, blue, opacity)
{
  this.gm.blackThreshold(red, green, blue, opacity);
};


/**
 * @param {int} x
 * @param {int} y
 */
App.GM.prototype.bluePrimary = function (x, y)
{
  this.gm.bluePrimary(x, y);
};


/**
 * @param {int} width
 * @param {int} height
 */
App.GM.prototype.border = function (width, height)
{
  this.gm.border(width, height);
};


/**
 * @param {string} color
 */
App.GM.prototype.borderColor = function (color)
{
  this.gm.borderColor(color);
};


/**
 * @param {string} color
 */
App.GM.prototype.box = function (color)
{
  this.gm.box(color);
};


/**
 * @param {string} type
 */
App.GM.prototype.channel = function (type)
{
  this.gm.channel(type);
};


/**
 * @param {int} w
 * @param {int} h
 * @param {int} x
 * @param {int} y
 */
App.GM.prototype.chop = function (w, h, x, y)
{
  this.gm.chop(w, h, x, y);
};


App.GM.prototype.clip = function ()
{
  this.gm.clip();
};


App.GM.prototype.coalesce = function ()
{
  this.gm.coalesce();
};


/**
 * @param {int} r
 * @param {int} g
 * @param {int} b
 */
App.GM.prototype.colorize = function (r, g, b)
{
  this.gm.colorize(r, g, b);
};


/**
 * @param {string} type
 */
App.GM.prototype.colorMap = function (type)
{
  this.gm.colorMap(type);
};


/**
 * @param {App.File} fileImg
 * @param {App.File} fileMask
 */
App.GM.prototype.composite = function (fileImg, fileMask)
{
  // checks and calculates the real path
  var imgRealPath = fileImg.fs.checkPath(false, fileImg);
  var imgMaskPath = fileMask ? fileMask.fs.checkPath(false, fileMask) : true;
  //
  // if path is invalid, returns an exception
  if (!imgRealPath || !imgMaskPath)
    throw new Error("Invalid path");
  else if (fileMask) // if user specifies a mask file
    this.gm.composite(imgRealPath, imgMaskPath);
  else
    this.gm.composite(imgRealPath);
};


/**
 * @param {string} operator
 */
App.GM.prototype.compose = function (operator)
{
  this.gm.compose(operator);
};


/**
 * @param {string} type
 */
App.GM.prototype.compress = function (type)
{
  this.gm.compress(type);
};


/**
 * @param {object} kernel
 */
App.GM.prototype.convolve = function (kernel)
{
  this.gm.convolve(kernel);
};


App.GM.prototype.createDirectories = function ()
{
  this.gm.createDirectories();
};


App.GM.prototype.deconstruct = function ()
{
  this.gm.deconstruct();
};


/**
 * @param {string} value
 */
App.GM.prototype.define = function (value)
{
  this.gm.define(value);
};


/**
 * @param {int} value
 */
App.GM.prototype.delay = function (value)
{
  this.gm.delay(value);
};


/**
 * @param {int} horizontalScale
 * @param {int} verticalScale
 */
App.GM.prototype.displace = function (horizontalScale, verticalScale)
{
  this.gm.displace(horizontalScale, verticalScale);
};


/**
 * @param {string} value
 */
App.GM.prototype.display = function (value)
{
  this.gm.display(value);
};


/**
 * @param {string} method
 */
App.GM.prototype.dispose = function (method)
{
  this.gm.dispose(method);
};


/**
 * @param {int} percent
 */
App.GM.prototype.dissolve = function (percent)
{
  this.gm.dissolve(percent);
};


/**
 * @param {string} type
 */
App.GM.prototype.encoding = function (type)
{
  this.gm.encoding(type);
};


/**
 * @param {string} type
 */
App.GM.prototype.endian = function (type)
{
  this.gm.endian(type);
};


/**
 * @param {App.File} file
 */
App.GM.prototype.file = function (file)
{
  // checks and calculates the real path
  var imgPath = file.fs.checkPath(false, file);
  //
  // if path is invalid, returns an exception
  if (!imgPath)
    throw new Error("Invalid path");
  else
    this.gm.file(imgPath);
};


App.GM.prototype.flatten = function ()
{
  this.gm.flatten();
};


/**
 * @param {string} color
 */
App.GM.prototype.foreground = function (color)
{
  this.gm.foreground(color);
};


/**
 * @param {int} width
 * @param {int} height
 * @param {int} outerBevelWidth
 * @param {int} innerBevelWidth
 */
App.GM.prototype.frame = function (width, height, outerBevelWidth, innerBevelWidth)
{
  this.gm.frame(width, height, outerBevelWidth, innerBevelWidth);
};


/**
 * @param {int} distance
 * @param {int} percent
 */
App.GM.prototype.fuzz = function (distance, percent)
{
  this.gm.fuzz(distance, percent);
};


/**
 * @param {int} radius
 * @param {int} sigma
 */
App.GM.prototype.gaussian = function (radius, sigma)
{
  this.gm.gaussian(radius, sigma);
};


/**
 * @param {int} width
 * @param {int} height
 * @param {string} arg
 */
App.GM.prototype.geometry = function (width, height, arg)
{
  this.gm.geometry(width, height, arg);
};


/**
 * @param {int} x
 * @param {int} y
 */
App.GM.prototype.greenPrimary = function (x, y)
{
  this.gm.greenPrimary(x, y);
};


/**
 * @param {string} color
 */
App.GM.prototype.highlightColor = function (color)
{
  this.gm.highlightColor(color);
};


/**
 * @param {string} style
 */
App.GM.prototype.highlightStyle = function (style)
{
  this.gm.highlightStyle(style);
};


/**
 * @param {string} geometry
 */
App.GM.prototype.iconGeometry = function (geometry)
{
  this.gm.iconGeometry(geometry);
};


/**
 * @param {string} type
 */
App.GM.prototype.intent = function (type)
{
  this.gm.intent(type);
};


/**
 * @param {string} width
 * @param {string} height
 * @param {string} offset
 * @param {int} percent
 */
App.GM.prototype.lat = function (width, height, offset, percent)
{
  this.gm.lat(width, height, offset, percent);
};


/**
 * @param {string} blackPoint
 * @param {string} gamma
 * @param {string} whitePoint
 * @param {int} percent
 */
App.GM.prototype.level = function (blackPoint, gamma, whitePoint, percent)
{
  this.gm.level(blackPoint, gamma, whitePoint, percent);
};


/**
 * @param {string} type
 */
App.GM.prototype.list = function (type)
{
  this.gm.list(type);
};


/**
 * @param {string} string
 */
App.GM.prototype.log = function (string)
{
  this.gm.log(string);
};


/**
 * @param {string} iterations
 */
App.GM.prototype.loop = function (iterations)
{
  this.gm.loop(iterations);
};


/**
 * @param {App.File} file
 */
App.GM.prototype.map = function (file)
{
  // checks and calculates the real path
  var filePath = file.fs.checkPath(false, file);
  //
  // if path is not valid returns an error
  if (!filePath)
    throw new Error("Invalid path");
  else
    this.gm.map(filePath);
};


/**
 * @param {App.File} file
 */
App.GM.prototype.mask = function (file)
{
  // checks and calculates the real path
  var filePath = file.fs.checkPath(true, file);
  //
  // if path is not valid returns an error
  if (!filePath)
    throw new Error("Invalid path");
  else
    this.gm.mask(filePath);
};


App.GM.prototype.matte = function ()
{
  this.gm.matte();
};


/**
 * @param {string} color
 */
App.GM.prototype.matteColor = function (color)
{
  this.gm.matteColor(color);
};


/**
 * @param {string} limit
 */
App.GM.prototype.maximumError = function (limit)
{
  this.gm.maximumError(limit);
};


/**
 * @param {string} value
 */
App.GM.prototype.mode = function (value)
{
  this.gm.mode(value);
};


App.GM.prototype.monitor = function ()
{
  this.gm.monitor();
};


App.GM.prototype.mosaic = function ()
{
  this.gm.mosaic();
};


/**
 * @param {string} radius
 * @param {string} sigma
 * @param {string} angle
 */
App.GM.prototype.motionBlur = function (radius, sigma, angle)
{
  this.gm.motionBlur(radius, sigma, angle);
};


App.GM.prototype.name = function ()
{
  this.gm.name();
};


App.GM.prototype.noop = function ()
{
  this.gm.noop();
};


App.GM.prototype.normalize = function ()
{
  this.gm.normalize();
};


/**
 * @param {string} color
 */
App.GM.prototype.opaque = function (color)
{
  this.gm.opaque(color);
};


/**
 * @param {string} channel
 * @param {string} operator
 * @param {string} rvalue
 * @param {string} percent
 */
App.GM.prototype.operator = function (channel, operator, rvalue, percent)
{
  this.gm.operator(channel, operator, rvalue, percent);
};


/**
 * @param {string} channeltype
 * @param {string} NxN
 */
App.GM.prototype.orderedDither = function (channeltype, NxN)
{
  this.gm.orderedDither(channeltype, NxN);
};


/**
 * @param {App.Directory} directory
 */
App.GM.prototype.outputDirectory = function (directory)
{
  // checks and calculates the real path
  var dirPath = directory.fs.checkPath(true, directory);
  //
  // if path is not valid returns an error
  if (!dirPath)
    throw new Error("Invalid path");
  else
    this.gm.outputDirectory(dirPath);
};


/**
 * @param {string} width
 * @param {string} height
 * @param {string} arg
 */
App.GM.prototype.page = function (width, height, arg)
{
  this.gm.page(width, height, arg);
};


/**
 * @param {string} seconds
 */
App.GM.prototype.pause = function (seconds)
{
  this.gm.pause(seconds);
};


/**
 * @param {string} color
 */
App.GM.prototype.pen = function (color)
{
  this.gm.pen(color);
};


App.GM.prototype.ping = function ()
{
  this.gm.ping();
};


/**
 * @param {string} value
 */
App.GM.prototype.pointSize = function (value)
{
  this.gm.pointSize(value);
};


/**
 * @param {string} type
 */
App.GM.prototype.preview = function (type)
{
  this.gm.preview(type);
};


/**
 * @param {string} command
 */
App.GM.prototype.process = function (command)
{
  this.gm.process(command);
};


/**
 * @param {App.File} file
 */
App.GM.prototype.profile = function (file)
{
  // checks and calculates the real path
  var filePath = file.fs.checkPath(false, file);
  //
  // if path is not valid returns an error
  if (!filePath)
    throw new Error("Invalid path");
  else
    this.gm.profile(filePath);
};


App.GM.prototype.progress = function ()
{
  this.gm.progress();
};


/**
 * @param {string} channeltype
 * @param {string} LOWxHIGH
 */
App.GM.prototype.randomThreshold = function (channeltype, LOWxHIGH)
{
  this.gm.randomThreshold(channeltype, LOWxHIGH);
};


/**
 * @param {string} matrix
 */
App.GM.prototype.recolor = function (matrix)
{
  this.gm.recolor(matrix);
};


/**
 * @param {string} x
 * @param {string} y
 */
App.GM.prototype.redPrimary = function (x, y)
{
  this.gm.redPrimary(x, y);
};


App.GM.prototype.remote = function ()
{
  this.gm.remote();
};


App.GM.prototype.render = function ()
{
  this.gm.render();
};


/**
 * @param {string} width
 * @param {string} height
 * @param {string} xoff
 * @param {string} yoff
 * @param {string} arg
 */
App.GM.prototype.repage = function (width, height, xoff, yoff, arg)
{
  this.gm.repage(width, height, xoff, yoff, arg);
};


/**
 * @param {string} geometry
 */
App.GM.prototype.sample = function (geometry)
{
  this.gm.sample(geometry);
};


/**
 * @param {string} horizontalFactor
 * @param {string} verticalFactor
 */
App.GM.prototype.samplingFactor = function (horizontalFactor, verticalFactor)
{
  this.gm.samplingFactor(horizontalFactor, verticalFactor);
};


/**
 * @param {string} value
 */
App.GM.prototype.scene = function (value)
{
  this.gm.scene(value);
};


/**
 * @param {string} start
 * @param {string} end
 */
App.GM.prototype.scenes = function (start, end)
{
  this.gm.scenes(start, end);
};


App.GM.prototype.screen = function ()
{
  this.gm.screen();
};


/**
 * @param {string} attribute
 * @param {string} value
 */
App.GM.prototype.set = function (attribute, value)
{
  this.gm.set(attribute, value);
};


/**
 * @param {string} clusterThreshold
 * @param {string} smoothingThreshold
 */
App.GM.prototype.segment = function (clusterThreshold, smoothingThreshold)
{
  this.gm.segment(clusterThreshold, smoothingThreshold);
};


/**
 * @param {string} azimuth
 * @param {string} elevation
 */
App.GM.prototype.shade = function (azimuth, elevation)
{
  this.gm.shade(azimuth, elevation);
};


/**
 * @param {string} radius
 * @param {string} sigma
 */
App.GM.prototype.shadow = function (radius, sigma)
{
  this.gm.shadow(radius, sigma);
};


App.GM.prototype.sharedMemory = function ()
{
  this.gm.sharedMemory();
};


/**
 * @param {string} width
 * @param {string} height
 * @param {string} percent
 */
App.GM.prototype.shave = function (width, height, percent)
{
  this.gm.shave(width, height, percent);
};


/**
 * @param {string} xDegrees
 * @param {string} yDegreees
 */
App.GM.prototype.shear = function (xDegrees, yDegreees)
{
  this.gm.shear(xDegrees, yDegreees);
};


/**
 * @param {string} color
 */
App.GM.prototype.silent = function (color)
{
  this.gm.silent(color);
};


/**
 * @param {string} width
 * @param {string} height
 * @param {string} offset
 */
App.GM.prototype.rawSize = function (width, height, offset)
{
  this.gm.rawSize(width, height, offset);
};


/**
 * @param {string} value
 */
App.GM.prototype.snaps = function (value)
{
  this.gm.snaps(value);
};


/**
 * @param {string} offset
 */
App.GM.prototype.stegano = function (offset)
{
  this.gm.stegano(offset);
};


App.GM.prototype.stereo = function ()
{
  this.gm.stereo();
};


/**
 * @param {string} name
 */
App.GM.prototype.textFont = function (name)
{
  this.gm.textFont(name);
};


/**
 * @param {App.File} file
 */
App.GM.prototype.texture = function (file)
{
  // checks and calculates the real path
  var filePath = file.fs.checkPath(false, file);
  //
  // if path is not valid returns an error
  if (!filePath)
    throw new Error("Invalid path");
  else
    this.gm.texture(filePath);
};


/**
 * @param {string} value
 * @param {string} percent
 */
App.GM.prototype.threshold = function (value, percent)
{
  this.gm.threshold(value, percent);
};


/**
 * @param {App.File} file
 */
App.GM.prototype.tile = function (file)
{
  // checks and calculates the real path
  var filePath = file.fs.checkPath(false, file);
  //
  // if path is not valid returns an error
  if (!filePath)
    throw new Error("Invalid path");
  else
    this.gm.tile(filePath);
};


/**
 * @param {string} color
 */
App.GM.prototype.transform = function (color)
{
  this.gm.transform(color);
};


/**
 * @param {string} color
 */
App.GM.prototype.transparent = function (color)
{
  this.gm.transparent(color);
};


/**
 * @param {int} value
 */
App.GM.prototype.treeDepth = function (value)
{
  this.gm.treeDepth(value);
};


/**
 * @param {int} seconds
 */
App.GM.prototype.update = function (seconds)
{
  this.gm.update(seconds);
};


/**
 * @param {string} type
 */
App.GM.prototype.units = function (type)
{
  this.gm.units(type);
};


/**
 * @param {string} radius
 * @param {string} sigma
 * @param {string} amount
 * @param {string} threshold
 */
App.GM.prototype.unsharp = function (radius, sigma, amount, threshold)
{
  this.gm.unsharp(radius, sigma, amount, threshold);
};


App.GM.prototype.usePixmap = function ()
{
  this.gm.usePixmap();
};


/**
 * @param {string} string
 */
App.GM.prototype.view = function (string)
{
  this.gm.view(string);
};


/**
 * @param {string} method
 */
App.GM.prototype.virtualPixel = function (method)
{
  this.gm.virtualPixel(method);
};


/**
 * @param {string} type
 */
App.GM.prototype.visual = function (type)
{
  this.gm.visual(type);
};


/**
 * @param {string} brightness
 * @param {string} saturation
 */
App.GM.prototype.watermark = function (brightness, saturation)
{
  this.gm.watermark(brightness, saturation);
};


/**
 * @param {string} amplitude
 * @param {string} wavelength
 */
App.GM.prototype.wave = function (amplitude, wavelength)
{
  this.gm.wave(amplitude, wavelength);
};


/**
 * @param {string} x
 * @param {string} y
 */
App.GM.prototype.whitePoint = function (x, y)
{
  this.gm.whitePoint(x, y);
};


/**
 * @param {string} red
 * @param {string} green
 * @param {string} blue
 * @param {string} opacity
 */
App.GM.prototype.whiteThreshold = function (red, green, blue, opacity)
{
  this.gm.whiteThreshold(red, green, blue, opacity);
};


/**
 * @param {string} id
 */
App.GM.prototype.window = function (id)
{
  this.gm.window(id);
};


App.GM.prototype.windowGroup = function ()
{
  this.gm.windowGroup();
};


App.GM.prototype.strip = function ()
{
  this.gm.strip();
};


/**
 * @param {string} type
 */
App.GM.prototype.interlace = function (type)
{
  this.gm.interlace(type);
};


/**
 * @param {string} format
 */
App.GM.prototype.setFormat = function (format)
{
  this.gm.setFormat(format);
};


/**
 * @param {string} w
 * @param {string} h
 * @param {string} options
 */
App.GM.prototype.resize = function (w, h, options)
{
  this.gm.resize(w, h, options);
};


/**
 * @param {string} w
 * @param {string} h
 * @param {string} options
 */
App.GM.prototype.scale = function (w, h, options)
{
  this.gm.scale(w, h, options);
};


/**
 * @param {string} val
 */
App.GM.prototype.filter = function (val)
{
  this.gm.filter(val);
};


/**
 * @param {string} w
 * @param {string} h
 */
App.GM.prototype.density = function (w, h)
{
  this.gm.density(w, h);
};


App.GM.prototype.noProfile = function ()
{
  this.gm.noProfile();
};


/**
 * @param {string} w
 * @param {string} h
 */
App.GM.prototype.resample = function (w, h)
{
  this.gm.resample(w, h);
};


/**
 * @param {string} color
 * @param {string} deg
 */
App.GM.prototype.rotate = function (color, deg)
{
  this.gm.rotate(color, deg);
};


App.GM.prototype.flip = function ()
{
  this.gm.flip();
};


App.GM.prototype.flop = function ()
{
  this.gm.flop();
};


/**
 * @param {string} w
 * @param {string} h
 * @param {string} x
 * @param {string} y
 * @param {string} percent
 */
App.GM.prototype.crop = function (w, h, x, y, percent)
{
  this.gm.crop(w, h, x, y, percent);
};


/**
 * @param {string} factor
 */
App.GM.prototype.magnify = function (factor)
{
  this.gm.magnify(factor);
};


App.GM.prototype.minify = function ()
{
  this.gm.minify();
};


/**
 * @param {string} val
 */
App.GM.prototype.quality = function (val)
{
  this.gm.quality(val);
};


/**
 * @param {string} radius
 * @param {string} sigma
 */
App.GM.prototype.blur = function (radius, sigma)
{
  this.gm.blur(radius, sigma);
};


/**
 * @param {string} factor
 */
App.GM.prototype.charcoal = function (factor)
{
  this.gm.charcoal(factor);
};


/**
 * @param {string} b
 * @param {string} s
 * @param {string} h
 */
App.GM.prototype.modulate = function (b, s, h)
{
  this.gm.modulate(b, s, h);
};


/**
 * @param {string} disable
 */
App.GM.prototype.antialias = function (disable)
{
  this.gm.antialias(disable);
};


/**
 * @param {string} val
 */
App.GM.prototype.bitdepth = function (val)
{
  this.gm.bitdepth(val);
};


/**
 * @param {string} val
 */
App.GM.prototype.colors = function (val)
{
  this.gm.colors(val);
};


/**
 * @param {string} val
 */
App.GM.prototype.colorspace = function (val)
{
  this.gm.colorspace(val);
};


/**
 * @param {string} mult
 */
App.GM.prototype.contrast = function (mult)
{
  this.gm.contrast(mult);
};


/**
 * @param {string} amount
 */
App.GM.prototype.cycle = function (amount)
{
  this.gm.cycle(amount);
};


App.GM.prototype.despeckle = function ()
{
  this.gm.despeckle();
};


/**
 * @param {string} on
 */
App.GM.prototype.dither = function (on)
{
  this.gm.dither(on);
};


App.GM.prototype.monochrome = function ()
{
  this.gm.monochrome();
};


/**
 * @param {string} radius
 */
App.GM.prototype.edge = function (radius)
{
  this.gm.edge(radius);
};


/**
 * @param {string} radius
 */
App.GM.prototype.emboss = function (radius)
{
  this.gm.emboss(radius);
};


App.GM.prototype.enhance = function ()
{
  this.gm.enhance();
};


App.GM.prototype.equalize = function ()
{
  this.gm.equalize();
};


/**
 * @param {string} r
 * @param {string} g
 * @param {string} b
 */
App.GM.prototype.gamma = function (r, g, b)
{
  this.gm.gamma(r, g, b);
};


/**
 * @param {string} factor
 */
App.GM.prototype.implode = function (factor)
{
  this.gm.implode(factor);
};


/**
 * @param {string} type
 * @param {string} val
 */
App.GM.prototype.limit = function (type, val)
{
  this.gm.limit(type, val);
};


/**
 * @param {string} radius
 */
App.GM.prototype.median = function (radius)
{
  this.gm.median(radius);
};


/**
 * @param {string} grayscale
 */
App.GM.prototype.negative = function (grayscale)
{
  this.gm.negative(grayscale);
};


/**
 * @param {string} radius
 */
App.GM.prototype.noise = function (radius)
{
  this.gm.noise(radius);
};


/**
 * @param {string} radius
 */
App.GM.prototype.paint = function (radius)
{
  this.gm.paint(radius);
};


/**
 * @param {string} w
 * @param {string} h
 */
App.GM.prototype.raise = function (w, h)
{
  this.gm.raise(w, h);
};


/**
 * @param {string} w
 * @param {string} h
 */
App.GM.prototype.lower = function (w, h)
{
  this.gm.lower(w, h);
};


/**
 * @param {string} w
 * @param {string} h
 * @param {string} x
 * @param {string} y
 */
App.GM.prototype.region = function (w, h, x, y)
{
  this.gm.region(w, h, x, y);
};


/**
 * @param {string} x
 * @param {string} y
 */
App.GM.prototype.roll = function (x, y)
{
  this.gm.roll(x, y);
};


/**
 * @param {string} radius
 * @param {string} sigma
 */
App.GM.prototype.sharpen = function (radius, sigma)
{
  this.gm.sharpen(radius, sigma);
};


/**
 * @param {string} factor
 */
App.GM.prototype.solarize = function (factor)
{
  this.gm.solarize(factor);
};


/**
 * @param {string} amount
 */
App.GM.prototype.spread = function (amount)
{
  this.gm.spread(amount);
};


/**
 * @param {string} degrees
 */
App.GM.prototype.swirl = function (degrees)
{
  this.gm.swirl(degrees);
};


/**
 * @param {string} type
 */
App.GM.prototype.type = function (type)
{
  this.gm.type(type);
};


App.GM.prototype.trim = function ()
{
  this.gm.trim();
};


/**
 * @param {string} w
 * @param {string} h
 * @param {string} options
 */
App.GM.prototype.extent = function (w, h, options)
{
  this.gm.extent(w, h, options);
};


/**
 * @param {string} type
 */
App.GM.prototype.gravity = function (type)
{
  this.gm.gravity(type);
};


/**
 * @param {string} color
 */
App.GM.prototype.fill = function (color)
{
  this.gm.fill(color);
};


/**
 * @param {string} color
 * @param {string} width
 */
App.GM.prototype.stroke = function (color, width)
{
  this.gm.stroke(color, width);
};


/**
 * @param {string} width
 */
App.GM.prototype.strokeWidth = function (width)
{
  this.gm.strokeWidth(width);
};


/**
 * @param {string || App.File} font
 */
App.GM.prototype.font = function (font)
{
  // if the font parameter is a file, is necessary get the full path of the font file
  if (font instanceof App.File) {
    // replaces font file with his absolute path
    if (font.fs instanceof App.NodeDriver)
      font = font.fs.appRoot + "/" + App.NodeDriver.typeMap[font.type] + "/" + font.path;
  }
  // sets font
  this.gm.font(font);
};


/**
 * @param {string} size
 */
App.GM.prototype.fontSize = function (size)
{
  this.gm.fontSize(size);
};


/**
 * @param {string} args
 */
App.GM.prototype.draw = function (args)
{
  this.gm.draw(args);
};


/**
 * @param {string} x
 * @param {string} y
 */
App.GM.prototype.drawPoint = function (x, y)
{
  this.gm.drawPoint(x, y);
};


/**
 * @param {string} x0
 * @param {string} y0
 * @param {string} x1
 * @param {string} y1
 */
App.GM.prototype.drawLine = function (x0, y0, x1, y1)
{
  this.gm.drawLine(x0, y0, x1, y1);
};


/**
 * @param {string} x0
 * @param {string} y0
 * @param {string} x1
 * @param {string} y1
 * @param {string} wc
 * @param {string} hc
 */
App.GM.prototype.drawRectangle = function (x0, y0, x1, y1, wc, hc)
{
  this.gm.drawRectangle(x0, y0, x1, y1, wc, hc);
};


/**
 * @param {string} x0
 * @param {string} y0
 * @param {string} x1
 * @param {string} y1
 * @param {string} a0
 * @param {string} a1
 */
App.GM.prototype.drawArc = function (x0, y0, x1, y1, a0, a1)
{
  this.gm.drawArc(x0, y0, x1, y1, a0, a1);
};


/**
 * @param {string} x0
 * @param {string} y0
 * @param {string} rx
 * @param {string} ry
 * @param {string} a0
 * @param {string} a1
 */
App.GM.prototype.drawEllipse = function (x0, y0, rx, ry, a0, a1)
{
  this.gm.drawEllipse(x0, y0, rx, ry, a0, a1);
};


/**
 * @param {string} x0
 * @param {string} y0
 * @param {string} x1
 * @param {string} y1
 */
App.GM.prototype.drawCircle = function (x0, y0, x1, y1)
{
  this.gm.drawCircle(x0, y0, x1, y1);
};


App.GM.prototype.drawPolyline = function ()
{
  this.gm.drawPolyline();
};


App.GM.prototype.drawPolygon = function ()
{
  this.gm.drawPolygon();
};


App.GM.prototype.drawBezier = function ()
{
  this.gm.drawBezier();
};


/**
 * @param {string} x0
 * @param {string} y0
 * @param {string} text
 * @param {string} gravity
 */
App.GM.prototype.drawText = function (x0, y0, text, gravity)
{
  this.gm.drawText(x0, y0, text, gravity);
};


/**
 * @param {string} prop
 * @param {string} x
 * @param {string} y
 * @param {string} method
 */
App.GM.prototype.setDraw = function (prop, x, y, method)
{
  this.gm.setDraw(prop, x, y, method);
};


/**
 * @param {int} w
 * @param {int} h
 * @param {App.File} outFile
 * @param {int} quality
 * @param {function} cb
 */
App.GM.prototype.thumb = function (w, h, outFile, quality, cb)
{
  // checks and calculates the real path
  var outFilePath = outFile.fs.checkPath(false, outFile);
  //
  // if path is not valid returns an error
  if (!outFilePath)
    return cb(null, new Error("Invalid path"));
  //
  this.gm.thumb(w, h, outFilePath, quality, function (err) {
    cb(null, err);
  });
};


/**
 * @param {App.File} fileImg
 * @param {App.File} outFile
 * @param {function} cb
 */
App.GM.prototype.morph = function (fileImg, outFile, cb)
{
  // checks and calculates the real paths
  var fileImgPath = fileImg.fs.checkPath(false, fileImg);
  var outFilePath = outFile.fs.checkPath(true, outFile);
  //
  // if all paths are valid proceed
  if (fileImgPath && outFilePath) {
    this.gm.morph(fileImgPath, outFilePath, function (err) {
      cb(null, err);
    });
  }
  else
    return cb(null, new Error("Invalid path"));
};


App.GM.prototype.sepia = function ()
{
  this.gm.sepia();
};


App.GM.prototype.autoOrient = function ()
{
  this.gm.autoOrient();
};


/**
 * Creates QR code image
 * @param {object} app
 * @param {String} text
 * @param {String} path
 * @param {Object} options
 * @param {function} cb
 */
App.GM.createQRCode = function (app, text, path, options, cb)
{
  app;
  //
  var qr = require('qr-image');
  //
  // protection  against null options
  options = options || {};
  //
  // if there is a path, creates the file
  if (path) {
    // create the file that contains qr code
    var qrCodeFile = app.fs.file(path);
    //
    // calculate the real path
    var realPath = qrCodeFile.fs.checkPath(true, qrCodeFile);
    //
    // if the path is invalid, returns an error
    if (!realPath)
      return cb(null, new Error("Invalid path"));
    //
    // call 'qr-image' to create the QR code
    var buffer = qr.image(text, options);
    //
    // creates the file phisically
    qrCodeFile.create(null, function (data, err) {
      //
      // pipe the buffer in file's writable stream
      buffer.pipe(qrCodeFile.wstream);
      //
      // file's writable stream event 'finish'
      qrCodeFile.wstream.on('finish', function () {
        return cb(qrCodeFile, null);
      });
      //
      // file's writable stream event "error
      qrCodeFile.wstream.once("error", function (error) {
        return cb(null, error);
      });
    });
  }
  else {
    // return only the SVG of generated QR code
    return cb(qr.svgObject(text, options), null);
  }
};


// export module for node
if (module)
  module.exports = App.GM;