/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, Infinity, ArrayBuffer */

var App = App || {};

/**
 * @class Definition of locale
 * @param {Session} app that owns this instance
 * */
App.Locale = function (app)
{
  if (module && !App.View)
    App = require("./requireAll");
  this.app = app;       // App that owns this instance
  this.timeZone = "";
  this.dateFormat = "";
  this.timeFormat = "";
};


/*
 * Init defaults for language
 * @param {String} lang
 */
App.Locale.prototype.setLanguage = function (lang)
{
  var lc = App.moment().locale(lang || "").localeData();
  this.dateFormat = lc.longDateFormat("L");
  this.timeFormat = lc.longDateFormat("LTS");
};


/*
 * Returns a moment.tz with locale timeZone
 */
App.Locale.prototype.moment = function ()
{
  // Create a moment.tz(arguments, timeZone);
  var args = Array.prototype.slice.call(arguments).concat([this.timeZone]);
  var m = App.moment.tz.apply(this, args);
  //
  // Set language if setted
  if (this.app.device.language)
    m.locale(this.app.device.language);
  //
  return m;
};


/*
 * Returns now Date
 */
App.Locale.prototype.now = function ()
{
  return this.moment();
};


/*
 * Returns today Date
 */
App.Locale.prototype.today = function ()
{
  var now = this.now();
  return this.moment([now.year(), now.month(), now.date()]);
};


/*
 * Returns time Date
 */
App.Locale.prototype.time = function ()
{
  return this.now().year(1970).month(0).date(1);
};


/**
 * Convert a value to another datatype
 * @param {AnyType} v
 * @param {Object} options
 *   el - element for wich convert the value
 *   datatype
 * @returns {Any type}
 */
App.Locale.prototype.convertValue = function (v, options)
{
  options = options || {};
  //
  // Without destination datatype or array no conversion
  if (!options.dstDT || (v instanceof Array && options.dstDT !== App.Database.DataTypes.JSON && options.srcDT !== App.Database.DataTypes.JSON))
    return v;
  //
  if (v === null || v === undefined || Number.isNaN(v) || v === Infinity)
    return null;
  //
  switch (typeof v) {
    case "boolean":
      {
        switch (App.Database.classDT(options.dstDT)) {
          case App.Database.DataTypeClasses.JSON:
          case App.Database.DataTypeClasses.BOOL:
            // Boolean -> Boolean
            return v;

          case App.Database.DataTypeClasses.NUMBER:
            // Boolean -> Number
            return (v ? -1 : 0);

          case App.Database.DataTypeClasses.STRING:
            // Boolean -> String
            return v.toString();
        }
      }
      break;

    case "number":
      {
        switch (App.Database.classDT(options.dstDT)) {
          case App.Database.DataTypeClasses.BOOL:
            // Number -> Boolean
            return (v ? true : false);

          case App.Database.DataTypeClasses.JSON:
          case App.Database.DataTypeClasses.NUMBER:
          {
            // Number -> Number
            if (options.dstDT === App.Database.DataTypes.INT || options.dstDT === App.Database.DataTypes.AUTOINCR)
              return Math.trunc(v);
            else
              return v;
          }

          case App.Database.DataTypeClasses.STRING:
            // Number -> String
            // TODO: consider decimal separator
            return v.toString();

          case App.Database.DataTypeClasses.DATETIME:
          {
            // Number -> Date
            var nv = new Date(v);
            if (!Number.isNaN(nv.getTime()))
              return nv;
            break;
          }
        }
      }
      break;

    case "string":
      {
        switch (App.Database.classDT(options.dstDT)) {
          case App.Database.DataTypeClasses.BOOL:
            // String -> Boolean
            return (v !== "false" && v !== "0" && v !== "");

          case App.Database.DataTypeClasses.NUMBER:
          {
            // String -> Number
            var nv = null;
            try {
              nv = JSON.parse(v);
              if (typeof nv === "number") {
                if (options.dstDT === App.Database.DataTypes.INT || options.dstDT === App.Database.DataTypes.AUTOINCR)
                  return Math.trunc(nv);
                else
                  return nv;
              }
            }
            catch (e) {
            }
            break;
          }

          case App.Database.DataTypeClasses.JSON:
          {
            // String -> JSON
            try {
              return JSON.parse(v);
            }
            catch (e) {
            }
            break;
          }

          case App.Database.DataTypeClasses.STRING:
            // JSON -> String
            if (options.srcDT === App.Database.DataTypes.JSON)
              return JSON.stringify(v);
            //
            // String -> String
            return v;

          case App.Database.DataTypeClasses.DATETIME:
          {
            // String - > Date
            var format = options.format;
            if (!format) {
              if (options.dstDT === App.Database.DataTypes.DATE)
                format = ["YYYY-MM-DD", this.dateFormat];
              else if (options.dstDT === App.Database.DataTypes.DATETIME)
                format = ["YYYY-MM-DDTHH:mm:ss.SSSZ", this.dateFormat + " " + this.timeFormat];
              else if (options.dstDT === App.Database.DataTypes.TIME)
                format = ["HH:mm:ss.SSS", "HH:mm:ss", "HH:mm", this.timeFormat];
            }
            //
            if (options.dstDT === App.Database.DataTypes.TIME) {
              if (!(format instanceof Array))
                format = [format];
              for (var i = 0; i < format.length; i++) {
                if (!format[i].startsWith("YYYY-MM-DD"))
                  format[i] = "YYYY-MM-DD " + format[i];
              }
              if (!v.startsWith("1970-01-01"))
                v = "1970-01-01 " + v;
            }
            //
            var m = this.moment(v, format);
            if (m.isValid())
              return m.toDate();
            break;
          }

          case App.Database.DataTypeClasses.BLOB:
          {
            // Base64 -> ArrayBuffer
            if (v.startsWith("data:"))
              return App.Utils.base64ToBuffer(v.split(";base64,")[1]);
            //
            // String -> ArrayBuffer
            var buf = new ArrayBuffer(v.length * 2);
            var bufView = new Uint16Array(buf);
            for (var i = 0; i < v.length; i++)
              bufView[i] = v.charCodeAt(i);
            return buf;
          }
        }
      }
      break;

    case "object":
    {
      if (v instanceof Date || v instanceof App.moment) {
        // Handles moment as Date
        if (v instanceof App.moment)
          v = v.toDate();
        //
        switch (App.Database.classDT(options.dstDT)) {
          case App.Database.DataTypeClasses.BOOL:
            // Date -> Boolean
            return true;

          case App.Database.DataTypeClasses.NUMBER:
            // Date -> Number
            return v.getTime();

          case App.Database.DataTypeClasses.STRING:
          {
            // Date -> String
            var format = options.format;
            if (!format) {
              if (options.srcDT === App.Database.DataTypes.DATE)
                format = "YYYY-MM-DD";
              else if (options.srcDT === App.Database.DataTypes.DATETIME)
                return v.toISOString();
              else if (options.srcDT === App.Database.DataTypes.TIME)
                format = "HH:mm:ss.SSS";
            }
            //
            return this.moment(v).format(format);
          }

          case App.Database.DataTypeClasses.DATETIME:
            // Date -> Date
            // TODO: remove date/time part
            return v;
        }
      }
      //
      switch (App.Database.classDT(options.dstDT)) {
        case App.Database.DataTypeClasses.BLOB:
          if (v instanceof ArrayBuffer)
            return v;
          if (v.buffer instanceof ArrayBuffer)
            return new Uint8Array(v).buffer;
          break;

        case App.Database.DataTypeClasses.STRING:
          if (v instanceof ArrayBuffer)
            return "data:application/octet-stream;base64," + App.Utils.bufferToBase64(v);
          if (options.srcDT === App.Database.DataTypes.JSON)
            return JSON.stringify(v);
          break;

        case App.Database.DataTypeClasses.JSON:
          try
          {
            // Check if object is JSON.stringifable
            JSON.stringify(v);
            return v;
          }
          catch (e) {
          }
          break;
      }
      break;
    }
  }
  //
  return null;
};


/**
 * Compare two value
 * @param {AnyType} v1
 * @param {AnyType} v2
 */
App.Locale.compareValues = function (v1, v2)
{
  if (typeof v1 !== typeof v2)
    return false;
  //
  if (v1 instanceof Date && v2 instanceof Date)
    return v1.getTime() === v2.getTime();
  //
  return v1 === v2;
};


//export module for node
if (module)
  module.exports = App.Locale;
