/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module, require */

var App = App || {};

if (module) {
  App.Database = require("./database");
  App.Phantom = require("node-phantom-simple");
  App.File = require("./fs/file.js")
}


/*
 * Interface to PhantomJS interpreter
 */
App.PhantomJS = function (app)
{
  this.app = app; // unused
};


/*
 * Spawn the phantomjs interpreter when the class is asked to do something.
 * It will die with the child/worker
 */
App.PhantomJS.prototype.init = function (cb)
{
  if (App.phantom) {
    if (cb)
      cb(App.phantom);
    return;
  }
  //
  App.Phantom.create({}, function (err, ph) {
    if (err)
      throw err;
    // the interpreter is in App hands, not app's!!
    App.phantom = ph;
    if (cb)
      cb(App.phantom);
  });
};


/*
 * Given a URL, get the HTML code.
 * @params {string} url
 * @options {object} options - options.delay: number of seconds to wait for the page to load
 */
App.PhantomJS.prototype.getHTML = function (url, options, cb)
{
  //
  if (typeof url !== "string")
    throw "Url not valid";
  //
  this.init(function (ph) {
    var _options = {
      delay: 0,
    };
    //
    if (typeof options === "object") {
      //
      if (typeof options.delay === "number")
        _options.delay = options.delay;
    }
    //
    ph.createPage(function (err, page) {
      if (err)
        throw err;
      page.open(url, function (err, status) {
        if (err)
          throw err;
        setTimeout(function () {
          page.evaluate(function () {
            return document.getElementsByTagName('html')[0].innerHTML;
          }, function (err, result) {
            if (err)
              throw err;
            cb("<html>" + result + "</html>", status);
          });
        }, _options.delay * 1000);
      });
    });
  });
};


/*
 * Render a page to pdf or image
 * @params {string} argument - url or html to render
 * @params {App.File} outputFile - FS file to write the output to
 * @params {object} options - rendering options (page size, delay before rendering...)
 */
App.PhantomJS.prototype.render = function (argument, outputFile, options, cb)
{
  var pthis = this;
  var type;
  //
  if (!(outputFile instanceof App.File))
    throw "No output file specified";
  //
  this.init(function (ph) {
    var _options = {
      fileType: "pdf",
      quality: "80",
      delay: 0,
      layout: {
        format: 'A4', //  'A3', 'A4', 'A5', 'Legal', 'Letter', 'Tabloid'.
        orientation: 'portrait', // 'portrait', 'landscape'
        margin: {// 'mm', 'cm', 'in', 'px'
          top: '1cm',
          left: '1cm',
          right: '1cm',
          bottom: '1cm'
        },
        // header: {
        //   height: "1cm",
        //   contents: ph.callback(function(pageNum, numPages) {
        //     return "<h1>Header <span style='float:right'>" + pageNum + " / " + numPages + "</span></h1>";
        //   })
        // },
        // footer: {
        //   height: "1cm",
        //   contents: ph.callback(function(pageNum, numPages) {
        //     return "<h1>Footer <span style='float:right'>" + pageNum + " / " + numPages + "</span></h1>";
        //   })
        // }
      }
    };
    //
    if (typeof options === "object") {
      // did we receive a filetype?
      if (["pdf", "jpeg", "png", "gif"].indexOf(options.fileType) >= 0)
        _options.fileType = options.fileType;
      // did we receive a quality setting?
      if (!isNaN(parseInt(options.quality)))
        _options.quality = options.quality;
      //
      if (typeof options.delay === "number")
        _options.delay = options.delay;
      // did we receive layout options?
      if (typeof options.layout === "object") {
        for (var key in options.layout) {
          _options.layout[key] = options.layout[key];
        }
      }
    }
    //
    ph.createPage(function (err, page) {
      if (err)
        throw err;
      page.set("paperSize", _options.layout, function (err) {
        if (err)
          throw err;
        outputFile._getFullPath(function (filename) {
          //
          if (argument instanceof App.File) {
            type = "html";
            argument.readAll(function (data, err) {
              if (err)
                throw err;
              //
              page.set("content", data, function (err) {
                if (err)
                  throw err;
                page.render(filename, {format: _options.fileType, quality: _options.quality}, function (err) {
                  if (err)
                    throw err;
                  //
                  if (cb)
                    cb(outputFile);
                });
              });
            });
          } else if (typeof argument === "string") {
            if (argument.indexOf("<") < 0) {
              // it's a url
              page.open(argument, function (err, status) {
                if (err)
                  throw err;
                //
                setTimeout(function () {
                  page.render(filename, {format: _options.fileType, quality: _options.quality}, function (err) {
                    if (err)
                      throw err;
                    //
                    if (cb)
                      cb(outputFile, status);
                  });
                }, _options.delay * 1000);
              });
            } else {
              // it's html
              page.set("content", argument, function (err) {
                if (err)
                  throw err;
                //
                page.render(filename, {format: _options.fileType, quality: _options.quality}, function (err) {
                  if (err)
                    throw err;
                  //
                  if (cb)
                    cb(outputFile);
                });
              });
            }
          } else {
            throw "Argument is not html nor a url";
          }
        });
      });
    });
  });
};


//export module for node
if (module)
  module.exports = App.PhantomJS;
