/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module, process, Buffer */

var App = App || {};


/**
 * @class A RemoteDriver Object
 * @param {Database} parent
 * */
App.RemoteDriver = function (parent)
{
  if (module && !App.View)
    App = require("./requireAll");
  //
  this.id = App.Utils.generateUID24();
  this.parent = parent;
  App.RemoteDriver.objectsMap[this.id] = this;
};

App.RemoteDriver.callbacks = {};
App.RemoteDriver.objectsMap = {};


/**
 * Open the connection to the remote database
 * @param {Function} callback - function to be called at the end
 */
App.RemoteDriver.prototype.connect = function (callback)
{
  this.sendMessage({cmd: "open"}, callback);
};


/**
 * Close the connection to the remote database
 * @param {Function} callback - function to be called at the end
 */
App.RemoteDriver.prototype.disconnect = function (callback)
{
  delete App.RemoteDriver.objectsMap[this.id];
  delete this.connection;
  this.sendMessage({cmd: "close"}, callback);
};


/**
 * Execute an SQL statement
 * @param {String} sql to execute
 * @param {Array} parameters
 * @param {Function} callback - function to be called at the end
 * @return {Integer} number of rows affected
 */
App.RemoteDriver.prototype.execute = function (sql, parameters, callback)
{
  // Serialize some parameters
  var pars = parameters.map(function (p) {
    if (p instanceof Buffer)
      return {type: "buffer", data: App.Utils.bufferToBase64(p.buffer)};
    else
      return p;
  });
  //
  var msg = {cmd: "execute", sql: sql, pars: pars};
  //
  if (sql.endsWith("into :counter"))
    msg.ct = true;
  //
  this.sendMessage(msg, callback);
};


/*
 * Begin a transaction
 * @param {Function} callback - function to be called at the end
 */
App.RemoteDriver.prototype.beginTransaction = function (callback)
{
  this.sendMessage({cmd: "begin"}, callback);
};


/*
 * Commit a transaction
 * @param {Function} callback - function to be called at the end
 */
App.RemoteDriver.prototype.commitTransaction = function (callback)
{
  this.sendMessage({cmd: "commit"}, callback);
};


/*
 * Rollback a transaction
 * @param {Function} callback - function to be called at the end
 */
App.RemoteDriver.prototype.rollbackTransaction = function (callback)
{
  this.sendMessage({cmd: "rollback"}, callback);
};


/*
 * Send message to remote database
 * @param {Object} msg - message to send
 * @param {Function} callback - function to be called at the end
 */
App.RemoteDriver.prototype.sendMessage = function (msg, callback)
{
  // Compose the message
  var m = {};
  m.type = "ccm";
  m.cnt = {};
  m.cnt.conn = this.parent.remoteConnector.split("/")[0];
  m.cnt.type = "remoteCmd";
  m.cnt.data = msg;
  m.cnt.key = this.parent.APIKey;
  msg.cid = this.id;
  //
  // If I'm in the IDE
  if (!this.parent.app)
    msg.dmid = this.parent.schema.name;
  msg.dm = this.parent.remoteConnector.split("/")[1];
  //
  if (callback) {
    msg.cbid = App.Utils.generateUID24();
    App.RemoteDriver.callbacks[msg.cbid] = {cb: callback, db: this, st: new Date()};
  }
  //
  // Dispatch to Node.Worker (master) or to InDe.AApp (IDE)
  process.send(m);
};


/**
 * Handle response of a remote command
 * @param {Object} msg
 */
App.RemoteDriver.handleResponse = function (msg)
{
  var cb = App.RemoteDriver.callbacks[msg.cbid];
  if (cb) {
    delete App.RemoteDriver.callbacks[msg.cbid];
    if (msg.data.result && cb.st)
      msg.data.result.times.net = (new Date() - cb.st);
    //
    if (msg.data.error)
      cb.cb(null, new Error(msg.data.error));
    else
      cb.cb(msg.data.result);
  }
};


/**
 * Handle disconnect of a remote connector
 * @param {Object} msg
 */
App.RemoteDriver.handleDisconnect = function (msg)
{
  // Invoke all callback awaiting with error
  var cbids = Object.keys(App.RemoteDriver.callbacks);
  for (var i = 0; i < cbids.length; i++) {
    var cbid = cbids[i];
    var cb = App.RemoteDriver.callbacks[cbid];
    if (cb.conn === msg.data.name) {
      delete App.RemoteDriver.callbacks[cbid];
      setImmediate(function () {
        cb.cb(null, new Error("Remote connector disconnected"));
      });
    }
  }
  //
  // Close all connections
  var ids = Object.keys(App.RemoteDriver.objectsMap);
  for (var i = 0; i < ids.length; i++) {
    var conn = App.RemoteDriver.objectsMap[ids[i]];
    if (conn.parent.remoteConnector.startsWith(msg.data.name + "/"))
      conn.parent.closeConnection();
  }
};


/**
 * Returns the i-th row of a resultSet
 * @param {Object} rs - recordset
 * @param {int} pos - position
 */
App.RemoteDriver.getRow = function (rs, pos)
{
  var row = rs.rows[pos];
  var newrow = {};
  for (var i = 0; i < rs.cols.length; i++)
    newrow[rs.cols[i]] = App.RemoteDriver.convertValue(row[i]);
  return newrow;
};


/**
 * Convert a native value into javascript value
 * @param {String} v
 */
App.RemoteDriver.convertValue = function (v)
{
  if (v && typeof v === "object" && v.type === "buffer" && v.data)
    return App.Utils.base64ToBuffer(v.data);
  //
  return v;
};


/*
 * Read some info from result
 * @param {String} sql
 * @param {Object} result
 */
App.RemoteDriver.prototype.readInfoFromResult = function (result)
{
  this.parent.lastRowsAffected = result.rowsAffected;
  this.parent.lastInsertId = result.insertId;
};


/*
 * Do nothing
 * @param {Function} callback - function to be called at the end
 */
App.RemoteDriver.prototype.ping = function (callback)
{
  this.sendMessage({cmd: "ping"}, function (result, error) {
    delete App.RemoteDriver.objectsMap[this.id];
    callback(result, error);
  }.bind(this));
};


//export module for node
if (module)
  module.exports = App.RemoteDriver;
