/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, process */

var App = App || {};


/**
 * @class RemotePlugin
 * Represents a remote plugin object interface
 * @param {Object} app
 * @param {String} name
 * @param {String} APIKey
 *
 */
App.RemotePlugin = function (app, name, APIKey)
{
  if (!module)
    throw new Error("Cannot create App.RemotePlugin instance in offline mode");
  //
  // Import modules
  if (module && !App.View)
    App = require("./requireAll");
  //
  this.app = app;
  this.name = name;
  this.APIKey = APIKey;
  //
  this.instances = {};
  this.instanceIndex = 0;
};


App.RemotePlugin.callbacks = {};


App.RemotePlugin.msgTypes = {
  callMethod: "cm",
  destroyObject: "do"
};


/*
 * Send message to remote plugin
 * @param {Object} msg
 * @param {Function} callback
 */
App.RemotePlugin.prototype.sendMessage = function (msg, callback)
{
  // Compose the message
  var m = {};
  m.type = "ccm";
  m.cnt = {};
  m.cnt.conn = this.name.split("://")[0];
  m.cnt.type = "remoteCmd";
  m.cnt.data = msg;
  m.cnt.key = this.APIKey;
  msg.plugin = this.name.split("://")[1];
  //
  // Store the callback and send only a callbackID
  msg.cbid = App.Utils.generateUID24();
  App.RemotePlugin.callbacks[msg.cbid] = {cb: callback, ccName: m.cnt.conn};
  //
  // Dispatch to Node.Worker (master) or to InDe.AApp (IDE)
  process.send(m);
};


/**
 * Call a method remotely
 * @param {Object/String} instance
 * @param {String} method
 * @param {Array} args
 * @param {Function} callback
 */
App.RemotePlugin.prototype.callMethod = function (instance, method, args, callback)
{
  var msg = {};
  msg.cmd = method;
  msg.type = App.RemotePlugin.msgTypes.callMethod;
  msg.obj = instance;
  //
  // Serialize object if needed
  if (typeof instance === "object") {
    var idx;
    var found = false;
    var keys = Object.keys(this.instances);
    for (var i = 0; i < keys.length; i++) {
      if (this.instances[keys[i]] === instance) {
        found = true;
        idx = keys[i];
        msg.obj = {};
        break;
      }
    }
    //
    if (!found) {
      msg.obj = this.serializeObject(instance) || {};
      //
      this.instances[this.instanceIndex] = instance;
      idx = this.instanceIndex++;
    }
    //
    msg.obj.instanceIndex = idx;
  }
  //
  msg.args = args || [];
  //
  // Send message
  this.sendMessage(msg, callback);
};


/**
 * Handle response of a remote command
 * @param {Object} msg
 */
App.RemotePlugin.handleResponse = function (msg)
{
  var cb = App.RemotePlugin.callbacks[msg.cbid];
  if (cb) {
    delete App.RemotePlugin.callbacks[msg.cbid];
    var args = [];
    args.push(msg.data.result);
    if (msg.data.error)
      args.push(new Error(msg.data.error));
    //
    cb.cb.apply(this, args);
  }
};


/**
 * Serialize given object
 * @param {Object} obj
 */
App.RemotePlugin.prototype.serializeObject = function (obj)
{

};


/**
 * Destroy given object
 * @param {Object} obj
 * @param {Function} callback
 */
App.RemotePlugin.prototype.destroyObject = function (obj, callback)
{
  var msg = {};
  msg.type = App.RemotePlugin.msgTypes.destroyObject;
  msg.obj = this.serializeObject(obj);
  //
  var keys = Object.keys(this.instances);
  for (var i = 0; i < keys.length; i++) {
    if (this.instances[keys[i]] === obj) {
      msg.obj.instanceIndex = keys[i];
      delete this.instances[keys[i]];
      break;
    }
  }
  //
  this.sendMessage(msg, callback);
};


/**
 * Handle disconnect of a remote connector
 * @param {Object} msg
 */ App.RemotePlugin.handleDisconnect = function (msg)
{
  // Invoke all callback awaiting with error
  var cbids = Object.keys(App.RemotePlugin.callbacks);
  for (var i = 0; i < cbids.length; i++) {
    var cbid = cbids[i];
    var cb = App.RemotePlugin.callbacks[cbid];
    if (cb.ccName === msg.data.name) {
      delete App.RemotePlugin.callbacks[cbid];
      setImmediate(function () {
        cb.cb(null, new Error("Remote connector disconnected"));
      });
    }
  }
};


// export module for node
if (module)
  module.exports = App.RemotePlugin;