/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module,Proxy */

var App = App || {};

if (module)
  App.Database = require("./database");

/**
 * @class A SQLite Database Object
 * @param {Session} app that owns this instance
 * @extends App.Database
 * */
App.SQLite = function (app)
{
  if (module && !App.View)
    App = require("./requireAll");
  App.Database.call(this, app);
  //
  this.typemap = App.SQLite.DataTypeMap; // Map of data types
  this.keywords = App.SQLite.Keywords;   // List of reserved keywords
  //
  // Definition of functions and operators
  this.functions = Object.assign({}, App.Database.Functions, App.SQLite.Functions);
};


// Make App.SQLite extend App.Database
App.SQLite.prototype = new App.Database();

App.SQLite.DataTypeMap = {
  i: ["integer"],
  f: ["real"],
  d: ["decimal"],
  vc: ["varchar"],
  c: ["character"],
  bl: ["blob"],
  cl: ["text"],
  b: ["boolean"],
  da: ["date"],
  ti: ["datetime"],
  dt: ["datetime"],
  id: ["character"],
  ai: ["integer"],
  j: ["varchar"]
};

App.SQLite.Keywords = [
  "ABORT",
  "ACTION",
  "ADD",
  "AFTER",
  "ALL",
  "ALTER",
  "ANALYZE",
  "AND",
  "AS",
  "ASC",
  "ATTACH",
  "AUTOINCREMENT",
  "BEFORE",
  "BEGIN",
  "BETWEEN",
  "BY",
  "CASCADE",
  "CASE",
  "CAST",
  "CHECK",
  "COLLATE",
  "COLUMN",
  "COMMIT",
  "CONFLICT",
  "CONSTRAINT",
  "CREATE",
  "CROSS",
  "CURRENT_DATE",
  "CURRENT_TIME",
  "CURRENT_TIMESTAMP",
  "DATABASE",
  "DEFAULT",
  "DEFERRABLE",
  "DEFERRED",
  "DELETE",
  "DESC",
  "DETACH",
  "DISTINCT",
  "DROP",
  "EACH",
  "ELSE",
  "END",
  "ESCAPE",
  "EXCEPT",
  "EXCLUSIVE",
  "EXISTS",
  "EXPLAIN",
  "FAIL",
  "FOR",
  "FOREIGN",
  "FROM",
  "FULL",
  "GLOB",
  "GROUP",
  "HAVING",
  "IF",
  "IGNORE",
  "IMMEDIATE",
  "IN",
  "INDEX",
  "INDEXED",
  "INITIALLY",
  "INNER",
  "INSERT",
  "INSTEAD",
  "INTERSECT",
  "INTO",
  "IS",
  "ISNULL",
  "JOIN",
  "KEY",
  "LEFT",
  "LIKE",
  "LIMIT",
  "MATCH",
  "NATURAL",
  "NO",
  "NOT",
  "NOTNULL",
  "NULL",
  "OF",
  "OFFSET",
  "ON",
  "OR",
  "ORDER",
  "OUTER",
  "PLAN",
  "PRAGMA",
  "PRIMARY",
  "QUERY",
  "RAISE",
  "RECURSIVE",
  "REFERENCES",
  "REGEXP",
  "REINDEX",
  "RELEASE",
  "RENAME",
  "REPLACE",
  "RESTRICT",
  "RIGHT",
  "ROLLBACK",
  "ROW",
  "SAVEPOINT",
  "SELECT",
  "SET",
  "TABLE",
  "TEMP",
  "TEMPORARY",
  "THEN",
  "TO",
  "TRANSACTION",
  "TRIGGER",
  "UNION",
  "UNIQUE",
  "UPDATE",
  "USING",
  "VACUUM",
  "VALUES",
  "VIEW",
  "VIRTUAL",
  "WHEN",
  "WHERE",
  "WITH",
  "WITHOUT"
];

/**
 * Definitions of functions and operators
 *   expr: implementation of the function/operator
 *   op: true if is an operator
 *   aggr: true if is an aggregated function
 * */
App.SQLite.Functions = {
  // Operators
  "^": {expr: "", op: true},
  "%": {expr: "@1 % @2", op: true},
  "all": {expr: "", op: true},
  "any": {expr: "", op: true},
  //
  // Aggregate functions
  stddev: {expr: "", aggr: true},
  variance: {expr: "", aggr: true},
  //
  // Date functions
  now: {expr: "datetime('now','localtime')"},
  today: {expr: "date('now','localtime')||'00:00:00'"},
  time: {expr: "time('now','localtime')"},
  toDate: {expr: "DATE((@1)||'-'||SUBSTR('0'||(@2),-2)||'-'||SUBSTR('0'||(@3),-2))"},
  toTime: {expr: "DATETIME('1899-12-30'||' '||SUBSTR('0'||(@1),-2)||':'||SUBSTR('0'||(@2),-2)||':'||SUBSTR('0'||(@3),-2))"},
  year: {expr: "CAST(STRFTIME('%Y', @1) as integer)"},
  month: {expr: "CAST(STRFTIME('%m', @1) as integer)"},
  day: {expr: "CAST(STRFTIME('%d', @1) as integer)"},
  hour: {expr: "CAST(STRFTIME('%H', @1) as integer)"},
  minute: {expr: "CAST(STRFTIME('%M', @1) as integer)"},
  second: {expr: "CAST(STRFTIME('%S', @1) as integer)"},
  weekDay: {expr: "CAST(STRFTIME('%w', @1) as integer)"},
  yearDay: {expr: "CAST(STRFTIME('%j', @1) as integer)"},
  week: {expr: "CAST(STRFTIME('%j', @1) as integer)"},
  getDate: {expr: "DATE(@1)"},
  getTime: {expr: "TIME(@1)"},
  //
  // Math functions
  int: {expr: "ROUND(@1)"},
  ceil: {expr: "ROUND((@1)+0.5)"},
  floor: {expr: "ROUND((@1)-0.5)"},
  round: {expr: "ROUND(@1, @2)"},
  trunc: {expr: ""},
  sign: {expr: "CASE WHEN @1>0 THEN 1 WHEN @1<0 THEN -1 ELSE 0 END"},
  sqrt: {expr: ""},
  log: {expr: ""},
  log10: {expr: ""},
  sin: {expr: ""},
  cos: {expr: ""},
  tan: {expr: ""},
  cot: {expr: ""},
  arcSin: {expr: ""},
  arcCos: {expr: ""},
  artTan: {expr: ""},
  //
  // String functions
  left: {expr: "SUBSTR(@1, 1, @2)"},
  right: {expr: "SUBSTR(@1, -(@2))"},
  substr: {expr: "SUBSTR(@1, @2, @3)"},
  spaces: {expr: "SUBSTR(REPLACE(HEX(ZEROBLOB((@1)/2+1)),'0',' '),1,@1)"},
  replicate: {expr: "SUBSTR(REPLACE(HEX(ZEROBLOB((@2)/2+1)),'0',@1),1,@2)"},
  fillLeft: {expr: "SUBSTR(REPLACE(HEX(ZEROBLOB((@2)/2+1)),'0',@3)||@1,-(@2))"},
  fillRight: {expr: "SUBSTR(@1||REPLACE(HEX(ZEROBLOB((@2)/2+1)),'0',@3),1,@2)"},
  //
  // Various function
  nullValue: {expr: "COALESCE(@1, @2)"},
  blobSize: {expr: "length(@1)"}
};


Object.defineProperty(App.SQLite.prototype, "driver", {
  get: function () {
    //
    var d = App.Database.DriverTypes.WebSQL;
    //
    // If the app is inside shell and it can use the native sqlite plugin, we can
    // select which driver we should activate
    if (this.app && this.app.device.sqlite && this.app.device.isInsideShell()) {
      //
      // We should select the native one, if the developer selected it, or if the the shell
      // runs inside a WKWebView that cannot use websql
      if (this.app.device.webview === "WK" || this._driver === App.Database.DriverTypes.Native)
        d = App.Database.DriverTypes.Native;
    }
    //
    return d;
  },
  set: function (newValue) {
    this._driver = newValue;
  }
});

/**
 * Open the connection to the database
 * @param {Function} callback - function to be called at the end
 */
App.SQLite.prototype._openConnection = function (callback)
{
  this.getPool().connect(callback);
};


/**
 * Init the application pool
 */
App.SQLite.prototype.getPool = function () {
  if (!App.SQLite.pools)
    App.SQLite.pools = {};
  //
  var poolName = this.schema.name;
  if (!App.SQLite.pools[poolName]) {
    console.warn(this.schema.name + " use " + this.driver + " driver");
    App.SQLite.pools[poolName] = new App.SQLitePool(this);
  }
  //
  return App.SQLite.pools[poolName];
};


App.SQLitePool = function (db)
{
  this.app = db.app;
  this.driver = db.driver;
  this.name = db.schema.name;
  this.connections = [];
};


/**
 * Open a new connection to the database
 * @param {Function} callback - function to be called at the end
 */
App.SQLitePool.prototype.connect = function (callback) {
  // Get a connection from pool
  var conn = this.connections.shift();
  if (conn) {
    // Stop auto closure and return the connection from pool
    clearTimeout(conn.closeTimer);
    delete conn.closeTimer;
    return callback(conn.connection);
  }
  //
  // Native driver connection
  if (this.driver === App.Database.DriverTypes.Native)
    return this.app.device.sqlite.openDatabase(this.name, callback);
  //
  // WebSQL driver connection
  try {
    var maxsize = 10 * 1024 * 1024;
    var connection = openDatabase(this.name, "", "", maxsize);
    if (!connection)
      throw new Error("Unable to open the database");
    //
    callback(connection);
  }
  catch (e) {
    callback(null, e);
  }
};


/**
 * Relase a connection into pool
 * @param {Object} connection to release
 */
App.SQLitePool.prototype.release = function (connection) {
  var conn = {};
  conn.connection = connection;
  //
  // Program closure of connection
  conn.closeTimer = setTimeout(function () {
    // Remove connection from pool
    this.connections.splice(this.connections.indexOf(conn), 1);
    //
    // Close connection
    if (this.driver === App.Database.DriverTypes.Native)
      this.app.device.sqlite.closeDatabase(connection, new Function());
  }.bind(this), 30000);
  //
  // Put connection into pool
  this.connections.push(conn);
};


/**
 * Close the connection to the database
 * @param {Function} callback - function to be called at the end
 */
App.SQLite.prototype._closeConnection = function (callback)
{
  this.getPool().release(this.connection);
  callback();
};


/**
 * Execute an SQL query
 * @param {String} sql to execute
 * @param {Boolean} readOnly - true if is a SELECT
 * @param {Array} parameters
 * @param {Function} callback - function to be called at the end
 * @return {Object/Integer}
 */
App.SQLite.prototype._executeSQL = function (sql, readOnly, parameters, callback)
{
  if (this.driver === App.Database.DriverTypes.Native) {
    // Await only if out of transaction or readOnly query
    if (readOnly || !this.isInTransaction())
      this.app.device.sqlite.query(this.connection, sql, parameters, callback);
    else {
      this.app.device.sqlite.query(this.connection, sql, parameters);
      callback({affectedRows: 1});
    }
    return;
  }
  //
  // If there is already an open transaction could still not be able
  // to execute the query in it if they are not in the same round of calls
  if (this.isInTransaction()) {
    try {
      // If you are already in transaction invoke the callback immediately after the query
      // else invoke the callback at the end of the transaction
      this.transaction.executeSql(sql, parameters, function (tx, res) {
        callback(res);
      }, function (tx, error) {
        // Query is terminated with error
        callback(null, error);
      });
      return;
    }
    catch (e) {
    }
  }
  //
  // Open transaction
  var r = null;
  this.connection[(readOnly ? "readTransaction" : "transaction")](function (tx) {
    tx.executeSql(sql, parameters, function (tx, res) {
      r = res;
    });
  }, function (err) {
    // Transaction is terminated with error
    callback(null, err);
  }, function () {
    // Transaction is terminated with success
    callback(r);
  });
};


/**
 * Execute an SQL query with offset paging
 * @param {Object} params
 * @param {Function} callback - function to be called at the end
 * @return {DataMap/Object}
 */
App.SQLite.prototype.queryOffsetPaging = function (params, callback)
{
  var sql = "select * from (";
  sql += params.sql;
  sql += ") a limit " + params.limit;
  sql += " offset " + params.offset;
  this.executeSQL(sql, true, params.parameters, callback);
};


/*
 * Read some info from result
 * @param {String} sql
 * @param {Object} result
 */
App.SQLite.prototype._readInfoFromResult = function (sql, result)
{
  if (this.driver === App.Database.DriverTypes.Native)
    this.lastRowsAffected = result.affectedRows;
  else
    this.lastRowsAffected = result.rowsAffected;
  if (sql.trim().substr(0, 6).toLowerCase() === "insert")
    this.lastInsertId = result.insertId;
};


/**
 * Create a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to create
 */
App.SQLite.prototype.createTable = function (sqls, t)
{
  // Getting create table stmt
  var s = "create table " + this.quote(t.name) + " (";
  //
  // Getting fields definition
  for (var i = 0; i < t.fields.length; i++)
    s += (i === 0 ? "" : ",") + "\n  " + this.getFieldDef(t.fields[i]);
  //
  // Creating PK constraint inside table definition
  var pkn = "";
  for (var i = 0; i < t.fields.length; i++) {
    var f = t.fields[i];
    if (f.pk && f.datatype !== App.Database.DataTypes.AUTOINCR)
      pkn += (pkn === "" ? "" : ", ") + this.quote(f.name);
  }
  if (pkn !== "")
    s += ",\n  primary key (" + pkn + ")";
  //
  // Creating FK constraints inside table definition
  if (t.fks) {
    for (var i = 0; i < t.fks.length; i++) {
      var fk = t.fks[i];
      s += ",\n  " + /*"constraint " + fk.name +*/ "foreign key(";
      var fkn = "";
      for (var ref in fk.refs) {
        s += (!fkn ? "" : ",") + this.quote(this.getObject(t.fields, {id: ref}).name);
        fkn += (!fkn ? "" : ",") + fk.refs[ref];
      }
      s += ") references " + fk.t + " (" + fkn + ")";
      s += this.getFKActions(fk);
    }
  }
  //
  s += "\n)";
  //
  sqls.push({sql: s});
  //
  // Return the table with childs
  var nt = {};
  nt.name = t.name;
  nt.fields = t.fields;
  nt.fks = t.fks;
  return nt;
};


/**
 * Modify a table
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} ot old version
 */
App.SQLite.prototype.modifyTable = function (sqls, t, ot)
{
  var ss = [];
  var needRecreate = !this.dropUnuselessTableFields(ss, t, ot);
  if (!needRecreate)
    needRecreate = !this.modifyTableFields(ss, t, ot);
  if (!needRecreate)
    needRecreate = !this.modifyTableFKs(ss, t, ot);
  //
  if (needRecreate)
    this.recreateTable(sqls, t, ot);
  else {
    for (var i = 0; i < ss.length; i++)
      sqls.push(ss[i]);
    if (t.name !== ot.name)
      this.renameTable(sqls, t, ot);
  }
  ot.name = t.name;
  ot.fields = t.fields;
  ot.fks = t.fks;
};


/**
 * Modify a field
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} f field to modify
 * @param {Object} of old version
 */
App.SQLite.prototype.modifyField = function (sqls, t, f, of)
{
  // SQLite not support changes to fields
  return (this.getFieldDef(f) === this.getFieldDef(of) && f.pk === of.pk);
};


/**
 * Drop a field
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of field
 * @param {Object} f field to drop
 */
App.SQLite.prototype.dropField = function (sqls, t, f)
{
  // SQLite not support drop of fields
  return false;
};


/**
 * Create a foreign key
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} fk foreign key to create
 */
App.SQLite.prototype.createFK = function (sqls, t, fk)
{
  // SQLite not support add of foreign keys
  return false;
};


/**
 * Modify a foreign key
 * @param {Array} sqls list of sql statements
 * @param {Object} t table to modify
 * @param {Object} ot table old version
 * @param {Object} fk foreign key to modify
 * @param {Object} ofk old version
 */
App.SQLite.prototype.modifyFK = function (sqls, t, ot, fk, ofk)
{
  // SQLite not support chances to foreign keys
  /*
   if (fk.name !== ofk.name)
   return false;
   */
  //
  if (Object.keys(fk.refs).length !== Object.keys(ofk.refs).length)
    return false;
  //
  for (var ref in fk.refs) {
    if (!ofk.refs[ref])
      return false;
  }
  //
  return true;
};


/**
 * Drop an foreign key
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of foreign key
 * @param {Object} fk foreign key to drop
 */
App.SQLite.prototype.dropFK = function (sqls, t, fk)
{
  // SQLite not support drop of foreign keys
  return false;
};


/**
 * Modify an index
 * @param {Array} sqls list of sql statements
 * @param {Object} t table owner of index
 * @param {Object} idx index to modify
 * @param {Object} oidx old version
 */
App.SQLite.prototype.modifyIndex = function (sqls, t, idx, oidx)
{
  var ok = true;
  if (idx.name !== oidx.name || idx.un !== oidx.un)
    ok = false;
  else if (idx.fields.length !== oidx.fields.length)
    ok = false;
  else {
    for (var i = 0; i < idx.fields.length; i++) {
      if (idx.fields[i] !== oidx.fields[i])
        ok = false;
    }
  }
  //
  if (!ok) {
    this.dropIndex(sqls, t, idx);
    this.createIndex(sqls, t, idx);
  }
  return true;
};


/**
 * Get the definition for the f field
 * @param {Object} f field to define
 */
App.SQLite.prototype.getFieldDef = function (f)
{
  var s = this.quote(f.name) + " " + this.getDataType(f);
  if (f.datatype === App.Database.DataTypes.AUTOINCR)
    s += " primary key autoincrement";
  if (f.notn)
    s += " not null";
  if (this.getFieldDefVal(f))
    s += " default(" + this.getFieldDefVal(f) + ")";
  return s;
};


/**
 * Get the datatype for the f field
 * @param {Object} f
 */
App.SQLite.prototype.getDataType = function (f)
{
  var maxlen = this.getFieldMaxLength(f);
  var scale = this.getFieldScale(f);
  switch (f.datatype) {
    case App.Database.DataTypes.INT:
    {
      if (maxlen < 5)
        return "smallint"; // -32768 to +32767
      else if (maxlen < 10)
        return "integer"; // -2147483648 to +2147483647
      else
        return "bigint"; // -9223372036854775808 to +9223372036854775807
    }

    case App.Database.DataTypes.AUTOINCR:
      return "integer";

    case App.Database.DataTypes.FLOAT:
    {
      if (maxlen < 8)
        return "real"; // 6 decimal digits precision
      else
        return "double precision"; // 15 decimal digits precision
    }

    case App.Database.DataTypes.DECIMAL:
      return "decimal(" + maxlen + "," + scale + ")";

    case App.Database.DataTypes.JSON:
    case App.Database.DataTypes.VARCHAR:
    {
      if (maxlen)
        return "varchar(" + maxlen + ")";
      else
        return "text";
    }

    case App.Database.DataTypes.CHAR:
    case App.Database.DataTypes.UUID:
      return "character(" + maxlen + ")";

    default:
      return this.typemap[f.datatype][0];
  }
};


/**
 * Compose select statement of a query
 * @param {Object} query to compose
 */
App.SQLite.prototype.createSelectStatement = function (query)
{
  // Add HEX function for blob columns
  if (query.schema && query.select) {
    for (var c in query.schema.columns) {
      var dc = query.schema.columns[c];
      if (dc.datatype !== App.Database.DataTypes.BLOB)
        continue;
      var s = dc.getSelectToken();
      var hex = {expr: "if", p: [{expr: "is null", p: [s], op: true}, s, {expr: "HEX", p: [s], funct: true}], alias: this.quote(c)};
      query.select[dc.selectTokenIndex] = hex;
    }
  }
  return App.Database.prototype.createSelectStatement.call(this, query);
};


/**
 * Compose insert statement of a query
 * @param {Object} query to compose
 */
App.SQLite.prototype.createInsertStatement = function (query)
{
  var sql = App.Database.prototype.createInsertStatement.call(this, query);
  //
  if (query.upsert)
    sql = sql.replace("insert into ", "insert or replace into ");
  //
  return sql;
};


/**
 * Compose sql for a simple token
 * @param {Object} token to compose
 */
App.SQLite.prototype.createSqlToken = function (token)
{
  if (token.expr === "true")
    return "1";
  else if (token.expr === "false")
    return "0";
  //
  return App.Database.prototype.createSqlToken.call(this, token);
};


/**
 * Returns the number of rows of a resultSet
 * @param {Object} rs
 */
App.SQLite.prototype.getRowCount = function (rs)
{
  return (rs.rows ? rs.rows.length : 0);
};


/**
 * Returns the i-th row of a resultSet
 * @param {Object} rs - recordset
 * @param {int} pos - position
 * @param {DataSchema} schema
 */
App.SQLite.prototype.getRow = function (rs, pos, schema)
{
  // Rows of resultset of SQLite are readonly
  // I create a copy of it for convert values
  var row = (this.driver === App.Database.DriverTypes.Native ? rs.rows[pos] : rs.rows.item(pos));
  var newrow = {};
  if (schema) {
    for (var c in schema.columns) {
      var v = row[c];
      if (v === undefined)
        continue;
      newrow[c] = this.convertValue(v, schema.columns[c].datatype);
    }
  }
  else {
    for (var c in row)
      newrow[c] = row[c];
  }
  return newrow;
};


/**
 * Returns the field default value
 * @param {Object} f
 */
App.SQLite.prototype.getFieldDefVal = function (f)
{
  if (f.defval) {
    if (f.datatype === App.Database.DataTypes.BOOL)
      return (f.defval === "true" ? "1" : "0");
  }
  return App.Database.prototype.getFieldDefVal.call(this, f);
};


/**
 * Prepare a constant string for SQL
 * @param {Object} v
 * @param {String} dt - dataType
 * @param {Integer} ml - maxLength
 * @param {Integer} sc - scale
 */
App.SQLite.prototype.toSql = function (v, dt, ml, sc)
{
  if (v !== null && v !== undefined) {
    switch (dt) {
      case App.Database.DataTypes.BLOB:
        return "x'" + App.Utils.bufferToHex(v) + "'";
    }
  }
  //
  return App.Database.prototype.toSql.call(this, v, dt, ml, sc);
};


/**
 * Convert a native value into javascript value
 * @param {String} v
 * @param {String} dt
 */
App.SQLite.prototype.convertValue = function (v, dt)
{
  if (v === null || v === undefined)
    return null;
  //
  if (dt === App.Database.DataTypes.BOOL)
    return (v !== 0);
  //
  if (dt === App.Database.DataTypes.JSON)
    return JSON.parse(v);
  //
  if (dt === App.Database.DataTypes.BLOB)
    return App.Utils.hexToBuffer(v);
  //
  return App.Database.prototype.convertValue.call(this, v, dt);
};


/*
 * Begin a transaction
 * @param {Function} callback - function to be called at the end
 */
App.SQLite.prototype._beginTransaction = function (callback)
{
  if (this.driver === App.Database.DriverTypes.Native)
    return this.app.device.sqlite.query(this.connection, "begin", [], callback);
  //
  // Open transaction
  this.connection.transaction(function (tx) {
    // Transaction is opened
    callback(tx);
  }, function (err) {
    // Transaction is terminated with error
    delete this.transaction;
    //
    // If you call the rollback during the transaction proceed
    if (this.rollbackCallback) {
      this.rollbackCallback();
      delete this.rollbackCallback;
    }
  }.bind(this), function () {
    // Transaction is terminated with success
    delete this.transaction;
    //
    // If you call the commit during the transaction proceed
    if (this.commitCallback) {
      this.commitCallback();
      delete this.commitCallback;
    }
  }.bind(this));
};


/*
 * Commit a transaction
 * @param {Function} callback - function to be called at the end
 */
App.SQLite.prototype._commitTransaction = function (callback)
{
  if (this.driver === App.Database.DriverTypes.Native)
    return this.app.device.sqlite.query(this.connection, "commit", [], callback);
  //
  // I sign the callback to be invoked at the end of the transaction
  this.commitCallback = callback;
  delete this.transaction;
};


/*
 * Rollback a transaction
 * @param {Function} callback - function to be called at the end
 */
App.SQLite.prototype._rollbackTransaction = function (callback)
{
  if (this.driver === App.Database.DriverTypes.Native)
    return this.app.device.sqlite.query(this.connection, "rollback", [], callback);
  //
  // I sign the callback to be invoked at the end of the transaction
  this.rollbackCallback = callback;
  //
  // I do a wrong query to cause the rollback of the transaction
  this.transaction.executeSql("rollback");
  delete this.transaction;
};


/*
 * Define a query parameter
 * @param {Object} token
 */
App.SQLite.prototype.defineQueryParameter = function (token)
{
  // BLOB inline
  if (token.dt === App.Database.DataTypes.BLOB)
    return this.toSql(token.value, token.dt);
  //
  this.parameters.push(this.forSql(token.value, token.dt));
  return "?";
};


/*
 * Verify if z_sync table exists
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.SQLite.prototype.syncExistsTable = function (handler, callback)
{
  this.executeSQL("select seq from sqlite_sequence where name = 'z_sync" + handler + "'", true, function (result) {
    App.Database.sequences = App.Database.sequences || {};
    if (result && this.getRowCount(result) === 1) {
      App.Database.sequences[handler] = this.getRow(result, 0).seq;
      callback(true);
    }
    else {
      App.Database.sequences[handler] = 0;
      callback(false);
    }
  }.bind(this));
};


/*
 * Create z_sync table
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.SQLite.prototype.syncCreateTable = function (handler, callback)
{
  var sql = "create table z_sync" + handler + "(id integer primary key autoincrement, uid char(24) unique, device text, sender text, topics text, body text, ts datetime default (datetime(CURRENT_TIMESTAMP,'localtime')), expire datetime)";
  this.executeSQL(sql, false, callback);
};


//export module for node
if (module)
  module.exports = App.SQLite;
