/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */


/* global module */

var App = App || {};


/**
 * @class Stripe
 * Represents an Stripe Object
 */
App.Stripe = function ()
{
  //
  // import modules
  if (module && !App.View) {
    App = require("./requireAll");
    //
    // create stripe object
    this.stripeObj = require("stripe")("");
  }
};


/**
 * authentication service
 * @param {string} privateKey
 */
App.Stripe.prototype.setKey = function (privateKey)
{
  this.stripeObj = require("stripe")(privateKey.trim());
};


/**
 * creates new pay token
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createPayToken = function (options, cb)
{
  this.stripeObj.tokens.create(options || {}, function (err, token) {
    cb(token, err);
  });
};


/**
 * gets an existing pay token
 * @param {string} tokenId
 * @param {function} cb
 */
App.Stripe.prototype.getPayToken = function (tokenId, cb)
{
  this.stripeObj.tokens.create(tokenId, function (err, token) {
    cb(token, err);
  });
};



/**
 * creates new stripe customer
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createCustomer = function (options, cb)
{
  this.stripeObj.customers.create(options || {}, function (err, customer) {
    cb(customer, err);
  });
};


/**
 * gets a specific customer detail
 * @param {string} customerId
 * @param {function} cb
 */
App.Stripe.prototype.getCustomer = function (customerId, cb)
{
  this.stripeObj.customers.retrieve(customerId, function (err, customer) {
    cb(customer, err);
  });
};


/**
 * gets customer's list
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getCustomerList = function (options, cb)
{
  this.stripeObj.customers.list(options || {}, function (err, customers) {
    cb(customers, err);
  });
};


/**
 * update customer
 * @param {string} customerId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateCustomer = function (customerId, newProperties, cb)
{
  this.stripeObj.customers.update(customerId, newProperties || {}, function (err, customer) {
    cb(customer, err);
  });
};


/**
 * delete customer
 * @param {string} customerId
 * @param {function} cb
 */
App.Stripe.prototype.deleteCustomer = function (customerId, cb)
{
  this.stripeObj.customers.del(customerId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * creates new payment source for customer
 * @param {string} customerId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createPaymentSource = function (customerId, options, cb)
{
  this.stripeObj.customers.createSource(customerId, options || {}, function (err, card) {
    cb(card, err);
  });
};


/**
 * gets a specific payment source
 * @param {string} customerId
 * @param {string} sourceId
 * @param {function} cb
 */
App.Stripe.prototype.getPaymentSource = function (customerId, sourceId, cb)
{
  this.stripeObj.customers.retrieveCard(customerId, sourceId, function (err, card) {
    cb(card, err);
  });
};


/**
 * gets the list of payment sources
 * @param {string} customerId
 * @param {function} cb
 */
App.Stripe.prototype.getPaymentSourceList = function (customerId, cb)
{
  this.stripeObj.customers.listCards(customerId, function (err, cards) {
    cb(cards, err);
  });
};


/**
 * update of specific payment source
 * @param {string} customerId
 * @param {string} sourcedId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updatePaymentSource = function (customerId, sourcedId, newProperties, cb)
{
  this.stripeObj.customers.update(customerId, sourcedId, newProperties || {}, function (err, card) {
    cb(card, err);
  });
};


/**
 * delete of specific payment source
 * @param {string} customerId
 * @param {string} sourceId
 * @param {function} cb
 */
App.Stripe.prototype.deletePaymentSource = function (customerId, sourceId, cb)
{
  this.stripeObj.customers.deleteCard(customerId, sourceId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * creates new payment
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createPayment = function (options, cb)
{
  var idempotencyKey = App.Utils.generateUID36();
  this.stripeObj.charges.create(options || {}, {
    idempotency_key: idempotencyKey
  }, function (err, charge) {
    cb(charge, err);
  });
};


/**
 * gets a specific payment
 * @param {string} paymentId
 * @param {function} cb
 */
App.Stripe.prototype.getPayment = function (paymentId, cb)
{
  this.stripeObj.charges.retrieve(paymentId, function (err, charge) {
    cb(charge, err);
  }
  );
};


/**
 * update of specific payment source
 * @param {string} paymentId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updatePayment = function (paymentId, newProperties, cb)
{
  this.stripeObj.charges.update(paymentId, newProperties || {}, function (err, charge) {
    cb(charge, err);
  });
};


/**
 * gets payment's list
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getPaymentList = function (options, cb)
{
  this.stripeObj.charges.list(options || {}, function (err, charge) {
    cb(charge, err);
  });
};


/**
 * confirm a payment
 * @param {string} paymentId
 * @param {function} cb
 */
App.Stripe.prototype.confirmPayment = function (paymentId, cb)
{

  this.stripeObj.charges.capture(paymentId, function (err, charge) {
    cb(charge, err);
  });
};


/**
 * creates new subscription plan
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createPlan = function (options, cb)
{
  this.stripeObj.plans.create(options || {}, function (err, plan) {
    cb(plan, err);
  });
};


/**
 * gets a specific subscription plan
 * @param {string} planId
 * @param {function} cb
 */
App.Stripe.prototype.getPlan = function (planId, cb)
{
  this.stripeObj.plans.retrieve(planId, function (err, plan) {
    cb(plan, err);
  }
  );
};


/**
 * update of subscription plan
 * @param {string} planId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updatePlan = function (planId, newProperties, cb)
{
  this.stripeObj.plans.update(planId, newProperties || {}, function (err, plan) {
    cb(plan, err);
  });
};


/**
 * gets list of plans
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getPlanList = function (options, cb)
{
  this.stripeObj.plans.list(options || {}, function (err, plans) {
    cb(plans, err);
  });
};


/**
 * delete a plan
 * @param {string} planId
 * @param {function} cb
 */
App.Stripe.prototype.deletePlan = function (planId, cb)
{
  this.stripeObj.plans.del(planId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * creates new subscription
 * @param {string} customerId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createSubscription = function (customerId, options, cb)
{
  options = options || {};
  options.customer = customerId;
  //
  this.stripeObj.subscriptions.create(options, function (err, subscription) {
    cb(subscription, err);
  });
};


/**
 * gets a specific subscription
 * @param {string} subscriptionId
 * @param {function} cb
 */
App.Stripe.prototype.getSubscription = function (subscriptionId, cb)
{
  this.stripeObj.subscriptions.retrieve(subscriptionId, function (err, subscription) {
    cb(subscription, err);
  });
};


/**
 * update subscription
 * @param {string} subscriptionId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateSubscription = function (subscriptionId, newProperties, cb)
{
  this.stripeObj.subscriptions.update(subscriptionId, newProperties || {}, function (err, subscription) {
    cb(subscription, err);
  });
};


/**
 * gets list of active customer's subscriptions
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getSubscriptionList = function (options, cb)
{
  this.stripeObj.subscriptions.list(options || {}, function (err, subscription) {
    cb(subscription, err);
  });
};


/**
 * delete a subscription
 * @param {string} subscriptionId
 * @param {boolean} atPeriodEnd
 * @param {function} cb
 */
App.Stripe.prototype.deleteSubscription = function (subscriptionId, atPeriodEnd, cb)
{
  if (typeof atPeriodEnd !== "boolean")
    atPeriodEnd = false;
  //
  this.stripeObj.subscriptions.del(subscriptionId, {at_period_end: atPeriodEnd}, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * creates new invoice
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createInvoice = function (options, cb)
{
  this.stripeObj.invoices.create(options || {}, function (err, invoice) {
    cb(invoice, err);
  });
};


/**
 * creates new invoice item
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createInvoiceItem = function (options, cb)
{
  this.stripeObj.invoiceItems.create(options || {}, function (err, invoiceItem) {
    cb(invoiceItem, err);
  });
};



/**
 * gets a specific invoice
 * @param {string} invoiceId
 * @param {function} cb
 */
App.Stripe.prototype.getInvoice = function (invoiceId, cb)
{
  this.stripeObj.invoices.retrieve(invoiceId, function (err, invoice) {
    cb(invoice, err);
  }
  );
};


/**
 * gets a line items of specific invoice
 * @param {string} invoiceId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getInvoiceLines = function (invoiceId, options, cb)
{
  this.stripeObj.invoices.retrieveLines(invoiceId, options || {}, function (err, lines) {
    cb(lines, err);
  });
};


/**
 * gets upcoming invoice for a specific customer
 * @param {string} customerId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getUpcomingInvoice = function (customerId, options, cb)
{
  this.stripeObj.invoices.retrieveUpcoming(customerId, options || {}, function (err, upcoming) {
    cb(upcoming, err);
  });
};


/**
 * gets a specific invoice item
 * @param {string} invoiceItemId
 * @param {function} cb
 */
App.Stripe.prototype.getInvoiceItem = function (invoiceItemId, cb)
{
  this.stripeObj.invoiceItems.retrieve(invoiceItemId, function (err, invoiceItem) {
    cb(invoiceItem, err);
  }
  );
};


/**
 * update a specific invoice
 * @param {string} invoiceId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateInvoice = function (invoiceId, newProperties, cb)
{
  this.stripeObj.invoices.update(invoiceId, newProperties || {}, function (err, invoice) {
    cb(invoice, err);
  });
};


/**
 * update a specific invoice item
 * @param {string} invoiceItemId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateInvoiceItem = function (invoiceItemId, newProperties, cb)
{
  this.stripeObj.invoiceItems.update(invoiceItemId, newProperties || {}, function (err, invoiceItem) {
    cb(invoiceItem, err);
  });
};


/**
 * gets invoces list
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getInvoiceList = function (options, cb)
{
  this.stripeObj.invoices.list(options || {}, function (err, invoice) {
    cb(invoice, err);
  });
};


/**
 * gets invoces item list
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getInvoiceItemList = function (options, cb)
{
  this.stripeObj.invoiceItems.list(options || {}, function (err, invoiceItem) {
    cb(invoiceItem, err);
  });
};


/**
 * pay a specific invoice
 * @param {string} invoiceId
 * @param {function} cb
 */
App.Stripe.prototype.payInvoice = function (invoiceId, cb)
{
  this.stripeObj.invoices.pay(invoiceId, function (err, invoice) {
    cb(invoice, err);
  });
};


/**
 * delete a specific invoice items
 * @param {string} invoiceItemId
 * @param {function} cb
 */
App.Stripe.prototype.deleteInvoiceItem = function (invoiceItemId, cb)
{
  this.stripeObj.invoiceItems.del(invoiceItemId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * creates new coupon
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createCoupon = function (options, cb)
{
  this.stripeObj.coupons.create(options || {}, function (err, coupon) {
    cb(coupon, err);
  });
};


/**
 * gets a specific coupon
 * @param {string} couponId
 * @param {function} cb
 */
App.Stripe.prototype.getCoupon = function (couponId, cb)
{
  this.stripeObj.coupons.retrieve(couponId, function (err, coupon) {
    cb(coupon, err);
  }
  );
};


/**
 * update a coupon
 * @param {string} couponId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateCoupon = function (couponId, newProperties, cb)
{
  this.stripeObj.coupons.update(couponId, newProperties || {}, function (err, coupon) {
    cb(coupon, err);
  });
};


/**
 * gets list of coupons
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getCouponList = function (options, cb)
{
  this.stripeObj.coupons.list(options || {}, function (err, coupons) {
    cb(coupons, err);
  });
};


/**
 * delete a coupon
 * @param {string} couponId
 * @param {function} cb
 */
App.Stripe.prototype.deleteCoupon = function (couponId, cb)
{
  this.stripeObj.coupons.del(couponId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * delete a customer discount
 * @param {string} customerId
 * @param {function} cb
 */
App.Stripe.prototype.deleteCustomerDiscount = function (customerId, cb)
{
  this.stripeObj.customers.deleteDiscount(customerId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * delete a subscription discount
 * @param {string} subscriptionId
 * @param {function} cb
 */
App.Stripe.prototype.deleteSubscriptionDiscount = function (subscriptionId, cb)
{
  this.stripeObj.subscriptions.deleteDiscount(subscriptionId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * create a refund
 * @param {string} chargeId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createRefund = function (chargeId, options, cb)
{
  if (!options)
    options = {};
  options.charge = chargeId;
  //
  this.stripeObj.refunds.create(options, function (err, refund) {
    cb(refund, err);
  });
};


/**
 * gets a specific refund
 * @param {string} refundId
 * @param {function} cb
 */
App.Stripe.prototype.getRefund = function (refundId, cb)
{
  this.stripeObj.refunds.retrieve(refundId, function (err, refund) {
    cb(refund, err);
  });
};


/**
 * update of refund
 * @param {string} refundId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateRefund = function (refundId, newProperties, cb)
{
  this.stripeObj.refunds.update(refundId, newProperties || {}, function (err, refund) {
    cb(refund, err);
  });
};


/**
 * gets list of refunds
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getRefundsList = function (options, cb)
{
  this.stripeObj.refunds.list(options || {}, function (err, refunds) {
    cb(refunds, err);
  });
};


/**
 * creates Stripe payout
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createPayout = function (options, cb)
{
  this.stripeObj.payouts.create(options || {}, function (err, payout) {
    cb(payout, err);
  });
};


/**
 * gets a specific payout
 * @param {string} payoutId
 * @param {function} cb
 */
App.Stripe.prototype.getPayout = function (payoutId, cb)
{
  this.stripeObj.payouts.retrieve(payoutId, function (err, payout) {
    cb(payout, err);
  });
};


/**
 * update a specific payout
 * @param {string} payoutId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updatePayout = function (payoutId, newProperties, cb)
{
  this.stripeObj.payouts.update(payoutId, newProperties || {}, function (err, payout) {
    cb(payout, err);
  });
};


/**
 * gets payouts list
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getPayoutList = function (options, cb)
{
  this.stripeObj.payouts.list(options || {}, function (err, payouts) {
    cb(payouts, err);
  });
};


/**
 * cancel a specific payout
 * @param {string} payoutId
 * @param {function} cb
 */
App.Stripe.prototype.getPayout = function (payoutId, cb)
{
  this.stripeObj.payouts.close(payoutId, function (err, payout) {
    cb(payout, err);
  });
};


/**
 * creates Stripe connect account
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createAccount = function (options, cb)
{
  this.stripeObj.accounts.create(options || {}, function (err, account) {
    cb(account, err);
  });
};


/**
 * gets Stripe connect account
 * @param {string} accountId
 * @param {function} cb
 */
App.Stripe.prototype.getAccount = function (accountId, cb)
{
  this.stripeObj.accounts.retrieve(accountId, function (err, account) {
    cb(account, err);
  });
};


/**
 * updates Stripe connect account
 * @param {string} accountId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateAccount = function (accountId, newProperties, cb)
{
  this.stripeObj.accounts.update(accountId, newProperties, function (err, account) {
    cb(account, err);
  });
};


/**
 * deletes Stripe connect account
 * @param {string} accountId
 * @param {function} cb
 */
App.Stripe.prototype.deleteAccount = function (accountId, cb)
{
  var response = this.stripeObj.accounts.del(accountId);
  cb(response);
};


/**
 * rejects Stripe connect account
 * @param {string} accountId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.rejectAccount = function (accountId, options, cb)
{
  this.stripeObj.accounts.reject(accountId, options, function (err, account) {
    cb(account, err);
  });
};


/**
 * gets Stripe accounts list
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getAccountsList = function (options, cb)
{
  this.stripeObj.accounts.list(options || {}, function (err, accounts) {
    cb(accounts, err);
  });
};


/**
 * creates Stripe login link
 * @param {string} accountId
 * @param {function} cb
 */
App.Stripe.prototype.createLoginLink = function (accountId, cb)
{
  this.stripeObj.accounts.createLoginLink(accountId, function (err, account) {
    cb(account, err);
  });
};


/**
 * creates an application fee refund
 * @param {string} feeId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createApplicationFeeRefund = function (feeId, options, cb)
{
  this.stripeObj.applicationFees.createRefund(feeId, options || {}, function (err, applicationFee) {
    cb(applicationFee, err);
  });
};


/**
 * retrieves an application fee refund
 * @param {string} refundId
 * @param {string} feeId
 * @param {function} cb
 */
App.Stripe.prototype.getApplicationFeeRefund = function (refundId, feeId, cb)
{
  this.stripeObj.applicationFees.retrieveRefund(refundId, feeId, function (err, refund) {
    cb(refund, err);
  });
};


/**
 * updates an application fee refund
 * @param {string} refundId
 * @param {string} feeId
 * @param {object} metadata
 * @param {function} cb
 */
App.Stripe.prototype.updateApplicationFeeRefund = function (refundId, feeId, metadata, cb)
{
  this.stripeObj.applicationFees.updateRefund(refundId, feeId, metadata || {}, function (err, refund) {
    cb(refund, err);
  });
};


/**
 * lists all application fee refunds
 * @param {string} refundId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getApplicationFeeRefundsList = function (refundId, options, cb)
{
  this.stripeObj.applicationFees.listRefunds(refundId, options || {}, function (err, refunds) {
    cb(refunds, err);
  });
};


/**
 * retrieves an application fee
 * @param {string} feeId
 * @param {function} cb
 */
App.Stripe.prototype.getApplicationFee = function (feeId, cb)
{
  this.stripeObj.applicationFees.retrieve(feeId, function (err, applicationFee) {
    cb(applicationFee, err);
  });
};


/**
 * lists all application fees
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getApplicationFeesList = function (options, cb)
{
  this.stripeObj.applicationFees.list(options || {}, function (err, applicationFees) {
    cb(applicationFees, err);
  });
};


/**
 * retrieves a country spec
 * @param {string} countryCode
 * @param {function} cb
 */
App.Stripe.prototype.getCountrySpec = function (countryCode, cb)
{
  this.stripeObj.countrySpecs.retrieve(countryCode, function (err, countrySpec) {
    cb(countrySpec, err);
  });
};


/**
 * lists country specs
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.getCountrySpecsList = function (options, cb)
{
  this.stripeObj.countrySpecs.list(options || {}, function (err, countrySpecs) {
    cb(countrySpecs, err);
  });
};


/**
 * creates a bank account
 * @param {string} externalAccountId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createBankAccount = function (externalAccountId, options, cb)
{
  this.stripeObj.accounts.createExternalAccount(externalAccountId, options || {}, function (err, bankAccount) {
    cb(bankAccount, err);
  });
};


/**
 * gets a bank account
 * @param {string} externalAccountId
 * @param {string} bankAccountId
 * @param {function} cb
 */
App.Stripe.prototype.getBankAccount = function (externalAccountId, bankAccountId, cb)
{
  this.stripeObj.accounts.retrieveExternalAccount(externalAccountId, bankAccountId, function (err, externalAccount) {
    cb(externalAccount, err);
  });
};


/**
 * updates a bank account
 * @param {string} externalAccountId
 * @param {string} bankAccountId
 * @param {object} metadata
 * @param {function} cb
 */
App.Stripe.prototype.updateBankAccount = function (externalAccountId, bankAccountId, metadata, cb)
{
  this.stripeObj.accounts.updateExternalAccount(externalAccountId, bankAccountId, metadata || {}, function (err, bankAccount) {
    cb(bankAccount, err);
  });
};


/**  DA CORREGGERE
 * deletes a bank account
 * @param {string} externalAccountId
 * @param {string} bankAccountId
 * @param {function} cb
 */
App.Stripe.prototype.deleteBankAccount = function (externalAccountId, bankAccountId, cb)
{
  this.stripeObj.accounts.deleteExternalAccount(externalAccountId, bankAccountId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * lists all bank accounts
 * @param {string} externalAccountId
 * @param {string} options
 * @param {function} cb
 */
App.Stripe.prototype.getBankAccountsList = function (externalAccountId, options, cb)
{
  this.stripeObj.accounts.listExternalAccounts(externalAccountId, options || {}, function (err, bankAccounts) {
    cb(bankAccounts, err);
  });
};


/**
 * creates a card for an account
 * @param {string} externalAccountId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createCardAccount = function (externalAccountId, options, cb)
{
  this.stripeObj.accounts.createExternalAccount(externalAccountId, options || {}, function (err, card) {
    cb(card, err);
  });
};


/**
 * gets a card for an account
 * @param {string} externalAccountId
 * @param {string} cardId
 * @param {function} cb
 */
App.Stripe.prototype.getCardAccount = function (externalAccountId, cardId, cb)
{
  this.stripeObj.accounts.retrieveCard(externalAccountId, cardId, function (err, card) {
    cb(card, err);
  });
};


/**
 * updates a card for an account
 * @param {string} externalAccountId
 * @param {string} cardId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateCardAccount = function (externalAccountId, cardId, newProperties, cb)
{
  this.stripeObj.accounts.updateExternalAccount(externalAccountId, cardId, newProperties || {}, function (err, bankAccount) {
    cb(bankAccount, err);
  });
};


/**
 * deletes a card for an account
 * @param {string} externalAccountId
 * @param {string} cardId
 * @param {function} cb
 */
App.Stripe.prototype.deleteCardAccount = function (externalAccountId, cardId, cb)
{
  this.stripeObj.accounts.deleteExternalAccount(externalAccountId, cardId, function (err, confirmation) {
    cb(confirmation, err);
  });
};


/**
 * lists all card for an account
 * @param {string} externalAccountId
 * @param {string} options
 * @param {function} cb
 */
App.Stripe.prototype.getCardsAccountList = function (externalAccountId, options, cb)
{
  options = options || {};
  options.object = "card";
  //
  this.stripeObj.accounts.listExternalAccounts(externalAccountId, options, function (err, cards) {
    cb(cards, err);
  });
};


/**
 * creates a transfer
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createTransfer = function (options, cb)
{
  this.stripeObj.transfers.create(options || {}, function (err, transfer) {
    cb(transfer, err);
  });
};


/**
 * gets a transfer
 * @param {string} transferId
 * @param {function} cb
 */
App.Stripe.prototype.getTransfer = function (transferId, cb)
{
  this.stripeObj.transfers.retrieve(transferId, function (err, transfer) {
    cb(transfer, err);
  });
};


/**
 * updates a transfer
 * @param {string} transferId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateTransfer = function (transferId, newProperties, cb)
{
  this.stripeObj.transfers.update(transferId, newProperties || {}, function (err, transfer) {
    cb(transfer, err);
  });
};


/**
 * lists all transfers
 * @param {string} options
 * @param {function} cb
 */
App.Stripe.prototype.getTransfersList = function (options, cb)
{
  this.stripeObj.transfers.list(options || {}, function (err, transfers) {
    cb(transfers, err);
  });
};


/**
 * creates a transfer reversal
 * @param {string} transferId
 * @param {object} options
 * @param {function} cb
 */
App.Stripe.prototype.createTransferReversal = function (transferId, options, cb)
{
  this.stripeObj.transfers.createReversal(transferId, options || {}, function (err, reversal) {
    cb(reversal, err);
  });
};


/**
 * gets a transfer reversal
 * @param {string} transferId
 * @param {string} reversalId
 * @param {function} cb
 */
App.Stripe.prototype.getTransferReversal = function (transferId, reversalId, cb)
{
  this.stripeObj.transfers.retrieveReversal(transferId, reversalId, function (err, reversal) {
    cb(reversal, err);
  });
};


/**
 * updates a transfer reversal
 * @param {string} transferId
 * @param {string} reversalId
 * @param {object} newProperties
 * @param {function} cb
 */
App.Stripe.prototype.updateTransferReversal = function (transferId, reversalId, newProperties, cb)
{
  this.stripeObj.transfers.updateReversal(transferId, reversalId, newProperties || {}, function (err, reversal) {
    cb(reversal, err);
  });
};


/**
 * lists all transfers reversal
 * @param {string} transferId
 * @param {string} options
 * @param {function} cb
 */
App.Stripe.prototype.getTransfersReversalsList = function (transferId, options, cb)
{
  this.stripeObj.transfers.listReversals(transferId, options || {}, function (err, reversals) {
    cb(reversals, err);
  });
};


//export module for node
if (module)
  module.exports = App.Stripe;
