/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module, process, Client, Error */

var App = App || {};

/**
 * @class Definition of Sync object
 * @param {Session} app that owns this sync instance
 * */
App.Sync = function (app)
{
  if (module && !App.View) {
    App = require("./requireAll");
    App.io = require("socket.io-client");
  }
  //
  this.app = app;    // App that owns this sync instance
  this.id = App.Utils.generateUID24();
  this.senderId = this.id;
  this.serverUrl = "";   // url of server application to connect
  this.appName = "";     // name of application server for IDE
  this.dataStore = null; // storage of permanent messages
  this.autoConnect = App.Sync.autoConnectTypes.automatic;
  this.maxMessages = 10;
  this.connectionStatus = App.Sync.statusTypes.disconnected;
  this.connectionTimeout = 5000;
  this.autoDisconnectTimeout = 30000;
  this.connectOptions = {};
  this.lastMessageSent = {};
  this.connectCallbacks = [];
  this.disconnectCallbacks = [];
  //
  this.addPlugin(new App.SyncCmd(this));
  this.addPlugin(new App.SyncDO(this));
};


App.Sync.statusTypes = {
  disconnected: "disconnected",
  connecting: "connecting",
  connected: "connected",
  disconnecting: "disconnecting"
};


App.Sync.messageTypes = {
  connect: "connect",
  disconnect: "disconnect",
  connected: "connected",
  msg: "msg",
  ack: "ack",
  sync: "sync",
  syncEnd: "syncEnd",
  topicsChg: "topicsChg"
};


App.Sync.autoConnectTypes = {
  alwaysConnected: 1,
  automatic: 2,
  manual: 3
};


Object.defineProperty(App.Sync.prototype, "enabled", {
  get: function () {
    return (this._enabled === true);
  },
  set: function (newValue) {
    if (newValue !== this._enabled) {
      this._enabled = newValue;
      //
      if (newValue) {
        this.checkConnection();
        //
        // Reconnect when change networkState
        this.deviceObs = function (changes) {
          for (var i = 0; i < changes.length; i++) {
            if (changes[i].name === "networkState") {
              this.checkConnection();
              break;
            }
          }
        }.bind(this);
        Object.observe(this.app.device, this.deviceObs);
      }
      else {
        this.disconnect();
        //
        // Remove observer on device
        if (this.deviceObs) {
          Object.unobserve(this.app.device, this.deviceObs);
          delete this.deviceObs;
        }
      }
    }
  }
});


Object.defineProperty(App.Sync.prototype, "dataStore", {
  get: function () {
    return this._dataStore;
  },
  set: function (newValue) {
    if (typeof newValue === "string")
      this._dataStore = new App[newValue](this.app);
    else
      this._dataStore = newValue;
  }
});


Object.defineProperty(App.Sync.prototype, "topics", {
  get: function () {
    return this._topics;
  },
  set: function (newValue) {
    this._topics = newValue;
    this.sendToProxy(App.Sync.messageTypes.topicsChg, {topics: this._topics});
  }
});


/**
 * Send a message to all
 * @param {Object} message to send
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.sendMessage = function (message, callback)
{
  this.initMessage(message);
  //
  var afterStorage = function (result, error) {
    if (!error)
      this.dispatchMessage(message);
    if (callback)
      callback(null, error);
  }.bind(this);
  //
  // If message is persistent store it
  if (message.ttl > 0)
    this.storeMessage(message, afterStorage);
  else
    afterStorage();
};


/**
 * Dispatch a message to all
 * @param {Object} message to send
 */
App.Sync.prototype.dispatchMessage = function (message)
{
  var handler = message.class || "";
  if (this.isClientProxy()) {
    // Autoconnect if needed
    this.checkConnection();
    //
    if (message.ttl > 0)
      this.sendStoredMessages(handler);
    else
      this.sendToProxy(App.Sync.messageTypes.msg, message);
  }
  else if (App.parentType === "master") {
    if (this.isServerProxy()) {
      // Messages marked as fromProxy should be send only to other proxy sessions
      if (message.fromProxy)
        delete message.fromProxy;
      else// Dispatch message also to the client
        this.sendToProxy(App.Sync.messageTypes.msg, message);
    }
    //
    // Dispatch message to master
    process.send({type: App.msgTypeMap.syncBroad, cnt: {sid: this.app.id, msg: message}});
  }
  else {
    // Dispatch message to all sessions of this app (except me)
    var sessionIds = Object.keys(App.sessionMap);
    for (var i = 0; i < sessionIds.length; i++) {
      var session = App.sessionMap[sessionIds[i]];
      //
      // If in febe mode I want to handle message because
      // sync session and online preview session are merged
      if (session.sync !== this || this.febe)
        session.sync.handleMessage(message);
    }
  }
  //
  // Set client synchronized
  if (this.isServerProxy() && message.ttl > 0)
    this.setSynchronized(handler, message.id);
};


/**
 * Init a message
 * @param {Object} message to init
 */
App.Sync.prototype.initMessage = function (message)
{
  // Set my senderId e uid
  if (!message.sender)
    message.sender = this.senderId;
  if (!message.uid)
    message.uid = App.Utils.generateUID24();
  if (!message.device && this.app.device.uuid)
    message.device = this.app.device.uuid;
};


/**
 * Store a message into datamodel
 * @param {Object} message to store
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.storeMessage = function (message, callback)
{
  var handler = message.class || "";
  var ds = this.getHandlerDataStore(handler);
  if (!ds) {
    callback(null, new Error("dataStore undefined"));
    return;
  }
  //
  ds.syncStoreMessage(handler, message, callback);
};


/**
 * Delete expired messages stored into datastore
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.deleteExpiredMessages = function (handler, callback)
{
  handler = handler || "";
  var ds = this.getHandlerDataStore(handler);
  if (!ds) {
    if (callback)
      callback(null, new Error("dataStore undefined"));
    return;
  }
  //
  ds.syncDeleteMessages(handler, {expired: true}, function (result, error) {
    if (callback)
      callback(null, error);
  });
};


/**
 * Update a message stored into datastore
 * @param {Object} message to update
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.updateStoredMessage = function (message, callback)
{
  var handler = message.class || "";
  var ds = this.getHandlerDataStore(handler);
  if (!ds) {
    if (callback)
      callback(null, new Error("dataStore undefined"));
    return;
  }
  //
  ds.syncUpdateMessage(handler, message, function (result, error) {
    if (callback)
      callback(null, error);
  });
};


/**
 * Load some messages from z_sync table
 * @param {String} handler - handler of message
 * @param {Object} filters
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.loadMessages = function (handler, filters, callback)
{
  handler = handler || "";
  var ds = this.getHandlerDataStore(handler);
  if (!ds)
    return callback(null, new Error("dataStore undefined"));
  //
  ds.syncLoadMessages(handler, filters, function (messages, error) {
    if (error)
      return callback(null, error);
    //
    if (handler) {
      for (var i = 0; i < messages.length; i++)
        messages[i].class = handler;
    }
    callback(messages);
  });
};


/**
 * Send messages stored into datastore
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.sendStoredMessages = function (handler, callback)
{
  handler = handler || "";
  //
  // Autoconnect if needed
  this.checkConnection();
  //
  if (!this.isConnected()) {
    if (callback)
      callback(false);
    return;
  }
  //
  // If plugin is paused nothing to do
  if (handler && (!this[handler] || this[handler].paused)) {
    if (callback)
      callback(false);
    return;
  }
  //
  // If this method is already been invoked nothing to do
  this.sendingStoredMessages = this.sendingStoredMessages || {};
  if (this.sendingStoredMessages[handler]) {
    if (callback)
      callback(false);
    return;
  }
  //
  this.sendingStoredMessages[handler] = true;
  //
  if (!this.lastMessageSent[handler])
    this.lastMessageSent[handler] = 1;
  //
  this.loadMessages(handler, {max: this.maxMessages, fromId: this.lastMessageSent[handler] + 1, expired: false}, function (messages, error) {
    var again = (!error && messages.length === this.maxMessages);
    var nextMessage = function (result, error) {
      if (error) {
        delete this.sendingStoredMessages[handler];
        if (callback)
          callback(null, error);
        return;
      }
      //
      // If in the meantime the connection has closed I finished
      if (!this.isConnected()) {
        delete this.sendingStoredMessages[handler];
        if (callback)
          callback(false);
        return;
      }
      //
      var msg = messages.shift();
      if (msg) {
        if (this.isServerProxy()) {
          // Filter the messages that not matches topics
          var options = {match: App.Sync.topicsMatches(this.topics, msg.topics)};
          this.fireOnMessageFiltering(msg, options, function (result, error) {
            if (error)
              return nextMessage(null, error);
            //
            // In topics don't matches ignore the message (but set client synchronized)
            if (!options.match)
              this.setSynchronized(handler, msg.id);
            else
              this.sendToProxy(App.Sync.messageTypes.msg, msg);
            //
            nextMessage();
          }.bind(this));
        }
        else {
          msg.ack = true;
          //
          this.sendToProxy(App.Sync.messageTypes.msg, msg);
          nextMessage();
        }
        //
        return;
      }
      //
      delete this.sendingStoredMessages[handler];
      //
      // If there could be other messages proceed
      if (again)
        this.sendStoredMessages(handler, callback);
      else {
        // Set client synchronized
        if (this.isServerProxy())
          this.setSynchronized(handler, this.lastMessageSent[handler]);
        //
        if (callback)
          callback(true);
      }
    }.bind(this);
    nextMessage(null, error);
  }.bind(this));
};


/**
 * Read last id and ts from z_sync of client
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 **/
App.Sync.prototype.getLastSynchronization = function (handler, callback)
{
  this.loadMessages(handler, {fromId: 1, toId: 1}, function (messages, error) {
    if (error)
      return callback(null, error);
    //
    if (messages.length !== 1)
      return callback(null, new Error("message with id = 1 not found"));
    //
    callback({id: messages[0].body, ts: messages[0].ts});
  });
};


/**
 * Request messages stored into datastore
 * @param {String} handler - handler of message
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.requestStoredMessages = function (handler, callback)
{
  this.getLastSynchronization(handler, function (ls, error) {
    if (error) {
      if (callback)
        callback(null, error);
      return;
    }
    //
    this.sendToProxy(App.Sync.messageTypes.sync, {from: ls.id, ts: ls.ts, class: handler});
    if (callback)
      callback();
  }.bind(this));
};


/**
 * Handle a message if topics matches
 * @param {Object} message received
 */
App.Sync.prototype.handleMessage = function (message)
{
  var options = {match: App.Sync.topicsMatches(this.topics, message.topics)};
  this.fireOnMessageFiltering(message, options, function (result, error) {
    // In topics don't matches ignore the message
    if (error || !options.match)
      return;
    //
    if (message.class && this[message.class] instanceof App.SyncPlugin)
      this[message.class].handleMessage(message);
    else if (this.isServerProxy() && !message.febe)
      this.sendToProxy(App.Sync.messageTypes.msg, message);
    else
      this.fireOnMessage(message);
  }.bind(this));
};


/**
 * Handle "connect" message (only server side)
 * @param {Object} message received
 */
App.Sync.prototype.handleProxyConnect = function (message)
{
  this.isServer = true;
  this.febe = message.sid.febe;
  this._topics = message.cnt.params.topics;
  this.senderId = message.cnt.params.senderId;
  this.id = message.cnt.params.id;
  this.maxMessages = message.cnt.params.maxMessages;
  this.connectOptions = message.cnt.params.connectOptions;
  //
  this.app.locale.timeZone = message.cnt.params.timeZone;
  this.app.locale.dateFormat = message.cnt.params.dateFormat;
  this.app.locale.timeFormat = message.cnt.params.timeFormat;
  this.app.device.language = message.cnt.params.language;
  this.app.device.uuid = message.cnt.params.uuid;
  //
  this.connectionStatus = App.Sync.statusTypes.connecting;
  //
  // Notify onConnect event
  var options = {cancel: false};
  this.fireOnConnect(options, function (result, error) {
    if (error)
      options.error = error;
    //
    // Tell the sessionID to the client (so that he knows how to comunicate with me)
    options.sidsrv = this.app.id;
    //
    // Send to client proxy "connected" command
    this.sendToProxy(App.Sync.messageTypes.connected, options);
    //
    if (!options.cancel && !options.error)
      this.connectionStatus = App.Sync.statusTypes.connected;
    //
    // Notify onTopicsChanged event
    this.onTopicsChanged();
  }.bind(this));
};


/**
 * Handle "connected" message (only client side)
 * @param {Object} message received
 */
App.Sync.prototype.handleProxyConnected = function (message)
{
  // Stop the timer for detection of connection timeout
  if (this.connectTimeoutTimer) {
    clearTimeout(this.connectTimeoutTimer);
    delete this.connectTimeoutTimer;
  }
  //
  var options = message.cnt.params;
  if (options.cancel || options.error)
    this.disconnect();
  else {
    this.connectionStatus = App.Sync.statusTypes.connected;
    options.success = true;
    //
    // Extract from the message the session ID (id of the server side session
    // that has been created for me)
    this.sessid = options.sidsrv;
    delete options.sidsrv;      // Remove it so that the programmer does not sees it inside the onConnect event
    //
    // Send buffered messages
    if (this.socketBuffer) {
      for (var i = 0; i < this.socketBuffer.length; i++)
        this.sendToProxy(this.socketBuffer[i].id, this.socketBuffer[i].params);
      delete this.socketBuffer;
    }
    //
    this.deleteExpiredMessages();
    this.sendStoredMessages();
    this.requestStoredMessages();
    //
    this.autoDisconnect();
  }
  //
  this.fireOnConnect(options);
};


/**
 * Handle "disconnect" message (only server side)
 * @param {Object} message received
 */
App.Sync.prototype.handleProxyDisconnect = function (message)
{
  this.connectionStatus = App.Sync.statusTypes.disconnected;
  this.onDisconnect();
};


/**
 * Handle "msg" message
 * @param {Object} message received
 */
App.Sync.prototype.handleProxyMessage = function (message)
{
  var msg = message.cnt.params;
  //
  // Save febe info into msg, otherwise I loose this info. msg.febe is used in handleMessage
  // to force fire on message even if this is a server proxy when I'm in febe mode
  if (message.sid.febe)
    msg.febe = true;
  //
  // Reconvert ts as Date
  if (msg.ts)
    msg.ts = new Date(msg.ts);
  //
  // Mark the message as coming from the proxy
  msg.fromProxy = true;
  //
  // If message has a class and I have a plugin with the same name
  // do handle message to plugin
  if (msg.class && this[msg.class] instanceof App.SyncPlugin)
    this[msg.class].onProxyMessage(msg);
  else if (this.isClientProxy())
    this.fireOnMessage(msg);
  else
    this.sendMessage(msg);
  //
  // If message require acnknowledgement send it
  if (msg.ack)
    this.sendToProxy(App.Sync.messageTypes.ack, msg.uid);
};


/**
 * Handle "ack" message
 * @param {Object} message received
 */
App.Sync.prototype.handleProxyMessageAcknowledgement = function (message)
{
  if (!this.messagesToAck)
    return;
  //
  var msg = this.messagesToAck[message.cnt.params];
  if (!msg)
    return;
  //
  delete this.messagesToAck[message.cnt.params];
  //
  // If message is permanent delete it
  var handler = msg.class || "";
  if (msg.ttl > 0 && this.isClientProxy())
    this.getHandlerDataStore(handler).syncDeleteMessages(handler, {uid: msg.uid}, function () {
    });
};


/**
 * Handle "sync" message (only server side)
 * @param {Object} message received
 */
App.Sync.prototype.handleProxySync = function (message)
{
  this.deleteExpiredMessages(handler);
  //
  var msg = message.cnt.params;
  var handler = msg.class || "";
  if (!this.lastMessageSent[handler])
    this.lastMessageSent[handler] = msg.from;
  //
  // Now plugin is not paused
  if (handler && this[handler])
    this[handler].paused = false;
  //
  this.sendStoredMessages(handler);
};


/**
 * Handle "syncEnd" message (only client side)
 * @param {Object} message received
 */
App.Sync.prototype.handleProxySyncEnd = function (message)
{
  this.setSynchronized(message.cnt.params.class, message.cnt.params.lastId);
};


/**
 * Handle "topicsChanged" message (only server side)
 * @param {Object} message received
 */
App.Sync.prototype.handleProxyTopicsChanged = function (message)
{
  var oldTopics = this._topics;
  this._topics = message.cnt.params.topics;
  this.onTopicsChanged(oldTopics);
};


/**
 * Update first record of z_sync with lastID of server
 * @param {String} handler
 * @param {Integer} lastId
 * @param {Function} callback - function to be called at the end
 **/
App.Sync.prototype.setSynchronized = function (handler, lastId, callback)
{
  if (this.isClientProxy()) {
    var msg = {};
    msg.id = 1;
    msg.ts = new Date();
    msg.class = handler;
    msg.body = lastId;
    //
    // Update first record of z_sync
    this.updateStoredMessage(msg, function (result, error) {
      if (handler)
        this[handler].onSynchronized();
      if (callback)
        callback(null, error);
    }.bind(this));
    return;
  }
  //
  if (this.isServerProxy()) {
    if (lastId && lastId >= this.lastMessageSent[handler]) {
      // Set last message sent
      this.lastMessageSent[handler] = lastId;
      //
      // Reset timer
      if (!this.synchronizedTimeout)
        this.synchronizedTimeout = {};
      else if (this.synchronizedTimeout[handler])
        clearTimeout(this.synchronizedTimeout[handler]);
      //
      // Set client synchronized (timerized)
      this.synchronizedTimeout[handler] = setTimeout(function () {
        this.sendToProxy(App.Sync.messageTypes.syncEnd, {class: handler, lastId: lastId});
        delete this.synchronizedTimeout[handler];
      }.bind(this), 200);
    }
  }
  //
  if (callback)
    callback();
};


/**
 * Connect to server
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.connect = function (callback)
{
  // Can connect only in some cases
  if (!this.isClientProxy() || !this.enabled || this.isOffline()) {
    if (callback)
      callback(false);
    return;
  }
  //
  if (callback) {
    if (this.connectionStatus === App.Sync.statusTypes.connected)
      return callback(true);
    //
    this.connectCallbacks.push(callback);
  }
  //
  // Just connected -> do nothing
  if (this.isConnected())
    return;
  //
  // Open socket connection
  var options = {};
  options.forceNew = true;
  options.reconnection = false;
  options.timeout = (this.connectionTimeout > 0 ? this.connectionTimeout : 9999999);
  this.socket = (module ? App.io : io)(this.serverUrl, options);
  //
  // Send command for connect
  var params = {};
  params.topics = this.topics;
  params.senderId = this.senderId;
  params.id = this.id;
  params.maxMessages = this.maxMessages;
  params.connectOptions = this.connectOptions;
  //
  params.timeZone = this.app.locale.timeZone;
  params.dateFormat = this.app.locale.dateFormat;
  params.timeFormat = this.app.locale.timeFormat;
  params.language = this.app.device.language;
  //
  if (this.app.device.isInsideShell()) {
    if (this.app.device.uuid)
      params.uuid = this.app.device.uuid;
    else
      this.app.dttobj.addWarning("", "Device uuid is undefined upon connection");
  }
  //
  this.sendToProxy(App.Sync.messageTypes.connect, params);
  this.connectionStatus = App.Sync.statusTypes.connecting;
  //
  this.socket.on("connect", function () {
    this.autoReconnect(true);
    //
    // Set a timer for detect connection timeout
    if (this.connectionTimeout > 0) {
      this.connectTimeoutTimer = setTimeout(function () {
        delete this.connectTimeoutTimer;
        this.disconnect();
        this.fireOnConnect({timeout: true});
      }.bind(this), this.connectionTimeout);
    }
  }.bind(this));
  //
  this.socket.on("connect_error", function (data) {
    this.connectionClosed();
    this.fireOnConnect({error: data});
  }.bind(this));
  //
  this.socket.on("connect_timeout", function () {
    this.connectionClosed();
    this.fireOnConnect({timeout: true, attempts: this.autoReconnectAttempts});
  }.bind(this));
  //
  this.socket.on("disconnect", function () {
    this.autoReconnect(true);
    this.connectionClosed();
    this.fireOnDisconnect();
  }.bind(this));
  //
  this.socket.on("sync", function (msg) {
    this.onProxyMessage(msg);
  }.bind(this));
};


/**
 * Disconnect from server
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.disconnect = function (callback)
{
  // Reset auto-reconnect timeout
  this.autoReconnect(true);
  //
  if (callback) {
    if (this.connectionStatus === App.Sync.statusTypes.disconnected)
      return callback();
    //
    this.disconnectCallbacks.push(callback);
  }
  //
  if (!this.isConnected())
    return;
  //
  this.connectionStatus = App.Sync.statusTypes.disconnecting;
  if (this.isClientProxy())
    this.socket.disconnect();
  else if (this.isServerProxy())
    this.sendToProxy(App.Sync.messageTypes.disconnect);
};


/**
 * Called when connection is closed
 */
App.Sync.prototype.connectionClosed = function ()
{
  this.connectionStatus = App.Sync.statusTypes.disconnected;
  delete this.socket;
  delete this.socketBuffer;
  this.sendingStoredMessages = {};
  this.lastMessageSent = {};
  delete this.messagesToAck;
  this.autoReconnect();
};


/**
 * Send a message to proxy
 * @param {String} id
 * @param {Object} params
 */
App.Sync.prototype.sendToProxy = function (id, params)
{
  if (this.isClientProxy()) {
    if (!this.isConnected() && id !== App.Sync.messageTypes.connect)
      return;
    //
    // If connecting buffer message; will send it when connected
    if (this.connectionStatus === App.Sync.statusTypes.connecting) {
      this.socketBuffer = this.socketBuffer || [];
      this.socketBuffer.push({id: id, params: params});
    }
    else {
      var sidcli = (module ? undefined : Client.Utils.getQueryVariable("sid"));
      this.socket.emit(App.msgTypeMap.sync, {sid: {sidcli: sidcli, sidsrv: this.sessid, appname: this.appName}, cnt: {id: id, params: params}});
      this.autoDisconnect();
    }
  }
  else if (this.isServerProxy())
    process.send({type: App.msgTypeMap.sync, sid: {sidsrv: this.app.id}, cnt: {id: id, params: params}});
  //
  if (id === App.Sync.messageTypes.msg && params.ttl > 0) {
    // Set client synchronized
    var handler = params.class || "";
    if (this.isServerProxy())
      this.setSynchronized(handler, params.id);
    else
      this.lastMessageSent[handler] = params.id;
  }
  //
  // Store message to acknowledge
  if (id === App.Sync.messageTypes.msg && params.ack) {
    if (!this.messagesToAck)
      this.messagesToAck = {};
    this.messagesToAck[params.uid] = params;
  }
};


/**
 * Handler of message reveived from proxy
 * @param {Object} message received
 */
App.Sync.prototype.onProxyMessage = function (message)
{
  this.autoDisconnect();
  //
  switch (message.cnt.id) {
    case App.Sync.messageTypes.connect:
      this.handleProxyConnect(message);
      break;

    case App.Sync.messageTypes.disconnect:
      this.handleProxyDisconnect(message);
      break;

    case App.Sync.messageTypes.connected:
      this.handleProxyConnected(message);
      break;

    case App.Sync.messageTypes.msg:
      this.handleProxyMessage(message);
      break;

    case App.Sync.messageTypes.ack:
      this.handleProxyMessageAcknowledgement(message);
      break;

    case App.Sync.messageTypes.sync:
      this.handleProxySync(message);
      break;

    case App.Sync.messageTypes.syncEnd:
      this.handleProxySyncEnd(message);
      break;

    case App.Sync.messageTypes.topicsChg:
      this.handleProxyTopicsChanged(message);
      break;
  }
};


/**
 * Test if is a client synchronizer
 */
App.Sync.prototype.isClientProxy = function ()
{
  return (this.serverUrl ? true : false);
};


/**
 * Test if is a server synchronizer
 */
App.Sync.prototype.isServerProxy = function ()
{
  return this.isServer;
};


/**
 * Check if two topics matches
 * @param {Object} senTopics
 * @param {Object} recTopics
 */
App.Sync.topicsMatches = function (senTopics, recTopics)
{
  // if sender topics is an Array
  if (senTopics instanceof Array) {
    // matches if al least one item matches
    for (var i = 0; i < senTopics.length; i++) {
      if (App.Sync.topicsMatches(senTopics[i], recTopics))
        return true;
    }
    return false;
  }
  else if (recTopics instanceof Array) { // if reveiver topics is an Array
    // matches if al least one item matches
    for (var i = 0; i < recTopics.length; i++) {
      if (App.Sync.topicsMatches(senTopics, recTopics[i]))
        return true;
    }
    return false;
  }
  else if (senTopics && typeof senTopics === "object") { // if sender topics is an Object
    if (!(senTopics && typeof recTopics === "object")) // if receiver topics isn't an Object
      return false; // no matches
    //
    // All keys must matches (also recursively)
    var senKeys = Object.keys(senTopics);
    var recKeys = Object.keys(recTopics);
    if (senKeys.length !== recKeys.length)
      return false;
    //
    for (var i = 0; i < senKeys.length; i++) {
      var topicKey = senKeys[i];
      if (recKeys.indexOf(topicKey) === -1 || !App.Sync.topicsMatches(senTopics[topicKey], recTopics[topicKey]))
        return false;
    }
    //
    return true;
  }
  else { // sender topic is a simple value (string, number, boolean)
    // if receiver topics is an Object
    if (recTopics && typeof recTopics === "object")
      return false; // no matches
    else { // receiver topic is a simple value (string, number, boolean)
      // if topics of sender or of receiver is "*" it matches
      if (senTopics === "*" || recTopics === "*")
        return true;
      //
      // matches if are equals
      return (senTopics === recTopics);
    }
  }
};


/**
 * Fired the event onMessage
 * @param {Object} message received
 */
App.Sync.prototype.fireOnMessage = function (message)
{
  this.onMessage(message);
};


/**
 * Event fired when a message arrived
 * @param {Object} message received
 */
App.Sync.prototype.onMessage = function (message)
{
};


/**
 * Fires the event onConnect
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.fireOnConnect = function (options, callback)
{
  // Fire event to me and to all my plugins
  var plugins = this.plugins;
  var next = function () {
    if (plugins.length > 0)
      plugins.shift().fireOnConnect(options, next);
    else {
      // Call all connect callbacks
      if (!options.timeout || !options.attempts) {
        while (this.connectCallbacks.length > 0)
          this.connectCallbacks.shift()(options.success);
      }
      //
      if (callback)
        callback();
    }
  }.bind(this);
  this.onConnect(options, next);
};


/**
 * Event fired when the connection has been established
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.onConnect = function (options, callback)
{
  if (callback)
    callback();
};


/**
 * Fires the event onDisconnect
 */
App.Sync.prototype.fireOnDisconnect = function ()
{
  // Fire event to me
  this.onDisconnect();
  //
  // and to all my plugins
  var plugins = this.plugins;
  for (var i = 0; i < plugins.length; i++)
    plugins[i].fireOnDisconnect();
  //
  // Call all disconnect callbacks
  while (this.disconnectCallbacks.length > 0)
    this.disconnectCallbacks.shift()();
};


/**
 * Event fired when the connection is closed
 */
App.Sync.prototype.onDisconnect = function ()
{
};


/**
 * Raised when the application enters background
 */
App.Sync.prototype.onPause = function ()
{
  // Close channel (only if connection type is alwaysConnected)
  // with a timer after 20 sec if inside shell, after 5 min if out of shell
  if (this.autoConnect === App.Sync.autoConnectTypes.alwaysConnected && this.enabled && !this.wasEnabled) {
    var timeout = (this.app.device.isInsideShell() ? 20 : 5 * 60) * 1000;
    this.wasEnabled = setTimeout(function () {
      clearTimeout(this.wasEnabled);
      this.enabled = false;
    }.bind(this), timeout);
  }
};


/**
 * Raised when the application exits from background
 */
App.Sync.prototype.onResume = function ()
{
  // Reopen channel (only if connection type is alwaysConnected)
  if (this.autoConnect === App.Sync.autoConnectTypes.alwaysConnected && this.wasEnabled) {
    clearTimeout(this.wasEnabled);
    delete this.wasEnabled;
    this.enabled = true;
  }
};


/**
 * Test if status is connected or connecting
 */
App.Sync.prototype.isConnected = function ()
{
  if (this.connectionStatus === App.Sync.statusTypes.connected)
    return true;
  if (this.connectionStatus === App.Sync.statusTypes.connecting)
    return true;
  else
    return false;
};


/**
 * Auto disconnection handler
 */
App.Sync.prototype.autoDisconnect = function ()
{
  // If there are no conditions I do not even bother
  if (!this.isConnected() || !this.isClientProxy() || this.autoConnect === App.Sync.autoConnectTypes.alwaysConnected)
    return;
  //
  // I plan disconnection
  if (this.autoDisconnectTimer)
    clearTimeout(this.autoDisconnectTimer);
  this.autoDisconnectTimer = setTimeout(function () {
    delete this.autoDisconnectTimer;
    this.disconnect();
  }.bind(this), this.autoDisconnectTimeout);
};


/**
 * Auto reconnection handler
 * @param {Boolean} disable - set for disable the handler
 */
App.Sync.prototype.autoReconnect = function (disable)
{
  // If I've reached a stop limit groped reconnection
  if (disable || this.autoReconnectTimeout === 512000) {
    if (this.autoReconnectTimer) {
      clearTimeout(this.autoReconnectTimer);
      delete this.autoReconnectTimer;
    }
    delete this.autoReconnectTimeout;
    delete this.autoReconnectAttempts;
    return;
  }
  //
  // If there are no conditions I do not even bother
  if (this.isConnected() || !this.enabled || !this.isClientProxy() || (this.autoConnect !== App.Sync.autoConnectTypes.alwaysConnected && !this.autoReconnectTimeout))
    return;
  //
  // Initialize or increase the reconnection timeout
  if (this.autoReconnectTimeout)
    this.autoReconnectTimeout *= 2;
  else
    this.autoReconnectTimeout = 125;
  //
  // I plan on trying to reconnect
  if (this.autoReconnectTimer)
    clearTimeout(this.autoReconnectTimer);
  this.autoReconnectTimer = setTimeout(function () {
    delete this.autoReconnectTimer;
    if (this.autoReconnectAttempts)
      this.autoReconnectAttempts += 1;
    else
      this.autoReconnectAttempts = 1;
    this.connect();
  }.bind(this), this.autoReconnectTimeout);
};


/**
 * Test if there is a network connection
 */
App.Sync.prototype.isOffline = function ()
{
  return (["unknown", "none"].indexOf(this.app.device.networkState) >= 0);
};


/**
 * Add a plugin
 * @param {Object} plugin to add
 */
App.Sync.prototype.addPlugin = function (plugin)
{
  this[plugin.id] = plugin;
};


Object.defineProperty(App.Sync.prototype, "plugins", {
  get: function () {
    var plugins = [];
    var keys = Object.keys(this);
    for (var i = 0; i < keys.length; i++) {
      if (this[keys[i]] instanceof App.SyncPlugin)
        plugins.push(this[keys[i]]);
    }
    return plugins;
  }
});


/**
 * Return the DataStore for an handler
 * @param {String} handler
 */
App.Sync.prototype.getHandlerDataStore = function (handler)
{
  if (!handler)
    return this.dataStore;
  else if (this[handler])
    return this[handler].dataStore;
};


/**
 * Check if can connect
 */
App.Sync.prototype.checkConnection = function ()
{
  if (this.autoConnect !== App.Sync.autoConnectTypes.manual)
    this.connect();
};


/**
 * Returns the number of message stored
 * @param {String} handler
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.storedMessageCount = function (handler, callback)
{
  this.getHandlerDataStore(handler).syncStoredMessageCount(handler, callback);
};


/**
 * Fire onMessageFiltering event
 * @param {Object} message
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.fireOnMessageFiltering = function (message, options, callback)
{
  if (message.class && this[message.class] instanceof App.SyncPlugin)
    this[message.class].fireOnMessageFiltering(message, options, callback);
  else
    this.onMessageFiltering(message, options, callback);
};


/**
 * onMessageFiltering event
 * @param {Object} message
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.Sync.prototype.onMessageFiltering = function (message, options, callback)
{
  callback();
};


/**
 * onTopicsChanged event
 * @param {Object} oldTopics
 */
App.Sync.prototype.onTopicsChanged = function (oldTopics)
{
};


/**
 * Serialize an error
 * @param {AnyType} error to serialize
 */
App.Sync.serializeError = function (error)
{
  var o = {};
  if (typeof error === "string")
    o.message = error;
  else if (error instanceof Error) {
    o.message = error.message;
    o.stack = error.stack;
  }
  else
    o = error;
  return o;
};


/**
 * Deserialize an error
 * @param {Object} error to deserialize
 */
App.Sync.deserializeError = function (error)
{
  var o = error;
  if (error.message) {
    o = new Error(error.message);
    o.stack = error.stack;
  }
  return o;
};


/**
 * Serialize an object
 * @param {AnyType} obj to serialize
 */
App.Sync.prototype.serializeObject = function (obj)
{
  var opts = {};
  opts.skipStatus = false;
  opts.skipErrors = false;
  opts.skipType = false;
  opts.skipOriginals = false;
  opts.asString = false;
  //
  var o = {};
  if (obj instanceof App.Collection) {
    o.data = obj.intoJSON(opts);
    o.type = "coll";
  }
  else if (obj instanceof App.Document) {
    o.data = obj.intoJSON(opts);
    o.type = "doc";
  }
  else if (obj instanceof App.DataMap) {
    o.data = obj;
    o.type = "dm";
  }
  else if (obj instanceof Array) {
    o = [];
    for (var i = 0; i < obj.length; i++)
      o.push(this.serializeObject(obj[i]));
  }
  else if (obj instanceof Date) {
    o.data = obj.toISOString();
    o.type = "dt";
  }
  else
    o = obj;
  //
  return o;
};


/**
 * Deserialize an object
 * @param {Object} obj to deserialize
 */
App.Sync.prototype.deserializeObject = function (obj)
{
  if (!obj)
    return obj;
  //
  switch (obj.type) {
    case "coll":
      // If object is a Collection load it from JSON
      var coll = new App.Collection(this.app);
      coll.loadFromJSON(obj.data);
      return coll;
      break;

    case "doc":
      // If object is a Document load it from JSON
      return App.Document.fromJSON(this.app, obj.data);

    case "dm":
      // If object is a DataMap load it from JSON
      var dm = new App.DataMap();
      dm.app = this.app;
      dm.loadFromJSON(obj.data);
      return dm;

    case "dt":
      // If object is a Date load it from JSON
      return new Date(obj.data);

    default:
      var obj = (typeof obj === "object" && "data" in obj ? obj.data : obj);
      if (obj instanceof Array) {
        // If object is an array deserialize each elements
        var a = [];
        for (var i = 0; i < obj.length; i++)
          a.push(this.deserializeObject(obj[i]));
        return a;
      }
      return obj;
  }
};


//export module for node
if (module)
  module.exports = App.Sync;
