/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module */

var App = App || {};

if (module)
  App.SyncPlugin = require("./syncPlugin");

/**
 * @class Definition of SyncCmd object
 * @param {Sync} sync that owns this instance
 * */
App.SyncCmd = function (sync)
{
  if (module && !App.View)
    App = require("./requireAll");
  //
  App.SyncPlugin.call(this, sync);
  //
  this.id = "cmd";
  this.commandTimeout = 30000;
};

// Make App.SyncCmd extend App.SyncPlugin
App.SyncCmd.prototype = new App.SyncPlugin(true);


App.SyncCmd.messageTypes = {
  cmd: "cmd",
  res: "res"
};


/**
 * Send a command to server
 * @param {Object} command to send
 * @param {Function} callback - function to be called at the end
 */
App.SyncCmd.prototype.sendCommand = function (command, callback)
{
  if (!this.sync.enabled)
    return callback(null, new Error("Sync is disabled"));
  //
  var msg = {};
  msg.type = App.SyncCmd.messageTypes.cmd;
  msg.cmd = command;
  msg.class = this.id;
  //
  // Send the command
  this.sync.sendMessage(msg);
  //
  // Set callback to call when response
  if (callback) {
    this.callbacks = this.callbacks || {};
    this.callbacks[msg.uid] = {callback: callback};
    //
    // Set a timer for detect command timeout
    if (this.commandTimeout > 0) {
      this.callbacks[msg.uid].timeoutTimer = setTimeout(function () {
        delete this.callbacks[msg.uid];
        //
        // Invoke callback with timeout error
        callback(null, new Error("Command timeout after " + this.commandTimeout + " ms"));
      }.bind(this), this.commandTimeout);
    }
  }
};


/**
 * Handler of message reveived from proxy
 * @param {Object} message received
 */
App.SyncCmd.prototype.onProxyMessage = function (message)
{
  switch (message.type) {
    case App.SyncCmd.messageTypes.cmd:
      this.handleProxyCommand(message);
      break;

    case App.SyncCmd.messageTypes.res:
      this.handleProxyCommandResponse(message);
      break;
  }
};


/**
 * Event fired when the connection has been established
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.SyncCmd.prototype.onConnect = function (options, callback)
{
  if (!options.success)
    this.connectionClosed();
  //
  if (callback)
    callback();
};


/**
 * Event fired when the connection is closed
 */
App.SyncCmd.prototype.onDisconnect = function ()
{
  this.connectionClosed();
};


/**
 * Called when connection is closed
 */
App.SyncCmd.prototype.connectionClosed = function ()
{
  if (this.callbacks) {
    var uids = Object.keys(this.callbacks);
    for (var i = 0; i < uids.length; i++) {
      var obj = this.callbacks[uids[i]];
      //
      // Cancel command timeout timer
      if (obj.timeoutTimer)
        clearTimeout(obj.timeoutTimer);
      //
      // Invoke callback with error
      obj.callback(null, new Error("Connection closed"));
    }
    delete this.callbacks;
  }
};


/**
 * Handle "cmd" message
 * @param {Object} message received
 */
App.SyncCmd.prototype.handleProxyCommand = function (message)
{
  var response = {};
  this.fireOnCommand(message.cmd, response, function (result, error) {
    var msg = {};
    msg.type = App.SyncCmd.messageTypes.res;
    msg.cmdId = message.uid;
    msg.class = this.id;
    //
    if (response.error || error)
      msg.error = App.Sync.serializeError(response.error || error);
    else {
      msg.result = this.sync.serializeObject(response.result || result);
      if (msg.result && msg.result.data === undefined)
        msg.result = {data: msg.result};
    }
    //
    this.sync.sendToProxy(App.Sync.messageTypes.msg, msg);
  }.bind(this));
};


/**
 * Handle "cmdRes" message
 * @param {Object} message received
 */
App.SyncCmd.prototype.handleProxyCommandResponse = function (message)
{
  if (!this.callbacks || !this.callbacks[message.cmdId])
    return;
  //
  var obj = this.callbacks[message.cmdId];
  delete this.callbacks[message.cmdId];
  //
  // Cancel command timeout timer
  if (obj.timeoutTimer)
    clearTimeout(obj.timeoutTimer);
  //
  // Invoke callback with error or result
  if (message.error)
    obj.callback(null, App.Sync.deserializeError(message.error));
  else
    obj.callback(this.sync.deserializeObject(message.result));
};


/**
 * Fired the event onCommand
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncCmd.prototype.fireOnCommand = function (command, response, callback)
{
  this.onCommand(command, response, callback);
};


/**
 * Event fired when a command arrived
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncCmd.prototype.onCommand = function (command, response, callback)
{
  if (callback)
    callback();
};


//export module for node
if (module)
  module.exports = App.SyncCmd;
