/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module */

var App = App || {};

if (module)
  App.SyncPlugin = require("./syncPlugin");

/**
 * @class Definition of SyncDO object
 * @param {Sync} sync that owns this instance
 * */
App.SyncDO = function (sync)
{
  if (module && !App.View)
    App = require("./requireAll");
  //
  App.SyncPlugin.call(this, sync);
  //
  this.id = "DO";
  this.ttl = 60 * 60 * 24 * 30; // 30 days
  this.app = sync.app;
  this.notifyDocUpdates = false;
  //
  this.initPlugins();
};

// Make App.SyncDO extend App.SyncPlugin
App.SyncDO.prototype = new App.SyncPlugin(true);


App.SyncDO.messageTypes = {
  error: "error",
  docUpdate: "docUpdate"
};


App.SyncDO.commandTypes = {
  resyncRequired: "resyncRequired",
  resyncClass: "resyncClass",
  lastId: "lastId",
  remoteLoad: "remoteLoad",
  remoteSave: "remoteSave",
  remoteCall: "remoteCall"
};

// List of classes to be synchronized
App.SyncDO.docToSync = [];

/**
 * Init private plugins
 */
App.SyncDO.prototype.initPlugins = function ()
{
  this.cmd = new App.SyncCmd(this.sync);
  this.cmd.id = "cmdDO";
  this.sync.addPlugin(this.cmd);
  //
  this.cmd.onCommand = function (command, response, callback) {
    this.onCommand(command, response, callback);
  }.bind(this);
};


Object.defineProperty(App.SyncDO.prototype, "commandTimeout", {
  get: function () {
    return this.cmd.commandTimeout;
  },
  set: function (newValue) {
    this.cmd.commandTimeout = newValue;
  }
});


/**
 * Handler of message reveived from proxy
 * @param {Object} message received
 */
App.SyncDO.prototype.onProxyMessage = function (message)
{
  switch (message.type) {
    case App.SyncDO.messageTypes.error:
      this.handleProxyError(message);
      break;

    case App.SyncDO.messageTypes.docUpdate:
      this.handleDocUpdate(message);
      break;

    default:
      this.handleProxyVariation(message);
      break;
  }
};


/**
 * Handle a message received from another session
 * @param {Object} message received
 */
App.SyncDO.prototype.handleMessage = function (message)
{
  switch (message.type) {
    case App.SyncDO.messageTypes.error:
      break;

    case App.SyncDO.messageTypes.docUpdate:
      if (!this.sync.isServerProxy() && !this.sync.isClientProxy())
        this.handleDocUpdate(message);
      break;

    default:
      if (this.sync.isServerProxy())
        this.sync.sendToProxy(App.Sync.messageTypes.msg, message);
      break;
  }
};


/**
 * Fires the event onConnect
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.fireOnConnect = function (options, callback)
{
  options.skip = false;
  this.log("fired onConnect event");
  this.onConnect(options, function (result, error) {
    if (this.sync.isClientProxy())
      this.log("connection " + (options.success ? "established successfully" : "refused by server"));
    if (options.skip)
      this.log("synchronization skipped");
    if (options.skip || !options.success)
      return callback();
    //
    if (this.sync.isServerProxy())
      this.paused = true;
    else
      this.startSynchronization();
    callback();
  }.bind(this));
};


/**
 * Start synchronization
 */
App.SyncDO.prototype.startSynchronization = function ()
{
  this.log("start synchronization");
  //
  // Delete expired messages
  this.deleteExpiredMessages(function (result, error) {
    if (error)
      return this.fireOnError({error: error});
    //
    // Send client variations
    this.sendStoredMessages(function (result, error) {
      if (error)
        return this.fireOnError({error: error});
      //
      // Check if is required a full resync
      this.resyncRequired(function (result, error) {
        if (error)
          return this.fireOnError({error: error});
        //
        if (result) {
          // Full resync required ...
          this.resyncAllClasses(function (result, error) {
            if (error)
              return;
            //
            this.requestStoredMessages(function (result, error) {
              if (error)
                this.fireOnError({error: error});
            }.bind(this));
          }.bind(this));
        }
        else // Differenzial sync -> require server variations
          this.requestStoredMessages(function (result, error) {
            if (error)
              this.fireOnError({error: error});
          }.bind(this));
      }.bind(this));
    }.bind(this));
  }.bind(this));
};


/**
 * Event fired when a command arrived
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.onCommand = function (command, response, callback)
{
  switch (command.id) {
    case App.SyncDO.commandTypes.resyncRequired:
      this.handleResyncRequired(command, response, callback);
      break;

    case App.SyncDO.commandTypes.resyncClass:
      this.handleResyncClass(command, response, callback);
      break;

    case App.SyncDO.commandTypes.lastId:
      this.handleLastId(command, response, callback);
      break;

    case App.SyncDO.commandTypes.remoteLoad:
      this.handleRemoteLoad(command, response, callback);
      break;

    case App.SyncDO.commandTypes.remoteSave:
      this.handleRemoteSave(command, response, callback);
      break;

    case App.SyncDO.commandTypes.remoteCall:
      this.handleRemoteCall(command, response, callback);
      break;
  }
};


/**
 * Event fired when an error occurred
 * @param {Object} params
 */
App.SyncDO.prototype.onError = function (params)
{
};


/**
 * Event fired when start/end the resync of all classes
 * @param {Object} params
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.onSync = function (params, callback)
{
  if (callback)
    callback();
};


/**
 * Send a command to server
 * @param {Object} command to send
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.sendCommand = function (command, callback)
{
  this.cmd.sendCommand(command, function (response, error) {
    if (error)
      error.serverSide = true;
    callback(response, error);
  });
};


/**
 * Handle "resyncRequired" command
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.handleResyncRequired = function (command, response, callback)
{
  this.deleteExpiredMessages(function (result, error) {
    if (error)
      return callback(null, error);
    //
    var filters = {fromId: command.lastId, toId: command.lastId};
    if (command.lastId > 1)
      filters.expired = false;
    this.sync.loadMessages(this.id, filters, function (messages, error) {
      if (error)
        return callback(null, error);
      //
      // Message not found -> resync required
      callback(messages.length === 0);
    });
  }.bind(this));
};


/**
 * Check if is required a resynchronization
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.resyncRequired = function (callback)
{
  this.log("check if a full synchronization is required");
  this.sync.getLastSynchronization(this.id, function (ls, error) {
    if (error)
      return callback(null, error);
    //
    // If lastId is null then resync is required
    if (!ls.id) {
      this.log("never synchronized then full synchronization is required");
      return callback(true);
    }
    //
    var cmd = {};
    cmd.id = App.SyncDO.commandTypes.resyncRequired;
    cmd.lastId = ls.id;
    cmd.ts = ls.ts;
    this.sendCommand(cmd, function (response, error) {
      if (error)
        return callback(null, error);
      //
      this.log("full synchronization is " + (response ? "" : "NOT") + " required");
      callback(response);
    }.bind(this));
  }.bind(this));
};


/**
 * Process the variation of a document or push it in a queue if already handling variations
 * @param {Object} message
 */
App.SyncDO.prototype.handleProxyVariation = function (message)
{
  // If already handling variations put it in the queue
  if (this.handlingVariations) {
    this.handlingVariations.push(message);
    return;
  }
  //
  // Process queue of variations
  this.handlingVariations = [message];
  var nextVariation = function ()
  {
    var nextVar = this.handlingVariations.shift();
    if (!nextVar) {
      delete this.handlingVariations;
      return;
    }
    //
    this.processVariation(nextVar, nextVariation);
  }.bind(this);
  nextVariation();
};


/**
 * Process the variation of a document
 * @param {Object} message
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.processVariation = function (message, callback)
{
  this.log("process variation of document " + message.body.dna, message);
  App.Document.loadByDNA(this.app, message.body.dna, {skipLoad: true}, function (doc, error) {
    var done = function (result, error) {
      if (error || !result)
        this.fireOnError({error: error, doc: doc});
      //
      // This is the last variation to process
      // Close transaction on all dataStores
      if (this.dsTrans && this.handlingVariations.length === 0) {
        Object.keys(this.dsTrans).forEach(function (dsName) {
          this.dsTrans[dsName].commitTransaction(function () {
            delete this.dsTrans[dsName];
            if (Object.keys(this.dsTrans).length === 0)
              delete this.dsTrans;
          }.bind(this));
        }.bind(this));
      }
      //
      callback();
    }.bind(this);
    //
    if (error)
      return done(null, error);
    //
    // Set dataStore of document (if it is the same as mine)
    var dsName = doc.getDataModel();
    if (dsName === this.dataStore.schema.name)
      doc.dataStore = this.dataStore;
    else
      doc.dataStore = (this.dsTrans && this.dsTrans[dsName]) || new App[dsName](this.app);
    //
    // Process variation in transaction
    (function (cb) {
      if (doc.dataStore.isInTransaction())
        return cb();
      //
      this.dsTrans = this.dsTrans || {};
      this.dsTrans[dsName] = doc.dataStore;
      doc.dataStore.beginTransaction(cb);
    }.bind(this))(function () {
      doc.load(null, function (result, error) {
        if (error)
          return done(null, error);
        //
        // Attach variation to document for some check in addVariation
        doc.syncVariation = message;
        //
        if (message.body.stt === "D")
          doc.deleted = true;
        else {
          // If document not found init it
          if (!doc.loaded)
            doc.inserted = true;
          //
          var keys = Object.keys(message.body.cnt);
          for (var i = 0; i < keys.length; i++) {
            var propName = keys[i];
            if (propName.endsWith("_old"))
              continue;
            //
            doc[propName] = message.body.cnt[propName];
          }
        }
        //
        // Validate the document
        var opts = {};
        opts.reason = App.Document.validationReasons.sync;
        opts.childLevel = 0;
        doc.validate(opts, function (result, error) {
          if (!result)
            return done(false);
          //
          // Save the document
          doc.save({revalidate: false, remote: false}, done);
        }.bind(this));
      }.bind(this));
    });
  }.bind(this));
};


/**
 * Handle "error" message
 * @param {Object} message
 */
App.SyncDO.prototype.handleProxyError = function (message)
{
  var params = {};
  //
  if (message.body.error) {
    params.error = App.Sync.deserializeError(message.body.error);
    params.error.serverSide = true;
  }
  //
  if (message.body.doc) {
    App.Document.loadByDNA(this.app, message.body.doc.dna, {skipLoad: true}, function (doc, error) {
      if (error) {
        this.fireOnError({error: error});
        return callback(null, error);
      }
      //
      params.doc = doc;
      if (message.body.doc) {
        if (message.body.doc.docError)
          params.doc.rowError = message.body.doc.docError;
        if (message.body.doc.propErrors)
          doc.columnsErrors = message.body.doc.propErrors;
      }
      //
      this.fireOnError(params);
    }.bind(this));
  }
  else
    this.fireOnError(params);
};


/**
 * Notify an error
 * @param {Object} params
 */
App.SyncDO.prototype.fireOnError = function (params)
{
  if (params.error) {
    params.serverSide = (params.error.serverSide === true);
    delete params.error.serverSide;
  }
  //
  // Notify onError event
  this.log("fired onError event", params);
  this.onError(params);
  //
  // Send error also to client
  if (this.sync.isServerProxy()) {
    var msg = {};
    msg.class = this.id;
    msg.type = App.SyncDO.messageTypes.error;
    msg.body = {};
    //
    if (params.error)
      msg.body.error = App.Sync.serializeError(params.error);
    //
    if (params.doc) {
      msg.body.doc = {};
      msg.body.doc.dna = params.doc.getDNA();
      if (params.doc.rowError)
        msg.body.doc.docError = params.doc.rowError;
      if (Object.keys(params.doc.columnsErrors).length > 0)
        msg.body.doc.propErrors = params.doc.columnsErrors;
    }
    //
    this.sync.sendToProxy(App.Sync.messageTypes.msg, msg);
  }
};


/**
 * Records changes to a document. Normally
 * is called after after save if the service is
 * synchronization is active.
 * @param {Document} doc
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.addVariation = function (doc, callback)
{
  // Only for document with synchronization service
  var md = doc.schema;
  if (md.services.indexOf(App.Document.ServiceTypes.synchronization) === -1)
    return callback(false);
  //
  // If is set syncf.serverUrl I leave manage the variation to syncf
  if (this.app.syncf.serverUrl)
    return this.app.syncf.addVariation(doc, callback);
  //
  // The document is really changed?
  if (!doc.inserted && !doc.updated && !doc.deleted)
    return callback(false);
  //
  // The document is deleted before being added ... nothing to do
  if (doc.inserted && doc.deleted)
    return callback(false);
  //
  // The document is marked to not save variation ... do nothing
  if (!doc.syncChanges) {
    if (this.logLevel === App.SyncPlugin.logLevels.verbose)
      this.log("variation not stored for document " + doc.getDNA() + " because syncChanges is false");
    return callback(false);
  }
  //
  var params = {};
  //
  // Set as default topics that of variation or *
  params.topics = (doc.syncVariation ? doc.syncVariation.topics : "*");
  //
  this.getTopics(doc, params, function (result, error) {
    // If document did not provide topics means that will not be synchronized
    if (params.topics === undefined) {
      if (this.logLevel === App.SyncPlugin.logLevels.verbose)
        this.log("variation not stored for document " + doc.getDNA() + " because topics is undefined");
      return callback(false);
    }
    //
    var msg = {};
    msg.class = this.id;
    msg.ttl = this.ttl;
    msg.topics = params.topics;
    msg.body = {};
    msg.body.dna = doc.getDNA();
    //
    // Calculating the current state
    msg.body.stt = "I";
    if (!doc.inserted) {
      if (doc.updated)
        msg.body.stt = "U";
      if (doc.deleted)
        msg.body.stt = "D";
    }
    //
    var msgMore = null;
    if (msg.body.stt !== "D") {
      var cnt = this.getContent(doc);
      //
      // If content is empty no variation
      if (Object.keys(cnt).length === 0)
        return callback(false);
      //
      msg.body.cnt = cnt;
      //
      // Check if there are more variations than synchronizing
      if (doc.syncVariation) {
        var md = doc.schema;
        var cntMore = {};
        for (var i = 0; i < md.properties.length; i++) {
          var mdp = md.properties[i];
          if (doc.syncVariation.body.cnt[mdp.name] !== cnt[mdp.name]) {
            cntMore[mdp.name] = cnt[mdp.name];
            cntMore[mdp.name + "_old"] = doc.syncVariation.body.cnt[mdp.name];
          }
        }
        //
        // There are more variations than synchronizing: create a second message
        if (Object.keys(cntMore).length > 0) {
          msgMore = JSON.parse(JSON.stringify(msg));
          this.sync.initMessage(msgMore);
          msgMore.body.stt = "U";
          msgMore.body.cnt = cntMore;
        }
        msg.fromProxy = true;
      }
    }
    //
    if (doc.syncVariation) {
      msg.sender = doc.syncVariation.sender;
      msg.uid = doc.syncVariation.uid;
      msg.device = doc.syncVariation.device;
    }
    else
      this.sync.initMessage(msg);
    //
    var afterSendMessages = function (result, error) {
      callback(true, error);
    };
    //
    // Store variation message(s)
    var storeVariation = function (m, cb) {
      // Client side variation must not be stored; only further variation
      if (m === msg && this.sync.isClientProxy() && doc.syncVariation)
        return cb();
      //
      if (m === msgMore && this.sync.isServerProxy()) {
        // Server side further variation must send only to client (don't store)
        this.sync.sendToProxy(App.Sync.messageTypes.msg, m);
        return cb();
      }
      //
      var transOpened = false;
      var ok = true;
      this.pendingMessages = this.pendingMessages || [];
      //
      // If dataStore is undefined callback with error
      if (!this.dataStore) {
        this.app.dttobj.addWarning("", "SyncDO: variation not stored because dataStore is undefined");
        return cb();
      }
      //
      // If doc dataStore is my same I store message with doc dataStore connection
      if (this.dataStore.schema.name === doc.dataStore.schema.name) {
        this.pendingMessages.push(m);
        doc.dataStore.syncStoreMessage(this.id, m, cb);
      }
      else if (!this.dataStore.isInTransaction()) {
        // If my dataStore is not in transaction start one
        this.dataStore.beginTransaction(function (result, error) {
          if (error) {
            ok = false;
            return cb(null, error);
          }
          //
          transOpened = true;
          this.pendingMessages.push(m);
          this.sync.storeMessage(m, cb);
        }.bind(this));
      }
      else {
        this.pendingMessages.push(m);
        this.sync.storeMessage(m, cb);
      }
      //
      // Wait end of doc dataStore transaction
      if (!doc.dataStore.endTransactionCallback) {
        this.awaitingDS = this.awaitingDS || [];
        this.awaitingDS.push(doc.dataStore);
        doc.dataStore.endTransactionCallback = function (success) {
          delete doc.dataStore.endTransactionCallback;
          this.awaitingDS.splice(this.awaitingDS.indexOf(doc.dataStore), 1);
          ok = ok && success;
          if (this.awaitingDS.length === 0) {
            delete this.awaitingDS;
            var endTransaction = function (result, error) {
              if (!error && ok)
                this.pendingMessages.forEach(App.Sync.prototype.dispatchMessage.bind(this.sync));
              delete this.pendingMessages;
            }.bind(this);
            //
            if (transOpened)
              this.dataStore[(ok ? "commit" : "rollback") + "Transaction"](endTransaction);
            else
              endTransaction();
          }
        }.bind(this);
      }
    }.bind(this);
    //
    storeVariation(msg, function (result, error) {
      if (error)
        callback(null, error);
      else if (msgMore)
        storeVariation(msgMore, afterSendMessages);
      else
        afterSendMessages();
    });
  }.bind(this));
};


/**
 * Picks up the document topics
 * @param {Document} doc
 * @param {Object} params
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.getTopics = function (doc, params, callback)
{
  doc.onGetTopics(params, callback);
};


/**
 * Returns an object that represents the changes made to the document
 * @param {Document} doc
 **/
App.SyncDO.prototype.getContent = function (doc)
{
  var cnt = {};
  var md = doc.schema;
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    //
    // Skip not synchronizable properties
    if (!mdp.synchronize)
      continue;
    //
    var vatt = doc[mdp.name];
    var vold = doc.getOriginalValue(mdp.name);
    if (doc.inserted || !App.Locale.compareValues(vatt, vold)) {
      if ([App.Database.DataTypeClasses.DATETIME, App.Database.DataTypeClasses.BLOB].indexOf(App.Database.classDT(mdp.datatype)) >= 0) {
        vatt = mdp.valueToString(vatt, null, this.app);
        vold = mdp.valueToString(vold, null, this.app);
      }
      cnt[mdp.name] = vatt;
      //
      if (md.optimisticLocking && mdp.isComparable())
        cnt[mdp.name + "_old"] = vold;
    }
  }
  return cnt;
};


/**
 * Resynchronize all classes
 * @param {Function} callback - function to be called at the end
 **/
App.SyncDO.prototype.resyncAllClasses = function (callback)
{
  var i = 0;
  var nextClass = function (result, error) {
    if (error) {
      this.log("full synchronization of all classes interrupted because of an error", error);
      this.fireOnError({error: error});
      this.onSync({status: "error", full: true, error: error}, function (result, error1) {
        if (callback)
          callback(null, error);
      });
      return;
    }
    //
    if (i < App.SyncDO.docToSync.length)
      return this.resyncClass(App.SyncDO.docToSync[i++], nextClass);
    //
    this.log("full synchronization of all classes completed");
    this.onSync({status: "completed", full: true}, function (result, error) {
      if (error) {
        if (callback)
          callback(null, error);
        return;
      }
      //
      this.setSynchronized(callback);
    }.bind(this));
  }.bind(this);
  //
  // Resync all classes
  this.log("start full synchronization of all classes");
  this.onSync({status: "started", full: true}, nextClass);
};


/**
 * Request messages stored into datastore
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.requestStoredMessages = function (callback)
{
  this.requestingStoredMessages = true;
  this.onSync({status: "started", full: false}, function () {
    App.SyncPlugin.prototype.requestStoredMessages.call(this, callback);
  }.bind(this));
};


/**
 * Event fired when synchronization of messages is ended
 */
App.SyncDO.prototype.onSynchronized = function ()
{
  if (!this.requestingStoredMessages)
    return;
  //
  // If I'm still processing variations postponement event notification
  if (this.handlingVariations && this.dsTrans)
    return _setTimeout(App.SyncDO.prototype.onSynchronized.bind(this), 100);
  //
  this.requestingStoredMessages = false;
  this.onSync({status: "completed", full: false});
};


/**
 * Update first record of z_sync with lastID of server
 * @param {Function} callback - function to be called at the end
 **/
App.SyncDO.prototype.setSynchronized = function (callback)
{
  this.getLastId(function (result, error) {
    if (error) {
      if (callback)
        callback(null, error);
      return;
    }
    //
    this.sync.setSynchronized(this.id, result, callback);
  }.bind(this));
};


/**
 * Get the last id of z_sync of server
 * @param {Function} callback - function to be called at the end
 **/
App.SyncDO.prototype.getLastId = function (callback)
{
  if (this.sync.isClientProxy()) {
    // Client side require lastId to server
    var cmd = {};
    cmd.id = App.SyncDO.commandTypes.lastId;
    this.sendCommand(cmd, callback);
  }
  else {
    // Server side read lastId from z_sync
    var filters = {maxId: true};
    this.sync.loadMessages(this.id, filters, function (messages, error) {
      if (error)
        callback(null, error);
      else
        callback(messages[0].id);
    });
  }
};


/**
 * Handle "lastId" command
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.handleLastId = function (command, response, callback)
{
  this.getLastId(function (result, error) {
    if (error)
      callback(null, error);
    else {
      response.result = result;
      callback();
    }
  });
};


/**
 * Fire onMessageFiltering event
 * @param {Object} message
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.fireOnMessageFiltering = function (message, options, callback)
{
  App.SyncPlugin.prototype.fireOnMessageFiltering.call(this, message, options, function (result, error) {
    if (error)
      return callback(null, error);
    //
    if (!message.type && this.sync.isServerProxy())
      this.onVariationFiltering(message, options, callback);
    else
      callback();
  }.bind(this));
};


/**
 * onVariationFiltering event
 * @param {Object} message
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.onVariationFiltering = function (message, options, callback)
{
  callback();
};


/**
 * Resynchronize a class
 * @param {String} className
 * @param {Function} callback - function to be called at the end
 **/
App.SyncDO.prototype.resyncClass = function (className, callback)
{
  this.log("start full synchronization of class " + className);
  var cmd = {};
  cmd.id = App.SyncDO.commandTypes.resyncClass;
  if (!className.startsWith("\""))
    cmd.dna = "\"" + className + "\"";
  else
    cmd.dna = className;
  //
  this.sendCommand(cmd, function (datamap, error) {
    if (error) {
      this.log("full synchronization of class " + className + " interrupted because of an error server side", error);
      if (callback)
        callback(null, error);
      return;
    }
    //
    this.log("full synchronization of class " + className + ": " + datamap.rows.length + " documents to save");
    if (datamap.rows.length === 0) {
      if (callback)
        callback();
    }
    else {
      // Save datamap in transaction
      datamap.upsert = true;
      //
      // Set dataStore of datamap (if it is the same as mine)
      if (datamap.getDataModel() === this.dataStore.schema.name)
        datamap.dataStore = this.dataStore;
      else
        datamap.dataStore = new App[datamap.getDataModel()](this.app);
      //
      datamap.dataStore.beginTransaction(function (result, error) {
        if (error) {
          if (callback)
            callback(null, error);
          return;
        }
        //
        // Intercept errors during datamap saving
        var dmError = "";
        datamap.onError = function (errorMessage) {
          dmError = errorMessage;
        };
        //
        datamap.save(null, function (result, error) {
          if (error || dmError) {
            this.log("full synchronization of class " + className + " interrupted because of an error client side", (error || dmError));
            datamap.dataStore.rollbackTransaction(function () {
              if (callback)
                callback(null, error || dmError);
            });
            return;
          }
          //
          datamap.dataStore.commitTransaction(function (result, error) {
            if (error) {
              if (callback)
                callback(null, error);
              return;
            }
            //
            // If datamap is full check if there are other data (except if datamap isn't partial)
            if (datamap.rows.length === this.sync.maxMessages && datamap.partial !== false) {
              this.log("full synchronization of class " + className + " next " + this.sync.maxMessages + " documents");
              var lastRow = datamap.rows[datamap.rows.length - 1];
              App.Document.loadByDNA(this.app, cmd.dna, {skipLoad: true}, function (lastDoc, error) {
                lastDoc.values = lastRow.values;
                this.resyncClass(lastDoc.getDNA(), callback);
              }.bind(this));
            }
            else {
              if (callback)
                callback();
            }
          }.bind(this));
        }.bind(this));
      }.bind(this));
    }
  }.bind(this));
};


/**
 * Handle "resyncClass" command
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.handleResyncClass = function (command, response, callback)
{
  App.Document.loadByDNA(this.app, command.dna, {skipLoad: true}, function (lastDoc, error) {
    if (error)
      return callback(null, error);
    //
    var doc = new (lastDoc.class)(this.app);
    var options = {skip: false};
    options.collection = new App.Collection();
    if (command.dna.indexOf(",") > 0)
      options.lastDoc = lastDoc;
    //
    // Notify onResyncClient
    doc.onResyncClient(options, function (result, error) {
      if (error)
        return callback(null, error);
      //
      var afterLoad = function (result, error) {
        if (error)
          return callback(null, error);
        //
        // Delete unbound columns from schema to avoid to send her
        var unbounds = 0;
        var columns = coll.schema.columns;
        for (var c in columns) {
          var dc = columns[c];
          if (dc.unbound) {
            unbounds++;
            delete columns[c];
          }
          else
            dc.index -= unbounds;
        }
        //
        response.result = coll;
        //
        // Change prototype beacause I want send data as DataMap
        Object.setPrototypeOf(response.result, App.DataMap.prototype);
        //
        callback();
      };
      //
      var coll = options.collection;
      if (options.skip) {
        // Set datamap as not partial
        if (!coll.partial)
          coll.partial = false;
        //
        // Set rowTemplate if undefined
        if (!coll.rowTemplate)
          coll.rowTemplate = doc;
        //
        return afterLoad();
      }
      //
      // Limit document to load
      coll.maxRows = this.sync.maxMessages;
      //
      // Set datamap as partial
      coll.partial = true;
      //
      // Intercept beforeLoad for add partialization's conditions
      coll.beforeLoad = function (query, callback) {
        var numkeys = command.dna.split(",").length - 1;
        var wcpart = [];
        var attpart = 0;
        var orderBy = [];
        var md = lastDoc.schema;
        for (var i = 0; i < md.properties.length; i++) {
          var mdp = md.properties[i];
          if (mdp.pk) {
            // I compose the where conditions for partialization
            var dnapart = lastDoc[mdp.name];
            for (var j = 0; j < numkeys - attpart; j++) {
              var wc = {expr: (j < numkeys - attpart - 1 ? "=" : ">"), p: []};
              wc.p.push(coll.schema.columns[mdp.name].getSelectToken());
              wc.p.push({expr: lastDoc.dataStore.toSql(dnapart, mdp.datatype)});
              if (!wcpart[j])
                wcpart[j] = wc;
              else
                wcpart[j] = {p: [{expr: "("}, wcpart[j], {expr: "and"}, wc, {expr: ")"}]};
            }
            attpart++;
            //
            // However the collection I want to sort by the pk
            orderBy.push(mdp.name);
          }
        }
        coll.orderBy = orderBy;
        //
        if (wcpart.length > 0) {
          var wc = null;
          for (var i = 0; i < wcpart.length; i++) {
            if (i === 0)
              wc = wcpart[i];
            else
              wc = {p: [{expr: "("}, wc, {expr: "or"}, wcpart[i], {expr: ")"}]};
          }
          query.where = query.where || [];
          query.where.push(wc);
        }
        else if (!attpart)
          return callback(null, new Error("Unable to synchronize class " + lastDoc.schema.className + " without primary keys"));
        //
        callback();
      }.bind(this);
      //
      doc.loadCollectionByExample(coll, {useQBE: true}, function (result, error) {
        afterLoad(result, error);
      });
    }.bind(this));
  }.bind(this));
};


/**
 * Check if an object allow remote call
 * @param {Document/Class} obj
 */
App.SyncDO.prototype.disallowedRemoteCall = function (obj)
{
  if (!obj)
    return "Document non found";
  //
  if (!((typeof obj === "function" ? obj.prototype : Object.getPrototypeOf(obj)) instanceof App.Document))
    return "Object is not a document";
  //
  if (!obj.schema.allowRemoteCall)
    return obj.schema.className + " not allow remote call";
};


/**
 * Fire onRemoteCall event
 * @param {Document/Class/Collection} obj
 * @param {String} method
 * @param {Array} args
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.fireOnRemoteCall = function (obj, method, args, callback)
{
  try {
    var options = {result: obj, skip: false, cancel: false};
    this.onRemoteCall(obj, method, args, options, function (result, error) {
      if (error)
        return callback(null, error);
      //
      if (options.skip)
        return callback(options.result);
      //
      if (options.cancel)
        return callback(null, options.error || new Error("Remote call canceled server side"));
      //
      callback();
    });
  }
  catch (e) {
    callback(null, e);
  }
};


/**
 * Fire onRemoteCall event
 * @param {Document/Class/Collection} obj
 * @param {String} method
 * @param {Array} args
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.onRemoteCall = function (obj, method, args, options, callback)
{
  callback();
};


/**
 * Load remotely a collection
 * @param {Collection} coll
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.remoteLoad = function (coll, options, callback)
{
  var cmd = {};
  cmd.id = App.SyncDO.commandTypes.remoteLoad;
  cmd.opts = JSON.parse(JSON.stringify(options));
  delete cmd.opts.remote;
  cmd.opts.maxRows = coll.maxRows;
  cmd.opts.orderBy = coll.orderBy;
  cmd.opts.groupBy = coll.groupBy;
  //
  // Add info for pagination
  if (coll.dataPagingMode) {
    cmd.pageMode = coll.dataPagingMode;
    cmd.preCount = coll.dataPagingPreCount;
    cmd.pageSize = coll.dataPageSize;
    if (options.page && coll.dataPagingMode === App.DataMap.dataPagingModes.keyset)
      cmd.lastRow = coll.rows[((options.page || 0) * coll.dataPageSize) - 1].intoJSON({childLevel: 0});
  }
  //
  var doc = null;
  switch (options.loadType) {
    case App.Document.loadTypes.loadChildCollection:
      doc = coll.parent;
      cmd.collName = coll.parent.schema.getCollDefinition(coll.index).name;
      break;

    case App.Document.loadTypes.load:
    case App.Document.loadTypes.loadCollection:
      doc = coll.rowTemplate;
      break;
  }
  //
  var opts = {};
  opts.childLevel = 0;
  opts.asString = false;
  opts.skipType = false;
  cmd.doc = doc.intoJSON(opts);
  //
  if (options.loadType === App.Document.loadTypes.loadCollection) {
    // Add filters
    cmd.filters = {};
    cmd.searchOptions = coll.searchOptions;
    for (var c in coll.schema.columns) {
      var dc = coll.schema.columns[c];
      if (dc.filters.length > 0)
        cmd.filters[dc.name] = dc.filters;
    }
    //
    // Add DO filters as filter
    if (coll.DOFilters) {
      cmd.DOFilters = [];
      for (var i = 0; i < coll.DOFilters.length; i++) {
        var filter = coll.DOFilters[i];
        if (coll.query.where && coll.query.where.indexOf(filter.wc) >= 0) {
          if (!cmd.filters[filter.col])
            cmd.filters[filter.col] = [];
          cmd.filters[filter.col].push({v: coll.getClientRefValue(filter)});
        }
      }
    }
  }
  //
  // Send command of remote load
  this.sendCommand(cmd, function (result, error) {
    if (error) {
      this.fireOnError({error: error});
      return callback(null, error);
    }
    //
    var c = result;
    if (coll.dataPagingPreCount && !options.page) {
      coll._calcRowCount = result.length;
      c = new App.Collection(this.app);
      c.loadFromJSON(result.coll);
    }
    //
    callback(c);
  }.bind(this));
};


/**
 * Handle "remoteLoad" command
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.handleRemoteLoad = function (command, response, callback)
{
  var doc = App.Document.fromJSON(this.app, command.doc);
  //
  // Check if document allow remote call
  var disallowedError = this.disallowedRemoteCall(doc);
  if (disallowedError)
    return callback(null, new Error(disallowedError));
  //
  var options = command.opts;
  var coll = null;
  switch (options.loadType) {
    case App.Document.loadTypes.loadChildCollection:
      coll = doc[command.collName];
      this.fireOnRemoteCall(coll, "load", [options], function (result, error) {
        if (result || error)
          return callback(result, error);
        //
        coll.load(options, function (result, error) {
          callback(coll, error);
        });
      });
      break;

    case App.Document.loadTypes.load:
      this.fireOnRemoteCall(doc, "load", [options], function (result, error) {
        if (result || error)
          return callback(result, error);
        //
        doc.load(options, function (result, error) {
          coll = new App.Collection(this.app);
          if (doc.loaded)
            coll.add(doc);
          callback(coll, error);
        }.bind(this));
      }.bind(this));
      break;

    case App.Document.loadTypes.loadCollection:
      this.fireOnRemoteCall(doc.class, "loadCollection", [doc, options], function (result, error) {
        if (result || error)
          return callback(result, error);
        //
        var coll = new App.Collection(this.app);
        coll.searchOptions = command.searchOptions;
        //
        // If collection has pagination set some properties
        if (command.pageMode) {
          coll.dataPagingMode = command.pageMode;
          coll.dataPagingPreCount = command.preCount;
          coll.dataPageSize = command.pageSize;
          if (command.opts.page && command.pageMode === App.DataMap.dataPagingModes.keyset)
            coll.rows[((options.page || 0) * coll.dataPageSize) - 1] = doc.class.fromJSON(this.app, command.lastRow);
        }
        //
        // Add filters
        coll.beforeLoad = function (query, callback) {
          coll.clearFilters();
          for (var c in coll.schema.columns) {
            var dc = coll.schema.columns[c];
            if (!command.filters[dc.name])
              continue;
            var filters = command.filters[dc.name];
            for (var i = 0; i < filters.length; i++)
              coll.addFilter(dc.name, filters[i].v, filters[i].or, filters[i].qbe);
          }
          callback();
        };
        //
        doc.loadCollectionByExample(coll, options, function (result, error) {
          if (error)
            return callback(null, error);
          //
          if (coll.dataPagingMode) {
            var length = coll.length;
            //
            // I leave in the collection only the rows loaded
            coll.rows = coll.rows.splice((command.opts.page || 0) * coll.dataPageSize, coll.dataPageSize);
            //
            // With precount send also the length
            if (coll.dataPagingPreCount && !command.opts.page)
              return callback({length: length, coll: coll.intoJSON({asString: false, skipType: false})});
          }
          //
          callback(coll);
        });
      }.bind(this));
      break;
  }
};


/**
 * Save remotely a document
 * @param {Document} doc
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.remoteSave = function (doc, options, callback)
{
  var cmd = {};
  cmd.id = App.SyncDO.commandTypes.remoteSave;
  cmd.opts = JSON.parse(JSON.stringify(options));
  delete cmd.opts.remote;
  //
  var opts = {};
  opts.skipStatus = false;
  opts.skipErrors = false;
  opts.skipType = false;
  opts.skipOriginals = false;
  opts.asString = false;
  opts.skipNulls = false;
  cmd.doc = doc.intoJSON(opts);
  //
  // Send command of remote save
  this.sendCommand(cmd, function (result, error) {
    if (error) {
      this.fireOnError({error: error});
      return callback(null, error);
    }
    //
    doc.copyFrom(App.Document.fromJSON(this.app, result, opts), 9999);
    callback(!doc.hasErrors());
  }.bind(this));
};


/**
 * Handle "remoteSave" command
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.handleRemoteSave = function (command, response, callback)
{
  var doc = App.Document.fromJSON(this.app, command.doc);
  //
  // Check if document allow remote call
  var disallowedError = this.disallowedRemoteCall(doc);
  if (disallowedError)
    return callback(null, new Error(disallowedError));
  //
  var options = command.opts;
  this.fireOnRemoteCall(doc, "save", [options], function (result, error) {
    if (result || error)
      return callback(result, error);
    //
    doc.save(options, function (result, error) {
      if (error)
        return callback(null, error);
      //
      var opts = {};
      opts.skipStatus = false;
      opts.skipErrors = false;
      opts.skipType = false;
      opts.skipOriginals = false;
      opts.asString = false;
      callback(doc.intoJSON(opts), error);
    });
  });
};


/**
 * Call remotely a method
 * @param {Document/Function} obj
 * @param {String/Function} method to call
 * @param {Array} args
 * @param {Object} options (only if obj is a Document)
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.remoteCall = function (obj, method, args, options, callback)
{
  options = options || {};
  var cmd = {};
  cmd.id = App.SyncDO.commandTypes.remoteCall;
  cmd.args = this.sync.serializeObject(args);
  //
  if (typeof method === "function") {
    // TODO: function -> name
  }
  else if (typeof method === "string") {
    if (typeof obj[method] === "function")
      cmd.met = method;
  }
  //
  if (!cmd.met)
    return callback(null, new Error("Invalid method"));
  //
  if (typeof obj === "function" && obj.prototype instanceof App.Document)
    cmd.class = obj.schema.className;
  else if (obj instanceof App.Document) {
    if (options.loadRemotely)
      cmd.dna = obj.getDNA();
    else {
      var opts = {};
      opts.skipStatus = false;
      opts.skipErrors = false;
      opts.skipType = false;
      opts.skipOriginals = false;
      opts.asString = false;
      opts.childLevel = options.childLevel;
      cmd.doc = obj.intoJSON(opts);
    }
  }
  else
    return callback(null, new Error("Invalid document"));
  //
  // Send command of remote call
  this.sendCommand(cmd, function (result, error) {
    if (error) {
      this.fireOnError({error: error});
      return callback(null, error);
    }
    //
    callback(result);
  }.bind(this));
};


/**
 * Handle "remoteCall" command
 * @param {Object} command received
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.SyncDO.prototype.handleRemoteCall = function (command, response, callback)
{
  var afterObj = function (obj) {
    // Check if document allow remote call
    var disallowedError = this.disallowedRemoteCall(obj);
    if (disallowedError)
      return callback(null, new Error(disallowedError));
    //
    var method = obj[command.met];
    if (typeof method !== "function")
      return callback(null, new Error(command.met + " is not a function"));
    //
    // Check if method is allowed to be invoked remotely
    if (obj.schema.allowRemoteCallMethods.indexOf(command.met) === -1)
      return callback(null, new Error(command.met + " is not allowed to be invoked remotely"));
    //
    var args = command.args;
    this.fireOnRemoteCall(obj, command.met, args, function (result, error) {
      if (result || error)
        return callback(result, error);
      //
      var argNames = App.Utils.functArguments(method);
      //
      // Add app as first argument if function is static
      if (argNames[0] === "app" && !(args[0] instanceof App.Session))
        args.unshift(this.app);
      //
      // Add callback as last argument if function is asynchronous
      var async = false;
      if (["_resumer", "callback"].indexOf(argNames[argNames.length - 1]) >= 0) {
        async = true;
        args[argNames.length - 1] = callback;
      }
      //
      try {
        // Call the method
        var res = method.apply(obj, args);
        if (!async)
          callback(res);
      }
      catch (e) {
        callback(null, e);
      }
    }.bind(this));
  }.bind(this);
  //
  if (command.class)
    afterObj(App.Session.getClassByName(command.class));
  else if (command.dna)
    App.Document.loadByDNA(this.app, command.dna, null, function (doc, error) {
      afterObj(doc);
    });
  else if (command.doc)
    afterObj(App.Document.fromJSON(this.app, command.doc));
};


/**
 * Informs all that document is changed
 * @param {App.Document} doc
 * @param {Object} options
 */
App.SyncDO.prototype.notifyDocUpdate = function (doc, options)
{
  options = options || {};
  //
  // Compose message
  var msg = {};
  msg.type = App.SyncDO.messageTypes.docUpdate;
  msg.class = this.id;
  msg.options = options;
  msg.topics = ("topics" in options ? options.topics : "*");
  msg.fromProxy = true;
  //
  // Serialize document in json
  var opts = {};
  opts.childLevel = 0;
  opts.skipStatus = false;
  opts.skipOriginals = false;
  opts.skipType = false;
  msg.doc = doc.intoJSON(opts);
  //
  // Handle doc update
  this.handleDocUpdate(msg);
};


/**
 * Handle changes of a document
 * @param {Object} message
 */
App.SyncDO.prototype.handleDocUpdate = function (message)
{
  // Load document from JSON
  var doc = App.Document.fromJSON(this.app, message.doc);
  //
  // Dispatch message to all others sessions
  if (message.options.allSessions && !this.sync.isClientProxy()) {
    delete message.options.allSessions;
    this.sync.sendMessage(message);
  }
  //
  // Clear class cache
  doc.class.clearCache(this.app, doc.getDNA());
  //
  var handleDocsUpdate = function () {
    delete this.docUpdateTimer;
    delete this.docUpdateLastTS;
    //
    // Dispatch messages to session
    this.docUpdateCache.forEach(function (d) {
      this.app.handleDocUpdate(d);
      //
      // Set document as original and loaded
      d.setOriginal(0);
      d.loaded = true;
    }.bind(this));
    //
    // Empty docUdpateCache
    this.docUpdateCache.length = 0;
  }.bind(this);
  //
  // I put the message in a cache to handle so merged updates
  if (!this.docUpdateCache)
    this.docUpdateCache = [];
  this.docUpdateCache.push(doc);
  //
  // Handle merged updates immediatly if it's been some time or if I already have a number of cached updates
  if ((this.docUpdateLastTS && (new Date() - this.docUpdateLastTS) >= this.docUpdateCacheTimeout)
          || this.docUpdateCache.length >= this.docUpdateCacheSize) {
    // Reset the timer (if started)
    if (!this.docUpdateTimer)
      clearTimeout(this.docUpdateTimer);
    //
    // Handle all docUpdate in cache
    handleDocsUpdate();
  }
  else {
    this.docUpdateLastTS = new Date();
    if (!this.docUpdateTimer)
      this.docUpdateTimer = setTimeout(handleDocsUpdate, this.docUpdateCacheTimeout);
  }
};


//export module for node
if (module)
  module.exports = App.SyncDO;
