/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

/* global module, Uint8Array, Buffer, ArrayBuffer, TextEncoder, TextDecoder */

var App = App || {};


App.Utils = function ()
{
  if (module && !App.View)
    App = require("./requireAll");
};

/**
 * Convert a string from camel case to dashed format
 * @param {String} str
 * @returns {String}
 */
App.Utils.unCamelCase = function (str)
{
  return str.replace(/([a-z])([A-Z])/g, "$1-$2").toLowerCase();
};

/**
 * Convert a string from dashed format to camel case
 * @param {String} str
 * @returns {String}
 */
App.Utils.camelCase = function (str)
{
  if (str.indexOf("-") === -1)
    return str;
  //
  return str.toLowerCase().replace(/-(.)/g, function (match, group1) {
    return group1.toUpperCase();
  });
};

/**
 * Replace parameters in the css
 * @param {String} name - the property to replace
 * @param {String} statements - the list of statements prop:name=value
 * @param {String} css - the css containing the parameters
 * @param {Device} device - device definition of the application
 * @param {Boolean} noDollar - if true, parameters to replace don't start with a dollar
 * @returns {String}
 */
App.Utils.replaceParameter = function (name, statements, css, device, noDollar)
{
  if (!statements)
    return css;
  //
  var s = statements.split(";");
  for (var i = 0; i < s.length; i++) {
    var res = this.getStatementValue(s[i], device);
    if (res && css && name) {
      res = this.unCamelCase(res);
      //
      // The parameters to replace don't start with "$".
      // It happens when replacing the parameters from a CSS Resource.
      if (noDollar) {
        css = this.replaceNoDollarParams(css, name, res);
        //
        // Replace the camel cased version of the parameter, too
        css = this.replaceNoDollarParams(css, App.Utils.camelCase(name), res);
        break;
      }
      else {
        var regexp = new RegExp("\\$" + name + "\\b", "g");
        css = css.replace(regexp, res);
        break;
      }
    }
  }
  //
  return css;
};


/**
 * Replace css parameters that not don't start with "$"
 * @param {String} css - the css containing the parameters
 * @param {String} name - the property to replace
 * @param {String} res - the new value of the property
 * @returns {String}
 */
App.Utils.replaceNoDollarParams = function (css, name, res)
{
  var chars = [" ", "=", "\"", ";", ":"];
  var prevChar = "";
  var nextChar = "";
  //
  // Search the parameter to substitute
  var idx = css.indexOf(name);
  while (idx !== -1) {
    if (idx > 0)
      prevChar = css.charAt(idx - 1);
    if (idx < css.length - 1)
      nextChar = css.charAt(idx + name.length);
    //
    // If the full name was found, replace it
    if ((chars.indexOf(prevChar) > -1 || prevChar === "$") && chars.indexOf(nextChar) > -1) {
      var start = idx;
      if (prevChar === "$")
        start = idx - 1;
      css = css.substring(0, start) + res + css.substring(idx + name.length);
    }
    //
    idx = css.indexOf(name, idx + 1);
  }
  //
  return css;
};


/**
 * Get the value in a statement
 * @param {String} statement - "param=name:value"
 * @param {Device} device
 * @returns {String} value if name match
 *
 * Examples:
 *  App.Utils.getStatementValue("$ua=ie10:border")        returns "border" if device.ua === "ie10"
 *  App.Utils.getStatementValue("$ua=ie9-11:border")      returns "border" if ["ie9", "ie10", "ie11"].indexOf(device.ua) >= 0
 *  App.Utils.getStatementValue("$platform=Win32:border") returns "border" if device.platform === "Win32"
 */
App.Utils.getStatementValue = function (statement, device)
{
  var s = statement.split(":");
  if (s.length > 1) {
    var name = s[0].split("=");
    if (name[0] === "$ua") {
      var res = this.decodeUA(name[0], name[1], s[1], device);
      if (res)
        return res;
    }
    else {
      if (device[name[0].substring(1)] === name[1].trim())
        return s[1].trim();
    }
  }
  else
    return statement.trim();
};


/**
 * Replacing theme parameters in requirements
 * @param {object} all object which properties will be replaced
 * @param {object} theme properties to replace
 */
App.Utils.replaceObjProps = function (all, theme)
{
  var k = Object.keys(all);
  for (var j = 0; j < k.length; j++) {
    var rn = k[j];
    if (rn.indexOf("$") > -1) {
      for (var prop in theme) {
        var p = theme[prop];
        var regexp = new RegExp("\\$" + prop, "g");
        rn = rn.replace(regexp, p);
      }
      if (rn !== k[j]) {
        all[rn] = all[k[j]];
        delete all[k[j]];
      }
    }
  }
};


/**
 * Decode a property of device object
 * @param {String} param
 * @param {String} name
 * @param {String} value
 * @param {Device} device
 * @returns {String}
 */
App.Utils.decodeUA = function (param, name, value, device)
{
  var agent = device[param.substring(1)];
  var matchingAgent = name.trim();
  var dashIdx = matchingAgent.indexOf("-");
  var numIdx = matchingAgent.search(/\d/);
  //
  // If there is no agent, do nothing
  if (!agent)
    return;
  //
  // There isn't a dash, it could be a generic or a specific user agent
  if (dashIdx < 0) {
    // It's a specific device
    if (numIdx > 0 && agent === matchingAgent)
      return value.trim();
    else if (agent.indexOf(matchingAgent) > -1)  // It's a generic user agent (no version number)
      return value.trim();
  }
  else {
    // It's a range of user agent
    var minv = 0;
    if (numIdx < dashIdx)
      minv = parseInt(matchingAgent.substring(numIdx, dashIdx));
    var maxv = 9999;
    if (dashIdx + 1 < matchingAgent.length)
      maxv = parseInt(matchingAgent.substring(dashIdx + 1));
    var version = parseInt(agent.substring(agent.search(/\d/)));
    //
    if (version >= minv && (version <= maxv))
      return value.trim();
  }
};


/* jshint ignore:start */

/**
 * Generate an UUID 24 characters long
 * @returns {String}
 */
App.Utils.generateUID24 = function ()
{
  var d0 = Math.random() * 0xffffffff | 0;
  var d1 = Math.random() * 0xffffffff | 0;
  var d2 = Math.random() * 0xffffffff | 0;
  var d3 = Math.random() * 0xffffffff | 0;
  var s = String.fromCharCode(d0 & 0xff) + String.fromCharCode(d0 >> 8 & 0xff) + String.fromCharCode(d0 >> 16 & 0xff) + String.fromCharCode(d0 >> 24 & 0xff) +
          String.fromCharCode(d1 & 0xff) + String.fromCharCode(d1 >> 8 & 0xff) + String.fromCharCode(d1 >> 16 & 0xff) + String.fromCharCode(d1 >> 24 & 0xff) +
          String.fromCharCode(d2 & 0xff) + String.fromCharCode(d2 >> 8 & 0xff) + String.fromCharCode(d2 >> 16 & 0xff) + String.fromCharCode(d2 >> 24 & 0xff) +
          String.fromCharCode(d3 & 0xff) + String.fromCharCode(d3 >> 8 & 0xff) + String.fromCharCode(d3 >> 16 & 0xff) + String.fromCharCode(d3 >> 24 & 0xff);
  //
  // btoa does not work in node therefore we use a function toBase64 defined below
  if (module)
    return (new Buffer(s || "", "ascii")).toString("base64");
  else
    return btoa(s);
};


/**
 * Generate an UUID 36 characters long
 * @returns {String}
 */
App.Utils.generateUID36 = function ()
{
  var lut = [];
  for (var i = 0; i < 256; i++)
    lut[i] = (i < 16 ? '0' : '') + (i).toString(16);
  var d0 = Math.random() * 0xffffffff | 0;
  var d1 = Math.random() * 0xffffffff | 0;
  var d2 = Math.random() * 0xffffffff | 0;
  var d3 = Math.random() * 0xffffffff | 0;
  return lut[d0 & 0xff] + lut[d0 >> 8 & 0xff] + lut[d0 >> 16 & 0xff] + lut[d0 >> 24 & 0xff] + '-' +
          lut[d1 & 0xff] + lut[d1 >> 8 & 0xff] + '-' + lut[d1 >> 16 & 0x0f | 0x40] + lut[d1 >> 24 & 0xff] + '-' +
          lut[d2 & 0x3f | 0x80] + lut[d2 >> 8 & 0xff] + '-' + lut[d2 >> 16 & 0xff] + lut[d2 >> 24 & 0xff] +
          lut[d3 & 0xff] + lut[d3 >> 8 & 0xff] + lut[d3 >> 16 & 0xff] + lut[d3 >> 24 & 0xff];
};

/* jshint ignore:end */


/**
 * Generate an UUID 20 characters long
 * @returns {String}
 */
App.Utils.generateUID20 = function ()
{
  return App.Utils.UID36ToUID20(App.Utils.generateUID36());
};


/**
 * Convert an UUID 36 (guid) into UUID 20 (docID)
 * @param {String} uid36
 * @returns {String}
 */
App.Utils.UID36ToUID20 = function (uid36)
{
  var guid = new Array(16);
  var uid20 = new Array(20);
  //
  guid[0] = parseInt(uid36.substr(6, 2), 16);
  guid[1] = parseInt(uid36.substr(4, 2), 16);
  guid[2] = parseInt(uid36.substr(2, 2), 16);
  guid[3] = parseInt(uid36.substr(0, 2), 16);
  guid[4] = parseInt(uid36.substr(11, 2), 16);
  guid[5] = parseInt(uid36.substr(9, 2), 16);
  guid[6] = parseInt(uid36.substr(16, 2), 16);
  guid[7] = parseInt(uid36.substr(14, 2), 16);
  guid[8] = parseInt(uid36.substr(19, 2), 16);
  guid[9] = parseInt(uid36.substr(21, 2), 16);
  guid[10] = parseInt(uid36.substr(24, 2), 16);
  guid[11] = parseInt(uid36.substr(26, 2), 16);
  guid[12] = parseInt(uid36.substr(28, 2), 16);
  guid[13] = parseInt(uid36.substr(30, 2), 16);
  guid[14] = parseInt(uid36.substr(32, 2), 16);
  guid[15] = parseInt(uid36.substr(34, 2), 16);
  //
  App.Utils.encode85(guid, uid20, 0);
  App.Utils.encode85(guid, uid20, 1);
  App.Utils.encode85(guid, uid20, 2);
  App.Utils.encode85(guid, uid20, 3);
  //
  return uid20.join("");
};


/**
 * Convert an UUID 20 (docID) into UUID 36 (guid)
 * @param {String} uid20
 * @returns {String}
 */
App.Utils.UID20ToUID36 = function (uid20)
{
  var docID = new Array(20);
  var guid = new Array(16);
  //
  for (var i = 0; i < 20; i++)
    docID[i] = uid20.charAt(i);
  //
  App.Utils.decode85(docID, guid, 0);
  App.Utils.decode85(docID, guid, 1);
  App.Utils.decode85(docID, guid, 2);
  App.Utils.decode85(docID, guid, 3);
  //
  var data12 = (guid[0] < 0 ? guid[0] + 256 : guid[0]) + (guid[1] < 0 ? guid[1] + 256 : guid[1]) * 256;
  var data11 = (guid[2] < 0 ? guid[2] + 256 : guid[2]) + (guid[3] < 0 ? guid[3] + 256 : guid[3]) * 256;
  var data2 = (guid[4] < 0 ? guid[4] + 256 : guid[4]) + (guid[5] < 0 ? guid[5] + 256 : guid[5]) * 256;
  var data3 = (guid[6] < 0 ? guid[6] + 256 : guid[6]) + (guid[7] < 0 ? guid[7] + 256 : guid[7]) * 256;
  var data4 = new Array(8);
  for (var i = 0; i < 8; i++)
    data4[i] = guid[8 + i];
  //
  var sb = "";
  sb += App.Utils.hexFull(data11, 4);
  sb += App.Utils.hexFull(data12, 4);
  sb += "-";
  sb += App.Utils.hexFull(data2, 4);
  sb += "-";
  sb += App.Utils.hexFull(data3, 4);
  sb += "-";
  for (var i = 0; i < 8; i++) {
    sb += App.Utils.hexFull((data4[i] < 0 ? data4[i] + 256 : data4[i]), 2);
    if (i === 1)
      sb += "-";
  }
  return sb;
};


/**
 * Encode85
 * @param {Array} dataIn
 * @param {Array} dataOut
 * @param {Integer} dw
 */
App.Utils.encode85 = function (dataIn, dataOut, dw)
{
  var j = 3 + dw * 4;
  var w = (dataIn[j] < 0 ? dataIn[j] + 256 : dataIn[j]) * 16777216;
  j--;
  w += (dataIn[j] < 0 ? dataIn[j] + 256 : dataIn[j]) * 65536;
  j--;
  w += (dataIn[j] < 0 ? dataIn[j] + 256 : dataIn[j]) * 256;
  j--;
  w += (dataIn[j] < 0 ? dataIn[j] + 256 : dataIn[j]);
  //
  dataOut[4 + dw * 5] = String.fromCharCode((w / 52200625) + 40);
  w = w % 52200625;
  dataOut[3 + dw * 5] = String.fromCharCode((w / 614125) + 40);
  w = w % 614125;
  dataOut[2 + dw * 5] = String.fromCharCode((w / 7225) + 40);
  w = w % 7225;
  dataOut[1 + dw * 5] = String.fromCharCode((w / 85) + 40);
  w = w % 85;
  dataOut[0 + dw * 5] = String.fromCharCode((w) + 40);
};


/**
 * Decode85
 * @param {Array} dataIn
 * @param {Array} dataOut
 * @param {Integer} dw
 */
App.Utils.decode85 = function (dataIn, dataOut, dw)
{
  var j = 4 + dw * 5;
  var w = (dataIn[j--].charCodeAt(0) - 40) * 52200625;
  w += (dataIn[j--].charCodeAt(0) - 40) * 614125;
  w += (dataIn[j--].charCodeAt(0) - 40) * 7225;
  w += (dataIn[j--].charCodeAt(0) - 40) * 85;
  w += (dataIn[j--].charCodeAt(0) - 40) * 1;
  //
  dataOut[3 + dw * 4] = (w / 16777216) & 0xFF;
  w = w % 16777216;
  dataOut[2 + dw * 4] = (w / 65536) & 0xFF;
  w = w % 65536;
  dataOut[1 + dw * 4] = (w / 256) & 0xFF;
  w = w % 256;
  dataOut[0 + dw * 4] = (w) & 0xFF;
};


/**
 * Return a fixed length hex string (filled with 0 if needed)
 * @param {Integer} num
 * @param {Integer} numDigits
 */
App.Utils.hexFull = function (num, numDigits)
{
  var sb = num.toString(16).toUpperCase();
  for (var i = sb.length; i < numDigits; i++)
    sb = sb.insert(0, '0');
  return sb;
};


/**
 * Try parse a partial date
 * @param {String} s - string to parse
 * @param {Int} add - quantity to add
 * @returns {Date}
 */
App.Utils.parsePartialDate = function (s, add)
{
  if (add === undefined)
    add = 0;
  //
  try {
    var d = null;
    //
    // Try format "yyyy"
    var d = App.moment(s, "YYYY", true);
    if (d.isValid()) {
      if (add !== 0)
        d.add(add, "y");
      return d.toDate();
    }
    //
    // Try formats "m/yyyy" or "mm/yyyy"
    d = App.moment(s, ["MM/YYYY", "M/YYYY", "MM-YYYY", "M-YYYY"], true);
    if (d.isValid()) {
      if (add !== 0)
        d.add(add, "M");
      return d.toDate();
    }
    //
    // Try formats "today", "yesterday" o "tomorrow"
    if (s.toLowerCase() === "today")
      d = App.moment(0, "HH");
    else if (s.toLowerCase() === "tomorrow")
      d = App.moment(0, "HH").add(1, "d");
    else if (s.toLowerCase() === "yesterday")
      d = App.moment(0, "HH").add(-1, "d");
    if (d.isValid()) {
      if (add !== 0)
        d = d.add(add, "d");
      return d.toDate();
    }
    //
    // Try format "jan", "feb", etc.
    if (s.length === 3) {
      d = App.moment(s + " " + App.moment().year(), "MMM YYYY", true);
      if (d.isValid()) {
        if (add !== 0)
          d.add(add, "M");
        return d.toDate();
      }
    }
    //
    // Try format "mmm yyyy"
    if (s.length === 7) {
      d = App.moment(s, "MMM YYYY", true);
      if (d.isValid()) {
        if (add !== 0)
          d.add(add, "M");
        return d.toDate();
      }
    }
  }
  catch (e) {
  }
  //
  return null;
};


/**
 * Function needed for synchronizing asynch methods
 * @param {Function} gen - generator function to be used as generator
 */
App.Utils.sync = function (gen)
{
  var iterable = gen(function (retVal, errVal) {
    var next = function () {
      // Restart simulated observers (if any)
      if (App.markObservers)
        App.markObservers();
      alreadyRunning = true;
      if (errVal)
        iterable.throw(errVal);
      else
        iterable.next(retVal);
      alreadyRunning = false;
    };
    //
    // Don't do it synchronously so that we
    // can call it directly if we need it inside our code
    if (alreadyRunning)
      setImmediate(next);
    else
      next();
  });
  //
  // First start
  var alreadyRunning = true;
  iterable.next();
  alreadyRunning = false;
};


/**
 * Clone an object
 * @param {Object} obj
 * @returns {Object}
 */
App.Utils.cloneObject = function (obj)
{
  var copy;
  //
  // Handle the 3 simple types, and null or undefined
  if (null === obj || "object" !== typeof obj)
    return obj;
  //
  // Handle Date
  if (obj instanceof Date) {
    copy = new Date();
    copy.setTime(obj.getTime());
    return copy;
  }
  //
  // Handle Array
  if (obj instanceof Array) {
    copy = [];
    for (var i = 0, len = obj.length; i < len; i++)
      copy[i] = App.Utils.cloneObject(obj[i]);
    return copy;
  }
  //
  // Handle Object
  if (obj instanceof Object) {
    copy = {};
    for (var attr in obj) {
      if (obj.hasOwnProperty(attr))
        copy[attr] = App.Utils.cloneObject(obj[attr]);
    }
    return copy;
  }
};


/**
 * Parse a string into an object without throwing exceptions
 * @param {Object|String} obj
 * @returns {Object}
 */
App.Utils.parseObject = function (obj)
{
  if (typeof obj === "string") {
    try {
      obj = JSON.parse(obj);
    }
    catch (ex) {
      obj = {};
    }
  }
  return obj;
};


/**
 * Encode html tag
 * @param {String} value
 * @param {Boolean} crlf
 * @param {Boolean} keepHtmlTags
 * @param {Boolean} removeIllegalXmlChars
 */
App.Utils.htmlEncode = function (value, crlf, keepHtmlTags, removeIllegalXmlChars)
{
  if (!value)
    return "";
  //
  var strval = "";
  var ns = value.length;
  for (var i = 0; i < ns; i++) {
    var ch = value.charAt(i);
    var cc = value.charCodeAt(i);
    //
    // If required, I do not add the illegal characters in XML
    // (http://www.w3.org/TR/2006/REC-xml-20060816/#charsets)
    if (removeIllegalXmlChars && !(ch === 0x9 || ch === 0xA || ch === 0xD
            || (ch >= 0x20 & ch <= 0xD7FF) || (ch >= 0xE000 && ch <= 0xFFFD)))
      continue;
    //
    if (keepHtmlTags) {
      if (cc > 126 || cc <= 13)
        strval = strval + "&#" + cc + ";";
      else
        strval = strval + ch;
    }
    else {
      switch (ch) {
        case "\"":
          strval = strval + "&quot;";
          break;
        case "&":
          strval = strval + "&amp;";
          break;
        case "<":
          strval = strval + "&lt;";
          break;
        case ">":
          strval = strval + "&gt;";
          break;
        case "\n":
          if (crlf)
            strval = strval + "<br/>";
          else
            strval = strval + "&#xA;";
          break;
        case "\r":
          if (!crlf)
            strval = strval + "&#xD;";
          break;
        default:
          if (cc > 126 || cc <= 13)
            strval = strval + "&#" + cc + ";";
          else
            strval = strval + ch;
          break;
      }
    }
  }
  return strval;
};


/**
 * Decode html tag
 * @param {String} value
 * @param {Boolean} allToken
 * @param {Boolean} decodeSpecialChar
 */
App.Utils.htmlDecode = function (value, allToken, decodeSpecialChar)
{
  if (!value)
    return "";
  //
  var strval = "";
  for (var i = 0; i < value.length; i++) {
    var ch = value.charAt(i);
    if (ch === "&" && i < value.length - 1 && value.charAt(i + 1) === "#") {
      // I seek ; and memorize what I have seen so far
      var sb = "";
      var ok = false;
      for (i += 2; i < value.length; i++) {
        if (value.charAt(i) !== ";")
          sb += value.charAt(i);
        else {
          ok = (sb.length > 0);
          break;
        }
      }
      //
      if (ok) {
        var c = "";
        if (sb.toLowerCase().startsWith("x"))
          c = new Number("0" + sb);
        else
          c = new Number(sb);
        //
        if (!isNaN(c))
          strval += String.fromCharCode(c);
        else
          ok = false;
      }
      //
      if (!ok) {
        strval += ch;
        i -= sb.length + 2;
      }
    }
    else if (allToken && ch === "&" && value.substr(i, 4) === "&lt;") {
      strval += '<';
      i += 3;
    }
    else if (allToken && ch === "&" && value.substr(i, 4) === "&gt;") {
      strval += '>';
      i += 3;
    }
    else if (allToken && ch === "&" && value.substr(i, 6) === "&quot;") {
      strval += '"';
      i += 5;
    }
    else if (allToken && ch === "&" && value.substr(i, 5) === "&amp;") {
      strval += '&';
      i += 4;
    }
    else if (decodeSpecialChar && ch === "&" && i < value.length - 1) {
      // I seek ; and memorize what I have seen so far
      var sb = "&";
      for (i += 1; i < value.length; i++) {
        if (value.charAt(i) !== ";")
          sb = sb + value.charAt(i);
        else {
          sb = sb + value.charAt(i);
          //
          switch (sb) {
            case "&quot;":
              strval = strval + "\"";
              break;
            case "&amp;":
              strval = strval + "&";
              break;
            case "&lt;":
              strval = strval + "<";
              break;
            case "&gt;":
              strval = strval + ">";
              break;
          }
          //
          break;
        }
      }
    }
    else
      strval = strval + ch;
  }
  //
  return strval;
};


/**
 * Translate a string
 * @param {String} text
 * @param {Object} params
 */
App.Utils.t = function (text, params)
{
  if (module && !App.View)
    App = require("./requireAll");
  //
  var s = text;
  //
  if (s.id) {
    var t;
    if (App.i18m && s.lc) {
      var m = App.i18m[s.lc];
      //
      // If I can't find a translation, I try to find it with the default language, if any
      if (!m && App.i18m.defaultLangCode)
        m = App.i18m[App.i18m.defaultLangCode];
      //
      if (m)
        t = m[s.id];
    }
    if (t)
      s = t;
    else
      s = s.t;
  }
  //
  if (params) {
    var k = Object.keys(params);
    for (var i = 0; i < k.length; i++) {
      var p = k[i];
      var v = params[p];
      if (typeof v === "number")
        v += "";      // Change it to string if needed
      s = s.split("@" + p).join(v || "");
    }
  }
  //
  return s;
};


/**
 * Return the names of the declared function arguments
 * @param {Function} func
 */
App.Utils.functArguments = function (func) {
  return (func + "")
          .replace(/[/][/].*$/mg, "") // strip single-line comments
          .replace(/\s+/g, "") // strip white space
          .replace(/[/][*][^/*]*[*][/]/g, "") // strip multi-line comments
          .split('){', 1)[0].replace(/^[^(]*[(]/, "") // extract the parameters
          .replace(/=[^,]+/g, "") // strip any ES6 defaults
          .split(',').filter(Boolean); // split & filter [""]
};


/**
 * Escapes an HTML string for security reasons (innerHTML)
 * @param {Obj} app - application object (for static invoking)
 * @param {String} str
 */
App.Utils.htmlEscape = function (app, str) {
  if (!str)
    return "";
  //
  // Convert object to String, using toString if defined or implicit conversion
  if (!str.replace)
    if (str.toISOString)
      str = str.toISOString();
    else if (str.toString)
      str = str.toString();
    else
      str = "" + str;
  //
  return str.replace(/&/g, '&amp;').replace(/"/g, '&quot;').replace(/'/g, '&#39;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
};


/**
 * Escapes an SQL string
 * @param {Obj} app - application object (for static invoking)
 * @param {String} str
 */
App.Utils.sqlEscape = function (app, str) {
  if (!str)
    return "";
  //
  if (module)
    App.Database = require("./database");
  return App.Database.quoteString(str);
};


/**
 * Convert an ArrayBuffer to exadecimal string
 * @param {ArrayBuffer} buffer
 */
App.Utils.bufferToHex = function (buffer) {
  if (module)
    return Buffer.from(new Uint8Array(buffer)).toString("hex");
  else
    return Array.prototype.map.call(new Uint8Array(buffer), function (x) {
      return("00" + x.toString(16)).slice(-2);
    }).join("");
};


/**
 * Convert an exadecimal string to ArrayBuffer
 * @param {String} hex
 */
App.Utils.hexToBuffer = function (hex) {
  var buf = new Uint8Array(hex.length / 2);
  for (var i = 0; i < hex.length; i += 2)
    buf[i / 2] = parseInt(hex.substring(i, i + 2), 16);
  return buf.buffer;
};


/**
 * Convert a buffer into ArrayBuffer
 * @param {Buffer} buffer
 */
App.Utils.bufferToArrayBuffer = function (buffer) {
  var ab = new ArrayBuffer(buffer.length);
  var view = new Uint8Array(ab);
  for (var i = 0; i < buffer.length; ++i)
    view[i] = buffer[i];
  return ab;
};


/**
 * Convert a string into ArrayBuffer
 * @param {String} str
 * @param {String} encoding
 */
App.Utils.stringToBuffer = function (str, encoding) {
  if (module)
    return App.Utils.bufferToArrayBuffer(Buffer.from(str, encoding));
  else if (typeof TextEncoder !== "undefined")
    return new TextEncoder(encoding).encode(str).buffer;
  else
    return new Uint8Array(Array.prototype.map.call(str, function (c) {
      return c.charCodeAt(0);
    })).buffer;
};


/**
 * Convert an ArrayBuffer to string
 * @param {ArrayBuffer} buffer
 * @param {String} encoding
 */
App.Utils.bufferToString = function (buffer, encoding) {
  if (module)
    return Buffer.from(new Uint8Array(buffer)).toString(encoding);
  else if (typeof TextDecoder !== "undefined")
    return new TextDecoder(encoding).decode(new Uint8Array(buffer));
  else
    return new Uint8Array(buffer).reduce(function (s, b) {
      return s + String.fromCharCode(b);
    }, "");
};


/**
 * Convert an ArrayBuffer to base64 string
 * @param {ArrayBuffer} buffer
 */
App.Utils.bufferToBase64 = function (buffer) {
  if (module)
    return Buffer.from(new Uint8Array(buffer)).toString("base64");
  else
    return btoa(String.fromCharCode.apply(null, new Uint8Array(buffer)));
};


/**
 * Convert an base64 string to ArrayBuffer
 * @param {String} base64
 */
App.Utils.base64ToBuffer = function (base64) {
  if (module)
    return App.Utils.bufferToArrayBuffer(Buffer.from(base64, "base64"));
  //
  return Uint8Array.from(atob(base64), function (c) {
    return c.charCodeAt(0);
  }).buffer;
};


/**
 * Convert an ArrayBuffer/String to base64 string
 * @param {ArrayBuffer/String} buffer
 * @param {Object} options
 */
App.Utils.toBase64 = function (buffer, options) {
  options = Object.assign({}, options);
  if (typeof buffer === "string")
    buffer = App.Utils.stringToBuffer(buffer, options.encoding);
  return App.Utils.bufferToBase64(buffer);
};


/**
 * Convert an exadecimal string to ArrayBuffer
 * @param {String} base64
 * @param {Object} options
 */
App.Utils.fromBase64 = function (base64, options) {
  options = Object.assign({}, options);
  var buff = App.Utils.base64ToBuffer(base64);
  return options.asBuffer ? buff : App.Utils.bufferToString(buff, options.encoding);
};


if (!module) {
  /**
   * Delay the exection of a function
   * @param {Function} funct - function to delay
   */
  if (!window.setImmediate) {
    setImmediate = function (funct) {
      return setTimeout(funct, 0);
    };
  }
  if (!window.clearImmediate) {
    clearImmediate = function (immediate) {
      return clearTimeout(immediate);
    };
  }
  //
  // Define translation function in the global scope for offline apps
  t = App.Utils.t;
}


//export module for node
if (module)
  module.exports = App.Utils;
