/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */

var App = App || {};

/**
 * @class A View Object
 * @param {Session} app - app that owns this view instance
 * */
App.View = function (app)
{
  if (module && !App.Database) {
    App = require("./requireAll");
    t = App.Utils.t;
  }
  if (!app)
    return;
  //
  this.elements = []; // Elements (RemElem/DataMap) child of this view
  this.app = app;     // App that owns this view instance
  //
  // Save id from IDE if set
  if (this.id)
    this.pid = this.id;
  this.id = App.Utils.generateUID24();
  //
  // Put the view in the element map
  this.app.eleMap[this.id] = this;
};


// Getter/setter of some properties
Object.defineProperties(App.View.prototype, {
  // View owner if this is a subview
  view: {
    get: function () {
      if (this.parent)
        return this.parent.view;
    },
    enumerable: true
  },
  parentView: {
    get: function () {
      return this.view;
    },
    enumerable: true
  },
  // Visibility of the view
  visible: {
    get: function () {
      if (this._visible === undefined)
        this._visible = true;
      return this._visible;
    },
    set: function (newValue) {
      if (this._visible !== newValue) {
        this._visible = newValue;
        if (this.elements) {
          for (var i = 0; i < this.elements.length; i++)
            this.elements[i].visible = newValue;
        }
      }
    },
    enumerable: true
  }
});


/**
 * Apply default parameters
 * @param {Object} viewDef
 */
App.View.prototype.applyDefaults = function (viewDef)
{
  if (viewDef) {
    for (var x in viewDef) {
      //
      // The ID property has a special meaning for subviews as it represents
      // the ID of this object at design time. Every instance should store in in the PID property
      // to allow "subview merging" work as expected.
      if (x === "id") {
        this.pid = viewDef[x];
        continue;
      }
      //
      // Don't copy some properties
      if (x === "child" || x === "req" || x === "class")
        continue;
      //
      // Copy other ones
      if (x.substr(0, 1) === "&")
        this[x.substring(1)] = App.Utils.t({id: viewDef[x].id + x, t: viewDef[x].t, lc: this.app.langCode});
      else
        this[x] = viewDef[x];
    }
    if (viewDef.child.elements) {
      // Apply default parameters also to my childs
      for (var i = 0; i < viewDef.child.elements.length; i++) {
        var cd = viewDef.child.elements[i];
        var b = false;
        for (var j = 0; j < this.elements.length; j++) {
          var e = this.elements[j];
          if (e.pid === cd.id || e.pid === cd.oid) {
            e.applyDefaults(cd);
            b = true;
            break;
          }
        }
        //
        if (!b) {
          // Not found: appending it
          var x;
          if (cd.child)
            x = this.createView(cd);
          else
            x = new App.RemElem(this, cd);
          this.elements.push(x);
          this[x.name] = x;
        }
      }
    }
  }
  //
  this.attachObservers(true, true);
};


/**
 * Tells something about this object to the DTT module
 */
App.View.prototype.getDebugInfo = function ()
{
  var info = {};
  info._class = "View";
  info._name = this.className;
  return info;
};


/**
 * Open the view
 * @param {Object} opt
 */
App.View.prototype.show = function (opt)
{
  if (this.isOpen) {
    this.visible = true;
  }
  else {
    // Do nothing, the user closed the form before the show (possibile only in the load event)
    if (this.closedBeforeOpening)
      return;
    //
    // Keep only the default options
    if (opt) {
      if (!this.opt)
        this.opt = {};
      var optKeys = Object.keys(opt);
      for (var i = 0; i < optKeys.length; i++) {
        var o = optKeys[i];
        if (o === "mode" || o === "modal" || o === "position" || o === "autoclose" || o === "timeout" || o === "ref" || o === "extcls" || o === "width" || o === "height")
          this.opt[o] = opt[o];
      }
      //
      // Copy the id of the view that opened me
      if (opt.owner)
        this.owner = opt.owner;
    }
    //
    // If the view (not a subview) is not in the app view list, add it
    // (it could happen if the user wrote:
    //    var f = new MyView();
    //    f.show()
    // and not
    //   app.show(MyView);
    if (!this.parent && this.app.views.indexOf(this) === -1)
      this.app.views.push(this);
    //
    // Tell the client to open the view
    var cmd = {id: "openView", cnt: this.getConfig(true)};
    this.app.sendCommands([cmd]);
    this.setIsOpen(true);
    this._visible = true;
  }
  //
  return this;
};


/**
 * Close the view
 * @param {Object} info
 */
App.View.prototype.close = function (info)
{
  if (this.closing)
    return;
  this.closing = true;
  //
  this.onClose(info, function (result, error) {
    if (result === false)
      return;
    //
    for (var i = 0; i < this.elements.length; i++)
      this.elements[i].close(true);
    //
    // Removing from map and view list
    delete this.app.eleMap[this.id];
    var idx = this.app.views.indexOf(this);
    if (idx >= 0)
      this.app.views.splice(idx, 1);
    //
    // Delete any observer on me
    if (App.deleteObservers)
      App.deleteObservers(this);
    //
    // Dereference static instance in my class
    if (this.app.defaultInstance[this.className] === this)
      delete this.app.defaultInstance[this.className];
    //
    // Notify onBack event to view that opened me
    if (this.owner) {
      var ow = this.app.eleMap[this.owner];
      if (ow)
        ow.onBack(this, info);
    }
    //
    // The close was called before the opening, so the show must do nothing; we remember that using a variable, and we must delete it in a timer because
    // for the default istance the 'show' is called two times (so we cannot delete the semaphore there)
    if (!this.isOpen) {
      this.closedBeforeOpening = true;
      setTimeout(function () {
        delete this.closedBeforeOpening;
      }, 0);
      return;
    }
    //
    this.isOpen = false;
    //
    // we use setImmediate to disengage closing lock after current microthread
    setImmediate(function () {
      this.closing = false;
    }.bind(this));
    //
    // Tell the client to close the view
    var cmd = {id: "closeView", cnt: {id: this.id}};
    this.app.sendCommands([cmd]);
  }.bind(this));
};


/**
 * Send a message to a view
 * @param {View} view - the recipient of the message
 * @param {Object} message - the message
 * @returns {Any Type}
 */
App.View.prototype.sendMessage = function (view, message)
{
  return view.onMessage(message);
};


/**
 * Activate the view
 */
App.View.prototype.activate = function ()
{
  this.onActivate();
  //
  // Tell the client to activate the view
  var cmd = {id: "activateView", content: {id: this.id}};
  this.app.sendCommands([cmd]);
};


/**
 * Get the properties of the view and its children
 * @param {bool} flReq - if true handles requirements of sub-views, otherwise only that of this view
 * @returns {Object}
 */
App.View.prototype.getConfig = function (flReq) {
  //
  // getting element configuration
  var el = [];
  for (var i = 0; i < this.elements.length; i++) {
    if (this.elements[i].getConfig) {
      var x = this.elements[i].getConfig();
      if (x !== undefined)
        el.push(x);
    }
  }
  //
  // Getting all requirements (including subviews)
  var all = this.requirements;
  if (flReq)
    this.allRequirements(all);
  //
  // Replacing theme parameters in requirements
  App.Utils.replaceObjProps(all, this.app.theme);
  //
  var r = {id: this.id, req: all, elements: el, type: "view", options: this.opt};
  return r;
};


/**
 * Return all the requirements (including subviews)
 * @param {Object} all - Map of all the requirements
 */
App.View.prototype.allRequirements = function (all)
{
  // Copy each requiremts if it does not already exist
  for (var r in this.requirements) {
    if (all[r] === undefined)
      all[r] = this.requirements[r];
  }
  //
  // Copy also requeriments of each children
  for (var i = 0; i < this.elements.length; i++) {
    if (this.elements[i].allRequirements)
      this.elements[i].allRequirements(all);
  }
};


/**
 * Event fired when the view is opened
 * @param {Object} options - options passed to the show method
 */
App.View.prototype.onLoad = function (options)
{
};


/**
 * Event fired when the view is closed
 * @param {Object} info - info passed to the close method
 * @param {Function} callback - function to be called at the end
 */
App.View.prototype.onClose = function (info, callback)
{
  callback();
};


/**
 * Event fired when the view receives a message
 * @param {Object} message - the message
 * @returns {Any Type}
 */
App.View.prototype.onMessage = function (message)
{
};


/**
 * Event fired when a view opened from me is closed
 * @param {Object} obj - info passed to the close method
 */
App.View.prototype.onBack = function (obj)
{
};


/**
 * Event fired when the view becomes the active
 */
App.View.prototype.onActivate = function ()
{
};


/**
 * Event fired when the device/screen resizes
 * @param {Object} event - event properties
 */
App.View.prototype.onResize = function (event)
{
};


/**
 * Append a child to the view
 * @param {RemElem} child - the element to append
 */
App.View.prototype.appendChild = function (child)
{
  this.insertBefore(child);
};


/**
 * Insert a child element before another element
 * @param {RemElem} child - the element to insert
 * @param {String} sibId - the id of the existing element. The new element is inserted before this element
 */
App.View.prototype.insertBefore = function (child, sibId)
{
  // Put the child in the elements array at the right position
  if (this.elements === undefined)
    this.elements = [];
  //
  var sib, idx = -1;
  if (sibId) {
    sib = this.app.eleMap[sibId];
    if (sib)
      idx = this.elements.indexOf(sib);
  }
  if (idx > -1)
    this.elements.splice(idx, 0, child);
  else
    this.elements.push(child);
  //
  // After attaching it, changes should be observed
  child.setParent(this, this);
  child.sendChanges(true);
  child.attachObservers(true, true);
  //
  // Tell the client to insert the child
  if (this.isOpen)
    this.app.sendCommands([{obj: this.id, id: "insertBefore", cnt: {child: child.getConfig(true), sib: sibId}}]);
};


/**
 * Insert a child element before another element, called by the IDE when the application is running and must be updated (AApp.updateCode)
 * @param {Object} childDef - the object representing the element to create and insert
 */
App.View.prototype.insertBeforeIDE = function (childDef)
{
  // Get the sibling object using the pid
  var ris = [];
  if (childDef.sib)
    this.getElementsByPid(childDef.sib, ris);
  var sib = null;
  if (ris.length > 0)
    sib = ris[0];
  //
  // The child can be an actual object or a string (node/netbeans), if is an object stringify it
  var ch = childDef.child;
  if (ch && typeof ch !== "string")
    ch = JSON.stringify(ch);
  //
  // Use a string to create the Object constructor rule
  var obj = "new App." + childDef.className + "(this," + ch + ", this);";
  try {
    // Create the real object and insert in the correct position
    obj = eval(obj);
    this.insertBefore(obj, sib);
  }
  catch (ex) {
  }
};


/**
 * Remove a view child
 * @param {RemElem} child - the element to remove
 */
App.View.prototype.removeChild = function (child)
{
  // If the element has no children, I can't remove anything
  if (this.elements === undefined)
    return;
  //
  // Search the child in the elements array and remove it
  var idx = this.elements.indexOf(child);
  if (idx > -1) {
    var el = this.elements[idx];
    el.close();
    this.elements.splice(idx, 1);
    //
    // Tell the client to remove the child
    if (this.isOpen) {
      var r = [{obj: this.id, id: "removeChild", cnt: child.id}];
      this.app.sendCommands(r);
    }
  }
};


/**
 * Remove a view child, called by the IDE when the application is running and must be updated (AApp.updateCode)
 * @param {Object} child - the object representing the element to remove (must use the pid to find the real element)
 */
App.View.prototype.removeChildIDE = function (child)
{
  // Use the PID to find all the objects related to the 'project object' and remove them
  var ris = [];
  this.getElementsByPid(child.id, ris);
  for (var i = 0; i < ris.length; i++)
    this.removeChild(ris[i]);
};


/**
 * Return a copy of the view
 * @param {Object} map of cloned elements
 * @returns {View}
 */
App.View.prototype.clone = function (map)
{
  var el = this.app.createView(this.className);
  if (map)
    map[this.id] = el;
  el.opt = this.opt;
  el.owner = this.owner;
  el.parent = this.parent;
  el._visible = this._visible;
  el.name = this.name;
  //
  el.elements = [];
  for (var i = 0; i < this.elements.length; i++) {
    var ne = this.elements[i].clone(map, el);
    el.elements.push(ne);
    //
    // Relink parent
    if (ne instanceof App.DataMap)
      ne.parentElement = el;
    else
      ne.parent = el;
    //
    // Relink direct child
    if (ne instanceof App.RemElem || ne instanceof App.View || ne instanceof App.DataMap)
      el[ne.name] = ne;
    //
    // Relink cloned listBox
    if (ne instanceof App.DataMap && ne.template && map)
      ne.schema.setListBox(map[this.elements[i].schema.listBox.id]);
  }
  //
  // Clone custom events and methods
  for (var k in this) {
    if (typeof this[k] === "function" && this[k] !== Object.getPrototypeOf(this)[k])
      el[k] = this[k];
  }
  //
  return el;
};

/**
 * Set the parent of the view
 * @param {View} view
 * @param {RemElem} parent
 */
App.View.prototype.setParent = function (view, parent)
{
  this.parent = parent;
  if (this.elements) {
    for (var i = 0; i < this.elements.length; i++) {
      this.elements[i].setParent(this, this);
    }
  }
};


/**
 * Sets the sendChg flag on each child remelem
 * @param {Boolean} flag
 */
App.View.prototype.sendChanges = function (flag)
{
  if (this.elements) {
    for (var i = 0; i < this.elements.length; i++) {
      if (this.elements[i].sendChanges)
        this.elements[i].sendChanges(flag);
    }
  }
};

/**
 * Attach the observers to each child remelem
 * @param {Boolean} flag
 * @param {Boolean} children
 */
App.View.prototype.attachObservers = function (flag, children)
{
  if (this.elements) {
    for (var i = 0; i < this.elements.length; i++) {
      if (this.elements[i].attachObservers)
        this.elements[i].attachObservers(flag, children);
    }
  }
};


/*
 * Returns all elements by pid
 * @param {String} pid
 * @param {Array} ris
 */
App.View.prototype.getElementsByPid = function (pid, ris)
{
  if (this.pid === pid)
    ris.push(this);
  if (this.elements) {
    for (var i = 0; i < this.elements.length; i++) {
      if (this.elements[i] && this.elements[i].getElementsByPid)
        this.elements[i].getElementsByPid(pid, ris);
    }
  }
};


/*
 * Returns all subviews
 * @param {Array} ris
 */
App.View.prototype.getSubviews = function (ris)
{
  if (this.elements) {
    for (var i = 0; i < this.elements.length; i++) {
      if (this.elements[i] instanceof App.View)
        ris.push(this.elements[i]);
      else if (this.elements[i].getSubviews)
        this.elements[i].getSubviews(ris);
    }
  }
};


/*
 * Set my isOpen property and of all my subviews
 * @param {Boolean} isOpen
 */
App.View.prototype.setIsOpen = function (isOpen)
{
  this.isOpen = isOpen;
  var ris = [];
  this.getSubviews(ris);
  for (var i = 0; i < ris.length; i++)
    ris[i].setIsOpen(isOpen);
};


/*
 * Informs all subview, datamap e do property that a document is changed
 * @param {Document} doc changed
 */
App.View.prototype.handleDocUpdate = function (doc)
{
  var options = {skip: false};
  this.onDocUpdate(doc, options);
  if (options.skip)
    return;
  //
  // Informs every datamap and document/collection property
  var keys = Object.keys(this);
  for (var i = 0; i < keys.length; i++) {
    if (this[keys[i]] instanceof App.Document || this[keys[i]] instanceof App.DataMap)
      this[keys[i]].handleDocUpdate(doc);
  }
  //
  // Informs every subview
  var subViews = [];
  this.getSubviews(subViews);
  for (var i = 0; i < subViews.length; i++)
    subViews[i].handleDocUpdate(doc);
};


/*
 * Event fired when a document is changed
 * @param {Document} doc changed
 * @param {Object} options
 */
App.View.prototype.onDocUpdate = function (doc, options)
{
};


//export module for node
if (module)
  module.exports = App.View;
