/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module */

var App = App || {};

/**
 * @class A WebAPIClient Object
 * @param {Session} app that owns this instance
 * @param {String} endpointUrl
 * if the object is created to extend a class the param req will be true and require
 * will be avoided. That's to prevent circular requires
 * */
App.WebApiEndpoint = function (app, endpointUrl)
{
  if (app === true)
    return;
  //
  if (module && !App.View)
    App = require("../requireAll");
  //
  this.app = app;
  this.endpointUrl = endpointUrl;
  //
  this.supportsHierarchyInsert = false;
  this.supportsSort = false;
  //
  this.formats = {};
  this.formats[App.Database.DataTypes.DATE] = "YYYY-MM-DD";
  this.formats[App.Database.DataTypes.DATETIME] = "YYYY-MM-DDTHH:mm:ss.SSSZ";
  this.formats[App.Database.DataTypes.TIME] = "HH:mm:ss.SSS";
};


App.WebApiEndpoint.actionTypes = {
  saveDocument: "saveDocument",
  loadDocument: "loadDocument",
  loadCollection: "loadCollection",
  loadChildCollection: "loadChildCollection",
  callMethod: "callMethod"
};


/**
 * Load remotely a collection
 * @param {Collection} coll
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.handleLoad = function (coll, options, callback)
{
  var done = function (result, error) {
    if (!error)
      result.setLoaded(options.childLevel, true);
    callback(result, error);
  };
  //
  switch (options.loadType) {
    case App.Document.loadTypes.load:
      if (this.getResourcePath(coll.rowTemplate, true))
        this.loadDocument(coll.rowTemplate, options, done);
      else
        this.loadCollection(coll, options, done);
      break;

    case App.Document.loadTypes.loadChildCollection:
      this.loadChildCollection(coll, options, done);
      break;

    case App.Document.loadTypes.loadCollection:
      this.loadCollection(coll, options, done);
      break;
  }
};


/**
 * Load a document
 * @param {Document} doc to load
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.loadDocument = function (doc, options, callback)
{
  var action = {obj: doc, options: options, type: App.WebApiEndpoint.actionTypes.loadDocument};
  var request = {path: this.getResourcePath(doc, true), method: "GET"};
  this.doHttpRequest(request, action, function (result, error) {
    if (error)
      return callback(null, error);
    //
    var opts = {};
    opts.action = action;
    opts.replacer = this.jsonParseReplacer.bind(this);
    //
    var coll = new App.Collection(this.app);
    coll.add(doc.class.fromJSON(this.app, result.body, opts));
    //
    callback(coll);
  }.bind(this));
};


/**
 * Load a collection
 * @param {Collection} coll
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.loadCollection = function (coll, options, callback)
{
  if (!this.supportsSort)
    options.clientSort = true;
  //
  var doc = coll.rowTemplate;
  var action = {obj: coll, options: options, type: App.WebApiEndpoint.actionTypes.loadCollection};
  var request = {path: this.getResourcePath(doc) + this.getQBEFilters(doc), method: "GET"};
  this.doHttpRequest(request, action, function (result, error) {
    if (error)
      return callback(null, error);
    //
    var opts = {};
    opts.action = action;
    opts.replacer = this.jsonParseReplacer.bind(this);
    opts.contentType = doc.class;
    //
    callback(App.Collection.fromJSON(this.app, result.body, opts));
  }.bind(this));
};


/**
 * Load a document's collection
 * @param {Collection} coll
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.loadChildCollection = function (coll, options, callback)
{
  var mdc = coll.parent.schema.getCollDefinition(coll.index);
  if (!mdc || !mdc.pks)
    return callback(null, new Error("Unable to load the collection"));
  //
  for (var i = 0; i < mdc.pks.length; i++)
    coll.rowTemplate[mdc.fks[i]] = coll.parent[mdc.pks[i]];
  //
  this.loadCollection(coll, options, callback);
};


/**
 * Save a document
 * @param {Document} doc to save
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.handleSave = function (doc, options, callback)
{
  if (doc.inserted && this.supportsHierarchyInsert)
    return this.saveDocument(doc, options, callback);
  //
  var phases = ["inserted", "updated", "deleted"];
  var nextPhase = function (result) {
    if (result === false || phases.length === 0)
      return callback(result);
    //
    this.doSavePhase(doc, options, phases.shift(), nextPhase);
  }.bind(this);
  nextPhase();
};


/**
 * Execute a saving phase for the document and his childs
 * @param {Document} doc
 * @param {String} phase
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 * */
App.WebApiEndpoint.prototype.doSavePhase = function (doc, options, phase, callback)
{
  var steps = [];
  //
  // Execute phase for the document
  if (doc[phase] && (
          (phase === "inserted" && !doc.deleted)
          || (phase === "updated" && !doc.deleted && !doc.inserted)
          || (phase === "deleted" && !doc.inserted))) {
    steps.push(function () {
      var opts = Object.assign({}, options);
      opts.childLevel = 0;
      App.WebApiEndpoint.prototype.saveDocument.call(this, doc, opts, nextStep);
    }.bind(this));
  }
  //
  // Execute phase for childs in parallel
  if (options.childLevel > 0) {
    var childs = doc.loadChildsToSave();
    var childsCount = childs.length;
    if (childsCount > 0) {
      steps.push(function () {
        var ok = true;
        var doneAll = function (result) {
          ok = ok && result;
          if (--childsCount === 0)
            nextStep(ok);
        };
        //
        var opts = Object.assign({}, options);
        opts.childLevel--;
        childs.forEach(function (child) {
          App.WebApiEndpoint.prototype.doSavePhase.call(this, child, opts, phase, doneAll);
        }.bind(this));
      }.bind(this));
    }
  }
  //
  if (phase === "deleted")
    steps.reverse();
  //
  var nextStep = function (result) {
    if (result === false || steps.length === 0)
      return callback(result);
    //
    steps.shift()();
  };
  nextStep(true);
};


/**
 * Save a document
 * @param {Document} doc to save
 * @param {Object} options
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.saveDocument = function (doc, options, callback)
{
  var done = function (result, error) {
    if (error) {
      doc.setError(error);
      return callback(false);
    }
    //
    doc.afterSaveReset(options.childLevel);
    callback(true);
  };
  //
  var request = {};
  if (doc.inserted && !doc.deleted)
    request.method = "POST";
  else if (doc.deleted)
    request.method = "DELETE";
  else if (doc.updated)
    request.method = "PUT";
  else
    return done();
  //
  request.path = this.getResourcePath(doc, request.method !== "POST");
  request.options = {};
  //
  var action = {obj: doc, options: options, type: App.WebApiEndpoint.actionTypes.saveDocument};
  if (request.method !== "DELETE") {
    var opts = {};
    opts.childLevel = options.childLevel;
    opts.action = action;
    opts.replacer = this.jsonStringifyReplacer.bind(this);
    opts.asString = false;
    request.options.body = doc.intoJSON(opts);
  }
  //
  this.doHttpRequest(request, action, function (result, error) {
    if (error)
      return done(null, error);
    //
    if (result.status === 200) {
      var md = doc.schema;
      for (var i = 0; i < md.properties.length; i++) {
        var mdp = md.properties[i];
        if (mdp.pk) {
          doc[mdp.name] = this.convertValueFromJSON(result.body, mdp);
          doc.reconnectFK(0);
          break;
        }
      }
    }
    //
    done();
  }.bind(this));
};


/**
 * Call remotely a method
 * @param {Document/Function} obj
 * @param {String/Function} method to call
 * @param {Array} args
 * @param {Object} options (only if obj is a Document)
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.handleCall = function (obj, method, args, options, callback)
{
  if (!((typeof obj === "function" ? obj.prototype : obj) instanceof App.Document))
    return callback(null, new Error("Invalid object"));
  //
  var action = {obj: obj, method: method, args: args, options: options, type: "callMethod"};
  var request = {};
  request.path = this.getResourcePath(obj, true);
  //
  if (typeof method === "function") {
    // TODO: function -> name
  }
  else if (typeof method === "string") {
    if (typeof obj[method] === "function")
      request.method = method;
  }
  //
  if (!request.method)
    return callback(null, new Error("Invalid method"));
  //
  var methMD = obj[method].md;
  if (args.length) {
    request.options = {};
    request.options.body = {};
    //
    var opts = {};
    opts.action = action;
    opts.replacer = this.jsonStringifyReplacer.bind(this);
    opts.asString = false;
    //
    if (args.length > methMD.arguments.length)
      return callback(null, new Error("Too many parameters: expected at most " + methMD.arguments.length));
    //
    for (var i = 0; i < args.length; i++) {
      var arg = args[i];
      var argMD = methMD.arguments[i];
      var argJSON = arg;
      if (typeof argMD.dataType === "string")
        argJSON = this.convertValueFromJSON(arg, {dstDT: argMD.dataType});
      else if (typeof argMD.dataType === "function") {
        if (arg && !(arg instanceof argMD.dataType))
          return callback(null, new Error("Type mismatch for parameter '" + argMD.name + "'"));
        //
        if (argMD.dataType === App.Collection)
          argJSON = arg.intoJSON(Object.assign({asObject: true}, opts));
        else if (argMD.dataType.prototype instanceof App.Document)
          argJSON = arg.intoJSON(opts);
      }
      //
      request.options.body[argMD.name] = argJSON;
    }
  }
  //
  // Do remote call
  this.doHttpRequest(request, action, function (result, error) {
    if (error)
      return callback(null, error);
    //
    if (result.status === 204)
      return callback();
    //
    if (typeof methMD.returnType === "string")
      return callback(this.convertValueFromJSON(result.body, {dstDT: methMD.returnType}));
    //
    var opts = {};
    opts.action = action;
    opts.replacer = this.jsonParseReplacer.bind(this);
    //
    if (methMD.returnType === App.Collection)
      return callback(App.Collection.fromJSON(this.app, result.body, Object.assign({contentType: methMD.contentType}, opts)));
    //
    if (methMD.returnType.prototype instanceof App.Document)
      return callback(methMD.returnType.fromJSON(this.app, result.body, opts));
    //
    callback(result.body);
  }.bind(this));
};


/**
 * Get the url path for a resource
 * @param {Document/Class} obj
 * @param {Boolean} withPK
 */
App.WebApiEndpoint.prototype.getResourcePath = function (obj, withPK)
{
  var path = "";
  var md = obj.schema;
  if (md.package)
    path += "/" + md.package.replace(/\./g, "/");
  path += "/" + md.className.replace(/\./g, "/");
  //
  if (withPK && obj instanceof App.Document) {
    var pks = [];
    for (var i = 0; i < md.properties.length; i++) {
      var mdp = md.properties[i];
      if (mdp.pk)
        pks.push(this.convertValueToJSON(obj[mdp.name], mdp));
    }
    //
    if (pks.indexOf(null) >= 0)
      return "";
    //
    path += "/" + pks.join("/");
  }
  return path;
};


/**
 * Get path for given QBE filters
 * @param {Document} doc
 */
App.WebApiEndpoint.prototype.getQBEFilters = function (doc) {
  return "";
};


/**
 * Do http request
 * @param {Object} request
 * @param {Object} action
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.doHttpRequest = function (request, action, callback)
{
  request.options = request.options || {};
  request.url = this.endpointUrl + request.path;
  //
  // Stringify body as JSON
  if (request.options.body && typeof request.options.body === "object") {
    request.options.body = JSON.stringify(request.options.body);
    request.options.bodyType = "application/json";
  }
  //
  // Fire beforeHttpRequest event
  this.fireBeforeHttpRequest(request, action, function (result, error) {
    if (error)
      return callback(null, error);
    //
    // Do the http request
    this.app.fs.url(request.url).request(request.method, request.options, function (response) {
      // Parse body if is JSON
      var contentType = (response.headers && response.headers["content-type"]) || "";
      if (contentType.startsWith("application/json")) {
        try {
          response.body = JSON.parse(response.body);
        }
        catch (e) {
          response.error = new Error("Error while parsing JSON response: " + e.message);
        }
      }
      //
      // Fire afterHttpRequest event
      this.fireAfterHttpRequest(request, action, response, function (result, error) {
        if (response.error || error)
          return callback(null, response.error || error);
        //
        if (response.status >= 400)
          return callback(null, new Error(response.body));
        //
        callback(response);
      });
    }.bind(this));
  }.bind(this));
};


/**
 * Fire beforeHttpRequest event
 * @param {Object} request
 * @param {Object} action
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.fireBeforeHttpRequest = function (request, action, callback)
{
  this.beforeHttpRequest(request, action, callback);
};


/**
 * Event fired before an http request
 * @param {Object} request
 * @param {Object} action
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.beforeHttpRequest = function (request, action, callback)
{
  callback();
};


/**
 * Fire afterHttpRequest event
 * @param {Object} request
 * @param {Object} action
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.fireAfterHttpRequest = function (request, action, response, callback)
{
  this.afterHttpRequest(request, action, response, callback);
};


/**
 * Event fired after an http request
 * @param {Object} request
 * @param {Object} action
 * @param {Object} response
 * @param {Function} callback - function to be called at the end
 */
App.WebApiEndpoint.prototype.afterHttpRequest = function (request, action, response, callback)
{
  callback();
};


/**
 * Replacer function called during parse from JSON
 * @param {Document/Collection} obj parsing
 * @param {Object} attr - {name: "", value: ""}
 * @param {Object} options
 */
App.WebApiEndpoint.prototype.jsonParseReplacer = function (obj, attr, options)
{
  if (obj instanceof App.Document) {
    // Call converter function for each property
    var md = obj.schema;
    for (var i = 0; i < md.properties.length; i++) {
      var mdp = md.properties[i];
      if (mdp.name === attr.name) {
        attr.value = this.convertValueFromJSON(attr.value, mdp);
        return true;
      }
    }
  }
  return false;
};


/**
 * Replacer function called during stringify to JSON
 * @param {Document/Collection} obj stringifing
 * @param {Object} attr - {name: "", value: ""}
 * @param {Object} options
 */
App.WebApiEndpoint.prototype.jsonStringifyReplacer = function (obj, attr, options)
{
  if (obj instanceof App.Document) {
    // Call converter function for each property
    var md = obj.schema;
    for (var i = 0; i < md.properties.length; i++) {
      var mdp = md.properties[i];
      if (mdp.name === attr.name) {
        // Skip values of type Document
        var skip = mdp.objectType;
        //
        // When updating skip unchanged values
        if (options.action.type === App.WebApiEndpoint.actionTypes.saveDocument
                && !obj.inserted && obj.updated && !(mdp.name in obj.originalValues))
          skip = true;
        //
        if (skip)
          attr.name = "";
        else
          attr.value = this.convertValueToJSON(obj[mdp.name], mdp);
        //
        return true;
      }
    }
  }
  //
  return false;
};


/**
 * Convert a value from JSON
 * @param {String} v
 * @param {DocumentProperty} mdp
 */
App.WebApiEndpoint.prototype.convertValueFromJSON = function (v, mdp)
{
  if (v === null || v === undefined)
    return null;
  //
  if (typeof v === "string" && App.Database.classDT(mdp.datatype) === App.Database.DataTypeClasses.DATETIME)
    return this.app.locale.convertValue(v, {format: this.formats[mdp.datatype], dstDT: mdp.datatype});
  //
  return v;
};


/**
 * Convert a value to JSON
 * @param {String} v
 * @param {DocumentProperty} mdp
 */
App.WebApiEndpoint.prototype.convertValueToJSON = function (v, mdp)
{
  if (v === null || v === undefined)
    return null;
  //
  if (v instanceof Date && App.Database.classDT(mdp.datatype) === App.Database.DataTypeClasses.DATETIME)
    return this.app.locale.convertValue(v, {format: this.formats[mdp.datatype], srcDT: mdp.datatype, dstDT: App.Database.DataTypes.VARCHAR});
  //
  return v;
};


//export module for node
if (module)
  module.exports = App.WebApiEndpoint;
