/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global module */

var App = App || {};

if (module)
  App.WebApiEndpoint = require("./webapiendpoint");

/**
 * @class A WebAPIClient Object
 * @param {Session} app that owns this instance
 * @param {String} endpointUrl
 * if the object is created to extend a class the param req will be true and require
 * will be avoided. That's to prevent circular requires
 * */
App.WebApiFoundation = function (app, endpointUrl)
{
  if (module && !App.View && app !== true)
    App = require("../requireAll");
  //
  App.WebApiEndpoint.call(this, app, endpointUrl);
  //
  this.supportsHierarchyInsert = true;
  this.formats[App.Database.DataTypes.DATETIME] = "YYYY-MM-DD HH:mm:ss";
  this.formats[App.Database.DataTypes.TIME] = "HH:mm:ss";
};

// Make App.WebApiFoundation extend App.WebApiEndpoint
App.WebApiFoundation.prototype = new App.WebApiEndpoint(true);


/**
 * Fire beforeHttpRequest event
 * @param {Object} request
 * @param {Object} action
 * @param {Function} callback - function to be called at the end
 */
App.WebApiFoundation.prototype.fireBeforeHttpRequest = function (request, action, callback)
{
  // Add child-level header for load and save
  if ((action.type.startsWith("load") || action.type.startsWith("save")) && "childLevel" in action.options) {
    request.options.headers = request.options.headers || {};
    request.options.headers["child-level"] = action.options.childLevel;
  }
  //
  // For update PK is required into body rather than into url
  if (action.type === App.WebApiEndpoint.actionTypes.saveDocument && request.method === "PUT")
    request.url = this.endpointUrl + this.getResourcePath(action.obj, false);
  //
  App.WebApiEndpoint.prototype.fireBeforeHttpRequest.apply(this, arguments);
};


/**
 * Replacer function called during parse from JSON
 * @param {Document/Collection} obj parsing
 * @param {Object} attr - {name: "", value: ""}
 * @param {Object} options
 */
App.WebApiFoundation.prototype.jsonParseReplacer = function (obj, attr, options)
{
  if (obj instanceof App.Document) {
    // Skip do_*
    if (attr.name.startsWith("do_")) {
      attr.name = "";
      return true;
    }
    //
    // Collections in Foundation are serialized with the name of content class
    var md = obj.schema;
    for (var i = 0; i < md.collections.length; i++) {
      var mdc = md.collections[i];
      if (mdc.contentType && mdc.contentType.split(".").pop() === attr.name) {
        attr.name = mdc.name;
        return true;
      }
    }
  }
  else if (obj instanceof App.Collection) {
    if (obj.rowTemplate && attr.name === obj.rowTemplate.schema.className.split(".").pop()) {
      attr.name = "rows";
      return true;
    }
  }
  //
  return App.WebApiEndpoint.prototype.jsonParseReplacer.apply(this, arguments);
};


/**
 * Replacer function called during stringify to JSON
 * @param {Document/Collection} obj stringifing
 * @param {Object} attr - {name: "", value: ""}
 * @param {Object} options
 */
App.WebApiFoundation.prototype.jsonStringifyReplacer = function (obj, attr, options)
{
  if (obj instanceof App.Document) {
    // When updating add pk values
    if (options.action.type === App.WebApiEndpoint.actionTypes.saveDocument && !obj.inserted && obj.updated)
    {
      var mdp = obj.schema.columns[attr.name];
      if (mdp && mdp.pk) {
        attr.value = this.convertValueToJSON(obj[mdp.name], mdp);
        return true;
      }
    }
    //
    // Collections in Foundation are serialized with the name of content class
    var md = obj.schema;
    for (var i = 0; i < md.collections.length; i++) {
      var mdc = md.collections[i];
      if (mdc.contentType && mdc.name === attr.name) {
        attr.name = mdc.contentType.split(".").pop();
        return true;
      }
    }
  }
  else if (obj instanceof App.Collection) {
    if (obj.rowTemplate && attr.name === "rows") {
      attr.name = obj.rowTemplate.schema.className.split(".").pop();
      return true;
    }
  }
  //
  return App.WebApiEndpoint.prototype.jsonStringifyReplacer.apply(this, arguments);
};


/**
 * Convert a value from JSON
 * @param {String} v
 * @param {DocumentProperty} mdp
 */
App.WebApiFoundation.prototype.convertValueFromJSON = function (v, mdp)
{
  if (v === null || v === undefined)
    return null;
  //
  switch (mdp.datatype) {
    case App.Database.DataTypes.BOOL:
      return (v === 1 ? true : false);

    case App.Database.DataTypes.UUID:
      if (mdp.maxlength === 20 && v.length === 36)
        return App.Utils.UID36ToUID20(v);
      break;
  }
  //
  return App.WebApiEndpoint.prototype.convertValueFromJSON.apply(this, arguments);
};


/**
 * Convert a value to JSON
 * @param {AnyType} v
 * @param {DocumentProperty} mdp
 */
App.WebApiFoundation.prototype.convertValueToJSON = function (v, mdp)
{
  if (v === null || v === undefined)
    return null;
  //
  switch (mdp.datatype) {
    case App.Database.DataTypes.BOOL:
      return (v ? 1 : 0);

    case App.Database.DataTypes.UUID:
      if (mdp.maxlength === 20 && v.length === 20)
        return App.Utils.UID20ToUID36(v);
      break;

    case App.Database.DataTypes.BLOB:
    case App.Database.DataTypes.JSON:
      return this.app.locale.convertValue(v, {srcDT: mdp.datatype, dstDT: App.Database.DataTypes.VARCHAR});
  }
  //
  return App.WebApiEndpoint.prototype.convertValueToJSON.apply(this, arguments);
};


/**
 * Create the path for the QBE filters
 * @param {Document} doc
 */
App.WebApiFoundation.prototype.getQBEFilters = function (doc) {
  var md = doc.schema;
  var path = "";
  var any = "";
  for (var i = 0; i < md.properties.length; i++) {
    var mdp = md.properties[i];
    if (!path && !any && mdp.pk)
      any = "?" + mdp.name + "=.";
    //
    var v = doc[mdp.name];
    if (v === undefined || v === null)
      continue;
    //
    path += (path ? "&" : "?") + mdp.name + "=" + v;
  }
  return path || any;
};


//export module for node
if (module)
  module.exports = App.WebApiFoundation;
