/*
 * Instant Developer Next
 * Copyright Pro Gamma Spa 2000-2016
 * All rights reserved
 */
/* global process, require, module, moment */

var App = App || {};
var Proxy;

if (module && !App.View) {
  Proxy = undefined;
  App = require("./requireAll");
  App.moment = require("moment-timezone");
  App.zlib = require("zlib");
  //
  // Inside Node declare the SQLite class (we don't import it 'cause it not supported
  // but if not there we can't detect the fact that we are using the wrong DB when using Node)
  App.SQLite = function () {
    throw new Error("SQlite not supported in Node");
  };
}

// Map of all active sessions
App.sessionMap = {};


App.msgTypeMap = {
  eval: "eval",
  connectionDB: "conDB",
  appmsg: "appmsg",
  deleteChildAppSession: "dcas",
  terminate: "terminate",
  appid: "appid",
  initApp: "initApp",
  sessionParams: "sespar",
  sendRestResponse: "rest",
  deleteSession: "delses",
  closeApp: "closeApp",
  evalResult: "evalResult",
  processRequest: "pr",
  log: "log",
  sync: "sync",
  syncBroad: "syncBroad",
  disconnectedClient: "disconnectedClient",
  deviceMsg: "dm",
  cloudConnectorMsg: "ccm",
  watchdogRequest: "wdreq",
  watchdogResponse: "wdres",
  install: "inst",
  installResult: "instres",
  testStart: "test",
  testStartResult: "testres",
  serverSession: "serverSession",
  createDBresult: "cdbres",
  cTokenOp: "ctop",
  changedAppParam: "chpar"
};


/**
 * Create a session for local testing
 */
App.createSession = function ()
{
  var s = App.sessionMap.S1;
  if (!s)
    s = new App.Session("S1");
  return s;
};


/**
 * Delete a session
 * @param {string} sid - session identifier
 */
App.deleteSession = function (sid)
{
  delete App.sessionMap[sid];
};


/**
 *  The app has been started
 */
App.startServerApp = function ()
{
  // I need to do something only if I'm on Node side or if I'm an offline app running in a browser/webView
  // If I'm running inside an IDE (loaded by ide2.min for instance) I don't need to do anything here
  if (!module && !App.isOffline())
    return;
  //
  var unhandledExceptionFnc = function (ex) {
    // If I'm an offline app "create" the ex object using the given arguments
    if (App.isOffline())
      ex = (arguments[4] || {message: arguments[0], stack: arguments[1] + ":" + arguments[2]}); // window.onerror(msg, url, lineNo, columnNo, error)
    //
    App.log("WARN", "Caught app exception: " + ex.message + " - " + ex.stack, "App.startServerApp");
    //
    // Get the main single-session if the app is inside an IDE or is offline
    var session;
    if (App.parentType === "ide" || App.isOffline())
      session = (App.sessionMap && Object.keys(App.sessionMap).length === 1 ? App.sessionMap[Object.keys(App.sessionMap)[0]] : null);
    //
    // Send exception to testAuto
    var sids = Object.keys(App.sessionMap);
    for (var i = 0; i < sids.length; i++) {
      var ses = App.sessionMap[sids[i]];
      if (ses.testAuto) {
        var e = [{id: "sendExceptionToTestAuto", cnt: {msg: ex.message, stack: ex.stack, uncought: true}}];
        ses.sendCommands(e);
      }
    }
    //
    // Log exception for analytics
    App.Analytics.logUncaughtException({msg: ex.message, stack: ex.stack, uncought: true});
    //
    // If I'm inside the IDE
    if (App.parentType === "ide") {
      // If I've found a valid session
      if (session) {
        // Tell the DTT that an error occurred and "close" the debug request so that everything goes to the IDE as soon as possible
        session.dttobj.addException("", ex);
        if (session.dttobj.currRequest)
          session.dttobj.currRequest.close();
      }
    }
    else if (!App.isOffline()) {  // If I'm a MASTER online app
      // Write an error message into console.error
      // If I'm insalling the app and there has been a problem the worker
      // needs to know that (see Worker::createChild)
      console.error("UncaughtException: " + ex + " - " + ex.stack);
      //
      // Remember that something went wrong...
      // If I'm installing this app I'll report this problem to the worker in the handleInstallMsg method
      App.uncaughtException = ex.message + "\nstack: " + ex.stack;
    }
    //
    // If the user implemented an uncaughtExceptionHandler
    if (App.uncaughtExceptionHandler) {
      // Tell him that a problem showed up and ask him if he wants do terminate the app
      var opt = {exception: ex};
      try {
        App.uncaughtExceptionHandler(opt);
      }
      catch (ex1) {
        App.log("WARN", "UncaughtExceptionHandler app exception: " + ex1, "App.startServerApp");
      }
      //
      // If requested, terminate the app
      if (opt.terminate) {
        App.log("INFO", "Application's process terminated by the user due to uncaught exception", "App.startServerApp");
        //
        // If the app is offline
        if (App.isOffline()) {
          // If I've a session
          if (session) {
            // Send terminate command to client
            var e = [{id: "terminate", content: {}}];
            session.sendCommands(e);
          }
        }
        else {    // If the app is online
          // If I'm inside the IDE stop the app
          if (App.parentType === "ide")
            process.send({type: App.msgTypeMap.closeApp, err: ex.message, stack: ex.stack});
          else // Master: Terminate the process
            process.exit();
        }
      }
    }
  };
  //
  // If the app is offline attach only global error handler
  if (App.isOffline())
    window.onerror = unhandledExceptionFnc;
  else {  // App is online
    process.on("disconnect", function () {    // Father process disconnected me -> I die
      App.log("DEBUG", "App Child terminated", "App.startServerApp");
      process.exit();
    });
    //
    process.on("uncaughtException", unhandledExceptionFnc);
    //
    process.on("message", function (m) {
      App.handleMessage(m);
    });
  }
};


/**
 * Log a new message
 * @param {string} level - message level (DEBUG, INFO, WARN, ERROR)
 * @param {string} message - text message
 * @param {string} sender - function that generated the message
 * @param {object} data - optional data to log
 */
App.log = function (level, message, sender, data)
{
  // If there is a logger (inside IDE), use it...
  if (App.logger) {
    data = (data ? JSON.parse(JSON.stringify(data)) : {});
    data.app = App.name;
    data.user = App.username;
    App.logger.log(level, message, sender, data);
  }
  else {    // No logger -> use standard console
    var s = new Date().toISOString() + " - " + level + " - " + message;
    //
    if (sender || data) {
      s += " - (";
      if (sender)
        s += "sender: " + sender;
      if (data)
        s += (sender ? ", " : "") + "data: " + JSON.stringify(data);
      s += ")";
    }
    //
    if (level === "ERROR" || App.isOffline())
      console.error(s);
    else
      console.log(s);
  }
};


/*
 * Handles a message received by parent process
 * @param {object} msg
 */
App.handleMessage = function (msg)
{
  switch (msg.type) {
    case App.msgTypeMap.eval:   // My parent process wants me to evaluate something
      try {
        eval(msg.cnt.text);     // jshint ignore:line
        process.send({type: App.msgTypeMap.evalResult, cnt: {id: msg.cnt.id, result: "OK"}});
      }
      catch (ex) {
        App.log("DEBUG", "Error while building the app (eval): " + ex.message, "App.handleMessage");
        process.send({type: App.msgTypeMap.evalResult, cnt: {id: msg.cnt.id, error: ex.message}});
      }
      break;

    case App.msgTypeMap.watchdogRequest:    // Tell my parent process I'm alive
      process.send({type: App.msgTypeMap.watchdogResponse});
      break;

    case App.msgTypeMap.connectionDB:   // Parent process tells me how to connect to any DB
      App.username = msg.username;
      App.defaultConString = msg.constring;
      App.remoteDBurls = msg.remoteDBurls;
      break;

    case App.msgTypeMap.initApp:
      // Init APP (sent by AApp.createChild (when IDE session) or Worker.init (when MASTER session))
      App.parentType = msg.sender;
      App.id = msg.appid;
      App.serverUrl = msg.url;
      App.name = msg.name;
      App.fs = App.FS.createFS({online: msg.online, rootPath: msg.path, publicUrl: msg.publicUrl});
      App.params = msg.params;
      //
      // Append appName to App.info structure
      // (N.B.: not for IDE session because App.info is created only for installed apps (MASTER))
      if (App.info && msg.path)
        App.info.appName = msg.path.substring(msg.path.lastIndexOf("/") + 1);   // Installed app
      //
      // If inside IDE, create a logger
      if (App.parentType === "ide") {
        App.Logger = require("../../../server/logger");
        App.logger = new App.Logger(this, "IDEAPP");
        App.log("DEBUG", "Inited app child", "App.handleMessage");
      }
      break;

    case App.msgTypeMap.appmsg:
      App.processAppMessage(msg);
      break;

    case App.msgTypeMap.deleteChildAppSession:
      App.handleDeleteChildApp(msg);
      break;

    case App.msgTypeMap.terminate:
      App.handleTerminate(msg);
      break;

    case App.msgTypeMap.sync:
      App.handleSync(msg);
      break;

    case App.msgTypeMap.syncBroad:
      App.handleSyncBroad(msg.cnt);
      break;

    case App.msgTypeMap.deviceMsg:
      App.phDeviceConnected = msg.cnt;
      break;

    case App.msgTypeMap.cloudConnectorMsg:
      App.handleCloudConnectorMsg(msg);
      break;

    case App.msgTypeMap.install:
      App.handleInstallMsg(msg);
      break;

    case App.msgTypeMap.createDBresult:
      App.handleCreateDBResultMsg(msg);
      break;

    case App.msgTypeMap.testStart:
      App.handleTestStartMsg(msg);
      break;

    case App.msgTypeMap.changedAppParam:
      App.handleChangedAppParamMsg(msg.cnt);
      break;
  }
};


/*
 * Process the App message received from the worker
 * @param {object} msg
 */
App.processAppMessage = function (msg)
{
  var i;
  //
  if (App.markObservers)
    App.markObservers();
  //
  // Get the session. If not there, create a new one
  var session = App.sessionMap[msg.sid];
  if (!session)
    session = new App.Session(msg.sid);
  //
  // Extract cookies from client message
  App.getCookiesFromClient(msg, session);
  //
  // Extract useful data from request object (see Config::processRun's newAppReq function)
  if (msg.request)
    session.getDataFromRequest(msg.request);
  //
  // If the request was a REST one
  if (msg.request && msg.request.query && msg.request.query.mode === "rest") {
    // Handle received files (if any)
    if (msg.request.files) {
      // Replace every file in the array with an InDe file
      // (NOTE: each file in the received array is an object like {fs: undefined, path: "uploaded/"+id+ext})
      var fsFiles = [];
      for (i = 0; i < msg.request.files.length; i++) {
        var f = msg.request.files[i];
        var file = session.fs.file(f.path, f.type, f.publicUrl);
        file.originalName = f.originalName;
        fsFiles.push(file);
      }
      msg.request.files = fsFiles;
    }
    else
      msg.request.files = [];
    //
    var pluginCB;
    try {
      if (msg.request.query.msgType) {
        // Route events to original plugins (used for upload events of camera and signaturepad)
        switch (msg.request.query.msgType) {
          case "device-camera":
            session.sendResponse(200);
            if (session.device && session.device.camera && session.device.camera.onUpload)
              pluginCB = function () {
                session.device.camera.onUpload(msg.request.files[0]);
              };
            break;
          case "signature-pad":
          case "input-upload":
            session.sendResponse(200);
            if (msg.request.query.objId) {
              if (session.eleMap[msg.request.query.objId] && session.eleMap[msg.request.query.objId].onUpload)
                pluginCB = function () {
                  session.eleMap[msg.request.query.objId].onUpload(msg.request.files);
                };
            }
            break;
          case "dropzone":
            session.sendResponse(200, JSON.stringify(msg.request.files));
            break;
        }
      }
      else
        pluginCB = function () {
          session.onCommand(msg.request);
        };
    }
    catch (e) {
      session.dttobj.addException("", e);
      //
      // Send exception to testAuto
      if (session.testAuto) {
        var e = [{id: "sendExceptionToTestAuto", cnt: {msg: e.message, stack: e.stack}}];
        session.sendCommands(e);
      }
    }
    return (pluginCB ? pluginCB() : undefined);
  }
  //
  // Start a new DTT request
  session.dttobj.beginRequest(msg.content);
  //
  // The static code is based on App.session
  App.session = session;
  session.cid = msg.cid;
  try {
    if (msg.content[0].id === "onStart") {
      session.dttobj.setRequestName("App.onStart");
      //
      // Delete the flag that tells me that the style have been initialized
      // (if the user presses REFRESH the session remains alive but the client needs a
      // new defineStyle, also when a new client starts for this session)
      delete session.styleDefined;
      session.processStart(msg.content, msg.request, msg.master);
      msg.content.splice(0, 1);
    }
    else if (msg.content[0].id === "onTerminate") {
      delete App.phDeviceConnected;
      session.dttobj.setRequestName("App.onTerminate");
      session.onTerminate();
      msg.content.splice(0, 1);
    }
    session.processRequest(msg.content);
  }
  catch (ex) {
    App.log("WARN", "App exception: " + ex.message, "App.processAppMessage", {sid: session.id, stack: ex.stack});
    //
    // Send exception to testAuto
    if (session.testAuto) {
      var e = [{id: "sendExceptionToTestAuto", cnt: {msg: ex.message, stack: ex.stack}}];
      session.sendCommands(e);
    }
    //
    // Save exception in analytics (if needed)
    if (session.analytics && session.analytics.enabled)
      session.analytics.logException({msg: ex.message, stack: ex.stack});
    //
    // If the user implemented the onError event
    if (session.onError) {
      // Tell him that a problem showed up and ask him if he wants do terminate the session
      var opt = {exception: ex};
      session.onError(opt, function (result, error) {
        if (error)
          App.log("WARN", "OnError app exception: " + error, "App.processAppMessage", {sid: session.id});
        //
        // If requested, terminate the session
        if (opt.terminate) {
          App.log("INFO", "Session terminated by the user due to session exception", "App.processAppMessage", {sid: session.id});
          //
          // If the app is offline
          if (App.isOffline()) {
            // Send terminate command to client
            var e = [{id: "terminate", content: {}}];
            session.sendCommands(e);
          }
          else {    // App is online
            // If I'm inside the IDE stop the app
            if (App.parentType === "ide")
              process.send({type: App.msgTypeMap.closeApp, err: ex.message, stack: ex.stack});
            else // Master: Terminate the session
              process.send({type: App.msgTypeMap.deleteSession, sid: session.id});
          }
        }
      });
    }
  }
  session.cid = null;
  session.CIDR = null;
  session.CIDE = null;
  //
  // End the current DTT request
  session.dttobj.endRequest();
  //
  App.session = null;
};


/*
 * Extract cookies from client message
 * @param {object} msg
 * @param {object} session
 */
App.getCookiesFromClient = function (msg, session)
{
  // Cookies come in two ways: from the http request for online app, if not "inplace",
  // or inside the start message if they are offline or inplace
  var cookies = (msg.content && msg.content.length > 0 && msg.content[0].cookies) || msg.cookies;
  if (cookies) {
    // If the session has no cookies, give it all cookies
    if (!session.cookies || Object.keys(session.cookies).length === 0) {
      //
      session.cookies = cookies;
      //
      // Copy startup parameters to device
      var sp = cookies["$device"];
      if (sp) {
        var spk = Object.keys(sp);
        for (var j = 0; j < spk.length; j++) {
          session.device[spk[j]] = sp[spk[j]];
        }
        delete cookies["$device"];
      }
      //
      // Copy info parameters to App.info struct
      sp = cookies["$info"];
      if (sp) {
        var spk = Object.keys(sp);
        for (var j = 0; j < spk.length; j++)
          App.info[spk[j]] = sp[spk[j]];
      }
    }
    else {
      // Session had cookies -> "merge" new cookies with previous ones
      // No startup parameters here
      var keys = Object.keys(cookies);
      for (var i = 0; i < keys.length; i++) {
        session.cookies[keys[i]] = cookies[keys[i]];
      }
    }
  }
  //
  // TODO remove after sending the new shell to the app stores
  // disable another cookie request (old cookies management)
  session.device.lscookiesNotYetAsked = false;
};


/*
 * Process the deleteChildApp message
 * (called when an app is stopped inside IDE or inside a worker)
 * @param {object} msg
 */
App.handleDeleteChildApp = function (msg)
{
  App.log("DEBUG", "Terminate child session (5)", "App.handleDeleteChildApp", {sid: msg.sid});
  //
  // Terminate and delete app session
  var session = App.sessionMap[msg.sid];
  if (session) {
    session.terminate({force: true});
    App.deleteSession(msg.sid);
  }
  else
    App.log("WARN", "Session not found during deleteChildAppSession from session (5)", "App.handleDeleteChildApp", {sid: msg.sid});
};


/*
 * Process the terminate message
 * (sent by Node.App terminate inside MASTER)
 * @param {object} msg
 */
App.handleTerminate = function (msg)
{
  // Terminate all sessions
  var sessionIds = Object.keys(App.sessionMap);
  for (var i = 0; i < sessionIds.length; i++) {
    var session = App.sessionMap[sessionIds[i]];
    session.terminate(msg.options);
  }
};


/*
 * Process the sync message
 * @param {object} msg
 */
App.handleSync = function (msg)
{
  var session = App.sessionMap[msg.sid.sidsrv];
  //
  // If the message is CONNECT handle connection, but only if I'm not in febe mode.
  // In fact, If I'm in febe mode, I have a session and that session is the one I want to dispatch message to
  if (msg.cnt.id === "connect" && !msg.sid.febe) {
    // Create a new session, if necessary
    if (!session)
      session = new App.Session(msg.sid.sidsrv);
    //
    // Extract useful data from request object (see Server::handleSyncMessage)
    if (msg.request)
      session.getDataFromRequest(msg.request);
  }
  //
  // Route message to proper session (if exists)
  if (!session)
    return App.log("WARN", "Can't handle SYNC message: session not found", "App.handleSync", msg);
  session.sync.onProxyMessage(msg);
  //
  // If the message is DISCONNECT handle disconnection
  if (msg.cnt.id === "disconnect" && !msg.sid.febe) {
    session.terminate({force: true});
    App.deleteSession(session.id);
  }
};


/*
 * Process the syncBroad message
 * @param {object} msg
 */
App.handleSyncBroad = function (msg)
{
  // Dispatch message to all sessions (except sender)
  var sessionIds = Object.keys(App.sessionMap);
  for (var i = 0; i < sessionIds.length; i++) {
    var session = App.sessionMap[sessionIds[i]];
    if (session.id !== msg.sid)
      session.sync.handleMessage(msg.msg);
  }
};


/*
 * Process the cloud connector message
 * @param {object} msg
 */
App.handleCloudConnectorMsg = function (msg)
{
  if (msg.cnt.type === "response") {
    if (msg.cnt.fs)
      App.RemoteFS.handleResponse(msg.cnt);
    else if (msg.cnt.plugin)
      App.RemotePlugin.handleResponse(msg.cnt);
    else
      App.RemoteDriver.handleResponse(msg.cnt);
  }
  else if (msg.cnt.type === "disconnect") {
    App.RemoteDriver.handleDisconnect(msg.cnt);
    App.RemoteFS.handleDisconnect(msg.cnt);
    App.RemotePlugin.handleDisconnect(msg.cnt);
  }
};


/*
 * Handle the install message
 * @param {object} msg
 */
App.handleInstallMsg = function (msg)   // jshint ignore:line
{
  // If there is no App.info it means that something went really bad!
  // The App.info structure is written at the end of the app.js file (this file) but,
  // somehow, the compiler did not reach the end of this file 'cause it stopped before (is that possible???
  // then why am I here and I can run this method????)...
  // Anyway, I can't continue...
  // If there was an "unhandled exception" is probably that the problem... otherwise report an unknown error
  if (!App.info) {
    var err = (App.uncaughtException || "Missing App.info structure");
    process.send({type: App.msgTypeMap.installResult, cnt: {err: err}});
    return;
  }
  //
  App.updateDbSchema(function (err) {
    if (err)
      process.send({type: App.msgTypeMap.installResult, cnt: {err: err}});
    else
      process.send({type: App.msgTypeMap.installResult, cnt: {appinfo: App.info}});
  });
};


/*
 * Update all datamodels' schema
 * @param {function} callback - (err)
 */
App.updateDbSchema = function (callback)    // jshint ignore:line
{
  // If there are no databases i've finished
  if (!App.databaseList || App.databaseList.length === 0)
    return callback();
  //
  var done = function () {
    // If this was the last one, report to callee
    if (++dbUpdated === App.databaseList.length) {
      if (!App.isOffline())
        App.log("INFO", "All database created/updated", "App.updateDbSchema");
      return callback(dbErrors.length ? dbErrors.join("\n") : undefined);
    }
  };
  //
  // Update all dbs (in parallel!)
  App.dbToInstall = {};
  var dbUpdated = 0;
  var dbErrors = [];
  for (var i = 0; i < App.databaseList.length; i++) {
    // Create/update db schema
    (function (dbname) {
      var db = new App[dbname](Proxy && Proxy.session);
      //
      // Online apps: don't create/update db with CC or with read-only schema
      if (!App.isOffline() && (db.phccn || db.readOnlySchema))
        return done();
      //
      // Offline apps: don't create/update db with driver=None (offline)
      if (App.isOffline() && db._driver === App.Database.DriverTypes.None)
        return done();
      //
      App.dbToInstall[dbname] = db;
      //
      if (App.isOffline())
        console.log(dbname + ".updateSchema");
      //
      // Create/update db schema
      db.updateSchema(function (result, err) {
        if (!App.isOffline())
          App.log("DEBUG", "Database created/updated", "App.updateDbSchema", {dbname: dbname});
        //
        // Close db connection
        db.closeConnection();
        //
        // If there are errors, I need to stop and report to callee
        if (err) {
          err = err.message || err;       // Note: Postgres return a complex error object
          if (!App.isOffline())
            App.log("WARN", "Database create/update failed: " + err, "App.updateDbSchema", {dbname: dbname});
          dbErrors.push("Can't create/update db " + dbname + ": " + err);
        }
        done();
      });
    }(App.databaseList[i]));  // jshint ignore:line
  }
};

/*
 * Handle a testStart message
 * @param {object} msg
 */
App.handleTestStartMsg = function (msg)
{
  App.updateDbSchema(function (err) {
    var result = {err: err, testAutoId: msg && msg.cnt ? msg.cnt.testAutoId : undefined};
    process.send({type: App.msgTypeMap.testStartResult, cnt: result});
  });
};


/*
 * Handle a CreateDBResult message
 * @param {object} msg
 */
App.handleCreateDBResultMsg = function (msg)
{
  // Search the DB by name
  var db = App.dbToInstall[msg.dbname];
  if (!db)
    return App.log("WARN", "Can't route CreateDBResult: datamodel not found", "App.handleCreateDBResultMsg", msg);
  //
  // If the DB has no callback I can't continue (it shouldn't happen!!)
  if (!db.createDBcallback)
    return App.log("WARN", "Can't route CreateDBResult: callback not found", "App.handleCreateDBResultMsg", msg);
  //
  App.log("DEBUG", "Database created", "App.handleCreateDBResultMsg", {dbname: db.name});
  //
  // Call and detach callback
  db.createDBcallback(msg.err);
  delete db.createDBcallback;
};


/*
 * Give the system an "uncaught exception handler" that will be fired if a
 * method thorws an exception that the system can't catch
 * @param {object} app
 * @param {function} handler - function that will be called by the system
 */
App.setExceptionHandler = function (app, handler)
{
  App.uncaughtExceptionHandler = handler;
};


/**
 * @class Session
 * @param {string} sid - session identifier
 */
App.Session = function (sid)
{
  // Offline apps and NetBeans
  if (!module)
    App.moment = moment;
  //
  this.id = sid;
  //
  this.eleMap = {};
  this.views = [];
  this.theme = {};
  this.style = {};
  this.defaultInstance = {};
  this.cmdCache = [];
  this.methodCache = [];
  this.popupBoxCallback = {};
  this.popupBoxContent = [];
  //
  this.device = new App.Device(this);
  if (module)
    this.phantom = new App.PhantomJS(this);
  //
  this.dttobj = new App.DTT(this);
  this.sync = new App.Sync(this);
  this.syncf = new App.SyncFoundation(this);
  this.analytics = new App.Analytics(this);
  this.locale = new App.Locale(this);
  this.head = new App.HeaderList(this);
  this.counter = 0;
  this.cookies = {};
  //
  App.sessionMap[sid] = this;
  this.initSession();
  //
  var pthis = this;
  //
  // Define console.test function
  console.test = function (name, value) {
    // Send console.test to test auto
    if (pthis.testAuto) {
      var e = [{id: "consoleTest", cnt: {name: name, value: value}}];
      pthis.sendCommands(e);
    }
    else if (pthis.dttobj && pthis.dttobj.enabled) // Redirect to console.log if debugger is enabled
      console.log({name: name, value: value});
  };
};


Object.defineProperty(App.Session.prototype, "fs", {
  get: function () {
    return App.fs;    // Every session uses the same FS "static" object
  }
});


Object.defineProperty(App.Session.prototype, "webRTCsettings", {
  get: function () {
    return this._webRTCsettings;
  },
  set: function (object) {
    this._webRTCsettings = object;
    var e = [{id: "setWebRTCProperties", cnt: object}];
    this.sendCommands(e);
  }
});


Object.defineProperty(App.Session.prototype, "sessionTimeout", {
  get: function () {
    return this._sessionTimeout || 60000;     // default: 1 min
  },
  set: function (newValue) {
    // Skip fake changes
    if (newValue === this._sessionTimeout)
      return;
    //
    // Store the value
    this._sessionTimeout = newValue;
    //
    // If I'm in the MASTER case, tell what happened to my owner (i.e. appSession)
    if (App.parentType === "master")
      process.send({type: App.msgTypeMap.sessionParams, sid: this.id, cnt: {sessionTimeout: this._sessionTimeout}});
  }
});


Object.defineProperty(App.Session.prototype, "debugEnabled", {
  get: function () {
    return this.dttobj.enabled;
  },
  set: function (newValue) {
    this.dttobj.enabled = newValue;
  }
});


/**
 * Generate session unique ids
 */
App.Session.prototype.generateCounter = function ()
{
  return (this.counter++).toString(36);
};


/**
 * Initialize the session
 */
App.Session.prototype.initSession = function ()
{
};


/**
 * Raised when the application starts
 * @param {request} request
 */
App.Session.prototype.onStart = function (request)    // jshint ignore:line
{
};


/**
 * Raised when the application enters background
 * @param {event} event
 */
App.Session.prototype.onPause = function (event)    // jshint ignore:line
{
};


/**
 * Raised when the application exits from background
 * @param {event} event
 */
App.Session.prototype.onResume = function (event)    // jshint ignore:line
{
};


/**
 * Raised when the application stops
 */
App.Session.prototype.onTerminate = function ()
{
};


/**
 * Raised when a new client is connecting
 * @param {string} settings
 */
App.Session.prototype.onConnectingClient = function (settings)    // jshint ignore:line
{
};


/**
 * Raised when the application receives a REST request
 * @param {request} request
 */
App.Session.prototype.onCommand = function (request)    // jshint ignore:line
{
  // Automatic response if the user does not handle REST requests
  this.sendResponse(500, "REST handler not implemented");
};


/**
 * Send back the response to a REST request
 * @param {int} code - HTTP status code
 * @param {string} text
 * @param {object} options
 */
App.Session.prototype.sendResponse = function (code, text, options)
{
  // Ask my parent (that owns the socket) to reply for me
  process.send({type: App.msgTypeMap.sendRestResponse, sid: this.id, code: code, text: text, options: (options || {})});
  //
  // If the session was not running (i.e. was not a "real/normal" session), stop it
  // (runningSession is set to TRUE in App.processStart when the "start" message is received by a client app)
  if (!this.runningSession)
    // Inside IDE stop the process after REST reply sent. Inside MASTER terminate gracefully
    if (App.parentType === "ide")
      process.exit(1);
    else
      this.terminate({force: true});
};


/**
 * Define application style
 */
App.Session.prototype.defineStyle = function ()
{
  this.styleDefined = true;
};


/**
 * Terminate the session
 * @param {object} options {force:(true/false), timeout:(ms), msg:("message")}
 */
App.Session.prototype.terminate = function (options)
{
  var pthis = this;
  var i;
  //
  // If the app is already terminating, do nothing
  if (this.terminating)
    return;
  //
  // Handle option "optionality"
  options = options || {};
  //
  // "Clone" options so that I can delete them when I've handled them
  options = JSON.parse(JSON.stringify(options));
  //
  // If there is a message, show it to the user
  if (options.msg) {
    this.alert(options.msg);
    delete options.msg;   // Message handled
  }
  //
  // If there is a timeout, procrastinate
  if (options.timeout) {
    this.killTimer = setTimeout(function () {
      delete options.timeout;   // Timeout handled
      pthis.terminate(options);
    }, options.timeout);
    //
    return;
  }
  //
  // Send termination message to the app
  var r = this.onTerminate();
  if (!options.force && r === false)
    return;   // session refuses to terminate and system is not forcing
  //
  // I'm terminating
  this.terminating = true;
  //
  // Close all views
  for (i = 0; i < this.views.length; i++)
    this.views[i].close();
  //
  // Close all database connections
  var keys = Object.keys(this.defaultInstance);
  for (i = 0; i < keys.length; i++) {
    var db = this.defaultInstance[keys[i]];
    if (db instanceof App.Database)
      db.closeConnection();
  }
  //
  // Send terminate command to client
  var e = [{id: "terminate", content: {}}];
  this.sendCommands(e);
  //
  if (module) { // server side on Node
    //
    // Wait 100ms before closing the session to wait for the terminate command reaching the client
    setTimeout(function () {
      if (App.parentType === "master")
        process.send({type: App.msgTypeMap.deleteSession, sid: this.id});
      else if (App.parentType === "ide") {
        if (!this.sync.isServerProxy())
          process.send({type: App.msgTypeMap.closeApp});
        else
          App.deleteSession(this.id);
      }
    }.bind(this), 100);
  }
  //
  // Stop analytics plugin and send data at the end of session
  if (this.analytics)
    this.analytics.terminate();
};


/*
 * Extract data from request
 * @param {object} req
 */
App.Session.prototype.getDataFromRequest = function (req)
{
  if (req.remoteAddress)
    this.device.ipAddress = req.remoteAddress;
};


/**
 * Substitute css parametric definitions
 * @param {String} css - the css containing parametric definitions
 */
App.Session.prototype.customizeStyle = function (css)
{
  var i, propk, prop, p, statements;
  //
  if (this.style) {
    propk = Object.keys(this.style);
    for (i = 0; i < propk.length; i++) {
      prop = propk[i];
      p = this.style[prop];
      if (p) {
        statements = p.replace(/(\n)+/g, "");
        if (statements) {
          var uncamelProp = App.Utils.unCamelCase(prop);
          css = App.Utils.replaceParameter(uncamelProp, statements, css, this.device, true);
        }
      }
    }
  }
  //
  if (this.theme) {
    propk = Object.keys(this.theme);
    for (i = 0; i < propk.length; i++) {
      prop = propk[i];
      p = this.theme[prop];
      if (p && p.replace) {
        statements = p.replace(/(\n)+/g, "");
        if (statements)
          css = App.Utils.replaceParameter(prop, statements, css, this.device);
      }
    }
  }
  //
  return css;
};


/**
 * Find the value of a style property
 * @param {string} prop - the property to decode
 * @param {bool} trans - translate or not
 */
App.Session.prototype.findStylePropValue = function (prop, trans)
{
  // Let's see if the property was defined in the style
  var pd = this.style[prop];
  if (pd === undefined) {
    if (trans) {
      var t = App.transStyleMap[prop];
      if (t)
        prop = t;
    }
    return prop;
  }
  //
  var statements = pd.replace(/(\n)+/g, "");
  var s = statements.split(";");
  for (var i = 0; i < s.length; i++) {
    var res = App.Utils.getStatementValue(s[i], this.device);
    if (res) {
      res = App.Utils.unCamelCase(res);
      return res;
    }
  }
};


/**
 * Execute a server request
 * @param {Object} req - Server request
 */
App.Session.prototype.processRequest = function (req)
{
  for (var i = 0; i < req.length; i++) {
    var cmd = req[i];
    //
    if (!cmd.obj) {
      //
      // Skip non valid commands
      if (!this.checkCommand(this, cmd))
        continue;
      //
      this.dttobj.setRequestName("App." + cmd.id);
      this[cmd.id](cmd.content);
      //
      if (cmd.id === "onPause" || cmd.id === "onResume") {
        if (this.analytics)
          this.analytics[cmd.id]();
        this.sync[cmd.id]();
      }
    }
    else {
      var ele = this.eleMap[cmd.obj];
      if (ele) {
        //
        // Skip non valid commands
        if (!this.checkCommand(ele, cmd))
          continue;
        //
        // Log click event for analytics
        if (this.analytics && cmd.id === "onClick" && ele.view)
          this.analytics.logEvent({eleId: ele.pid, eleName: ele.name, eleType: "element", eventName: "analytics_click", parentViewId: ele.view.pid});
        //
        if (ele[cmd.id]) {
          this.dttobj.setRequestName(ele.name + "." + cmd.id);
          ele[cmd.id](cmd.content);
        }
      }
      else {
        var ep = this.getElementsByPid(cmd.obj);
        for (var j = 0; j < ep.length; j++) {
          if (ep[j][cmd.id]) {
            this.dttobj.setRequestName(ep[j].name + "." + cmd.id);
            ep[j][cmd.id](cmd.content);
          }
        }
      }
    }
    //
    // If the command must be broadcast to other client sessions resend it
    if (cmd.bc)
      this.sendCommands([cmd], true);
  }
};


/**
 * Returns true if the cmd is valid
 * @param {RemElem} ele - element
 * @param {Object} cmd - client request
 */
App.Session.prototype.checkCommand = function (ele, cmd)
{
  // Command validation
  var ok = false;
  //
  if (ele === this && this.testAuto) {
    var testautoCmd = ["calculateTagResults", "setTaggingData", "saveTaggingData",
      "saveTestAuto", "initTestAuto", "playTestAuto", "stepForwardTestAuto", "getTestProperties"];
    if (cmd.id && testautoCmd.indexOf(cmd.id) > -1)
      ok = true;
  }
  //
  if (ele === this) {
    var appCmd = ["defineStyle", "popupBoxReturn", "chgProp", "closeAllGuests", "setTestAuto"];
    if (cmd.id && (cmd.id.startsWith("on") || appCmd.indexOf(cmd.id) > -1))
      ok = true;
  }
  else if (cmd.obj.startsWith("device-")) {
    var devCmd = ["setProp", "startGuests"];
    if (cmd.id && (cmd.id.endsWith("CB") || cmd.id.startsWith("on") || devCmd.indexOf(cmd.id) > -1))
      ok = true;
  }
  else if (ele instanceof App.View) {
    var viewCmd = ["close"];
    if (cmd.id && viewCmd.indexOf(cmd.id) > -1)
      ok = true;
  }
  else if (ele instanceof App.RemElem) {
    var eleCmd = ["cb", "chgProp", "chgData", "animateSwipeMenu", "endAnimation", "close"];
    if (cmd.id && (cmd.id.startsWith("on") || eleCmd.indexOf(cmd.id) > -1))
      ok = true;
  }
  //
  if (!ok) {
    return false;
  }
  //
  // Status validation
  if (cmd.id === "onClick" || cmd.id === "onDblClick") {
    if (ele.disabled)
      return false;
  }
  if (cmd.id === "onChange" || cmd.id === "onInput") {
    if (ele.disabled || ele.readOnly)
      return false;
  }
  if (cmd.id === "chgProp" && cmd.content && cmd.content.name === "value") {
    if (ele.disabled || ele.readOnly)
      return false;
  }
  //
  return true;
};


/**
 * * Process start message
 * @param {obj} dp device property
 * @param {obj} request
 * @param {type} master
 */
App.Session.prototype.processStart = function (dp, request, master)
{
  // The cide and cidr are set in order to manage the situation when more clients are connected
  this.CIDE = true;
  this.CIDR = this.cid;
  if (!master && this.runningSession) {
    var settings = {};
    settings = {clientRole: "ro", ownerStream: "a", guestStream: "av", ownerPeerId: this.ownerPeerId, webRTCsettings: this.webRTCsettings};
    //
    // If I was started from the IDE, client role must be read/write and webrtc connection there must not be
    if (App.parentType === "ide")
      settings = {clientRole: "rw", ownerStream: "", guestStream: "", ownerPeerId: this.ownerPeerId, webRTCsettings: this.webRTCsettings};
    //
    // Fire onConnectingClient event
    this.onConnectingClient(settings);
    //
    // Set custom settings or default ones for client
    this.setClientProperties(settings, "guest");
    //
    if (this.multiClient === undefined) {
      // Unset CIDR and CIDE in order to call setClientProperties on all clients. After call reset them to old values
      var cidr = this.CIDR;
      var cide = this.CIDE;
      this.CIDR = null;
      this.CIDE = null;
      //
      // This is the first client connected with this session, so client 0 has to open its webrtc component
      this.setClientProperties(settings, "owner");
      this.multiClient = 1;
      this.CIDR = cidr;
      this.CIDE = cide;
    }
    else
      this.multiClient++;
  }
  //
  // Refresh UI of the new client
  if (this.runningSession) {
    this.defineStyle();
    this.refreshingClient = true;
    for (var i = 0; i < this.views.length; i++) {
      this.views[i].setIsOpen(false);
      this.views[i].show();
      this.refreshingClient = false;
    }
  }
  else {
    // This is a "true" running session (not REST)
    this.runningSession = true;
    //
    if (this.device)
      this.device.setProp(dp[0].content);
    //
    // Set server session's name
    if (dp[0].name) {
      this.isSS = true;
      this.sessionName = dp[0].name;
    }
    //
    // Extract app params from app cookies
    if (!module && this.cookies["app_params"]) {
      try {
        App.params = JSON.parse(this.cookies["app_params"]);
      }
      catch (ex) {
        App.log("WARN", "Unable to read app parameters: " + ex, "App.Session.processStart", {params: this.cookies["app_params"]});
      }
    }
    //
    // For application offline before I must to update db schema
    if (App.isOffline())
      App.updateDbSchema(function () {
        this.onStart(request);
      }.bind(this));
    else
      this.onStart(request);
  }
  //
  // Set presence of physical device
  this.setPhDevice(App.phDeviceConnected);
};


/**
 * Cache a command collection. it will be sent to client next tick
 * @param {Array} cmds - array of commands to send
 * @param {Boolean} isMethod - if true the commands are calls to methods
 */
App.Session.prototype.sendCommands = function (cmds, isMethod)
{
  // Translate only if online
  if (!App.isOffline())
    this.translate(cmds);
  //
  var push = true;
  var last;
  if (!isMethod && this.cmdCache.length > 0) {
    last = this.cmdCache[this.cmdCache.length - 1];
    push = (last.cidr !== this.CIDR || last.cide !== this.CIDE);
  }
  else if (isMethod && this.methodCache.length > 0) {
    last = this.methodCache[this.methodCache.length - 1];
    push = (last.cidr !== this.CIDR || last.cide !== this.CIDE);
  }
  //
  if (push) {
    if (isMethod)
      this.methodCache.push({cide: this.CIDE, cidr: this.CIDR, cmds: cmds});
    else
      this.cmdCache.push({cide: this.CIDE, cidr: this.CIDR, cmds: cmds});
  }
  else {
    for (var i = 0; i < cmds.length; i++)
      last.cmds.push(cmds[i]);
  }
  //
  if (!this.immediateID) {
    var pthis = this;
    if (module || App.isOffline())
      this.immediateID = setImmediate(function () {
        pthis.sendCmdCache();
      });
    else {
      // Netbeans debugger is not able to dispatch timeout accordingly
      this.sendCmdCache();
    }
  }
};


/**
 * Sends cached commands to client
 */
App.Session.prototype.sendCmdCache = function ()
{
  var pthis = this;
  this.immediateID = undefined;
  //
  // Reset last remelem command as it will be sent to the client
  App.RemElem.lastr = undefined;
  //
  this.cmdCache = this.cmdCache.concat(this.methodCache);
  this.methodCache = [];
  //
  // If I'm on the server side
  if (module) {
    if (this.cmdCache.length > 0) {
      var cmd = this.cmdCache[0];
      this.cmdCache.splice(0, 1);
      //
      // Send the message to the process owner (AApp.js on IDE server, or worker.js on MASTER server)
      process.send({type: App.msgTypeMap.processRequest, sid: pthis.id, cidr: cmd.cidr, cide: cmd.cide, content: cmd.cmds});
      if (pthis.cmdCache.length > 0)
        pthis.sendCmdCache();
    }
  }
  else if (this.proxy) { // Local: direct client access
    for (var i = 0; i < this.cmdCache.length; i++)
      this.proxy.send(this, this.cmdCache[i].cmds, this.cmdCache[i].cidr);
    this.cmdCache = [];
  }
  else
    App.log("WARN", "App.Session.sendCmdCache: No proxy set", "Session.sendCmdCache");
};


/**
 * Compress commands
 * @param {Array} cmds
 */
App.Session.prototype.translate = function (cmds)
{
  for (var i = 0; i < cmds.length; i++) {
    var t = App.transCmdMap[cmds[i].id];
    if (t)
      cmds[i].id = t;
  }
};


/**
 * Tells something about this object to the DTT module
 */
App.Session.prototype.getDebugInfo = function ()
{
  var info = {};
  info._class = "Session";
  info.id = this.id;
  return info;
};


/**
 * Show an alert message on the client
 * @param {String} txt
 */
App.Session.prototype.alert = function (txt)
{
  if (typeof txt === "object") {
    try {
      txt = JSON.stringify(txt);
    }
    catch (ex) {
      txt = "object";
    }
  }
  if (this.device.isInsideShell())
    this.device.notification.alert(txt);
  else {
    var e = [{id: "alert", cnt: txt}];
    this.popupBoxContent.push(txt);
    this.sendCommands(e);
  }
};


/**
 * Show a popup on the client
 * @param {object} options
 * @param {function} callback
 */
App.Session.prototype.popup = function (options, callback)
{
  // Generate callback reference
  var id = App.Utils.generateUID24();
  this.popupBoxCallback[id] = callback;
  this.popupBoxContent.push(options ? JSON.stringify(options) : "");
  //
  var e = [{id: "popup", cnt: {options: options, cbId: id}}];
  this.sendCommands(e);
};


/**
 * Sleep
 * @param {int} ms
 * @param {function} callback
 */
App.Session.prototype.sleep = function (ms, callback)
{
  setTimeout(callback, ms);
};


/**
 * setCookie
 * @param {string} name
 * @param {string} value
 * @param {objects} options
 */
App.Session.prototype.setCookie = function (name, value, options)
{
  if (!options)
    options = {};
  //
  // Update the cookie in the cookie map (if the cookie is expired, delete it)
  if (options.exdays <= 0)
    delete this.cookies[name];
  else
    this.cookies[name] = value;
  //
  var e = [{id: "setCookie", cnt: {name: name, value: value, options: options}}];
  this.sendCommands(e);
};


/**
 * Open url on new window on the client
 * @param {String} href
 * @param {string} target
 * @param {string} options
 * @param {function} callback
 */
App.Session.prototype.open = function (href, target, options, callback)
{
  // If we pass an object File from fs to an offline device, say the device to open it from its file system.
  var e;
  if (typeof href !== "string" &&
          typeof href.path === "string" &&
          typeof href.driver !== "undefined" &&
          typeof href.driver.type === "string" &&
          typeof href.driver.name === "string" &&
          href.driver.name === "shelldriver") {
    e = [{id: "open", cnt: {href: href.path, target: target, options: options, shelldriver: {type: href.driver.type}}}];
  }
  else
    e = [{id: "open", cnt: {href: href, target: target, options: options}}];
  //
  if (callback) {
    var id = App.Utils.generateUID24();
    this.popupBoxCallback[id] = callback;
    e[0].cnt.cbId = id;
  }
  //
  this.sendCommands(e);
};


/**
 * Show a confirm message on the client
 * @param {String} txt - the text to display in the dialog box
 * @param {Function} callback
 */
App.Session.prototype.confirm = function (txt, callback)
{
  if (typeof txt === "object") {
    try {
      txt = JSON.stringify(txt);
    }
    catch (ex) {
      txt = "object";
    }
  }
  //
  if (this.device.isInsideShell()) {
    this.device.notification.confirm(txt, undefined, undefined, function (res) {
      if (callback)
        callback(res === 1);
    });
  }
  else {
    // Generate callback reference
    var id = App.Utils.generateUID24();
    this.popupBoxCallback[id] = callback;
    this.popupBoxContent.push(txt);
    //
    var e = [{id: "confirm", cnt: {txt: txt, cbId: id}}];
    this.sendCommands(e);
  }
};


/**
 * Show a prompt message on the client
 * @param {String} txt - the text to display in the dialog box
 * @param {String} def - the default input text
 * @param {Function} callback
 */
App.Session.prototype.prompt = function (txt, def, callback)
{
  if (typeof txt === "object") {
    try {
      txt = JSON.stringify(txt);
    }
    catch (ex) {
      txt = "object";
    }
  }
  if (typeof def === "object") {
    try {
      def = JSON.stringify(def);
    }
    catch (ex) {
      def = "object";
    }
  }
  //
  // If the second argument is a function and the third argument is undefined
  // assume that the default was not indicated and that the second argument is the callback
  if (typeof def === "function" && !callback) {
    callback = def;
    def = undefined;
  }
  //
  if (this.device.isInsideShell()) {
    this.device.notification.prompt(txt, undefined, undefined, def, function (res) {
      if (callback) {
        if (res.buttonIndex === 1)
          callback(res.input1);
        else
          callback(null);
      }
    });
  }
  else {
    // generate callback reference
    var id = App.Utils.generateUID24();
    this.popupBoxCallback[id] = callback;
    this.popupBoxContent.push(txt);
    //
    var e = [{id: "prompt", cnt: {txt: txt, def: def, cbId: id}}];
    this.sendCommands(e);
  }
};


/**
 * Called when a popup box (confirm or prompt or open) returns
 * @param {String|Boolean} response
 */
App.Session.prototype.popupBoxReturn = function (response)
{
  var callback = this.popupBoxCallback[response.cbId];
  if (callback) {
    // I can't just call the callback... the confirm is an asynchronous
    // method and uses Generators... Calling it directly gives an error
    setImmediate(function () {
      callback(response.res);
    });
    //
    delete this.popupBoxCallback[response.cbId];
    this.popupBoxContent.pop();
  }
  //
  this.sendCommands([{id: "closePopup"}]);
  //
  return response;
};


/**
 * Called when something bad happened while updating a client object
 * @param {Object} info - error details
 */
App.Session.prototype.onUpdateElementError = function (info)
{
  var ele = this.eleMap[info.id];
  this.dttobj.addException((ele ? ele.pid : null), info.error);
  //
  // Send exception to testAuto
  if (this.testAuto) {
    var e = [{id: "sendExceptionToTestAuto", cnt: {msg: info.error.message, stack: info.error.stack, uncought: true}}];
    this.sendCommands(e);
  }
};


/**
 * Show a view
 * @param {Object} view
 * @param {Object} options
 */
App.Session.prototype.show = function (view, options)
{
  // Push stlye definition before opening one
  this.defineStyle();
  //
  var ViewCls = view;
  var x = new ViewCls(this, undefined, options);
  this.views.push(x);
  //
  // Log view show for analytics
  if (this.analytics)
    this.analytics.logEvent({eleId: x.pid, eleName: x.className, eleType: "page", eventName: "analytics_show", parentViewId: x.pid});
  //
  x.show(options);
  return x;
};


/**
 * Create a new object view
 * @param {string|object} viewDef
 */
App.Session.prototype.createView = function (viewDef)
{
  var lib = App;
  //
  // CN is the qualified class name to create
  var cn = (typeof viewDef === "string") ? viewDef : viewDef.class;
  //
  // Starting from APP dig down to the class path
  var cp = cn.split(".");
  for (var i = 0; i < cp.length - 1; i++)
    lib = lib[cp[i]];
  //
  // Creating the class from the full qualified library
  return new lib[cp[cp.length - 1]](this, (typeof viewDef === "string") ? undefined : viewDef);
};


/**
 * Adds an object to the history
 * @param {object} obj
 */
App.Session.prototype.historyPush = function (obj)
{
  var cmd = {id: "historyPush", cnt: obj};
  this.sendCommands([cmd]);
};


/**
 * Called then an object is removed from the history
 * @param {object} obj
 */
App.Session.prototype.onHistoryPop = function (obj)   // jshint ignore:line
{
};


/**
 * Called when the app is resized
 * @param {object} event
 */
App.Session.prototype.onResize = function (event)
{
  // Inform every open view
  for (var i = 0; i < this.views.length; i++)
    this.views[i].onResize(event);
};


/*
 * Returns all elements by pid
 * @param {String} pid
 */
App.Session.prototype.getElementsByPid = function (pid)
{
  var ris = [];
  for (var i = 0; i < this.views.length; i++)
    this.views[i].getElementsByPid(pid, ris);
  return ris;
};


/*
 * Create link to allow clients to connect with session's owner
 */
App.Session.prototype.requestConnectionUrl = function ()
{
  // If I'm an IDE session, don't use CTOKEN
  if (App.parentType === "ide")
    return App.serverUrl + "/app/client/preview.html?appUrl=/" + this.id + "/" + App.id.replace(/\//g, "-") + "/run";
  //
  // App session -> create a new ctoken
  var ctoken = App.Utils.generateUID36();
  //
  // Send ctoken to server (so that it can wait for the new user to come)
  process.send({type: App.msgTypeMap.cTokenOp, sid: this.id, cnt: {op: "add", ctoken: ctoken}});
  //
  return App.serverUrl + "/" + App.name + "/?sid=" + this.id + "&ctoken=" + ctoken;
};


/*
 * Send command to clients to notify them to set they own webrtc properties
 * @param {Object} settings
 * @param {String} client - client type (owner or guest)
 */
App.Session.prototype.setClientProperties = function (settings, client)
{
  var e = [{id: "setClientProperties", cnt: {settings: settings, client: client}}];
  this.sendCommands(e);
};


/**
 * Called when a property changes
 * @param {Object} pv property and value
 */
App.Session.prototype.chgProp = function (pv)
{
  //Start multiclient session when peerId property changes
  if (pv.name === "ownerPeerId") {
    this.ownerPeerId = pv.value;
    var e = [{id: "callOwner", cnt: this.ownerPeerId}];
    this.sendCommands(e);
  }
};


/**
 * Called when owner closes its rtc popup
 */
App.Session.prototype.closeAllGuests = function ()
{
  this.multiClient = undefined;
  this.ownerPeerId = undefined;
  //
  // Send command to guests to disconnect themselves
  var e = [{id: "closeGuest"}];
  this.sendCommands(e);
};


/**
 * Called when a guest closes its rtc popup
 */
App.Session.prototype.onDisconnectingGuest = function ()
{
  this.multiClient--;
  //
  // If the last guest disconnect, owner closes its rtc popup
  if (this.multiClient === 0) {
    this.multiClient = undefined;
    this.ownerPeerId = undefined;
    var e = [{id: "closeOwnerRtc"}];
    this.sendCommands(e);
  }
};


/**
 * Return the default instance of a class (and create it if not exists)
 * @param {String} name
 * @param {Object} options
 */
App.Session.prototype.getDefaultInstance = function (name, options)
{
  if (!this.defaultInstance[name]) {
    var DbCls = App.Session.getClassByName(name);
    if (DbCls.prototype instanceof App.Database)
      this.defaultInstance[name] = new DbCls(this);
    else
      this.defaultInstance[name] = this.show(DbCls, options);
  }
  return this.defaultInstance[name];
};


/**
 * Return a class by name
 * @param {String} name
 */
App.Session.getClassByName = function (name)
{
  var tks = name.split(".");
  var cls = App;
  for (var i = 0; i < tks.length && cls; i++)
    cls = cls[tks[i]];
  return cls;
};


/*
 * Sinchronise viewport of collaborating devices
 * @param {object} msg
 */
App.Session.prototype.onChangeViewport = function (msg)
{
  if (this.device)
    this.device.setViewport(msg.devicetype, msg.orientation, msg.origin, msg.physical);
  if (module) {
    var e = [{id: "restoreCall"}];
    this.sendCommands(e);
  }
};


/*
 * Inform server of device properties (screenSize, clientSize, deviceOrientation, ua...)
 * @param {object} msg
 */
App.Session.prototype.onSetDeviceProperties = function (msg)
{
  if (this.device)
    this.device.setProp(msg.properties);
};


/*
 * Inform View of presence of physical device
 * @param {object} device
 */
App.Session.prototype.setPhDevice = function (device)
{
  var e = [{id: "setPhDevice", cnt: {device: device}}];
  this.sendCommands(e);
};


/**
 * Sends an error message to the client, to be shown on the preview
 * @param {String} msg - error message
 * @param {Bool} warning - if true is a warning, otherwise an error
 */
App.Session.prototype.notifyClientError = function (msg, warning)
{
  var e = [{obj: (!warning ? "preview-exception" : "preview-warning"), txt: msg}];
  this.sendCommands(e);
};


/*
 * Checks if running in offline mode
 */
App.isOffline = function ()
{
  return Proxy !== undefined;
};


/*
 * Checks if running in offline mode (also for NetBeans)
 */
App.Session.prototype.runsLocally = function ()
{
  return this.offline || App.isOffline();
};


/*
 * Informs all views and all property of type Document/Collecion that a document is changed
 * @param {Document} doc changed
 */
App.Session.prototype.handleDocUpdate = function (doc)
{
  // Informs every document/collection property
  var i;
  var keys = Object.keys(this);
  for (i = 0; i < keys.length; i++) {
    if (this[keys[i]] instanceof App.Document || this[keys[i]] instanceof App.Collection)
      this[keys[i]].handleDocUpdate(doc);
  }
  //
  // Informs every open view
  for (i = 0; i < this.views.length; i++)
    this.views[i].handleDocUpdate(doc);
};


/*
 * Set test auto
 */
App.Session.prototype.setTestAuto = function ()
{
  this.testAuto = true;
};


/*
 * Calculate tag results (i.e. differences between recorded tagging data and current ones)
 * @param {Object} data
 */
App.Session.prototype.calculateTagResults = function (data)    // jshint ignore:line
{
  var elToCheck = [];
  //
  // Analyze taggingData and look for differences with current remelems status
  for (var i = 0; i < data.length; i++) {
    // Get element
    var element = this.eleMap[data[i].elId];
    //
    // Check if element is a popup
    if (data[i].elId === "popup") {
      element = {id: "popup"};
      data[i].elName = "popup";
    }
    //
    var obj;
    var propsToCheck = data[i].properties || [];
    //
    obj = {
      elId: element ? element.id : data[i].elId,
      elNotFound: data[i].elNotFound
    };
    //
    if (!element)
      obj.elNotFound = true;
    //
    for (var j = 0; j < propsToCheck.length; j++) {
      var prop = element ? element[propsToCheck[j].name] : undefined;
      if (element && element.id === "popup")
        prop = this.popupBoxContent[this.popupBoxContent.length - 1];
      //
      if (typeof prop === "object")
        prop = JSON.stringify(prop);
      //
      if (typeof propsToCheck[j].value === "object")
        propsToCheck[j].value = JSON.stringify(propsToCheck[j].value);
      //
      if ((prop !== propsToCheck[j].value) || obj.elNotFound) {
        obj.error = true;
        obj.elName = data[i].elName;
        obj.props = obj.props || [];
        var propObj = {};
        propObj.propName = propsToCheck[j].name;
        propObj.oldValue = propsToCheck[j].value || "undefined";
        propObj.newValue = prop || "undefined";
        obj.props.push(propObj);
      }
    }
    //
    elToCheck.push(obj);
  }
  //
  // Send results to client
  var e = [{id: "getTaggingDataResults", cnt: elToCheck}];
  this.sendCommands(e);
};


/*
 * Give tagging data to testAuto
 * @param {Object} data
 */
App.Session.prototype.setTaggingData = function (data)    // jshint ignore:line
{
  if (!data)
    return;
  //
  var keys = Object.keys(this.eleMap);
  for (var i = 0; i < data.length; i++) {
    // Find the element using tag elId
    var element;
    for (var j = 0; j < keys.length; j++) {
      if (this.eleMap[keys[j]].id === data[i].elId) {
        element = this.eleMap[keys[j]];
        break;
      }
    }
    //
    // If I found the element and there are properties to check
    if (element && data[i].properties) {
      // For each property name get property value from element
      for (var k = 0; k < data[i].properties.length; k++)
        data[i].properties[k].value = element[data[i].properties[k].name];
    }
  }
  //
  // Send results to client
  var e = [{id: "setTaggingData", cnt: data}];
  this.sendCommands(e);
};


/*
 * Save test auto tagging data
 * @param {Object} data test auto description
 */
App.Session.prototype.saveTaggingData = function (data)    // jshint ignore:line
{

};


/*
 * Save a test auto
 * @param {String} desc test auto description
 */
App.Session.prototype.saveTestAuto = function (desc)    // jshint ignore:line
{

};


/*
 * Initialize a test auto session
 */
App.Session.prototype.initTestAuto = function ()
{

};


/*
 * Play/pause test auto
 * @param {Boolean} play
 */
App.Session.prototype.playTestAuto = function (play)    // jshint ignore:line
{

};


/*
 * Play next step in test auto
 */
App.Session.prototype.stepForwardTestAuto = function ()
{

};


/*
 * Get test properties
 */
App.Session.prototype.getTestProperties = function (data)
{
  var excludedProp = ["id", "pid", "oid", "animations"];
  data = data || {};
  //
  var props = {};
  //
  // Handle special case of popup element
  if (data.elId === "popup")
    props.content = this.popupBoxContent[this.popupBoxContent.length - 1];
  else {
    var el = this.eleMap[data.elId];
    if (el) {
      var keys = Object.keys(el);
      for (var i = 0; i < keys.length; i++) {
        if (!App.RemElem.excludedProp[keys[i]] && typeof el[keys[i]] !== "function" && excludedProp.indexOf(keys[i]) === -1) {
          if (typeof el[keys[i]] === "object") {
            try {
              JSON.stringify(el[keys[i]]);
            } catch (ex) {
              continue;
            }
          }
          props[keys[i]] = el[keys[i]];
        }
      }
    }
  }
  //
  // Send results to test auto
  var e = [{id: "sendMessageToTestAuto", cnt: {type: "testAutoMsg", content: {id: "getTestProperties", data: {elId: data.elId, props: props}}}}];
  this.sendCommands(e);
};


/*
 * Receive IP address from client and set geolocation using given ip
 */
App.Session.prototype.onIPAddress = function (ip)
{
  if (this.analytics && this.analytics.enabled)
    this.analytics.setGeolocation(ip);
};


/*
 * Checks if is a server session
 */
App.Session.prototype.isServerSession = function ()
{
  return !!this.isSS;
};


/*
 * Response of server session's messages
 * @param {object} response
 */
App.Session.prototype.serverSessionCB = function (response)
{
  var callback = App.SessionManager.callbacks[response.cbId];
  if (callback) {
    delete App.SessionManager.callbacks[response.cbId];
    callback(response);
  }
};


/**
 * @class SessionManager
 */
App.SessionManager = function ()
{
};
App.SessionManager.callbacks = {};

/*
 * Start a new server session
 * @param {Session} app
 * @param {string} name
 * @param {object} request
 * @param {function} callback
 */
App.SessionManager.startServerSession = function (app, name, request, callback)
{
  // If the app is offline
  if (App.isOffline()) {
    app.dttobj.addWarning("", "Server sessions not supported in offline app");
    return callback();
  }
  //
  // If I'm inside the IDE
  if (App.parentType === "ide") {
    app.dttobj.addWarning("", "Server sessions not supported in app launched by IDE");
    return callback();
  }
  //
  // Check if a session with same name already exists
  App.SessionManager.runningServerSession(app, name, function (running) {
    if (running) {
      app.dttobj.addWarning("", "A server session with the same name is already running");
      return callback(false);
    }
    //
    // Master: send message to worker
    process.send({type: App.msgTypeMap.serverSession, cnt: {cmd: "start", name: name, request: request}});
    callback(true);
  });
};


/*
 * Stop a server session
 * @param {Session} app
 * @param {string} name
 * @param {function} callback
 */
App.SessionManager.stopServerSession = function (app, name, callback)
{
  // If the app is offline
  if (App.isOffline()) {
    app.dttobj.addWarning("", "Server sessions not supported in offline app");
    return callback();
  }
  //
  // If I'm inside the IDE
  if (App.parentType === "ide") {
    app.dttobj.addWarning("", "Server sessions not supported in app launched by IDE");
    return callback();
  }
  //
  // Check if session exists
  App.SessionManager.runningServerSession(app, name, function (running) {
    if (!running) {
      app.dttobj.addWarning("", "Server session isn't running");
      return callback(false);
    }
    //
    // Master: send message to worker
    process.send({type: App.msgTypeMap.serverSession, cnt: {cmd: "stop", name: name}});
    callback(true);
  });
};


/*
 * Check if exists a server session
 * @param {Session} app
 * @param {string} name
 * @param {function} callback
 */
App.SessionManager.runningServerSession = function (app, name, callback)
{
  // If the app is offline
  if (App.isOffline()) {
    app.dttobj.addWarning("", "Server sessions not supported in offline app");
    return callback();
  }
  //
  // If I'm inside the IDE
  if (App.parentType === "ide") {
    app.dttobj.addWarning("", "Server sessions not supported in app launched by IDE");
    return callback();
  }
  //
  // Generate callback reference
  var id = App.Utils.generateUID24();
  App.SessionManager.callbacks[id] = function (response) {
    callback(response.running);
  };
  //
  // Master: send message to worker
  process.send({type: App.msgTypeMap.serverSession, cnt: {cmd: "running", name: name, cbId: id, sid: app.id}});
};


/*
 * Loads a resource file to translate apps
 * @param {string} url
 * @param {string} langCode
 * @param {function} callback
 */
App.loadResourceFile = function (url, langCode, callback)
{
  url = url + "";
  //
  // Create map
  if (!App.i18m)
    App.i18m = {};
  if (!App.i18m.cache)
    App.i18m.cache = {};
  //
  // Already loaded?
  if (App.i18m.cache[url]) {
    if (callback)
      callback();
    return;
  }
  //
  App.i18m.cache[url] = true;
  //
  if (!App.i18m[langCode])
    App.i18m[langCode] = {};
  //
  var typ = "";
  if (url.indexOf("resources/") > -1)
    typ = "resource";
  //
  var f = App.fs.file(url, typ);
  f.encoding = "utf8";
  f.readAll(function (data, err) {
    if (err) {
      if (callback)
        callback();
      return;
    }
    //
    App.Xml.parse(null, data, {}, function (d, err) {
      if (err) {
        if (callback)
          callback();
        return;
      }
      //
      if (d && d.root && d.root.stringlist) {
        var m = App.i18m[langCode];
        for (var i = 0; i < d.root.stringlist.length; i++) {
          var n = d.root.stringlist[i];
          m[n.id[0]] = n.translation[0];
        }
      }
      //
      if (callback)
        callback();
    });
  });
};
App.Session.prototype.loadResourceFile = App.loadResourceFile;


/**
 * Handle a parameter change
 * @param {String} paramName
 * @param {String} oldValue
 * @param {String} newValue
 */
App.Session.prototype.fireParameterChange = function (paramName, oldValue, newValue)    // jshint ignore:line
{
  // Fire event
  this.onParameterChange(paramName, oldValue, newValue);
  //
  // Handle "system-parameters" change
  switch (paramName) {
    case "syncServerUrl":
      if (this.sync.enabled) {
        // Change server url
        this.sync.enabled = false;
        this.sync.serverUrl = newValue;
        this.sync.enabled = true;
      }
      break;

    case "analyticsServerUrl":
      if (this.analytics && this.analytics.enabled) {
        // Change server url
        this.analytics.enabled = false;
        this.analytics.serverUrl = newValue;
        this.analytics.enabled = true;
      }
      break;
  }
};


/**
 * Raised when an app parameter changes
 * @param {String} paramName
 * @param {String} oldValue
 * @param {String} newValue
 */
App.Session.prototype.onParameterChange = function (paramName, oldValue, newValue)    // jshint ignore:line
{
};


/**
 * Return the value of an app parameter
 * @param {String} parName
 */
App.Session.prototype.getParameter = function (parName)
{
  return (App.params ? App.params[parName] : undefined);
};


/**
 * Sets the value of an app parameter
 * @param {String} parName
 * @param {String} newValue
 */
App.Session.prototype.setParameter = function (parName, newValue)
{
  // If the value hasn't changed... do nothing
  var oldValue = this.getParameter(parName);
  if (oldValue === newValue)
    return;
  //
  // Tell my owner that a parameter has changed
  if (module)
    process.send({type: App.msgTypeMap.changedAppParam, cnt:
              {sid: this.id, appid: App.id, par: parName, old: oldValue, new : newValue}});
  else {
    // Store the parameter in my "local" params copy
    App.params = App.params || {};
    App.params[parName] = newValue;
    //
    // No need to reinvent the wheel... use a "permanent" cookie... my owner (browser, launcher) will store it for me
    this.setCookie("app_params", JSON.stringify(App.params), {exdays: 100000});   // 273 years
    //
    // Send parameter change event
    this.fireParameterChange(parName, oldValue, newValue);
  }
};


/*
 * Process the change app parameter message
 * @param {object} msg
 */
App.handleChangedAppParamMsg = function (msg)
{
  // Update parameter value
  App.params = App.params || {};
  App.params[msg.par] = msg.new;
  //
  // Notify all sessions in this process that a parameter has changed
  var sessionIds = Object.keys(App.sessionMap);
  for (var i = 0; i < sessionIds.length; i++) {
    var session = App.sessionMap[sessionIds[i]];
    session.fireParameterChange(msg.par, msg.old, msg.new);
  }
};


App.startServerApp();


// NEVER EVER EVER export this!
// if (module)
//   module.exports = App;


if (module)
  var App = require("./requireClasses");

 App.Session.prototype.onStart = function(request, _resumer) {
    var app = this;
    var _this = this;
    App.Utils.sync($traceurRuntime.initGeneratorFunction(function _onStart(__resumer) {
      var aa,
          res,
          _ex;
      return $traceurRuntime.createGeneratorInstance(function($ctx) {
        while (true)
          switch ($ctx.state) {
            case 0:
              $ctx.pushTry(9, null);
              $ctx.state = 12;
              break;
            case 12:
              app.langCode = "en";
              aa = app.fs.publicUrl;
              app.purl = aa;
              app.showInfo = true;
              app.theme.infoSize = "1";
              App.ViewMain.show(app);
              App.ViewMain.open(app, App.Splash);
              $ctx.state = 6;
              break;
            case 6:
              $ctx.state = 2;
              return App.sbDB.query(app, {
                "id": "EfJ03GFt0D3IhShZXfEFcA==",
                "name": "newQuery",
                "qry": {
                  "id": "EfJ03GFt0D3IhShZXfEFcA==",
                  "name": "newQuery",
                  "type": "s",
                  "from": [{"expr": "Settings"}],
                  "select": [{
                    "expr": "Codice",
                    "from": ""
                  }, {
                    "expr": "Username",
                    "from": ""
                  }, {
                    "expr": "PasswordU",
                    "from": ""
                  }, {
                    "expr": "ShowInfo",
                    "from": ""
                  }, {
                    "expr": "Autologin",
                    "from": ""
                  }],
                  "dm": "sbDB"
                },
                "sch": {
                  "id": "EfJ03GFt0D3IhShZXfEFcA==",
                  "p": [{
                    "name": "Codice",
                    "maxlength": 6,
                    "datatype": "vc",
                    "nullable": true,
                    "unbound": false
                  }, {
                    "name": "Username",
                    "datatype": "vc",
                    "nullable": true,
                    "unbound": false
                  }, {
                    "name": "PasswordU",
                    "datatype": "vc",
                    "nullable": true,
                    "unbound": false
                  }, {
                    "name": "ShowInfo",
                    "datatype": "i",
                    "nullable": true,
                    "unbound": false,
                    "default": "-1"
                  }, {
                    "name": "Autologin",
                    "datatype": "i",
                    "nullable": true,
                    "unbound": false,
                    "default": "0"
                  }]
                },
                "this": _this
              }, __resumer);
            case 2:
              res = $ctx.sent;
              $ctx.state = 4;
              break;
            case 4:
              if (res.length > 0) {
                app.lCodice = res.row.Codice;
                app.lUsername = res.row.Username;
                app.lPassword = res.row.PasswordU;
                App.ViewMain.open(app, App.ViewCataloghi);
              }
              if (_resumer)
                _resumer();
              $ctx.state = 8;
              break;
            case 8:
              $ctx.popTry();
              $ctx.state = -2;
              break;
            case 9:
              $ctx.popTry();
              $ctx.maybeUncatchable();
              _ex = $ctx.storedException;
              $ctx.state = 15;
              break;
            case 15:
              if (_resumer)
                _resumer(null, _ex);
              else
                __resumer(null, _ex);
              $ctx.state = -2;
              break;
            default:
              return $ctx.end();
          }
      }, _onStart, this);
    }));
  };
  App.Session.prototype.onHistoryPop = function(obj, _resumer) {
    var app = this;
    var _this = this;
    App.Utils.sync($traceurRuntime.initGeneratorFunction(function _onHistoryPop(__resumer) {
      var ris,
          ret,
          _ex;
      return $traceurRuntime.createGeneratorInstance(function($ctx) {
        while (true)
          switch ($ctx.state) {
            case 0:
              $ctx.pushTry(19, null);
              $ctx.state = 22;
              break;
            case 22:
              $ctx.state = (obj && obj.page === "login") ? 1 : 14;
              break;
            case 1:
              $ctx.state = 2;
              return app.device.notification.confirm("Vuoi eseguire il logout?", "StyleBook", undefined, __resumer);
            case 2:
              ris = $ctx.sent;
              $ctx.state = 4;
              break;
            case 4:
              $ctx.state = (ris === 1) ? 5 : 11;
              break;
            case 5:
              $ctx.state = 6;
              return App.sbDB.execute(app, {
                "id": "TmoaYm1F2nSuC7a9eHkFDw==",
                "name": "newQuery",
                "qry": {
                  "id": "TmoaYm1F2nSuC7a9eHkFDw==",
                  "name": "newQuery",
                  "type": null,
                  "table": "Settings",
                  "custom": [{"expr": "delete\n from Settings"}],
                  "dm": "sbDB"
                },
                "sch": {
                  "id": "TmoaYm1F2nSuC7a9eHkFDw==",
                  "p": []
                },
                "this": _this
              }, __resumer);
            case 6:
              ret = $ctx.sent;
              $ctx.state = 8;
              break;
            case 8:
              App.ViewMain.back(app);
              $ctx.state = 10;
              break;
            case 11:
              app.historyPush({page: "login"});
              $ctx.state = 10;
              break;
            case 14:
              if (obj) {
                App.ViewMain.back(app);
              }
              $ctx.state = 10;
              break;
            case 10:
              if (_resumer)
                _resumer();
              $ctx.state = 18;
              break;
            case 18:
              $ctx.popTry();
              $ctx.state = -2;
              break;
            case 19:
              $ctx.popTry();
              $ctx.maybeUncatchable();
              _ex = $ctx.storedException;
              $ctx.state = 25;
              break;
            case 25:
              if (_resumer)
                _resumer(null, _ex);
              else
                __resumer(null, _ex);
              $ctx.state = -2;
              break;
            default:
              return $ctx.end();
          }
      }, _onHistoryPop, this);
    }));
  };
  App.Session.prototype.defineStyle = function() {
    if (this.styleDefined)
      return ;
    this.styleDefined = true;
    var sty = {
      "css": "   .HTitle  {   color : #fff;  }  .mainpage {   height:100%;   width:100%;   overflow:hidden; }   .maincnt {       position: relative;     z-index: 1;     height: 100%;     width: 100%;     top: 0;  } .header {   background-color:#cc1818;   color:#fff;     display: flex !important; } .header i {    color:#fff;    } .cellavm {     /*height: 0;*/     position: relative;     /*min-height: 100px;*/ } .mdl-button .material-icons {     vertical-align: middle;     margin: 0px; } .mdl-button i {     margin: 0 10px; } .header-button {      height: 34px;     width: 34px;     padding: 5px;     margin: -5px 0px 0 5px;     shape-rendering: geometricPrecision;     fill:#fff; } .mdl-menu__item:hover{   background-color:#dd2828; } .imgCat  {   position : absolute;   display : inline-block;   width : 12px;   height : 12px;  opacity: $infoSize;  top : 0%;   left : 70%;   z-index : 91;  }  .imgLis  {   position : absolute;   display : inline-block;    width : 12px;   height : 12px;  opacity: $infoSize;  top : 0%;   left : 60%;   z-index : 92;  }  .imgAnn  {   position : absolute;   display : inline-block;   width : 12px;   height : 12px;  opacity: $infoSize;  top : 0%;   left : 80%;   z-index : 93;  }  .imgArt  {   /*position : absolute; */  display : inline-block;   width : 100%;   height : 100%; }  .txtCnt  {   position : absolute;   display : inline-block;   font-size : 4pt;  line-height : 4pt;  width : 100%;   height : 100%;  }     .titolo-cat {            /* position: absolute; */     top: 0px;     left: 0px;     margin: 0;     margin-left: 0;     width: 100%;     color: #757575;     background: #fff;     padding: 2px 8px;     border-radius: 0 0 3px;     font-family:\"RobotoDraft\",\"Roboto\",sans-serif;     text-align:right;     font-weight:700;     font-size:14px; }     .titolo-pagina {    text-transform:uppercase; }  .header-button-back {      display: flex;     position: relative;     margin-left: -60px;     margin-right: 30px; }    .noscrollbar{ }   .dialog {   width:90%;        left: 5%;   box-shadow: 0 -2px 25px 0 rgba(0, 0, 0, 0.15), 0 13px 25px 0 rgba(0, 0, 0, 0.3);   background-color: #FFFFFF;   /* Internal padding */   padding-top: 21px;   padding-left: 24px;   padding-right: 24px;   padding-bottom: 12px;   /* External padding */   margin-top: 0;   margin-left: auto;   margin-right: auto;   margin-bottom: 0;    }      .dialogTitle {   /* Form */      font-weight: 500;   font-size: 20px;    line-height: 24px;       /* Internal padding */   padding-top: 0;   padding-left: 0;   padding-right: 0;   padding-bottom: 0;   /* External padding */   margin-top: 0;   margin-left: 0;   margin-right: 0;   margin-bottom: 0;     }           .dialogCont {       /* Form */      width: auto;   height: auto;   background-color: #FFFFFF;   /* Corner radius */   border-top-left-radius:2px;   border-top-right-radius:2px;   border-bottom-right-radius:0;   border-bottom-left-radius:0;       /* Internal padding */   padding-top: 21px;   padding-left: 24px;   padding-right: 24px;   padding-bottom: 12px;   /* External padding */   margin-top: 0;   margin-left: 0;   margin-right: 0;   margin-bottom: 0;          } .dialogContentTitle {    font-weight: 500;   font-size: 20px;    line-height: 24px;     text-align: left;   letter-spacing: 0.03px;   /* Internal padding */   padding-top: 0;   padding-left: 0;   padding-right: 0;   padding-bottom: 0;   /* External padding */   margin-top: 0;   margin-left: 0;   margin-right: 0;   margin-bottom: 0;     } .dialogContentBody {   display:block;   /* Typography */   color: rgba(0,0,0,.5);   font-size: 14px;    line-height: 20px;       text-align: left;    /* Internal padding */   padding-top: 14px;   padding-left: 0;   padding-right: 0;   padding-bottom: 0;      /* External padding */   margin-top: 0;   margin-left: 0;   margin-right: 0;   margin-bottom: 0;        }   .dialogContainer {    width:100%;   height:100%;   background-color: rgba(0,0,0,.5);   z-index:9999;   position:absolute;   transition:all 0.2s;    } .bgloader{      background-position: center;     background-repeat: no-repeat;     background-size: 48px; }  .header .mdl-textfield__label, .header .mdl-textfield__label:after{   background-color:#fff;   border-color:#fff; } .mdl-layout__drawer {     background: #fff;     border-bottom: 1px solid #e0e0e0; } .margin-top-10{ margin-top:10px;} .margin-top-20{ margin-top:20px;} .margin-top-30{ margin-top:30px;} .margin-top-40{ margin-top:40px;} .margin-top-50{ margin-top:50px;}   .menu-link { } .menu-link i {vertical-align:middle; margin:0 5px 0 0; }  .hlogo  {   width : 170px;   margin : 0 0 0 -20px;  }   .MenuTitle  {   border-bottom : 1px solid #e6e6e6;  }   .zoom  {   position : absolute;   bottom : 0px;   right : 0px;   z-index : 1000;   width : 30px;   height : 30px;   margin : 5px; }   .imgNote  {   position : absolute;   display : inline-block;      width: 14px;     height: 14px;   opacity: $infoSize;  top : 100%;   left : 50%;   z-index : 90;  }      @media (min-width: 480px) {       .txtCnt        {         font-size : 6pt;          line-height: 6pt;       }       .imgLis  , .imgAnn , .imgCat , .imgNote  {          width: 18px;         height: 18px;        }        .zoom {         width: 30px;         height: 30px;                }     }          @media (min-width: 768px) {       .txtCnt        {         font-size : 10pt;          line-height: 11pt;       }       .imgLis   , .imgAnn , .imgCat, .imgNote {          width: 30px;         height: 30px;        }        .zoom {         width: 40px;         height: 40px;                }     }          @media (min-width: 1280px) {     .txtCnt        {         font-size : 10pt;      line-height: 11pt;       }       .zoom {         width: 60px;         height: 60px;                }           }   .footer  {  \tpadding-top : 10px;  \tpadding-bottom : 10px;  \tjustify-content : center;  \tmargin-bottom : 0px;  \tposition : fixed;  \twidth : 100%;  \tbottom : 0px;  }   .header-nomenu  {  \tpadding-left : 10px;  }   .zoom-splash  {  \ttext-align : center;  \tcolor : rgb(255,255,255);  \tfont-size : 20pt;  \tposition : static;  \topacity : 0.6;  \tmargin-top : 10px;  }   .zoom-image  {  \tposition : static;  \ttop : 25%;  \talign-self : center;  \tmargin-left : 10px;  \tmargin-top : 10px;  \tmargin-right : 10px;  }    .splash-btn  {  \twidth : 100%;  \tmargin-bottom : 16px;  \tfont-weight : bold;  }   .bottone-login  {  \twidth : 201px;  \tbackground-color : #cc1818;  \tcolor : rgb(255,255,255);  }   .imgZ  {  \tposition : absolute;  \tdisplay : block;  \tbottom : 0px;  \tleft : 0px;  \tz-index : 20000;  \twidth : 34px !important;  \theight : 34px !important;  \tpadding-top : 10px;  } ",
      "svg": [],
      "req": {"svgdefs.svg": {
          "type": "svg",
          "name": "svgdefs.svg"
        }},
      "theme": {
        "mdPrimaryColor": "indigo",
        "mdAccentColor": "pink",
        "mdVersion": "1.1.3"
      }
    };
    sty.css = this.customizeStyle(sty.css);
    sty.theme = this.theme;
    this.sendCommands([{
      id: "loadStyle",
      cnt: sty
    }]);
    this.device.sendViewport();
  };
  App.Session.prototype.initSession = function() {
    var app = this;
    this.theme = {
      "mdPrimaryColor": "indigo",
      "mdAccentColor": "pink",
      "mdVersion": "1.1.3"
    };
    this.style = {
      "alignItems": "$ua=safari:-webkit-align-items;align-items",
      "flex": "$ua=safari:-webkit-flex;flex",
      "flexDirection": "$ua=safari:-webkit-flex-direction;flex-direction",
      "flexWrap": "$ua=safari:-webkit-flex-wrap;flex-wrap",
      "justifyContent": "$ua=safari:-webkit-justify-content;justify-content",
      "transform": "$ua=safari:-webkit-transform;transform",
      "flexGrow": "$ua=safari:-webkit-flex-grow;flex-grow",
      "flexShrink": "$ua=safari:-webkit-flex-shrink;flex-shrink",
      "alignSelf": "$ua=safari:-webkit-align-self;align-self",
      "flexFlow": "$ua=safari:-webkit-flex-flow;flex-flow",
      "inlineFlex": "$ua=safari:-webkit-inline-flex;inline-flex",
      "order": "$ua=safari:-webkit-order;order",
      "alignContent": "$ua=safari:-webkit-align-content;align-content"
    };
    this.device.barcodeScanner = {
      device: this.device,
      id: "device-barcodeScanner"
    };
    this.device.barcodeScanner.scan = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "scan",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.barcodeScanner.scanCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-barcodeScanner"] = this.device.barcodeScanner;
    this.device.vibration = {
      device: this.device,
      id: "device-vibration"
    };
    this.device.vibration.vibrate = function(time) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "vibrate",
        cnt: {time: time}
      }]);
    };
    this.eleMap["device-vibration"] = this.device.vibration;
    this.device.camera = {
      device: this.device,
      id: "device-camera"
    };
    this.device.camera.getPicture = function(options) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getPicture",
        cnt: {options: options}
      }]);
    };
    this.device.camera.saveImageToGallery = function(path, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "saveImageToGallery",
        cbId: this.device.gencb(callback),
        cnt: {path: path}
      }]);
    };
    this.device.camera.saveImageToGalleryCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-camera"] = this.device.camera;
    this.device.accelerometer = {
      device: this.device,
      id: "device-accelerometer"
    };
    this.device.accelerometer.getCurrentAcceleration = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getCurrentAcceleration",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.accelerometer.getCurrentAccelerationCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.accelerometer.watchAcceleration = function(frequency) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "watchAcceleration",
        cnt: {frequency: frequency}
      }]);
    };
    this.device.accelerometer.clearWatch = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "clearWatch",
        cnt: {}
      }]);
    };
    this.eleMap["device-accelerometer"] = this.device.accelerometer;
    this.device.compass = {
      device: this.device,
      id: "device-compass"
    };
    this.device.compass.getCurrentHeading = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getCurrentHeading",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.compass.getCurrentHeadingCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.compass.watchHeading = function(frequency) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "watchHeading",
        cnt: {frequency: frequency}
      }]);
    };
    this.device.compass.clearWatch = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "clearWatch",
        cnt: {}
      }]);
    };
    this.eleMap["device-compass"] = this.device.compass;
    this.device.geolocation = {
      device: this.device,
      id: "device-geolocation"
    };
    this.device.geolocation.getCurrentPosition = function(options, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getCurrentPosition",
        cbId: this.device.gencb(callback),
        cnt: {options: options}
      }]);
    };
    this.device.geolocation.getCurrentPositionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.geolocation.watchPosition = function(options) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "watchPosition",
        cnt: {options: options}
      }]);
    };
    this.device.geolocation.clearWatch = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "clearWatch",
        cnt: {}
      }]);
    };
    this.eleMap["device-geolocation"] = this.device.geolocation;
    this.device.notification = {
      device: this.device,
      id: "device-notification"
    };
    this.device.notification.alert = function(message, title, buttonName, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "alert",
        cbId: this.device.gencb(callback),
        cnt: {
          message: message,
          title: title,
          buttonName: buttonName
        }
      }]);
    };
    this.device.notification.alertCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.notification.confirm = function(message, title, buttonLabels, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "confirm",
        cbId: this.device.gencb(callback),
        cnt: {
          message: message,
          title: title,
          buttonLabels: buttonLabels
        }
      }]);
    };
    this.device.notification.confirmCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.notification.prompt = function(message, title, buttonLabels, defaultText, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "prompt",
        cbId: this.device.gencb(callback),
        cnt: {
          message: message,
          title: title,
          buttonLabels: buttonLabels,
          defaultText: defaultText
        }
      }]);
    };
    this.device.notification.promptCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.notification.beep = function(number) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "beep",
        cnt: {number: number}
      }]);
    };
    this.device.notification.hasPermission = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "hasPermission",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.notification.hasPermissionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.notification.schedule = function(notification) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "schedule",
        cnt: {notification: notification}
      }]);
    };
    this.device.notification.update = function(notification) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "update",
        cnt: {notification: notification}
      }]);
    };
    this.device.notification.clear = function(notification) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "clear",
        cnt: {notification: notification}
      }]);
    };
    this.device.notification.clearAll = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "clearAll",
        cnt: {}
      }]);
    };
    this.device.notification.cancel = function(notification) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "cancel",
        cnt: {notification: notification}
      }]);
    };
    this.device.notification.cancelAll = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "cancelAll",
        cnt: {}
      }]);
    };
    this.device.notification.register = function(senderID, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "register",
        cbId: this.device.gencb(callback),
        cnt: {senderID: senderID}
      }]);
    };
    this.device.notification.registerCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.notification.push = function(message, token, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "push",
        cbId: this.device.gencb(callback),
        cnt: {
          message: message,
          token: token
        }
      }]);
    };
    this.device.notification.pushCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.notification.getBadge = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getBadge",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.notification.getBadgeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.notification.setBadge = function(count) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "setBadge",
        cnt: {count: count}
      }]);
    };
    this.eleMap["device-notification"] = this.device.notification;
    this.device.statusBar = {
      device: this.device,
      id: "device-statusBar"
    };
    this.device.statusBar.show = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "show",
        cnt: {}
      }]);
    };
    this.device.statusBar.hide = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "hide",
        cnt: {}
      }]);
    };
    this.device.statusBar.setBackgroundColor = function(color) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "setBackgroundColor",
        cnt: {color: color}
      }]);
    };
    this.device.statusBar.overlaysWebView = function(status) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "overlaysWebView",
        cnt: {status: status}
      }]);
    };
    this.eleMap["device-statusBar"] = this.device.statusBar;
    this.device.contacts = {
      device: this.device,
      id: "device-contacts"
    };
    this.device.contacts.find = function(filter, fieldsToReturn, fieldsToMatch, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "find",
        cbId: this.device.gencb(callback),
        cnt: {
          filter: filter,
          fieldsToReturn: fieldsToReturn,
          fieldsToMatch: fieldsToMatch
        }
      }]);
    };
    this.device.contacts.findCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.contacts.create = function(info, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "create",
        cbId: this.device.gencb(callback),
        cnt: {info: info}
      }]);
    };
    this.device.contacts.createCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.contacts.read = function(id, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "read",
        cbId: this.device.gencb(callback),
        cnt: {id: id}
      }]);
    };
    this.device.contacts.readCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.contacts.update = function(id, info, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "update",
        cbId: this.device.gencb(callback),
        cnt: {
          id: id,
          info: info
        }
      }]);
    };
    this.device.contacts.updateCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.contacts.delete = function(id, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "delete",
        cbId: this.device.gencb(callback),
        cnt: {id: id}
      }]);
    };
    this.device.contacts.deleteCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-contacts"] = this.device.contacts;
    this.device.nfc = {
      device: this.device,
      id: "device-nfc"
    };
    this.device.nfc.listen = function(type, options) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "listen",
        cnt: {
          type: type,
          options: options
        }
      }]);
    };
    this.device.nfc.unListen = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "unListen",
        cnt: {}
      }]);
    };
    this.device.nfc.isAvailable = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isAvailable",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.nfc.isAvailableCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-nfc"] = this.device.nfc;
    this.device.keyboard = {
      device: this.device,
      id: "device-keyboard"
    };
    this.device.keyboard.hide = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "hide",
        cnt: {}
      }]);
    };
    this.device.keyboard.hideViewAccessoryBar = function(hide) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "hideViewAccessoryBar",
        cnt: {hide: hide}
      }]);
    };
    this.device.keyboard.copy = function(text) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "copy",
        cnt: {text: text}
      }]);
    };
    this.device.keyboard.paste = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "paste",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.keyboard.pasteCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-keyboard"] = this.device.keyboard;
    this.device.backgroundLocation = {
      device: this.device,
      id: "device-backgroundLocation"
    };
    this.device.backgroundLocation.start = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "start",
        cnt: {}
      }]);
    };
    this.device.backgroundLocation.stop = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "stop",
        cnt: {}
      }]);
    };
    this.device.backgroundLocation.configure = function(options) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "configure",
        cnt: {options: options}
      }]);
    };
    this.device.backgroundLocation.isLocationEnabled = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isLocationEnabled",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.backgroundLocation.isLocationEnabledCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.backgroundLocation.showAppSettings = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "showAppSettings",
        cnt: {}
      }]);
    };
    this.device.backgroundLocation.showLocationSettings = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "showLocationSettings",
        cnt: {}
      }]);
    };
    this.device.backgroundLocation.getLocations = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getLocations",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.backgroundLocation.getLocationsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.backgroundLocation.getValidLocations = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getValidLocations",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.backgroundLocation.getValidLocationsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.backgroundLocation.deleteLocation = function(locationId) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "deleteLocation",
        cnt: {locationId: locationId}
      }]);
    };
    this.device.backgroundLocation.deleteAllLocations = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "deleteAllLocations",
        cnt: {}
      }]);
    };
    this.device.backgroundLocation.getLogEntries = function(limit, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getLogEntries",
        cbId: this.device.gencb(callback),
        cnt: {limit: limit}
      }]);
    };
    this.device.backgroundLocation.getLogEntriesCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-backgroundLocation"] = this.device.backgroundLocation;
    this.device.facebook = {
      device: this.device,
      id: "device-facebook"
    };
    this.device.facebook.login = function(permissions, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "login",
        cbId: this.device.gencb(callback),
        cnt: {permissions: permissions}
      }]);
    };
    this.device.facebook.loginCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.facebook.getLoginStatus = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getLoginStatus",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.facebook.getLoginStatusCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.facebook.logout = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "logout",
        cnt: {}
      }]);
    };
    this.device.facebook.showDialog = function(options) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "showDialog",
        cnt: {options: options}
      }]);
    };
    this.device.facebook.api = function(path, permissions, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "api",
        cbId: this.device.gencb(callback),
        cnt: {
          path: path,
          permissions: permissions
        }
      }]);
    };
    this.device.facebook.apiCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.facebook.init = function(appId, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "init",
        cbId: this.device.gencb(callback),
        cnt: {appId: appId}
      }]);
    };
    this.device.facebook.initCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-facebook"] = this.device.facebook;
    this.device.beacon = {
      device: this.device,
      id: "device-beacon"
    };
    this.device.beacon.getAuthorizationStatus = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getAuthorizationStatus",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.beacon.getAuthorizationStatusCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.beacon.requestWhenInUseAuthorization = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "requestWhenInUseAuthorization",
        cnt: {}
      }]);
    };
    this.device.beacon.requestAlwaysAuthorization = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "requestAlwaysAuthorization",
        cnt: {}
      }]);
    };
    this.device.beacon.isAdvertisingAvailable = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isAdvertisingAvailable",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.beacon.isAdvertisingAvailableCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.beacon.isAdvertising = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isAdvertising",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.beacon.isAdvertisingCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.beacon.startAdvertising = function(beacon) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "startAdvertising",
        cnt: {beacon: beacon}
      }]);
    };
    this.device.beacon.stopAdvertising = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "stopAdvertising",
        cnt: {}
      }]);
    };
    this.device.beacon.isBluetoothEnabled = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isBluetoothEnabled",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.beacon.isBluetoothEnabledCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.beacon.enableBluetooth = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "enableBluetooth",
        cnt: {}
      }]);
    };
    this.device.beacon.disableBluetooth = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "disableBluetooth",
        cnt: {}
      }]);
    };
    this.device.beacon.startMonitoringForRegion = function(beacon) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "startMonitoringForRegion",
        cnt: {beacon: beacon}
      }]);
    };
    this.device.beacon.stopMonitoringForRegion = function(beacon) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "stopMonitoringForRegion",
        cnt: {beacon: beacon}
      }]);
    };
    this.device.beacon.requestStateForRegion = function(beacon) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "requestStateForRegion",
        cnt: {beacon: beacon}
      }]);
    };
    this.device.beacon.getMonitoredRegions = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getMonitoredRegions",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.beacon.getMonitoredRegionsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.beacon.startRangingBeaconsInRegion = function(beacon) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "startRangingBeaconsInRegion",
        cnt: {beacon: beacon}
      }]);
    };
    this.device.beacon.stopRangingBeaconsInRegion = function(beacon) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "stopRangingBeaconsInRegion",
        cnt: {beacon: beacon}
      }]);
    };
    this.device.beacon.isRangingAvailable = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isRangingAvailable",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.beacon.isRangingAvailableCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.beacon.getRangedRegions = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getRangedRegions",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.beacon.getRangedRegionsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-beacon"] = this.device.beacon;
    this.device.sms = {
      device: this.device,
      id: "device-sms"
    };
    this.device.sms.send = function(number, message, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "send",
        cbId: this.device.gencb(callback),
        cnt: {
          number: number,
          message: message
        }
      }]);
    };
    this.device.sms.sendCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.sms.hasPermission = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "hasPermission",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.sms.hasPermissionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-sms"] = this.device.sms;
    this.device.sqlite = {
      device: this.device,
      id: "device-sqlite"
    };
    this.device.sqlite.openDatabase = function(name, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "openDatabase",
        cbId: this.device.gencb(callback),
        cnt: {name: name}
      }]);
    };
    this.device.sqlite.openDatabaseCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.sqlite.deleteDatabase = function(name, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "deleteDatabase",
        cbId: this.device.gencb(callback),
        cnt: {name: name}
      }]);
    };
    this.device.sqlite.deleteDatabaseCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.sqlite.closeDatabase = function(id, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "closeDatabase",
        cbId: this.device.gencb(callback),
        cnt: {id: id}
      }]);
    };
    this.device.sqlite.closeDatabaseCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.sqlite.query = function(id, sql, params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "query",
        cbId: this.device.gencb(callback),
        cnt: {
          id: id,
          sql: sql,
          params: params
        }
      }]);
    };
    this.device.sqlite.queryCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-sqlite"] = this.device.sqlite;
    this.device.haptic = {
      device: this.device,
      id: "device-haptic"
    };
    this.device.haptic.feedback = function(type, style) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "feedback",
        cnt: {
          type: type,
          style: style
        }
      }]);
    };
    this.eleMap["device-haptic"] = this.device.haptic;
    this.device.media = {
      device: this.device,
      id: "device-media"
    };
    this.device.media.getCurrentPosition = function(src, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getCurrentPosition",
        cbId: this.device.gencb(callback),
        cnt: {src: src}
      }]);
    };
    this.device.media.getCurrentPositionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.getDuration = function(src, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getDuration",
        cbId: this.device.gencb(callback),
        cnt: {src: src}
      }]);
    };
    this.device.media.getDurationCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.pause = function(src) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "pause",
        cnt: {src: src}
      }]);
    };
    this.device.media.play = function(src, options) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "play",
        cnt: {
          src: src,
          options: options
        }
      }]);
    };
    this.device.media.release = function(src) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "release",
        cnt: {src: src}
      }]);
    };
    this.device.media.seekTo = function(src, milliseconds) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "seekTo",
        cnt: {
          src: src,
          milliseconds: milliseconds
        }
      }]);
    };
    this.device.media.setVolume = function(src, volume) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "setVolume",
        cnt: {
          src: src,
          volume: volume
        }
      }]);
    };
    this.device.media.stop = function(src) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "stop",
        cnt: {src: src}
      }]);
    };
    this.device.media.startRecord = function(src, sampleRate) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "startRecord",
        cnt: {
          src: src,
          sampleRate: sampleRate
        }
      }]);
    };
    this.device.media.pauseRecord = function(src) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "pauseRecord",
        cnt: {src: src}
      }]);
    };
    this.device.media.resumeRecord = function(src) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "resumeRecord",
        cnt: {src: src}
      }]);
    };
    this.device.media.stopRecord = function(src) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "stopRecord",
        cnt: {src: src}
      }]);
    };
    this.device.media.getRecordLevels = function(src, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getRecordLevels",
        cbId: this.device.gencb(callback),
        cnt: {src: src}
      }]);
    };
    this.device.media.getRecordLevelsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.remove = function(src, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "remove",
        cbId: this.device.gencb(callback),
        cnt: {src: src}
      }]);
    };
    this.device.media.removeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.size = function(src, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "size",
        cbId: this.device.gencb(callback),
        cnt: {src: src}
      }]);
    };
    this.device.media.sizeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.exists = function(src, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "exists",
        cbId: this.device.gencb(callback),
        cnt: {src: src}
      }]);
    };
    this.device.media.existsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.upload = function(src, options, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "upload",
        cbId: this.device.gencb(callback),
        cnt: {
          src: src,
          options: options
        }
      }]);
    };
    this.device.media.uploadCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.download = function(src, url, options, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "download",
        cbId: this.device.gencb(callback),
        cnt: {
          src: src,
          url: url,
          options: options
        }
      }]);
    };
    this.device.media.downloadCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.initStorage = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "initStorage",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.media.initStorageCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.media.url = function(src, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "url",
        cbId: this.device.gencb(callback),
        cnt: {src: src}
      }]);
    };
    this.device.media.urlCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-media"] = this.device.media;
    this.device.socialSharing = {
      device: this.device,
      id: "device-socialSharing"
    };
    this.device.socialSharing.shareWithOptions = function(options, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "shareWithOptions",
        cbId: this.device.gencb(callback),
        cnt: {options: options}
      }]);
    };
    this.device.socialSharing.shareWithOptionsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.socialSharing.checkAvailability = function(scheme, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "checkAvailability",
        cbId: this.device.gencb(callback),
        cnt: {scheme: scheme}
      }]);
    };
    this.device.socialSharing.checkAvailabilityCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-socialSharing"] = this.device.socialSharing;
    this.device.calendar = {
      device: this.device,
      id: "device-calendar"
    };
    this.device.calendar.listCalendars = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "listCalendars",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.calendar.listCalendarsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.createCalendar = function(calendarName, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "createCalendar",
        cbId: this.device.gencb(callback),
        cnt: {calendarName: calendarName}
      }]);
    };
    this.device.calendar.createCalendarCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.deleteCalendar = function(calendarName, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "deleteCalendar",
        cbId: this.device.gencb(callback),
        cnt: {calendarName: calendarName}
      }]);
    };
    this.device.calendar.deleteCalendarCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.openCalendar = function(date, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "openCalendar",
        cbId: this.device.gencb(callback),
        cnt: {date: date}
      }]);
    };
    this.device.calendar.openCalendarCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.hasPermission = function(read, write, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "hasPermission",
        cbId: this.device.gencb(callback),
        cnt: {
          read: read,
          write: write
        }
      }]);
    };
    this.device.calendar.hasPermissionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.requestPermission = function(read, write, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "requestPermission",
        cbId: this.device.gencb(callback),
        cnt: {
          read: read,
          write: write
        }
      }]);
    };
    this.device.calendar.requestPermissionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.listEventsInRange = function(startDate, endDate, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "listEventsInRange",
        cbId: this.device.gencb(callback),
        cnt: {
          startDate: startDate,
          endDate: endDate
        }
      }]);
    };
    this.device.calendar.listEventsInRangeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.findAllEventsInNamedCalendar = function(calendarName, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "findAllEventsInNamedCalendar",
        cbId: this.device.gencb(callback),
        cnt: {calendarName: calendarName}
      }]);
    };
    this.device.calendar.findAllEventsInNamedCalendarCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.createEvent = function(options, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "createEvent",
        cbId: this.device.gencb(callback),
        cnt: {options: options}
      }]);
    };
    this.device.calendar.createEventCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.findEvent = function(options, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "findEvent",
        cbId: this.device.gencb(callback),
        cnt: {options: options}
      }]);
    };
    this.device.calendar.findEventCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.deleteEvent = function(options, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "deleteEvent",
        cbId: this.device.gencb(callback),
        cnt: {options: options}
      }]);
    };
    this.device.calendar.deleteEventCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.calendar.modifyEvent = function(oldEvent, newEvent, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "modifyEvent",
        cbId: this.device.gencb(callback),
        cnt: {
          oldEvent: oldEvent,
          newEvent: newEvent
        }
      }]);
    };
    this.device.calendar.modifyEventCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-calendar"] = this.device.calendar;
    this.device.touchid = {
      device: this.device,
      id: "device-touchid"
    };
    this.device.touchid.isAvailable = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isAvailable",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.touchid.isAvailableCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.touchid.verify = function(key, message, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "verify",
        cbId: this.device.gencb(callback),
        cnt: {
          key: key,
          message: message
        }
      }]);
    };
    this.device.touchid.verifyCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.touchid.has = function(key, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "has",
        cbId: this.device.gencb(callback),
        cnt: {key: key}
      }]);
    };
    this.device.touchid.hasCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.touchid.delete = function(key, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "delete",
        cbId: this.device.gencb(callback),
        cnt: {key: key}
      }]);
    };
    this.device.touchid.deleteCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.touchid.save = function(key, password, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "save",
        cbId: this.device.gencb(callback),
        cnt: {
          key: key,
          password: password
        }
      }]);
    };
    this.device.touchid.saveCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-touchid"] = this.device.touchid;
    this.device.speech = {
      device: this.device,
      id: "device-speech"
    };
    this.device.speech.isRecognitionAvailable = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isRecognitionAvailable",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.speech.isRecognitionAvailableCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.speech.hasPermission = function(onlyMicrophone, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "hasPermission",
        cbId: this.device.gencb(callback),
        cnt: {onlyMicrophone: onlyMicrophone}
      }]);
    };
    this.device.speech.hasPermissionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.speech.requestPermission = function(onlyMicrophone, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "requestPermission",
        cbId: this.device.gencb(callback),
        cnt: {onlyMicrophone: onlyMicrophone}
      }]);
    };
    this.device.speech.requestPermissionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.speech.speak = function(msg, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "speak",
        cbId: this.device.gencb(callback),
        cnt: {msg: msg}
      }]);
    };
    this.device.speech.speakCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.speech.startListening = function(options, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "startListening",
        cbId: this.device.gencb(callback),
        cnt: {options: options}
      }]);
    };
    this.device.speech.startListeningCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.speech.stopListening = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "stopListening",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.speech.stopListeningCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.speech.getSupportedLanguages = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getSupportedLanguages",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.speech.getSupportedLanguagesCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-speech"] = this.device.speech;
    this.device.preferences = {
      device: this.device,
      id: "device-preferences"
    };
    this.device.preferences.show = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "show",
        cnt: {}
      }]);
    };
    this.device.preferences.fetch = function(key, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "fetch",
        cbId: this.device.gencb(callback),
        cnt: {key: key}
      }]);
    };
    this.device.preferences.fetchCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.preferences.store = function(key, value) {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "store",
        cnt: {
          key: key,
          value: value
        }
      }]);
    };
    this.eleMap["device-preferences"] = this.device.preferences;
    this.device.linkedin = {
      device: this.device,
      id: "device-linkedin"
    };
    this.device.linkedin.login = function(scopes, prompt, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "login",
        cbId: this.device.gencb(callback),
        cnt: {
          scopes: scopes,
          prompt: prompt
        }
      }]);
    };
    this.device.linkedin.loginCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.linkedin.logout = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "logout",
        cnt: {}
      }]);
    };
    this.device.linkedin.getActiveSession = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getActiveSession",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.linkedin.getActiveSessionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.linkedin.get = function(path, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "get",
        cbId: this.device.gencb(callback),
        cnt: {path: path}
      }]);
    };
    this.device.linkedin.getCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.linkedin.post = function(path, body, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "post",
        cbId: this.device.gencb(callback),
        cnt: {
          path: path,
          body: body
        }
      }]);
    };
    this.device.linkedin.postCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.linkedin.openProfile = function(memberId, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "openProfile",
        cbId: this.device.gencb(callback),
        cnt: {memberId: memberId}
      }]);
    };
    this.device.linkedin.openProfileCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.linkedin.init = function(appId, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "init",
        cbId: this.device.gencb(callback),
        cnt: {appId: appId}
      }]);
    };
    this.device.linkedin.initCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-linkedin"] = this.device.linkedin;
    this.device.ble = {
      device: this.device,
      id: "device-ble"
    };
    this.device.ble.initialize = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "initialize",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.initializeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.enable = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "enable",
        cnt: {}
      }]);
    };
    this.device.ble.disable = function() {
      this.device.app.sendCommands([{
        obj: this.id,
        id: "disable",
        cnt: {}
      }]);
    };
    this.device.ble.getAdapterInfo = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "getAdapterInfo",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.getAdapterInfoCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.isInitialized = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isInitialized",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.isInitializedCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.isEnabled = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isEnabled",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.isEnabledCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.hasPermission = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "hasPermission",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.hasPermissionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.requestPermission = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "requestPermission",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.requestPermissionCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.isLocationEnabled = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isLocationEnabled",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.isLocationEnabledCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.requestLocation = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "requestLocation",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.requestLocationCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.startScan = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "startScan",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.startScanCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.stopScan = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "stopScan",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.stopScanCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.retrieveConnected = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "retrieveConnected",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.retrieveConnectedCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.isScanning = function(callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isScanning",
        cbId: this.device.gencb(callback),
        cnt: {}
      }]);
    };
    this.device.ble.isScanningCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.connect = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "connect",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.connectCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.reconnect = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "reconnect",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.reconnectCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.disconnect = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "disconnect",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.disconnectCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.close = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "close",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.closeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.rssi = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "rssi",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.rssiCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.mtu = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "mtu",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.mtuCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.requestConnectionPriority = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "requestConnectionPriority",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.requestConnectionPriorityCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.wasConnected = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "wasConnected",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.wasConnectedCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.isConnected = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isConnected",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.isConnectedCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.bond = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "bond",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.bondCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.unbond = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "unbond",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.unbondCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.isBonded = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isBonded",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.isBondedCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.discover = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "discover",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.discoverCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.services = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "services",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.servicesCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.characteristics = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "characteristics",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.characteristicsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.descriptors = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "descriptors",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.descriptorsCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.isDiscovered = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "isDiscovered",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.isDiscoveredCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.read = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "read",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.readCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.subscribe = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "subscribe",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.subscribeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.unsubscribe = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "unsubscribe",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.unsubscribeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.write = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "write",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.writeCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.readDescriptor = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "readDescriptor",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.readDescriptorCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.device.ble.writeDescriptor = function(params, callback) {
      this.cb = callback;
      this.device.app.sendCommands([{
        obj: this.id,
        id: "writeDescriptor",
        cbId: this.device.gencb(callback),
        cnt: {params: params}
      }]);
    };
    this.device.ble.writeDescriptorCB = function(content) {
      var cb = this.cb;
      delete this.cb;
      if (content.cbId)
        this.device.cb(content);
      else
        cb(content.result, content.error);
    };
    this.eleMap["device-ble"] = this.device.ble;
    this.device.initPlugins();
    App.ide = false;
    App.Device = App.Device || {};
    App.Device.orientation = {
      horizontal: "horizontal",
      vertical: "vertical"
    };
    App.Device.orientation.getValues = function() {
      return [{
        "n": "horizontal",
        "v": "horizontal"
      }, {
        "n": "vertical",
        "v": "vertical"
      }];
    };
    App.Device.orientation.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Device = App.Device || {};
    App.Device.deviceType = {
      none: "none",
      desktop: "desktop",
      smartphone: "smartphone",
      tablet: "tablet"
    };
    App.Device.deviceType.getValues = function() {
      return [{
        "n": "none",
        "v": "none"
      }, {
        "n": "desktop",
        "v": "desktop"
      }, {
        "n": "smartphone",
        "v": "smartphone"
      }, {
        "n": "tablet",
        "v": "tablet"
      }];
    };
    App.Device.deviceType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.View = App.View || {};
    App.View.showModes = {
      normal: "normal",
      popup: "popup"
    };
    App.View.showModes.getValues = function() {
      return [{
        "n": "normal",
        "v": "normal"
      }, {
        "n": "popup",
        "v": "popup"
      }];
    };
    App.View.showModes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.View = App.View || {};
    App.View.popupPosition = {
      center: "center",
      top: "top",
      right: "right",
      bottom: "bottom",
      left: "left",
      topleft: "topleft",
      topright: "topright",
      bottomleft: "bottomleft",
      bottomright: "bottomright"
    };
    App.View.popupPosition.getValues = function() {
      return [{
        "n": "center",
        "v": "center"
      }, {
        "n": "top",
        "v": "top"
      }, {
        "n": "right",
        "v": "right"
      }, {
        "n": "bottom",
        "v": "bottom"
      }, {
        "n": "left",
        "v": "left"
      }, {
        "n": "topleft",
        "v": "topleft"
      }, {
        "n": "topright",
        "v": "topright"
      }, {
        "n": "bottomleft",
        "v": "bottomleft"
      }, {
        "n": "bottomright",
        "v": "bottomright"
      }];
    };
    App.View.popupPosition.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.View = App.View || {};
    App.View.popupRelativePosition = {
      top: "top",
      right: "right",
      bottom: "bottom",
      left: "left"
    };
    App.View.popupRelativePosition.getValues = function() {
      return [{
        "n": "top",
        "v": "top"
      }, {
        "n": "right",
        "v": "right"
      }, {
        "n": "bottom",
        "v": "bottom"
      }, {
        "n": "left",
        "v": "left"
      }];
    };
    App.View.popupRelativePosition.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Theme = App.Theme || {};
    App.Theme.Devices = {
      iPhone5V: "{\"devicetype\":\"smartphone\",\"devicesize\":{\"w\":320,\"h\":576},\"position\":{\"l\":0,\"t\":0},\"orientation\":\"vertical\",\"color\":\"#FFFFFF\"}",
      iPhone5H: "{\"devicetype\":\"smartphone\",\"devicesize\":{\"w\":320,\"h\":576},\"position\":{\"l\":0,\"t\":0},\"orientation\":\"horizontal\",\"color\":\"#FFFFFF\"}",
      iPadV: "{\"devicetype\":\"tablet\",\"devicesize\":{\"w\":768,\"h\": 1024},\"position\":{\"l\":0,\"t\":0},\"orientation\":\"vertical\",\"color\":\"#FFFFFF\"}",
      iPadH: "{\"devicetype\":\"tablet\",\"devicesize\":{\"w\":768,\"h\": 1024},\"position\":{\"l\":0,\"t\":0},\"orientation\":\"horizontal\",\"color\":\"#FFFFFF\"}",
      None: "{\"devicetype\":\"none\"}"
    };
    App.Theme.Devices.getValues = function() {
      return [{
        "src": "icon-device-smartphone-portrait",
        "n": "iPhone5V",
        "v": "{\"devicetype\":\"smartphone\",\"devicesize\":{\"w\":320,\"h\":576},\"position\":{\"l\":0,\"t\":0},\"orientation\":\"vertical\",\"color\":\"#FFFFFF\"}"
      }, {
        "src": "icon-device-smartphone-landscape",
        "n": "Template for viewport of the smartphone type.",
        "v": "{\"devicetype\":\"smartphone\",\"devicesize\":{\"w\":320,\"h\":576},\"position\":{\"l\":0,\"t\":0},\"orientation\":\"horizontal\",\"color\":\"#FFFFFF\"}"
      }, {
        "src": "icon-device-tablet-portrait",
        "n": "iPadV",
        "v": "{\"devicetype\":\"tablet\",\"devicesize\":{\"w\":768,\"h\": 1024},\"position\":{\"l\":0,\"t\":0},\"orientation\":\"vertical\",\"color\":\"#FFFFFF\"}"
      }, {
        "src": "icon-device-tablet-landscape",
        "n": "Template for viewport of the tablet type.",
        "v": "{\"devicetype\":\"tablet\",\"devicesize\":{\"w\":768,\"h\": 1024},\"position\":{\"l\":0,\"t\":0},\"orientation\":\"horizontal\",\"color\":\"#FFFFFF\"}"
      }, {
        "src": "icon-device-desktop",
        "n": "None",
        "v": "{\"devicetype\":\"none\"}"
      }];
    };
    App.Theme.Devices.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Container = App.Container || {};
    App.Container.Layouts = {
      horizontal: "horizontal",
      vertical: "vertical",
      absolute: "absolute",
      table: "table",
      none: "none"
    };
    App.Container.Layouts.getValues = function() {
      return [{
        "n": "horizontal",
        "v": "horizontal"
      }, {
        "n": "vertical",
        "v": "vertical"
      }, {
        "n": "absolute",
        "v": "absolute"
      }, {
        "n": "table",
        "v": "table"
      }, {
        "n": "none",
        "v": "none"
      }];
    };
    App.Container.Layouts.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.SplitView = App.SplitView || {};
    App.SplitView.VisibilityStatus = {
      Auto: "auto",
      Never: "never",
      Always: "always",
      DeviceDependant: "device"
    };
    App.SplitView.VisibilityStatus.getValues = function() {
      return [{
        "n": "Auto",
        "v": "auto"
      }, {
        "n": "Never",
        "v": "never"
      }, {
        "n": "Always",
        "v": "always"
      }, {
        "n": "DeviceDependant",
        "v": "device"
      }];
    };
    App.SplitView.VisibilityStatus.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Gmap = App.Gmap || {};
    App.Gmap.mapType = {
      roadmap: "ROADMAP",
      satellite: "SATELLITE",
      terrain: "TERRAIN",
      hybrid: "HYBRID"
    };
    App.Gmap.mapType.getValues = function() {
      return [{
        "n": "The map displays the streets, without satellite images.",
        "v": "ROADMAP"
      }, {
        "n": "The map displays the satellite images, but not the streets.",
        "v": "SATELLITE"
      }, {
        "n": "The map shows the terrain levels and the streets.",
        "v": "TERRAIN"
      }, {
        "n": "The map shows both the satellite images and the streets.",
        "v": "HYBRID"
      }];
    };
    App.Gmap.mapType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.ChartJs = App.ChartJs || {};
    App.ChartJs.chartType = {
      line: "line",
      bar: "bar",
      horizontalBar: "horizontalBar",
      radar: "radar",
      polarArea: "polarArea",
      pie: "pie",
      doughnut: "doughnut"
    };
    App.ChartJs.chartType.getValues = function() {
      return [{
        "n": "line",
        "v": "line"
      }, {
        "n": "bar",
        "v": "bar"
      }, {
        "n": "horizontalBar",
        "v": "horizontalBar"
      }, {
        "n": "radar",
        "v": "radar"
      }, {
        "n": "polarArea",
        "v": "polarArea"
      }, {
        "n": "pie",
        "v": "pie"
      }, {
        "n": "doughnut",
        "v": "doughnut"
      }];
    };
    App.ChartJs.chartType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.GoogleCharts = App.GoogleCharts || {};
    App.GoogleCharts.chartType = {
      line: "LineChart",
      bar: "BarChart",
      pie: "PieChart",
      geo: "GeoChart",
      scatter: "ScatterChart",
      column: "ColumnChart",
      histogram: "Histogram",
      combo: "ComboChart",
      gauge: "Gauge",
      area: "AreaChart",
      steppedArea: "SteppedAreaChart",
      bubble: "BubbleChart",
      donut: "DonutChart",
      org: "OrgChart",
      treemap: "TreeMap",
      table: "Table",
      timeline: "Timeline",
      candlestick: "CandlestickChart",
      sankey: "Sankey"
    };
    App.GoogleCharts.chartType.getValues = function() {
      return [{
        "n": "line",
        "v": "LineChart"
      }, {
        "n": "bar",
        "v": "BarChart"
      }, {
        "n": "pie",
        "v": "PieChart"
      }, {
        "n": "geo",
        "v": "GeoChart"
      }, {
        "n": "scatter",
        "v": "ScatterChart"
      }, {
        "n": "column",
        "v": "ColumnChart"
      }, {
        "n": "histogram",
        "v": "Histogram"
      }, {
        "n": "combo",
        "v": "ComboChart"
      }, {
        "n": "gauge",
        "v": "Gauge"
      }, {
        "n": "area",
        "v": "AreaChart"
      }, {
        "n": "steppedArea",
        "v": "SteppedAreaChart"
      }, {
        "n": "bubble",
        "v": "BubbleChart"
      }, {
        "n": "donut",
        "v": "DonutChart"
      }, {
        "n": "org",
        "v": "OrgChart"
      }, {
        "n": "treemap",
        "v": "TreeMap"
      }, {
        "n": "table",
        "v": "Table"
      }, {
        "n": "timeline",
        "v": "Timeline"
      }, {
        "n": "candlestick",
        "v": "CandlestickChart"
      }, {
        "n": "sankey",
        "v": "Sankey"
      }];
    };
    App.GoogleCharts.chartType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Calendar = App.Calendar || {};
    App.Calendar.views = {
      month: "month",
      basicWeek: "basicWeek",
      basicDay: "basicDay ",
      agendaWeek: "agendaWeek",
      agendaDay: "agendaDay"
    };
    App.Calendar.views.getValues = function() {
      return [{
        "n": "month",
        "v": "month"
      }, {
        "n": "basicWeek",
        "v": "basicWeek"
      }, {
        "n": "basicDay",
        "v": "basicDay "
      }, {
        "n": "agendaWeek",
        "v": "agendaWeek"
      }, {
        "n": "agendaDay",
        "v": "agendaDay"
      }];
    };
    App.Calendar.views.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.ColorPicker = App.ColorPicker || {};
    App.ColorPicker.pickerTypes = {
      basic: "basic",
      full: "full"
    };
    App.ColorPicker.pickerTypes.getValues = function() {
      return [{
        "n": "A simpler color picker with the basic features.",
        "v": "basic"
      }, {
        "n": "A complete color picker, with some advanced features.",
        "v": "full"
      }];
    };
    App.ColorPicker.pickerTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.ColorPicker = App.ColorPicker || {};
    App.ColorPicker.formatTypes = {
      HSL: "HSL",
      HSV: "HSV"
    };
    App.ColorPicker.formatTypes.getValues = function() {
      return [{
        "n": "RGB color display mode of the Hue Saturation Lightness type.",
        "v": "HSL"
      }, {
        "n": "RGB color display mode of the Hue Saturation Value type.",
        "v": "HSV"
      }];
    };
    App.ColorPicker.formatTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.ProgressBar = App.ProgressBar || {};
    App.ProgressBar.progressType = {
      Line: "Line",
      Circle: "Circle",
      SemiCircle: "SemiCircle",
      Custom: "Path"
    };
    App.ProgressBar.progressType.getValues = function() {
      return [{
        "n": "Line",
        "v": "Line"
      }, {
        "n": "Circle",
        "v": "Circle"
      }, {
        "n": "SemiCircle",
        "v": "SemiCircle"
      }, {
        "n": "Custom",
        "v": "Path"
      }];
    };
    App.ProgressBar.progressType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.ProgressBar = App.ProgressBar || {};
    App.ProgressBar.progressEasing = {
      Linear: "linear",
      EaseOut: "easeOut",
      EaseIn: "easeIn",
      EaseInOut: "easeInOut",
      Bounce: "bounce"
    };
    App.ProgressBar.progressEasing.getValues = function() {
      return [{
        "n": "Linear",
        "v": "linear"
      }, {
        "n": "EaseOut",
        "v": "easeOut"
      }, {
        "n": "EaseIn",
        "v": "easeIn"
      }, {
        "n": "EaseInOut",
        "v": "easeInOut"
      }, {
        "n": "Bounce",
        "v": "bounce"
      }];
    };
    App.ProgressBar.progressEasing.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.QRCode = App.QRCode || {};
    App.QRCode.errorLevels = {
      Low: 1,
      Medium: 0,
      High: 3,
      Highest: 2
    };
    App.QRCode.errorLevels.getValues = function() {
      return [{
        "n": "Low",
        "v": 1
      }, {
        "n": "Medium",
        "v": 0
      }, {
        "n": "High",
        "v": 3
      }, {
        "n": "Highest",
        "v": 2
      }];
    };
    App.QRCode.errorLevels.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.ScrollContainer = App.ScrollContainer || {};
    App.ScrollContainer.scrollDirections = {
      vertical: 1,
      horizontal: 2,
      both: 3
    };
    App.ScrollContainer.scrollDirections.getValues = function() {
      return [{
        "n": "vertical",
        "v": 1
      }, {
        "n": "horizontal",
        "v": 2
      }, {
        "n": "both",
        "v": 3
      }];
    };
    App.ScrollContainer.scrollDirections.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.DropZone = App.DropZone || {};
    App.DropZone.Status = {
      Added: "added",
      Accepted: "queued",
      Uploading: "Uploading",
      Canceled: "canceled",
      Error: "error",
      Success: "success"
    };
    App.DropZone.Status.getValues = function() {
      return [{
        "n": "Added",
        "v": "added"
      }, {
        "n": "Accepted",
        "v": "queued"
      }, {
        "n": "Uploading",
        "v": "Uploading"
      }, {
        "n": "Canceled",
        "v": "canceled"
      }, {
        "n": "Error",
        "v": "error"
      }, {
        "n": "Success",
        "v": "success"
      }];
    };
    App.DropZone.Status.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Interactable = App.Interactable || {};
    App.Interactable.acceptedElements = {
      parent: "parent",
      siblings: "siblings",
      children: "children",
      view: "view"
    };
    App.Interactable.acceptedElements.getValues = function() {
      return [{
        "n": "parent",
        "v": "parent"
      }, {
        "n": "siblings",
        "v": "siblings"
      }, {
        "n": "children",
        "v": "children"
      }, {
        "n": "view",
        "v": "view"
      }];
    };
    App.Interactable.acceptedElements.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Interactable = App.Interactable || {};
    App.Interactable.axis = {
      x: "x",
      y: "y"
    };
    App.Interactable.axis.getValues = function() {
      return [{
        "n": "x",
        "v": "x"
      }, {
        "n": "y",
        "v": "y"
      }];
    };
    App.Interactable.axis.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Interactable = App.Interactable || {};
    App.Interactable.invertValues = {
      none: "none",
      negate: "negate",
      reposition: "reposition"
    };
    App.Interactable.invertValues.getValues = function() {
      return [{
        "n": "none",
        "v": "none"
      }, {
        "n": "negate",
        "v": "negate"
      }, {
        "n": "reposition",
        "v": "reposition"
      }];
    };
    App.Interactable.invertValues.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Interactable = App.Interactable || {};
    App.Interactable.dragTypes = {
      drag: "drag",
      move: "move",
      copy: "copy"
    };
    App.Interactable.dragTypes.getValues = function() {
      return [{
        "n": "drag",
        "v": "drag"
      }, {
        "n": "move",
        "v": "move"
      }, {
        "n": "copy",
        "v": "copy"
      }];
    };
    App.Interactable.dragTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Interactable = App.Interactable || {};
    App.Interactable.Edges = {
      top: 1,
      left: 2,
      bottom: 3,
      right: 4
    };
    App.Interactable.Edges.getValues = function() {
      return [{
        "n": "top",
        "v": 1
      }, {
        "n": "left",
        "v": 2
      }, {
        "n": "bottom",
        "v": 3
      }, {
        "n": "right",
        "v": 4
      }];
    };
    App.Interactable.Edges.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.DataMap = App.DataMap || {};
    App.DataMap.likeTypes = {
      noLike: 0,
      equals: 1,
      startsWith: 2,
      contains: 3
    };
    App.DataMap.likeTypes.getValues = function() {
      return [{
        "n": "Do not use the like operator, but the = operator",
        "v": 0
      }, {
        "n": "Use the like operator in the format \"field like value\"",
        "v": 1
      }, {
        "n": "Use the like operator in the format \"field like value%\"",
        "v": 2
      }, {
        "n": "Use the like operator in the format \"field like %value%\"",
        "v": 3
      }];
    };
    App.DataMap.likeTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.DataMap = App.DataMap || {};
    App.DataMap.dataPagingModes = {
      none: 0,
      offset: 1,
      keyset: 2
    };
    App.DataMap.dataPagingModes.getValues = function() {
      return [{
        "n": "Pagination disabled.",
        "v": 0
      }, {
        "n": "Pagination with offset operator.",
        "v": 1
      }, {
        "n": "Pagination with filter by key.\n@--\nRequires that the datamap be ordered.",
        "v": 2
      }];
    };
    App.DataMap.dataPagingModes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.DataMap = App.DataMap || {};
    App.DataMap.errorSources = {
      load: "load",
      save: "save",
      validation: undefined
    };
    App.DataMap.errorSources.getValues = function() {
      return [{
        "n": "Error occurred during loading",
        "v": "load"
      }, {
        "n": "Error occurred during saving",
        "v": "save"
      }, {"n": "Error set on a row using the @-setError-@ method"}];
    };
    App.DataMap.errorSources.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Document = App.Document || {};
    App.Document.loadTypes = {
      load: 0,
      loadChildCollection: 1,
      loadCollection: 2
    };
    App.Document.loadTypes.getValues = function() {
      return [{
        "n": "Loading of a document instance.",
        "v": 0
      }, {
        "n": "Loading of a child collection of the document.",
        "v": 1
      }, {
        "n": "Loading of a collection of documents.",
        "v": 2
      }];
    };
    App.Document.loadTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Document = App.Document || {};
    App.Document.validationReasons = {
      change: "change",
      save: "save",
      sync: "sync"
    };
    App.Document.validationReasons.getValues = function() {
      return [{
        "n": "Validation following a change to a document property from the user interface.",
        "v": "change"
      }, {
        "n": "Validation before the document save.",
        "v": "save"
      }, {
        "n": "Validation during synchronization.",
        "v": "sync"
      }];
    };
    App.Document.validationReasons.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Document = App.Document || {};
    App.Document.savePhases = {
      beforeSave: "beforeSave",
      inserting: "inserting",
      updating: "updating",
      deleting: "deleting",
      afterSave: "afterSave"
    };
    App.Document.savePhases.getValues = function() {
      return [{
        "n": "Preliminary phase.",
        "v": "beforeSave"
      }, {
        "n": "Insertion phase.",
        "v": "inserting"
      }, {
        "n": "Update phase.",
        "v": "updating"
      }, {
        "n": "Deletion phase.",
        "v": "deleting"
      }, {
        "n": "Concluding phase.",
        "v": "afterSave"
      }];
    };
    App.Document.savePhases.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Fs = App.Fs || {};
    App.Fs.internalType = {
      permanent: "permanent",
      temp: "temp",
      resource: "resource",
      private: "private"
    };
    App.Fs.internalType.getValues = function() {
      return [{
        "n": "Permanent file/directory (default).",
        "v": "permanent"
      }, {
        "n": "Temporary file/directory.",
        "v": "temp"
      }, {
        "n": "File/directory of the resource type; the objects are read-only. The resources loaded in the project are present in this part of the file system.",
        "v": "resource"
      }, {
        "n": "Private files/directories. They cannot be accessed externally and therefore cannot be downloaded in the browser.",
        "v": "private"
      }];
    };
    App.Fs.internalType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.File = App.File || {};
    App.File.fileEncoding = {
      binary: "null",
      utf8: "utf-8",
      ascii: "ascii",
      ucs2: "ucs2",
      base64: "base64"
    };
    App.File.fileEncoding.getValues = function() {
      return [{
        "n": "binary",
        "v": "null"
      }, {
        "n": "utf8",
        "v": "utf-8"
      }, {
        "n": "ascii",
        "v": "ascii"
      }, {
        "n": "ucs2",
        "v": "ucs2"
      }, {
        "n": "base64",
        "v": "base64"
      }];
    };
    App.File.fileEncoding.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Sync = App.Sync || {};
    App.Sync.autoConnectTypes = {
      alwaysConnected: 1,
      automatic: 2,
      manual: 3
    };
    App.Sync.autoConnectTypes.getValues = function() {
      return [{
        "n": "Always connected.\n@--\nAt the start of the session, the connection is immediately opened, and upon disconnection reconnection is attempted at increasing intervals.",
        "v": 1
      }, {
        "n": "Connect as needed.\n@--\nThe connection is opened automatically when a message is sent. After the sending of a message, the connection remains open for the time specified by the @-Sync.autoDisconnectTimeout-@ property.",
        "v": 2
      }, {
        "n": "Manual connection.\n@--\nThe connection must be opened by calling the @-Sync.connect-@ method. After the sending of a message, the connection remains open for the time specified by the @-Sync.autoDisconnectTimeout-@ property.",
        "v": 3
      }];
    };
    App.Sync.autoConnectTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Sync = App.Sync || {};
    App.Sync.statusTypes = {
      disconnected: "disconnected",
      connecting: "connecting",
      connected: "connected",
      disconnecting: "disconnecting"
    };
    App.Sync.statusTypes.getValues = function() {
      return [{
        "n": "The connection is closed.",
        "v": "disconnected"
      }, {
        "n": "The connection is still closed, but an attempt to connect is in progress.",
        "v": "connecting"
      }, {
        "n": "The connection is open.",
        "v": "connected"
      }, {
        "n": "The connection is still open, but disconnection is in progress.",
        "v": "disconnecting"
      }];
    };
    App.Sync.statusTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.SyncPlugin = App.SyncPlugin || {};
    App.SyncPlugin.logLevels = {
      disabled: "disabled",
      complete: "complete",
      verbose: "verbose"
    };
    App.SyncPlugin.logLevels.getValues = function() {
      return [{
        "n": "disabled",
        "v": "disabled"
      }, {
        "n": "complete",
        "v": "complete"
      }, {
        "n": "verbose",
        "v": "verbose"
      }];
    };
    App.SyncPlugin.logLevels.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.WebApiEndpoint = App.WebApiEndpoint || {};
    App.WebApiEndpoint.actionTypes = {
      saveDocument: "saveDocument",
      loadDocument: "loadDocument",
      loadCollection: "loadCollection",
      loadChildCollection: "loadChildCollection",
      callMethod: "callMethod"
    };
    App.WebApiEndpoint.actionTypes.getValues = function() {
      return [{
        "n": "saveDocument",
        "v": "saveDocument"
      }, {
        "n": "loadDocument",
        "v": "loadDocument"
      }, {
        "n": "loadCollection",
        "v": "loadCollection"
      }, {
        "n": "loadChildCollection",
        "v": "loadChildCollection"
      }, {
        "n": "callMethod",
        "v": "callMethod"
      }];
    };
    App.WebApiEndpoint.actionTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Analytics = App.Analytics || {};
    App.Analytics.categories = {
      sessions: "sessions",
      users: "users",
      events: "events",
      pageviews: "pageviews",
      exceptions: "exceptions"
    };
    App.Analytics.categories.getValues = function() {
      return [{
        "n": "Category that represents the application usage data about the sessions.",
        "v": "sessions"
      }, {
        "n": "Category that represents the application usage data about the users.",
        "v": "users"
      }, {
        "n": "Category that represents the application usage data about the user interactions.",
        "v": "events"
      }, {
        "n": "Category that represents the application usage data about the pages visited.",
        "v": "pageviews"
      }, {
        "n": "Category that represents the application usage data about the exceptions that occurred.",
        "v": "exceptions"
      }];
    };
    App.Analytics.categories.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Analytics = App.Analytics || {};
    App.Analytics.lists = {
      apps: "apps",
      browsers: "browsers",
      os: "os",
      countries: "countries",
      languages: "languages",
      customEvents: "customEvents"
    };
    App.Analytics.lists.getValues = function() {
      return [{
        "n": "Category used to obtain the list of applications for which the Analytics server has data.",
        "v": "apps"
      }, {
        "n": "Category used to obtain the list of browsers known to the Analytics server.",
        "v": "browsers"
      }, {
        "n": "Category used to obtain the list of operating systems known to the Analytics server.",
        "v": "os"
      }, {
        "n": "Category used to obtain the list of countries known to the Analytics server.",
        "v": "countries"
      }, {
        "n": "Category used to obtain the list of languages known to the Analytics server.",
        "v": "languages"
      }, {
        "n": "Category used to obtain the list of custom events known to the Analytics server.",
        "v": "customEvents"
      }];
    };
    App.Analytics.lists.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Analytics = App.Analytics || {};
    App.Analytics.sessions = {
      groupBy: {
        duration: "duration",
        weeklyUsage: "weeklyUsage"
      },
      statsType: {
        newUsers: "newUsers",
        returningUsers: "returningUsers",
        allUsers: "allUsers"
      }
    };
    App.Analytics.sessions.getValues = function() {
      return [{
        "n": "groupBy",
        "v": "{duration:\"duration\",weeklyUsage:\"weeklyUsage\"}"
      }, {
        "n": "statsType",
        "v": "{newUsers:\"newUsers\",returningUsers:\"returningUsers\",allUsers:\"allUsers\"}"
      }];
    };
    App.Analytics.sessions.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Analytics = App.Analytics || {};
    App.Analytics.users = {
      groupBy: {
        place: "place",
        os: "os",
        browser: "browser",
        lang: "lang",
        city: "city"
      },
      statsType: {
        sessions: "sessions",
        sessionsPerUsers: "sessionsPerUsers",
        sessionsDuration: "sessionsDuration",
        interactionsPerPage: "interactionsPerPage",
        interactionsPerMinute: "interactionsPerMinute",
        newUsers: "newUsers",
        returningUsers: "returningUsers",
        allUsers: "allUsers",
        pageviews: "pageviews",
        exceptions: "exceptions",
        customEvent: "customEvent"
      }
    };
    App.Analytics.users.getValues = function() {
      return [{
        "n": "groupBy",
        "v": "{place:\"place\",os:\"os\",browser:\"browser\",lang:\"lang\",city:\"city\"}"
      }, {
        "n": "statsType",
        "v": "{sessions:\"sessions\",sessionsPerUsers:\"sessionsPerUsers\",sessionsDuration:\"sessionsDuration\",interactionsPerPage:\"interactionsPerPage\",interactionsPerMinute:\"interactionsPerMinute\",newUsers:\"newUsers\",returningUsers:\"returningUsers\",allUsers:\"allUsers\",pageviews:\"pageviews\",exceptions:\"exceptions\",customEvent:\"customEvent\"}"
      }];
    };
    App.Analytics.users.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Analytics = App.Analytics || {};
    App.Analytics.events = {
      groupBy: {
        view: "view",
        properties: "properties"
      },
      statsType: {
        interactionsPerPage: "interactionsPerPage",
        interactionsPerSession: "interactionsPerSession",
        interactionsPerPageNewUsers: "interactionsPerPageNewUsers",
        interactionsPerPageReturningUsers: "interactionsPerPageReturningUsers",
        interactionsPerSessionNewUsers: "interactionsPerSessionNewUsers",
        interactionsPerSessionReturningUsers: "interactionsPerSessionReturningUsers"
      }
    };
    App.Analytics.events.getValues = function() {
      return [{
        "n": "groupBy",
        "v": "{view:\"view\",properties:\"properties\"}"
      }, {
        "n": "statsType",
        "v": "{interactionsPerPage:\"interactionsPerPage\",interactionsPerSession:\"interactionsPerSession\",interactionsPerPageNewUsers:\"interactionsPerPageNewUsers\",interactionsPerPageReturningUsers:\"interactionsPerPageReturningUsers\",interactionsPerSessionNewUsers:\"interactionsPerSessionNewUsers\",interactionsPerSessionReturningUsers:\"interactionsPerSessionReturningUsers\"}"
      }];
    };
    App.Analytics.events.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Analytics = App.Analytics || {};
    App.Analytics.pageviews = {
      groupBy: {
        view: "view",
        navigationFlow: "navigationFlow"
      },
      statsType: {
        duration: "duration",
        allUsers: "allUsers",
        newUsers: "newUsers",
        returningUsers: "returningUsers",
        interactions: "interactions"
      }
    };
    App.Analytics.pageviews.getValues = function() {
      return [{
        "n": "groupBy",
        "v": "{view:\"view\",navigationFlow:\"navigationFlow\"}"
      }, {
        "n": "statsType",
        "v": "{duration:\"duration\",allUsers:\"allUsers\",newUsers:\"newUsers\",returningUsers:\"returningUsers\",interactions:\"interactions\"}"
      }];
    };
    App.Analytics.pageviews.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Analytics = App.Analytics || {};
    App.Analytics.exceptions = {
      groupBy: {type: "type"},
      statsType: {
        totalExceptions: "totalExceptions",
        sessions: "sessions",
        usersAffected: "usersAffected",
        newUsersAffected: "newUsersAffected",
        returningUsersAffected: "returningUsersAffected"
      }
    };
    App.Analytics.exceptions.getValues = function() {
      return [{
        "n": "groupBy",
        "v": "{type:\"type\"}"
      }, {
        "n": "statsType",
        "v": "{totalExceptions:\"totalExceptions\",sessions:\"sessions\",usersAffected:\"usersAffected\",newUsersAffected:\"newUsersAffected\",returningUsersAffected:\"returningUsersAffected\"}"
      }];
    };
    App.Analytics.exceptions.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Analytics = App.Analytics || {};
    App.Analytics.timeUnit = {
      hour: "hour",
      day: "day",
      week: "week",
      month: "month",
      dayHour: "dayHour",
      weekDay: "weekDay"
    };
    App.Analytics.timeUnit.getValues = function() {
      return [{
        "n": "hour",
        "v": "hour"
      }, {
        "n": "day",
        "v": "day"
      }, {
        "n": "week",
        "v": "week"
      }, {
        "n": "month",
        "v": "month"
      }, {
        "n": "dayHour",
        "v": "dayHour"
      }, {
        "n": "weekDay",
        "v": "weekDay"
      }];
    };
    App.Analytics.timeUnit.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Camera = App.Camera || {};
    App.Camera.sourceType = {
      camera: "camera",
      photoLibrary: "photolibrary",
      savedPhotoAlbum: "savedphotoalbum"
    };
    App.Camera.sourceType.getValues = function() {
      return [{
        "n": "camera",
        "v": "camera"
      }, {
        "n": "photo library",
        "v": "photolibrary"
      }, {
        "n": "saved Photo Album",
        "v": "savedphotoalbum"
      }];
    };
    App.Camera.sourceType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Database = App.Database || {};
    App.Database.DataTypes = {
      INT: "i",
      FLOAT: "f",
      DECIMAL: "d",
      VARCHAR: "vc",
      CHAR: "c",
      BLOB: "bl",
      CLOB: "cl",
      BOOL: "b",
      DATE: "da",
      TIME: "ti",
      DATETIME: "dt",
      UUID: "id",
      AUTOINCR: "ai",
      JSON: "j"
    };
    App.Database.DataTypes.getValues = function() {
      return [{
        "n": "INT",
        "v": "i"
      }, {
        "n": "FLOAT",
        "v": "f"
      }, {
        "n": "DECIMAL",
        "v": "d"
      }, {
        "n": "VARCHAR",
        "v": "vc"
      }, {
        "n": "CHAR",
        "v": "c"
      }, {
        "n": "BLOB",
        "v": "bl"
      }, {
        "n": "CLOB",
        "v": "cl"
      }, {
        "n": "BOOL",
        "v": "b"
      }, {
        "n": "DATE",
        "v": "da"
      }, {
        "n": "TIME",
        "v": "ti"
      }, {
        "n": "DATETIME",
        "v": "dt"
      }, {
        "n": "UUID",
        "v": "id"
      }, {
        "n": "AUTOINCR",
        "v": "ai"
      }, {
        "n": "JSON",
        "v": "j"
      }];
    };
    App.Database.DataTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.Database = App.Database || {};
    App.Database.DriverTypes = {
      WebSQL: "WebSQL",
      Native: "Native",
      None: "None"
    };
    App.Database.DriverTypes.getValues = function() {
      return [{
        "n": "WebSQL",
        "v": "WebSQL"
      }, {
        "n": "Native",
        "v": "Native"
      }, {
        "n": "None",
        "v": "None"
      }];
    };
    App.Database.DriverTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDHeader = App.MDHeader || {};
    App.MDHeader.ScreenSizes = {
      large: "large",
      small: "small"
    };
    App.MDHeader.ScreenSizes.getValues = function() {
      return [{
        "n": "large",
        "v": "large"
      }, {
        "n": "small",
        "v": "small"
      }];
    };
    App.MDHeader.ScreenSizes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDContent = App.MDContent || {};
    App.MDContent.MDContentTypes = {
      main: "main",
      section: "section",
      page: "page",
      grid: "grid"
    };
    App.MDContent.MDContentTypes.getValues = function() {
      return [{
        "n": "main",
        "v": "main"
      }, {
        "n": "section",
        "v": "section"
      }, {
        "n": "page",
        "v": "page"
      }, {
        "n": "grid",
        "v": "grid"
      }];
    };
    App.MDContent.MDContentTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDCell = App.MDCell || {};
    App.MDCell.MDCellAlign = {
      top: "top",
      middle: "middle",
      bottom: "bottom",
      stretch: "stretch"
    };
    App.MDCell.MDCellAlign.getValues = function() {
      return [{
        "n": "top",
        "v": "top"
      }, {
        "n": "middle",
        "v": "middle"
      }, {
        "n": "bottom",
        "v": "bottom"
      }, {
        "n": "stretch",
        "v": "stretch"
      }];
    };
    App.MDCell.MDCellAlign.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDFooter = App.MDFooter || {};
    App.MDFooter.MDFooterType = {
      mini: "mini",
      mega: "mega"
    };
    App.MDFooter.MDFooterType.getValues = function() {
      return [{
        "n": "mini",
        "v": "mini"
      }, {
        "n": "mega",
        "v": "mega"
      }];
    };
    App.MDFooter.MDFooterType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDFooterSection = App.MDFooterSection || {};
    App.MDFooterSection.MDSectionType = {
      left: "left",
      right: "right",
      top: "top",
      middle: "middle",
      bottom: "bottom",
      dropdown: "drop-down"
    };
    App.MDFooterSection.MDSectionType.getValues = function() {
      return [{
        "n": "left",
        "v": "left"
      }, {
        "n": "right",
        "v": "right"
      }, {
        "n": "top",
        "v": "top"
      }, {
        "n": "middle",
        "v": "middle"
      }, {
        "n": "bottom",
        "v": "bottom"
      }, {
        "n": "dropdown",
        "v": "drop-down"
      }];
    };
    App.MDFooterSection.MDSectionType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDTableSection = App.MDTableSection || {};
    App.MDTableSection.MDTableSecTypes = {
      head: "head",
      body: "body",
      foot: "foot"
    };
    App.MDTableSection.MDTableSecTypes.getValues = function() {
      return [{
        "n": "head",
        "v": "head"
      }, {
        "n": "body",
        "v": "body"
      }, {
        "n": "foot",
        "v": "foot"
      }];
    };
    App.MDTableSection.MDTableSecTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDText = App.MDText || {};
    App.MDText.MDTextTypes = {
      title: "title",
      spacer: "spacer",
      icon: "icon",
      logo: "logo",
      heading: "heading",
      span: "span",
      h1: "h1",
      h2: "h2",
      h3: "h3",
      h4: "h4",
      h5: "h5",
      h6: "h6"
    };
    App.MDText.MDTextTypes.getValues = function() {
      return [{
        "n": "Title of the application.\n\nIt must be used inside the first @-MDL.MDHeaderRow-@ element.",
        "v": "title"
      }, {
        "n": "Allows positioning of subsequent elements on the right side.",
        "v": "spacer"
      }, {
        "n": "Icon.\n@--\nTo select the icon, its name must be written in the innerText property. \n\nYou can use an icon from the set [Material Design Icons](https://www.google.com/design/icons/). \n\nIf the name of the icon contains spaces, they must be replaced with underscore characters.",
        "v": "icon"
      }, {
        "n": "It must be used in the first @-MDText-@ element of an @-MDL.MDFooterSection-@ of the *left* type, and in this case represents the title.",
        "v": "logo"
      }, {
        "n": "It must be used in the first text object of an @-MDL.MDFooterSection-@ of the *dropdown* type, and in this case represents the title.",
        "v": "heading"
      }, {
        "n": "Represents a *span* object that contains a text in the Material Design visual style.",
        "v": "span"
      }, {
        "n": "Represents an *h1* object that contains a text in the Material Design visual style.",
        "v": "h1"
      }, {
        "n": "Represents an *h2* object that contains a text in the Material Design visual style.",
        "v": "h2"
      }, {
        "n": "Represents an *h3* object that contains a text in the Material Design visual style.",
        "v": "h3"
      }, {
        "n": "Represents an *h4* object that contains a text in the Material Design visual style.",
        "v": "h4"
      }, {
        "n": "Represents an *h5* object that contains a text in the Material Design visual style.",
        "v": "h5"
      }, {
        "n": "Represents an *h6* object that contains a text in the Material Design visual style.",
        "v": "h6"
      }];
    };
    App.MDText.MDTextTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDButton = App.MDButton || {};
    App.MDButton.MDButtonTypes = {
      flat: "flat",
      icon: "icon",
      raised: "raised",
      fab: "fab",
      miniFab: "mini-fab"
    };
    App.MDButton.MDButtonTypes.getValues = function() {
      return [{
        "n": "flat",
        "v": "flat"
      }, {
        "n": "icon",
        "v": "icon"
      }, {
        "n": "raised",
        "v": "raised"
      }, {
        "n": "fab",
        "v": "fab"
      }, {
        "n": "miniFab",
        "v": "mini-fab"
      }];
    };
    App.MDButton.MDButtonTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDToggle = App.MDToggle || {};
    App.MDToggle.MDToggleTypes = {
      checkbox: "checkbox",
      radio: "radio",
      iconToggle: "icon-toggle",
      switch: "switch"
    };
    App.MDToggle.MDToggleTypes.getValues = function() {
      return [{
        "n": "checkbox",
        "v": "checkbox"
      }, {
        "n": "radio",
        "v": "radio"
      }, {
        "n": "iconToggle",
        "v": "icon-toggle"
      }, {
        "n": "switch",
        "v": "switch"
      }];
    };
    App.MDToggle.MDToggleTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDCardSection = App.MDCardSection || {};
    App.MDCardSection.MDCardSectionTypes = {
      title: "title",
      image: "image",
      text: "supporting-text",
      actions: "actions",
      menu: "menu"
    };
    App.MDCardSection.MDCardSectionTypes.getValues = function() {
      return [{
        "n": "title",
        "v": "title"
      }, {
        "n": "image",
        "v": "image"
      }, {
        "n": "text",
        "v": "supporting-text"
      }, {
        "n": "actions",
        "v": "actions"
      }, {
        "n": "menu",
        "v": "menu"
      }];
    };
    App.MDCardSection.MDCardSectionTypes.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDMenu = App.MDMenu || {};
    App.MDMenu.MDMenuPosition = {
      topLeft: "top-left",
      topRight: "top-right",
      bottomLeft: "bottom-left",
      bottomRight: "bottom-right"
    };
    App.MDMenu.MDMenuPosition.getValues = function() {
      return [{
        "n": "topLeft",
        "v": "top-left"
      }, {
        "n": "topRight",
        "v": "top-right"
      }, {
        "n": "bottomLeft",
        "v": "bottom-left"
      }, {
        "n": "bottomRight",
        "v": "bottom-right"
      }];
    };
    App.MDMenu.MDMenuPosition.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDListContent = App.MDListContent || {};
    App.MDListContent.MDListContentType = {
      primary: "primary",
      secondary: "secondary",
      info: "info",
      text: "text"
    };
    App.MDListContent.MDListContentType.getValues = function() {
      return [{
        "n": "Primary content.",
        "v": "primary"
      }, {
        "n": "Secondary content, shown on the right.",
        "v": "secondary"
      }, {
        "n": "Informative content.",
        "v": "info"
      }, {
        "n": "Text description.",
        "v": "text"
      }];
    };
    App.MDListContent.MDListContentType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.MDListIcon = App.MDListIcon || {};
    App.MDListIcon.MDListIconType = {
      icon: "icon",
      avatar: "avatar"
    };
    App.MDListIcon.MDListIconType.getValues = function() {
      return [{
        "n": "icon",
        "v": "icon"
      }, {
        "n": "avatar",
        "v": "avatar"
      }];
    };
    App.MDListIcon.MDListIconType.decodeValue = function(val) {
      return this.getValues().find(function(el) {
        return el.v === val;
      });
    };
    App.resMap = {};
    App.resMap["logo-trasp-2_png"] = "../resources/46bc18ee-0a36-4fc8-a255-b3b6ffb37ffa.png";
    App.resMap["rolling_gif"] = "../resources/565288bb-000b-4228-8936-497e13e877db.gif";
    App.resMap["note-venduto_png"] = "../resources/1413e40b-6128-4225-8154-133d8acd7965.png";
    App.resMap["note-catalogo_png"] = "../resources/5804edfc-83da-4175-ba1b-2c4721f18b5a.png";
    App.resMap["note-listino_png"] = "../resources/88c22918-0ebd-489d-9625-e94b2ceba233.png";
    App.resMap["tasto-note_png"] = "../resources/860def55-7421-4352-ac6c-7a339fe75b87.png";
    App.resMap["logo-splash"] = "../resources/0a46be8e-f549-453e-8af0-ffab3113316c.png";
    App.resMap["StyleBook-en"] = "../resources/f2b988ca-9540-4479-bae1-c2aad117c585.xml";
    App.resMap["it.js"] = "../resources/cad6b13d-4497-4784-9861-a47098b03a3a.js";
  };
  App.databaseList = [];
  App.databaseList.push('sbDB');
  App.info = {
    "buildVersion": "8614566e-a2f2-c423-3100-4c0e67f34b0e",
    "buildName": "sb2017-4",
    "date": "2017-12-07T17:23:58.257Z",
    "id": "iMU0dhHd2H+ihDBIwaUrtA=="
  };
  App.i18m = {};
  //# sourceURL=StyleBook.js